/*
 * Copyright 2007-2008 SpringSource Inc.
 */

package com.springsource.json.parser;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;

import org.antlr.runtime.ANTLRInputStream;
import org.antlr.runtime.ANTLRStringStream;
import org.antlr.runtime.CharStream;
import org.antlr.runtime.CommonTokenStream;
import org.antlr.runtime.RecognitionException;
import org.antlr.runtime.tree.CommonTreeNodeStream;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.springsource.json.parser.generated.JsonFormatParser.objects_return;

/**
 * This implementation of the JSONParser interface uses an Antlr generated parser to process JSON formatted data, either
 * from a URL or directly from a String. It performs two passes, the first is a regular parse to create an Ast then a
 * tree walker is used to restructure that as a more useful tree of Node instances. The root node of the tree is
 * returned from either parse method. Unless an instance of JSONParserLogger is supplied during construction, the parser
 * will use an instance of StandardJSONParserLogger which reports errors/warnings/info to System.out.
 * 
 * This class is thread safe.
 * 
 * @author Andy Clement
 * @author Rob Harrop
 * @since 1.0
 */
public final class AntlrJSONParser implements JSONParser {

    private static final Log LOG = LogFactory.getLog(AntlrJSONParser.class);

    private final JSONParserLogger parserLogger;

    /**
     * Construct an AntlrJsonParser that will use the StandardJsonParserLogger to report info/warn/error messages.
     */
    public AntlrJSONParser() {
        this.parserLogger = new StandardJSONParserLogger();
    }

    /**
     * Construct an AntlrJsonParser that will use the supplied logger to report info/warn/error messages.
     */
    public AntlrJSONParser(JSONParserLogger logger) {
        this.parserLogger = logger;
    }

    /**
     * @inheritDoc
     */
    public Node parse(URL jsonDataURL) throws JSONParseException {
        try {
            return parse(jsonDataURL.openStream(), jsonDataURL.toString());
        } catch (IOException e) {
            throw new JSONParseException("Unable to access JSON data at URL '" + jsonDataURL + "'", e);
        }
    }

    /**
     * @inheritDoc
     */
    public Node parse(File file) throws JSONParseException {
        try {
            return parse(new FileInputStream(file), file.getAbsolutePath());
        } catch (FileNotFoundException e) {
            throw new JSONParseException("File '" + file.getAbsolutePath() + "' does not exist.");
        }
    }

    private Node parse(InputStream inputStream, String location) throws JSONParseException {
        ANTLRInputStream stream = null;
        try {
            stream = new ANTLRInputStream(inputStream);
        } catch (IOException ioe) {
            throw new JSONParseException("Unable to create ANTLR input stream", ioe);
        }

        try {
            Node n = parseInternal(stream);
            return n;
        } catch (RecognitionException re) {
            throw new JSONParseException("JSON format structure at '" + location + "' failed to parse at line " + re.line + ", character "
                + re.charPositionInLine + ". See message log for further information.", re);
        } catch (InternalParseException fpe) {
            // A InternalParseException can be thrown in case of lexing issues - so here we throw a new exception with
            // the right context in it. Not performant but this is the error case
            int line = 0;
            int col = 0;
            Throwable cause = fpe.getCause();
            if (cause instanceof RecognitionException) {
                RecognitionException re = (RecognitionException) cause;
                line = re.line;
                col = re.charPositionInLine;
            }
            throw new JSONParseException("JSON format structure at '" + location + "' failed to parse at line " + line + ", character " + col
                + ". See message log for further information.", cause);
        } finally {
            if (inputStream != null) {
                try {
                    inputStream.close();
                } catch (IOException e) {
                    if (LOG.isWarnEnabled()) {
                        LOG.warn("Error closing stream '" + inputStream + "'.", e);
                    }
                }
            }
        }
    }

    /**
     * @inheritDoc
     */
    public Node parse(String jsonData) throws JSONParseException {
        try {
            Node n = parseInternal(new ANTLRStringStream(jsonData));
            return n;
        } catch (RecognitionException re) {
            // re.printStackTrace();
            throw new JSONParseException("JSON format data '" + jsonData + "' failed to parse at line " + re.line + ", character "
                + re.charPositionInLine + ". See message log for further information.", re);
        } catch (InternalParseException fpe) {
            // A InternalParseException can be thrown in case of lexing issues - so here we throw a new one exception
            // with the right
            // context in it. Not performant but this is the error case
            int line = 0;
            int col = 0;
            Throwable cause = fpe.getCause();
            if (cause instanceof RecognitionException) {
                RecognitionException re = (RecognitionException) cause;
                line = re.line;
                col = re.charPositionInLine;
            }
            throw new JSONParseException("JSON format data '" + jsonData + "' failed to parse at line " + line + ", character " + col
                + ". See message log for further information.", cause);
        }
    }

    /**
     * Internal parse method that just wants an inputStream, which can be from a URL or simple String data. This method
     * will do the regular parse of the data and then the tree walk to produce the final tree of Node objects.
     * 
     * @param inputStream an Antlr input stream of data to be parsed
     * @return the root Node in the parse tree
     * @throws RecognitionException thrown if any problems occur with either pass over the data
     */
    private Node parseInternal(CharStream inputStream) throws RecognitionException {
        LoggedJsonFormatLexer lexer = new LoggedJsonFormatLexer(parserLogger, inputStream);
        CommonTokenStream tokenStream = new CommonTokenStream(lexer);
        LoggedJsonFormatParser parser = new LoggedJsonFormatParser(parserLogger, tokenStream);

        objects_return result = null;
        try {
            result = parser.objects();
        } catch(RecognitionException ex) {
            parser.logError(ex, LoggedJsonFormatParser.tokenNames);
            throw ex;
        } catch(InternalParseException ex) {
            Throwable cause = ex.getCause();
            if(cause instanceof RecognitionException) {
                RecognitionException re = (RecognitionException) cause;
                //parser.logError(re, LoggedJsonFormatParser.tokenNames);
            }
            throw ex;
        }
        // printTree((Tree) result.getTree());
        CommonTreeNodeStream treeNodeStream = new CommonTreeNodeStream(result.getTree());
        LoggedJsonFormatTreeWalker walker = new LoggedJsonFormatTreeWalker(parserLogger, treeNodeStream);
        walker.objects();

        return walker.getRootNode();
    }

    // DEBUG code
    // private void printTree(Tree t) {
    // printIndentedTree(t, "");
    // }
    //
    // private void printIndentedTree(Tree t, String indent) {
    // System.out.print(indent);
    // System.out.println(t.toString());
    // if (t.getChildCount() > 0) {
    // for (int c = 0; c < t.getChildCount(); c++) {
    // printIndentedTree(t.getChild(c), indent + " ");
    // }
    // }
    // }
}
