/*
 * Copyright 2007-2008 SpringSource Inc.
 */

package com.springsource.json.parser;

import java.util.List;

import org.antlr.runtime.CharStream;
import org.antlr.runtime.MismatchedRangeException;
import org.antlr.runtime.MismatchedTokenException;
import org.antlr.runtime.NoViableAltException;
import org.antlr.runtime.RecognitionException;

import com.springsource.json.parser.InternalParseException;
import com.springsource.json.parser.JSONParserLogger;
import com.springsource.json.parser.generated.JsonFormatLexer;

/**
 * The LoggedJsonFormatLexer extends the Antlr generated JsonFormatLexer and allows a logger to be plugged in. This
 * logger will get all info/warn/error events that occur during parser processing. If no parser logger is plugged in, a
 * StandardJsonParserLogger instance will be created and used, which reports events to System.out.
 * <p />
 * 
 * This class is thread safe.
 * 
 * @author Andy Clement
 * @since 1.0
 */
class LoggedJsonFormatLexer extends JsonFormatLexer {

    private final JSONParserLogger logger;

    public LoggedJsonFormatLexer(JSONParserLogger logger, CharStream input) {
        super(input);
        this.logger = logger;
    }

    /**
     * recover() attempts to provide better error messages once something has gone wrong. It then throws a
     * JSONParseException (has to be this unchecked exception as the exception must flow through Antlr lexer methods
     * that do not have declared exceptions). The JSONParseException will be caught at the top level and altered to
     * include context (line,column) information before being rethrown.<br>
     * 
     * This error analysis code is in recover() rather than reportError() because reportError() isn't always called by
     * the lexer and there is no way to add the calls to it by editing the .g file.
     */
    @Override
    public void recover(RecognitionException re) {
        List<?> rules = getRuleInvocationStack(re, JsonFormatLexer.class.getName());
        String failedRule = (String) rules.get(rules.size() - 1);
        // System.out.println("DBG: lexer rule " + failedRule);
        if (re instanceof NoViableAltException) {
            NoViableAltException nvae = (NoViableAltException) re;
            // example error data: { "abc": def }
            if (failedRule.equals("mTokens") && Character.isLetter((char) (nvae.getUnexpectedType()))) {
                logger.error(ParserMessage.ERROR_STRINGS_MUST_BE_QUOTED, re.line, re.charPositionInLine);
            }

        } else if (re instanceof MismatchedRangeException) {
            // MismatchedRangeException mre = (MismatchedRangeException) re;
            // example error data: [ 123e ]
            if (failedRule.equals("mDIGIT") && rules.size() > 3 && ((String) rules.get(rules.size() - 3)).equals("mExponent")) {
                logger.error(ParserMessage.ERROR_INVALID_EXPONENT, re.line, re.charPositionInLine);
            }
        } else if (re instanceof MismatchedTokenException) {
            MismatchedTokenException mte = (MismatchedTokenException) re;
            logger.error(ParserMessage.ERROR_MISMATCHED_CHARACTER, mte.charPositionInLine, mte.charPositionInLine, getCharErrorDisplay(mte.expecting), getCharErrorDisplay(mte.c));
        }

        throw new InternalParseException(re);
    }

    
    @Override
    public void reportError(RecognitionException re) {
        // Do not report anything. If better messages could be reported they will have been reported
        // by the recover() method above.
    }

}
