/*
 * Copyright 2007-2008 SpringSource Inc.
 */

package com.springsource.json.parser;

import org.antlr.runtime.BitSet;
import org.antlr.runtime.CommonTokenStream;
import org.antlr.runtime.EarlyExitException;
import org.antlr.runtime.FailedPredicateException;
import org.antlr.runtime.IntStream;
import org.antlr.runtime.MismatchedNotSetException;
import org.antlr.runtime.MismatchedSetException;
import org.antlr.runtime.MismatchedTokenException;
import org.antlr.runtime.MismatchedTreeNodeException;
import org.antlr.runtime.NoViableAltException;
import org.antlr.runtime.RecognitionException;
import org.antlr.runtime.Token;
import org.antlr.runtime.TokenStream;

import com.springsource.json.parser.generated.JsonFormatParser;

/**
 * The LoggedJsonFormatParser extends the Antlr generated JsonFormatParser and allows a logger to be plugged in. This
 * logger will get all info/warn/error events that occur during parser processing. If no parser logger is plugged in, a
 * StandardJsonParserLogger instance will be created and used, which reports events to System.out.
 * <p />
 * 
 * This class is thread safe.
 * 
 * @author Andy Clement
 * @since 1.0
 */
class LoggedJsonFormatParser extends JsonFormatParser {

    private final JSONParserLogger logger;

    public LoggedJsonFormatParser(JSONParserLogger logger, TokenStream input) {
        super(input);
        this.logger = logger;
    }

    public String getTokenForId(int id) {
        if (id == -1)
            return "EOF";
        return getTokenNames()[id];
    }

    /**
     * This method does not actually recover but attempts to produce better error messages than a generic Antlr parser
     * would do - by utilising knowledge of the JSON format.
     * 
     */
    @Override
    public void recoverFromMismatchedToken(IntStream input, RecognitionException re, int ttype, BitSet follow) throws RecognitionException {
        CommonTokenStream tokStream = (CommonTokenStream) input;
        int prevToken = tokStream.LA(-1);
        int nextToken = tokStream.LA(1);
        String prevTokenText = tokStream.LT(-1).getText();
        String expectedToken = getTokenForId(ttype);

        if (prevToken == SimpleNumber && nextToken == SimpleNumber && prevTokenText.equals("0")) {
            logger.error(ParserMessage.ERROR_NO_LEADING_ZERO, re.line, re.charPositionInLine);
        } else if (ttype == EOF) {
            logger.error(ParserMessage.ERROR_EXTRANEOUS_DATA_FOUND, re.line, re.charPositionInLine);
        } else if (nextToken == EOF) {
            logger.error(ParserMessage.ERROR_UNEXPECTED_END_OF_DATA, re.line, re.charPositionInLine, expectedToken);
        }
        throw re;
    }

    /**
     * Similar to the BaseRecognizer getErrorMessage() but uses the ParserMessages class to get the text of the message
     */
    public void logError(RecognitionException e, String[] tokenNames) {
 
        if (e instanceof MismatchedTokenException) {
            MismatchedTokenException mte = (MismatchedTokenException) e;
            String tokenName = "<unknown>";
            if (mte.expecting == Token.EOF) {
                tokenName = "EOF";
            } else {
                tokenName = tokenNames[mte.expecting];
            }
            logger.error(ParserMessage.ERROR_MISMATCHED_TOKEN,e.line, e.charPositionInLine, getTokenErrorDisplay(e.token), tokenName);
        } else if (e instanceof MismatchedTreeNodeException) {
            MismatchedTreeNodeException mtne = (MismatchedTreeNodeException) e;
            String tokenName = "<unknown>";
            if (mtne.expecting == Token.EOF) {
                tokenName = "EOF";
            } else {
                tokenName = tokenNames[mtne.expecting];
            }
            logger.error(ParserMessage.ERROR_MISMATCHED_TREE_NODE, e.line, e.charPositionInLine, mtne.node.toString(), tokenName);
            // msg = "mismatched tree node: " + mtne.node + " expecting " + tokenName;
        } else if (e instanceof NoViableAltException) {
            NoViableAltException nvae = (NoViableAltException) e;
            // for development, can add "decision=<<"+nvae.grammarDecisionDescription+">>"
            // and "(decision="+nvae.decisionNumber+") and
            // "state "+nvae.stateNumber
            logger.error(ParserMessage.ERROR_NO_VIABLE_ALTERNATIVE,nvae.line, nvae.charPositionInLine, getTokenErrorDisplay(e.token));
            // msg = "no viable alternative at input " + getTokenErrorDisplay(e.token);
        } else if (e instanceof EarlyExitException) {
            // for development, can add "(decision="+eee.decisionNumber+")"
            logger.error(ParserMessage.ERROR_EARLY_EXIT, e.line, e.charPositionInLine, getTokenErrorDisplay(e.token));
        } else if (e instanceof MismatchedSetException) {
            MismatchedSetException mse = (MismatchedSetException) e;
            logger.error(ParserMessage.ERROR_MISMATCHED_INPUT, e.line, e.charPositionInLine, getTokenErrorDisplay(e.token), mse.expecting.toString());
        } else if (e instanceof MismatchedNotSetException) {
            MismatchedNotSetException mse = (MismatchedNotSetException) e;
            logger.error(ParserMessage.ERROR_MISMATCHED_INPUT, e.line, e.charPositionInLine, getTokenErrorDisplay(e.token), mse.expecting.toString());
        } else if (e instanceof FailedPredicateException) {
            FailedPredicateException fpe = (FailedPredicateException) e;
            logger.error(ParserMessage.ERROR_FAILED_PREDICATE, e.line,e.charPositionInLine, fpe.ruleName, fpe.predicateText);
            // TODO: msg = "rule " + fpe.ruleName + " failed predicate: {" + fpe.predicateText + "}?";
        }
    }

    @Override
    public void recoverFromMismatchedSet(IntStream input, RecognitionException e, BitSet follow) throws RecognitionException {
        throw e;
    }

    @Override
    public String getTokenErrorDisplay(Token t) {
        if(t == null) {
            return "<unknown>";
        }
        return super.getTokenErrorDisplay(t);
    }

    @Override
    public void displayRecognitionError(String[] tokenNames, RecognitionException re) {
        logger.error(ParserMessage.ERROR_RECOGNITION, re.line, re.charPositionInLine, re);
    }
}
