/*
 * Copyright 2007-2008 SpringSource Inc.
 */

package com.springsource.json.parser;

import java.util.List;


/**
 * A Node is a node in a tree representing a JSON data structure and may be either a scalar node (boolean, string,
 * number, null), a map node, or a list node. A scalar node contains a single value. A map node contains a map of key to
 * node. A list node contains a list of nodes.
 * 
 * Implementations of this interface must be thread safe.
 * 
 * @author Glyn Normington
 * @author Andy Clement
 * @since 1.0
 */
public interface Node {

    /**
     * Get the node of the tree identified by the specified path from this node. Each element of the path is a key which
     * is used to navigate to the appropriate child node.
     * 
     * @param path a list of keys
     * @return the node identified by the path or null if there is no such node
     */
    Node getNode(List<String> path);

    /**
     * Add the given node at the given path in the tree.
     * 
     * If there is already a node at the given path, do not modify the tree.
     * 
     * @param path a list of keys identifying where to add the new node
     * @param newNode the node to be added
     * @return true if and only if the node was added successfully
     */
    boolean addNode(List<String> path, Node newNode);

    /**
     * Return the path of this node including the config. point.
     * 
     * @return a string path identifying this node in the whole config. tree
     */
    String getPath();

    /**
     * Return the path of the given node including the config. point. The given node must be a direct child of this
     * node.
     * 
     * @return a string path identifying the given node in the whole config. tree
     */
    String getPath(Node childNode);

    /**
     * Return the parent node of this node or null if this node is a config. point.
     * 
     * @return the parent ConfigNode
     */
    Node getParent();

}
