
package com.springsource.json.parser;

import java.text.MessageFormat;

import com.springsource.json.parser.ParserMessage;

/**
 * Encapsulates all messages that can come from the parser. Some of the error messages include inserts, denoted by {n}
 * in their text. When an instance of the error is to be reported, the context of the error (line/column) and relevant
 * inserts are supplied to the format() method which returns a fully formatted error string.<br>
 * 
 * Encapsulating the messages here will make I18N support easier. Note: some of the tests check parts of these error
 * messages.
 * 
 * This class is thread safe.
 * 
 * @author Andy Clement
 * @author Rob Harrop
 * @since 1.1
 */
public enum ParserMessage {

    ERROR_NO_LEADING_ZERO(0, "Numbers should never have a leading zero"),

    ERROR_STRINGS_MUST_BE_QUOTED(1, "Strings must be quoted"),

    ERROR_INVALID_EXPONENT(2, "Invalid exponent"),

    ERROR_EXTRANEOUS_DATA_FOUND(3, "Extraneous data found at end of input"),

    ERROR_UNEXPECTED_END_OF_DATA(4, "Unexpectedly run out of data, expecting to find {0}"),

    ERROR_MISMATCHED_TOKEN(5, "Mismatched input {0} expecting {1}"),

    ERROR_MISMATCHED_TREE_NODE(6, "Mismatched tree node: {0} expecting {1}"),

    ERROR_NO_VIABLE_ALTERNATIVE(7, "No viable alternative at input {0}"),

    ERROR_MISMATCHED_INPUT(8, "Mismatched input {0} expecting set {1}"),

    ERROR_FAILED_PREDICATE(9, "Rule '{0}' failed predicate: '{1}'"),

    ERROR_EARLY_EXIT(10, "Required (...)+ loop did not match anything at input {0}"),
    
    ERROR_MISMATCHED_CHARACTER(11, "Mismatched character. Expecting {0}, got {1}"),

    ERROR_RECOGNITION(100, "Recognition error");

    private final String defaultMessage;

    private final int code;

    private ParserMessage(int code, String defaultMessage) {
        this.code = code;
        this.defaultMessage = defaultMessage;
    }

    public String getDefaultMessage() {
        return defaultMessage;
    }

    public int getCode() {
        return code;
    }

    /**
     * Returns a string containing a message header that includes line and column information, then the message body
     * where the appropriate inserts have been applied.
     * 
     * @param line the line number in the input Json structure where the problem occurred
     * @param col the column number in the input Json structure where the problem occurred
     * @param inserts any inserts to be placed into the message text
     * @return formatted string
     */
    public String format(int line, int col, Object... inserts) {
        StringBuilder errorText = new StringBuilder();
        errorText.append("Parser Error (line ").append(line).append(", col ").append(col).append("): ");
        errorText.append(MessageFormat.format(this.defaultMessage, inserts));
        return errorText.toString();
    }
}
