/*
 * Copyright 2007-2008 SpringSource Inc.
 */

package com.springsource.json.parser.internal;

import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.springsource.json.parser.InternalParseException;
import com.springsource.json.parser.MapNode;
import com.springsource.json.parser.Node;
import com.springsource.json.parser.internal.AbstractNode;
import com.springsource.json.parser.internal.StandardMapNode;

/**
 * Implementation of the MapNode. A MapNode is represented as a java.util.Map of key/value entries, where the key is
 * always a String, and the value may be any node kind.
 * 
 * this class is thread safe.
 * 
 * @author Andy Clement
 * @since 1.0
 */
public final class StandardMapNode extends AbstractNode implements MapNode {

    private final Map<String, Node> map = Collections.synchronizedMap(new LinkedHashMap<String, Node>());

    public StandardMapNode(Node parent) {
        super(parent);
    }

    public Node getNode(String key) {
        return this.map.get(key);
    }

    public void setNode(String key, Node value) {
        if (value.getParent() != this) {
            throw new InternalParseException("Cannot set node of a map node which is not a child of the map node", null);
        }

        if (key.charAt(0) == '"')
            key = key.substring(1, key.length() - 1);
        this.map.put(key, value);
    }

    public Node getNode(List<String> path) {
        Node result = null;
        if (!path.isEmpty()) {
            String head = head(path);
            if (this.map.containsKey(head)) {
                result = this.map.get(head).getNode(tail(path));
            }
        } else {
            result = this;
        }
        return result;
    }

    public boolean addNode(List<String> path, Node newNode) {
        boolean result = false;
        if (!path.isEmpty()) {
            String head = head(path);
            List<String> tail = tail(path);
            if (tail.isEmpty()) {
                if (!this.map.containsKey(head)) {
                    this.map.put(head, newNode);
                    result = true;
                }
            } else {
                if (!this.map.containsKey(head)) {
                    this.map.put(head, new StandardMapNode(this));
                }
                if (this.map.get(head) instanceof MapNode) {
                    MapNode child = (MapNode) this.map.get(head);
                    result = child.addNode(tail, newNode);
                }
            }
        }
        return result;
    }

    @Override
    public String toString() {
        StringBuffer result = new StringBuffer();
        result.append('{');
        boolean first = true;
        for (String k : this.map.keySet()) {
            if (first == false) {
                result.append(", ");
            }
            result.append(k);
            result.append(" : ");
            result.append(this.map.get(k).toString());
            first = false;
        }
        result.append('}');
        return result.toString();
    }

    public Set<String> getKeys() {
        return Collections.unmodifiableSet(this.map.keySet());
    }

    /**
     * @inheritDoc
     */
    @Override
    protected String doGetPath(Node childNode) {
        StringBuffer path = new StringBuffer();
        path.append(getPath());
        if (getParent() != null) {// && getParent().getParent() != null) {
            path.append(',');
        }
        for (String k : this.map.keySet()) {
            if (this.map.get(k) == childNode) {
                path.append(k);
                return path.toString();
            }
        }
        throw new InternalParseException("Child node not found in parent map", null);
    }
}
