/*
 * Copyright 2007-2008 SpringSource Inc.
 */

package com.springsource.json.validation;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.regex.Pattern;

import com.springsource.json.parser.ListNode;
import com.springsource.json.parser.MapNode;
import com.springsource.json.parser.Node;
import com.springsource.json.parser.ScalarNode;
import com.springsource.json.parser.internal.StandardBooleanNode;
import com.springsource.json.parser.internal.StandardListNode;
import com.springsource.json.parser.internal.StandardMapNode;
import com.springsource.json.parser.internal.StandardNullNode;
import com.springsource.json.parser.internal.StandardNumberNode;
import com.springsource.json.parser.internal.StandardStringNode;

/**
 * The JSONValidator will validate a JSON formatted structure against a JSON schema that defines the expected format of
 * the structure.
 * 
 * <p>
 * The spec followed is here: <a
 * href="http://groups.google.com/group/json-schema/web/json-schema-proposal---second-draft">
 * http://groups.google.com/group/json-schema/web/json-schema-proposal---second-draft</a>
 * 
 * <h4>Using the JSON Validator</h4>
 * 
 * <p>
 * In order to use the JSON Validator a schema must be defined against which JSON data can be validated. There is a
 * schema-schema which can be used to verify that any defined schema is well formed. Any schema is well formed JSON and
 * so the same parsing functions can be used to parse the schema as parse the data. Here is a simple example:<br>
 * <i><b>NOTE</b>: single quotes are used as the delimiters within the expressions to keep them looking pleasant in
 * the example. In a real program they would need to be double quotes (escaped double quotes if specified within a
 * string literal like this)</i>
 * 
 * <code><pre>
 * JSONParser parser = new AntlrJSONParser();
 * 
 * Node person = parser.parse(&quot;{'name':'Derek','age':16}&quot;);
 * 
 * MapNode personSchema = (MapNode) parser.parse(&quot;{'id':'person','type':'object','properties':{'name':{'type':'string'},'age':{'type':'integer'}}}&quot;);
 * 
 * JSONValidator personValidator = new JSONValidator(personSchema);
 * 
 * personValidator.validate(person);
 * 
 * if (personValidator.foundProblems()) {
 *     List&lt;JSONValidationProblem&gt; problems = personValidator.getProblems();
 * }
 * </pre></code>
 * 
 * <p>
 * The comments at the start of JSONSchemaTests discuss what the various Schema Attributes mean. Within that test suite
 * there are examples of using all the attributes too. Giving the schema an id is optional, it is used when referred to
 * from other schemas:
 * 
 * <pre><code>
 * MapNode marriedPersonSchema = parseSchema(&quot;{'id':'marriedperson','extends':{'$ref':'person'},&quot;
 *     + &quot; 'properties':{'age':{'type':'integer','minimum':18}}}&quot;);
 * </pre></code>
 * <p>
 * Here the marriedPerson extends the person schema (through extends $ref=person) and specifies a further constraint
 * that the age must be at least 18. How do we ensure validators know about various previously defined schemas? A very
 * simple SchemaRegistry maps ids to the root of the schema.
 * 
 * <pre><code>
 * SchemaRegistry simpleRegistry = new SchemaRegistry();
 * 
 * simpleRegistry.register(personSchema);
 * 
 * JSONValidator marriedPersonValidator = new JSONValidator(marriedPersonSchema);
 * 
 * marriedPersonValidator.setSchemaRegistry(simpleRegistry);
 * 
 * marriedPersonValidator.validate(person);
 * </code></pre>
 * 
 * <p>
 * There are now problems reported because Derek is too young:
 * 
 * <pre><code>
 * if (marriedPersonValidator.foundProblems()) {
 * 		List&lt;JSONValidationProblem&gt; problems = marriedPersonValidator.getProblems();
 * 		System.out.println(problems);
 * }
 * 
 * Error 14: 'age': value of '16' specified for the property is lower than minimum allowed '18'
 * </code></pre>
 * 
 * <p>
 * Validation errors include the error message (one of those listed in ValidationMessages.java), the path to the
 * problematic node ('age') and explanation of the violation.
 * </p>
 * 
 * <h4>Validating schemas</h4>
 * 
 * <p>
 * Every schema can be validated against the JSON schema schema to check it is well formed:
 * 
 * <pre><code>
 * JSONParser parser = new AntlrJSONParser();
 * 
 * MapNode schemaschema = (MapNode) parser.parse(JSONValidator.getSchemaSchema());
 * 
 * MapNode personSchema = (MapNode) parser.parse(&quot;{'id':'person','type':'object','properties':{'name':{'type':'string'},'age':{'type':'integer'}}}&quot;);
 * 
 * JSONValidator schemaValidator = new JSONValidator(schemaschema);
 * 
 * schemaValidator.validate(personSchema);
 * </pre></code>
 * <p>
 * Any problems found in that validate() call are due to the schema itself being badly formed.
 * 
 * 
 * @author Andy Clement
 * @since 1.0
 */
public class JSONValidator {

    /* list of validation problems found during a validation run */
    private final List<JSONValidationProblem> validationProblems = new ArrayList<JSONValidationProblem>();

    /* if true then default values from the schema will be inserted into the structure being validated */
    private boolean applyDefaults;

    /* if true after a validate() call then the structure was modified (defaults were inserted) */
    private boolean wasUpdated;

    /* schema to be used when asking this JSONValidator to validate a structure */
    private final MapNode schema;

    /*
     * registered format checkers that will be invoked to check 'format':'xxx' entries in the schema - if no checker is
     * registered for a format then we assume the data is valid.
     */
    private final Map<String, FormatChecker> formatCheckers = new HashMap<String, FormatChecker>();

    /*
     * maintains a map of schema id's to schemas - for use when resolving schema references. if no registry is attached
     * to a validator instance upon creation, a default registry will be created when necessary.
     */
    private SchemaRegistry registry;

    /**
     * Creates a JSONValidator that will validate JSON structures against the specified schema.
     * 
     * @param schema the JSON schema to use to validate JSON structures on future validate calls
     */
    public JSONValidator(MapNode schema) {
        this.schema = schema;
    }

    /**
     * Validate a JSON structure, default values specified in the schema will not be added to the input structure.
     * 
     * @param jsonStructure A JSON structure as produced by the JSONParser
     */
    public void validate(Node jsonStructure) {
        validate(jsonStructure, false);
    }

    /**
     * Validate a JSON structure, default values specified in the schema will be added to the input structure.
     * 
     * @param jsonStructure A JSON structure as produced by the JSONParser
     * @param applyDefaults If true then default values in the schema will be added to the input structure during
     *        validation
     */
    public void validate(Node jsonStructure, boolean applyDefaults) {
        reset();
        this.applyDefaults = applyDefaults;
        checkValue(jsonStructure, schema);
    }

    /**
     * @return true if schema violations were found during validation
     */
    public boolean foundProblems() {
        return !validationProblems.isEmpty();
    }

    /**
     * @return list of schema violations that were found during validation
     */
    public List<JSONValidationProblem> getProblems() {
        return validationProblems;
    }

    /**
     * @return true if the input structure was updated (defaults applied) during validation
     */
    public boolean wasInputStructureUpdated() {
        return wasUpdated;
    }

    public void registerFormatChecker(VersionFormatChecker versionFormatChecker) {
        formatCheckers.put(versionFormatChecker.getFormat(), versionFormatChecker);
    }

    public void setSchemaRegistry(SchemaRegistry registry) {
        this.registry = registry;
    }

    public SchemaRegistry getSchemaRegistry() {
        if (this.registry == null) {
            // Create a default one if none has been set
            this.registry = new SchemaRegistry();
        }
        return this.registry;
    }

    // --- internals

    private void reset() {
        validationProblems.clear(); // in case this validator is used twice
        wasUpdated = false;
        alreadyFollowed.clear();
        defaultsChecked.clear();
    }

    private void checkValue(Node data, Node schema) {
        if (isList(schema)) {
            checkTupleTyping((ListNode) data, (ListNode) schema);
        } else {
            checkValueAgainstMap(data, (MapNode) schema);
        }
    }

    /*
     * For tuple typing the data is a list and the schema is a list, each list element should match each corresponding
     * schema list element
     */
    private void checkTupleTyping(Node data, ListNode schema) {
        if (!isList(data)) {
            recordError(ValidationMessages.LIST_EXPECTED, data.getPath(), getKind(data));
        }
        List<Node> dataElements = ((ListNode) data).getNodes();
        List<Node> schemaElements = schema.getNodes();
        int listIndex = 0;
        for (Node schemaListElement : schemaElements) {
            checkValue(dataElements.get(listIndex), schemaListElement);
        }
    }

    /*
     * When the schema is a map then it may define the various schema attribute constraints against which the data must
     * be checked.
     */
    private void checkValueAgainstMap(Node data, MapNode schema) {
        checkMaxDecimal(data, schema);
        checkMaxLength(data, schema);
        checkMinLength(data, schema);
        checkPattern(data, schema);
        checkMinimum(data, schema);
        checkMaximum(data, schema);
        checkOptions(data, schema);
        checkExtends(data, schema);
        checkFormat(data, schema);
        checkType(data, schema);
        // uniqueness check could go here too
        if (isList(data)) {
            ListNode dataAsList = (ListNode) data;
            checkItems(data, schema, dataAsList);
            checkMinItems(dataAsList, schema);
            checkMaxItems(dataAsList, schema);
        } else if (isMap(data)) {
            checkSpecificity((MapNode) data, schema);
            if (defines(schema, "properties") || defines(schema, "additionalProperties")) {
                Node properties = getNode(schema, "properties");
                Node additionalProperties = getNode(schema, "additionalProperties");
                checkMap((MapNode) data, (MapNode) properties, additionalProperties);
            }
        }
    }

    private void checkItems(Node data, MapNode schemaAsMap, ListNode dataAsList) {
        if (defines(schemaAsMap, "items")) {
            List<Node> nodes = dataAsList.getNodes();
            Node items = getNode(schemaAsMap, "items");
            if (isMap(items)) { // it is the schema that all items should be matched against
                for (int i = 0; i < nodes.size(); i++) {
                    checkValue(nodes.get(i), items);
                }
            } else if (isList(items)) { // each element is the schema for the corresponding data element
                List<Node> schemaList = ((ListNode) items).getNodes();
                if (nodes.size() < schemaList.size()) {
                    recordError(ValidationMessages.VALUE_LIST_TOO_SHORT, data.getPath(), nodes.size(), schemaList.size());
                }
                Node aProps = getNode(schemaAsMap, "additionalProperties");
                for (int i = 0; i < nodes.size(); i++) {
                    MapNode schemaForElementValidation = null;
                    if (i < schemaList.size()) {
                        schemaForElementValidation = (MapNode) schemaList.get(i);
                    } else {
                        // If additionalProperties allows then use it as the schema against which to validate any extra entries
                        if (aProps != null) {
                            if (isBoolean(aProps) && toBoolean(aProps) == false) {
                                // no additional properties allowed!
                                recordError(ValidationMessages.VALUE_LIST_TOO_LONG, data.getPath(), nodes.size(), schemaList.size());
                            } else {
                                schemaForElementValidation = (MapNode) aProps;
                            }
                        }
                    }
                    if (schemaForElementValidation != null) {
                        checkValue(nodes.get(i), schemaForElementValidation);
                    }
                }
            }
        }
    }

    @SuppressWarnings("unchecked")
    private void checkExtends(Node data, MapNode schemaAsMap) {
        if (defines(schemaAsMap, "extends")) {
            Node parentSchemaReference = getNode(schemaAsMap, "extends");
            String parentSchemaId = getReference(schemaAsMap, "extends");
            if (!isMap(parentSchemaReference)) {
                recordError(ValidationMessages.SCHEMA_REFERENCE_INVALID, data.getPath(), parentSchemaReference);
            } else {
                MapNode parentSchema = (MapNode) parentSchemaReference;
                MapNode parentSchemaProperties = (MapNode) getNode(parentSchema, "properties");
                Node parentSchemaAdditionalProperties = getNode(parentSchema, "additionalProperties");

                boolean violation = false;

                // Check if the parent schema forbids extra properties is this schema
                if (parentSchemaAdditionalProperties != null && isBoolean(parentSchemaAdditionalProperties)
                    && toBoolean(parentSchemaAdditionalProperties) == false) {
                    Set<String> propertyKeys = (Set<String>) ((defines(schemaAsMap, "properties") ? ((MapNode) getNode(schemaAsMap, "properties")).getKeys()
                        : Collections.emptySet()));
                    Set<String> parentPropertyKeys = (Set<String>) (parentSchemaProperties != null ? ((parentSchemaProperties).getKeys())
                        : Collections.emptySet());

                    Set<String> extraProperties = new HashSet<String>();
                    extraProperties.addAll(propertyKeys);
                    extraProperties.removeAll(parentPropertyKeys);
                    if (extraProperties.size() != 0) {
                        recordError(ValidationMessages.EXTENDING_SCHEMA_CANNOT_ADD_NEW_PROPERTIES, data.getPath(), parentSchemaId, extraProperties);
                        violation = true;
                    }
                }

                // If no violations, check the data against the parent schema
                if (!violation) {
                    checkValue(data, parentSchema);
                }
            }
        }
    }

    /**
     * Check if the data should conform to a specificity, if it does then other entries may be required in the same
     * structure (map/list).
     */
    private void checkSpecificity(MapNode data, MapNode schema) {
        Set<String> dataKeys = data.getKeys();
        MapNode properties = (MapNode) getNode(schema, "properties");
        if (properties == null)
            return;
        Set<String> schemaKeys = properties.getKeys();
        Map<Integer, String> included = new HashMap<Integer, String>();
        Map<Integer, String> excluded = new HashMap<Integer, String>();
        int maxValue = -1;
        for (String schemaKey : schemaKeys) {
            Node n = getNode(properties, schemaKey);
            if (defines(n, "specificity")) {
                boolean hasDefault = defines(n, "default");
                boolean isOptional = isOptional(n);
                int specValue = getSpecificityValue(n);
                if ((hasDefault && applyDefaults && !isOptional) || dataKeys.contains(schemaKey)) {
                    // it will be in the validated output
                    included.put(specValue, schemaKey);
                    if (specValue > maxValue)
                        maxValue = specValue;
                } else {
                    excluded.put(specValue, schemaKey);
                }

            }
        }
        // Now we have been through all the schema keys and determined whether they are:
        // - included, meaning defined in the data or will be defaulted into the data
        // - excluded, meaning they are not defined and will not be default in
        // produce an error message for any that are excluded
        if (maxValue != -1) {
            for (int i = maxValue; i >= 0; i--) {
                if (excluded.get(i) != null) {
                    // there is a lower specificity option not included
                    recordError(ValidationMessages.MORE_SPECIFIC_PROPERTY_REQUIRED, data.getPath(), included.get(maxValue), excluded.get(i));
                }
            }
        }
    }

    private int getSpecificityValue(Node node) {
        Node v = getNode(((MapNode) node), "specificity");
        return toInteger(v);
    }

    /* do not chase down references multiple times as it can lead to infinite recursion */
    private final Set<Node> alreadyFollowed = new HashSet<Node>();

    /* only validate default values once */
    private final Set<Node> defaultsChecked = new HashSet<Node>();

    private boolean areAdditionalPropertiesAllowed(Node additionalProperties) {
        if (additionalProperties == null)
            return true;
        if (isBoolean(additionalProperties) && toBoolean(additionalProperties) == false)
            return false;
        return true;
    }

    /*
     * Look through all the entries in the schema and check the input data against them, then look through all the
     * remaining keys in the data and see if they are allowed and if allowed whether they obey the necessary constraints
     */
    private void checkMap(MapNode data, Node mapSchema, Node additionalProps) {
        if (isMap(mapSchema)) {
            checkMapSchemaKeys(data, (MapNode) mapSchema);
        }
        for (String key : data.getKeys()) {
            if (defines(mapSchema, key))
                continue;
            Node value = getNode(data, key);
            if (!areAdditionalPropertiesAllowed(additionalProps)) {
                recordError(ValidationMessages.PROPERTY_NOT_ALLOWED, data.getPath() + "," + key, key);
            }
            if (additionalProps != null && !isBoolean(additionalProps)) {
                MapNode addProps = (MapNode) additionalProps;
                if (addProps.getKeys().size() != 0) {
                    checkValue(value, additionalProps);
                }
            }
        }
    }

    /*
     * For every constraint in the schema, check if the data defines a value and if so check the data value meets the
     * constraint.
     */
    private void checkMapSchemaKeys(MapNode data, MapNode mapSchema) {
        for (String key : mapSchema.getKeys()) {
            if (alreadyFollowedReference(data, key))
                continue;
            Node value = getNode(data, key);
            checkKeyValuePair(key, value, mapSchema, data);
        }
    }

    private void checkKeyValuePair(String key, Node value, MapNode mapSchema, MapNode parentMap) {
        Node valueDefinition = getNode(mapSchema, key);
        boolean defaultedIn = false;
        if (applyDefaults) {
            if (value == null && valueDefinition != null && isMap(valueDefinition) && !isOptional(valueDefinition)) {
                Node defaultValue = getNode((MapNode) valueDefinition, "default");
                if (defaultValue != null) {
                    List<String> newpath = new ArrayList<String>();
                    newpath.add(key);
                    defaultedIn = true;
                    if (!defaultsChecked.contains(defaultValue)) {
                        defaultsChecked.add(defaultValue);
                        value = copy(defaultValue, parentMap);
                        parentMap.addNode(newpath, value);
                    } else {
                        parentMap.addNode(newpath, copy(defaultValue, parentMap));
                    }
                }
            }
        }

        if (value == null) {
            if (!defaultedIn && !isOptional(valueDefinition)) {
                recordError(ValidationMessages.PROPERTY_NOT_DEFINED_AND_NOT_OPTIONAL, parentMap + "," + key);
            }
        } else {
            checkValue(value, valueDefinition);
        }
    }

    /**
     * Return true if the key in the map is a reference (ie. it is of the form "{'$ref':'something'}" )
     */
    private boolean isReference(MapNode node, String key) {
        Node n = node.getNode(key);
        if (n == null)
            return false;
        if (!isMap(n))
            return false;
        MapNode mn = (MapNode) n;
        Set<String> keys = mn.getKeys();
        if (keys.size() != 1)
            return false;
        if (!keys.contains("$ref"))
            return false;
        return true;
    }

    /*
     * Retrieve a key from a map and if the value turns out to be a reference ( {'$ref':'somewhere'} ) then resolve the
     * reference and return the result. References can be local (within the same schema) or to another schema. Local
     * references start with a '$'. '$' by itself refers to the root of the current schema, and then sub-elements of
     * that schema can be selected by specifying a path from the schema root, eg. '$.properties' is the properties key
     * within the root schema map. References to another schema do not start $, they just start with the id of the
     * schema they are referring to. For example 'person.properties' would be a reference to the properties key in the
     * person schema. Schemas are given ids using the 'id' top level schema attribute and validator instances can share
     * a SchemaRegistry that maps ids to schemas.
     */
    private Node getNode(MapNode map, String key) {
        Node n = map.getNode(key);
        if (n != null && isMap(n)) {
            MapNode mn = ((MapNode) n);
            Set<String> keys = mn.getKeys();
            if (keys.size() == 1) {
                if (mn.getKeys().contains("$ref")) {
                    String value = toString(mn.getNode("$ref"));
                    if (value.charAt(0) == '$') { // local schema reference

                        Node root = mn;
                        while (root.getParent() != null)
                            root = root.getParent();

                        if (value.equals("$")) {
                            return root;
                        }
                        if (value.startsWith("$.")) {
                            List<String> pathString = new ArrayList<String>();
                            StringTokenizer st = new StringTokenizer(value.substring(2), ".");
                            while (st.hasMoreTokens()) {
                                pathString.add(st.nextToken());
                            }
                            return root.getNode(pathString);
                        }
                        recordError(ValidationMessages.UNRESOLVABLE_REFERENCE, map.getPath(), value);
                        return n;
                    } else {
                        if (value.indexOf('.') != -1) { // relative eg. "1d6.properties"
                            MapNode rootSchema = getSchemaRegistry().getSchema(value.substring(0, value.indexOf('.')));
                            if (rootSchema == null) {
                                recordError(ValidationMessages.UNRESOLVABLE_REFERENCE, map.getPath(), value);
                                return n;
                            }
                            String path = value.substring(value.indexOf('.') + 1);
                            List<String> pathString = new ArrayList<String>();
                            StringTokenizer st = new StringTokenizer(path, ".");
                            while (st.hasMoreTokens()) {
                                pathString.add(st.nextToken());
                            }
                            Node nn = rootSchema.getNode(pathString);
                            if (nn == null) {
                                recordError(ValidationMessages.UNRESOLVABLE_REFERENCE, map.getPath(), value);
                                return n;
                            }
                            return nn;
                        } else {
                            MapNode referencedSchema = getSchemaRegistry().getSchema(value);
                            if (referencedSchema == null) {
                                recordError(ValidationMessages.UNRESOLVABLE_REFERENCE, map.getPath(), value);
                                return n;
                            }
                            return referencedSchema;
                        }
                    }
                }
            }
        }
        return n;
    }

    /*
     * For a key that maps to a reference, return the reference value (as opposed to getNode() which goes on to resolve
     * the reference)
     */
    private String getReference(MapNode map, String key) {
        Node n = map.getNode(key);
        if (n != null && isMap(n)) {
            MapNode mn = ((MapNode) n);
            Set<String> keys = mn.getKeys();
            if (keys.size() == 1) {
                if (mn.getKeys().contains("$ref")) {
                    String value = toString(mn.getNode("$ref"));
                    return value;
                }
            }
        }
        return null;
    }

    /**
     * Deep copy a tree structure of nodes.
     */
    private Node copy(Node value, Node parent) {
        if (isMap(value)) {
            MapNode from = ((MapNode) value);
            Set<String> fromKeys = from.getKeys();
            MapNode to = new StandardMapNode(parent);
            for (String key : fromKeys) {
                List<String> path = new ArrayList<String>();
                path.add(key);
                to.addNode(path, copy(from.getNode(key), to));
            }
            return to;
        } else if (isList(value)) {
            ListNode from = ((ListNode) value);
            List<Node> fromNodes = from.getNodes();
            StandardListNode to = new StandardListNode(parent);
            for (Node node : fromNodes) {
                to.addNode(copy(node, to));
            }
            return to;
        } else if (isScalar(value)) {
            if (value instanceof StandardBooleanNode) {
                return new StandardBooleanNode(parent, ((StandardBooleanNode) value).getValue());
            } else if (value instanceof StandardStringNode) {
                return new StandardStringNode(parent, ((StandardStringNode) value).getValue());
            } else if (value instanceof StandardNumberNode) {
                return new StandardNumberNode(parent, ((StandardNumberNode) value).getValue());
            } else if (value instanceof StandardNullNode) {
                return new StandardNullNode(parent);
            }
        }
        return null;
    }

    private boolean isScalar(Node value) {
        return (value instanceof ScalarNode);
    }

    /**
     * If a format is specified for the data and we have a format checker registered for it, then check the data.
     */
    private void checkFormat(Node data, MapNode schemaMap) {
        if (alreadyFollowedReference(schemaMap, "format"))
            return;
        Node format = getNode(schemaMap, "format");
        if (format != null) {
            if (!isString(format))
                return;
            String formatName = toString(format);
            if (formatName.equals("uri")) {
                if (!isString(data)) {
                    recordError(ValidationMessages.INCORRECT_FORMAT_WRONG_TYPE, data.getPath(), formatName, "String", getKind(data));
                }
            } else {
                // Check any registered format checkers
                FormatChecker fChecker = formatCheckers.get(formatName);
                if (fChecker != null) {
                    if (!fChecker.isFormatOK(toString(data))) {
                        recordError(ValidationMessages.VALUE_IS_INCORRECT_FORMAT, data.getPath(), data, format);
                    }
                }
            }
        }
    }

    private boolean alreadyFollowedReference(MapNode data, String key) {
        if (isReference(data, key)) {
            if (alreadyFollowed.contains(data.getNode(key)))
                return true;
            alreadyFollowed.add(data.getNode(key));
        }
        return false;
    }

    /**
     * If the data is supposed to be one of a number of options, check that it is. Also allows for the set of properties
     * to be considered unconstrained, meaning the data does not actually have to be of those specified.
     */
    private void checkOptions(Node data, MapNode schemaAsMap) {
        Node optionsNode = getNode(schemaAsMap, "options");
        if (optionsNode != null && !isUnconstrained(schemaAsMap)) {
            if (!isList(optionsNode)) {
                recordError(ValidationMessages.OPTIONS_NODE_SHOULD_BE_LIST, data.getPath(), getKind(optionsNode));
                return;
            }
            ListNode options = (ListNode) getNode(schemaAsMap, "options");
            Object dataValue = ((ScalarNode) data).getValue();
            boolean found = false;
            for (Node value : options.getNodes()) {
                if (isMap(value)) {
                    MapNode mNode = (MapNode) value;
                    if (((ScalarNode) getNode(mNode, "value")).getValue().equals(dataValue)) {
                        found = true;
                        break;
                    }
                } else {
                    if (((ScalarNode) value).getValue().equals(dataValue)) {
                        found = true;
                        break;
                    }
                }
            }
            if (!found) {
                recordError(ValidationMessages.INVALID_OPTION_VALUE_SPECIFIED, data.getPath(), data, stringifyOptions(options));
            }
        }
    }

    /**
     * Check if the data is limited to a particular number of decimal places with the schema element 'maxDecimal'. If it
     * is then record a validation error if the limit is exceeded.
     */
    private void checkMaxDecimal(Node data, MapNode schemaMap) {
        if (alreadyFollowedReference(schemaMap, "maxDecimal"))
            return;
        Node maxDecimal = getNode(schemaMap, "maxDecimal");
        if (maxDecimal != null) {
            int mD = toInteger(maxDecimal);
            double decimalData = Double.parseDouble((String) ((ScalarNode) data).getValue());
            String decimalDataAsString = Double.toString(decimalData);
            int decimalPointIndex = decimalDataAsString.indexOf(".");
            if (decimalPointIndex == -1)
                return;
            int decimalPlaces = decimalDataAsString.substring(decimalPointIndex + 1).length();
            if (decimalPlaces > mD) {
                recordError(ValidationMessages.TOO_MANY_DECIMAL_PLACES, data.getPath(), decimalDataAsString, mD);
            }
        }
    }

    /**
     * Check if the data is limited to a maximum value according to the schema, and if it is then report an error if the
     * maximum value is exceeded.
     */
    private void checkMaximum(Node data, MapNode schemaAsMap) {
        if (alreadyFollowedReference(schemaAsMap, "maximum"))
            return;
        Node max = getNode(schemaAsMap, "maximum");
        if (max == null)
            return;
        if (sameKind(data, max)) {
            if (isString(data)) { // both nodes are Strings
                if (toString(data).compareTo(toString(max)) > 0) {
                    recordError(ValidationMessages.VALUE_TOO_HIGH, data.getPath(), data, max);
                }
            } else if (isNumber(data)) { // both nodes are Numbers
                if (toNumber(data) > toNumber(max)) {
                    recordError(ValidationMessages.VALUE_TOO_HIGH, data.getPath(), data, max);
                }
            }
        } else if (isString(data) && isNumber(max)) { // data is a String but max value is a Number
            if (toNumber(data) > toNumber(max)) {
                recordError(ValidationMessages.VALUE_TOO_HIGH, data.getPath(), data, max);
            }
        }
    }

    /**
     * Check if the data is limited to a minimum value according to the schema, and if it is then report an error if the
     * data is lower than the minimum value.
     */
    private void checkMinimum(Node data, MapNode schemaAsMap) {
        if (alreadyFollowedReference(schemaAsMap, "minimum"))
            return;
        Node min = getNode(schemaAsMap, "minimum");
        if (min == null)
            return;
        if (sameKind(data, min)) {
            if (isString(data)) { // both nodes are Strings
                if (toString(data).compareTo(toString(min)) < 0) {
                    recordError(ValidationMessages.VALUE_TOO_LOW, data.getPath(), data, min);
                }
            } else if (isNumber(data)) { // both nodes are Numbers
                if (toNumber(data) < toNumber(min)) {
                    recordError(ValidationMessages.VALUE_TOO_LOW, data.getPath(), data, min);
                }
            }
        } else if (isString(data) && isNumber(min)) { // data is a String but max value is a Number
            if (toNumber(data) < toNumber(min)) {
                recordError(ValidationMessages.VALUE_TOO_LOW, data.getPath(), data, min);
            }
        }
    }

    /**
     * Check that the data obeys any regex pattern specified in the schema.
     */
    private void checkPattern(Node data, MapNode schema) {
        if (alreadyFollowedReference(schema, "pattern"))
            return;
        Node patternNode = getNode(schema, "pattern");
        if (patternNode != null) {
            if (isString(patternNode) && isString(data)) {
                String patternString = toString(patternNode);
                String value = toString(data);
                if (!Pattern.matches(patternString, value)) {
                    recordError(ValidationMessages.VALUE_DOES_NOT_MATCH_PATTERN, data.getPath(), value, patternString);
                }
            }
        }
    }

    /**
     * Check if the list must contain a minimum number of items according to the schema.
     */
    private void checkMinItems(ListNode data, MapNode schema) {
        if (alreadyFollowedReference(schema, "minItems"))
            return;
        Node minItems = getNode(schema, "minItems");
        if (minItems != null) {
            int listsize = data.getNodes().size();
            int minitems = toInteger(minItems);
            if (listsize < minitems) {
                recordError(ValidationMessages.VALUE_LIST_TOO_SHORT, data.getPath(), listsize, minitems);
            }
        }
    }

    /**
     * Check if the list must contain a maximum number of items according to the schema.
     */
    private void checkMaxItems(ListNode data, MapNode schema) {
        if (alreadyFollowedReference(schema, "maxItems"))
            return;
        Node maxItems = getNode(schema, "maxItems");
        if (maxItems != null) {
            int listsize = data.getNodes().size();
            int maxitems = toInteger(maxItems);
            if (listsize > maxitems) {
                recordError(ValidationMessages.VALUE_LIST_TOO_LONG, data.getPath(), listsize, maxitems);
            }
        }
    }

    /**
     * Convert a list of options into a string for inclusion in an error message
     */
    private String stringifyOptions(ListNode options) {
        StringBuilder s = new StringBuilder();
        int i = 0;
        for (Node n : options.getNodes()) {
            if (i > 0)
                s.append(",");
            if (isMap(n)) {
                MapNode mN = (MapNode) n;
                s.append(((ScalarNode) mN.getNode("value")).getValue());
            } else {
                s.append(((ScalarNode) n).getValue());
            }
            i++;
        }
        return s.toString();
    }

    /**
     * Check if the data violates a minimum length specified in the schema. Only applies if the data is string data and
     * the schema specifies minLength.
     */
    private void checkMinLength(Node data, MapNode schema) {
        if (alreadyFollowedReference(schema, "minLength"))
            return;
        if (!isString(data))
            return;
        Node minLength = getNode(schema, "minLength");
        if (minLength != null) {
            int minlength = toInteger(minLength);
            if (toString(data).length() < minlength) {
                recordError(ValidationMessages.VALUE_STRING_TOO_SHORT, data.getPath(), data, minlength);
            }
        }
    }

    /**
     * Check if the data violates a minimum length specified in the schema. Only applies if the data is string data and
     * the schema specifies maxLength.
     */
    private void checkMaxLength(Node data, MapNode schemaMap) {
        if (!isString(data))
            return;
        Node maxLength = getNode(schemaMap, "maxLength");
        if (maxLength != null) {
            int maxlength = toInteger(maxLength);
            if (toString(data).length() > maxlength) {
                recordError(ValidationMessages.VALUE_STRING_TOO_LONG, data.getPath(), data, maxlength);
            }
        }
    }

    /**
     * Return an integer representation of the data passed in - any problems in parsing a string node as an integer will
     * be reported as an error.
     */
    private int toInteger(Node possibleInteger) {
        if (isNumber(possibleInteger) || isString(possibleInteger)) {
            try {
                return Integer.parseInt((String) ((ScalarNode) possibleInteger).getValue());
            } catch (Exception e) {
            }
        }
        recordError(ValidationMessages.VALUE_CANNOT_CONVERT_TO_INTEGER, possibleInteger.getPath(), possibleInteger);
        return 0;
    }

    /**
     * Return a double representation of the data passed in - any problems in parsing a string as a double will be
     * reported as an error.
     */
    private Double toNumber(Node possibleNumber) {
        if (isNumber(possibleNumber) || isString(possibleNumber)) {
            try {
                return Double.parseDouble((String) ((ScalarNode) possibleNumber).getValue());
            } catch (Exception e) {
            }
        }
        recordError(ValidationMessages.VALUE_CANNOT_CONVERT_TO_NUMBER, possibleNumber.getPath(), possibleNumber);
        return 0.0d;
    }

    /**
     * Check if the unconstrained property is specified, and return its value if it is. Return the default (false) if it
     * is not specified.
     */
    private boolean isUnconstrained(Node schema) {
        if (isMap(schema)) {
            Node n = getNode((MapNode) schema, "unconstrained");
            if (n != null) {
                if (!isBoolean(n)) {
                    return false;
                } else {
                    return toBoolean(n);
                }
            }
        }
        return false;
    }

    /**
     * Return true if a node is a map and it contains a particular key
     */
    private boolean defines(Node map, String key) {
        if (map == null || !isMap(map))
            return false;
        return getNode(((MapNode) map), key) != null;
    }

    /**
     * Check if the data obeys any specified type restriction.
     */
    private void checkType(Node data, MapNode schema) {
        Node type = getNode(schema, "type");
        if (type == null)
            return;
        boolean nodeIsOfValidType = false;
        if (isList(type)) {
            List<Node> types = ((ListNode) type).getNodes();
            for (Node aType : types) {
                if (matches(data, toString(aType))) {
                    nodeIsOfValidType = true;
                    break;
                }
            }
        } else {
            nodeIsOfValidType = matches(data, toString(type));
        }
        if (!nodeIsOfValidType) {
            // produce a nice string of types allowed for inclusion in the message, eg. "string,boolean"
            StringBuilder s = new StringBuilder();
            if (isList(type)) {
                int i = 0;
                for (Node n : ((ListNode) type).getNodes()) {
                    if (i > 0)
                        s.append(",");
                    s.append(((ScalarNode) n).getValue());
                    i++;
                }
            } else {
                s.append(((ScalarNode) type).getValue());
            }
            recordError(ValidationMessages.VALUE_IS_INCORRECT_TYPE, data.getPath(), getKind(data), s.toString());
        }
    }

    /**
     * return the kind of node for use in validation messages
     */
    private String getKind(Node data) {
        if (isNumber(data)) {
            return "number";
        } else if (isMap(data)) {
            return "map";
        } else if (isList(data)) {
            return "list";
        } else if (isBoolean(data)) {
            return "boolean";
        } else if (isNull(data)) {
            return "null";
        } else if (isString(data)) {
            return "string";
        }
        return null;
    }

    private boolean isOptional(Node n) {
        if (!isMap(n))
            return false;
        Node value = getNode(((MapNode) n), "optional");
        if (value == null)
            return true;
        return toBoolean(value);
    }

    /**
     * Return true if the node matches the type specified in the string.
     */
    private boolean matches(Node n, String typeAsString) {
        if (typeAsString.equals("any")) {
            return true;
        } else if (typeAsString.equals("string")) {
            return isString(n);
        } else if (typeAsString.equals("number")) {
            return isNumber(n);
        } else if (typeAsString.equals("integer")) {
            if (isNumber(n) || isString(n)) {
                try {
                    Integer.parseInt((String) ((ScalarNode) n).getValue());
                    return true;
                } catch (NumberFormatException nfe) {
                    return false;
                }
            } else {
                return false;
            }
        } else if (typeAsString.equals("boolean")) {
            return isBoolean(n);
        } else if (typeAsString.equals("object")) {
            return isMap(n);
        } else if (typeAsString.equals("array")) {
            return isList(n);
        } else if (typeAsString.equals("null")) {
            return isNull(n);
        } else {
            // type is actually unrestricted, so if the typeAsString is something else, we just have to say yes
            return true;
        }
    }

    /**
     * Record an error when the data violates the schema.
     * 
     * @param message the error message
     * @param path the path to the node that failed validation
     * @param otherInserts other message inserts
     */
    private void recordError(ValidationMessages message, String path, Object... otherInserts) {
        Object[] data = new Object[(otherInserts == null ? 0 : otherInserts.length + 1)];
        data[0] = (path.startsWith(",") ? path.substring(1) : path);
        if (otherInserts != null)
            System.arraycopy(otherInserts, 0, data, 1, otherInserts.length);
        validationProblems.add(new JSONValidationProblem(message, data));
    }

    private boolean isMap(Node n) {
        return (n instanceof MapNode);
    }

    private boolean isList(Node data) {
        return (data instanceof ListNode);
    }

    private boolean isNull(Node data) {
        return (data instanceof StandardNullNode);
    }

    private boolean isString(Node data) {
        return (data instanceof StandardStringNode);
    }

    private String toString(Node node) {
        return (String) ((ScalarNode) node).getValue();
    }

    private boolean isNumber(Node data) {
        return (data instanceof StandardNumberNode);
    }

    private boolean isBoolean(Node data) {
        return (data instanceof StandardBooleanNode);
    }

    private boolean toBoolean(Node boolNode) {
        if (!isBoolean(boolNode)) {
            recordError(ValidationMessages.VALUE_CANNOT_CONVERT_TO_BOOLEAN, boolNode.getPath(), boolNode, getKind(boolNode));
            return false;
        }
        return ((Boolean) ((ScalarNode) boolNode).getValue()).booleanValue();
    }

    /**
     * return true if the two nodes are the exact same kind
     * 
     * @return
     */
    private boolean sameKind(Node nodeOne, Node nodeTwo) {
        return (getKind(nodeOne).equals(getKind(nodeTwo)));
    }

    /**
     * Return the JSON Schema schema definition - once parsed it can be used to verify other schemas.
     */
    public static String getSchemaSchema() {
        return theSchemaSchema.replaceAll("'", "\"");
    }

    /**
     * This is the JSON definition of the schema schema as defined here:
     * http://groups.google.com/group/json-schema/web/json-schema-proposal---second-draft
     * 
     * Single quotes are used to make it less ugly! they need transforming to double quotes before a JSON parser should
     * accept it and that is done when getSchemaSchema() is called.
     */
    private final static String theSchemaSchema = "{'description':'This is the JSON Schema for JSON Schemas.',\n"
        + " 'type':['object','array'],\n"
        + " 'items':{\n"
        + "     'type':'object',\n"
        + "     'properties':{'$ref':'$.properties'},\n"
        + "     'description':'When the schema is an array, it indicates that it is enforcing tuple typing. Each item in the instance array must correspond to the item in the schema array'},\n"
        + " 'additionalProperties':false,\n"
        + " 'properties':{\n"
        + "     'type':{\n"
        + "     	'type':['string','array'],\n"
        + "     	'items':{'$ref':'$.properties.type'},\n"
        + "	    'description':'This is a type definition value. This can be a simple type, or a union type',\n"
        + "	    'options':['string','object','array','boolean','number','integer','null','any'],\n"
        + "	    'unconstrained':true,\n"
        + "	    'optional':true,\n"
        + "	    'default':'any'},\n"
        + "     'optional':{\n"
        + "     	'type':'boolean',\n"
        + "     	'description':'This indicates that the instance property in the instance object is not required.',\n"
        + "	    'optional':true,\n"
        + "	    'default':false},\n"
        + "     'properties':{\n"
        + "     	'type':'object',\n"
        + "     	'additionalProperties':{'$ref':'$'},\n"
        + "     	'description':'This is a definition for the properties of an object value',\n"
        + "     	'optional':true,\n"
        + "     	'default':{}\n"
        + "     },\n"
        + "     'items':{\n"
        + "     	'type':['object','array'],\n"
        + "     	'properties':{'$ref':'$.properties'},\n"
        + "     	'description':'When the value is an array, this indicates the schema to use to validate each item in an array',\n"
        + "	    'optional':true,\n"
        + "	    'default':{}},\n"
        + "     'additionalProperties':{\n"
        + "     	'type':['boolean','object'],\n"
        + "     	'properties':{'$ref':'$.properties'},\n"
        + "     	'description':'This provides a default property definition for all properties that are not explicitly defined in an object type definition. The value must be a schema. If false is provided, no additional properties are allowed, and the schema can not be extended. The default value is an empty schema which allows any value for additional properties.',\n"
        + "	    'optional':true,\n"
        + "	    'default':{}},\n"
        + "     'specificity':{\n"
        + "        'type':'number',\n"
        + "        'description':'This indicates an order of specificity of properties. If an instance defines another property with a higher specificity than this one, than this instance property is required.',\n"
        + "	    'optional':true,\n"
        + "        'default':false},\n"
        + "     'unique':{\n"
        + "     	'type':'boolean',\n"
        + "     	'description':'This indicates that the instance property should have unique values. No other property with the same name in the instance object tree should have the same value.',\n"
        + "	    'optional':true,\n"
        + "	    'default':false},\n"
        + "     'minimum':{\n"
        + "     	'type':'number',\n"
        + "	    'optional':true,\n"
        + "     	'description':'This indicates the minimum value for the instance property when the type of the instance value is a number, or it indicates the minimum number of values in an array when an array is the instance value.'},\n"
        + "     'maximum':{\n"
        + "     	'type':'number',\n"
        + "	    'optional':true,\n"
        + "     	'description':'This indicates the maximum value for the instance property when the type of the instance value is a number, or it indicates the maximum number of values in an array when an array is the instance value.'},\n"
        + "     'pattern':{\n"
        + "     	'type':'string',\n"
        + "     	'format':'regex',\n"
        + "     	'description':'When the instance value is a string, this provides a regular expression that a instance string value should match in order to be valid.',\n"
        + "	    'optional':true,\n"
        + "	    'default':'.*'},\n"
        + "     'maxLength' :{\n"
        + "     	'type':'number',\n"
        + "	    'optional':true,\n"
        + "     	'description':'When the instance value is a string, this indicates maximum length of the string.'},\n"
        + "     'minLength' :{\n"
        + "     	'type':'number',\n"
        + "	    'optional':true,\n"
        + "     	'description':'When the instance value is a string, this indicates minimum length of the string.'},\n"
        + "     'maxItems' :{\n"
        + "     	'type':'number',\n"
        + "	    'optional':true,\n"
        + "     	'description':'When the instance value is an array, this indicates maximum number of items.'},\n"
        + "     'minItems' :{\n"
        + "     	'type':'number',\n"
        + "	    'optional':true,\n"
        + "     	'description':'When the instance value is an array, this indicates minimum number of items.'},\n"
        + "     'options' : {\n"
        + "     	'type':'array',\n"
        + "     	'items':{\n"
        + "     	   'properties':{\n"
        + "     	   	'label':{\n"
        + "     	   	    'type':'string',\n"
        + "     	   	    'description':'This is the label for this option',\n"
        + "     	   	    'optional':true\n"
        + "     	   	    },\n"
        + "     	   	'value':{\n"
        + "     	   	    'description':'This is the value for this option'\n"
        + "     	   	}\n"
        + "     	   },\n"
        + "     	   'description':'This is an option for list of possible values'\n"
        + "     	},\n"
        + "	    'optional':true,\n"
        + "     	'description':'This provides an enumeration of possible values that are valid for the instance property.'},\n"
        + "     'unconstrained':{\n"
        + "     	'type':'boolean',\n"
        + "     	'description':'When used in conjunction with the options property, this indicates a value can be used that is not in the list of options. This has no meaning when the options property is not a sibling.',\n"
        + "	    'optional':true,\n"
        + "	    'default':false},\n"
        + "     'readonly':{\n"
        + "     	'type':'boolean',\n"
        + "     	'description':'This indicates that the instance property should not be changed (this is only for interaction, it has no effect for standalone validation).',\n"
        + "	    'optional':true,\n"
        + "	    'default':false},\n"
        + "     'description':{\n"
        + "     	'type':'string',\n"
        + "	    'optional':true,\n"
        + "     	'description':'This provides a description of the purpose the instance property. The value can be a string or it can be an object with properties corresponding to various different instance languages (with an optional default property indicating the default description).'},\n"
        + "     'format':{\n"
        + "     	'type':'string',\n"
        + "	    'optional':true,\n"
        + "     	'description':'This indicates what format the data is among some predefined formats'},\n"
        + "     'default':{\n"
        + "     	'type':'any',\n"
        + "	    'optional':true,\n"
        + "     	'description':'This indicates the default for the instance property.'},\n"
        + "     'transient':{\n"
        + "     	'type':'boolean',\n"
        + "	    'optional':true,\n"
        + "     	'description':'This indicates that the property will be used for transient/volatile values that should not be persisted.',\n"
        + "	    'default':false},\n"
        + "     'maxDecimal':{\n"
        + "     	'type':'integer',\n"
        + "	    'optional':true,\n"
        + "     	'description':'This indicates the maximum number of decimal places in a floating point number.'},\n"
        + "     'hidden':{\n"
        + "     	'type':'boolean',\n"
        + "	    'optional':true,\n"
        + "     	'description':'This indicates whether the property should be hidden in user interfaces.'},     	\n"
        + "     'extends':{\n"
        + "     	'type':'object',\n"
        + "     	'properties':{'$ref':'$.properties'},\n"
        + "     	'description':'This indicates the schema extends the given schema. All instances of this schema must be valid to by the extended schema also.',\n"
        + "	    'optional':true,\n" + "	    'default':{}},     	\n" + "     'id':{\n" + "     	'type':['string','number'],\n"
        + "	    'optional':true,\n" + "	    'format':'url',\n" + "      	'unique':true},\n" + "     '$ref':{\n" + "        'type':'string'}\n"
        + "   }\n" + "}\n";

}