/*
 * Copyright 2007-2008 SpringSource Inc.
 */

package com.springsource.json.validation;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import com.springsource.json.parser.MapNode;
import com.springsource.json.parser.Node;
import com.springsource.json.parser.ScalarNode;

/**
 * A simple schema registry that maps from schema id to the root node in the schema. It is used by the JSONValidator
 * which records entries in it for each schema it is asked to validate. A SchemaRegistry instance can be shared amongst
 * multiple JSONValidators and if non is provided for a JSONValidator instance it will create one the first time it
 * needs it. 
 * 
 * @author Andy Clement
 * @since 1.0
 */
public class SchemaRegistry {

    private final Map<String, MapNode> schemas = new HashMap<String, MapNode>();

    /**
     * Register a schema with the registry - this will only succeed if the there is a
     * top level 'id' node within the schema.
     * @param schema the schema to register
     */
    public void register(MapNode schema) {
        Node idNode = schema.getNode("id");
        if (idNode == null)
            return;
        String id = (String) ((ScalarNode) idNode).getValue();
        schemas.put(id, schema);
    }

    /**
     * Remove a specified schema from the registry.
     * @param id the identity of the schema to remove
     */
    public void remove(String id) {
        schemas.remove(id);
    }

    /**
     * Return the ids of all registered schemas.
     * @return set of strings representings the ids of registered schemas.
     */
    public Set<String> getRegisteredSchemas() {
        Set<String> allRegistered = new HashSet<String>();
        allRegistered.addAll(schemas.keySet());
        return allRegistered;
    }

    /**
     * Retrieve any schema registered under the specified id.
     * @param id the identity of the schema to look up
     * @return any registered schema, or null if the id is unrecognized
     */
    public MapNode getSchema(String id) {
        return schemas.get(id);
    }

    /**
     * Clear the registry, forgetting all schemas.
     */
    public void clear() {
        schemas.clear();
    }

}
