/*
 * Copyright 2008, SpringSource.
 */

package com.springsource.json.writer.xml;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.OptionBuilder;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.apache.commons.cli.PosixParser;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import com.springsource.json.writer.JSONObject;
import com.springsource.json.writer.JSONWriter;
import com.springsource.json.writer.JSONWriterImpl;



/**
 * Converts any XML file to JSON. The output is compressed and not optimized for readability. Use
 * http://curiousconcept.com/jsonformatter/ if you need a human-readable output.
 * TODO Order output to make it more readable and provide pretty output. 
 * 
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * XmlToJSONConverter is thread safe
 * 
 * @author Paul Kuzan
 * @since 1.0
 */
public class XmlToJSONConverter {

    private static String SERVER = "Server";

    /**
     * Converts XML to JSON
     * 
     * @param xml - The XML to convert
     * @param writer - The writer to stream JSON output to
     * 
     */
    public static void converntXMLToJSON(File xml, Writer writer) {
        if (xml == null) {
            throw new IllegalArgumentException("Please specify an XML file to convert");
        }
        if (writer == null) {
            throw new IllegalArgumentException("Please specify a writer to stream JSON output to");
        }
        if (!xml.exists()) {
            throw new JSONConverterException("XML file at " + xml.getAbsolutePath() + " does not exist");
        }

        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        DocumentBuilder docBuilder = null;
        try {
            docBuilder = factory.newDocumentBuilder();
        } catch (ParserConfigurationException parserConfigurationException) {
            throw new JSONConverterException("Failed to create document builder", parserConfigurationException);
        }
        org.w3c.dom.Document doc = null;
        try {
            doc = docBuilder.parse(xml);
        } catch (SAXException exception) {
            throw new JSONConverterException("Failed to parse XML", exception);
        } catch (IOException exception) {
            throw new JSONConverterException("Failed to open XML file", exception);
        }

        org.w3c.dom.Element elt = doc.getDocumentElement();

        JSONWriter jsonWriter = new JSONWriterImpl(writer);
        JSONObject server = new JSONObject();
        parseElement(elt, server);
        jsonWriter.object().key(SERVER).value(server).endObject();
    }

    private static void parseElement(Element elt, JSONObject object) {
        NodeList list = elt.getChildNodes();
        Element child = null;
        Node node;
        for (int i = 0; i < list.getLength(); i++) {
            node = list.item(i);
            if (node instanceof org.w3c.dom.Element) {
                child = (Element) node;
                JSONObject childObject = new JSONObject();
                object.accumulate(child.getTagName(), childObject);
                parseElement(child, childObject);
            }
        }
        NamedNodeMap attributes = elt.getAttributes();

        for (int i = 0; i < attributes.getLength(); i++) {
            node = attributes.item(i);
            object.accumulate(node.getNodeName(), node.getTextContent());
        }
    }

    @SuppressWarnings("static-access")
    public static void main(String[] args) {
        Options options = new Options();

        Option json = OptionBuilder.withDescription("The path where the generated JSON will be created").withLongOpt("json").withValueSeparator('=').isRequired().hasArg().create(
            'j');

        Option xml = OptionBuilder.withDescription("The path to XML file to be converted").withLongOpt("xml").withValueSeparator('=').isRequired().hasArg().create(
            'x');

        options.addOption(xml);
        options.addOption(json);

        CommandLineParser parser = new PosixParser();
        CommandLine cmd;
        try {
            cmd = parser.parse(options, args);
        } catch (ParseException pe) {
            usage(options);
            return;
        }

        File xmlFile = new File(cmd.getOptionValue("xml"));
        if (!xmlFile.exists()) {
            System.err.println("XML file does not extist at path " + xmlFile.getAbsolutePath());
            return;
        }

        File jsonFile = new File(cmd.getOptionValue("json"));
        if (jsonFile.isDirectory()) {
            System.err.println("Please specify a full path for the output JSON file, a driectory was sepecified." + xmlFile.getAbsolutePath());
            return;
        }

        FileWriter writer = null;
        try {
            writer = new FileWriter(jsonFile);
            XmlToJSONConverter.converntXMLToJSON(xmlFile, writer);
            writer.flush();
        } catch (IOException e) {
            throw new RuntimeException("Problem writing to JSON file.", e);
        }
    }

    private static void usage(Options options) {
        HelpFormatter formatter = new HelpFormatter();
        formatter.printHelp("XMLToJSON", options);
    }

}
