/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.bootstrap.client;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import javax.management.MBeanServerConnection;
import javax.management.ObjectName;
import javax.management.remote.JMXConnector;
import javax.management.remote.JMXConnectorFactory;
import javax.management.remote.JMXServiceURL;
import javax.naming.ServiceUnavailableException;

/**
 * A stand-alone client for shutdown of the Server.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * Thread-safe.
 * 
 * @author Andy Wilkinson
 * @since 1.0
 */
public class Shutdown {

    private static final String DEFAULT_PASSWORD = "springsource";
    private static final String DEFAULT_USERNAME = "admin";
    private static final String SERVICE_URL_TEMPLATE = "service:jmx:rmi:///jndi/rmi://127.0.0.1:%s/jmxrmi";
    private static final int DEFAULT_PORT = 9875;

    public static void main(String[] args) {

	int port = DEFAULT_PORT;
	boolean immediate = false;
	String username = DEFAULT_USERNAME;
	String password = DEFAULT_PASSWORD;

	if (args.length > 0) {
	    for (int i = 0; i < args.length; i++) {
		if (args[i].equals("-username")) {
		    if (i < args.length - 1) {
			username = args[++i];
		    } else {
			displayUsageAndExit();
		    }
		} else if (args[i].equals("-password")) {
		    if (i < args.length - 1) {
			password = args[++i];
		    } else {
			displayUsageAndExit();
		    }
		} else if (args[i].equals("-port")) {
		    if (i < args.length - 1) {
			try {
			    port = Integer.parseInt(args[++i]);
			} catch (NumberFormatException nfe) {
			    displayUsageAndExit();
			}
		    } else {
			displayUsageAndExit();
		    }
		} else if (args[i].equals("-immediate")) {
		    immediate = true;
		} else {
		    displayUsageAndExit();
		}
	    }
	}

	try {
	    JMXServiceURL jmxServiceURL = new JMXServiceURL(String.format(SERVICE_URL_TEMPLATE, port));
	    
	    Map<String, Object> jmxEnvironment = new HashMap<String, Object>();
	    jmxEnvironment.put("jmx.remote.credentials", new String[] { username , password });
	    JMXConnector connector = JMXConnectorFactory.connect(jmxServiceURL, jmxEnvironment);
	    MBeanServerConnection connection = connector.getMBeanServerConnection();
	    ObjectName shutdownMBeanName = new ObjectName("com.springsource.server", "type", "Shutdown");
	    if (immediate) {
		connection.invoke(shutdownMBeanName, "immediateShutdown", null, null);
	    } else {
		connection.invoke(shutdownMBeanName, "shutdown", null, null);
	    }
	    connector.close();
	} catch (IOException ioe) {
	    Throwable cause = ioe.getCause();
	    if (cause instanceof ServiceUnavailableException) {
		System.out.println("The Server could not be reached, it may already be stopped.");
	    } else {
		ioe.printStackTrace();
	    }
	} catch (Exception e) {
	    e.printStackTrace();
	}
    }

    private static void displayUsageAndExit() {
	System.out.println("Usage: shutdown [-options]");
	System.out.println("Available options:");
	System.out.println("    -port nnnn        Specifies the control port of the Server");
	System.out.println("                      instance which is to be shutdown.");
	System.out.println("    -immediate        Specifies that the Server should be shutdown");
	System.out.println("                      immediately.");
	System.out.println("    -username         Specifies the username to use for the");
	System.out.println("                      connection to the server.");
	System.out.println("    -password         Specifies the password to use for the");
	System.out.println("                      connection to the server.");
	System.exit(1);
    }
}
