/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.common.math;

import com.springsource.server.serviceability.ffdc.NonNull;

/**
 * Defines an ordered pair of types <code>F</code> and <code>S</code>.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * Implementation is immutable.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
public final class OrderedPair<F, S> {

    private final F first;

    private final S second;

    /**
     * Creates a new <code>OrderedPair</code>.
     * 
     * @param first the first member of the <code>OrderedPair</code>.
     * @param second the second member of the <code>OrderedPair</code>.
     */
    public OrderedPair(@NonNull F first, @NonNull S second) {
        this.first = first;
        this.second = second;
    }

    /**
     * Gets the first member of the <code>OrderedPair</code>.
     * 
     * @return the first member of the <code>OrderedPair</code>.
     */
    public F getFirst() {
        return this.first;
    }

    /**
     * Gets the second member of the <code>OrderedPair</code>.
     * 
     * @return the second member of the <code>OrderedPair</code>.
     */
    public S getSecond() {
        return this.second;
    }

    /**
     * {@inheritDoc}
     */
    @Override public int hashCode() {
        final int prime = 31;
        int result = this.first == null ? 0 : this.first.hashCode();
        result = prime * result + (this.second == null ? 0 : this.second.hashCode());
        return result;
    }

    /**
     * {@inheritDoc}
     */
    @SuppressWarnings("unchecked") @Override public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final OrderedPair other = (OrderedPair) obj;
        return this.first.equals(other.first) && this.second.equals(other.second);
    }

    /**
     * {@inheritDoc}
     */
    @Override public String toString() {
        return "(" + this.first.toString() + ", " + this.second.toString() + ")";
    }

}
