/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.common.math;

import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

import com.springsource.server.serviceability.output.Level;
import com.springsource.server.serviceability.tracing.EntryExitTraceLevel;

/**
 * Defines common operations on {@link Set Sets}.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
public class Sets {

    /**
     * Calculates the difference between <code>a</code> and <code>b</code>.
     * 
     * @param <T> the type of the <code>Set</code>.
     * @param a the lhs of the difference.
     * @param b the rhs of the difference.
     * @return the difference between <code>a</code> and <code>b</code>.
     */
    @EntryExitTraceLevel(Level.DEBUG) public static <T> Set<T> difference(Set<T> a, Set<T> b) {
        Set<T> result = new HashSet<T>(a);
        result.removeAll(b);
        return result;
    }

    /**
     * Calculates the intersection between <code>a</code> and <code>b</code>.
     * 
     * @param <T> the type of the <code>Set</code>.
     * @param a the lhs of the intersection.
     * @param b the rhs of the intersection.
     * @return the intersection between <code>a</code> and <code>b</code>.
     */
    public static <T> Set<T> intersection(Set<T> a, Set<T> b) {
        Set<T> iter = a.size() < b.size() ? a : b;
        Set<T> other = iter == a ? b : a;

        Set<T> result = new HashSet<T>();
        for (T t : iter) {
            if (other.contains(t)) {
                result.add(t);
            }
        }
        return result;
    }

    /**
     * Creates a <code>Set</code> containing the supplied items.
     * 
     * @param <T> the type of the <code>Set</code> entries.
     * @param items the <code>Set</code> contents.
     * @return the resultant <code>Set</code>.
     */
    public static <T> Set<T> asSet(T... items) {
        Set<T> set = new HashSet<T>(items.length);
        Collections.addAll(set, items);
        return set;
    }
}
