/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.concurrent.core;

import java.util.concurrent.ExecutorService;
import java.util.concurrent.ThreadPoolExecutor;

import com.springsource.server.concurrent.core.accessor.ThreadLocalAccessor;
import com.springsource.server.concurrent.core.accessor.ThreadLocalAccessorFactory;
import com.springsource.server.concurrent.core.accessor.ThreadLocalData;
import com.springsource.server.serviceability.output.Level;
import com.springsource.server.serviceability.tracing.EntryExitTraceLevel;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * Delegate object that encapsulates common operations for {@link ExecutorService} implementations.<p/> Each
 * {@link ExecutorService} should maintain its own instance of <code>ExecutorServiceDelegate</code>.
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
final class ExecutorServiceDelegate {

    private static final Tracer TRACER = TracerFactory.createTracer(ExecutorServiceDelegate.class);

    private final ThreadLocalAccessor threadLocalAccessor = ThreadLocalAccessorFactory.newThreadLocalAccessor();

    // always false for now
    private final boolean clearThreadLocals = false;

    private final Object monitor = new Object();

    private long totalExecutionTime;

    /**
     * Clears the <code>ThreadLocals</code> for this <code>Thread</code>. Safe due to invocation against current
     * <code>Thread</code>.
     */
    @EntryExitTraceLevel(Level.DEBUG)
    public void clearThreadLocals() {
        if (this.clearThreadLocals) {
            Thread thread = Thread.currentThread();
            ThreadLocalData[] threadLocals = this.threadLocalAccessor.threadLocals(thread);
            for (ThreadLocalData threadLocal : threadLocals) {
                if (!(threadLocal.getThreadLocal() instanceof PersistentThreadLocal)) {
                    TRACER.traceDebug("Clearing ThreadLocal '%s' for thread '%s'.", threadLocal.getThreadLocal(), thread);
                    this.threadLocalAccessor.removeThreadLocal(threadLocal.getThreadLocal(), thread);
                }
            }
        }
    }

    /**
     * Gets an estimate of the average amount of time spent processing successful tasks.
     * 
     * @return the estimate of time spent.
     * @see ThreadPoolExecutor#getCompletedTaskCount()
     */
    public long getAverageExecutionTime(long completedTaskCount) {
        synchronized (this.monitor) {
            return completedTaskCount == 0 ? this.totalExecutionTime : this.totalExecutionTime / completedTaskCount;
        }
    }

    /**
     * Gets an estimate of the total amount of time spent processing successful tasks.
     * 
     * @return the estimate of time spent.
     * @see ThreadPoolExecutor#getCompletedTaskCount()
     */
    public long getExecutionTime() {
        synchronized (this.monitor) {
            return this.totalExecutionTime;
        }
    }

    /**
     * Creates a {@link Runnable} wrapper that gathers execution statistics for the supplied {@link Runnable}.
     * 
     * @param delegate the <code>Runnable</code> to gather the statistics for.
     * @return the wrapper.
     */
    public Runnable decorateForStatisticsTracking(Runnable delegate) {
        return new StatisticsGatheringRunnable(delegate);
    }

    /**
     * Simple {@link Runnable} that tracks execution statistics for another, wrapped <code>Runnable</code> instance.<p/>
     * 
     * <strong>Concurrent Semantics</strong><br/>
     * 
     * Threadsafe.
     * 
     * @author Rob Harrop
     * @since 1.0
     */
    private class StatisticsGatheringRunnable implements Runnable {

        private final Runnable delegate;

        /**
         * @param delegate
         */
        public StatisticsGatheringRunnable(Runnable delegate) {
            this.delegate = delegate;
        }

        /**
         * {@inheritDoc}
         */
        @EntryExitTraceLevel(Level.DEBUG)
        public void run() {
            long timeBefore = System.currentTimeMillis();
            try {
                this.delegate.run();
            } finally {
                ExecutorServiceDelegate outer = ExecutorServiceDelegate.this;
                long time = System.currentTimeMillis() - timeBefore;
                synchronized (outer.monitor) {
                    outer.totalExecutionTime += time;
                }
            }
        }
    }
}
