/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.concurrent.core;

import java.util.concurrent.RejectedExecutionHandler;
import java.util.concurrent.ScheduledThreadPoolExecutor;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * Extension of {@link ScheduledThreadPoolExecutor} that enforces thread creation via a {@link ThreadManager}.
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @author Christopher Frost
 * @since 1.0
 */
final class ServerScheduledThreadPoolExecutor extends ScheduledThreadPoolExecutor implements ServerScheduledExecutorService {

    static final String DEFAULT_POOL_NAME_PREFIX = "ServerScheduledThreadPool";

    private static final AtomicInteger POOL_NAME_COUNTER = new AtomicInteger(0);

    private final ExecutorServiceDelegate delegate = new ExecutorServiceDelegate();

    private final String poolName;

    /**
     * Creates a new <code>ServerScheduledThreadPoolExecutor</code>.
     * 
     * @param corePoolSize the number of threads in the pool.
     * @param threadManager the {@link ThreadManager}.
     * @param handler the {@link RejectedExecutionHandler}.
     * @see ScheduledThreadPoolExecutor#ScheduledThreadPoolExecutor(int, RejectedExecutionHandler)
     */
    public ServerScheduledThreadPoolExecutor(int corePoolSize, ThreadManager threadManager, RejectedExecutionHandler handler) {
        this(corePoolSize, threadManager, newPoolName(), handler);
    }

    /**
     * Creates a new <code>ServerScheduledThreadPoolExecutor</code>.
     * 
     * @param corePoolSize the number of threads in the pool.
     * @param threadManager the {@link ThreadManager}.
     * @see ScheduledThreadPoolExecutor#ScheduledThreadPoolExecutor(int)
     */
    public ServerScheduledThreadPoolExecutor(int corePoolSize, ThreadManager threadManager) {
        this(corePoolSize, threadManager, newPoolName());
    }

    /**
     * Creates a new <code>ServerScheduledThreadPoolExecutor</code>.
     * 
     * @param corePoolSize the number of threads in the pool.
     * @param threadManager the {@link ThreadManager}.
     * @param poolName the name of the pool.
     * @param handler the {@link RejectedExecutionHandler}.
     * @see ScheduledThreadPoolExecutor#ScheduledThreadPoolExecutor(int, RejectedExecutionHandler)
     */
    public ServerScheduledThreadPoolExecutor(int corePoolSize, ThreadManager threadManager, String poolName, RejectedExecutionHandler handler) {
        super(corePoolSize, new ThreadManagerPoolThreadFactory(poolName, threadManager), handler);
        this.poolName = poolName;
    }

    /**
     * Creates a new <code>ServerScheduledThreadPoolExecutor</code>.
     * 
     * @param corePoolSize the number of threads in the pool.
     * @param threadManager the {@link ThreadManager}.
     * @param poolName the name of the pool.
     * @see ScheduledThreadPoolExecutor#ScheduledThreadPoolExecutor(int)
     */
    public ServerScheduledThreadPoolExecutor(int corePoolSize, ThreadManager threadManager, String poolName) {
        super(corePoolSize, new ThreadManagerPoolThreadFactory(poolName, threadManager));
        this.poolName = poolName;
    }

    /**
     * {@inheritDoc}
     */
    @Override protected void afterExecute(Runnable r, Throwable t) {
        super.afterExecute(r, t);
        this.delegate.clearThreadLocals();
    }

    /**
     * {@inheritDoc}
     */
    @Override public void execute(Runnable command) {
        super.execute(this.delegate.decorateForStatisticsTracking(command));
    }

    /**
     * Gets an estimate of the average amount of time spent processing successful tasks.
     * 
     * @return the estimate of time spent.
     * @see ThreadPoolExecutor#getCompletedTaskCount()
     */
    public long getAverageExecutionTime() {
        return this.delegate.getAverageExecutionTime(getCompletedTaskCount());
    }

    /**
     * Gets an estimate of the total amount of time spent processing successful tasks.
     * 
     * @return the estimate of time spent.
     * @see ThreadPoolExecutor#getCompletedTaskCount()
     */
    public long getExecutionTime() {
        return this.delegate.getExecutionTime();
    }

    /**
     * Get the unique name of the Pool used in this executor service
     * 
     * {@inheritDoc}
     */
    public String getPoolName() {
        return this.poolName;
    }

    /**
     * Get the name of the type of this Thread Pool Executor.
     * @return the name of this thread pool executor's type
     */
    public String getTypeName() {
        return DEFAULT_POOL_NAME_PREFIX;
    }

    private static String newPoolName() {
        return String.format("%1$s-%2$d", DEFAULT_POOL_NAME_PREFIX, POOL_NAME_COUNTER.getAndIncrement());
    }

}
