/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.concurrent.core;

import java.util.Properties;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.RejectedExecutionHandler;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;

import org.osgi.framework.BundleContext;
import org.osgi.framework.ServiceRegistration;

import com.springsource.server.concurrent.core.diagnostics.FatalConcurrentException;
import com.springsource.server.kernel.core.Subsystem;
import com.springsource.server.kernel.core.SubsystemRegistry;
import com.springsource.server.management.proxy.concurrent.ExecutorServiceInfo;
import com.springsource.server.serviceability.tracing.TracingService;

/**
 * Standard implementation of {@link ExecutorManager}. <p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe
 * 
 * @author Rob Harrop
 * @author Andy Wilkinson
 * @author Christopher Frost
 * @since 1.0
 */
final class StandardExecutorManager implements ExecutorManager {

    private final ThreadManager threadManager;

    private final ConcurrentWeakServiceRegistrationTracker<ExecutorServiceStatistics, Subsystem> subsystemExecutors = new ConcurrentWeakServiceRegistrationTracker<ExecutorServiceStatistics, Subsystem>();

    private final ServiceRegistration defaultExecutorRegistration;

    private final ServiceRegistration defaultScheduledExecutorRegistration;

    private static final String DEFAULT_EXECUTOR_NAME = "default-executor";

    private static final String DEFAULT_SCHEDULED_EXECUTOR_NAME = "default-scheduled-executor";

    private final BundleContext bundleContext;

    private final TracingService tracingService;
    
    private static final String EXECUTOR_NAME_KEY = "com.springsource.server.concurrent.executor.name";
    
    private static final String SCHEDULED_EXECUTOR_NAME_KEY = "com.springsource.server.concurrent.scheduledexecutor.name";

    /**
     * Creates a new <code>StandardEngineExecutorManager</code>.
     * 
     * @param threadManager the {@link ThreadManager} for creating {@link Thread Threads}.
     */
    public StandardExecutorManager(ThreadManager threadManager, SubsystemRegistry subsystemRegistry, BundleContext bundleContext,
        TracingService tracingService) {
        this.threadManager = threadManager;
        this.bundleContext = bundleContext;
        this.tracingService = tracingService;

        Subsystem thisSubsystem = subsystemRegistry.getSubsystemByBundle(bundleContext.getBundle());

        ExecutorService defaultExecutor = newExecutorService(DEFAULT_EXECUTOR_NAME, thisSubsystem, 5, 5, 30, TimeUnit.SECONDS,
            new LinkedBlockingQueue<Runnable>());
        Properties properties = new Properties();
        properties.put(EXECUTOR_NAME_KEY, DEFAULT_EXECUTOR_NAME);
        this.defaultExecutorRegistration = bundleContext.registerService(ExecutorService.class.getName(), defaultExecutor, properties);

        ScheduledExecutorService defaultScheduledExecutor = newScheduledExecutorService(DEFAULT_SCHEDULED_EXECUTOR_NAME, thisSubsystem, 5);
        properties = new Properties();
        properties.put(SCHEDULED_EXECUTOR_NAME_KEY, DEFAULT_SCHEDULED_EXECUTOR_NAME);
        this.defaultScheduledExecutorRegistration = bundleContext.registerService(ScheduledExecutorService.class.getName(), defaultScheduledExecutor,
            properties);
    }

    /**
     * {@inheritDoc}
     */
    public ServerExecutorService newExecutorService(String name, Subsystem owner, int coreSize, int maxSize, long keepAlivePeriod, TimeUnit keepAliveUnit,
        BlockingQueue<Runnable> queue, RejectedExecutionHandler rejectedExecutionHandler) {
        ServerThreadPoolExecutor executorService = new ServerThreadPoolExecutor(coreSize, maxSize, keepAlivePeriod, keepAliveUnit,
            this.threadManager, queue, name, rejectedExecutionHandler, this.tracingService);
        return registerAndRecord(owner, executorService);
    }

    /**
     * {@inheritDoc}
     */
    public ServerExecutorService newExecutorService(String name, Subsystem owner, int coreSize, int maxSize, long keepAlivePeriod, TimeUnit keepAliveUnit,
        BlockingQueue<Runnable> queue) {
        ServerThreadPoolExecutor executorService = new ServerThreadPoolExecutor(coreSize, maxSize, keepAlivePeriod, keepAliveUnit,
            this.threadManager, queue, name, this.tracingService);
        return registerAndRecord(owner, executorService);
    }

    /**
     * {@inheritDoc}
     */
    public ServerScheduledExecutorService newScheduledExecutorService(String name, Subsystem owner, int coreSize) {
        ServerScheduledThreadPoolExecutor executorService = new ServerScheduledThreadPoolExecutor(coreSize, this.threadManager, name);
        return registerAndRecord(owner, executorService);
    }

    /**
     * Registers the <code>ExecutorServiceStatistics</code> in the service registry, under the appropriate info object,
     * and records it against the supplied owner.
     * 
     * @param owner the owner of the <code>ExecutorService</code>.
     * @param executorService the <code>ExecutorService</code>.
     */
    private <T extends ExecutorServiceStatistics> T registerAndRecord(Subsystem owner, T executorService) {
        ServiceRegistration sr = publishThreadPoolExecutorInfo(executorService);
        this.subsystemExecutors.record(executorService, owner, sr);
        return executorService;
    }

    /**
     * Create a new instance of <code>StandardExecutorServiceInfo</code> to wrap the new EAP
     * <code>ExecutorService</code> and publish it in the service registry.
     * 
     * @param service
     */
    private ServiceRegistration publishThreadPoolExecutorInfo(ExecutorServiceStatistics service) {
        try {
            StandardExecutorServiceInfo info = new StandardExecutorServiceInfo(service);
            return this.bundleContext.registerService(ExecutorServiceInfo.class.getName(), info, null);
        } catch (Exception e) {
            throw new FatalConcurrentException("Unable to publish thread pool information.", e);
        }
    }

    public void shutdown() {
        this.defaultExecutorRegistration.unregister();
        this.defaultScheduledExecutorRegistration.unregister();
    }
}
