/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.concurrent.core;

import java.lang.ref.Reference;
import java.lang.ref.WeakReference;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.ScheduledExecutorService;

import com.springsource.server.management.proxy.concurrent.ExecutorServiceInfo;

/**
 * Standard implementation of {@link ExecutorServiceInfo}. Maintains a {@link WeakReference} to the underlying
 * {@link ExecutorServiceStatistics} to prevent pinning the executor in memory.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Christopher Frost
 * @since 1.0
 */
public class StandardExecutorServiceInfo implements ExecutorServiceInfo {

    private final Reference<ExecutorServiceStatistics> managedExecutorService;

    /**
     * Constructor that deals with figuring out what kind of service there is and setting the references accordingly.
     * 
     * @param service the {@link ExecutorServiceStatistics} being exposed.
     */
    StandardExecutorServiceInfo(ExecutorServiceStatistics service) {
        this.managedExecutorService = new WeakReference<ExecutorServiceStatistics>(service);
    }

    /**
     * {@inheritDoc}
     */
    public long getAverageExecutionTime() {
        ExecutorServiceStatistics executorService = this.managedExecutorService.get();
        return executorService == null ? -1 : executorService.getAverageExecutionTime();
    }

    /**
     * {@inheritDoc}
     */
    public long getExecutionTime() {
        ExecutorServiceStatistics executorService = this.managedExecutorService.get();
        return executorService == null ? -1 : executorService.getExecutionTime();
    }

    /**
     * {@inheritDoc}
     */
    public int getPoolSize() {
        ExecutorServiceStatistics executorService = this.managedExecutorService.get();
        return executorService == null ? -1 : executorService.getPoolSize();
    }

    /**
     * {@inheritDoc}
     */
    public int getLargestPoolSize() {
        ExecutorServiceStatistics executorService = this.managedExecutorService.get();
        return executorService == null ? -1 : executorService.getLargestPoolSize();
    }

    /**
     * {@inheritDoc}
     */
    public String getTypeName() {
        ExecutorServiceStatistics executorService = this.managedExecutorService.get();
        if(executorService == null) {
            return null;
        } else {
            return (executorService instanceof ScheduledExecutorService ? ScheduledExecutorService.class.getSimpleName() : ExecutorService.class.getSimpleName());
        }
    }

    /**
     * {@inheritDoc}
     */
    public int getActiveCount() {
        ExecutorServiceStatistics executorService = this.managedExecutorService.get();
        return executorService == null ? -1 : executorService.getActiveCount();
    }

    /**
     * {@inheritDoc}
     */
    public int getMaximumPoolSize() {
        ExecutorServiceStatistics executorService = this.managedExecutorService.get();
        return executorService == null ? -1 : executorService.getMaximumPoolSize();
    }

    /**
     * {@inheritDoc}
     */
    public String getPoolName() {
        ExecutorServiceStatistics executorService = this.managedExecutorService.get();
        return executorService == null ? null : executorService.getPoolName();
    }

}
