/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.concurrent.core;

import com.springsource.server.concurrent.core.accessor.ThreadLocalAccessor;
import com.springsource.server.concurrent.core.accessor.ThreadLocalAccessorFactory;
import com.springsource.server.concurrent.core.accessor.ThreadLocalData;

/**
 * Standard implementation of {@link ThreadManager}.
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
final class StandardThreadManager implements ThreadManager {

    private static final EngineDubThreadLocal DUB_THREAD_LOCAL = new EngineDubThreadLocal();

    private final ThreadLocalAccessor accessor = ThreadLocalAccessorFactory.newThreadLocalAccessor();

    private final Object dubMonitor = new Object();

    /**
     * {@inheritDoc}
     */
    public Thread newThread(Runnable runnable) {
        return makeEngineThread(new Thread(runnable));
    }

    /**
     * {@inheritDoc}
     */
    public Thread newThread(Runnable runnable, String name) {
        return makeEngineThread(new Thread(runnable, name));
    }

    /**
     * {@inheritDoc}
     */
    public void dub(Thread thread) {
        makeEngineThread(thread);
    }

    /**
     * {@inheritDoc}
     */
    public boolean isEngineManagedThread(Thread thread) {
        synchronized (this.dubMonitor) {
            ThreadLocalData[] threadLocals = this.accessor.threadLocals(thread);
            for (ThreadLocalData threadLocalData : threadLocals) {
                if (threadLocalData.getThreadLocal() instanceof EngineDubThreadLocal) {
                    return Boolean.TRUE.equals(threadLocalData.get());
                }
            }
            return false;
        }
    }

    private Thread makeEngineThread(Thread thread) {
        synchronized (this.dubMonitor) {
            this.accessor.setThreadLocalValue(DUB_THREAD_LOCAL, Boolean.TRUE, thread);
        }
        return thread;
    }

    /**
     * Marker {@link ThreadLocal} for Engine-managed threads.
     */
    private static class EngineDubThreadLocal extends ThreadLocal<Boolean> implements PersistentThreadLocal {

        @Override public String toString() {
            return "Engine-managed Thread";
        }
    }
}
