/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.concurrent.core.accessor;

import java.util.concurrent.locks.ReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;

/**
 * Extension to {@link ThreadLocal} that guarantees safe access from multiple threads. Useful when a
 * <code>ThreadLocal</code> is accessed for diagnostic purposes.
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
public class ConcurrentThreadLocal<T> extends ThreadLocal<T> {

    private final ReadWriteLock lock = new ReentrantReadWriteLock();

    /**
     * {@inheritDoc}
     */
    @Override public T get() {
        this.lock.readLock().lock();
        try {
            return super.get();
        } finally {
            this.lock.readLock().unlock();
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override public void remove() {
        this.lock.writeLock().lock();
        try {
            super.remove();
        } finally {
            this.lock.writeLock().unlock();
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override public void set(T value) {
        this.lock.writeLock().lock();
        try {
            super.set(value);
        } finally {
            this.lock.writeLock().unlock();
        }
    }

    /**
     * Gets the lock for this <code>ConcurrentThreadLocal</code>. Used to preserve happens-before when accessed from
     * a {@link ThreadLocalAccessor}.
     * 
     * @return the lock for this <code>ConcurrentThreadLocal</code>/
     */
    ReadWriteLock lock() {
        return this.lock;
    }
}
