/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.concurrent.core.config;

import java.util.concurrent.BlockingQueue;

import org.springframework.beans.factory.FactoryBean;

import com.springsource.server.concurrent.core.queue.WorkQueueFactory;
import com.springsource.server.serviceability.ffdc.Assert;

/**
 * {@link FactoryBean} for creating managed {@link BlockingQueue work queue} instances using the
 * {@link WorkQueueFactory}.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
public final class WorkQueueFactoryBean implements FactoryBean {

    private static final int DEFAULT_QUEUE_SIZE = Integer.MAX_VALUE;

    private final WorkQueueFactory workQueueFactory;

    private final int capacity;

    /**
     * Creates a new <code>WorkQueueFactoryBean</code> for an unbounded queue.
     * 
     * @param workQueueFactory the {@link WorkQueueFactory}.
     */
    public WorkQueueFactoryBean(WorkQueueFactory workQueueFactory) {
        this(workQueueFactory, DEFAULT_QUEUE_SIZE);
    }

    /**
     * Creates a new <code>WorkQueueFactoryBean</code> for an unbounded queue.
     * 
     * @param workQueueFactory
     * @param capacity
     */
    public WorkQueueFactoryBean(WorkQueueFactory workQueueFactory, int capacity) {
        Assert.notNull(workQueueFactory, "WorkQueueFactory must not be null");
        this.workQueueFactory = workQueueFactory;
        this.capacity = capacity;
    }

    /**
     * {@inheritDoc}
     */
    public Object getObject() throws Exception {
        return this.workQueueFactory.newStealingWorkQueue(this.capacity);
    }

    /**
     * {@inheritDoc}
     */
    public Class<?> getObjectType() {
        return BlockingQueue.class;
    }

    /**
     * {@inheritDoc}
     */
    public boolean isSingleton() {
        return true;
    }
}
