/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.concurrent.core.queue;

import java.lang.ref.Reference;
import java.lang.ref.WeakReference;

import com.springsource.server.management.proxy.concurrent.WorkQueueInfo;

/**
 * Standard implementation of {@link WorkQueueInfo}.Maintains a {@link WeakReference} to the underlying
 * {@link WorkQueueInfo} to prevent pinning the executor in memory. Public to allow reflection to work 
 * in management.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Christopher Frost
 * @since 1.0
 */
final public class StandardWorkQueueInfo implements WorkQueueInfo {

    private final Reference<WorkStealingQueue> workStealingQueue;

    private final String name;
    
    private final String type;

    /**
     * Creates a new {@link StandardWorkQueueInfo} for the supplied {@link WorkStealingQueue}.
     * 
     * @param queue the queue.
     * @param type the name of the queue.
     */
    public StandardWorkQueueInfo(WorkStealingQueue queue, String type) {
        this.workStealingQueue = new WeakReference<WorkStealingQueue>(queue);
        this.name = queue.getClass().getSimpleName() + Integer.toHexString(queue.hashCode());
        this.type = type;
    }

    /**
     * {@inheritDoc}
     */
    public long getQueueEnqueueCount() {
        WorkStealingQueue queue = this.workStealingQueue.get();
        return queue == null ? -1 : queue.getEnqueueCount();
    }

    /**
     * {@inheritDoc}
     */
    public int getQueueRemainingCapacity() {
        WorkStealingQueue queue = this.workStealingQueue.get();
        return queue == null ? -1 : queue.remainingCapacity();
    }

    /**
     * {@inheritDoc}
     */
    public int getQueueSize() {
        WorkStealingQueue queue = this.workStealingQueue.get();
        return queue == null ? -1 : queue.size();
    }

    /**
     * {@inheritDoc}
     */
    public long getQueueStealCount() {
        WorkStealingQueue queue = this.workStealingQueue.get();
        return queue == null ? -1 : queue.getStealCount();
    }

    /**
     * {@inheritDoc}
     */
    public long getQueueStolenCount() {
        WorkStealingQueue queue = this.workStealingQueue.get();
        return queue == null ? -1 : queue.getStolenCount();
    }

    /**
     * {@inheritDoc}
     */
    public String getQueueName() {
        return this.name;
    }
    
    /**
     * {@inheritDoc}
     */
    public String getQueueType() {
        return this.type;
    }
}
