/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.concurrent.core.queue;

import java.util.Set;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicLong;

import com.springsource.server.serviceability.output.Level;
import com.springsource.server.serviceability.tracing.EntryExitTraceLevel;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * {@link BlockingQueue} that supports work stealing. When a {@link #take()} or {@link #poll()} call is made against an
 * empty instance, it will attempt to steal work from another managed queue. Work is always stolen from the bottom of a
 * queue. This reduces contention with the queues own pool which is accessing the head of the queue.<p/>
 * 
 * Operations other than <code>take()</code> and <code>poll()</code> do not support stealing. In particular, the
 * {@link #size()} operation reports on the size of the callee only, and not on any queues that it may steal from.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
final class WorkStealingQueue extends DelegatingBlockingDeque<Runnable> implements BlockingQueue<Runnable> {

    private static final Tracer TRACER = TracerFactory.createTracer(WorkStealingQueue.class);

    private final AtomicLong stealCount = new AtomicLong(0);

    private final AtomicLong enqueueCount = new AtomicLong(0);

    private final AtomicLong stolenCount = new AtomicLong(0);

    private final Set<WorkStealingQueue> knownQueues;

    /**
     * Creates a new <code>StealingBlockingQueue</code>.
     * 
     * @param delegate the delegate queue.
     * @param knownQueues the known queues.
     */
    public WorkStealingQueue(BlockingDeque<Runnable> delegate, Set<WorkStealingQueue> knownQueues) {
        super(delegate);
        this.knownQueues = knownQueues;
    }

    /**
     * {@inheritDoc}
     */
    @Override public boolean offer(Runnable o, long timeout, TimeUnit unit) throws InterruptedException {
        boolean result = super.offer(o, timeout, unit);
        if (result) {
            this.enqueueCount.incrementAndGet();
        }
        return result;
    }

    /**
     * {@inheritDoc}
     */
    @Override public boolean offer(Runnable o) {
        boolean result = super.offer(o);
        if (result) {
            this.enqueueCount.incrementAndGet();
        }
        return result;
    }

    /**
     * Poll the this queue for {@link Runnable work}. If no work is available in this queue, attempts to
     * {@link WorkStealingQueue#trySteal() steal work} from another queue. This operation does not block, and will
     * return <code>null</code> if no work is available.
     * 
     * @return the <code>Runnable</code> if available; otherwise <code>null</code>.
     * @see edu.emory.mathcs.backport.java.util.concurrent.BlockingDeque#poll()
     */
    @Override public Runnable poll() {
        Runnable runnable = getDelegate().poll();
        if (runnable == null) {
            runnable = tryStealWorkFromOtherQueues();
        }
        return runnable;
    }

    /**
     * Take some {@link Runnable work} from this queue. If no work is immediately available in this queue, attempts to
     * {@link WorkStealingQueue#trySteal() steal work} from another queue. This operation will block until work is
     * available directly in this queue, or a steal attempt is successful.
     * 
     * @return the <code>Runnable</code> from this queue or another queue
     * @throws InterruptedException
     * @see edu.emory.mathcs.backport.java.util.concurrent.BlockingDeque#take()
     */
    @Override public Runnable take() throws InterruptedException {
        Runnable runnable = poll();
        while (runnable == null) {
            runnable = tryStealWorkFromOtherQueues();
            if (runnable == null) {
                runnable = getDelegate().poll(100, TimeUnit.MILLISECONDS);
            }
        }
        return runnable;
    }

    /**
     * Attempts to steal work from this queue.
     * 
     * @return any stolen work, or <code>null</code> if no work can be stolen.
     */
    @EntryExitTraceLevel(Level.VERBOSE) public Runnable trySteal() {
        Runnable r = getDelegate().pollLast();
        if (r != null) {
            TRACER.traceDebug("Work '%s' was stolen from queue '%s'.", r, this);
            this.stolenCount.incrementAndGet();
        }
        return r;
    }

    /**
     * Attempts to steal work from one of the managed <code>BlockingDeques</code>.
     * 
     * @return the stolen <code>Runnable</code>, or <code>null</code> if there is no work available.
     */
    @EntryExitTraceLevel(Level.VERBOSE) private Runnable tryStealWorkFromOtherQueues() {
        Runnable runnable = null;
        for (WorkStealingQueue queue : this.knownQueues) {
            runnable = queue.trySteal();
            if (runnable != null) {
                TRACER.traceDebug("Queue '%s' stole work '%s' from queue '%s'.", this, runnable, queue);
                this.stealCount.incrementAndGet();
                break;
            }
        }
        return runnable;
    }

    /**
     * Gets the number of items that have been stolen from other queues.
     * 
     * @return the number of items that been stolen from other queues.
     */
    public long getStealCount() {
        return this.stealCount.get();
    }

    /**
     * Gets the number of items that have been stolen from this queue.
     * 
     * @return the number of items that been stolen from this queue.
     */
    public long getStolenCount() {
        return this.stolenCount.get();
    }

    /**
     * Gets the number of items that have been enqueued in this queue.
     * 
     * @return the number of items that have been enqueued in this queue.
     */
    public long getEnqueueCount() {
        return this.enqueueCount.get();
    }
}
