/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.config.core;

import java.util.List;

/**
 * A ConfigNode is a node in a ConfigTree and may be either a scalar node, a map node, or a list node. A scalar node
 * contains a single value. A map node contains a map of key to node. A list node contains a list of nodes.
 * 
 * Implementations of this interface must be thread safe.
 * 
 * @author Glyn Normington
 * @author Sam Brannen
 * @since 1.0
 */
public interface ConfigNode {

    /**
     * Get the node of the tree identified by the specified path from this node. Each element of the path is a key which
     * is used to navigate to the appropriate child node.
     * 
     * @param path a list of keys
     * @return the node identified by the path or null if there is no such node
     */
    ConfigNode getNode(List<String> path);

    /**
     * Get the node of the tree identified by the specified path from this node. Each element of the path is a key which
     * is used to navigate to the appropriate child node.
     * 
     * @param path a list of keys
     * @return the node identified by the path or null if there is no such node
     */
    ConfigNode getNode(String... path);

    /**
     * Add the given node at the given path in the tree.
     * 
     * If there is already a node at the given path, do not modify the tree.
     * 
     * @param path a list of keys identifying where to add the new node
     * @param newNode the node to be added
     * @return true if and only if the node was added successfully
     */
    boolean addNode(List<String> path, ConfigNode newNode);

    /**
     * Return the path of this node including the config. point.
     * 
     * @return a string path identifying this node in the whole config. tree
     */
    String getPath();

    /**
     * Return the path of the given node including the config. point. The given node must be a direct child of this
     * node.
     * 
     * @return a string path identifying the given node in the whole config. tree
     */
    String getPath(ConfigNode childNode);

    /**
     * Return the parent node of this node or null if this node is a config. point.
     * 
     * @return the parent ConfigNode
     */
    ConfigNode getParent();

}
