/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.config.core.internal;

import java.util.Arrays;
import java.util.List;

import com.springsource.json.parser.MapNode;
import com.springsource.server.config.core.ConfigNode;
import com.springsource.server.config.core.ConfigPoint;
import com.springsource.server.config.core.FatalConfigurationException;

/**
 * Implementation of {@link ConfigPoint} backed by JSON.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Thread-safe.
 * 
 * @author Rob Harrop
 * @author Sam Brannen
 * @since 1.0
 */
final class JsonConfigPoint implements ConfigPoint {

    private final String name;

    private final ConfigNode rootConfigNode;

    private final String toString;

    /**
     * Creates a new <code>JsonConfigPoint</code>.
     * 
     * @param name the name of the config point.
     * @param rootNode the root node in the JSON tree.
     */
    public JsonConfigPoint(String name, MapNode rootNode) {
        this.name = name;
        this.rootConfigNode = JsonConfigNodeFactory.createConfigNode(this, rootNode);
        this.toString = rootNode.toString();
    }

    /**
     * {@inheritDoc}
     */
    public String getName() {
        return this.name;
    }

    /**
     * {@inheritDoc}
     */
    public ConfigNode getRootNode() {
        return this.rootConfigNode;
    }

    /**
     * {@inheritDoc}
     */
    public boolean addNode(List<String> path, ConfigNode newNode) {
        return this.rootConfigNode.addNode(path, newNode);
    }

    /**
     * {@inheritDoc}
     */
    public ConfigNode getNode(List<String> path) {
        return this.rootConfigNode.getNode(path);
    }

    /**
     * {@inheritDoc}
     */
    public ConfigNode getNode(String... path) {
        ConfigNode result = null;
        if (path != null && path.length > 0) {
            result = getNode(Arrays.asList(path));
        } else {
            result = this;
        }
        return result;
    }

    /**
     * {@inheritDoc}
     */
    public ConfigNode getParent() {
        return null;
    }

    /**
     * {@inheritDoc}
     */
    public String getPath() {
        return this.name + ":";
    }

    /**
     * {@inheritDoc}
     */
    public String getPath(ConfigNode childNode) {
        if (childNode == this.rootConfigNode) {
            return getPath();
        } else {
            throw new FatalConfigurationException("Invalid child node", null);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String toString() {
        return this.toString;
    }

}
