/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.config.core.internal;

import java.io.File;
import java.io.FilenameFilter;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;

import com.springsource.server.bootstrap.BootstrapConfiguration;
import com.springsource.server.common.math.Range;
import com.springsource.server.common.math.Version;
import com.springsource.server.config.core.ConfigTree;
import com.springsource.server.config.core.ConfigTreeFactory;
import com.springsource.server.config.core.FatalConfigurationException;
import com.springsource.server.config.core.diagnostics.ConfigLogEvent;
import com.springsource.server.config.core.internal.validate.ConfigValidator;
import com.springsource.server.config.core.internal.validate.EngineConfigPointValidatorFactory;
import com.springsource.server.config.core.internal.validate.StandardEngineConfigPointValidatorFactory;
import com.springsource.server.config.core.internal.validate.StandardMapConfigNodeValidator;
import com.springsource.server.config.core.internal.validate.StandardVersionRangeConfigNodeValidator;
import com.springsource.server.kernel.core.SubsystemRegistry;
import com.springsource.server.serviceability.logging.LogService;
import com.springsource.server.serviceability.logging.LogServiceFactory;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * Standard implementation of the {@link ConfigTreeFactory} interface.
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * This class is thread-safe.
 * 
 * @author Glyn Normington
 * @author Sam Brannen
 * @since 1.0
 */
public class StandardConfigTreeFactory implements ConfigTreeFactory {

    private static final String CONFIG_RESOURCE_DIRECTORY = "/META-INF/";

    private static final Tracer TRACER = TracerFactory.createTracer(StandardConfigTreeFactory.class);

    private static final LogService LOGGER = LogServiceFactory.getLogService();

    private static final String BASENAME_VERSION_SEPARATOR = "-";

    public static final String CONFIG_FILEEXTENSION = ".config";

    public static final String CONFIG_SCHEMA_FILEEXTENSION = ".configschema";

    private final EngineConfigPointValidatorFactory configPointValidatorFactory;

    private final BootstrapConfiguration bootstrapConfiguration;

    private final Map<String, ConfigTree> configTreesToDump = new ConcurrentHashMap<String, ConfigTree>();

    private final BundleContext myContext;

    public StandardConfigTreeFactory(BootstrapConfiguration bootstrapConfiguration, SubsystemRegistry subsystemRegistry, BundleContext myContext) {
        this.configPointValidatorFactory = new StandardEngineConfigPointValidatorFactory();
        this.bootstrapConfiguration = bootstrapConfiguration;
        this.myContext = myContext;
    }

    /**
     * {@inheritDoc}
     */
    public ConfigTree createConfigTree(String filename, BundleContext configOwnerContext) {
        Bundle configOwner = configOwnerContext.getBundle();

        Range<Version> versionRange = getPreferredVersionRange(filename, configOwner);

        // Get any user-supplied configuration file.
        URL configURL = getUserSuppliedConfigFile(filename, versionRange);

        /*
         * If no user-supplied configuration file was found, use the owner's default (with no version in the filename).
         * This must be included in (or imported by) the owning bundle.
         */
        if (configURL == null) {
            configURL = configOwner.getResource(CONFIG_RESOURCE_DIRECTORY + filename + CONFIG_FILEEXTENSION);
        }

        // If there is no user-supplied config. file or a default config. file,
        // give up.
        if (configURL == null) {
            throw new FatalConfigurationException("Server configuration file '" + filename + "' not found");
        }

        // Parse the config. file into an abstract syntax tree.
        JsonConfigTree configTree = new JsonConfigTree(configURL);

        for (String cpn : configTree.getConfigPointNames()) {
            ConfigValidator cpValidator = getConfigPointValidator(cpn, configOwner);
            if (cpValidator != null) {
                configTree.setConfigPointValidator(cpn, cpValidator);
            }

            configTree.registerInfoObject(cpn, this.myContext);

            // Following would be better as a debug trace if we can work out how
            // to configure debug trace this early.
            TRACER.traceInfo("Configuration point '%s' is set to '%s'", cpn, configTree.getConfigPoint(cpn));
        }

        if (!configTree.validate()) {
            throw new FatalConfigurationException("Server configuration file '" + configURL + "' has invalid contents");
        } else {
            this.configTreesToDump.put(filename, configTree);
            return configTree;
        }
    }

    private Range<Version> getPreferredVersionRange(String filename, Bundle configOwner) {
        Range<Version> versionRange = null;
        ConfigValidator v = getConfigPointValidator(filename, configOwner);
        if (v instanceof StandardMapConfigNodeValidator) {
            ConfigValidator versionValidator = ((StandardMapConfigNodeValidator) v).getValidator("version");
            if (versionValidator instanceof StandardVersionRangeConfigNodeValidator) {
                versionRange = ((StandardVersionRangeConfigNodeValidator) versionValidator).getVersionRange();
            }
        }
        if (versionRange == null) {
            versionRange = new Range<Version>(Version.MINIMUM_VERSION, false, Version.MINIMUM_VERSION, false);
        }
        return versionRange;
    }

    private ConfigValidator getConfigPointValidator(String configPointName, Bundle configOwner) {
        URL configPointSchemaURL = configOwner.getResource(CONFIG_RESOURCE_DIRECTORY + configPointName + CONFIG_SCHEMA_FILEEXTENSION);
        if (configPointSchemaURL == null) {
            TRACER.traceWarning("Configuration schema '%s' missing from bundle '%s'", configPointName, configOwner);
            return null;
        }
        return this.configPointValidatorFactory.createConfigPointValidator(configPointSchemaURL);
    }

    /**
     * Get a user supplied configuration file.
     * 
     * Search the configuration directories for a configuration file with the specified base name.
     * 
     * @param baseName the base name of the configuration file
     * @param versionRange the preferred version range
     * @return a URL of the configuration file or <code>null</code> if no file was found
     * @see #getUserSuppliedConfigFile(File, String, Range)
     */
    private URL getUserSuppliedConfigFile(final String baseName, Range<Version> versionRange) {
        for (File configDir : getConfigDirectories()) {
            URL configURL = getUserSuppliedConfigFile(configDir, baseName, versionRange);
            if (configURL != null) {
                return configURL;
            }
        }
        return null;
    }

    /**
     * Get a user supplied configuration file.
     * 
     * Search the supplied configuration directory for a configuration file with the specified base name.
     * 
     * @param configDir the directory in which to search for the configuration file
     * @param baseName the base name of the configuration file
     * @param versionRange the preferred version range
     * @return a URL of the configuration file or <code>null</code> if no file was found
     */
    private URL getUserSuppliedConfigFile(final File configDir, final String baseName, Range<Version> versionRange) {
        URL configURL = null;

        if (configDir != null) {
            String[] candidates = configDir.list(new FilenameFilter() {

                public boolean accept(File dir, String name) {
                    return name.startsWith(baseName) && name.endsWith(CONFIG_FILEEXTENSION);
                }
            });

            String configFile = selectCandidate(baseName, versionRange, candidates);

            if (configFile != null) {
                File file = new File(configDir.getAbsolutePath() + File.separatorChar + configFile);
                if (file.canRead()) {
                    TRACER.traceInfo("Selected configuration file '%s' for '%s' configuration", file.getAbsolutePath(), baseName);
                    try {
                        configURL = file.toURL();
                    } catch (MalformedURLException e) {
                        throw new FatalConfigurationException("Bad URL created for configuration file", e);
                    }
                } else {
                    throw new FatalConfigurationException("Cannot read Server configuration file '" + file.getAbsolutePath() + "'");
                }
            }
        }
        return configURL;
    }

    /**
     * Select a configuration file with a given base name from an array of candidates all of which must start with the
     * given base name end with CONFIG_FILE_EXTENSION.
     * 
     * The filename containing the highest version in the given range is preferred or, if there is no such file, the
     * filename with no version, or if there is no such file, the highest version filename in the directory (as
     * defaulting to the owning bundle's configuration file might wrongly ignore a user-supplied configuration file).
     * 
     * Note that this method is declared to be package private simply so that it can be tested properly. It is not
     * intended to be called from outside this class.
     * 
     * @param baseName the base name of the configuration file
     * @param versionRange the preferred version range
     * @param candidates the candidates from the configuration directory
     * @return the chosen candidate filename or null if none was chosen
     */
    String selectCandidate(final String baseName, Range<Version> versionRange, String[] candidates) {
        if (candidates.length == 0) {
            return null;
        }
        String highestCandidate = null;
        Version highestVersion = Version.MINIMUM_VERSION;
        String highestInRangeCandidate = null;
        Version highestVersionInRange = Version.MINIMUM_VERSION;
        for (String c : candidates) {
            // Extract version from filename
            Version v = null;
            String delimitedVersionPortion = c.substring(baseName.length(), c.length() - CONFIG_FILEEXTENSION.length());
            String versionPortion;
            if (delimitedVersionPortion.startsWith(BASENAME_VERSION_SEPARATOR)) {
                versionPortion = delimitedVersionPortion.substring(BASENAME_VERSION_SEPARATOR.length());
                if (versionPortion.length() != 0) {
                    try {
                        v = new Version(versionPortion);
                    } catch (Exception e) {
                    }
                }
            } else if (delimitedVersionPortion.length() == 0) {
                v = Version.MINIMUM_VERSION;
            }

            // If the version is valid, compare it to current best candidates.
            if (v != null) {
                /*
                 * Keep track of highest version and its filename. If there is more than one filename with the highest
                 * version, use the last one in the directory list.
                 */
                if (v.compareTo(highestVersion) >= 0) {
                    highestVersion = v;
                    highestCandidate = c;
                }

                /*
                 * Keep track of highest version in the preferred range and its filename. If there is more than one
                 * filename with the highest version, use the last one in the directory list.
                 */
                if (versionRange.contains(v)) {
                    if (v.compareTo(highestVersionInRange) >= 0) {
                        highestVersionInRange = v;
                        highestInRangeCandidate = c;
                    }
                }
            } else {
                LOGGER.log(ConfigLogEvent.INVALID_CONFIG_FILENAME, c);
            }
        }
        return highestInRangeCandidate != null ? highestInRangeCandidate : highestCandidate;
    }

    /**
     * Get the list of the Server's configuration directories which hold the Server's configuration files.
     * 
     * The configuration path is either specified using a system property or defaults to the current directory.
     * 
     * @return the configuration directories; may be <em>empty</em> if no directories were found.
     */
    private List<File> getConfigDirectories() {
        final List<File> configDirectories = new ArrayList<File>(1);
        String[] configPaths = this.bootstrapConfiguration.getConfigPaths();

        for (String configDirName : configPaths) {
            File configDir = new File(configDirName);
            if (!configDir.isAbsolute()) {
                configDir = new File(this.bootstrapConfiguration.getHomeDirectory(), configDirName);
            }
            if (!configDir.isDirectory()) {
                LOGGER.log(ConfigLogEvent.INVALID_CONFIG_DIRECTORY, configDir.getAbsolutePath());
            } else {
                configDirectories.add(configDir);
            }
        }

        return configDirectories;
    }

    public Map<String, ConfigTree> getConfigTreesMap() {
        return this.configTreesToDump;
    }

}
