/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.config.core.internal;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import com.springsource.server.config.core.ConfigNode;
import com.springsource.server.config.core.FatalConfigurationException;
import com.springsource.server.config.core.ListConfigNode;

/**
 * Standard implementation of {@link ListConfigNode}. <p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Glyn Normington
 * @since 1.0
 */
public final class StandardListConfigNode extends AbstractConfigNode implements ListConfigNode {

    private final List<ConfigNode> nodeList = Collections.synchronizedList(new ArrayList<ConfigNode>());

    /**
     * Creates a new <code>StandardListConfigNode</code>.
     * 
     * @param parent the parent {@link ConfigNode}.
     */
    public StandardListConfigNode(ConfigNode parent) {
        super(parent);
    }

    /**
     * {@inheritDoc}
     */
    public List<ConfigNode> getNodes() {
        return Collections.unmodifiableList(this.nodeList);
    }

    /**
     * Adds the supplied {@link ConfigNode} to this {@link ListConfigNode}.
     * 
     * @param newNode the child node.
     * @return <code>true</code> if the add succeeds.
     */
    public boolean addNode(ConfigNode newNode) {
        if (newNode.getParent() != this) {
            throw new FatalConfigurationException("Cannot add node to a list node which is not a child of the list node", null);
        }
        return this.nodeList.add(newNode);
    }

    /**
     * {@inheritDoc}
     */
    public boolean addNode(List<String> path, ConfigNode newNode) {
        return false;
    }

    /**
     * {@inheritDoc}
     */
    public ConfigNode getNode(List<String> path) {
        ConfigNode result = null;
        if (!path.isEmpty()) {
            String head = head(path);
            try {
                int index = Integer.valueOf(head).intValue();

                if (index >= 0 && index < this.nodeList.size()) {
                    result = this.nodeList.get(index).getNode(tail(path));
                }
            } catch (NumberFormatException e) {
                // TODO: trace e and ignore
            }
        } else {
            result = this;
        }
        return result;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String toString() {
        StringBuffer result = new StringBuffer();
        result.append('[');
        boolean first = true;
        for (ConfigNode n : this.nodeList) {
            if (first == false) {
                result.append(", ");
            }
            result.append(n.toString());
            first = false;
        }
        result.append(']');
        return result.toString();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected String doGetPath(ConfigNode childNode) {
        StringBuffer path = new StringBuffer();
        path.append(getPath());
        if (getParent().getParent() != null) {
            path.append(',');
        }
        for (int i = 0; i < this.nodeList.size(); i++) {
            if (this.nodeList.get(i) == childNode) {
                path.append(i);
                return path.toString();
            }
        }
        throw new FatalConfigurationException("Child node not found in parent list", null);
    }
}
