/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.config.core.internal.validate;

import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import com.springsource.server.config.core.ConfigNode;
import com.springsource.server.config.core.MapConfigNode;
import com.springsource.server.config.core.diagnostics.ConfigLogEvent;
import com.springsource.server.serviceability.logging.LogService;
import com.springsource.server.serviceability.logging.LogServiceFactory;

/**
 * {@link StandardMapConfigNodeValidator} is a standard implementation of {@link ConfigValidator} for maps.
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * This class is thread safe provided he validators in the range of the validator map are thread safe.
 * 
 * @author Glyn Normington
 * @since 1.0
 */
public final class StandardMapConfigNodeValidator implements ConfigValidator {

    private static final LogService LOGGER = LogServiceFactory.getLogService();

    private final Map<String, ConfigValidator> validatorMap;

    /**
     * Create a validator for maps from the given map of validators.
     * 
     * @param validatorMap the map of validators to be used to validate maps
     */
    public StandardMapConfigNodeValidator(Map<String, ConfigValidator> validatorMap) {
        this.validatorMap = validatorMap;
    }

    /**
     * {@inheritDoc}
     */
    public boolean validate(ConfigNode configNode) {
        boolean result = true;
        try {
            MapConfigNode map = (MapConfigNode) configNode;
            if (!map.getKeys().equals(this.validatorMap.keySet())) {
                {
                    Set<String> ks = new HashSet<String>(map.getKeys().size());
                    ks.addAll(map.getKeys());
                    ks.removeAll(this.validatorMap.keySet());
                    for (String k : ks) {
                        LOGGER.log(ConfigLogEvent.EXTRA_MAPLET, map.getPath(), k);
                    }
                }
                {
                    Set<String> ks = new HashSet<String>(this.validatorMap.keySet().size());
                    ks.addAll(this.validatorMap.keySet());
                    ks.removeAll(map.getKeys());
                    for (String k : ks) {
                        LOGGER.log(ConfigLogEvent.MISSING_MAPLET, map.getPath(), k);
                    }
                }
                result = false; // but continue to validate any elements with valid keys
            }
            for (String k : this.validatorMap.keySet()) {
                if (map.getNode(k) != null && !getValidator(k).validate(map.getNode(k))) {
                    result = false; // but continue validating any remaining elements
                }
            }
        } catch (ClassCastException e) {
            LOGGER.log(ConfigLogEvent.INVALID_MAP, configNode.getPath());
            result = false;
        }
        return result;
    }

    public ConfigValidator getValidator(String key) {
        return this.validatorMap.get(key);
    }
}
