/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.control.core.config;

import java.util.Map;

import org.springframework.beans.factory.DisposableBean;
import org.springframework.beans.factory.InitializingBean;

import com.springsource.server.control.core.ControlExporter;
import com.springsource.server.control.core.ControlHandle;

/**
 * An {@link InitializingBean} for creating and exporting controls. Controls
 * that are exported using an instance of  this bean will be automatically
 * retracted when the instance is disposed.
 * </p>
 * <strong>Concurrent Semantics</strong><br />
 * Thread-safe.
 *
 * @author Andy Wilkinson
 * @since 1.0
 */
public class ControlInitializingBean implements InitializingBean, DisposableBean {
    
    private final ControlExporter controlExporter;
    
    private final Object implementation;
    private final Class<?> definition;
    private final Map<String, String> properties;
    
    private volatile ControlHandle controlHandle;
    
    /**
     * Creates and exports a new control using the given control exporter. The control with be backed by
     * the given implementation and have its interface defined by the given definition.
     * @param controlExporter the <code>ControlExporter</code> with which the new control will be exported.
     * @param implementation the control's implementation.
     * @param definition the control's definition.
     */
    public ControlInitializingBean(ControlExporter controlExporter, Object implementation, Class<?> definition) {
        this(controlExporter, implementation, definition, null);
    }
    
    /**
     * Creates and exports a new control using the given control exporter. The control with be backed by
     * the given implementation and have its interface defined by the given definition. The given properties
     * will be associated with the exported control and should be used to distinguish the control from others
     * that have an identical definition.
     * 
     * @param controlExporter the <code>ControlExporter</code> with which the new control will be exported.
     * @param implementation the control's implementation.
     * @param definition the control's definition.
     * @param properties the control's distinguishing properties
     */
    public ControlInitializingBean(ControlExporter controlExporter, Object implementation, Class<?> definition, Map<String, String> properties) {
        this.controlExporter = controlExporter;
        this.implementation = implementation;
        this.definition = definition;
        this.properties = properties;
    }

    /** 
     * {@inheritDoc}
     */
    public void afterPropertiesSet() throws Exception {
        this.controlHandle = controlExporter.exportControl(implementation, definition, properties);       
    }

    /** 
     * {@inheritDoc}
     */
    public void destroy() throws Exception {
        this.controlHandle.retractControl();      
    }
}
