/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.deployer.app;

import java.util.ArrayList;
import java.util.List;

import com.springsource.server.deployer.core.Module;
import com.springsource.server.deployer.core.ModuleImportExpander;
import com.springsource.server.osgi.framework.ImportExpander;
import com.springsource.server.osgi.framework.ImportMergeException;
import com.springsource.server.osgi.framework.UnableToSatisfyDependenciesException;
import com.springsource.server.osgi.framework.ImportExpander.ImportPromotionVector;
import com.springsource.server.osgi.manifest.BundleManifest;
import com.springsource.server.osgi.manifest.Scoper;

/**
 * {@link ApplicationImportExpander} keeps track of the {@link ImportPromotionVector} for an application.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * This class is thread safe.
 * 
 * @author Glyn Normington
 * @since 1.0
 */
public final class ApplicationImportExpander implements ModuleImportExpander {

    private final List<Module> modules;

    private final ImportPromotionVector importPromotionVector;

    private final ImportExpander importExpander;

    private final Object monitor = new Object();

    public ApplicationImportExpander(List<Module> modules, ImportExpander importExpander) throws ImportMergeException,
        UnableToSatisfyDependenciesException {
        this.modules = modules;
        this.importExpander = importExpander;
        this.importPromotionVector = this.importExpander.expandImports(getBundleManifests());
        for (Module module : this.modules) {
            module.setImportExpander(this);
        }
    }

    /**
     * Get the bundle manifests corresponding to the modules of the application.
     * 
     * @param modules
     * @return
     */
    private List<BundleManifest> getBundleManifests() {
        List<BundleManifest> bundleManifests = new ArrayList<BundleManifest>();
        for (Module module : this.modules) {
            bundleManifests.add(module.getModuleData().getBundleManifest());
        }
        return bundleManifests;
    }

    /**
     * {@inheritDoc}
     */
    public boolean expandImports(Module module, Scoper scoper) throws UnableToSatisfyDependenciesException, ImportMergeException {
        synchronized (this.monitor) {
            return this.importExpander.expandOnRefresh(getBundleManifests(), module.getUnscopedSymbolicName(), this.importPromotionVector, scoper);
        }
    }

}
