/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.deployer.app;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import com.springsource.server.deployer.core.DeployerLogEvent;
import com.springsource.server.deployer.core.DeploymentException;
import com.springsource.server.deployer.core.FatalDeploymentException;
import com.springsource.server.deployer.core.Module;
import com.springsource.server.deployer.core.ModuleData;
import com.springsource.server.deployer.core.ModuleScoper;
import com.springsource.server.osgi.manifest.BundleManifest;
import com.springsource.server.osgi.manifest.BundleManifestFactory;
import com.springsource.server.osgi.manifest.Scoper;
import com.springsource.server.osgi.manifest.Scoper.DuplicateBundleSymbolicNameException;
import com.springsource.server.osgi.manifest.Scoper.DuplicateExportException;
import com.springsource.server.osgi.manifest.Scoper.UnsupportedBundleManifestVersionException;
import com.springsource.server.serviceability.logging.LogService;
import com.springsource.server.serviceability.logging.LogServiceFactory;

/**
 * {@link ApplicationScoper} provides scoping of bundle manifests and Spring
 * services for {@link MultiBundleApplication MultiBundleApplications}.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * This class is not thread safe.
 * 
 * @author Glyn Normington
 * @since 1.0
 */
final class ApplicationScoper implements ModuleScoper {

	private static final LogService LOGGER = LogServiceFactory.getLogService();

	private final Scoper scoper;

	private final ServiceScoper serviceScoper;

	private final Collection<Module> modules;

	private final String scopeName;

	private final org.osgi.framework.Version scopeVersion;

	/**
	 * Construct a new application scoper.
	 * 
	 * @param modules
	 * @param osgiApplicationDescriptor
	 */
	ApplicationScoper(Collection<Module> modules,
			MultiBundleApplicationDescriptor osgiApplicationDescriptor, BundleManifestFactory bundleManifestFactory) {
		this.modules = modules;
		this.scopeName = osgiApplicationDescriptor.getSymbolicName();
		this.scopeVersion = new org.osgi.framework.Version(
				osgiApplicationDescriptor.getVersion().toString());
		this.scoper = new Scoper(getBundleManifests(), this.scopeName,
				this.scopeVersion, bundleManifestFactory);
		this.serviceScoper = new ServiceScoper(getModuleData(), this.scopeName,
				this.scopeVersion);
		for (Module module : modules) {
			module.setScoper(this);
		}
	}

	private List<BundleManifest> getBundleManifests() {
		List<BundleManifest> bundleManifests = new ArrayList<BundleManifest>();
		for (Module module : this.modules) {
			bundleManifests.add(module.getModuleData().getBundleManifest());
		}
		return bundleManifests;
	}

	private Set<ModuleData> getModuleData() {
		Set<ModuleData> moduleData = new HashSet<ModuleData>();
		for (Module module : this.modules) {
			moduleData.add(module.getModuleData());
		}
		return moduleData;
	}

	/**
	 * Scope the application.
	 * @throws DeploymentException 
	 */
	void scope() throws DeploymentException {
		try {
			// Transform the modules' bundle manifests to scope the OSGi
			// application.
			this.scoper.scope();
		} catch (UnsupportedBundleManifestVersionException ubmve) {
			// This represents a failure to upgrade the manifest.
			throw new FatalDeploymentException(
					"Cannot scope a bundle which does not specify a bundle manifest version of at least "
							+ ubmve.getLowestSupportedVersion());
		} catch (DuplicateBundleSymbolicNameException dbsne) {
		    LOGGER.log(DeployerLogEvent.DUPLICATE_BSN_IN_SCOPE, dbsne, this.scopeName, this.scopeVersion, dbsne.getBundleSymbolicName());
			throw new DeploymentException(
					"More than one bundle in scope '" + this.scopeName
							+ "' version '" + this.scopeVersion
							+ "' has bundle symbolic name '"
							+ dbsne.getBundleSymbolicName() + "'");
		} catch (DuplicateExportException dee) {
			String packageName = dee.getPackageName();
			String exporters = dee.getExporters();
			LOGGER.log(DeployerLogEvent.DUPLICATE_PACKAGE_DURING_SCOPING,
					this.scopeName, this.scopeVersion, packageName, exporters
							.toString());
			throw new DeploymentException("Package '" + packageName
					+ "' exported by more than one bundle ["
					+ exporters.toString() + "] in scope '" + this.scopeName
					+ "' version '" + this.scopeVersion + "'");
		}
		this.serviceScoper.scope();
	}

	/**
	 * @throws DeploymentException 
	 * {@inheritDoc}
	 */
	public void rescope(Module module) throws DeploymentException {
		try {
			this.scoper.rescope(module.getModuleData().getBundleManifest());
		} catch (UnsupportedBundleManifestVersionException ubmve) {
			// This is a failure to upgrade the manifest.
			throw new FatalDeploymentException(
					"Cannot scope a bundle which does not specify a bundle manifest version of at least "
							+ ubmve.getLowestSupportedVersion());
		} catch (DuplicateBundleSymbolicNameException dbsne) {
			LOGGER.log(DeployerLogEvent.DUPLICATE_BSN_IN_SCOPE, dbsne, this.scopeName, this.scopeVersion, dbsne.getBundleSymbolicName());
			throw new DeploymentException(
					"More than one bundle in scope '" + this.scopeName
							+ "' version '" + this.scopeVersion
							+ "' has bundle symbolic name '"
							+ dbsne.getBundleSymbolicName() + "'");
		} catch (DuplicateExportException dee) {
			String packageName = dee.getPackageName();
			String exporters = dee.getExporters();
			LOGGER.log(DeployerLogEvent.DUPLICATE_PACKAGE_DURING_SCOPING,
					this.scopeName, this.scopeVersion, packageName, exporters
							.toString());
			throw new DeploymentException("Package '" + packageName
					+ "' exported by more than one bundle ["
					+ exporters.toString() + "] in scope '" + this.scopeName
					+ "' version '" + this.scopeVersion + "'");
		}
		this.serviceScoper.rescope(module.getModuleData());
	}

    /** 
     * {@inheritDoc}
     */
    public Scoper getScoper() {
        return this.scoper;
    }

}
