/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.deployer.app;

import java.io.IOException;
import java.io.Reader;
import java.net.URI;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.osgi.framework.Bundle;

import com.springsource.server.deployer.core.CompletableDeploymentContext;
import com.springsource.server.deployer.core.ContextModuleFactory;
import com.springsource.server.deployer.core.DeployableFactory;
import com.springsource.server.deployer.core.DeployerLogEvent;
import com.springsource.server.deployer.core.DeploymentArtefact;
import com.springsource.server.deployer.core.DeploymentContextFactory;
import com.springsource.server.deployer.core.DeploymentException;
import com.springsource.server.deployer.core.FatalDeploymentException;
import com.springsource.server.deployer.core.Module;
import com.springsource.server.deployer.core.ModuleDeployer;
import com.springsource.server.deployer.core.ModuleDeployerSet;
import com.springsource.server.deployer.core.ModuleFactory;
import com.springsource.server.deployer.core.ModuleInfo;
import com.springsource.server.deployer.core.ApplicationDeployer.DeploymentOptions;
import com.springsource.server.deployer.core.DeploymentArtefact.Entry;
import com.springsource.server.deployer.core.internal.AbstractInformedApplicationDeployable;
import com.springsource.server.deployer.core.internal.ExplodedDirectoryDeploymentArtefact;
import com.springsource.server.deployer.core.parse.DeployerBundleManifest;
import com.springsource.server.osgi.framework.BundleInstaller;
import com.springsource.server.osgi.framework.ImportMergeException;
import com.springsource.server.osgi.framework.OsgiFramework;
import com.springsource.server.osgi.framework.UnableToSatisfyDependenciesException;
import com.springsource.server.osgi.manifest.BundleManifestFactory;
import com.springsource.server.osgi.manifest.parse.FatalParseException;
import com.springsource.server.serviceability.ffdc.Assert;
import com.springsource.server.serviceability.logging.LogService;
import com.springsource.server.serviceability.logging.LogServiceFactory;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;
import com.springsource.server.serviceability.tracing.TracingService;

/**
 * This class represents a deployable OSGi application.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * This class is not thread safe and so depends on the synchronization provided by StandardApplicationDeployer.
 * 
 * @author Glyn Normington
 * @author Andy Wilkinson
 * @since 1.0
 */
final class MultiBundleApplication extends AbstractInformedApplicationDeployable {

    private static final String MANIFEST_PATH = "/META-INF/MANIFEST.MF";

    private static final LogService LOGGER = LogServiceFactory.getLogService();

    private static final Tracer TRACER = TracerFactory.createTracer(MultiBundleApplication.class);

    private final OsgiFramework osgi;

    private final DeploymentArtefact applicationDeploymentArtefact;

    private final MultiBundleApplicationDescriptor osgiApplicationDescriptor;

    private final List<Module> modules;

    private final ModuleDeployerSet moduleDeployerSet;

    private final List<ModuleDeployer> personalityDeployers = new ArrayList<ModuleDeployer>(1);

    private final ModuleFactory moduleFactory;

    private final ContextModuleFactory contextFactory;

    private final BundleInstaller bundleInstaller;

    private final DeployableFactory libraryFactory;

    private final DeploymentContextFactory deploymentContextFactory;

    private CompletableDeploymentContext deploymentContext;

    private final BundleManifestFactory bundleManifestFactory;

    public MultiBundleApplication(OsgiFramework osgi, URI location, URI sourceLocation, DeploymentOptions deploymentOptions,
        ModuleDeployerSet moduleDeployerSet, ModuleFactory moduleFactory, ContextModuleFactory contextFactory, BundleInstaller bundleInstaller,
        DeployableFactory libraryFactory, DeploymentContextFactory deploymentContextFactory, BundleManifestFactory bundleManifestFactory,
        TracingService tracingService) throws DeploymentException, UnableToSatisfyDependenciesException {
        super(sourceLocation, Type.MULTI_BUNDLE_APPLICATION, deploymentOptions, osgi, tracingService);
        this.osgi = osgi;
        this.moduleFactory = moduleFactory;
        this.contextFactory = contextFactory;
        this.bundleInstaller = bundleInstaller;
        this.libraryFactory = libraryFactory;
        this.deploymentContextFactory = deploymentContextFactory;
        this.bundleManifestFactory = bundleManifestFactory;
        this.applicationDeploymentArtefact = new ExplodedDirectoryDeploymentArtefact(location);
        Reader manifestReader = createManifestReader();
        DeployerBundleManifest manifest;
        try {
            manifest = new DeployerBundleManifest(manifestReader);
        } catch (FatalParseException fpe) {
            LOGGER.log(DeployerLogEvent.APPLICATION_DESCRIPTOR_PARSE_ERROR, fpe);
            throw new DeploymentException("Error parsing application manifest", fpe);
        }
        super.setApplicationManifest(manifest);
        this.osgiApplicationDescriptor = new MultiBundleApplicationDescriptorParser().parse(manifest);
        this.setName(this.osgiApplicationDescriptor.getSymbolicName());
        this.setVersion(this.osgiApplicationDescriptor.getVersion());
        try {
            manifestReader.close();
        } catch (IOException ioe) {
            throw new RuntimeException("Failed to close manifest reader", ioe);
        }
        this.deploymentContext = this.deploymentContextFactory.create(this);
        this.moduleDeployerSet = moduleDeployerSet;

        this.modules = getModules(this.osgi, this.moduleFactory, this.applicationDeploymentArtefact);

        for (Module module : this.modules) {
            determineModulePersonality(module);
        }
    }

    /**
     * Creates a {@link Reader} for the application's manifest.
     * 
     * @return the manifest <code>Reader</code>.
     * @throws DeploymentException if the manifest cannot be found.
     */
    private Reader createManifestReader() throws DeploymentException {
        Reader manifestReader = null;
        try {
            manifestReader = this.applicationDeploymentArtefact.getEntry(MANIFEST_PATH).getReader();
        } catch (FatalDeploymentException ex) {
            LOGGER.log(DeployerLogEvent.APP_MANIFEST_NOT_FOUND, this.applicationDeploymentArtefact.getFilename());
            throw new DeploymentException("Unable to get application manifest for artefact '" + this.applicationDeploymentArtefact + "'", ex);
        }
        return manifestReader;
    }

    private List<Module> getModules(OsgiFramework osgi, ModuleFactory moduleFactory, DeploymentArtefact applicationDeploymentArtefact)
        throws DeploymentException, UnableToSatisfyDependenciesException {
        List<Module> modules = new ArrayList<Module>();
        Set<Entry> contents = applicationDeploymentArtefact.getContents();
        boolean hasContent = false;
        for (Entry entry : contents) {
            String entryName = entry.getName();
            if (entryName.endsWith(".jar") || entryName.endsWith(".war")) {
                modules.add(moduleFactory.create(this, entry.getDeploymentArtefact().getURI()));
                hasContent = true;
            } else if (entryName.endsWith(".lib") || entryName.endsWith(".libd")) {
                this.libraryFactory.create(entry.getDeploymentArtefact().getURI(), getSourceUri(), getDeploymentOptions()).deploy();
            }
        }
        // If there are no nested contents, treat the application file as its
        // own module.
        if (!hasContent) {
            modules.add(moduleFactory.create(this, applicationDeploymentArtefact.getURI()));
        }
        return modules;
    }

    /**
     * {@inheritDoc}
     */
    @Override public void deploy() throws DeploymentException, UnableToSatisfyDependenciesException {
        try {
            createSyntheticContextModule();

            expandImports();

            for (Module module : this.modules) {
                module.transform();
            }

            scope();

            for (Module module : this.modules) {
                module.install();
            }

            checkModuleConstraints(true);

            super.deploy();

            for (Module module : this.modules) {
                module.deploy();
            }

            for (Module module : this.modules) {
                module.checkDeployed();
            }

            super.deployed();

            this.deploymentContext.completed();
        } catch (DeploymentException e) {
            undeployAfterFailedDeploy();
            throw e;
        } catch (UnableToSatisfyDependenciesException utsde) {
            undeployAfterFailedDeploy();
            throw utsde;
        } catch (RuntimeException e) {
            undeployAfterFailedDeploy();
            throw e;
        }

    }

    /**
     * Expand any Import-Library and Import-Bundle headers in this application's bundles into package imports and
     * propagate promoted imports across the application's bundles.
     * 
     * @throws UnableToSatisfyDependenciesException
     * @throws DeploymentException
     */
    private void expandImports() throws UnableToSatisfyDependenciesException, DeploymentException {
        try {
            new ApplicationImportExpander(this.modules, this.bundleInstaller.getImportExpander());
        } catch (ImportMergeException e) {
            LOGGER.log(DeployerLogEvent.CONFLICTING_PACKAGE_EXPORTS, this.osgiApplicationDescriptor.getSymbolicName(),
                this.osgiApplicationDescriptor.getVersion(), e.getConflictingPackageName(), e.getSources());
            throw new DeploymentException("Clashing exports of package '" + e.getConflictingPackageName() + "' while expanding imports: "
                + e.getSources(), e);
        }
    }

    private void undeployAfterFailedDeploy() {
        try {
            undeploy();
        } catch (Exception ex) {
            TRACER.traceWarning("ApplicationUndeploying failed deployment for application '%s' at version '%s' was not successful.",
                this.osgiApplicationDescriptor.getSymbolicName(), this.osgiApplicationDescriptor.getVersion());
        }
    }

    /**
     * @throws DeploymentException
     * 
     */
    private void createSyntheticContextModule() throws DeploymentException {
        Module module = this.contextFactory.createContext(this.modules, this);
        determineModulePersonality(module);
        this.modules.add(0, module);
    }

    /**
     * Scope the manifests and Spring services of all the modules associated with the application.
     * 
     * @throws DeploymentException
     */
    private void scope() throws DeploymentException {
        new ApplicationScoper(this.modules, this.osgiApplicationDescriptor, this.bundleManifestFactory).scope();
    }

    /**
     * Check that the modules will resolve properly and optionally diagnose any unsatisfiable constraints.
     * 
     * @param diagnose <code>true</code> if and only if failures are to be reported
     * @return <code>true</code> if and only if all module constraints are satisfiable
     * @throws DeploymentException if a constraint cannot be satisfied, but only in certain obscure cases
     * @throws UnableToSatisfyDependenciesException if a constraint cannot be satisfied
     */
    private boolean checkModuleConstraints(boolean diagnose) throws DeploymentException, UnableToSatisfyDependenciesException {
        try {
            this.bundleInstaller.satisfyConstraints(getBundleArray());
            return true;
        } catch (RuntimeException e) {
            /*
             * A RuntimeException, such as a NPE, may be thrown in some extreme cases.
             */
            if (diagnose) {
                LOGGER.log(DeployerLogEvent.APPLICATION_RESOLUTION_ERROR, this.osgiApplicationDescriptor.getSymbolicName(),
                    this.osgiApplicationDescriptor.getVersion().toString(), e.getLocalizedMessage());
                throw new DeploymentException("The modules of application '" + this.osgiApplicationDescriptor.getSymbolicName() + "' version '"
                    + this.osgiApplicationDescriptor.getVersion().toString() + "' cannot be resolved, reason '" + e.getLocalizedMessage() + "'", e);
            }
        } catch (UnableToSatisfyDependenciesException e) {
            if (diagnose) {
                throw e;
            }
        }
        return false;
    }

    /**
     * @return
     */
    private Bundle[] getBundleArray() {
        Bundle[] bundlesArray;
        List<Bundle> bundles = new ArrayList<Bundle>();
        for (Module module : this.modules) {
            bundles.add(module.getBundle());
        }
        bundlesArray = bundles.toArray(new Bundle[bundles.size()]);
        return bundlesArray;
    }

    /**
     * Determine the personality of the given module.
     * 
     * @param module the module whose personality is to be determined
     */
    private void determineModulePersonality(Module module) {
        this.personalityDeployers.add(this.moduleDeployerSet.determinePersonality(this.deploymentContext, module));
    }

    /**
     * {@inheritDoc}
     */
    public void start() throws DeploymentException {
        for (Module module : this.modules) {
            module.start();
        }
    }

    /**
     * {@inheritDoc}
     */
    public void stop() throws DeploymentException {
        for (Module module : this.modules) {
            module.stop();
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override public void undeploy() throws DeploymentException {
        DeploymentException firstFailure = null;
        for (int i = this.modules.size() - 1; i >= 0; i--) {
            try {
                this.modules.get(i).undeploy();
            } catch (DeploymentException e) {
                firstFailure = e;
            }
        }
        super.undeploy();
        if (firstFailure != null) {
            throw firstFailure;
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override public Set<ModuleInfo> getModuleInfo() {
        Set<ModuleInfo> result = new HashSet<ModuleInfo>(this.modules.size());
        result.addAll(this.modules);
        return result;
    }

    /**
     * {@inheritDoc}
     */
    @Override public String toString() {
        return String.format("%s - %s, %s, %s", getClass().getSimpleName(), this.osgiApplicationDescriptor.getSymbolicName(),
            this.osgiApplicationDescriptor.getVersion().toString(), this.applicationDeploymentArtefact.getURI().toString());
    }

    /**
     * {@inheritDoc}
     */
    public boolean refresh(String symbolicName, ModuleDeployerSet moduleDeployerSet, CompletableDeploymentContext deploymentContext)
        throws DeploymentException {
        Assert.isNull(moduleDeployerSet, "non-null moduleDeployerSet");
        Assert.isNull(deploymentContext, "non-null deploymentContext");
        boolean found = false;
        boolean result = false;
        if (symbolicName != null) {
            for (Module module : this.modules) {
                if (module.getUnscopedSymbolicName().equals(symbolicName)) {
                    result = module.refresh(symbolicName, this.moduleDeployerSet, this.deploymentContextFactory.create(this));
                    found = true;
                }
            }
            if (!found) {
                LOGGER.log(DeployerLogEvent.MODULE_NOT_FOUND, symbolicName, this);
                throw new DeploymentException("No module with bundle symbolic name '" + symbolicName + "' was found in application '" + this + "'");
            }
        }
        return result;
    }

    /**
     * {@inheritDoc}
     */
    public void updateResource(String symbolicName, URI resourceLocation, String targetPath) throws DeploymentException {
        boolean found = false;
        for (Module module : this.modules) {
            if (module.getUnscopedSymbolicName().equals(symbolicName)) {
                module.updateResource(symbolicName, resourceLocation, targetPath);
                found = true;
            }
        }
        if (!found) {
            LOGGER.log(DeployerLogEvent.MODULE_NOT_FOUND, symbolicName, this);
            throw new DeploymentException("No module with bundle symbolic name '" + symbolicName + "' was found in application '" + this + "'");
        }
    }

    /**
     * {@inheritDoc}
     */
    public void deleteResource(String symbolicName, String targetPath) throws DeploymentException {
        boolean found = false;
        for (Module module : this.modules) {
            if (module.getUnscopedSymbolicName().equals(symbolicName)) {
                module.deleteResource(symbolicName, targetPath);
                found = true;
            }
        }
        if (!found) {
            LOGGER.log(DeployerLogEvent.MODULE_NOT_FOUND, symbolicName, this);
            throw new DeploymentException("No module with bundle symbolic name '" + symbolicName + "' was found in application '" + this + "'");
        }
    }

}
