/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.deployer.app;

import java.io.BufferedReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.osgi.framework.Filter;
import org.osgi.framework.FrameworkUtil;
import org.osgi.framework.InvalidSyntaxException;
import org.osgi.framework.Version;
import org.springframework.util.StringUtils;
import org.springframework.util.xml.XmlValidationModeDetector;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import com.springsource.server.deployer.core.DeployerLogEvent;
import com.springsource.server.deployer.core.DeploymentException;
import com.springsource.server.deployer.core.FatalDeploymentException;
import com.springsource.server.deployer.core.ModuleData;
import com.springsource.server.deployer.core.DeploymentArtefact.Entry;
import com.springsource.server.osgi.manifest.Scoper;
import com.springsource.server.serviceability.logging.LogService;
import com.springsource.server.serviceability.logging.LogServiceFactory;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * Scopes service publication and export in an application to ensure that application bind to the services they export.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Not threadsafe.
 * 
 * @author Rob Harrop
 * @author Glyn Normington
 * @since 1.0
 */
final class ServiceScoper {

    static final String SCOPE_SEPARATOR = "-";

    private static final Tracer TRACER = TracerFactory.createTracer(ServiceScoper.class);

    private static final LogService LOGGER = LogServiceFactory.getLogService();

    /**
     * The service property used to scope services.
     */
    static final String SCOPE_KEY = "com.springsource.server.app.name";

    private static final String ATTRIBUTE_FILTER = "filter";

    private static final String ATTRIBUTE_INTERFACE = "interface";

    private static final String ELEMENT_VALUE = "value";

    private static final String ELEMENT_INTERFACES = "interfaces";

    private static final String ATTRIBUTE_VALUE = ELEMENT_VALUE;

    private static final String ATTRIBUTE_KEY = "key";

    private static final String ELEMENT_ENTRY = "entry";

    private static final String ELEMENT_SERVICE_PROPERTIES = "service-properties";

    private static final String SPRING_DM_NAMESPACE = "http://www.springframework.org/schema/osgi";

    private static final String SPRING_BEANS_NAMESPACE = "http://www.springframework.org/schema/beans";

    private static final String ELEMENT_SERVICE = "service";

    private static final String ELEMENT_REFERENCE = "reference";

    private final Map<String, List<Properties>> serviceInterfaces = new HashMap<String, List<Properties>>();

    private final Set<ModuleData> modules;

    private final String scopeName;

    private final DocumentBuilder builder;

    private final XmlValidationModeDetector xmlValidationModeDetector = new XmlValidationModeDetector();

    /**
     * Creates a new <code>ServiceScoper</code> for the supplied modules and scope name.
     * 
     * @param modules the modules of the application.
     * @param scopeName the name of the scope.
     * @param scopeVersion the version of the scope.
     */
    public ServiceScoper(Set<ModuleData> modules, String scopeName, Version scopeVersion) {
        this.modules = modules;
        this.scopeName = scopeName + SCOPE_SEPARATOR + versionToShortString(scopeVersion);
        this.builder = createDocumentBuilder();
    }

    private static String versionToShortString(Version version) {
        String result = version.toString();
        while (result.endsWith(".0")) {
            result = result.substring(0, result.length() - 2);
        }
        return result;
    }

    /**
     * Creates a {@link DocumentBuilder}.
     * 
     * @return the <code>DocumentBuilder</code>.
     */
    private DocumentBuilder createDocumentBuilder() {
        try {
            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            dbf.setNamespaceAware(true);
            return dbf.newDocumentBuilder();
        } catch (ParserConfigurationException e) {
            throw new FatalDeploymentException("Unable to create DocumentBuilder - JAXP parser configuration error.", e);
        }
    }

    /**
     * Scopes the application's services.
     * 
     * @throws DeploymentException
     */
    public void scope() throws DeploymentException {
        Map<Entry, ModuleData> configFiles = new HashMap<Entry, ModuleData>();
        for (ModuleData moduleData : modules) {
            findConfigFiles(moduleData, configFiles);
        }
        scopeServices(configFiles);
        scopeReferences(configFiles);

        if (TRACER.isDebugEnabled()) {
            for (Entry configFile : configFiles.keySet()) {
                dumpEntry("after", configFile);
            }
        }
    }

    private void findConfigFiles(ModuleData moduleData, Map<Entry, ModuleData> configFiles) {
        Entry entry = moduleData.getEntry("/META-INF/spring");
        findConfigFiles(moduleData, entry, configFiles);
    }

    private void findConfigFiles(ModuleData moduleData, Entry entry, Map<Entry, ModuleData> configFiles) {
        Set<Entry> children = entry.children();
        for (Entry e : children) {
            if (e.isDirectory()) {
                findConfigFiles(moduleData, e, configFiles);
            } else if (e.getName().endsWith(".xml")) {
                if (TRACER.isDebugEnabled()) {
                    dumpEntry("before", e);
                }
                try {
                    int validationMode = xmlValidationModeDetector.detectValidationMode(e.getInputStream());
                    if (validationMode != XmlValidationModeDetector.VALIDATION_DTD) {
                        configFiles.put(e, moduleData);
                    } else {
                        if (TRACER.isDebugEnabled()) {
                            TRACER.traceDebug("Skipping entry '%s' as it uses a DTD.", e);
                        }
                    }
                } catch (IOException ioe) {
                    if (TRACER.isDebugEnabled()) {
                        TRACER.traceDebug("Unexpected error detecting validation mode of entry '%s'", ioe, e);
                    }
                    configFiles.put(e, moduleData);
                }
            }
        }
    }

    /**
     * Re-scope the service exports and service references of the given {@link ModuleData}.
     * 
     * @param moduleData the {@link ModuleData} to be re-scoped.
     * @throws DeploymentException
     */
    public void rescope(ModuleData moduleData) throws DeploymentException {
        Map<Entry, ModuleData> configFiles = new HashMap<Entry, ModuleData>();
        findConfigFiles(moduleData, configFiles);
        scopeServices(configFiles);
        scopeReferences(configFiles);

        if (TRACER.isDebugEnabled()) {
            for (Entry configFile : configFiles.keySet()) {
                dumpEntry("after", configFile);
            }
        }
    }

    /**
     * Scopes the service exports.
     * 
     * @param configFiles the config files to scope.
     * @throws DeploymentException
     */
    private void scopeServices(Map<Entry, ModuleData> configFiles) throws DeploymentException {
        for (Entry configFile : configFiles.keySet()) {
            scopeServicesForEntry(configFile, configFiles.get(configFile));
        }
    }

    /**
     * Scopes the service references for services that are also published by the application.
     * 
     * @param configFiles the config files to scope.
     * @throws DeploymentException
     */
    private void scopeReferences(Map<Entry, ModuleData> configFiles) throws DeploymentException {
        for (Entry configFile : configFiles.keySet()) {
            scopeReferencesForEntry(configFile, configFiles.get(configFile));
        }
    }

    /**
     * Scopes the <code>service</code> tags in the supplied config file.
     * 
     * @param configFile the config file.
     * @throws DeploymentException
     */
    private void scopeServicesForEntry(Entry configFile, ModuleData moduleData) throws DeploymentException {
        try {
            Document doc = this.builder.parse(new InputSource(configFile.getReader()));
            Element documentElement = doc.getDocumentElement();
            doScopeServices(doc, documentElement.getChildNodes());
            writeXml(doc, configFile);
        } catch (SAXException ex) {
            LOGGER.log(DeployerLogEvent.CONFIG_FILE_ERROR, ex, configFile.getName(), Scoper.getUnscopedSymbolicName(moduleData.getBundleManifest()));
            throw new DeploymentException("Error parsing configuration file '" + configFile.getName() + "'.", ex);
        } catch (IOException ex) {
            throw new FatalDeploymentException("Error accessing configuration file '" + configFile.getName() + "'.", ex);
        }
    }

    /**
     * Scopes the <code>service</code> tags in the supplied {@link NodeList}.
     * 
     * @param doc the config file <code>Document</code>.
     * @param childNodes the nodes to scope.
     */
    private void doScopeServices(Document doc, NodeList childNodes) {
        for (int x = 0; x < childNodes.getLength(); x++) {
            Node node = childNodes.item(x);
            if (node instanceof Element) {
                if (isServiceElement(node)) {
                    Element elem = (Element) node;
                    Element servicePropertiesElement = null;
                    Element interfacesElement = null;
                    
                    Properties serviceProperties = new Properties();

                    NodeList serviceChildren = elem.getChildNodes();
                    for (int y = 0; y < serviceChildren.getLength(); y++) {
                        Node serviceChild = serviceChildren.item(y);
                        if (isServicePropertiesElement(serviceChild)) {
                            servicePropertiesElement = (Element) serviceChild;
                            break;
                        } else if(isInterfacesElement(serviceChild)) {
                            interfacesElement = (Element) serviceChild;
                        }
                    }


                    if (servicePropertiesElement == null) {
                        servicePropertiesElement = doc.createElementNS(SPRING_DM_NAMESPACE, ELEMENT_SERVICE_PROPERTIES);
                        
                        if(interfacesElement == null) {
                            Node firstChild = elem.getFirstChild();
                            if(firstChild == null) {
                                elem.appendChild(servicePropertiesElement);
                            } else {
                                elem.insertBefore(servicePropertiesElement, firstChild);
                            }
                        } else {
                            Node sibling = interfacesElement.getNextSibling();
                            if(sibling == null) {
                                elem.appendChild(servicePropertiesElement);
                            } else {
                                elem.insertBefore(servicePropertiesElement, sibling);
                            }
                        }
                        
                    } else {
                        readServiceProperties(servicePropertiesElement, serviceProperties);
                    }

                    Element entry = doc.createElementNS(SPRING_BEANS_NAMESPACE, ELEMENT_ENTRY);
                    entry.setAttribute(ATTRIBUTE_KEY, SCOPE_KEY);
                    entry.setAttribute(ATTRIBUTE_VALUE, this.scopeName);
                    servicePropertiesElement.appendChild(entry);

                    Set<String> interfaces = extractInterfaces(elem);
                    for (String iface : interfaces) {
                        List<Properties> propList = this.serviceInterfaces.get(iface);
                        if(propList == null) {
                            propList = new ArrayList<Properties>();
                            this.serviceInterfaces.put(iface, propList);
                        }
                        propList.add(serviceProperties);
                    }
                }
            }
            doScopeServices(doc, node.getChildNodes());
        }
    }

    private void readServiceProperties(Element servicePropertiesElement, Properties serviceProperties) {
        NodeList childNodes = servicePropertiesElement.getChildNodes();
        for (int y = 0; y < childNodes.getLength(); y++) {
            Node child = childNodes.item(y);
            if (isEntryElement(child)) {
                Element entry = (Element) child;
                serviceProperties.setProperty(entry.getAttribute(ATTRIBUTE_KEY), entry.getAttribute(ATTRIBUTE_VALUE));
            }
        }
    }

    /**
     * Scopes the <code>reference</code> tags in the supplied config file.
     * 
     * @param configFile the config file.
     * @throws DeploymentException
     */
    private void scopeReferencesForEntry(Entry configFile, ModuleData moduleData) throws DeploymentException {
        try {
            Document doc = this.builder.parse(new InputSource(configFile.getReader()));
            Element documentElement = doc.getDocumentElement();
            doScopeReferences(doc, documentElement.getChildNodes(), configFile.getName());
            writeXml(doc, configFile);
        } catch (SAXException ex) {
            LOGGER.log(DeployerLogEvent.CONFIG_FILE_ERROR, ex, configFile.getName(), Scoper.getUnscopedSymbolicName(moduleData.getBundleManifest()));
            throw new DeploymentException("Error parsing configuration file '" + configFile.getName() + "'.", ex);
        } catch (IOException ex) {
            throw new FatalDeploymentException("Error accessing configuration file '" + configFile.getName() + "'.", ex);
        }
    }

    /**
     * Scopes the <code>reference</code> tags in the supplied {@link NodeList}.
     * 
     * @param doc the config file <code>Document</code>.
     * @param childNodes the nodes to scope.
     * @param source a String describing the source of the config file
     */
    private void doScopeReferences(Document doc, NodeList childNodes, String source) {
        for (int x = 0; x < childNodes.getLength(); x++) {
            Node node = childNodes.item(x);
            if (node instanceof Element) {
                if (isReferenceElement(node)) {
                    Element elem = (Element) node;
                    String filterValue = null;
                    if (isReferenceForExportedService(elem, source)) {
                        filterValue = String.format("(%s=%s)", SCOPE_KEY, this.scopeName, SCOPE_KEY);
                    } else {
                        // prevent referencing services from other pars
                        filterValue = String.format("(! (%s=*))", SCOPE_KEY);
                    }
                    String existingFilter = elem.getAttribute(ATTRIBUTE_FILTER);
                    if (StringUtils.hasText(existingFilter)) {
                        filterValue = String.format("(& %s %s)", existingFilter, filterValue);
                    }

                    elem.setAttribute(ATTRIBUTE_FILTER, filterValue);
                }
            }
            doScopeReferences(doc, node.getChildNodes(), source);
        }
    }

    /**
     * Queries whether the supplied <code>reference</code> {@link Element} refers to a <code>service</code> in the
     * application being scoped.
     * 
     * @param reference the <code>reference</code> <code>Element</code>.
     * @return <code>true</code> if the <code>reference</code> refers to an exported service, otherwise
     *         <code>false</code>.
     */
    private boolean isReferenceForExportedService(Element reference, String source) {
        Set<String> referencedInterfaces = extractInterfaces(reference);
        for (String referencedInterface : referencedInterfaces) {
            if (this.serviceInterfaces.containsKey(referencedInterface)) {
                String existingFilter = reference.getAttribute(ATTRIBUTE_FILTER);
                if (StringUtils.hasText(existingFilter)) {
                    List<Properties> propList = this.serviceInterfaces.get(referencedInterface);
                    for (Properties serviceProperties : propList) {
                        try {
                            Filter filter = FrameworkUtil.createFilter(existingFilter);
                            if(filter.match(serviceProperties)) {
                                return true;
                            }
                        } catch (InvalidSyntaxException e) {
                            throw new FatalDeploymentException("Filter '" + existingFilter + "' in '" + source + "' is invalid.");
                        }
                        
                    }
                } else {
                    return true;
                }
            }
        }
        return false;
    }

    /**
     * Extracts all the interfaces from the supplied <code>reference</code> or <code>service</code> {@link Element}.
     * 
     * @param e the <code>Element</code> to parse.
     * @return the interfaces.
     */
    private Set<String> extractInterfaces(Element e) {
        Set<String> exportedInterfaces = new HashSet<String>();
        String iface = StringUtils.trimWhitespace(e.getAttribute(ATTRIBUTE_INTERFACE));
        if (StringUtils.hasText(iface)) {
            exportedInterfaces.add(iface);
        } else {
            NodeList children = e.getChildNodes();
            for (int y = 0; y < children.getLength(); y++) {
                Node child = children.item(y);
                if (child instanceof Element) {
                    if (isInterfacesElement(child)) {
                        Element elem = (Element) child;
                        NodeList intChildren = elem.getChildNodes();
                        for (int i = 0; i < intChildren.getLength(); i++) {
                            Node intChild = intChildren.item(i);
                            if (isValueElement(intChild)) {
                                exportedInterfaces.add(StringUtils.trimWhitespace(intChild.getTextContent()));
                            }
                        }
                    }
                }
            }
        }
        return exportedInterfaces;
    }

    private boolean isValueElement(Node node) {
        return SPRING_BEANS_NAMESPACE.equals(node.getNamespaceURI()) && ELEMENT_VALUE.equals(node.getLocalName())
            && node.getNodeType() == Node.ELEMENT_NODE;
    }

    private boolean isInterfacesElement(Node node) {
        return SPRING_DM_NAMESPACE.equals(node.getNamespaceURI()) && ELEMENT_INTERFACES.equals(node.getLocalName())
            && node.getNodeType() == Node.ELEMENT_NODE;
    }

    private boolean isServicePropertiesElement(Node node) {
        return SPRING_DM_NAMESPACE.equals(node.getNamespaceURI()) && ELEMENT_SERVICE_PROPERTIES.equals(node.getLocalName())
            && node.getNodeType() == Node.ELEMENT_NODE;
    }

    private boolean isServiceElement(Node node) {
        return SPRING_DM_NAMESPACE.equals(node.getNamespaceURI()) && ELEMENT_SERVICE.equals(node.getLocalName())
            && node.getNodeType() == Node.ELEMENT_NODE;
    }

    private boolean isReferenceElement(Node node) {
        return SPRING_DM_NAMESPACE.equals(node.getNamespaceURI()) && ELEMENT_REFERENCE.equals(node.getLocalName())
            && node.getNodeType() == Node.ELEMENT_NODE;
    }

    private boolean isEntryElement(Node node) {
        return SPRING_BEANS_NAMESPACE.equals(node.getNamespaceURI()) && ELEMENT_ENTRY.equals(node.getLocalName())
            && node.getNodeType() == Node.ELEMENT_NODE;
    }

    /**
     * Writes the supplied {@link Document} into the supplied {@link Entry}.
     * 
     * @param doc the <code>Document</code> to write.
     * @param entry the target <code>Entry</code>.
     */
    private static void writeXml(Document doc, Entry entry) {
        try {
            Source source = new DOMSource(doc);
            Result result = new StreamResult(entry.getWriter(false));
            Transformer xformer = TransformerFactory.newInstance().newTransformer();
            xformer.transform(source, result);
        } catch (Exception ex) {
            throw new FatalDeploymentException("Error writing XML config file for entry '" + entry + "'", ex);
        }
    }

    /**
     * Dumps the supplied entry.
     * 
     * @param phase the phase of processing.
     * @param entry the entry.
     */
    private void dumpEntry(String phase, Entry entry) {
        StringBuilder sb = new StringBuilder(entry.getName()).append(" <").append(phase).append(">\n");
        BufferedReader reader = new BufferedReader(entry.getReader());
        String line;
        try {
            while ((line = reader.readLine()) != null) {
                sb.append(line).append("\n");
            }
        } catch (IOException e) {
            TRACER.traceError("Error tracing config file contents for '%s'", e, entry.getName());
        } finally {
            try {
                reader.close();
            } catch (IOException e) {
                TRACER.traceWarning("Error closing Reader", e);
            }
        }
        TRACER.traceDebug(sb.toString());
    }
}
