/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.deployer.app.context;

import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import com.springsource.server.common.math.Version;
import com.springsource.server.deployer.core.ContextModuleFactory;
import com.springsource.server.deployer.core.DeploymentException;
import com.springsource.server.deployer.core.FatalDeploymentException;
import com.springsource.server.deployer.core.InformedDeployable;
import com.springsource.server.deployer.core.Module;
import com.springsource.server.deployer.core.ModuleFactory;
import com.springsource.server.deployer.core.internal.DeployArea;
import com.springsource.server.io.core.PathReference;
import com.springsource.server.osgi.framework.BundleInstaller;
import com.springsource.server.osgi.manifest.BundleManifest;
import com.springsource.server.osgi.manifest.BundleManifestFactory;
import com.springsource.server.osgi.manifest.BundleManifest.ImportBundleHeader.BundleImport;
import com.springsource.server.osgi.manifest.BundleManifest.ImportPackageHeader.PackageImport;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * {@link StandardContextModuleFactory} is the implementation of {@link ContextModuleFactory}.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * This class is thread safe.
 * 
 * @author Glyn Normington
 * @since 1.0
 */
final class StandardContextModuleFactory implements ContextModuleFactory {

    private static final int SYNTHETIC_BUNDLE_MANIFEST_VERSION = 2;

    private static final String SYNTHETIC_CONTEXT_SUFFIX = "-synthetic.context";

    private static final String WORK_AREA_NAME = "Context";

    private static final Tracer TRACER = TracerFactory.createTracer(StandardContextModuleFactory.class);

    private final DeployArea deployArea;

    private final BundleManifestFactory bundleManifestFactory;

    private final ModuleFactory moduleFactory;

    StandardContextModuleFactory(DeployArea deployArea, BundleManifestFactory bundleManifestFactory, BundleInstaller bundleInstaller,
        ModuleFactory moduleFactory) {
        this.deployArea = deployArea.createWorkArea(WORK_AREA_NAME, true);
        this.bundleManifestFactory = bundleManifestFactory;
        this.moduleFactory = moduleFactory;
    }

    /**
     * @throws DeploymentException {@inheritDoc}
     */
    public Module createContext(Collection<Module> modules, InformedDeployable informedDeployable) throws DeploymentException {
        // Create the bundle name of the context module
        StringBuffer contextBundleName = new StringBuffer(informedDeployable.getName());
        contextBundleName.append(SYNTHETIC_CONTEXT_SUFFIX);
        // Create directory to hold the context bundle
        StringBuffer contextDirectoryName = new StringBuffer(informedDeployable.getName());
        contextDirectoryName.append("-");
        contextDirectoryName.append(informedDeployable.getVersion());
        contextDirectoryName.append(SYNTHETIC_CONTEXT_SUFFIX);
        PathReference bundleDir = this.deployArea.createDirectory(contextDirectoryName.toString());
        try {
            PathReference manifestFile = bundleDir.newChild("META-INF/MANIFEST.MF");
            manifestFile.createFile();
            Writer manifestWriter;
            try {
                manifestWriter = new FileWriter(manifestFile.toFile());
            } catch (IOException e) {
                // Fatal since it is not the user's fault.
                throw new FatalDeploymentException("Unable to write manifest of synthetic context bundle", e);
            }
            // Create bundle manifest for context module
            BundleManifest bundleManifest = this.bundleManifestFactory.createBundleManifest();
            bundleManifest.setBundleSymbolicName(this.bundleManifestFactory.createBundleSymbolicName(contextBundleName.toString()));
            bundleManifest.setBundleVersion(informedDeployable.getVersion().toString());
            bundleManifest.setBundleManifestVersion(SYNTHETIC_BUNDLE_MANIFEST_VERSION);

            // Set the bundle scope. Not strictly necessary, but could be a red herring during debugging
            StringBuffer contextScope = new StringBuffer(informedDeployable.getName());
            contextScope.append("-");
            contextScope.append(informedDeployable.getVersion());
            bundleManifest.setModuleScope(contextScope.toString());

            computeContextImports(modules, bundleManifest, informedDeployable.getName(), informedDeployable.getVersion());

            bundleManifest.write(manifestWriter);
            try {
                manifestWriter.close();
            } catch (IOException e) {
                TRACER.traceError("Problem closing manifest output stream", e);
            }

            TRACER.traceInfo("Synthetic context bundle manifest for application '%s' version '%s':\n%s", informedDeployable.getName(),
                informedDeployable.getVersion().toString(), bundleManifest);

            Module context = this.moduleFactory.create(informedDeployable, bundleDir.toURI());

            for (Module module : modules) {
                module.setContext(context);
            }
            return context;
        } finally {
            bundleDir.delete(true);
        }
    }

    private void computeContextImports(Collection<Module> modules, BundleManifest bundleManifest, String applicationName, Version applicationVersion)
        throws DeploymentException {
        // Import all the application bundles
        List<BundleImport> syntheticBundleImports = new ArrayList<BundleImport>(1);
        for (Module module : modules) {
            String bundleSymbolicName = module.getUnscopedSymbolicName();
            if (bundleSymbolicName != null) {
                syntheticBundleImports.add(this.bundleManifestFactory.createBundleImport(bundleSymbolicName));
            }
        }
        if (syntheticBundleImports.size() > 0) {
            bundleManifest.setImportBundleHeader(this.bundleManifestFactory.createImportBundleHeader(syntheticBundleImports));
        }

        // Import the necessary Spring DM package
        List<PackageImport> syntheticImports = new ArrayList<PackageImport>(1);
        PackageImport dmContextPackageImport = this.bundleManifestFactory.createPackageImport();
        dmContextPackageImport.getPackageNames().add("org.springframework.osgi.context");
        syntheticImports.add(dmContextPackageImport);
        bundleManifest.setImportPackageHeader(this.bundleManifestFactory.createImportPackageHeader(syntheticImports));
    }

}
