/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.deployer.controls;

import java.net.URI;

import javax.management.openmbean.CompositeData;
import javax.management.openmbean.CompositeDataSupport;
import javax.management.openmbean.CompositeType;
import javax.management.openmbean.OpenDataException;
import javax.management.openmbean.OpenType;
import javax.management.openmbean.SimpleType;

import com.springsource.server.control.core.ControlExporter;
import com.springsource.server.control.core.ControlHandle;
import com.springsource.server.deployer.core.ApplicationDeployer;
import com.springsource.server.deployer.core.DeploymentException;
import com.springsource.server.deployer.core.DeploymentIdentity;
import com.springsource.server.deployer.core.FatalDeploymentException;
import com.springsource.server.deployer.core.ApplicationDeployer.DeploymentOptions;
import com.springsource.server.deployer.core.internal.StandardDeploymentIdentity;

/**
 * Standard implementation of the deployer's control.
 * 
 * <strong>Concurrent Semantics</strong><br />
 * Thread-safe assuming that the ApplicationDeployer is thread-safe.
 * 
 * @author Andy Wilkinson
 * @author Glyn Normington
 * @since 1.0
 */
public class StandardDeployerControl implements Deployer {

    private static final String ITEM_SYMBOLIC_NAME = "symbolicName";

    private static final String ITEM_VERSION = "version";

    private static final String DEPLOYMENT_IDENTITY_TYPE_NAME = DeploymentIdentity.class.getName();

    private static final String[] DEPLOYMENT_IDENTITY_ITEM_NAMES = { ITEM_SYMBOLIC_NAME, ITEM_VERSION };

    private static final OpenType[] DEPLOYMENT_IDENTITY_ITEM_TYPES = { SimpleType.STRING, SimpleType.STRING };

    private static final CompositeType DEPLOYMENT_IDENTITY_TYPE = constructDeploymentIdentityType();

    private final ApplicationDeployer applicationDeployer;

    private final ControlHandle controlHandle;

    public StandardDeployerControl(ApplicationDeployer applicationDeployer, ControlExporter controlExporter) {
        this.applicationDeployer = applicationDeployer;
        this.controlHandle = controlExporter.exportControl(this, Deployer.class, null);
    }

    /**
     * @return
     */
    private static CompositeType constructDeploymentIdentityType() {
        try {
            return new CompositeType(DEPLOYMENT_IDENTITY_TYPE_NAME, DEPLOYMENT_IDENTITY_TYPE_NAME, DEPLOYMENT_IDENTITY_ITEM_NAMES,
                DEPLOYMENT_IDENTITY_ITEM_NAMES, DEPLOYMENT_IDENTITY_ITEM_TYPES);
        } catch (OpenDataException e) {
            throw new FatalDeploymentException("Unable to construct DeploymentIdentity OpenType.");
        }
    }

    /**
     * {@inheritDoc}
     * 
     * @return
     */
    public CompositeData deploy(String uri) throws DeploymentException {
        return toCompositeData(this.applicationDeployer.deploy(URI.create(uri)));
    }

    /**
     * {@inheritDoc}
     * 
     * @return
     */
    public CompositeData deploy(String uri, boolean recoverable) throws DeploymentException {
        return toCompositeData(this.applicationDeployer.deploy(URI.create(uri), new DeploymentOptions(recoverable, false)));
    }

    /**
     * {@inheritDoc}
     */
    public void undeploy(String applicationSymbolicName, String version) throws DeploymentException {
        this.applicationDeployer.undeploy(applicationSymbolicName, version);
    }

    /**
     * Retract this control so it is no longer visible.
     */
    public void retract() {
        this.controlHandle.retractControl();
    }

    /**
     * {@inheritDoc}
     */
    public void refresh(String uri, String symbolicName) throws DeploymentException {
        this.applicationDeployer.refresh(URI.create(uri), symbolicName);
    }

    private CompositeData toCompositeData(DeploymentIdentity identity) throws DeploymentException {
        Object[] values = { identity.getSymbolicName(), identity.getVersion() };
        try {
            return new CompositeDataSupport(DEPLOYMENT_IDENTITY_TYPE, DEPLOYMENT_IDENTITY_ITEM_NAMES, values);
        } catch (OpenDataException e) {
            throw new DeploymentException("Unable to convert DeploymentIdentity to OpenType", e);
        }
    }

    /**
     * {@inheritDoc}
     */
    public void updateResource(String applicationSymbolicName, String version, String resourceLocation, String targetPath) throws DeploymentException {
        this.applicationDeployer.updateResource(new StandardDeploymentIdentity(applicationSymbolicName, version), URI.create(resourceLocation),
            targetPath);
    }

    /**
     * {@inheritDoc}
     */
    public void updateResource(String applicationSymbolicName, String version, String symbolicName, String resourceLocation, String targetPath)
        throws DeploymentException {
        this.applicationDeployer.updateResource(new StandardDeploymentIdentity(applicationSymbolicName, version), symbolicName,
            URI.create(resourceLocation), targetPath);
    }

    /**
     * {@inheritDoc}
     */
    public void deleteResource(String applicationSymbolicName, String version, String targetPath) throws DeploymentException {
        this.applicationDeployer.deleteResource(new StandardDeploymentIdentity(applicationSymbolicName, version), targetPath);
    }

    /**
     * {@inheritDoc}
     */
    public void deleteResource(String applicationSymbolicName, String version, String symbolicName, String targetPath) throws DeploymentException {
        this.applicationDeployer.deleteResource(new StandardDeploymentIdentity(applicationSymbolicName, version), symbolicName, targetPath);
    }

}
