/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.deployer.core;

import java.net.URI;
import java.util.concurrent.TimeUnit;

/**
 * The ApplicationDeployer interface is the programmatic interface to the deployer subsystem for deploying applications
 * and libraries.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Implementations of this interface must be thread safe.
 * 
 * @author Rob Harrop
 * @author Glyn Normington
 * @since 1.0
 */
public interface ApplicationDeployer {

    /**
     * {@link DeploymentOptions} provides a collection of deployment options.
     * <p />
     * 
     * <strong>Concurrent Semantics</strong><br />
     * 
     * This class is immutable and therefore thread safe.
     */
    public class DeploymentOptions {

        public static final DeploymentOptions DEFAULT_DEPLOYMENT_OPTIONS = new DeploymentOptions();

        private final boolean recoverable;

        private final boolean deployerOwned;

        /**
         * Create default deployment options.
         */
        public DeploymentOptions() {
            this.recoverable = true;
            this.deployerOwned = false;
        }

        /**
         * Create deployment options with the given recoverability and ownership.
         * 
         * @param recoverable <code>true</code> if and only if the application is to persist across Server restarts
         * @param deployerOwned <code>true</code> if and only if the application at the location specified on
         *        deployment is to be deleted when the application is undeployed
         */
        public DeploymentOptions(boolean recoverable, boolean deployerOwned) {
            this.recoverable = recoverable;
            this.deployerOwned = deployerOwned;
        }

        /**
         * Get the recoverability option.
         * 
         * @return <code>true</code> if and only if the application is to persist across Server restarts
         */
        public boolean getRecoverable() {
            return this.recoverable;
        }

        /**
         * Get the ownership option.
         * 
         * @return <code>true</code> if and only if the application at the location specified on deployment is to be
         *         deleted when the application is undeployed
         */
        public boolean getDeployerOwned() {
            return this.deployerOwned;
        }
    }

    /**
     * Deploy an application which may be either an OSGi application or a legacy application such as a WAR. This is used
     * by admin. and hot deployment. A successfully deployed application is always recovered on warm restart but the
     * application at the given location is not deleted when the application is undeployed.
     * <p />
     * This method is equivalent to calling {@link deploy(location, new DeploymentOptions())}.
     * 
     * @param location the location of the application JAR file
     * @return the {@link DeploymentIdentity} of the deployed application
     * @throws DeploymentException
     */
    DeploymentIdentity deploy(URI location) throws DeploymentException;

    /**
     * Deploy an application which may be either an OSGi application or a legacy application such as a WAR. This is used
     * by admin. and hot deployment.
     * 
     * @param location the location of the application JAR file
     * @param options the options for this deployment
     * @return the {@link DeploymentIdentity} of the deployed application
     * @throws DeploymentException
     */
    DeploymentIdentity deploy(URI location, DeploymentOptions options) throws DeploymentException;

    /**
     * Undeploy an application with a given symbolic name and version. The application may be either an OSGi application
     * or a legacy application such as a WAR.
     * 
     * @param applicationSymbolicName the symbolic name of the application
     * @param version the version of the application in string form
     * @throws DeploymentException
     */
    void undeploy(String applicationSymbolicName, String version) throws DeploymentException;

    /**
     * Undeploy an application with a given {@link DeploymentIdentity}. The application may be either an OSGi
     * application or a legacy application such as a WAR.
     * 
     * @param deploymentIdentity the <code>DeploymentIdentity</code> of the application
     * @throws DeploymentException
     */
    void undeploy(DeploymentIdentity deploymentIdentity) throws DeploymentException;

    /**
     * Refresh a single module of the application which was deployed from the given URI.
     * 
     * @param location the location from which the application was deployed
     * @param symbolicName the bundle symbolic name of the module to be refreshed
     * @throws DeploymentException
     */
    void refresh(URI location, String symbolicName) throws DeploymentException;

    /**
     * Update files or directories of the module of the single-bundle application with the given
     * {@link DeploymentIdentity}. The application must have been deployed with the non-recoverable deployment option.
     * <p />
     * The given location refers to a file or directory to be copied recursively over the file or directory at a given
     * path relative to the root directory of the module.
     * <p />
     * <b>WARNING</b>
     * <p />
     * The intended use of this method is to update static resources which are loaded directly from the exploded form of
     * the module in the deployer's work area. This method does <b>not</b> update or refresh the given module and so
     * the updated resources <b>cannot</b> be accessed using the module's class loader. The updated resources will be
     * <b>overwritten</b> if the given module is refreshed or redeployed.
     * 
     * @param deploymentIdentity the <code>DeploymentIdentity</code> of the application
     * @param resourceLocation the location of the file or directory to be overlaid on the file or directory at the
     *        given target path of the deployed module
     * @param targetPath the relative URI path of the location of the file to be updated within the deployed module
     * @throws DeploymentException
     */
    void updateResource(DeploymentIdentity deploymentIdentity, URI resourceLocation, String targetPath) throws DeploymentException;

    /**
     * Update files or directories of a module of the multi-bundle application with the given {@link DeploymentIdentity}.
     * The application must have been deployed with the non-recoverable deployment option.
     * <p />
     * The given location refers to a file or directory to be copied recursively over the file or directory at a
     * given path relative to the root directory of the module.
     * <p />
     * <b>WARNING</b>
     * <p />
     * The intended use of this method is to update static resources which are loaded directly from the exploded form of
     * the module in the deployer's work area. This method does <b>not</b> update or refresh the given module and so
     * the updated resources <b>cannot</b> be accessed using the module's class loader. The updated resources will be
     * <b>overwritten</b> if the given module is refreshed or redeployed.
     * 
     * @param deploymentIdentity the <code>DeploymentIdentity</code> of the application
     * @param symbolicName the bundle symbolic name of the module to be updated
     * @param resourceLocation the location of the file or directory to be overlaid on the file or directory at the
     *        given target path of the deployed module or <code>null</code> if the file or directory at the given
     *        target path is to be deleted
     * @param targetPath the relative URI path of the location of the file to be updated within the deployed module
     * @throws DeploymentException
     */
    void updateResource(DeploymentIdentity deploymentIdentity, String symbolicName, URI resourceLocation, String targetPath)
        throws DeploymentException;

    /**
     * Delete a file or directory at the given path relative to the root directory of the module of the single-bundle
     * application with the given {@link DeploymentIdentity}. The application must have been deployed with the
     * non-recoverable deployment option.
     * <p />
     * <b>WARNING</b>
     * <p />
     * The intended use of this method is to delete static resources which are loaded directly from the exploded form of
     * the module in the deployer's work area. This method does <b>not</b> update or refresh the given module and so
     * the updated resources <b>may still</b> be accessed using the module's class loader.
     * 
     * @param deploymentIdentity the <code>DeploymentIdentity</code> of the application
     * @param targetPath the relative URI path of the location of the file to be deleted within the deployed module
     * @throws DeploymentException
     */
    void deleteResource(DeploymentIdentity deploymentIdentity, String targetPath) throws DeploymentException;

    /**
     * Delete a file or directory at the given path relative to the root directory of a module of the multi-bundle
     * application with the given {@link DeploymentIdentity}. The application must have been deployed with the
     * non-recoverable deployment option.
     * <p />
     * <b>WARNING</b>
     * <p />
     * The intended use of this method is to delete static resources which are loaded directly from the exploded form of
     * the module in the deployer's work area. This method does <b>not</b> update or refresh the given module and so
     * the updated resources <b>may still</b> be accessed using the module's class loader.
     * 
     * @param deploymentIdentity the <code>DeploymentIdentity</code> of the application
     * @param symbolicName the bundle symbolic name of the module to be updated
     * @param targetPath the relative URI path of the location of the file to be deleted within the deployed module
     * @throws DeploymentException
     */
    void deleteResource(DeploymentIdentity deploymentIdentity, String symbolicName, String targetPath) throws DeploymentException;

    /**
     * Get a list of deployed applications.
     * 
     * @return Array of deployed application identities.
     */
    DeploymentIdentity[] getDeploymentIdentities();

    /**
     * Get the {@link DeploymentIdentity} of an application deployed from the given location. If no such application was
     * found, return <code>null</code>.
     * <p />
     * Although the deployer currently prevents two applications being deployed from the same location without
     * undeploying the first application, this restriction may be lifted in the future in which case this method will
     * return one of potentially many applications deployed from the given location.
     * 
     * @param location the {@link URI} used to deploy the application
     * @return a <code>DeploymentIdentity</code> or <code>null</code>
     */
    DeploymentIdentity getDeploymentIdentity(URI location);

    /**
     * Waits until the deployer recovery process has completed. Returns immediately if recovery has already completed.
     * 
     * @param timeout the amount of time to wait.
     * @param unit the unit of time to wait.
     * @return <code>true</code> if recovery completed successfully, otherwise <code>false</code>.
     * @throws InterruptedException if the thread was interrupted while waiting
     */
    boolean awaitRecovery(long timeout, TimeUnit unit) throws InterruptedException;

    /**
     * Determine whether or not the given artefact is already deployed. Return <code>true</code> if the given artefact
     * at its file's last modified time is already deployed. The last modified time is only taken into account if the
     * file was deployed on a previous run of the Server.
     * 
     * @param sourceArtefact the source artefact URI string
     * @return <code>true</code> if and only if the given artefact at its file's last modified time is already
     *         deployed
     */
    boolean isDeployed(URI location);

}
