/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.deployer.core;

import java.util.Set;

import org.springframework.core.Ordered;

import com.springsource.server.osgi.manifest.VersionRange;

/**
 * A module deployer is capable of deploying one or more particular kinds of application module. This interface is
 * provided by the deployer subsystem and is implemented by the various application containers. <p/> Implementations of
 * this interface should extend AbstractModuleDeployer which polices the interface protocol. <p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Implementations of this interface must be thread safe.
 * 
 * @author Rob Harrop
 * @author Glyn Normington
 * @author Sam Brannen
 * @since 1.0
 */
public interface ModuleDeployer extends Ordered {

    static final String DEPLOYER_EXTENSION_POINT_NAME = "ModuleDeployer";

    /**
     * Module type manifest header.
     */
    static final String MODULE_TYPE_MANIFEST_HEADER = "Module-Type";

    /**
     * Module type for generic bundle modules.
     */
    static final String BUNDLE_MODULE_TYPE = "Bundle";

    static final int TEST_ORDER = 50;

    static final int WEB_ORDER = 100;

    static final int WAR_ORDER = 200;

    static final int BUNDLE_ORDER = Ordered.LOWEST_PRECEDENCE;

    /**
     * Determine whether this deployer can deploy the given module. If true is returned, this deployer will be used to
     * deploy the module. The deployer has the opportunity to register a {@link ModuleDeployerContext} if it needs to
     * know when the current deployment has completed.
     * 
     * @param deploymentContext the current deployment context
     * @param module the module to be deployed
     * @return true if and only if this deployer can deploy the given module, otherwise return false
     */
    boolean canDeploy(DeploymentContext deploymentContext, Module module);

    /**
     * Get the personality name.
     * 
     * @return the personality name
     */
    String getPersonality();

    /**
     * Get any personality-specific bundles e.g. for creating application contexts on behalf of this deployer.
     * 
     * @param deploymentContext the current deployment context
     * @return a set of PersonalityBundle describing the bundles to be included in the application scope (or
     *         <code>null</code> which is equivalent to the empty set)
     */
    Set<PersonalityBundle> getBundles(DeploymentContext deploymentContext);

    /**
     * {@link PersonalityBundle} identifies a bundle to be included in the application scope on behalf of this deployer.
     * <p />
     * 
     * <strong>Concurrent Semantics</strong><br />
     * 
     * Implementations of this interface must be thread safe.
     * 
     * @author Glyn Normington
     * @since 1.0
     */
    public interface PersonalityBundle {

        /**
         * Get the bundle symbolic name of the personality bundle.
         * 
         * @return the bundle symbolic name
         */
        String getBundleSymbolicName();

        /**
         * Get the bundle version range of the personality bundle.
         * 
         * @return the bundle version range
         */
        VersionRange getBundleVersionRange();
    }
}
