/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.deployer.core.internal;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.osgi.framework.Bundle;
import org.springframework.beans.BeansException;
import org.springframework.beans.InvalidPropertyException;
import org.springframework.beans.TypeMismatchException;
import org.springframework.beans.factory.BeanCreationException;
import org.springframework.beans.factory.CannotLoadBeanClassException;

import com.springsource.server.deployer.core.DeployerLogEvent;
import com.springsource.server.osgi.framework.ServerClassNotFoundException;
import com.springsource.server.osgi.framework.ServerNoClassDefFoundError;
import com.springsource.server.serviceability.logging.LogService;
import com.springsource.server.serviceability.logging.LogServiceFactory;

/**
 * <code>BeansExceptionHandler</code> is a helper class for handling {@link BeansException BeansExceptions} that are
 * thrown during the processing of an application context.
 * <p />
 * <strong>Concurrent Semantics</strong><br />
 * This class is <strong>thread-safe</strong>.
 * 
 * @author Andy Wilkinson
 * @since 1.0
 */
public class BeansExceptionHandler {

    private static final LogService LOGGER = LogServiceFactory.getLogService();
    
    private static final Map<Class<? extends BeansException>, Handler<?>> handlers = new HashMap<Class<? extends BeansException>, Handler<?>>();
    
    static {
        handlers.put(InvalidPropertyException.class, new Handler<InvalidPropertyException>() {
            public boolean handleException(InvalidPropertyException ipe, String beanName, String bundleDescription) {
                LOGGER.log(DeployerLogEvent.BUNDLE_CONTEXT_FAILED, bundleDescription);
                LOGGER.log(DeployerLogEvent.APP_CONTEXT_CREATION_FAILURE_INVALID_PROPERTY, ipe.getPropertyName(), beanName, ipe.getBeanClass());
                return true;
            }            
        });
        handlers.put(TypeMismatchException.class, new Handler<TypeMismatchException>() {
            public boolean handleException(TypeMismatchException tme, String beanName, String bundleDescription) {
                LOGGER.log(DeployerLogEvent.BUNDLE_CONTEXT_FAILED, bundleDescription);
                LOGGER.log(DeployerLogEvent.APP_CONTEXT_CREATION_FAILURE_TYPE_MISMATCH, tme.getValue(), beanName, tme.getRequiredType());
                return true;
            }            
        });
        handlers.put(CannotLoadBeanClassException.class, new Handler<CannotLoadBeanClassException>() {
            public boolean handleException(CannotLoadBeanClassException clbce, String beanName, String bundleDescription) {
                ClassLoader classLoader = null;
                Throwable cause = clbce.getCause();
                Throwable serverThrowable = null;
                
                while (cause != null) {
                    if (cause instanceof ServerNoClassDefFoundError) {
                        classLoader = ((ServerNoClassDefFoundError)cause).getClassLoader();
                        serverThrowable = cause.getCause();
                    } else if (cause instanceof ServerClassNotFoundException) {
                        classLoader = ((ServerClassNotFoundException)cause).getClassLoader();
                        serverThrowable = cause.getCause();
                    }
                    cause = cause.getCause();
                }
                
                LOGGER.log(DeployerLogEvent.BUNDLE_CONTEXT_FAILED, bundleDescription);
                if (serverThrowable == null) {
                    serverThrowable = clbce;
                }
                LOGGER.log(DeployerLogEvent.APP_CONTEXT_CREATION_FAILURE_CANNOT_LOAD_BEAN_CLASS, serverThrowable, clbce.getBeanClassName(), clbce.getBeanName(), classLoader);
                return true;
            }            
        });
    }
    
    private static String determineBeanName(List<Throwable> invertedHierarchy) {
        for (Throwable throwable : invertedHierarchy) {
            if (throwable instanceof BeanCreationException) {
                return ((BeanCreationException)throwable).getBeanName();
            }
        }
        return null;
    }

    /**
     * Handles the supplied {@link BeansException} which was thrown as a result of processing an application context
     * from the supplied {@link Bundle}.
     * 
     * @param be The <code>BeansException</code> to handle.
     * @param bundle The bundle that contains the application context
     */
    @SuppressWarnings("unchecked")
    public static void handleBeansException(BeansException be, Bundle bundle) {
        List<Throwable> invertedHierarchy = getInvertedExceptionHierarchy(be);
        for (Throwable throwable : invertedHierarchy) {
            Class throwableClass = throwable.getClass();
            Handler handler = handlers.get(throwableClass);
            if (handler != null) {
                Class<? extends BeansException> handleableClass = throwableClass;
                String beanName = determineBeanName(invertedHierarchy); 
                String bundleDescription = getBundleIdentifier(bundle);
                if (handler.handleException(handleableClass.cast(throwable), beanName, bundleDescription)) {
                    return; 
                }
            }
        }

        LOGGER.log(DeployerLogEvent.BUNDLE_CONTEXT_FAILED, be, getBundleIdentifier(bundle));
    }
    
    private static List<Throwable> getInvertedExceptionHierarchy(Throwable throwable) {
        List<Throwable> hierarchy = new ArrayList<Throwable>();
        Throwable current = throwable;
        do {
            hierarchy.add(current);
        } while ((current = current.getCause()) != null);
        Collections.reverse(hierarchy);
        return hierarchy;
    }

    /**
     * Returns an identifier for the supplied {@link Bundle} which is suitable for inclusion in log messages.
     * 
     * @param bundle The <code>Bundle</code> for which an identifier is required
     * @return The identifier for the <code>Bundle</code>.
     */
    public static String getBundleIdentifier(Bundle bundle) {
        String identifer = (String) bundle.getHeaders().get("Original-Bundle-SymbolicName");
        if (identifer == null) {
            identifer = bundle.getSymbolicName();

        }
        return identifer;
    }
    
    private interface Handler<T extends BeansException> {
        boolean handleException(T e, String beanName, String bundleDescription);
    }
}
