/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.deployer.core.internal;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;

import org.springframework.util.FileCopyUtils;

import com.springsource.server.deployer.core.DeployerLogEvent;
import com.springsource.server.deployer.core.DeploymentException;
import com.springsource.server.deployer.core.FatalDeploymentException;
import com.springsource.server.io.core.JarUtils;
import com.springsource.server.io.core.PathReference;
import com.springsource.server.serviceability.ffdc.Assert;
import com.springsource.server.serviceability.logging.LogService;
import com.springsource.server.serviceability.logging.LogServiceFactory;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * Provides operations for working with the deployment area in the Server work directory.<p/>
 * 
 * Deployment artefacts can be {@link #copyToDeployArea(URI) copied to} and (@link
 * {@link #removeFromDeployArea(URI) removed from} the deploy area. Existence of an artefact in the deploy
 * area can be queried with {@link #existsInDeployArea(URI)}.
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @author Glyn Normington
 * @since 1.0
 */
public final class DeployArea {

    private static final String LIBRARY_DESCRIPTOR_EXTENSION = ".libd";

    private static final String URI_PATH_SEPARATOR = "/";

    private static final Tracer TRACER = TracerFactory.createTracer(DeployArea.class);

    private static final LogService LOGGER = LogServiceFactory.getLogService();

    static final String PACKED_DIR_NAME = "packed";

    /**
     * The directory to use for deployment.
     */
    private final PathReference deployDir;

    /**
     * Create a new <code>DeployArea</code> using the supplied PathReference.
     * 
     * @param deployDir the deploy directory for this <code>DeployArea</code>
     */
    public DeployArea(PathReference deployDir) {
        this.deployDir = deployDir;
    }

    /**
     * Create a deploy area which is a subdirectory of this deploy area named using the given work area name. An
     * existing deploy area can be reused or a new one can be created. Any old contents are deleted if the clean
     * parameter is <code>true</code>.
     * 
     * @param workAreaName the subdirectory name
     * @parm clean <code>true</code> if and only if the work area content is to be deleted
     * @return the new deploy area
     */
    public DeployArea createWorkArea(String workAreaName, boolean clean) {
        return new DeployArea(format(this.deployDir.newChild(workAreaName), clean));
    }

    /**
     * Copies the artefact located at the supplied {@link URI} to the deploy area.
     * 
     * @param artefact the artefact to copy to the deploy area.
     * @return the location of the artefact in the deploy area.
     * @throws DeploymentException if the supplied artefact does not exist, or if it already exists inside the deploy
     *         area.
     */
    public PathReference copyToDeployArea(URI artefact) throws DeploymentException {
        PathReference artefactPath = new PathReference(artefact);
        if (!artefactPath.exists()) {
            LOGGER.log(DeployerLogEvent.ARTEFACT_NOT_FOUND, artefactPath);
            throw new DeploymentException("Artefact not found '" + artefactPath + "'");
        }

        PathReference deployPath = createDeployPathReference(artefact);
        if (deployPath.exists()) {
            TRACER.traceWarning("Deleting existing content from deploy area '%s'.", deployPath);
            deployPath.delete(true);
        }

        if (artefactPath.isDirectory()) {
            TRACER.traceInfo("Copying exploded artefact '%s' to path '%s'.", artefactPath, deployPath);
            deployPath = artefactPath.copy(deployPath, true);
        } else if (artefactPath.getName().endsWith(LIBRARY_DESCRIPTOR_EXTENSION)) {
            TRACER.traceInfo("Copying library descriptor artefact '%s' to path '%s'.", artefactPath, deployPath);
            deployPath = artefactPath.copy(deployPath, false);
        } else {
            PathReference copyPath = copyArtefactToDeployArea(artefact);
            TRACER.traceInfo("Expanding artefact copy '%s' to path '%s'.", copyPath, deployPath);
            try {
                JarUtils.unpackTo(copyPath, deployPath);
            } catch (IOException e) {
                LOGGER.log(DeployerLogEvent.ARTEFACT_UNPACK_ERROR, e, copyPath);
                throw new DeploymentException("Exception unpacking '" + copyPath + "'", e);
            }
            // Leave the packed form around as the web deployer, for instance, may have use of it.
        }
        return deployPath;
    }

    /**
     * @param artefact
     * @param deployPath
     * @throws DeploymentException
     */
    private PathReference copyArtefactToDeployArea(URI artefact) throws DeploymentException {
        try {
            PathReference copyPath = createArtefactCopyPathReference(artefact);

            InputStream is;

            try {
                is = artefact.toURL().openStream();
            } catch (MalformedURLException e) {
                throw new DeploymentException("Artefact URI '" + artefact + "' is invalid.", e);
            } catch (IOException e) {
                throw new DeploymentException("Artefact '" + artefact + "' does not exist.", e);
            }

            TRACER.traceInfo("Copying artefact '%s' to '%s'", artefact, copyPath);

            try {
                copyPath.createFile();
                FileCopyUtils.copy(is, new FileOutputStream(copyPath.toFile()));
            } catch (FileNotFoundException e) {
                throw new DeploymentException("Error copying artefact '" + artefact + "' to deploy area.", e);
            } catch (IOException e) {
                throw new DeploymentException("Unable to copy artefact '" + artefact + "' to '" + copyPath + "'.", e);
            }
            return copyPath;
        } catch (DeploymentException e) {
            LOGGER.log(DeployerLogEvent.ARTEFACT_COPY_ERROR, e.getCause(), artefact);
            throw e;
        }
    }

    /**
     * Removes the artefact located at the supplied {@link URI} from the deploy area.
     * 
     * @param artefact the artefact to remove.
     * @return <code>true</code> if the artefact was in the deploy area and thus removed, otherwise returns
     *         <code>false</code>.
     */
    public boolean removeFromDeployArea(URI artefact) {
        PathReference deployFile = createDeployPathReference(artefact);
        boolean removed = deployFile.delete(true);
        TRACER.traceInfo("File '%s', deleted '%s'", deployFile.toString(), removed);
        if (removed) {
            PathReference copyFile = createArtefactCopyPathReference(artefact);
            boolean copyRemoved = copyFile.delete();
            TRACER.traceInfo("File '%s', deleted '%s'", copyFile.toString(), copyRemoved);
        }
        return removed;
    }

    /**
     * Queries whether the artefact located at the supplied {@link URI} already exists in the deploy area.
     * 
     * @param artefact the artefact
     * @return <code>true</code> if the artefact exists in the deploy area, otherwise <code>false</code>.
     */
    public boolean existsInDeployArea(URI artefact) {
        PathReference deployFile = createDeployPathReference(artefact);
        return deployFile.exists();
    }

    /**
     * Creates the {@link PathReference} pointing to the exploded deploy area location for the supplied artefact.
     * 
     * @param artefact the artefact.
     * @return the deploy area location of the artefact.
     */
    public PathReference createDeployPathReference(URI artefact) {
        return PathReference.concat(this.deployDir.getAbsolutePath(), getName(artefact));
    }

    private static String getName(URI uriPath) {
        String path = uriPath.getPath();
        path = path.replace('/', File.separatorChar);
        return getName(path);
    }

    private static String getName(String path) {
        if (path.endsWith(File.separator)) {
            return getName(path.substring(0, path.length() - 1));
        }
        int separatorIndex = path.lastIndexOf(File.separator);
        return separatorIndex != -1 ? path.substring(separatorIndex + 1) : path;
    }

    /**
     * Creates the {@link PathReference} pointing to the deploy area copy of the supplied artefact.
     * 
     * @param artefact the artefact.
     * @return the deploy area location of the artefact.
     */
    private PathReference createArtefactCopyPathReference(URI artefact) {
        return PathReference.concat(this.deployDir.getAbsolutePath(), PACKED_DIR_NAME, getName(artefact));
    }

    /**
     * Create a directory in this deploy area with the specified name which must not contain path separators.
     * 
     * @param dirName the name of the directory to be created which must not contain path separators
     * @return the PathReference of the created directory
     */
    public PathReference createDirectory(String dirName) {
        Assert.isFalse(dirName.contains(File.separator), "Cannot create a directory with name '" + dirName + "' which contains a path separator");
        StringBuffer dirURI = new StringBuffer(URI_PATH_SEPARATOR);
        dirURI.append(dirName);
        try {
            PathReference dirPath = createDeployPathReference(new URI(dirURI.toString()));
            ensureDirectory(dirPath);
            return dirPath;
        } catch (URISyntaxException e) {
            throw new FatalDeploymentException("Invalid directory name '" + dirName + "'", e);
        }
    }

    private static PathReference format(PathReference stagingDir, boolean clean) {
        if (clean) {
            stagingDir.delete(true);
        }
        ensureDirectory(stagingDir);
        PathReference unpackedDir = stagingDir.newChild(PACKED_DIR_NAME);
        ensureDirectory(unpackedDir);
        return stagingDir;
    }

    private static void ensureDirectory(PathReference ref) {
        if (ref.isFile()) {
            TRACER.traceDebug("Deleting stray file from deploy area location '%s'.", ref);
            ref.delete();
        }
        if (!ref.exists()) {
            TRACER.traceDebug("Creating deploy area directory at '%s'.", ref);
            ref.createDirectory();
        }
    }

    /**
     * Create a file in this deploy area with the specified name which must not contain path separators.
     * 
     * @param fileName the name of the directory to be created which must not contain path separators
     * @return the PathReference of the created directory
     */
    public PathReference createFile(String fileName) {
        Assert.isFalse(fileName.contains(File.separator), "Cannot create a file with name '" + fileName + "' which contains a path separator");
        StringBuffer fileURI = new StringBuffer(URI_PATH_SEPARATOR);
        fileURI.append(fileName);
        try {
            PathReference filePath = createDeployPathReference(new URI(fileURI.toString()));
            ensureFile(filePath);
            return filePath;
        } catch (URISyntaxException e) {
            throw new FatalDeploymentException("Invalid file name '" + fileName + "'", e);
        }
    }

    private static void ensureFile(PathReference ref) {
        if (ref.isDirectory()) {
            TRACER.traceDebug("Deleting stray directory from deploy area location '%s'.", ref);
            ref.delete(true);
        }
    }

    /**
     * Delete the contents of the work area.
     */
    public void clean() {
        this.deployDir.delete(true);
    }

    /**
     * Get the filename of the artefact at the given PathReference in this deploy area.
     * 
     * @param artefact the artefact
     * @return the filename of the artefact
     */
    public String getFilename(PathReference artefact) {
        return getName(artefact.getAbsolutePath());
    }

    /**
     * Rename the artefact at the given PathReference in this deploy area to have the given filename.
     * 
     * @param artefact the artefact to be renamed
     * @param newFilename the new filename to be used
     * @return a new PathReference to the artefact
     */
    public PathReference renameArtefact(PathReference artefact, String newFilename) {
        String oldFilename = getFilename(artefact);
        String dirPath = artefactDirectoryPath(artefact.getAbsolutePath());
        PathReference newArtefact = PathReference.concat(dirPath, newFilename);
        artefact.toFile().renameTo(newArtefact.toFile());
        PathReference oldPackedArtefact = PathReference.concat(dirPath, PACKED_DIR_NAME, oldFilename);
        if (oldPackedArtefact.exists()) {
            PathReference newPackedArtefact = PathReference.concat(dirPath, PACKED_DIR_NAME, newFilename);
            oldPackedArtefact.toFile().renameTo(newPackedArtefact.toFile());
        }
        return newArtefact;
    }
    
    /**
     * Delete the artefact at the given PathReference in this deploy area.
     * 
     * @param artefact the artefact to be deleted
     */
    public void deleteArtefact(PathReference artefact) {
        String filename = getFilename(artefact);
        String dirPath = artefactDirectoryPath(artefact.getAbsolutePath());
        artefact.delete(true);
        PathReference.concat(dirPath, PACKED_DIR_NAME, filename).delete(true);
    }

    private static String artefactDirectoryPath(String artefactPath) {
        if (artefactPath.endsWith(File.separator)) {
            return artefactDirectoryPath(artefactPath.substring(0, artefactPath.length() - 1));
        }
        int separatorIndex = artefactPath.lastIndexOf(File.separator);
        return separatorIndex != -1 ? artefactPath.substring(0, separatorIndex) : null;
    }

}
