/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.deployer.core.internal;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.util.Arrays;
import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.Set;

import org.springframework.util.Assert;

import com.springsource.server.deployer.core.DeployerLogEvent;
import com.springsource.server.deployer.core.DeploymentArtefact;
import com.springsource.server.deployer.core.DeploymentException;
import com.springsource.server.deployer.core.FatalDeploymentException;
import com.springsource.server.io.core.PathReference;
import com.springsource.server.serviceability.logging.LogService;
import com.springsource.server.serviceability.logging.LogServiceFactory;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * {@link DeploymentArtefact} implementation for exploded directories. <p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Thread-safe.
 * 
 * @author Rob Harrop
 * @author Sam Brannen
 * @since 1.0
 */
public final class ExplodedDirectoryDeploymentArtefact extends AbstractDeploymentArtefact {

    private static final Tracer TRACER = TracerFactory.createTracer(ExplodedDirectoryDeploymentArtefact.class);
    
    private static final LogService LOGGER = LogServiceFactory.getLogService();

    private static final String EXPLOSION_ROOT = "/";

    private final PathReference path;

    /**
     * Creates a new <code>ExplodedDirectoryDeploymentArtefact</code> for the supplied location.
     * 
     * @param location the location of the exploded deployment artefact.
     * @throws ArtefactNotFoundException 
     */
    public ExplodedDirectoryDeploymentArtefact(URI location) throws ArtefactNotFoundException {
        super(location);
        this.path = new PathReference(location);
        if (!this.path.exists()) {
            LOGGER.log(DeployerLogEvent.ARTEFACT_NOT_FOUND, location);
            throw new ArtefactNotFoundException("The artefact '" + location + "' was not found");
        }
        Assert.isTrue(this.path.isDirectory(), "ExplodedDirectoryDeploymentArtefact cannot be constructed from a non-exploded file");
    }

    /**
     * {@inheritDoc}
     */
    public Entry getEntry(String entryPath) {
        return new ExplodedEntry(entryPath);
    }

    /**
     * {@inheritDoc}
     */
    public Set<Entry> getContents() {
        return new ExplodedEntry(EXPLOSION_ROOT).children();
    }

    /**
     * {@inheritDoc}
     */
    public Entry createDirectoryEntry(String path) throws IllegalStateException, IllegalArgumentException {
        Assert.notNull(path, "'path' must not be null");
        if (!path.startsWith("/") && path.length() < 2) {
            throw new IllegalArgumentException("'path' must start with a '/'");
        }
        this.path.newChild(path).createDirectory();
        return new ExplodedEntry(path);
    }

    /**
     * {@link DeploymentArtefact.Entry} implementation for exploded directory entries <p/>
     * 
     * <strong>Concurrent Semantics</strong><br />
     * 
     * Thread-safe
     * 
     * @author Rob Harrop
     * @author Sam Brannen
     * @author Glyn Normington
     * @since 1.0
     */
    private class ExplodedEntry implements DeploymentArtefact.Entry {

        private final PathReference path;

        private final String entryPath;

        private ExplodedEntry(String entryPath) {
            this.entryPath = entryPath;
            this.path = ExplodedDirectoryDeploymentArtefact.this.path.newChild(entryPath);
        }

        /**
         * {@inheritDoc}
         */
        public Set<Entry> children() {
            String[] children = this.path.toFile().list();
            Set<Entry> set = new LinkedHashSet<Entry>();
            if (children != null) {
                Arrays.sort(children);
                for (String childName : children) {
                    String newName = this.entryPath + (this.entryPath.endsWith(EXPLOSION_ROOT) ? "" : EXPLOSION_ROOT) + childName;
                    set.add(new ExplodedEntry(newName));
                }
            }
            return Collections.unmodifiableSet(set);
        }

        /**
         * {@inheritDoc}
         */
        public boolean exists() {
            return this.path.exists();
        }

        /**
         * {@inheritDoc}
         */
        public boolean isDirectory() {
            return this.path.isDirectory();
        }

        /**
         * {@inheritDoc}
         */
        public String getName() {
            return this.entryPath;
        }

        /**
         * {@inheritDoc}
         */
        public String getFilename() {
            return this.path.getName();
        }

        /**
         * {@inheritDoc}
         */
        public BufferedReader getReader() throws FatalDeploymentException {
            if (!exists()) {
                throw new FatalDeploymentException("Cannot create Reader - entry '" + getName() + "' does not exist.");
            }
            if (this.path.isDirectory()) {
                throw new FatalDeploymentException("Cannot create Reader for directory entry '" + getName() + "'.");
            }
            try {
                return new BufferedReader(new FileReader(this.path.toFile()));
            } catch (FileNotFoundException e) {
                throw new FatalDeploymentException("Unexpected error creating Reader for entry '" + getName() + "'.", e);
            }
        }

        /**
         * {@inheritDoc}
         */
        public BufferedWriter getWriter(boolean append) {
            if (!exists()) {
                this.path.createFile();
            }
            if (this.path.isDirectory()) {
                throw new FatalDeploymentException("Cannot create Writer for directory entry '" + getName() + "'.");
            }
            try {
                return new BufferedWriter(new FileWriter(this.path.toFile(), append));
            } catch (IOException e) {
                throw new FatalDeploymentException("Unexpected error creating Writer for entry '" + getName() + "'.", e);
            }
        }

        /**
         * {@inheritDoc}
         */
        public boolean delete() {
            return this.path.delete(true);
        }

        /**
         * {@inheritDoc}
         */
        public Entry move(String destination) {
            if (!this.exists()) {
                throw new IllegalArgumentException("Can not move a non-existent entry [" + this + "]");
            }
            Assert.notNull(destination, "'destination' must not be null");
            if (!destination.startsWith("/")) {
                throw new IllegalArgumentException("'destination' must start with a '/'");
            }

            TRACER.traceVerbose("Moving [%s] to [%s].", this.path, destination);

            PathReference target = ExplodedDirectoryDeploymentArtefact.this.path.newChild(destination);
            this.path.copy(target, true);
            this.path.delete(true);

            return new ExplodedEntry(destination);
        }

        /**
         * {@inheritDoc}
         */
        public Entry moveContents(String destination) throws IllegalArgumentException {
            return moveContents(destination, true);
        }

        /**
         * {@inheritDoc}
         */
        public Entry moveContents(String destination, boolean deleteSource) throws IllegalArgumentException {
            if (!exists()) {
                throw new IllegalArgumentException("Can not move a non-existent entry [" + this + "]");
            }
            Assert.notNull(destination, "'destination' must not be null");
            if (!destination.startsWith("/")) {
                throw new IllegalArgumentException("'destination' must start with a '/'");
            }

            for (Entry child : children()) {
                child.move(destination + "/" + child.getFilename());
            }

            if (deleteSource) {
                delete();
            }

            return new ExplodedEntry(destination);
        }

        /**
         * {@inheritDoc}
         */
        public DeploymentArtefact getDeploymentArtefact() throws DeploymentException {
            return this.path.isDirectory() ? new ExplodedDirectoryDeploymentArtefact(this.path.toURI()) : new FileDeploymentArtefact(
                this.path.toURI());
        }

        /**
         * {@inheritDoc}
         */
        @Override public String toString() {
            return getName() + "@" + ExplodedDirectoryDeploymentArtefact.this;
        }

        /**
         * {@inheritDoc}
         */
        public InputStream getInputStream() throws FatalDeploymentException {
            if (!exists()) {
                throw new FatalDeploymentException("Cannot create InputStream - entry '" + getName() + "' does not exist.");
            }
            if (this.path.isDirectory()) {
                throw new FatalDeploymentException("Cannot create InputStream for directory entry '" + getName() + "'.");
            }
            try {
                return new FileInputStream(this.path.toFile());
            } catch (FileNotFoundException e) {
                throw new FatalDeploymentException("Unexpected error creating InputStream for entry '" + getName() + "'.", e);
            }
        }

    }

    /**
     * {@inheritDoc}
     */
    public void copy(PathReference source, String path) {
        PathReference target = this.path.newChild(path);
        if (source.isFile()) {
            if (target.exists()) {
                if (target.isFile()) {
                    target.delete();
                } else {
                    PathReference targetFile = target.newChild(source.getName());
                    if (targetFile.exists() && targetFile.isFile()) {
                        targetFile.delete();
                    }
                }
            }
            source.copy(target);
        } else {
            source.copy(target, true);
        }
    }

}
