/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.deployer.core.internal;

import java.io.File;
import java.net.URI;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeUnit;

import org.osgi.service.packageadmin.PackageAdmin;

import com.springsource.server.concurrent.core.ThreadManager;
import com.springsource.server.deployer.controls.RecoveryMonitor;
import com.springsource.server.deployer.core.ApplicationDeployer;
import com.springsource.server.deployer.core.DeployableFactory;
import com.springsource.server.deployer.core.DeployableInfo;
import com.springsource.server.deployer.core.DeployerLogEvent;
import com.springsource.server.deployer.core.DeploymentException;
import com.springsource.server.deployer.core.DeploymentIdentity;
import com.springsource.server.deployer.core.FatalDeploymentException;
import com.springsource.server.deployer.core.InformedDeployable;
import com.springsource.server.deployer.core.event.ApplicationDeployed;
import com.springsource.server.deployer.core.event.ApplicationDeploying;
import com.springsource.server.deployer.core.event.ApplicationUndeployed;
import com.springsource.server.deployer.core.event.ApplicationUndeploying;
import com.springsource.server.deployer.core.event.DeploymentListener;
import com.springsource.server.io.core.PathReference;
import com.springsource.server.osgi.framework.BundleInstaller;
import com.springsource.server.osgi.framework.UnableToSatisfyDependenciesException;
import com.springsource.server.osgi.provisioning.repository.BundleRepository;
import com.springsource.server.profile.core.Profile;
import com.springsource.server.serviceability.ffdc.Assert;
import com.springsource.server.serviceability.logging.LogService;
import com.springsource.server.serviceability.logging.LogServiceFactory;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;
import com.springsource.server.serviceability.tracing.TracingService;

/**
 * This class provides the main programmatic interface to the deployer.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * This class is thread safe. It provides synchronization of the deployment process and holds an object lock across
 * deployment calls. Therefore personality deployers must not perform any actions that cause applications to be deployed
 * otherwise deadlock could ensue.
 * 
 * @author Glyn Normington
 * @since 1.0
 */
public final class StandardApplicationDeployer implements ApplicationDeployer {

    private static final int WAIT_TIMEOUT_SECONDS = 60;

    private static final String URI_PATH_SEPARATOR = "/";

    private static final Tracer TRACER = TracerFactory.createTracer(StandardApplicationDeployer.class);

    private static final LogService LOGGER = LogServiceFactory.getLogService();

    // The following maps obey the invariant:
    // dom deployableByUri = ran deployableByIdentity
    // forall uri : dom deployableByUri @
    // let deployable = deployableByUri uri within
    // let nv = (deployable.name, deployable.version) within
    // nv in dom deployableByIdentity &
    // deployableByIdentity nv = uri &
    // uri in ran deployableByFileName

    private final Map<URI, InformedDeployable> deployableByUri = new ConcurrentHashMap<URI, InformedDeployable>();

    private final Map<DeploymentIdentity, URI> deployableByIdentity = new ConcurrentHashMap<DeploymentIdentity, URI>();

    private final Map<String, URI> deployableByFileName = new ConcurrentHashMap<String, URI>();

    private final CountDownLatch recoveryLatch = new CountDownLatch(1);

    private final DeployArea deployArea;

    private final DeployableFactory osgiApplicationFactory;

    private final DeployableFactory legacyApplicationFactory;

    private final DeployableFactory libraryFactory;

    private final Object monitor = new Object();

    private final Profile profile;

    private final RecoveryMonitor recoveryMonitor;

    private final BundleRepository bundleRepository;

    private final PackageAdmin packageAdmin;

    private final TracingService tracingService;

    private final DeploymentListener applicationDeploymentEventNotifier;

    private final DeployerRecoveryLog deployerRecoveryLog;

    public StandardApplicationDeployer(DeployableFactory osgiApplicationFactory, DeployableFactory legacyApplicationFactory,
        DeployableFactory libraryFactory, DeployArea deployArea, Profile profile, ThreadManager threadManager, RecoveryMonitor recoveryMonitor,
        BundleInstaller bundleInstaller, PackageAdmin packageAdmin, TracingService tracingService,
        DeploymentListener applicationDeploymentEventNotifier) {
        this.osgiApplicationFactory = osgiApplicationFactory;
        this.legacyApplicationFactory = legacyApplicationFactory;
        this.libraryFactory = libraryFactory;
        this.deployArea = deployArea;
        this.profile = profile;
        this.deployerRecoveryLog = new DeployerRecoveryLog(this.deployArea);
        this.recoveryMonitor = recoveryMonitor;
        this.bundleRepository = bundleInstaller.getBundleRepository();
        this.packageAdmin = packageAdmin;
        this.tracingService = tracingService;
        this.applicationDeploymentEventNotifier = applicationDeploymentEventNotifier;

        waitUntilProfileStarted();

        Thread recoveryThread = threadManager.newThread(new Runnable() {

            public void run() {
                StandardApplicationDeployer.this.recover();
                StandardApplicationDeployer.this.recoveryLatch.countDown();
            }
        }, "Deployer Recovery");
        recoveryThread.start();
    }

    /**
     * {@inheritDoc}
     */
    public boolean awaitRecovery(long timeout, TimeUnit unit) throws InterruptedException {
        return this.recoveryLatch.await(timeout, unit);
    }

    /**
     * Recover the deployer state by redeploying any previously deployed applications. The current implementation is
     * highly conservative and does not depend on any server state being preserved other than the redeploy dataset and
     * the targets of URIs that were previously deployed (either in the hot deploy directory or in user directories).
     * 
     * Any redeployment failures are logged but do not prevent the server from coming up.
     * 
     * @throws DeploymentException
     */
    private void recover() {
        // Read the redeploy dataset.
        Map<URI, DeploymentOptions> recoveryState = this.deployerRecoveryLog.getRecoveryState();

        // Redeploy the applications to be redeployed.
        for (URI uri : recoveryState.keySet()) {
            try {
                this.recoverDeployment(uri, recoveryState.get(uri));
            } catch (DeploymentException e) {
                LOGGER.log(DeployerLogEvent.RECOVERY_FAILED, e, uri.toString());
            } catch (FatalDeploymentException e) {
                LOGGER.log(DeployerLogEvent.RECOVERY_FAILED, e, uri.toString());
            }
        }
        this.recoveryMonitor.signalRecoveryComplete();
    }

    /**
     * Recover a previous deployment of the specified URI.
     * 
     * @param location
     * @param deploymentOptions
     * @throws DeploymentException
     */
    private void recoverDeployment(URI location, DeploymentOptions deploymentOptions) throws DeploymentException {
        try {
            synchronized (this.monitor) {
                Assert.notNull(location, "Cannot recover from a null URI");
                Assert.isFalse(this.deployableByUri.containsKey(location), "Cannot recover a URI which has already been deployed");
                InformedDeployable deployableArtefact = null;
                try {
                    deployableArtefact = recoverDeployableArtefact(location, deploymentOptions);
                    if (deployableArtefact != null) {
                        notifyApplicationDeploying(deployableArtefact);
                        try {
                            deployableArtefact.deploy();
                        } catch (UnableToSatisfyDependenciesException e) {
                            notifyApplicationUndeployed(deployableArtefact);
                            logDependencySatisfactionException(location, e);
                            throw new DeploymentException("Dependency satisfaction failed", e);
                        } catch (RuntimeException e) {
                            notifyApplicationUndeployed(deployableArtefact);
                            throw new FatalDeploymentException("Unexpected exception caught during deployment", e);
                        } catch (DeploymentException e) {
                            notifyApplicationUndeployed(deployableArtefact);
                            throw e;
                        }

                        this.deployableByIdentity.put(getDeploymentIdentity(deployableArtefact), location);
                        this.deployableByUri.put(location, deployableArtefact);
                        this.deployableByFileName.put(getFileName(location), location);

                        notifyApplicationDeployed(deployableArtefact);

                        logCompletion(DeployerLogEvent.DEPLOYMENT_COMPLETED, deployableArtefact);
                    } else {
                        /*
                         * Deployer owned artefact has not been recovered since its source had been deleted, so avoid
                         * recovering it in future.
                         */
                        this.deployerRecoveryLog.remove(location);
                    }
                } catch (UnableToSatisfyDependenciesException utsde) {
                    logDependencySatisfactionException(location, utsde);
                    throw new FatalDeploymentException("Dependency satisfaction failed", utsde);
                } catch (RuntimeException e) {
                    throw new FatalDeploymentException("Unexpected exception during deployment recovery", e);
                } catch (DeploymentException e) {
                    throw new FatalDeploymentException("Recovery failed", e);
                }
            }
        } catch (FatalDeploymentException e) {
            this.deployArea.removeFromDeployArea(location);
            throw e;
        }
    }

    private void waitUntilProfileStarted() {
        if (!this.profile.waitUntilStarted(WAIT_TIMEOUT_SECONDS)) {
            throw new FatalDeploymentException("Timed out waiting for profile subsystems to start");
        }
    }

    /**
     * {@inheritDoc}
     */
    public DeploymentIdentity deploy(URI location) throws DeploymentException {
        return deploy(location, DeploymentOptions.DEFAULT_DEPLOYMENT_OPTIONS);
    }

    /**
     * {@inheritDoc}
     */
    public DeploymentIdentity deploy(URI location, DeploymentOptions deploymentOptions) throws DeploymentException {
        this.bundleRepository.refresh();
        InformedDeployable deployableArtefact = null;
        try {
            synchronized (this.monitor) {
                Assert.notNull(location, "Cannot deploy from a null URI");

                /*
                 * If the location is already deployed, undeploy it and proceed or, in the case of a single bundle
                 * artefact, refresh.
                 * 
                 * In the latter case, refresh is used so that any bundles dependent on it are restarted, but note that
                 * the order of recovery is *not* modified, so redeploying a single bundle artefact where the redeployed
                 * version depends on another bundle deployed after the original version will cause a deployment failure
                 * on server restart. This could be fixed by installing at least all the single bundle artefacts during
                 * recovery before starting any of the artefacts.
                 */
                if (this.deployableByUri.containsKey(location)) {
                    InformedDeployable informedDeployable = this.deployableByUri.get(location);
                    if (informedDeployable.getType() == DeployableInfo.Type.SINGLE_BUNDLE_APPLICATION) {
                        updateDeployArea(location);
                        // Attempt to refresh the given module and continue redeploying normally if this fails.
                        boolean refreshed = false;
                        try {
                            refreshed = informedDeployable.refresh(null, null, null);
                        } catch (DeploymentException e) {
                            TRACER.traceError("Exception thrown by refresh, continuing with normal redeploy", e);
                        }
                        if (refreshed) {
                            // Touch the deployer recovery log so the refreshed application is not deployed and also
                            // redeployed on server restart.
                            this.deployerRecoveryLog.setRedeployFileLastModified();
                            
                            DeploymentIdentity deploymentIdentity = getDeploymentIdentity(informedDeployable);
                            
                            notifyApplicationDeployed(informedDeployable);
                            logCompletion(DeployerLogEvent.DEPLOYMENT_COMPLETED, informedDeployable);
                            
                            return deploymentIdentity;
                        }
                    }
                    undeployByOriginalLocation(location, true);
                }

                String fileName = getFileName(location);
                if (this.deployableByFileName.containsKey(fileName)) {
                    URI clashingDeployableUri = this.deployableByFileName.get(fileName);
                    InformedDeployable clashingDeployable = this.deployableByUri.get(clashingDeployableUri);
                    DeploymentIdentity clashingDeployableIdentity = getDeploymentIdentity(clashingDeployable);
                    LOGGER.log(DeployerLogEvent.CLASHING_LOCATIONS, location, fileName, clashingDeployableIdentity);
                    throw new DeploymentException("Artefact at URI '" + location + "' cannot be deployed as it has the same file or directory name '"
                        + fileName + "' as application " + clashingDeployableIdentity);
                }

                try {
                    deployableArtefact = createDeployableArtefact(location, deploymentOptions);
                } catch (UnableToSatisfyDependenciesException e) {
                    logDependencySatisfactionException(location, e);
                    throw new DeploymentException("Dependency satisfaction failed", e);
                }
                DeploymentIdentity deployableIdentity = getDeploymentIdentity(deployableArtefact);
                if (this.deployableByIdentity.containsKey(deployableIdentity)) {
                    LOGGER.log(DeployerLogEvent.DUPLICATE_APP_AT_DIFFERENT_LOCATION, location, deployableIdentity);
                    throw new DeploymentException("Artefact at URI '" + location + "' cannot be deployed as the application " + deployableIdentity
                        + " is already deployed");
                }

                notifyApplicationDeploying(deployableArtefact);

                try {
                    deployableArtefact.deploy();
                } catch (UnableToSatisfyDependenciesException e) {
                    notifyApplicationUndeployed(deployableArtefact);
                    logDependencySatisfactionException(location, e);
                    throw new DeploymentException("Dependency satisfaction failed", e);
                } catch (RuntimeException e) {
                    notifyApplicationUndeployed(deployableArtefact);
                    throw new FatalDeploymentException("Unexpected exception caught during deployment", e);
                } catch (DeploymentException e) {
                    notifyApplicationUndeployed(deployableArtefact);
                    throw e;
                }
                this.deployableByIdentity.put(deployableIdentity, location);
                this.deployableByUri.put(location, deployableArtefact);
                this.deployableByFileName.put(fileName, location);
                this.deployerRecoveryLog.add(location, deploymentOptions);

                notifyApplicationDeployed(deployableArtefact);

                logCompletion(DeployerLogEvent.DEPLOYMENT_COMPLETED, deployableArtefact);
                return deployableIdentity;
            }
        } catch (FatalDeploymentException e) {
            this.tracingService.applicationUndeployed(getApplicationName(deployableArtefact));
            this.deployArea.removeFromDeployArea(location);
            throw e;
        } catch (DeploymentException e) {
            this.tracingService.applicationUndeployed(getApplicationName(deployableArtefact));
            this.deployArea.removeFromDeployArea(location);
            if (deployableArtefact != null) {
                LOGGER.log(DeployerLogEvent.DEPLOYMENT_FAILED, deployableArtefact.getName(), deployableArtefact.getVersion().toString());
            } else {
                LOGGER.log(DeployerLogEvent.DEPLOYMENT_FAILED_URI, location);
            }
            throw e;
        }
    }

    /**
     * @param deployableArtefact
     */
    private void notifyApplicationDeploying(InformedDeployable deployableArtefact) {
        if (deployableArtefact.getType() != DeployableInfo.Type.LIBRARY) {
            this.applicationDeploymentEventNotifier.onEvent(new ApplicationDeploying(deployableArtefact.getName(), deployableArtefact.getVersion()));
        }
    }

    /**
     * @param deployableArtefact
     */
    private void notifyApplicationDeployed(InformedDeployable deployableArtefact) {
        if (deployableArtefact.getType() != DeployableInfo.Type.LIBRARY) {
            this.applicationDeploymentEventNotifier.onEvent(new ApplicationDeployed(deployableArtefact.getName(), deployableArtefact.getVersion()));
        }
    }

    /**
     * @param deployableArtefact
     */
    private void notifyApplicationUndeploying(InformedDeployable deployableArtefact) {
        if (deployableArtefact.getType() != DeployableInfo.Type.LIBRARY) {
            this.applicationDeploymentEventNotifier.onEvent(new ApplicationUndeploying(deployableArtefact.getName(), deployableArtefact.getVersion()));
        }
    }

    /**
     * @param deployableArtefact
     */
    private void notifyApplicationUndeployed(InformedDeployable deployableArtefact) {
        if (deployableArtefact.getType() != DeployableInfo.Type.LIBRARY) {
            this.applicationDeploymentEventNotifier.onEvent(new ApplicationUndeployed(deployableArtefact.getName(), deployableArtefact.getVersion()));
        }
    }

    /**
     * Get the file or directory name portion of a URI.
     * 
     * @param location the URI
     * @return the file or directory name
     */
    private String getFileName(URI location) {
        String path = location.getPath();
        if (path.endsWith(URI_PATH_SEPARATOR)) {
            path = path.substring(0, path.length() - 1);
        }
        int separatorIndex = path.lastIndexOf(URI_PATH_SEPARATOR);
        return separatorIndex != -1 ? path.substring(separatorIndex + 1) : path;
    }

    private void logDependencySatisfactionException(URI location, UnableToSatisfyDependenciesException ex) {
        LOGGER.log(DeployerLogEvent.UNABLE_TO_SATISFY_CONSTRAINTS, location, ex.getSymbolicName(), ex.getVersion(), ex.getFailureDescription());
    }

    /**
     * Get the name and version of a {@link DeployableInfo} as an ordered pair.
     * 
     * @param deployableInfo the {@link DeployableInfo} to be turned into an ordered pair
     * @return the ordered pair of name and version
     */
    private DeploymentIdentity getDeploymentIdentity(DeployableInfo deployableInfo) {
        return deployableInfo == null ? null : new StandardDeploymentIdentity(deployableInfo.getName(), deployableInfo.getVersion().toString());
    }

    /**
     * Log the given event for the given artefact from the given location.
     * 
     * @param deployableArtefact the deployed artefact
     */
    private void logCompletion(DeployerLogEvent event, InformedDeployable deployableArtefact) {
        LOGGER.log(event, deployableArtefact.getName(), deployableArtefact.getVersion());
    }

    /**
     * Create a deployable artefact with the given deployment options from the file at the given location.
     * 
     * @param sourceLocation
     * @param deploymentOptions
     * @return the deployable artefact
     * @throws DeploymentException
     * @throws UnableToSatisfyDependenciesException
     */
    private InformedDeployable createDeployableArtefact(URI sourceLocation, DeploymentOptions deploymentOptions) throws DeploymentException,
        UnableToSatisfyDependenciesException {
        PathReference copiedArtefact = this.deployArea.copyToDeployArea(sourceLocation);
        try {
            URI staged = copiedArtefact.toFile().toURI();
            return getDeployableFactory(staged).create(staged, sourceLocation, deploymentOptions);
        } catch (RuntimeException e) {
            this.deployArea.removeFromDeployArea(sourceLocation);
            throw e;
        } catch (DeploymentException e) {
            this.deployArea.removeFromDeployArea(sourceLocation);
            throw e;
        }
    }

    /**
     * Optionally recover a deployable artefact from the deploy area using the source URI to identify the artefact. The
     * source URI need no longer be valid unless the artefact is owned by the deployer. Non-recoverable artefacts are
     * "recovered" by deleting them.
     * 
     * @param sourceLocation the source URI of the deployable artefact
     * @param deploymentOptions the {@link DeploymentOptions} of the deployable artefact
     * @return the {@link InformedDeployable} that was previously copied from the source URI or <code>null</code>
     * @throws UnableToSatisfyDependenciesException
     */
    private InformedDeployable recoverDeployableArtefact(URI sourceLocation, DeploymentOptions deploymentOptions)
        throws UnableToSatisfyDependenciesException {
        PathReference copiedArtefact = this.deployArea.createDeployPathReference(sourceLocation);
        try {
            URI staged = copiedArtefact.toFile().toURI();
            InformedDeployable informedDeployable = null;
            if (deploymentOptions.getRecoverable() && (!deploymentOptions.getDeployerOwned() || (new PathReference(sourceLocation)).exists())) {
                informedDeployable = getDeployableFactory(staged).create(staged, sourceLocation, deploymentOptions);
            } else {
                this.deployArea.removeFromDeployArea(sourceLocation);
            }
            return informedDeployable;
        } catch (RuntimeException e) {
            this.deployArea.removeFromDeployArea(sourceLocation);
            throw e;
        } catch (UnableToSatisfyDependenciesException e) {
            this.deployArea.removeFromDeployArea(sourceLocation);
            throw e;
        } catch (DeploymentException e) {
            this.deployArea.removeFromDeployArea(sourceLocation);
            TRACER.traceWarning("An error occurred during recovery of artefact '%s'", sourceLocation, e);
            return null;
        }
    }

    private DeployableFactory getDeployableFactory(URI location) {
        String fileName = new File(location).getName().toLowerCase();
        // Identify libraries by their specific extensions. Exploded directory
        // libraries are not yet supported.
        if (fileName.endsWith(".lib") || fileName.endsWith(".libd")) {
            TRACER.traceInfo("Library detected at '%s'", location);
            return this.libraryFactory;
        }
        /*
         * Note: this test will classify exploded directories as non-legacy which is good enough.
         */
        boolean legacyJar = !fileName.endsWith(".par");
        TRACER.traceInfo(legacyJar ? "Legacy JAR detected at '%s'" : "OSGi application detected at '%s'", location);
        return legacyJar ? this.legacyApplicationFactory : this.osgiApplicationFactory;
    }

    /**
     * {@inheritDoc}
     */
    public void undeploy(String applicationSymbolicName, String version) throws DeploymentException {
        synchronized (this.monitor) {
            DeploymentIdentity deploymentIdentity = new StandardDeploymentIdentity(applicationSymbolicName, version);
            URI location = this.deployableByIdentity.get(deploymentIdentity);
            if (location != null) {
                undeployByOriginalLocation(location, false);
            } else {
                LOGGER.log(DeployerLogEvent.UNDEPLOY_ARTEFACT_NOT_FOUND, applicationSymbolicName, version);
                throw new DeploymentException("Undeploy not possible as no application with name " + applicationSymbolicName + " and version "
                    + version + " is deployed");
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    public void undeploy(DeploymentIdentity deploymentIdentity) throws DeploymentException {
        undeploy(deploymentIdentity.getSymbolicName(), deploymentIdentity.getVersion());
    }

    /**
     * @param location
     * @param undeployingDuringDeploy
     * @throws DeploymentException
     */
    private void undeployByOriginalLocation(URI location, boolean undeployingDuringDeploy) throws DeploymentException {
        InformedDeployable deployableArtefact = this.deployableByUri.remove(location);

        notifyApplicationUndeploying(deployableArtefact);

        try {
            this.deployerRecoveryLog.remove(location);
            this.deployableByIdentity.remove(getDeploymentIdentity(deployableArtefact));
            this.deployableByFileName.remove(getFileName(location));
            deployableArtefact.undeploy();
            // This call must be made after the application's bundles have been stopped and uninstalled
            this.tracingService.applicationUndeployed(getApplicationName(deployableArtefact));

            // Clean up files
            this.deployArea.removeFromDeployArea(location);
            if (deployableArtefact.getDeploymentOptions().getDeployerOwned() && !undeployingDuringDeploy) {
                new PathReference(location).delete(true);
            }

            this.packageAdmin.refreshPackages(null);

            notifyApplicationUndeployed(deployableArtefact);

            logCompletion(DeployerLogEvent.UNDEPLOYMENT_COMPLETED, deployableArtefact);
        } catch (FatalDeploymentException e) {
            LOGGER.log(DeployerLogEvent.UNDEPLOYMENT_FAILED, e, deployableArtefact.getName(), deployableArtefact.getVersion());
            throw e;
        }
    }

    /**
     * {@inheritDoc}
     */
    public DeploymentIdentity[] getDeploymentIdentities() {
        synchronized (this.monitor) {
            return this.deployableByIdentity.keySet().toArray(new DeploymentIdentity[this.deployableByIdentity.keySet().size()]);
        }
    }

    /**
     * {@inheritDoc}
     */
    public DeploymentIdentity getDeploymentIdentity(URI location) {
        synchronized (this.monitor) {
            return getDeploymentIdentity(this.deployableByUri.get(location));
        }
    }

    /**
     * {@inheritDoc}
     */
    public void refresh(URI location, String symbolicName) throws DeploymentException {
        synchronized (this.monitor) {
            if (this.deployableByUri.containsKey(location)) {
                InformedDeployable deployableArtefact = this.deployableByUri.get(location);
                try {
                    updateDeployArea(location);
                    // Attempt to refresh the given module and escalate to redeploying the whole deployable if this
                    // fails.
                    boolean refreshed = false;
                    try {
                        refreshed = deployableArtefact.refresh(symbolicName, null, null);
                    } catch (DeploymentException e) {
                        TRACER.traceError("Exception thrown by refresh: escalating to redeploy", e);
                    }
                    // Touch the deployer recovery log so the refreshed application is not deployed and also redeployed
                    // on server restart.
                    if (!refreshed || !this.deployerRecoveryLog.setRedeployFileLastModified()) {
                        deploy(location);
                    }
                    LOGGER.log(DeployerLogEvent.REFRESH_COMPLETED, symbolicName, deployableArtefact.getName(), deployableArtefact.getVersion());
                } catch (FatalDeploymentException e) {
                    LOGGER.log(DeployerLogEvent.REFRESH_FAILED, e, symbolicName, deployableArtefact.getName(), deployableArtefact.getVersion());
                    throw e;
                }
            } else {
                TRACER.traceInfo("Refresh of application that is not deployed treated as a request to deploy the application");
                deploy(location);
            }
        }
    }

    /**
     * Update the copy of the given artefact in the deploy area.
     * 
     * @param sourceLocation the location of the artefact to be updated
     * @return the {@link PathReference} of the updated copy
     * @throws DeploymentException
     */
    private PathReference updateDeployArea(URI sourceLocation) throws DeploymentException {
        this.deployArea.removeFromDeployArea(sourceLocation);
        return this.deployArea.copyToDeployArea(sourceLocation);
    }

    public static String getApplicationName(DeployableInfo info) {
        if (info != null) {
            return info.getName() + "-" + info.getVersion().toString();
        } else {
            return null;
        }
    }

    /**
     * {@inheritDoc}
     */
    public boolean isDeployed(URI location) {
        return getDeploymentIdentity(location) != null
            && new PathReference(location).toFile().lastModified() < this.deployerRecoveryLog.getRedeployFileLastModified();
    }

    /**
     * {@inheritDoc}
     */
    public void updateResource(DeploymentIdentity deploymentIdentity, URI resourceLocation, String targetPath) throws DeploymentException {
        updateResource(deploymentIdentity, null, resourceLocation, targetPath);
    }

    /**
     * {@inheritDoc}
     */
    public void updateResource(DeploymentIdentity deploymentIdentity, String symbolicName, URI resourceLocation, String targetPath)
        throws DeploymentException {
        synchronized (this.monitor) {
            URI location = this.deployableByIdentity.get(deploymentIdentity);
            if (location != null) {
                InformedDeployable deployableArtefact = this.deployableByUri.get(location);
                if (deployableArtefact.getDeploymentOptions().getRecoverable()) {
                    LOGGER.log(DeployerLogEvent.UPDATE_ARTEFACT_RECOVERABLE, deploymentIdentity.getSymbolicName(), deploymentIdentity.getVersion());
                    throw new DeploymentException("Resource update not possible as application with name " + deploymentIdentity.getSymbolicName()
                        + " and version " + deploymentIdentity.getVersion() + " is recoverable");
                }
                deployableArtefact.updateResource(symbolicName, resourceLocation, targetPath);
            } else {
                LOGGER.log(DeployerLogEvent.UPDATE_ARTEFACT_NOT_FOUND, deploymentIdentity.getSymbolicName(), deploymentIdentity.getVersion());
                throw new DeploymentException("Resource update not possible as no application with name " + deploymentIdentity.getSymbolicName()
                    + " and version " + deploymentIdentity.getVersion() + " is deployed");
            }
        }

    }

    /**
     * {@inheritDoc}
     */
    public void deleteResource(DeploymentIdentity deploymentIdentity, String targetPath) throws DeploymentException {
        deleteResource(deploymentIdentity, null, targetPath);
    }

    /**
     * {@inheritDoc}
     */
    public void deleteResource(DeploymentIdentity deploymentIdentity, String symbolicName, String targetPath) throws DeploymentException {
        synchronized (this.monitor) {
            URI location = this.deployableByIdentity.get(deploymentIdentity);
            if (location != null) {
                InformedDeployable deployableArtefact = this.deployableByUri.get(location);
                if (deployableArtefact.getDeploymentOptions().getRecoverable()) {
                    LOGGER.log(DeployerLogEvent.UPDATE_ARTEFACT_RECOVERABLE, deploymentIdentity.getSymbolicName(), deploymentIdentity.getVersion());
                    throw new DeploymentException("Resource update not possible as application with name " + deploymentIdentity.getSymbolicName()
                        + " and version " + deploymentIdentity.getVersion() + " is recoverable");
                }
                deployableArtefact.deleteResource(symbolicName, targetPath);
            } else {
                LOGGER.log(DeployerLogEvent.UPDATE_ARTEFACT_NOT_FOUND, deploymentIdentity.getSymbolicName(), deploymentIdentity.getVersion());
                throw new DeploymentException("Resource update not possible as no application with name " + deploymentIdentity.getSymbolicName()
                    + " and version " + deploymentIdentity.getVersion() + " is deployed");
            }
        }

    }

}
