/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.deployer.core.internal;

import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import com.springsource.server.common.math.Version;
import com.springsource.server.deployer.core.CompletableDeploymentContext;
import com.springsource.server.deployer.core.FatalDeploymentException;
import com.springsource.server.deployer.core.InformedDeployable;
import com.springsource.server.deployer.core.ModuleDeployer;
import com.springsource.server.deployer.core.ModuleDeployerContext;

/**
 * {@link StandardDeploymentContext} provides a per-deployment context to contain personality deployer per-deployment
 * state.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * This class is thread safe.
 * 
 * @author Glyn Normington
 * @since 1.0
 */
class StandardDeploymentContext implements CompletableDeploymentContext {

    private Map<ModuleDeployer, ModuleDeployerContext> moduleDeployerContexts = new ConcurrentHashMap<ModuleDeployer, ModuleDeployerContext>();
    
    private final String applicationSymbolicName;
    
    private final Version applicationVersion;
    
    /**
     * 
     */
    public StandardDeploymentContext(InformedDeployable informedDeployable) {
        this.applicationSymbolicName = informedDeployable.getName();
        this.applicationVersion = informedDeployable.getVersion();
    }

    /**
     * {@inheritDoc}
     */
    public ModuleDeployerContext getModuleDeployerContext(ModuleDeployer moduleDeployer) {
        return this.moduleDeployerContexts.get(moduleDeployer);
    }

    /**
     * {@inheritDoc}
     */
    public void registerModuleDeployerContext(ModuleDeployer moduleDeployer, ModuleDeployerContext moduleDeployerContext) {
        if (this.moduleDeployerContexts.put(moduleDeployer, moduleDeployerContext) != null) {
            throw new FatalDeploymentException("A module deployer context was registered multiple times by the '" + moduleDeployer.getPersonality()
                + "' personality deployer");
        }

    }

    /**
     * {@inheritDoc}
     */
    public void completed() {
        for (ModuleDeployerContext deploymentListener : this.moduleDeployerContexts.values()) {
            deploymentListener.completed();
        }
    }

    /** 
     * {@inheritDoc}
     */
    public String getApplicationSymbolicName() {
        return this.applicationSymbolicName;
    }

    /** 
     * {@inheritDoc}
     */
    public Version getApplicationVersion() {
        return this.applicationVersion;
    }

}
