/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.deployer.core.internal;

import java.util.Set;

import com.springsource.server.deployer.core.DeploymentContext;
import com.springsource.server.deployer.core.Module;
import com.springsource.server.deployer.core.ModuleDeployer;
import com.springsource.server.deployer.core.ModuleDeployerSet;
import com.springsource.server.serviceability.ffdc.Assert;


/**
 * {@link ModuleDeployerSet} encapsulates a set of {@link ModuleDeployer module deployers} and provides the means of
 * finding the module deployer for a module.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * TODO Document concurrent semantics of ModuleDeployerSet
 * 
 * @author Glyn Normington
 * @since 1.0
 */
class StandardModuleDeployerSet implements ModuleDeployerSet {

    private final Set<ModuleDeployer> deployers;
    
    private final ModuleDeployer defaultDeployer = new BundleDeployer();

    StandardModuleDeployerSet(Set<ModuleDeployer> deployers) {
        this.deployers = deployers;
    }
    
    /** 
     * {@inheritDoc}
     */
    public ModuleDeployer determinePersonality(DeploymentContext deploymentContext, Module module) {
        boolean deployerFound = false;
        ModuleDeployer personalityDeployer = null;
        /*
         * Find the module deployer, but don't stop looking after the first one is found as we want to detect the error
         * where multiple module deployers claim to be able to deploy the same module. Note that this error is detected
         * by protocol checks in the ModuleData implementation.
         */
        for (ModuleDeployer moduleDeployer : deployers) {
            if (moduleDeployer.canDeploy(deploymentContext, module)) {
                deployerFound = true;
                personalityDeployer = moduleDeployer;
            }
        }
        if (this.defaultDeployer.canDeploy(deploymentContext, module)) {
            deployerFound = true;
            personalityDeployer = this.defaultDeployer;
        }
        Assert.isTrue(deployerFound, "No deployer found for module '%s'", module);
        return personalityDeployer;
    }
}
