/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.deployer.core.module;

import java.net.URI;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import com.springsource.server.deployer.core.CompletableDeploymentContext;
import com.springsource.server.deployer.core.DeploymentException;
import com.springsource.server.deployer.core.Module;
import com.springsource.server.deployer.core.ModuleData;
import com.springsource.server.deployer.core.ModuleDeployerSet;
import com.springsource.server.deployer.core.ModuleImportExpander;
import com.springsource.server.deployer.core.ModulePersonality;
import com.springsource.server.deployer.core.ModuleScoper;
import com.springsource.server.osgi.framework.ImportMergeException;
import com.springsource.server.osgi.framework.UnableToSatisfyDependenciesException;
import com.springsource.server.osgi.manifest.BundleManifest;
import com.springsource.server.osgi.manifest.BundleManifestFactory;
import com.springsource.server.osgi.manifest.Scoper;
import com.springsource.server.osgi.manifest.BundleManifest.ExportPackageHeader;
import com.springsource.server.osgi.manifest.BundleManifest.ImportPackageHeader;
import com.springsource.server.osgi.manifest.BundleManifest.ExportPackageHeader.PackageExport;
import com.springsource.server.osgi.manifest.BundleManifest.ImportPackageHeader.PackageImport;
import com.springsource.server.serviceability.ffdc.Assert;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * {@link AbstractModule} holds a module's {@link ModuleData} and delegates personality-specific methods to an
 * associated {@link ModulePersonality}.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * This class is thread safe.
 * 
 * @author Glyn Normington
 * @since 1.0
 */
public abstract class AbstractModule implements Module {

    private static final String KERNEL_DM_FRAGMENT_PACKAGE = "com.springsource.server.kernel.dm";

    private static final Tracer TRACER = TracerFactory.createTracer(AbstractModule.class);

    private final ModuleData moduleData;

    private final BundleManifestFactory bundleManifestFactory;

    /**
     * Invariant: personalityModule == null => transformed = false && deployed == false && <br/> transformed == false =>
     * deployed == false
     */
    private ModulePersonality personalityModule;

    private boolean transformed = false;

    private boolean deployed = false;

    private ModuleScoper moduleScoper;

    private ModuleImportExpander moduleImportExpander;

    private ModulePersonality stackedPersonalityModule;

    private final Object monitor = new Object();

    public AbstractModule(ModuleData moduleData, BundleManifestFactory bundleManifestFactory) {
        this.moduleData = moduleData;
        this.bundleManifestFactory = bundleManifestFactory;
    }

    public ModuleData getModuleData() {
        synchronized (this.monitor) {
            return this.moduleData;
        }
    }

    /**
     * {@inheritDoc}
     */
    public final void setPersonality(ModulePersonality personalityModule) {
        synchronized (this.monitor) {
            Assert.isNull(this.personalityModule, "Module personality already set");
            this.personalityModule = personalityModule;
        }
    }

    /**
     * {@inheritDoc}
     */
    public boolean hasPersonality() {
        synchronized (this.monitor) {
            return this.personalityModule != null;
        }
    }
    
    /**
     * @throws DeploymentException
     */
    protected abstract void normaliseManifest() throws DeploymentException;

    /**
     * @throws DeploymentException {@inheritDoc}
     */
    public void transform() throws DeploymentException {
        ModulePersonality persMod;
        synchronized (this.monitor) {
            persMod = this.personalityModule;
            Assert.notNull(persMod, "A module without a personality cannot be transformed");
            Assert.isFalse(this.transformed, "A module cannot be transformed more than once");
        }
        persMod.transform();
        synchronized (this.monitor) {
            addStandardImports();
            this.transformed = true;
        }
    }

    /**
     * Add the package imports required by the generic deployer.
     */
    private void addStandardImports() {
        BundleManifest bundleManifest = this.moduleData.getBundleManifest();
        ImportPackageHeader importPackageHeader = bundleManifest.getImportPackageHeader();
        List<PackageImport> packageImports;
        if (importPackageHeader != null) {
            packageImports = importPackageHeader.getPackageImports();
        } else {
            packageImports = new ArrayList<PackageImport>(1);
            ImportPackageHeader createdImportPackageHeader = this.bundleManifestFactory.createImportPackageHeader(packageImports);
            bundleManifest.setImportPackageHeader(createdImportPackageHeader);
        }

        PackageImport packageImport = this.bundleManifestFactory.createPackageImport();
        packageImport.getPackageNames().add(KERNEL_DM_FRAGMENT_PACKAGE);
        packageImports.add(packageImport);
    }

    /**
     * {@inheritDoc}
     */
    public void onStarting() throws DeploymentException {
        ModulePersonality persMod;
        synchronized (this.monitor) {
            Assert.notNull(this.personalityModule, "A module without a personality cannot be be driven for onStarting");
            Assert.isTrue(this.transformed, "A module must be transformed before being driven for onStarting");
            persMod = this.personalityModule;
        }
        persMod.onStarting();
    }

    /**
     * {@inheritDoc}
     */
    public void onStopped() throws DeploymentException {
        ModulePersonality persMod;
        synchronized (this.monitor) {
            Assert.notNull(this.personalityModule, "A module without a personality cannot be be driven for onStopped");
            Assert.isTrue(this.transformed, "A module must be transformed before being driven for onStopped");
            persMod = this.personalityModule;
        }
        persMod.onStopped();
    }

    /**
     * {@inheritDoc}
     */
    public final void onUnresolved() {
        synchronized (this.monitor) {
            popModulePersonality();
        }
    }

    /**
     * Push the new module personality during refresh.
     */
    private void pushModulePersonality() {
        Assert.isNull(this.stackedPersonalityModule, "Module personality already pushed");
        Assert.isFalse(this.deployed, "Module personality cannot be pushed after deploy called");
        this.stackedPersonalityModule = this.personalityModule;
    }

    /**
     * This method only has an effect during refresh when there is a new module personality ready to be switched in.
     */
    private void popModulePersonality() {
        if (this.stackedPersonalityModule != null) {
            this.personalityModule = this.stackedPersonalityModule;
            this.stackedPersonalityModule = null;
            this.deployed = false;
        }
    }

    /**
     * Clear the refresh state when refresh has failed or is being escalated.
     */
    protected final void clearStackedModulePersonality() {
        synchronized (this.monitor) {
            this.stackedPersonalityModule = null;
        }
    }

    /**
     * {@inheritDoc}
     */
    public void deploy() throws DeploymentException, UnableToSatisfyDependenciesException {
        ModulePersonality persMod;
        synchronized (this.monitor) {
            Assert.isTrue(this.transformed, "A module cannot be deployed before being transformed");
            Assert.isFalse(this.deployed, "A module cannot be deployed more than once");
            persMod = this.personalityModule;
        }
        persMod.deploy();
        synchronized (this.monitor) {
            this.deployed = true;
        }
    }

    /**
     * {@inheritDoc}
     */
    public void start() throws DeploymentException {
        ModulePersonality persMod;
        synchronized (this.monitor) {
            Assert.isTrue(this.deployed, "A module cannot be started before being deployed");
            persMod = this.personalityModule;
        }
        persMod.start();
    }

    /**
     * {@inheritDoc}
     */
    public void stop() throws DeploymentException {
        ModulePersonality persMod;
        synchronized (this.monitor) {
            Assert.isTrue(this.deployed, "A module cannot be stopped before being deployed");
            persMod = this.personalityModule;
        }
        persMod.stop();
    }

    /**
     * {@inheritDoc}
     */
    public void undeploy() throws DeploymentException {
        ModulePersonality persMod = null;
        synchronized (this.monitor) {
            if (this.deployed) {
                persMod = this.personalityModule;
            }
        }
        if (persMod != null) {
            persMod.undeploy();
            synchronized (this.monitor) {
                this.deployed = false;
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    public String getPersonality() {
        ModulePersonality persMod;
        synchronized (this.monitor) {
            persMod = this.personalityModule;
        }
        return persMod != null ? persMod.getPersonality() : null;
    }

    /**
     * {@inheritDoc}
     */
    public String getPersonalityIdentifier() {
        ModulePersonality persMod;
        synchronized (this.monitor) {
            persMod = this.personalityModule;
        }
        return persMod != null ? persMod.getPersonalityIdentifier() : null;
    }

    /**
     * {@inheritDoc}
     */
    public String getUnscopedSymbolicName() {
        synchronized (this.monitor) {
            return Scoper.getUnscopedSymbolicName(this.moduleData.getBundleManifest());
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override public String toString() {
        synchronized (this.monitor) {
            return this.moduleData != null ? this.moduleData.toString() : super.toString();
        }
    }

    /**
     * {@inheritDoc}
     */
    public void setScoper(ModuleScoper moduleScoper) {
        synchronized (this.monitor) {
            this.moduleScoper = moduleScoper;
        }
    }

    /**
     * {@inheritDoc}
     */
    public void setImportExpander(ModuleImportExpander moduleImportExpander) {
        synchronized (this.monitor) {
            this.moduleImportExpander = moduleImportExpander;
        }
    }

    /**
     * {@inheritDoc}
     */
    public boolean refresh(String symbolicName, ModuleDeployerSet moduleDeployerSet, CompletableDeploymentContext deploymentContext)
        throws DeploymentException {
        boolean proceed = true;
        final ModulePersonality oldModulePersonality;
        synchronized (this.monitor) {
            /*
             * A post-condition of refresh is that this.personalityModule, this.transformed, and this.deployed are equal
             * to the values on entry to refresh.
             */
            oldModulePersonality = this.personalityModule;
            Assert.isTrue(this.transformed, "Refresh on untransformed module '%s'", this);
            Assert.isTrue(this.deployed, "Refresh on undeployed module '%s'", this);
        }
        try {
            ExportPackageHeader oldExportPackageHeader;
            synchronized (this.monitor) {
                this.personalityModule = null;
                this.transformed = false;
                this.deployed = false;
                oldExportPackageHeader = this.moduleData.getBundleManifest().getExportPackageHeader();
                this.moduleData.refreshContents();
                normaliseManifest();
            }
            moduleDeployerSet.determinePersonality(deploymentContext, this);
            synchronized (this.monitor) {
                if (!oldModulePersonality.getPersonality().equals(this.personalityModule.getPersonality())) {
                    TRACER.traceInfo("Incremental refresh of '%s' not possible as its personality has changed - escalating refresh", this);
                    proceed = false;
                }
            }
            if (proceed) {
                synchronized (this.monitor) {
                    try {
                        proceed = expandImportsOnRefresh();
                    } catch (UnableToSatisfyDependenciesException e) {
                        proceed = false;
                    } catch (ImportMergeException e) {
                        proceed = false;
                    }
                }
                if (proceed) {
                    transform();
                    synchronized (this.monitor) {
                        if (moduleScoper != null) {
                            moduleScoper.rescope(this);
                        }
                        // Replace the bundle manifest with any updates.
                        this.moduleData.replaceManifest();
                        // Check whether the bundle exports have changed.
                        ExportPackageHeader newExportPackageHeader = this.moduleData.getBundleManifest().getExportPackageHeader();
                        if (!getPackageExportSet(oldExportPackageHeader).equals(getPackageExportSet(newExportPackageHeader))) {
                            TRACER.traceWarning("Refresh packages of '%s' not possible as the exports appear to have changed - escalating refresh",
                                this);
                            proceed = false;
                        } else {
                            pushModulePersonality();
                        }
                    }
                }
            }
        } finally {
            this.personalityModule = oldModulePersonality;
            this.transformed = true;
            this.deployed = true;
        }
        return proceed;
    }

    /**
     * Expand the bundle and library imports in this module on refresh.
     * 
     * @throws UnableToSatisfyDependenciesException 
     */
    private boolean expandImportsOnRefresh() throws UnableToSatisfyDependenciesException, ImportMergeException {
        return this.moduleImportExpander == null || this.moduleImportExpander.expandImports(this, this.moduleScoper.getScoper());
    }

    /**
     * Convert a given {@link ExportPackageHeader} into a set of {@link PackageExport PackageExports}.
     * 
     * @param exportPackageHeader the {@link ExportPackageHeader} to be converted
     * @return the set of {@link PackageExport PackageExports} corresponding to the given {@link ExportPackageHeader}
     */
    private Set<PackageExport> getPackageExportSet(ExportPackageHeader exportPackageHeader) {
        Set<PackageExport> packageExports = new HashSet<PackageExport>();
        if (exportPackageHeader != null) {
            packageExports.addAll(exportPackageHeader.getPackageExports());
        }
        return packageExports;
    }

    /** 
     * {@inheritDoc}
     */
    public void updateResource(String symbolicName, URI resourceLocation, String targetPath) throws DeploymentException {
        this.moduleData.updateResource(resourceLocation, targetPath);
    }
    
    /** 
     * {@inheritDoc}
     */
    public void deleteResource(String symbolicName, String targetPath) throws DeploymentException {
        this.moduleData.deleteResource(targetPath);
    }
    
}
