/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.deployer.core.module;

import java.net.URI;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.ReentrantLock;

import org.osgi.framework.Bundle;
import org.osgi.framework.BundleException;
import org.osgi.framework.FrameworkEvent;
import org.osgi.framework.FrameworkListener;
import org.osgi.framework.Version;
import org.osgi.service.packageadmin.PackageAdmin;
import org.springframework.beans.BeansException;
import org.springframework.context.ApplicationContext;

import com.springsource.server.deployer.core.CompletableDeploymentContext;
import com.springsource.server.deployer.core.DeployerConfiguration;
import com.springsource.server.deployer.core.DeployerLogEvent;
import com.springsource.server.deployer.core.DeploymentException;
import com.springsource.server.deployer.core.FatalDeploymentException;
import com.springsource.server.deployer.core.InformedDeployable;
import com.springsource.server.deployer.core.Module;
import com.springsource.server.deployer.core.ModuleDeployerSet;
import com.springsource.server.deployer.core.event.DeploymentListener;
import com.springsource.server.deployer.core.event.ApplicationBundleDeployed;
import com.springsource.server.deployer.core.event.ApplicationBundleDeploying;
import com.springsource.server.deployer.core.event.ApplicationBundleStarted;
import com.springsource.server.deployer.core.event.ApplicationBundleStarting;
import com.springsource.server.deployer.core.event.ApplicationBundleStopped;
import com.springsource.server.deployer.core.event.ApplicationBundleStopping;
import com.springsource.server.deployer.core.event.ApplicationBundleUndeployed;
import com.springsource.server.deployer.core.event.ApplicationBundleUndeploying;
import com.springsource.server.deployer.core.internal.DeployArea;
import com.springsource.server.deployer.core.internal.StandardApplicationDeployer;
import com.springsource.server.kernel.core.BundleStarter;
import com.springsource.server.osgi.framework.BundleInstaller;
import com.springsource.server.osgi.framework.OsgiFramework;
import com.springsource.server.osgi.framework.OsgiFrameworkException;
import com.springsource.server.osgi.framework.UnableToSatisfyBundleDependenciesException;
import com.springsource.server.osgi.framework.UnableToSatisfyDependenciesException;
import com.springsource.server.osgi.manifest.BundleManifest;
import com.springsource.server.osgi.manifest.BundleManifestFactory;
import com.springsource.server.osgi.manifest.BundleManifest.BundleActivationPolicy;
import com.springsource.server.osgi.manifest.BundleManifest.BundleSymbolicName;
import com.springsource.server.serviceability.ffdc.Assert;
import com.springsource.server.serviceability.logging.LogService;
import com.springsource.server.serviceability.logging.LogServiceFactory;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;
import com.springsource.server.serviceability.tracing.TracingService;

import edu.emory.mathcs.backport.java.util.concurrent.BlockingDeque;
import edu.emory.mathcs.backport.java.util.concurrent.LinkedBlockingDeque;

/**
 * {@link StandardModule} provides the OSGi bundle capabilities of a module.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Implementations of this interface need not be thread safe. It is the
 * responsibility of callers to synchronize access.
 * 
 * @author Glyn Normington
 * @author Sam Brannen
 * @since 1.0
 */
final class StandardModule extends AbstractModule {

    private static final long REFRESH_WAIT_INTERVAL_MS = 500;

    private static final int BUNDLE_MANIFEST_VERSION_FOR_OSGI_R4 = 2;

    private static final Tracer TRACER = TracerFactory.createTracer(StandardModule.class);

    private static final LogService LOGGER = LogServiceFactory.getLogService();

    private static final String WRAPPED_NULL = "WRAPPED_NULL";

    private Bundle moduleBundle;

    private Module context;

    private final BundleInstaller bundleInstaller;

    private final BundleManifestFactory bundleManifestFactory;

    private boolean fragment = false;

    private BundleActivationPolicy bundleActivationPolicy = null;

    private final BlockingDeque contextClassLoaderStack = new LinkedBlockingDeque();

    private final BlockingDeque applicationTraceNameStack = new LinkedBlockingDeque();

    private final OsgiFramework osgi;

    private final StandardModuleRegistry standardModuleRegistry;

    private final PackageAdmin packageAdmin;

    private volatile boolean refreshComplete;

    private final TracingService tracingService;

    private final InformedDeployable informedDeployable;

    private String applicationTraceName;

    private final DeployerConfiguration deployerConfiguration;

    private final BundleStarter bundleStarter;

    private final DeploymentListener applicationDeploymentEventNotifier;

    private volatile Future<ApplicationContext> startFuture;

    private volatile BundleStartedFuture startedFuture;

    public StandardModule(InformedDeployable informedDeployable, DeployArea deployArea, URI bundleLocation,
            BundleInstaller bundleInstaller, BundleManifestFactory bundleManifestFactory, OsgiFramework osgi,
            StandardModuleRegistry standardModuleRegistry, PackageAdmin packageAdmin, TracingService tracingService,
            DeployerConfiguration deployerConfiguration, BundleStarter bundleStarter,
            DeploymentListener applicationDeploymentEventNotifier) throws DeploymentException {
        super(new StandardModuleData(informedDeployable, bundleLocation, deployArea), bundleManifestFactory);
        this.context = this;
        this.bundleInstaller = bundleInstaller;
        this.bundleManifestFactory = bundleManifestFactory;
        this.osgi = osgi;
        this.standardModuleRegistry = standardModuleRegistry;
        this.packageAdmin = packageAdmin;
        this.tracingService = tracingService;
        Assert.notNull(informedDeployable, "InformedDeployable must not be null");
        this.informedDeployable = informedDeployable;
        refreshApplicationTraceName();
        this.deployerConfiguration = deployerConfiguration;
        this.bundleStarter = bundleStarter;
        this.applicationDeploymentEventNotifier = applicationDeploymentEventNotifier;
        normaliseManifest();
    }

    /**
     * {@inheritDoc}
     */
    public void refreshApplicationTraceName() {
        this.applicationTraceName = StandardApplicationDeployer.getApplicationName(this.informedDeployable);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void transform() throws DeploymentException {

        super.transform();
    }

    /**
     * @throws DeploymentException
     */
    protected void normaliseManifest() throws DeploymentException {
        BundleManifest bundleManifest = this.getModuleData().getBundleManifest();

        /*
         * Upgrade the bundle manifest version to OSGi R4 if it is older. Do not
         * issue a log message to avoid flooding the log.
         * 
         * This upgrade is necessary to support application scoping.
         */
        if (bundleManifest.getBundleManifestVersion() < BUNDLE_MANIFEST_VERSION_FOR_OSGI_R4) {
            TRACER.traceWarning("OSGi R3 or earlier manifest detected:\n'" + bundleManifest + "'\nfor module '" + this
                    + "' - upgrading manifest to OSGi R4.");
            bundleManifest.setBundleManifestVersion(BUNDLE_MANIFEST_VERSION_FOR_OSGI_R4);
            if (bundleManifest.getBundleSymbolicName() == null) {
                BundleSymbolicName bundleSymbolicName = this.bundleManifestFactory.createBundleSymbolicName(this.getModuleData().getFilename());
                bundleManifest.setBundleSymbolicName(bundleSymbolicName);
            }
        }
        else {
            if (bundleManifest.getBundleSymbolicName() == null) {
                LOGGER.log(DeployerLogEvent.MISSING_BUNDLE_SYMBOLIC_NAME, bundleManifest, this);
                throw new DeploymentException("OSGi R4 manifest with no bundle symbolic name detected:\n"
                        + bundleManifest + "'\nfor module '" + this + "'.");
            }
        }

        // Detect whether the bundle is a fragment to avoid starting and
        // stopping it.
        this.fragment = bundleManifest.getFragmentHostHeader() != null;

        this.bundleActivationPolicy = bundleManifest.getBundleActivationPolicy();

        // Expand semicolon separated package lists to avoid Equinox bug.
        bundleManifest.expandPackageLists();

        // Remove any bundle update location header to avoid it interfering with
        // refresh.
        String bundleUpdateLocation = bundleManifest.getBundleUpdateLocation();
        if (bundleUpdateLocation != null) {
            LOGGER.log(DeployerLogEvent.DISCARDING_BUNDLE_UPDATE_LOCATION,
                    bundleUpdateLocation,
                    this,
                    this.informedDeployable.getName(),
                    this.informedDeployable.getVersion().toString());
            bundleManifest.setBundleUpdateLocation(null);
        }
    }

    /**
     * {@inheritDoc}
     */
    public Bundle getBundle() {
        return this.moduleBundle;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String toString() {
        return super.toString();
    }

    /**
     * {@inheritDoc}
     */
    public void install() throws UnableToSatisfyDependenciesException {
        // Replace the bundle manifest with any updates.
        this.getModuleData().replaceManifest();

        // Complete the installation.
        try {
            /*
             * Defer satisfying the bundle's constraints in case it depends on
             * other bundles in this application which are yet to be installed
             */
            this.moduleBundle = this.bundleInstaller.installBundleFromLocation(this.getModuleData().getFile(), false);
            this.standardModuleRegistry.associateModule(this.moduleBundle, this);
        }
        catch (BundleException be) {
            BundleManifest bundleManifest = this.getModuleData().getBundleManifest();
            BundleSymbolicName bundleSymbolicName = bundleManifest.getBundleSymbolicName();
            throw new UnableToSatisfyBundleDependenciesException(bundleSymbolicName == null ? ""
                    : bundleSymbolicName.getSymbolicName(),
                    Version.parseVersion(bundleManifest.getBundleVersion()),
                    "Unable to install bundle from module '" + this + "':" + be.toString());
        }
        catch (OsgiFrameworkException ofe) {
            Throwable cause = ofe.getCause() == null ? ofe : ofe.getCause();
            throw new FatalDeploymentException("Unable to install bundle for module '" + this + "'.", cause);
        }
    }

    /**
     * {@inheritDoc}
     */
    public void uninstall() {
        Assert.notNull(this.moduleBundle, "A module cannot be uninstalled unless it has been installed");
        try {
            this.standardModuleRegistry.dissociateModule(this.moduleBundle);
            this.moduleBundle.uninstall();
        }
        catch (final BundleException e) {
            throw new FatalDeploymentException("Unable to uninstall bundle for module '" + this + "'.", e);
        }
    }

    /**
     * @throws DeploymentException {@inheritDoc}
     */
    @Override
    public void deploy() throws DeploymentException {
        // notify application bundle deploying
        this.applicationDeploymentEventNotifier.onEvent(new ApplicationBundleDeploying(this.informedDeployable.getName(),
                this.informedDeployable.getVersion(),
                this.moduleBundle));

        pushThreadContext();
        try {
            startBundle();
        }
        finally {
            popThreadContext();
        }
    }

    /**
     * {@inheritDoc}
     */
    public void checkDeployed() throws DeploymentException {
        if (!fragment && this.bundleActivationPolicy == null) {
            try {
                TRACER.traceInfo("Entering wait for start of module '%s' bundle '%s'", this, this.moduleBundle);
                awaitStart(this.deployerConfiguration.getDeploymentTimeoutSeconds());
                // notify application bundle deployed
                this.applicationDeploymentEventNotifier.onEvent(new ApplicationBundleDeployed(this.informedDeployable.getName(),
                        this.informedDeployable.getVersion(),
                        this.moduleBundle));
                TRACER.traceInfo("Exited wait for start of module '%s' bundle '%s'", this, this.moduleBundle);
            }
            catch (InterruptedException e) {
                // notify application bundle not deployed
                this.applicationDeploymentEventNotifier.onEvent(new ApplicationBundleUndeployed(this.informedDeployable.getName(),
                        this.informedDeployable.getVersion(),
                        this.moduleBundle));
                TRACER.traceError("Wait for start of module '%s' bundle '%s' interrupted", this, this.moduleBundle);
                LOGGER.log(DeployerLogEvent.DEPLOY_INTERRUPTED, this);
                throw new DeploymentException("Deployment of '" + this + "' failed due to interruption", e);
            }
            catch (ExecutionException e) {
                // notify application bundle not deployed
                this.applicationDeploymentEventNotifier.onEvent(new ApplicationBundleUndeployed(this.informedDeployable.getName(),
                        this.informedDeployable.getVersion(),
                        this.moduleBundle));
                TRACER.traceError("Exception returned after wait for start of module '%s' bundle '%s'",
                        e,
                        this,
                        this.moduleBundle);
                Throwable cause = e.getCause();
                if (cause != null && cause instanceof DeploymentException) {
                    throw (DeploymentException) cause;
                }
                else {
                    Throwable report = cause != null ? cause : e;

                    // Failures due to a BeansException have already been logged
                    // by DeployerFailureListener
                    if (!(cause instanceof BeansException)) {
                        LOGGER.log(DeployerLogEvent.MODULE_DEPLOY_FAILURE, report, this);
                    }

                    throw new DeploymentException("Module '" + this + "' failed to deploy", report);
                }
            }
            catch (TimeoutException e) {
                // notify application bundle not deployed
                this.applicationDeploymentEventNotifier.onEvent(new ApplicationBundleUndeployed(this.informedDeployable.getName(),
                        this.informedDeployable.getVersion(),
                        this.moduleBundle));
                TRACER.traceError("Wait for start of module '%s' bundle '%s' timed out", e, this, this.moduleBundle);
                LOGGER.log(DeployerLogEvent.DEPLOY_TIMEOUT, this);
                throw new DeploymentException("Deployment of '" + this + "' failed due to time out", e);
            }
            finally {
                this.startFuture = null;
                this.startedFuture = null;
            }
        }
    }

    /**
     * Wait for the futures associated with starting this module's bundle to
     * report success or an exception. Time out the wait after the given period.
     * 
     * @param timeoutSeconds the amount of time to wait
     * @throws InterruptedException
     * @throws ExecutionException
     * @throws TimeoutException
     */
    private void awaitStart(int timeoutSeconds) throws InterruptedException, ExecutionException, TimeoutException {
        long waitStartTimeMillis = System.currentTimeMillis();
        long remainingWaitIntervalMillis = timeoutSeconds * 1000;
        this.startFuture.get(remainingWaitIntervalMillis, TimeUnit.MILLISECONDS);
        long elapsedIntervalMillis = System.currentTimeMillis() - waitStartTimeMillis;
        this.startedFuture.get(Math.max(remainingWaitIntervalMillis - elapsedIntervalMillis, 0), TimeUnit.MILLISECONDS);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void undeploy() throws DeploymentException {
        try {
            if (this.moduleBundle != null) {
                // notify application bundle undeploying
                this.applicationDeploymentEventNotifier.onEvent(new ApplicationBundleUndeploying(this.informedDeployable.getName(),
                        this.informedDeployable.getVersion(),
                        this.moduleBundle));
                pushThreadContext();
                try {
                    stopBundle();
                }
                finally {
                    popThreadContext();
                }
                // notify application bundle undeployed
                this.applicationDeploymentEventNotifier.onEvent(new ApplicationBundleUndeployed(this.informedDeployable.getName(),
                        this.informedDeployable.getVersion(),
                        this.moduleBundle));
                uninstall();
            }
        }
        finally {
            getModuleData().cleanup();
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void start() throws DeploymentException {
        pushThreadContext();
        try {
            startBundle();
            super.start();
        }
        finally {
            popThreadContext();
        }
    }

    /**
     * Start this module's bundle.
     * 
     * @throws DeploymentException
     */
    private void startBundle() throws DeploymentException {
        if (!this.fragment && this.bundleActivationPolicy == null) {
            Assert.notNull(this.moduleBundle, "A module cannot be started unless it is installed");
            Assert.isTrue(this.standardModuleRegistry.getModule(this.moduleBundle) == this, "Broken module registry");
            try {
                this.startedFuture = new BundleStartedFuture();
                this.startFuture = this.bundleStarter.start(this.moduleBundle);
            }
            catch (BundleException e) {
                LOGGER.log(DeployerLogEvent.START_BUNDLE_FAILURE, e, this);
                throw new DeploymentException("Failed to start bundle for module '" + this + "'.", e);
            }
            catch (RuntimeException e) {
                LOGGER.log(DeployerLogEvent.START_BUNDLE_FAILURE, e, this);
                throw new DeploymentException("Failed to start bundle for module '" + this + "'.", e);
            }
        }
    }

    /**
     * Update this module's bundle.
     * 
     * Pre-condition: the bundle is in INSTALLED or RESOLVED state.
     * 
     * @throws DeploymentException
     */
    private void updateBundle() throws DeploymentException {
        if (!this.fragment) {
            Assert.notNull(this.moduleBundle, "A module cannot be updated unless it is installed");
            Assert.isTrue(this.moduleBundle.getState() == Bundle.INSTALLED
                    || this.moduleBundle.getState() == Bundle.RESOLVED,
                    "A module cannot be updated unless its bundle is in INSTALLED or RESOLVED state");
            try {
                this.bundleInstaller.updateBundle(this.moduleBundle);
            }
            catch (UnableToSatisfyDependenciesException e) {
                throw new DeploymentException("Cannot satisfy dependencies of module '" + this + "' during update", e);
            }
            catch (BundleException e) {
                throw new DeploymentException("Failed to update bundle for module '" + this + "'.", e);
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void stop() throws DeploymentException {
        pushThreadContext();
        try {
            super.stop();
            stopBundle();
        }
        finally {
            popThreadContext();
        }
    }

    /**
     * Stop this module's bundle if necessary.
     * 
     * @return <code>true</code> if and only if Bundle.stop was issued
     * @throws DeploymentException
     */
    private boolean stopBundle() throws DeploymentException {
        boolean stopped = false;
        if (!this.fragment) {
            Assert.notNull(this.moduleBundle, "A module cannot be stopped unless it is installed");
            int bundleState = this.moduleBundle.getState();
            if (bundleState == Bundle.STARTING || bundleState == Bundle.ACTIVE) {
                try {
                    this.moduleBundle.stop();
                    stopped = true;
                }
                catch (final BundleException e) {
                    throw new FatalDeploymentException("Failed to stop bundle for module '" + this + "'.", e);
                }
            }
        }
        return stopped;
    }

    /**
     * {@inheritDoc}
     */
    public void setContext(Module context) {
        this.context = context;
    }

    /**
     * {@inheritDoc}
     */
    public Module getContext() {
        return this.context;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void onStarting() throws DeploymentException {
        try {
            this.applicationDeploymentEventNotifier.onEvent(new ApplicationBundleStarting(this.informedDeployable.getName(),
                    this.informedDeployable.getVersion(),
                    this.moduleBundle));

            pushThreadContext();
            try {
                super.onStarting();
            }
            finally {
                popThreadContext();
            }
        }
        catch (Exception e) {
            TRACER.traceError("onStarting of module '%s' caught an exception", e, this);
            if (this.startedFuture != null) {
                this.startedFuture.handOff(e);
            }
            throw new FatalDeploymentException("Exception onStarting module '" + this + "'", e);
        }
    }

    /**
     * {@inheritDoc}
     */
    public void onStarted() throws DeploymentException {
        try {
            pushThreadContext();
            try {
                super.deploy();
            }
            finally {
                popThreadContext();
                TRACER.traceInfo("Popped thread context for module '%s' bundle '%s'", this, this.moduleBundle);
                this.applicationDeploymentEventNotifier.onEvent(new ApplicationBundleStarted(this.informedDeployable.getName(),
                        this.informedDeployable.getVersion(),
                        this.moduleBundle));
            }
        }
        catch (Exception e) {
            TRACER.traceError("onStarted of module '%s' caught an exception", e, this);
            if (!(e instanceof DeploymentException)) {
                LOGGER.log(DeployerLogEvent.MODULE_START_FAILURE, e, this);
            }
            if (this.startedFuture != null) {
                this.startedFuture.handOff(e);
            }
            throw new DeploymentException("Exception onStarted module '" + this + "'", e);
        }
        if (this.startedFuture != null) {
            this.startedFuture.notifyOnStarted();
        }
    }

    /**
     * {@inheritDoc}
     */
    public void onStopping() throws DeploymentException {
        try {
            this.applicationDeploymentEventNotifier.onEvent(new ApplicationBundleStopping(this.informedDeployable.getName(),
                    this.informedDeployable.getVersion(),
                    this.moduleBundle));
            pushThreadContext();
            try {
                super.undeploy();
            }
            finally {
                popThreadContext();
            }
        }
        catch (Exception e) {
            TRACER.traceError("onStopping of module '%s' caught an exception", e, this);
            LOGGER.log(DeployerLogEvent.MODULE_STOPPING_FAILURE, e, this);
            throw new DeploymentException("Exception onStopping module '" + this + "'", e);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void onStopped() throws DeploymentException {
        try {
            pushThreadContext();
            try {
                super.onStopped();
            }
            finally {
                popThreadContext();
                this.applicationDeploymentEventNotifier.onEvent(new ApplicationBundleStopped(this.informedDeployable.getName(),
                        this.informedDeployable.getVersion(),
                        this.moduleBundle));
            }
        }
        catch (Exception e) {
            TRACER.traceError("onStopped of module '%s' caught an exception", e, this);
            LOGGER.log(DeployerLogEvent.MODULE_ON_STOPPED_FAILURE, e, this);
            throw new DeploymentException("Exception onStopped module '" + this + "'", e);
        }
    }

    /**
     * {@inheritDoc}
     */
    public boolean refresh(String symbolicName, ModuleDeployerSet moduleDeployerSet,
            CompletableDeploymentContext deploymentContext) throws DeploymentException {
        String unscopedSymbolicName = this.getUnscopedSymbolicName();
        Assert.isTrue((unscopedSymbolicName == null && symbolicName == null)
                || unscopedSymbolicName.equals(symbolicName),
                "refresh called with symbolic name '%s' on the wrong module '%s'",
                symbolicName,
                this);
        String bundleVersion = this.getModuleData().getBundleManifest().getBundleVersion();
        boolean refreshed = super.refresh(symbolicName, moduleDeployerSet, deploymentContext);
        /*
         * If the bundle symbolic name or bundle version has changed, the
         * refresh is unsuccessful and the module must be redeployed.
         */
        String refreshedSymbolicName = this.getUnscopedSymbolicName();
        String refreshedBundleVersion = this.getModuleData().getBundleManifest().getBundleVersion();
        refreshed = refreshed
                && ((refreshedSymbolicName == null && symbolicName == null) || refreshedSymbolicName.equals(symbolicName))
                && ((refreshedBundleVersion == null && bundleVersion == null) || refreshedBundleVersion.equals(bundleVersion));

        if (refreshed) {
            try {
                /*
                 * To avoid this module's bundle from being stopped and started
                 * by each of update and refresh packages, stop it if necessary
                 * and restart it later if we had to stop it.
                 */
                int bundleState = this.moduleBundle.getState();
                boolean stop = bundleState == Bundle.STARTING || bundleState == Bundle.ACTIVE;
                pushThreadContext();
                try {
                    if (stop) {
                        // notify application bundle undeploying
                        this.applicationDeploymentEventNotifier.onEvent(new ApplicationBundleUndeploying(this.informedDeployable.getName(),
                                this.informedDeployable.getVersion(),
                                this.moduleBundle));
                        stopBundle();
                        // notify application bundle undeployed
                        this.applicationDeploymentEventNotifier.onEvent(new ApplicationBundleUndeployed(this.informedDeployable.getName(),
                                this.informedDeployable.getVersion(),
                                this.moduleBundle));
                    }

                    updateBundle();

                    // Refresh this module's packages and wait for the refresh
                    // to complete.
                    this.refreshComplete = false;
                    this.osgi.getBundleContext().addFrameworkListener(new FrameworkListener() {

                        public void frameworkEvent(FrameworkEvent event) {
                            if (event.getType() == FrameworkEvent.PACKAGES_REFRESHED) {
                                TRACER.traceInfo("Received 'PACKAGES_REFRESHED' framework event");
                                StandardModule.this.refreshComplete = true;
                            }
                        }
                    });
                    refreshBundle();
                    for (int i = 0; i < 50; i++) {
                        if (!this.refreshComplete) {
                            try {
                                Thread.sleep(REFRESH_WAIT_INTERVAL_MS);
                            }
                            catch (InterruptedException e) {
                            }
                        }
                    }
                    if (!this.refreshComplete) {
                        TRACER.traceWarning("Refresh packages of '%s' failed - escalating refresh", this);
                        refreshed = false;
                    }
                    else {
                        /*
                         * Refresh succeeded so, if necessary, restart this
                         * module's bundle and wait for the start to complete.
                         */
                        if (stop) {
                            // notify application bundle deploying
                            this.applicationDeploymentEventNotifier.onEvent(new ApplicationBundleDeploying(this.informedDeployable.getName(),
                                    this.informedDeployable.getVersion(),
                                    this.moduleBundle));
                            startBundle();
                            checkDeployed();
                        }
                    }
                }
                finally {
                    popThreadContext();
                }
            }
            catch (DeploymentException e) {
                TRACER.traceWarning("Update of '%s' failed - escalating refresh", e, this);
                refreshed = false;
            }
            finally {
                clearStackedModulePersonality();
            }
        }
        return refreshed;
    }

    /**
     * Refresh this module's bundle.
     * 
     * Pre-condition: this module's bundle must be in INSTALLED or RESOLVED
     * state.
     * 
     * @throws DeploymentException
     */
    private void refreshBundle() throws DeploymentException {
        Assert.notNull(this.moduleBundle, "A module cannot be refreshed unless it is installed");
        Assert.isTrue(this.moduleBundle.getState() == Bundle.INSTALLED
                || this.moduleBundle.getState() == Bundle.RESOLVED,
                "A module cannot be refreshed unless its bundle is in INSTALLED or RESOLVED state");
        this.packageAdmin.refreshPackages(new Bundle[] { this.moduleBundle });
    }

    private void pushThreadContext() {
        pushThreadContextClassLoader();
        this.applicationTraceNameStack.push(wrapNull(this.tracingService.getCurrentApplicationName()));
        this.tracingService.setCurrentApplicationName(applicationTraceName);
    }

    private String wrapNull(String string) {
        return string != null ? string : WRAPPED_NULL;
    }

    private String unwrapNull(String string) {
        return string != WRAPPED_NULL ? string : null;
    }

    private void popThreadContext() {
        popThreadContextClassLoader();
        this.tracingService.setCurrentApplicationName(unwrapNull((String) this.applicationTraceNameStack.pop()));
    }

    private void pushThreadContextClassLoader() {
        Thread currentThread = Thread.currentThread();
        ClassLoader oldContextClassLoader = currentThread.getContextClassLoader();
        this.contextClassLoaderStack.push(oldContextClassLoader);

        int state = context.getBundle().getState();
        if (state != Bundle.INSTALLED && state != Bundle.UNINSTALLED) {
            if (this.context != null) {
                Bundle contextBundle = this.context.getBundle();
                if (contextBundle != null) {
                    ClassLoader newContextClassLoader = this.osgi.getBundleClassLoader(contextBundle);
                    if (newContextClassLoader != null) {
                        currentThread.setContextClassLoader(newContextClassLoader);
                        TRACER.traceInfo("Thread context class loader '%s' pushed and set to '%s'",
                                oldContextClassLoader,
                                newContextClassLoader);
                    }
                    else {
                        TRACER.traceInfo("Thread context class loader not found for bundle '%s'",
                                contextBundle.getSymbolicName());
                    }
                }
            }
        }
    }

    private void popThreadContextClassLoader() {
        Thread currentThread = Thread.currentThread();
        ClassLoader oldContextClassLoader = currentThread.getContextClassLoader();
        ClassLoader newContextClassLoader = (ClassLoader) this.contextClassLoaderStack.pop();
        currentThread.setContextClassLoader(newContextClassLoader);
        TRACER.traceInfo("Thread context class loader '%s' popped and set to '%s'",
                oldContextClassLoader,
                newContextClassLoader);
    }

    /**
     * {@inheritDoc}
     */
    public int getDeploymentTimeoutSeconds() {
        return this.deployerConfiguration.getDeploymentTimeoutSeconds();
    }

    final class BundleStartedFuture implements Future<Void> {

        private final ReentrantLock lock = new ReentrantLock();

        private final Condition available = this.lock.newCondition();

        private boolean done = false;

        private Throwable throwable = null;

        /**
         * Signals successful processing of the onStarted call back.
         */
        public void notifyOnStarted() {
            this.lock.lock();
            try {
                if (this.throwable == null) {
                    this.done = true;
                    this.available.signalAll();
                }
            }
            finally {
                this.lock.unlock();
            }
        }

        /**
         * Signals failed processing of the onStarted call back.
         * 
         * @param throwable the error that occurred.
         */
        public void handOff(Throwable throwable) {
            this.lock.lock();
            try {
                this.throwable = throwable;
                this.available.signalAll();
            }
            finally {
                this.lock.unlock();
            }
        }

        /**
         * {@inheritDoc}
         */
        public boolean cancel(boolean mayInterruptIfRunning) {
            return false;
        }

        /**
         * {@inheritDoc}
         */
        public Void get() throws InterruptedException, ExecutionException {
            this.lock.lock();
            if (isDone()) {
                return processResult();
            }
            try {
                this.available.await();
                return processResult();
            }
            finally {
                this.lock.unlock();
            }
        }

        /**
         * {@inheritDoc}
         */
        public Void get(long timeout, TimeUnit unit) throws InterruptedException, ExecutionException, TimeoutException {
            this.lock.lock();
            if (isDone()) {
                return processResult();
            }
            try {
                if (this.available.await(timeout, unit)) {
                    return processResult();
                }
                else {
                    throw new TimeoutException("Timeout while waiting.");
                }
            }
            finally {
                this.lock.unlock();
            }
        }

        /**
         * {@inheritDoc}
         */
        public boolean isCancelled() {
            return false;
        }

        /**
         * {@inheritDoc}
         */
        public boolean isDone() {
            this.lock.lock();
            try {
                return this.done || this.throwable != null;
            }
            finally {
                this.lock.unlock();
            }
        }

        private Void processResult() throws ExecutionException {
            if (this.throwable != null) {
                throw new ExecutionException(this.throwable);
            }
            else {
                return null;
            }
        }
    }

}
