/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.ffdc.test.support;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

import java.io.File;
import java.io.FileReader;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.StringWriter;

import org.junit.After;
import org.junit.Before;
import org.junit.runner.RunWith;
import org.osgi.framework.BundleContext;
import org.osgi.framework.ServiceReference;
import org.springframework.util.FileCopyUtils;

import com.springsource.server.osgi.framework.OsgiFramework;
import com.springsource.server.serviceability.ffdc.FFDCService;
import com.springsource.server.test.framework.junit4.ServerOsgiIntegrationTestJUnit4ClassRunner;

/**
 * Abstract base class for integration tests involving
 * {@link com.springsource.server.serviceability.ffdc.DumpExtension DumpExtensions}.
 * 
 * @author Sam Brannen
 * @since 1.0
 */
@RunWith(ServerOsgiIntegrationTestJUnit4ClassRunner.class)
public abstract class AbstractDumpExtensionIntegrationTests {

    @com.springsource.server.test.framework.OsgiFramework
    protected OsgiFramework osgiFramework;

    protected BundleContext bundleContext;

    protected ServiceReference ffdcServiceReference;

    protected FFDCService ffdcService;

    protected String dumpContents;

    protected File dumpDir = new File("target");

    protected final FilenameFilter dumpFilenameFilter = new FilenameFilter() {

        public boolean accept(File dir, String name) {
            return name.endsWith(".dump");
        }
    };

    protected void assertDumpContains(String expectedText) {
        assertTrue("The generated dump file should contain [" + expectedText + "].", this.dumpContents.contains(expectedText));
    }

    protected void deleteDumpFiles(File dumpDir) {
        for (String filename : dumpDir.list(this.dumpFilenameFilter)) {
            new File(dumpDir, filename).delete();
        }
    }

    /**
     * Prepares the file system by removing any existing dump files; retrieves the {@link FFDCService} from the OSGi
     * service registry; {@link FFDCService#generateDump(Throwable) generates a dump} with the specified
     * {@see #getCauseOfDump() cause}; and the retrieves the contents of the generated dump file, which can be
     * {@link #assertDumpContains(String) asserted} in subclasses.
     * 
     * @see #getCauseOfDump()
     */
    @Before
    public void setUpAndGenerateDump() throws IOException {
        this.bundleContext = this.osgiFramework.getBundleContext();
        this.ffdcServiceReference = this.bundleContext.getServiceReference(FFDCService.class.getName());
        assertNotNull("The FFDCService should have been registered as an OSGi service via Spring-DM.", this.ffdcServiceReference);
        this.ffdcService = (FFDCService) this.bundleContext.getService(this.ffdcServiceReference);
        assertNotNull("The FFDCService should have been retrievable.", this.ffdcService);

        deleteDumpFiles(this.dumpDir);

        String[] filenames = this.dumpDir.list(this.dumpFilenameFilter);
        assertEquals("There should initially not be any dump files.", 0, filenames.length);
        this.ffdcService.generateDump(getCauseOfDump());
        filenames = this.dumpDir.list(this.dumpFilenameFilter);
        assertEquals("Only one dump file should have been generated.", 1, filenames.length);

        StringWriter stringWriter = new StringWriter();
        FileCopyUtils.copy(new FileReader(new File(this.dumpDir, filenames[0])), stringWriter);
        stringWriter.flush();
        this.dumpContents = stringWriter.toString();
    }

    /**
     * Subclasses can override this method to provide a non-null {@link Throwable cause} for the generated dump.
     * <p>
     * This default implementation simply returns <code>null</code>.
     * </p>
     * 
     * @return the cause of the dump
     */
    protected Throwable getCauseOfDump() {
        return null;
    }

    @After
    public void cleanUpAfterDump() throws Exception {
        if (this.ffdcServiceReference != null) {
            this.bundleContext.ungetService(this.ffdcServiceReference);
        }
    }

}
