/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.io.core;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.net.URI;

import org.springframework.util.FileCopyUtils;

import com.springsource.server.serviceability.ffdc.NonNull;
import com.springsource.server.serviceability.output.Level;
import com.springsource.server.serviceability.tracing.EntryExitTraceLevel;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * Represents a reference, by path, to a location on the file system. This location may be a file or a directory, or it
 * may point to a non-existent location.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @author Andy Wilkinson
 * @author Glyn Normington
 * @since 1.0
 */
public final class PathReference {

    private static final Tracer TRACER = TracerFactory.createTracer(PathReference.class);

    /**
     * The {@link File} handle that backs this reference.
     */
    private final File file;

    /**
     * Creates a new <code>PathReference</code> for the supplied path.
     * 
     * @param path the path to reference.
     */
    public PathReference(@NonNull String path) {
        this(new File(path));
    }

    /**
     * Creates a new <code>PathReference</code> for the supplied {@link URI} which must point to a file location.
     * 
     * @param uri the file URI to reference
     */
    public PathReference(@NonNull URI uri) {
        this.file = new File(uri);
    }

    /**
     * Creates a new <code>PathReference</code> for the supplied {@link File}.
     * 
     * @param file the file to reference.
     */
    public PathReference(@NonNull File file) {
        this.file = file;
    }

    /**
     * Converts this <code>PathReference</code> to a {@link File}.
     * 
     * @return the <code>File</code>.
     */
    @EntryExitTraceLevel(Level.DEBUG)
    public File toFile() {
        return this.file;
    }

    /**
     * Queries whether or not this <code>PathReference</code> points to a vaild file or directory.
     * 
     * @return <code>true</code> if the location is valid, otherwise <code>false</code>.
     */
    @EntryExitTraceLevel(Level.DEBUG)
    public boolean exists() {
        return this.file.exists();
    }

    /**
     * Queries whether or not this <code>PathReference</code> points to a file.
     * 
     * @return <code>true</code> if this location points to a file, otherwise <code>false</code>.
     */
    @EntryExitTraceLevel(Level.DEBUG)
    public boolean isFile() {
        return this.file.isFile();
    }

    /**
     * Queries whether or not this <code>PathReference</code> points to a directory.
     * 
     * @return <code>true</code> if this location points to a directory, otherwise <code>false</code>.
     */
    @EntryExitTraceLevel(Level.DEBUG)
    public boolean isDirectory() {
        return this.file.isDirectory();
    }

    /**
     * Gets the name of the file or directory referenced by this <code>PathReference</code>.
     * 
     * @return the file or directory name.
     * @see File#getName()
     */
    @EntryExitTraceLevel(Level.DEBUG)
    public String getName() {
        return this.file.getName();
    }

    /**
     * Gets the parent directory of this <code>PathReference</code>.
     * 
     * @return the parent directory.
     * @see File#getParentFile()
     */
    @EntryExitTraceLevel(Level.DEBUG)
    public PathReference getParent() {
        File parentFile = this.file.getParentFile();
        return parentFile == null ? null : new PathReference(parentFile);
    }

    /**
     * Queries whether or not the path of this <code>PathReference</code> is absolute.
     * 
     * @return <code>true</code> if the path is absolute, otherwise <code>false</code>.
     */
    @EntryExitTraceLevel(Level.DEBUG)
    public boolean isAbsolute() {
        return this.file.isAbsolute();
    }

    /**
     * Gets the absolute path of this <code>PathReference</code>.
     * 
     * @return the absolute path.
     */
    @EntryExitTraceLevel(Level.DEBUG)
    public String getAbsolutePath() {
        return this.file.getAbsolutePath();
    }

    /**
     * Converts this <code>PathReference</code> into a {@link URI}.
     * 
     * @return the URI of this <code>PathReference</code>.
     * @see File#toURI()
     */
    @EntryExitTraceLevel(Level.DEBUG)
    public URI toURI() {
        return this.file.toURI();
    }

    /**
     * Gets the canonical path of this <code>PathReference</code>.
     * 
     * @return the canonical path.
     */
    @EntryExitTraceLevel(Level.DEBUG)
    public String getCanonicalPath() {
        try {
            return this.file.getCanonicalPath();
        } catch (IOException e) {
            throw new FatalIOException("Unable to canonicalize path for file '" + this + "'.", e);
        }
    }

    /**
     * Creates a <code>PathReference</code> that points to the absolute location of this <code>PathReference</code>.
     * 
     * @return the new <code>PathReference</code>.
     */
    @EntryExitTraceLevel(Level.DEBUG)
    public PathReference toAbsoluteReference() {
        return new PathReference(getAbsolutePath());
    }

    /**
     * Creates a <code>PathReference</code> that points to the canonical location of this <code>PathReference</code>.
     * 
     * @return the new <code>PathReference</code>.
     */
    @EntryExitTraceLevel(Level.DEBUG)
    public PathReference toCanonicalReference() {
        return new PathReference(getCanonicalPath());
    }

    /**
     * {@inheritDoc}
     */
    @EntryExitTraceLevel(Level.DEBUG)
    @Override public String toString() {
        return this.file.toString();
    }

    /**
     * Delete the file/directory referenced by this <code>PathReference</code>. Has the same effect as calling
     * {@link #delete(boolean) delete(false)}.
     * 
     * @return <code>true</code> if the file is deleted, otherwise <code>false</code>.
     */
    @EntryExitTraceLevel(Level.DEBUG)
    public boolean delete() {
        return delete(false);
    }

    /**
     * Delete the file/directory referenced by this <code>PathReference</code>. Nested contents can optionally be
     * deleted by passing <code>recursive = true</code>. Non-recursive deletes will fail (return <code>false</code>
     * for non-empty directories).
     * 
     * @param recursive should the delete process nested contents recursively.
     * @return <code>true</code> if the file is deleted, otherwise <code>false</code>.
     */
    @EntryExitTraceLevel(Level.DEBUG)
    public boolean delete(boolean recursive) {
        boolean deleted = recursive ? FileSystemUtils.deleteRecursively(this.file) : this.file.delete();
        if (deleted) {
            TRACER.traceDebug("Recursive '%s', deleted '%s'", recursive, deleted);
        }
        else {
            TRACER.traceInfo("Delete of '%s' failed. Delete request was recursive: '%s'", this, recursive);
        }
        return deleted;
    }

    /**
     * Same as calling {@link #copy(PathReference, boolean) copy(to, false)}.
     */
    @EntryExitTraceLevel(Level.DEBUG)
    public PathReference copy(@NonNull PathReference to) {
        return copy(to, false);
    }

    /**
     * Same as calling {@link #copy(PathReference, boolean, PathFilter) copy(to, recursive, null)};
     */
    @EntryExitTraceLevel(Level.DEBUG)
    public PathReference copy(@NonNull PathReference to, boolean recursive) {
        return copy(to, recursive, null);
    }

    /**
     * Copies the contents of the file/directory to location referenced supplied <code>PathReference</code>. The
     * exact semantics of the copy depend on the state of both the source and destination locations.<p/>
     * 
     * When copying a file, if the destination location already exists and is a file then a {@link FatalIOException} is
     * thrown. If the destination location exists and is a directory, the file is copied into this directory with the
     * same {@link #getName name} as the source location. If the destination location does not exist, then the file is
     * copied to this location directly.<p/>
     * 
     * When copying a directory, if the destination location exists then the directory is copied into the already
     * existing directory with the same name as the source location. If the destination location does not exist, then
     * the directory is copied directly to the destination location and thus gets the name defined by the destination
     * location.<p/>
     * 
     * A {@link PathFilter} can be used to specify exactly which files will be included in the copy. Passing
     * <code>null</code> for the filter includes all files.<p/>
     * 
     * Non-recursive copies on non-empty directories will fail with {@link FatalIOException}.
     * 
     * @param to the destination to copy to.
     * @param recursive whether the copy should be recursive or not.
     * @param filter a <code>PathFilter</code> controlling which files are included in the copy.
     * @return the final destination <code>PathReference</code>.
     */
    @EntryExitTraceLevel(Level.DEBUG)
    public PathReference copy(@NonNull PathReference to, boolean recursive, PathFilter filter) {
        PathReference dest = to;
        if (!exists()) {
            throw new FatalIOException("Cannot copy path '" + this + "' to '" + dest + "'. Source path does not exist.");
        }
        if (isFile()) {
            if (dest.exists()) {
                if (dest.isFile()) {
                    throw new FatalIOException("Cannot copy path '" + this + "' to '" + dest + "'. Destination path already exists.");
                } else {
                    dest = dest.newChild(this.file.getName());
                    if (dest.exists()) {
                        throw new FatalIOException("Cannot copy path '" + this + "' to '" + dest + "'. Destination path already exists.");
                    } else {
                        dest.createFile();
                    }
                }
            }
            copyFile(this.file, dest.file);
        } else {
            int length = this.file.list().length;
            if (length > 0 && !recursive) {
                throw new FatalIOException("Cannot perform non-recursive copy on non-empty directory '" + this + "'");
            }
            if (dest.exists()) {
                dest = dest.newChild(this.file.getName()).createDirectory();
            } else {
                // copy to this dir
                dest.createDirectory();
            }
            recursiveCopy(this.file, dest.file, filter);
        }
        return dest;
    }

    /**
     * Creates a new <code>PathReference</code> by concatenating the path of this <code>PathReference</code> with
     * the supplied <code>name</code>. Note that this method does <strong>not</strong> create a physical file or
     * directory at the child location.
     * 
     * @param name the name of the new child entry.
     * @return the child <code>PathReference</code>.
     */
    @EntryExitTraceLevel(Level.DEBUG)
    public PathReference newChild(@NonNull String name) {
        return PathReference.concat(this.file.getPath(), name);
    }

    /**
     * Creates a new file at the location of this <code>PathReference</code>.
     * 
     * @return this <code>PathReference</code> for chaining purposes.
     */
    @EntryExitTraceLevel(Level.DEBUG)
    public PathReference createFile() {
        if (this.exists()) {
            return this;
        }
        File parent = this.file.getParentFile();
        if (!parent.exists()) {
            parent.mkdirs();
        }
        try {
            this.file.createNewFile();
            return this;
        } catch (IOException e) {
            throw new FatalIOException("Unable to create file '" + this + "'.", e);
        }
    }

    /**
     * Creates a directory at the location of this <code>PathReference</code>.
     * 
     * @return this <code>PathReference</code> for chaining purposes.
     */
    @EntryExitTraceLevel(Level.DEBUG)
    public PathReference createDirectory() {
        this.file.mkdirs();
        return this;
    }

    /**
     * Creates a new <code>PathReference</code> by concatenating the supplied <code>parts</code>, treating each
     * part as a separate path element.
     * 
     * @param parts the path elements.
     * @return the <code>PathReference</code> for the concatenated path.
     */
    @EntryExitTraceLevel(Level.DEBUG)
    public static PathReference concat(String... parts) {
        StringBuilder sb = new StringBuilder(256);
        for (int x = 0; x < parts.length; x++) {
        	String part = parts[x];
            sb.append(part);
            if (x < parts.length && !part.endsWith(File.separator)) {
                sb.append(File.separatorChar);
            }
        }
        return new PathReference(sb.toString());
    }

    /**
     * Recursively copy the contents of <code>src</code> to <code>dest</code>.
     * 
     * @param src the source file.
     * @param dest the destination file.
     */
    @EntryExitTraceLevel(Level.DEBUG)
    private static void recursiveCopy(File src, File dest, PathFilter filter) {
        File[] entries = src.listFiles();
        for (File file : entries) {
            PathReference newFile = PathReference.concat(dest.getAbsolutePath(), file.getName());
            if (filter != null && !filter.matches(new PathReference(file))) {
                continue;
            }
            if (file.isFile()) {
                newFile.createFile();
                copyFile(file, newFile.toFile());
            } else {
                newFile.createDirectory();
                recursiveCopy(file, newFile.toFile(), filter);
            }
        }
    }

    /**
     * Copies <code>src</code> to <code>dest</code>.
     * 
     * @param src the source <code>File</code>.
     * @param dest the destination <code>File</code>.
     */
    @EntryExitTraceLevel(Level.DEBUG)
    private static void copyFile(File src, File dest) {
        try {
            FileCopyUtils.copy(src, dest);
        } catch (IOException e) {
            throw new FatalIOException("Cannot copy " + (src.isFile() ? "file" : "directory") + " '" + src + "' to '" + dest + "'.", e);
        }
    }

    /**
     * {@inheritDoc}
     */
    @EntryExitTraceLevel(Level.DEBUG)
    @Override public int hashCode() {
        return this.file.hashCode();
    }

    /**
     * {@inheritDoc}
     */
    @EntryExitTraceLevel(Level.DEBUG)
    @Override public boolean equals(Object that) {
        if (this == that) {
            return true;
        }
        if (!(that instanceof PathReference)) {
            return false;
        }
        PathReference thatRef = (PathReference) that;
        return this.file.equals(thatRef.file);
    }

    /**
     * Gets the contents of the {@link File} pointed to by this <code>PathReference</code>.
     * 
     * @return the file contents as a <code>String</code>.
     * @throws IOException if the file cannot be found or read.
     */
    @EntryExitTraceLevel(Level.DEBUG)
    public String fileContents() throws IOException {
        if (!this.file.isFile()) {
            throw new FileNotFoundException("PathReference refers to a directory which has no file contents");
        }
        StringBuffer contents = new StringBuffer(100);
        BufferedReader fileReader = new BufferedReader(new FileReader(this.file));
        char[] chars = new char[100];
        int charsRead = 0;
        while ((charsRead = fileReader.read(chars)) != -1) {
            contents.append(String.valueOf(chars, 0, charsRead));
        }
        fileReader.close();
        return contents.toString();
    }
    
    /**
     * Set the last modified time of the {@link File} pointed to by this <code>PathReference</code> to the current time.
     * 
     * Note: the behaviour is not the same as UNIX(TM) touch as the latter also sets the file access time.
     */
    public boolean touch() {
        return this.file.setLastModified(System.currentTimeMillis());
    }

    /**
     * Allows filtering during {@link PathReference} operations. <p/>
     * 
     * <strong>Concurrent Semantics</strong><br />
     * 
     * Implementations needn't be threadsafe if they are not shared by multiple threads.
     * 
     * @author Rob Harrop
     * @since 1.0
     */
    public static interface PathFilter {

        /**
         * Match against the supplied {@link PathReference} to query whether it is included in the outer operation.
         * 
         * @param path the path to match against.
         * @return <code>true</code> if the <code>PathReference</code> should be included, otherwise
         *         <code>false</code>.
         */
        @EntryExitTraceLevel(Level.DEBUG)
        boolean matches(PathReference path);
    }
}
