/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.io.core.fs;

import java.io.File;
import java.util.List;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.regex.Pattern;

import com.springsource.server.concurrent.core.ThreadManager;
import com.springsource.server.serviceability.ffdc.Assert;
import com.springsource.server.serviceability.ffdc.NonNull;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * Monitors a directory on the file system for modifications.<p/>
 * 
 * Modification notifications are published to a set of configured {@link FileSystemListener FileSystemListeners}. New
 * listeners can be safely added at runtime.<p/>
 * 
 * By default, all files in the directory are monitored. Files can be excluded from monitoring using a regex pattern.
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
public final class FileSystemWatcher {

    private static final Tracer TRACER = TracerFactory.createTracer(FileSystemWatcher.class);

    private final File watchDir;

    private final List<FileSystemListener> listeners = new CopyOnWriteArrayList<FileSystemListener>();

    private final ThreadManager threadManager;

    private final Object lifecycleLock = new Object();

    private final Pattern excludePattern;

    private Thread watchThread;

    /**
     * Creates a new, inactive <code>FileSystemWatcher</code>. Notifies for changes on all files in
     * <code>watchDir</code>.
     * 
     * @param watchDir the directory to watch.
     * @param threadManager the {@link ThreadManager} to create threads with.
     */
    public FileSystemWatcher(@NonNull File watchDir, @NonNull ThreadManager threadManager) {
        this(watchDir, threadManager, null);
    }

    /**
     * Creates a new, inactive <code>FileSystemWatcher</code>. Notifies for changes on all files that match
     * <code>excludePattern</code>.
     * 
     * @param watchDir the directory to watch.
     * @param threadManager the {@link ThreadManager} to create threads with.
     * @param excludePattern regex pattern for files to exclude.
     */
    public FileSystemWatcher(@NonNull File watchDir, @NonNull ThreadManager threadManager, @NonNull String excludePattern) {
        Assert.isTrue(watchDir.isDirectory(), "Watch directory '%s' must exist and must be a directory.", watchDir.getAbsolutePath());
        this.watchDir = watchDir;
        this.threadManager = threadManager;
        this.excludePattern = (excludePattern == null ? null : Pattern.compile(excludePattern));
    }

    /**
     * Add a new {@link  FileSystemListener} to this <code>FileSystemWatcher</code>.
     * 
     * @param listener the listener to add.
     */
    public void addListener(@NonNull FileSystemListener listener) {
        this.listeners.add(listener);
    }

    /**
     * Start the <code>FileSystemWatcher</code>.
     */
    public void start() {
        synchronized (this.lifecycleLock) {
            if (this.watchThread == null) {
                this.watchThread = this.threadManager.newThread(new WatchTask(this.watchDir, this.listeners, this.excludePattern), "fs-watcher");
                this.watchThread.start();
                TRACER.traceInfo("Started file system watcher on '%s'.", this.watchDir);
            }
        }
    }

    /**
     * Stop the <code>FileSystemWatcher</code>,
     */
    public void stop() {
        synchronized (this.lifecycleLock) {
            if (this.watchThread != null) {
                TRACER.traceInfo("Stopping file system watcher");
                this.watchThread.interrupt();
            }
        }
    }
}
