/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.io.core.fs;

import java.io.File;
import java.io.FilenameFilter;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Pattern;

import com.springsource.server.common.math.Sets;
import com.springsource.server.serviceability.output.Level;
import com.springsource.server.serviceability.tracing.EntryExitTraceLevel;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * Task that monitors a given directory and notifies configured {@link FileSystemListener FileSystemListeners}.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @author Glyn Normington
 * @since 1.0
 */
final class WatchTask implements Runnable {

    private final static Tracer TRACER = TracerFactory.createTracer(WatchTask.class);

    private final Map<String, Long> fileState = new HashMap<String, Long>(32);
    
    private final Map<File, MonitorRecord> monitorRecords = new HashMap<File, MonitorRecord>(16);

    private final int scanIntervalMillis = 1000;

    private final File watchDir;

    private final List<FileSystemListener> listeners;

    private final FilenameFilter includeFilter;

    /**
     * Creates a new <code>WatchTask</code>.
     * 
     * @param watchDir the directory to watch.
     * @param listeners the listeners to notify.
     * @param excludePattern regex used to match against file for exclusion.
     */
    public WatchTask(File watchDir, List<FileSystemListener> listeners, final Pattern excludePattern) {
        this.watchDir = watchDir;
        this.listeners = listeners;
        this.includeFilter = new FilenameFilter() {

            @EntryExitTraceLevel(Level.DEBUG)
            public boolean accept(File dir, String name) {
                return excludePattern == null || !excludePattern.matcher(name).matches();
            }

        };
        populateInitialState();
    }

    /**
     * Watches the configured directory for modifications.
     */
    public void run() {
        notifyInitialState();
        while (!Thread.currentThread().isInterrupted()) {
            try {
                Thread.sleep(this.scanIntervalMillis);
            } catch (InterruptedException e) {
                break;
            }

            File[] files = listFiles();
            try {
                Set<String> currentFileKeys = new HashSet<String>(files.length);
                for (File file : files) {
                    if(TRACER.isVerboseEnabled()) {
                        TRACER.traceVerbose("Scanning file '%s'", file);
                    }
                    MonitorRecord monitorRecord = monitorRecords.get(file);

                    if (monitorRecord != null) {
                        processMonitoredRecord(file, monitorRecord);
                    } else {
                        processUnmonitoredFile(file);
                    }
                    recordState(file);
                    
                    // record seen files to allow comparison for delete
                    currentFileKeys.add(key(file));
                }

                Set<String> deletedFiles = Sets.difference(this.fileState.keySet(), currentFileKeys);
                for (String deletedFile : deletedFiles) {
                    if(TRACER.isVerboseEnabled()) {
                        TRACER.traceVerbose("Detected deleted file '%s'", deletedFile);
                    }
                    notifyListeners(deletedFile, FileSystemEvent.DELETED);
                    this.fileState.remove(deletedFile);
                }
            } catch (Exception e) {
                TRACER.traceError("Error watching directory '%s'", e, this.watchDir.getAbsolutePath());
            }
        }
    }

    /**
     * Notify listeners of the initial state of the watched directory.
     */
    private void notifyInitialState() {
        for (String path : this.fileState.keySet()) {
            notifyListeners(path, FileSystemEvent.INITIAL);
        }
    }

    /**
     * Processes a {@link File} that we are not yet monitoring.
     */
    private void processUnmonitoredFile(File file) {
        // possibly an update or a new file
        Long lastModified = lastModified(file);

        if (lastModified != null) {
            if (file.lastModified() > lastModified) {
                if(TRACER.isVerboseEnabled()) {
                    TRACER.traceVerbose("Detected modified file '%s'", file);
                }
                monitorRecords.put(file, new MonitorRecord(file.length(), FileSystemEvent.MODIFIED));
            }
        } else {
            if(TRACER.isVerboseEnabled()) {
                TRACER.traceVerbose("Detected new file '%s'", file);
            }
            // no recorded last modified means this is a new file
            monitorRecords.put(file, new MonitorRecord(file.length(), FileSystemEvent.CREATED));
        }
    }

    /**
     * Processes a {@link File} that we are already monitoring.
     */
    private void processMonitoredRecord(File file, MonitorRecord monitorRecord) {
        if(TRACER.isVerboseEnabled()) {
            TRACER.traceVerbose("Found monitor record for file '%s'", file);
        }
        // we are in the process of tracking this file for notification
        long size = file.length();
        if (size > monitorRecord.getSize()) {
            if(TRACER.isVerboseEnabled()) {
                TRACER.traceVerbose("File size changed - probably still updating: '%s'", file);
            }
            // still updating
            monitorRecord.setSize(size);
        } else {
            if(TRACER.isVerboseEnabled()) {
                TRACER.traceVerbose("No change to file size: '%s'", file);
            }
            notifyListeners(file, monitorRecord.getEvent());
            monitorRecords.remove(file);
        }
    }

    private void notifyListeners(File file, FileSystemEvent event) {
        notifyListeners(key(file), event);
    }

    private void notifyListeners(String file, FileSystemEvent event) {
        for (FileSystemListener listener : this.listeners) {
            TRACER.traceVerbose("Notifying  listener '%s' of event '%s' on path '%s'.", listener, event, file);
            listener.onChange(file, event);
        }
    }

    /**
     * Takes an initial snapshot of the watch directory state.
     */
    private void populateInitialState() {
        File[] files = listFiles();
        TRACER.traceDebug("Populating initial state of directory '%s'.", this.watchDir.getAbsolutePath());
        for (File file : files) {
            recordState(file);
        }
    }

    /**
     * Lists the {@link File Files} currently in the watch directory.
     * 
     * @return the <code>Files</code> that are in the watch directory.
     */
    private File[] listFiles() {
        File[] files = this.watchDir.listFiles(this.includeFilter);
        return (files == null ? new File[0] : files);
    }

    /**
     * Records the state of the supplied {@link  File}.
     * 
     * @param the <code>File</code> to record state for.
     */
    private void recordState(File file) {
        String key = key(file);
        long lastModified = file.lastModified();
        TRACER.traceDebug("Recording file state [key=%s,lastModified=%d]", key, lastModified);
        this.fileState.put(key, lastModified);
    }

    /**
     * Gets the recorded last modified timestamp for the supplied {@link File}.
     * 
     * @param file the <code>File</code> to check against.
     * @return the last modified timestamp, or <code>null</code> if no timestamp is recorded.
     */
    private Long lastModified(File file) {
        return this.fileState.get(key(file));
    }

    /**
     * Gets the record key for the supplied {@link  File}.
     * 
     * @param file the <code>File</code> to get the key for.
     * @return the record key.
     */
    private String key(File file) {
        String key =  file.getAbsolutePath();
        if(file.isDirectory()) {
            key += File.separator;
        }
        return key;
    }

    private static class MonitorRecord {

        private final FileSystemEvent event;

        private long size;

        public MonitorRecord(long size, FileSystemEvent event) {
            this.size = size;
            this.event = event;
        }

        public long getSize() {
            return size;
        }

        public void setSize(long size) {
            this.size = size;
        }

        public FileSystemEvent getEvent() {
            return event;
        }

    }
}
