/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.kernel.bootstrap;

import java.io.File;

import org.osgi.framework.BundleContext;

import com.springsource.server.bootstrap.BootstrapConfiguration;
import com.springsource.server.osgi.framework.OsgiFramework;
import com.springsource.server.osgi.framework.OsgiLauncher;
import com.springsource.server.osgi.framework.equinox.EquinoxOsgiConfiguration;
import com.springsource.server.osgi.framework.equinox.EquinoxOsgiLauncher;
import com.springsource.server.serviceability.logging.LogServiceFactory;
import com.springsource.server.serviceability.logging.internal.LoggingBootstrap;
import com.springsource.server.serviceability.output.DestinationRegistry;
import com.springsource.server.serviceability.output.EntryPublisherFactory;
import com.springsource.server.serviceability.output.internal.ServiceabilityOutputBootstrap;
import com.springsource.server.serviceability.tracing.internal.TracingBootstrap;

/**
 * Engine bootstrap launcher.
 * 
 * @author Rob Harrop
 * @author Andy Wilkinson
 */
public class Bootstrap {

    private static final String EQUINOX_INSTALL = "equinox-install";
	private static final String EQUINOX_CONFIG = "equinox-config";

	/*
     * Example VM Argument for running from Eclipse -Dcom.springsource.server.home=.
     */
    public static void main(String[] args) {
        boolean loggingStarted = false;       
        try {
            
            StandardBootstrapConfiguration bootstrapConfiguration = new StandardBootstrapConfiguration();
            performCleanIfRequired(bootstrapConfiguration);
    
            startServiceability(bootstrapConfiguration);
            loggingStarted = true;
            
            LogServiceFactory.getLogService().log(KernelBootstrapLogEvent.SERVER_STARTING);
    
            EquinoxOsgiConfiguration configuration = createEquinoxConfiguration(bootstrapConfiguration);
    
            OsgiLauncher launcher = new EquinoxOsgiLauncher(configuration);
            OsgiFramework osgi = launcher.start();
    
            BundleContext bundleContext = osgi.getBundleContext();
            publishServiceabilityServices(bundleContext);
    
            KernelBootstrap kernelBootstrap = new KernelBootstrap(osgi, bootstrapConfiguration);
            kernelBootstrap.start();
        } catch (Exception e) {
            if (loggingStarted) {
                LogServiceFactory.getLogService().log(KernelBootstrapLogEvent.SERVER_START_FAILED, e);
            } else {
                e.printStackTrace(TracingBootstrap.getOriginalSystemErr());
            }
        }
    }

    private static void publishServiceabilityServices(BundleContext bundleContext) {
        TracingBootstrap.publishServices(bundleContext);
        ServiceabilityOutputBootstrap.publishServices(bundleContext);
    }

    private static EquinoxOsgiConfiguration createEquinoxConfiguration(StandardBootstrapConfiguration bootstrapConfiguration) {
        String[] searchPath = bootstrapConfiguration.getSearchPaths();
        String installPath = bootstrapConfiguration.getInstallPath();
   
        String workDirectory = bootstrapConfiguration.getWorkDirectory();
   
        EquinoxOsgiConfiguration configuration = new EquinoxOsgiConfiguration(workDirectory, installPath, bootstrapConfiguration.getHomeDirectory(), searchPath);
        configuration.setConfigAreaPath(bootstrapConfiguration.getWorkDirectory() + File.separatorChar + EQUINOX_CONFIG);
        configuration.setInstallAreaPath(bootstrapConfiguration.getWorkDirectory() + File.separatorChar + EQUINOX_INSTALL);
        configuration.setJavaProfilePath(bootstrapConfiguration.getOSGiProfileLocation());
   
        if (bootstrapConfiguration.isOsgiConsoleEnabled()) {
            configuration.setRunConsole(true);
            configuration.setConsolePort(bootstrapConfiguration.getOSGiConsolePort());
        }
        
        configuration.setClean(true);
        return configuration;
    }

    private static void startServiceability(StandardBootstrapConfiguration bootstrapConfiguration) {
        ServiceabilityOutputBootstrap.start();
        
        DestinationRegistry destinationRegistry = ServiceabilityOutputBootstrap.getDestinationRegistry();
        EntryPublisherFactory entryPublisherFactory = ServiceabilityOutputBootstrap.getEntryPublisherFactory();
        
        TracingBootstrap.start(destinationRegistry, ServiceabilityOutputBootstrap.getEntryFilterFactory(),
            entryPublisherFactory, bootstrapConfiguration.getTraceLevels(), bootstrapConfiguration.getTraceDirectory());
        LoggingBootstrap.start(bootstrapConfiguration.getLogsDirectory(), destinationRegistry, entryPublisherFactory);
    }

    private static void performCleanIfRequired(BootstrapConfiguration bootstrapConfiguration) {
        if (Boolean.getBoolean("com.springsource.server.clean")) {
            deleteRecursively(new File(bootstrapConfiguration.getWorkDirectory()));
            deleteRecursively(new File(bootstrapConfiguration.getTraceDirectory()));
            deleteRecursively(new File(bootstrapConfiguration.getLogsDirectory()));
            deleteRecursively(new File(bootstrapConfiguration.getDumpDirectory()));
        }
        ensureTempDir();
    }

    private static void deleteRecursively(File fileToDelete) {
        if (fileToDelete.exists()) {
            if (fileToDelete.isDirectory()) {
                File[] files = fileToDelete.listFiles();
                if (files != null) {
                    for (File file : files) {
                        deleteRecursively(file);
                    }
                }
            }
            if (!fileToDelete.delete()) {
                System.out.println("Failed to delete " + fileToDelete);
            }
        }
    }

    private static void ensureTempDir() {
        String tempDirPath = System.getProperty("java.io.tmpdir");
        if (tempDirPath != null && tempDirPath.length() > 0) {
            new File(tempDirPath).mkdirs();
        }
    }
}
