/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.kernel.bootstrap;

import org.osgi.framework.Bundle;
import org.osgi.framework.BundleException;
import org.osgi.framework.ServiceRegistration;

import com.springsource.server.bootstrap.BootstrapConfiguration;
import com.springsource.server.osgi.framework.BundleInstaller;
import com.springsource.server.osgi.framework.OsgiFramework;
import com.springsource.server.osgi.framework.UnableToSatisfyDependenciesException;

/**
 * Basic bootstrapper for the kernel.
 * 
 * @author Rob Harrop
 * @author Glyn Normington
 * @since 1.0
 */
public final class KernelBootstrap {

    public static final String SERVICEABILITY_SUBSYSTEM_NAME = "com.springsource.server.serviceability";

    public static final String COMMON_SUBSYSTEM_NAME = "com.springsource.server.common";

    private final OsgiFramework osgi;

    private final SubsystemContents subsystemContents = new StandardSubsystemContents();

    private ServiceRegistration subsystemContentsReg;

    /**
     * @param osgi
     */
    public KernelBootstrap(OsgiFramework osgi, BootstrapConfiguration bootstrapConfiguration) {
        this.osgi = osgi;
        this.osgi.getBundleContext().registerService(BootstrapConfiguration.class.getName(), bootstrapConfiguration, null);
    }

    /**
     * Install the bundles on which the kernel bundle depends, then install and start the kernel.
     * 
     * @throws BundleException
     * @throws UnableToSatisfyDependenciesException 
     */
    public void start() throws BundleException, UnableToSatisfyDependenciesException {

        installBundle(SERVICEABILITY_SUBSYSTEM_NAME, "com.springsource.server.serviceability.tracing.aspects").start();
        
        BundleInstaller installer = this.osgi.getBundleInstaller();
        
        Bundle equinoxConfigAdmin = installer.installBundle("org.eclipse.equinox.cm");
        equinoxConfigAdmin.start();
        
        Bundle dmExtender = installer.installBundle("org.springframework.osgi.extender", false);
        Bundle dmFragment = installer.installBundle("com.springsource.server.kernel.dm", false);
        installer.satisfyConstraints(new Bundle[]{dmExtender, dmFragment});
        dmExtender.start();

        installBundle(SERVICEABILITY_SUBSYSTEM_NAME, "com.springsource.server.serviceability.ffdc").start();
        
        //this.osgi.waitForBundleApplicationContext(ffdcBundle, 5);
        installBundle(COMMON_SUBSYSTEM_NAME, "com.springsource.server.common.math").start();

        // Register the subsystem contents in the service registry so the kernel can obtain it
        // during startup
        publishSubsystemContents();

        Bundle bundle = installer.installBundle("com.springsource.server.kernel.core");

        if (bundle.getState() != Bundle.ACTIVE) {
            bundle.start();
        }

        // Kernel startup is synchronous, so unregister the bootstrap configuration.
        retractSubsystemContents();
    }

    private void publishSubsystemContents() {
        this.subsystemContentsReg = this.osgi.getBundleContext().registerService(SubsystemContents.class.getName(), this.subsystemContents, null);

    }

    private void retractSubsystemContents() {
        this.subsystemContentsReg.unregister();
    }

    private Bundle installBundle(String subsystemName, String bundleSymbolicName) throws BundleException, UnableToSatisfyDependenciesException {
        Bundle bundle = this.osgi.getBundleInstaller().installBundle(bundleSymbolicName);
        this.subsystemContents.addBundle(subsystemName, bundle);
        return bundle;
    }
}
