/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.kernel.core;

import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;

import org.osgi.framework.Bundle;
import org.osgi.framework.BundleException;
import org.springframework.context.ApplicationContext;

/**
 * Provides a mechanism to start {@link Bundle bundles} and then access any asynchronously created
 * {@link ApplicationContext} via a {@link Future}.<p/>
 * 
 * If a started <code>Bundle</code> is Spring DM-powered and starts successfully, then its {@link ApplicationContext}
 * is available from the monitor {@link Future}. If the start fails, then the failure cause will be wrapped in an
 * {@link ExecutionException} when calling {@link Future#get()}.<p/>
 * 
 * For non-Spring DM <code>Bundles</code>, the monitor <code>Future</code> returns <code>null</code> immediately
 * when <code>get()</code> is called.
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Implementations <strong>must</strong> be threadsafe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
public interface BundleStarter {

    /**
     * Starts the supplied {@link Bundle} and returns a {@link Future} that allows for lazy-access to any
     * {@link ApplicationContext} that might be constructed for the <code>Bundle</code>. If the <code>Bundle</code>
     * doesn't construct an <code>ApplicationContext</code> the <code>Future</code> will return <code>null</code>.
     * 
     * @param bundle the <code>Bundle</code> to start.
     * @return a {@link Future} that can be used to access the constructed <code>AppicationContext</code>.
     * @throws BundleException if the {@link Bundle#start} call fails.
     */
    Future<ApplicationContext> start(Bundle bundle) throws BundleException;
    
    /**
     * Waits for the supplied {@link Bundle} to start and returns a {@link Future} that allows for lazy-access to any
     * {@link ApplicationContext} that might be constructed for the <code>Bundle</code>. If the <code>Bundle</code>
     * doesn't construct an <code>ApplicationContext</code> the <code>Future</code> will return <code>null</code>.
     * 
     * @param bundle the <code>Bundle</code> to start.
     * @return a {@link Future} that can be used to access the constructed <code>AppicationContext</code>.
     * @throws BundleException if the {@link Bundle#start} call fails.
     */
    Future<ApplicationContext> awaitStart(Bundle bundle);
}
