/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.kernel.core.internal;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.osgi.framework.Bundle;
import org.osgi.framework.BundleException;

import com.springsource.server.common.math.ConcurrentMapRelation;
import com.springsource.server.common.math.ConcurrentRelation;
import com.springsource.server.kernel.core.BundleDescriptor;
import com.springsource.server.kernel.core.FatalKernelException;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * KernelAwareBase is an implementation of KernelAware that is passed the kernel on a constructor.
 * 
 * This class is thread safe.
 * 
 * @author Glyn Normington
 * @since 1.0
 * 
 */
public abstract class AbstractSubsystem implements ManagedSubsystem {

    private static final Tracer TRACER = TracerFactory.createTracer(AbstractSubsystem.class);

    private final String name;

    private volatile List<BundleDescriptor> bundleDescriptors;

    private volatile List<Bundle> bundles;

    private final Object stateTransitionMonitor = new Object();

    private volatile State state = State.INSTALLING;

    private final ManagedSubsystemRegistry subsystemRegistry;

    private final ConcurrentRelation<State, State> validSubsystemStateTransitions = new ConcurrentMapRelation<State, State>();

    public AbstractSubsystem(ManagedSubsystemRegistry subsystemRegistry, String name) {
        this.subsystemRegistry = subsystemRegistry;
        this.name = name;
        this.subsystemRegistry.registerSubsystem(this);
        setValidSubsystemStateTransitions();
    }

    private void setValidSubsystemStateTransitions() {
        this.validSubsystemStateTransitions.add(State.INSTALLING, State.INSTALLED);
        this.validSubsystemStateTransitions.add(State.INSTALLING, State.UNINSTALLING);
        this.validSubsystemStateTransitions.add(State.INSTALLED, State.STARTING);
        this.validSubsystemStateTransitions.add(State.STARTING, State.ACTIVE);
        this.validSubsystemStateTransitions.add(State.STARTING, State.STOPPING);
        this.validSubsystemStateTransitions.add(State.ACTIVE, State.STOPPING);
        this.validSubsystemStateTransitions.add(State.STOPPING, State.INSTALLED);
        this.validSubsystemStateTransitions.add(State.INSTALLED, State.UNINSTALLING);
        this.validSubsystemStateTransitions.add(State.UNINSTALLING, State.UNINSTALLED);
        for (State s : State.values()) {
            this.validSubsystemStateTransitions.add(s, State.ERROR);
        }
    }

    /**
     * @param bundles The Subsystem's bundles
     */
    protected void setBundles(List<Bundle> bundles) {
        this.bundles = Collections.unmodifiableList(bundles);
        this.bundleDescriptors = createBundleDescriptors(this.bundles);
    }

    private <T> List<T> reverse(List<T> l) {
        List<T> result = new ArrayList<T>(l.size());
        result.addAll(l);
        Collections.reverse(result);
        return result;
    }

    /**
     * {@inheritDoc}
     */
    public String getName() {
        return this.name;
    }

    /**
     * {@inheritDoc}
     */
    public List<BundleDescriptor> getBundleDescriptors() {
        return this.bundleDescriptors;
    }

    /**
     * {@inheritDoc}
     */
    public List<Bundle> getBundles() {
        return this.bundles;
    }

    /**
     * {@inheritDoc}
     */
    public State getState() {
        return this.state;
    }

    /**
     * Set the subsystem to the specified state.
     * 
     * @param state the new state of the subsystem
     */
    private void setState(State state) {
        if (this.validSubsystemStateTransitions.contains(this.state, state)) {
            TRACER.traceInfo("Subsystem '%s' state changing from %s to %s", this.name, this.state, state);
            synchronized (this.stateTransitionMonitor) {
                this.state = state;
            }
            this.subsystemRegistry.getSubsystemListenerManager().subsystemStateChange(this, state);
        } else {
            TRACER.traceError("Subsystem '%s' state change from %s to %s is not supported", this.name, this.state, state);
            throw new FatalKernelException("Subsystem state change from " + this.state + " to " + state + " is not supported", null);
        }
    }

    /**
     * {@inheritDoc}
     */
    public final ManagedSubsystemRegistry getSubsystemRegistry() {
        return this.subsystemRegistry;
    }

    /**
     * Set the subsystem state to INSTALLED.
     */
    protected final void setInstalled() {
        setState(State.INSTALLED);
    }

    public final void start() {
        if (transitionState(State.STARTING, State.STARTING, State.ACTIVE)) {
            try {
                doStart();
            } catch (RuntimeException e) {
                setState(State.ERROR);
                throw e;
            }
            setState(State.ACTIVE);
        }
    }

    protected void doStart() {
        for (Bundle b : this.bundles) {
            try {
                if (b.getState() != Bundle.STARTING) {
                    TRACER.traceInfo("Subsystem '%s' starting bundle '%s'", this.name, b.getSymbolicName());
                    b.start();
                    TRACER.traceInfo("Subsystem '%s' started bundle '%s'", this.name, b.getSymbolicName());
                }
            } catch (BundleException e) {
                throw new FatalKernelException("Failed to start the bundle '" + b + "'", e);
            }
        }
    }

    public final void stop() {
        if (transitionState(State.STOPPING, State.STOPPING, State.INSTALLED)) {
            try {
                doStop();
            } catch (RuntimeException e) {
                setState(State.ERROR);
                throw e;
            }
            setState(State.INSTALLED);
        }
    }

    protected void doStop() {
        for (Bundle b : reverse(this.bundles)) {
            try {
                if (b.getState() != Bundle.STOPPING) {
                    TRACER.traceInfo("Subsystem '%s' stopping bundle '%s'", this.name, b.getSymbolicName());
                    b.stop();
                    TRACER.traceInfo("Subsystem '%s' stopped bundle '%s'", this.name, b.getSymbolicName());
                }
            } catch (BundleException e) {
                throw new FatalKernelException("Failed to stop the bundle '" + b + "'", e);
            }
        }
    }

    public final void uninstall() {
        if (transitionState(State.UNINSTALLING, State.UNINSTALLING)) {
            try {
                doUninstall();
            } catch (RuntimeException e) {
                setState(State.ERROR);
                throw e;
            }
            setState(State.UNINSTALLED);
            this.subsystemRegistry.unregisterSubsystem(this);
        }
    }

    protected void doUninstall() {
        for (Bundle b : reverse(this.bundles)) {
            try {
                if (b.getState() != Bundle.UNINSTALLED) {
                    TRACER.traceInfo("Subsystem '%s' uninstalling bundle '%s'", this.name, b.getSymbolicName());
                    b.uninstall();
                    TRACER.traceInfo("Subsystem '%s' uninstalled bundle '%s'", this.name, b.getSymbolicName());
                }
            } catch (BundleException e) {
                throw new FatalKernelException("Failed to uninstall the bundle '" + b + "'", e);
            }
        }
    }

    private boolean transitionState(State toSet, State... equivalents) {
        synchronized (this.stateTransitionMonitor) {
            for (State equivalent : equivalents) {
                if (this.state.equals(equivalent)) {
                    return false;
                }
            }
            this.state = toSet;
            return true;
        }
    }
    
    /**
     * Creates a {@link List} of {@link BundleDescriptor BundleDescriptors} from a <code>List</code> of
     * {@link Bundle Bundles}.
     * 
     * @param bundles the source <code>Bundles</code>.
     * @return the resultant <code>BundleDescriptors</code>.
     */
    private List<BundleDescriptor> createBundleDescriptors(List<Bundle> bundles) {
        List<BundleDescriptor> descriptors = new ArrayList<BundleDescriptor>(bundles.size());
        for (Bundle bundle : bundles) {
            descriptors.add(new ImmutableBundleDescriptor(bundle));
        }
        return Collections.unmodifiableList(descriptors);
    }

    /**
     * {@inheritDoc}
     */
    @Override public String toString() {
        return "Subsystem " + this.name + " (" + super.toString() + ")";
    }
}
