/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.kernel.core.internal;

import com.springsource.server.kernel.core.Subsystem;
import com.springsource.server.kernel.core.SubsystemRegistry;
import com.springsource.server.osgi.framework.OsgiFramework;

/**
 * Extension to {@link SubsystemRegistry} that provides additional management capabilities.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * Implementations of this interface must be thread safe.
 * 
 * @author Rob Harrop
 * @author Glyn Normington
 * @since 1.0
 */
interface ManagedSubsystemRegistry extends SubsystemRegistry {

    /**
     * Gets the {@link OsgiFramework} this registry is running in.
     * 
     * @return the <code>OsgiFramework</code>.
     */
    OsgiFramework getOsgiFramework();

    /**
     * Add a subsystem to the subsystem registry.
     * 
     * @param subsystem the subsystem to be registered.
     */
    void registerSubsystem(ManagedSubsystem subsystem);

    /**
     * Remove an uninstalled subsystem from the subsystem registry.
     * 
     * @param subsystem the subsystem to be unregistered.
     */
    void unregisterSubsystem(ManagedSubsystem subsystem);

    /**
     * Retrieve the {@link Subsystem} identified by the supplied name.
     * 
     * @param name the name of the <code>Subsystem</code> to be returned
     * @return the named <code>Subsystem</code> or <code>null</code> if no such <code>Subsystem</code> exists.
     */
    ManagedSubsystem getSubsystemByName(String name);

    /**
     * Return the subsystem that the given class belongs to or null if the class does not belong to a subsystem.
     * 
     * @param clazz the class whose subsystem we want
     * @return the subsystem of the class or null if the class does not belong to a subsystem
     */
    ManagedSubsystem getSubsystemByClass(Class<?> clazz);

    /**
     * Gets the {@link SubsystemListenerManager} associated with this registry.
     * 
     * @return the <code>SubsystemListenerManager</code>.
     */
    SubsystemListenerManager getSubsystemListenerManager();

    /**
     * Set the primordial subsystem that will own classes which do not belong to a bundle installed in a subsystem.
     */
    void setPrimordialSubsystem(ManagedSubsystem primordialSubsystem);
}
