/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.kernel.core.internal;

import java.util.ArrayList;
import java.util.List;

import org.osgi.framework.Bundle;
import org.osgi.framework.BundleException;

import com.springsource.server.common.math.OrderedPair;
import com.springsource.server.kernel.core.FatalKernelException;
import com.springsource.server.kernel.core.Subsystem;
import com.springsource.server.osgi.framework.UnableToSatisfyDependenciesException;
import com.springsource.server.osgi.manifest.VersionRange;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * Provides the default implementation of {@link Subsystem}.<p/>
 * 
 * Implements {@link ManagedSubsystem} to allow collaborating Kernel components to access {@link Bundle} state.
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * Implementation is concurrent.
 * 
 * @author Rob Harrop
 * @author Glyn Normington
 * @since 1.0
 */
final class StandardSubsystem extends AbstractSubsystem {

    private static final Tracer TRACER = TracerFactory.createTracer(StandardSubsystem.class);

    private final List<OrderedPair<String, VersionRange>> bundleLocations;

    /**
     * Creates a new <code>StandardSubsytem</code>
     * @param subsystemRegistry The subsystem registry
     * @param name The name of the subsystem
     * @param bundleLocations The locations of the subsystem's bundles
     */
    StandardSubsystem(ManagedSubsystemRegistry subsystemRegistry, String name, List<OrderedPair<String, VersionRange>> bundleLocations) {
        super(subsystemRegistry, name);
        this.bundleLocations = bundleLocations;
    }

    void install() {
        List<Bundle> bundles = new ArrayList<Bundle>(this.bundleLocations.size());
        for (OrderedPair<String, VersionRange> bundleDefinition : this.bundleLocations) {
            String symbolicName = bundleDefinition.getFirst();
            VersionRange versionRange = bundleDefinition.getSecond();
            try {
                TRACER.traceInfo("Subsystem '%s' installing bundle '%s:%s'", getName(), symbolicName, versionRange.toString());
                Bundle bundle = getSubsystemRegistry().getOsgiFramework().getBundleInstaller().installBundle(symbolicName, versionRange);
                bundles.add(bundle);
            } catch (BundleException ex) {
                throw new FatalKernelException("Unable to install bundle with symbolic name '" + symbolicName + "' and version range '"
                    + versionRange + "'.", ex);
            } catch (UnableToSatisfyDependenciesException utsde) {
                throw new FatalKernelException("Unable to install bundle with symbolic name '" + symbolicName + "' and version range '"
                    + versionRange + "'.", utsde);
            }
        }
        setBundles(bundles);
        setInstalled();
    }
}
