/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.kernel.core.internal;

import java.util.Collection;
import java.util.Collections;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import org.osgi.framework.Bundle;
import org.osgi.framework.BundleEvent;
import org.osgi.framework.SynchronousBundleListener;

import com.springsource.server.kernel.core.FatalKernelException;
import com.springsource.server.kernel.core.KernelException;
import com.springsource.server.kernel.core.Subsystem;
import com.springsource.server.kernel.core.SubsystemInstaller;
import com.springsource.server.kernel.core.SubsystemListener;
import com.springsource.server.kernel.core.SubsystemRegistry;
import com.springsource.server.kernel.core.Subsystem.State;
import com.springsource.server.osgi.framework.OsgiFramework;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * Standard implementation of {@link SubsystemRegistry}.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @author Glyn Normington
 * @since 1.0
 */
final class StandardSubsystemRegistry implements ManagedSubsystemRegistry {

    private static final Tracer TRACER = TracerFactory.createTracer(StandardSubsystemRegistry.class);

    /**
     * The name of the {@link Subsystem} representing the kernel itself.
     */
    public static final String KERNEL_SUBSYSTEM_NAME = "com.springsource.server.kernel";

    private final OsgiFramework osgiFramework;

    private final Map<Bundle, ManagedSubsystem> bundleSubsystem = new ConcurrentHashMap<Bundle, ManagedSubsystem>();

    private final Map<String, ManagedSubsystem> subsystems = new ConcurrentHashMap<String, ManagedSubsystem>();

    private final SubsystemListenerManager listenerManager = new SubsystemListenerManager();

    private volatile ManagedSubsystem primordialSubsystem;

    /**
     * Creates a new <code>StandardSubsystemRegistry</code>.
     * 
     * @param osgiFramework the {@link OsgiFramework} to install into.
     */
    public StandardSubsystemRegistry(OsgiFramework osgiFramework) {
        this.osgiFramework = osgiFramework;
        this.listenerManager.registerSubsystemListener(new SubsystemListener() {

            public void subSystemStateChange(Subsystem subsystem, State newState) {
                if (newState == State.INSTALLED) {
                    registerBundles((ManagedSubsystem) subsystem);
                }

            }
        });

        /*
         * Respond to bundle lifecycle events by managing any corresponding subsystem accordingly. Once this OSGi
         * framework starts shutting down, abandon this approach as Spring DM stops all the Server's beans which turns
         * each service proxy into a source of deadlock. A little later in shutdown, the OSGi framework goes through all
         * the installed bundles in reverse order of installation and stops them, which would provoke many deadlocks if
         * this also caused the Server subsystems to stop.
         */
        this.osgiFramework.getBundleContext().addBundleListener(new SynchronousBundleListener() {

            private volatile boolean shuttingDown = false;

            private String bundleEventAsString(BundleEvent event) {
                switch (event.getType()) {
                    case BundleEvent.INSTALLED:
                        return "INSTALLED";
                    case BundleEvent.LAZY_ACTIVATION:
                        return "LAZY_ACTIVATION";
                    case BundleEvent.RESOLVED:
                        return "RESOLVED";
                    case BundleEvent.STARTED:
                        return "STARTED";
                    case BundleEvent.STARTING:
                        return "STARTING";
                    case BundleEvent.STOPPED:
                        return "STOPPED";
                    case BundleEvent.STOPPING:
                        return "STOPPING";
                    case BundleEvent.UNINSTALLED:
                        return "UNINSTALLED";
                    case BundleEvent.UNRESOLVED:
                        return "UNRESOLVED";
                    case BundleEvent.UPDATED:
                        return "UPDATED";
                    default:
                        return "???";
                }
            }

            public void bundleChanged(BundleEvent event) {
                TRACER.traceInfo("Bundle event '%s' for bundle with symbolic name '%s'", bundleEventAsString(event),
                    event.getBundle().getSymbolicName());
                if (event.getBundle().getBundleId() == 0 && event.getType() == BundleEvent.STOPPING) {
                    this.shuttingDown = true;
                } else if (!this.shuttingDown) {
                    ManagedSubsystem subsystem = StandardSubsystemRegistry.this.bundleSubsystem.get(event.getBundle());
                    if (subsystem != null) {
                        switch (event.getType()) {
                            case BundleEvent.STARTING:
                            case BundleEvent.STARTED:
                                // The kernel is already ACTIVE when its
                                // bundle(s) become STARTED.
                                if (!subsystem.getName().equals(KERNEL_SUBSYSTEM_NAME)) {
                                    subsystem.start();
                                }
                                break;
                            case BundleEvent.STOPPING:
                            case BundleEvent.STOPPED:
                                subsystem.stop();
                                break;
                            case BundleEvent.UNINSTALLED:
                                subsystem.uninstall();
                                break;
                            default:
                                break;
                        }
                    }
                }
            }
        });
    }

    /**
     * {@inheritDoc}
     */
    public void registerSubsystem(ManagedSubsystem subsystem) {
        if (this.subsystems.containsKey(subsystem.getName())) {
            throw new FatalKernelException("Subsystem already present in the subsystem registry", null);
        }
        this.subsystems.put(subsystem.getName(), subsystem);
        TRACER.traceInfo("Subsystem '%s' registered", subsystem);
    }

    /**
     * {@inheritDoc}
     */
    public void unregisterSubsystem(ManagedSubsystem subsystem) {
        if (!this.subsystems.containsKey(subsystem.getName())) {
            throw new FatalKernelException("Subsystem not present in the subsystem registry", null);
        }
        if (subsystem.getState() != Subsystem.State.UNINSTALLED) {
            throw new FatalKernelException("Subsystem must be uninstalled before being removed from the subsystem registry", null);
        }

        // Clean up the state corresponding to the subsystem being unregistered.
        for (Bundle bundle : this.bundleSubsystem.keySet()) {
            if (this.bundleSubsystem.get(bundle) == subsystem) {
                this.bundleSubsystem.remove(bundle);
            }
        }
        this.subsystems.remove(subsystem.getName());
        TRACER.traceInfo("Subsystem '%s' unregistered", subsystem);
    }

    /**
     * @param subsystem
     */
    private void registerBundles(ManagedSubsystem subsystem) {
        for (Bundle b : subsystem.getBundles()) {
            this.bundleSubsystem.put(b, subsystem);
        }
    }

    /**
     * {@inheritDoc} <p/>
     * 
     * Returns a "weakly consistent" view of the <code>Subsystems</code> that is guaranteed to be consistent with the
     * view of <code>Subsystems</code> at the time of calling. Concurrent modifications are not guaranteed to be
     * reflected.
     * @see ConcurrentHashMap#values()
     * 
     */
    public Collection<Subsystem> getSubsystems() {
        Collection<? extends Subsystem> subsystems = this.subsystems.values();
        return Collections.unmodifiableCollection(subsystems);
    }

    /**
     * {@inheritDoc}
     */
    public ManagedSubsystem getSubsystemByName(String name) {
        return this.subsystems.get(name);
    }

    /**
     * {@inheritDoc}
     */
    public Subsystem getSubsystemByBundle(Bundle bundle) {
        return this.bundleSubsystem.get(bundle);
    }

    /**
     * {@inheritDoc}
     */
    public ManagedSubsystem getSubsystemByClass(Class<?> clazz) {
        ManagedSubsystem subsystem = null;
        Bundle bundle = this.osgiFramework.getClassBundle(clazz);
        if (bundle != null) {
            subsystem = this.bundleSubsystem.get(bundle);
            if (subsystem == null) {
                throw new FatalKernelException("Subsystem for class '" + clazz + "' cannot be determined as the class's bundle '" + bundle
                    + "' does not belong to a subsystem known by the kernel", null);
            }
        } else if (this.primordialSubsystem != null) {
            subsystem = this.primordialSubsystem;
        } else {
            throw new RuntimeException("Subsystem for class '" + clazz
                + "' cannot be determined as the class does not belong to a bundle which has been installed by the kernel");
        }
        return subsystem;
    }

    /**
     * {@inheritDoc}
     */
    public void addSubsystemListener(SubsystemListener listener) {
        this.listenerManager.registerSubsystemListener(listener);
    }

    /**
     * {@inheritDoc}
     */
    public SubsystemListenerManager getSubsystemListenerManager() {
        return this.listenerManager;
    }

    /**
     * {@inheritDoc}
     */
    public OsgiFramework getOsgiFramework() {
        return this.osgiFramework;
    }

    /**
     * {@inheritDoc}
     */
    public Subsystem getKernelSubsystem() {
        return getSubsystemByName(KERNEL_SUBSYSTEM_NAME);
    }

    /**
     * {@inheritDoc}
     */
    public void setPrimordialSubsystem(ManagedSubsystem primordialSubsystem) {
        if (this.primordialSubsystem != null) {
            throw new FatalKernelException("Primordial subsystem already set", null);
        }
        this.primordialSubsystem = primordialSubsystem;
    }

    /**
     * {@inheritDoc}
     */
    public Subsystem installSubsystem(String name) throws KernelException {
        SubsystemInstaller kernelSubsystemInstaller = (KernelSubsystem) getKernelSubsystem();
        return kernelSubsystemInstaller.installSubsystem(name);
    }
}
