/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.kernel.core.internal.install;

import static java.io.File.separator;

import java.io.File;

import org.springframework.core.io.FileSystemResource;

import com.springsource.server.bootstrap.BootstrapConfiguration;
import com.springsource.server.kernel.core.KernelException;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * Standard {@link SubsystemResolver} implementation that looks for subsystems in the
 * {@link BootstrapConfiguration#getSubsystemsDirectory() subsystem directory}.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
public class StandardSubsystemResolver implements SubsystemResolver {

    private static final String SUBSYSTEM_PLAN_FILENAME = "subsystem.xml";
    
    private final Tracer TRACE = TracerFactory.createTracer(StandardSubsystemResolver.class);

    private final XmlSubsystemDescriptorParser parser = new XmlSubsystemDescriptorParser();

    private final String subsystemsDirectoryPath;

    /**
     * Creates a new <code>StandardSubsystemResolver</code>.
     * 
     * @param bootstrapConfiguration the <code>BootstrapConfiguration</code>.
     */
    public StandardSubsystemResolver(BootstrapConfiguration bootstrapConfiguration) {
        this.subsystemsDirectoryPath = determineSubsystemsDirectory(bootstrapConfiguration);
    }

    /**
     * Determines the root subsystem directory.
     * 
     * @param bootstrapConfiguration the supplied {@link BootstrapConfiguration}.
     * @return the subsystem directory path.
     */
    private String determineSubsystemsDirectory(BootstrapConfiguration bootstrapConfiguration) {
        String subsystemsDirectory = null;

        if (bootstrapConfiguration != null) {
            subsystemsDirectory = bootstrapConfiguration.getSubsystemsDirectory();
        }

        if (subsystemsDirectory == null) {
            subsystemsDirectory = ".." + File.separatorChar + "..";
        }

        if (!subsystemsDirectory.endsWith(File.separator)) {
            subsystemsDirectory = subsystemsDirectory + File.separatorChar;
        }

        return subsystemsDirectory;
    }

    /**
     * {@inheritDoc}
     */
    public SubsystemDescriptor resolveSubsystemDescriptor(String subsystemName) throws KernelException {
        String subsystemDecriptorFilePath = new StringBuilder(this.subsystemsDirectoryPath).append(subsystemName).append(separator).append(SUBSYSTEM_PLAN_FILENAME).toString();
        FileSystemResource fileSystemResource = new FileSystemResource(subsystemDecriptorFilePath.toString());
        if(fileSystemResource.exists()){
            return this.parser.parse(fileSystemResource);
        }else{
            TRACE.traceWarning("The descriptor for subsystem '%s' is not in the expected place '%s' on the file system.", subsystemName, subsystemDecriptorFilePath);
            return null;
        }
    }
}
