/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.kernel.core.internal.install;

import java.util.ArrayList;
import java.util.List;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.springframework.core.io.Resource;
import org.springframework.util.StringUtils;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import com.springsource.server.common.math.OrderedPair;
import com.springsource.server.kernel.core.KernelException;
import com.springsource.server.kernel.core.diagnostics.KernelLogEvent;
import com.springsource.server.osgi.manifest.VersionRange;
import com.springsource.server.serviceability.logging.LogService;
import com.springsource.server.serviceability.logging.LogServiceFactory;

/**
 * Parses XML-formatted {@link SubsystemDescriptor SubsystemDescriptors}.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * Implementation is entirely stateless and as such is suitable for concurrent access.
 * 
 * @author Rob Harrop
 * @author Glyn Normington
 * @since 1.0
 */
public final class XmlSubsystemDescriptorParser {

    /**
     * The name of the '<code>name</code>' attribute.
     */
    private static final String NAME_ATTRIBUTE = "name";

    /**
     * The name of the '<code>description</code>' attribute.
     */
    private static final String DESCRIPTION_ATTRIBUTE = "description";

    /**
     * The name of the '<code>bundle</code>' element.
     */
    private static final String BUNDLE_ELEMENT = "bundle";

    /**
     * The name of the '<code>symbolic-name</code>' attribute.
     */
    private static final String SYMBOLIC_NAME_ATTRIBUTE = "symbolic-name";

    /**
     * The name of the '<code>version</code>' attribute.
     */
    private static final String VERSION_ATTRIBUTE = "version";
    
    private static final LogService LOGGER = LogServiceFactory.getLogService();

    /**
     * Parses the supplied XML {@link Resource} and creates a corresponding {@link SubsystemDescriptor}.
     * 
     * @param xmlResource the <code>Resource</code> to parse.
     * @return the resultant {@link  SubsystemDescriptor}.
     * @throws KernelException 
     */
    public SubsystemDescriptor parse(Resource xmlResource) throws KernelException {
        Document doc = readDocument(xmlResource);
        Element element = doc.getDocumentElement();
        return parseSubsystemElement(element);
    }

    private XmlSubsystemDescriptor parseSubsystemElement(Element subsystemElement) {
        String name = subsystemElement.getAttribute(NAME_ATTRIBUTE);
        String description = subsystemElement.getAttribute(DESCRIPTION_ATTRIBUTE);
        NodeList bundleElements = subsystemElement.getElementsByTagName(BUNDLE_ELEMENT);
        List<OrderedPair<String, VersionRange>> bundleLocations = new ArrayList<OrderedPair<String, VersionRange>>(bundleElements.getLength());
        for (int n = 0; n < bundleElements.getLength(); n++) {
            Element bundleElement = (Element) bundleElements.item(n);
            parseBundleElement(bundleElement, bundleLocations);
        }
        return new XmlSubsystemDescriptor(name, description, bundleLocations);
    }

    /**
     * Parses the <code>&lt;bundle&gt;</code> element and adds the resultant {@link OrderedPair} to
     * <code>bundleDefinitions</code>.
     * 
     * @param bundleElement the {@link Element} being parsed.
     * @param bundleDefinitions the {@link List} to add the bundle definition <code>OrderedPair</code> to.
     */
    private void parseBundleElement(Element bundleElement, List<OrderedPair<String, VersionRange>> bundleDefinitions) {
        String symbolicName = bundleElement.getAttribute(SYMBOLIC_NAME_ATTRIBUTE);
        if (!StringUtils.hasText(symbolicName)) {
            LOGGER.log(KernelLogEvent.PARSE_DESCRIPTOR_BUNDLE_SYMBOLIC_NAME_MISSING, bundleElement);
            throw new SubsystemDescriptorParseException("Bundle symbolic name is missing");
        }
        String versionRange = bundleElement.getAttribute(VERSION_ATTRIBUTE);
        bundleDefinitions.add(new OrderedPair<String, VersionRange>(symbolicName, new VersionRange(versionRange)));
    }

    /**
     * Reads the DOM {@link Document} at the supplied {@link  Resource}.
     * 
     * @param xmlResource the XML <code>Resource</code>.
     * @return the parsed <code>Document</code>.
     * @throws KernelException 
     */
    private Document readDocument(Resource xmlResource) throws KernelException {
        try {
            DocumentBuilder builder = createDocumentBuilderFactory().newDocumentBuilder();
            return builder.parse(xmlResource.getInputStream());
        } catch (Exception e) {
            throw new KernelException("Exception reading subsystem descriptor", e);
        }
    }

    /**
     * Creates a new {@link DocumentBuilderFactory} instance.
     * 
     * @return the new <code>DocumentBuilderFactory</code>.
     */
    private DocumentBuilderFactory createDocumentBuilderFactory() {
        return DocumentBuilderFactory.newInstance();
    }

}
