/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.kernel.dm;

import java.util.HashSet;
import java.util.Set;

import org.osgi.framework.BundleContext;
import org.osgi.framework.BundleException;
import org.osgi.framework.InvalidSyntaxException;
import org.springframework.beans.BeansException;
import org.springframework.beans.MutablePropertyValues;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.beans.factory.config.ConfigurableListableBeanFactory;
import org.springframework.osgi.extender.OsgiBeanFactoryPostProcessor;
import org.springframework.osgi.service.importer.support.ImportContextClassLoader;
import org.springframework.osgi.service.importer.support.OsgiServiceCollectionProxyFactoryBean;
import org.springframework.osgi.service.importer.support.OsgiServiceProxyFactoryBean;

/**
 * {@link OsgiBeanFactoryPostProcessor} that ensures that all service references are in unmanaged mode for thread
 * context class loader propagation.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
public class ContextClassLoaderOsgiPostProcessor implements OsgiBeanFactoryPostProcessor {

    private static final String PROPERTY_CONTEXT_CLASS_LOADER = "contextClassLoader";

    private static final Set<String> IMPORTER_CLASS_NAMES = createImportClassNames();

    /**
     * {@inheritDoc}
     */
    public void postProcessBeanFactory(BundleContext bundleContext, ConfigurableListableBeanFactory beanFactory) throws BeansException,
        InvalidSyntaxException, BundleException {
        String[] beanDefinitionNames = beanFactory.getBeanDefinitionNames();
        for (String name : beanDefinitionNames) {
            BeanDefinition beanDefinition = beanFactory.getBeanDefinition(name);
            if (isServiceImportDefinition(beanDefinition)) {
                MutablePropertyValues propertyValues = beanDefinition.getPropertyValues();
                propertyValues.addPropertyValue(PROPERTY_CONTEXT_CLASS_LOADER, ImportContextClassLoader.UNMANAGED.getLabel());
            }
        }
    }

    private static Set<String> createImportClassNames() {
        Set<String> names = new HashSet<String>();
        names.add(OsgiServiceProxyFactoryBean.class.getName());
        names.add(OsgiServiceCollectionProxyFactoryBean.class.getName());
        return names;
    }

    private boolean isServiceImportDefinition(BeanDefinition beanDefinition) {
        return IMPORTER_CLASS_NAMES.contains(beanDefinition.getBeanClassName());
    }

}
