/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.kernel.dm;

import java.util.concurrent.Executor;
import java.util.concurrent.Executors;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.atomic.AtomicInteger;

import org.osgi.framework.BundleContext;
import org.osgi.framework.ServiceReference;
import org.springframework.beans.factory.DisposableBean;
import org.springframework.core.task.TaskExecutor;
import org.springframework.osgi.context.BundleContextAware;

import com.springsource.server.serviceability.tracing.TracingService;

/**
 * {@link TaskExecutor} implementation that propagates trace context to spawned threads.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
public class ContextPropagatingTaskExecutor implements TaskExecutor, BundleContextAware, DisposableBean {

    private final AtomicInteger threadCount = new AtomicInteger();

    private final Executor executor;

    private volatile ServiceReference tracingServiceRef;

    private volatile TracingService tracingService;

    private volatile BundleContext bundleContext;

    private ContextPropagatingTaskExecutor(int poolSize) {
        this.executor = Executors.newFixedThreadPool(poolSize, new ThreadFactory() {

            public Thread newThread(Runnable r) {
                Thread t = new Thread(r);
                t.setName("server-dm-" + ContextPropagatingTaskExecutor.this.threadCount.getAndIncrement());
                return t;
            }

        });
    }

    /**
     * {@inheritDoc}
     */
    public void execute(final Runnable task) {
        final String applicationName = this.tracingService.getCurrentApplicationName();
        final ClassLoader threadContextClassLoader = Thread.currentThread().getContextClassLoader();
        this.executor.execute(new Runnable() {

            public void run() {
                ClassLoader originalContextClassLoader = Thread.currentThread().getContextClassLoader();
                try {
                    Thread.currentThread().setContextClassLoader(threadContextClassLoader);
                    ContextPropagatingTaskExecutor.this.tracingService.setCurrentApplicationName(applicationName);
                    task.run();
                } finally {
                    ContextPropagatingTaskExecutor.this.tracingService.setCurrentApplicationName(null);
                    Thread.currentThread().setContextClassLoader(originalContextClassLoader);
                }
            }

        });
    }

    /**
     * {@inheritDoc}
     */
    public void setBundleContext(BundleContext bundleContext) {
        this.bundleContext = bundleContext;
        this.tracingServiceRef = bundleContext.getServiceReference(TracingService.class.getName());
        if (this.tracingServiceRef == null) {
            throw new IllegalStateException("Unable to access TracingService");
        }
        this.tracingService = (TracingService) bundleContext.getService(this.tracingServiceRef);
    }

    /**
     * {@inheritDoc}
     */
    public void destroy() throws Exception {
        if (this.tracingServiceRef != null && this.bundleContext != null) {
            this.bundleContext.ungetService(this.tracingServiceRef);
        }
    }

}
