/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.kernel.dm;

import org.osgi.framework.BundleContext;
import org.osgi.framework.BundleEvent;
import org.osgi.framework.InvalidSyntaxException;
import org.osgi.framework.ServiceReference;
import org.osgi.framework.SynchronousBundleListener;
import org.springframework.context.ApplicationContext;
import org.springframework.osgi.service.importer.support.OsgiServiceProxyFactoryBean;

import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * A {@link SynchronousBundleListener} that reacts to a {@link BundleEvent.STOPPING STOPPING} event
 * for the system bundle. Upon such an event being received the listener retrieves all
 * {@link ApplicationContext ApplicationContexts} from the service registry and disables retry on any
 * {@link OsgiServiceProxyFactoryBean OsgiServiceProxyFactoryBeans} which they contain.
 * <p />
 *
 * <strong>Concurrent Semantics</strong><br />
 *
 * Thread-safe.
 *
 * @since 1.0.1
 * 
 * @see ApplicationContextShutdownBean#disableServiceProxyRetry(ApplicationContext)
 */
class ServiceProxyRetryDisablingBundleListener implements SynchronousBundleListener {
    
    private Tracer TRACER = TracerFactory.createTracer(ServiceProxyRetryDisablingBundleListener.class);
    
    ServiceProxyRetryDisablingBundleListener(BundleContext bundleContext) {
        bundleContext.addBundleListener(this);
    }
    
    public void bundleChanged(BundleEvent event) {
        if (event.getBundle().getBundleId() == 0 && event.getType() == BundleEvent.STOPPING) {
            BundleContext bundleContext = event.getBundle().getBundleContext();
            try {
                ServiceReference[] applicationContextServiceReferences = event.getBundle().getBundleContext().getAllServiceReferences(ApplicationContext.class.getName(), null);
                
                for (ServiceReference applicationContextServiceReference : applicationContextServiceReferences) {
                    ApplicationContext applicationContext = (ApplicationContext)bundleContext.getService(applicationContextServiceReference);
                    ApplicationContextShutdownBean.disableServiceProxyRetry(applicationContext);
                    bundleContext.ungetService(applicationContextServiceReference);
                }
            } catch (InvalidSyntaxException ise) {
                TRACER.traceError("Failed to retrieve all application contexts from service registry", ise);
            }
        }
    }            
}