/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.management.proxy.server;

import java.util.Properties;

import org.osgi.framework.BundleContext;
import org.osgi.framework.ServiceReference;

/**
 * Standard implementation of <code>ServerPropertiesResourceInfo</code>. Only one should exist per instance of the
 * server.
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * This class is thread safe
 * 
 * @author Christopher Frost
 * @since 1.0
 */
class StandardServerPropertiesResourceInfo implements ServerPropertiesResourceInfo {

    private final Properties properties;

    /**
     * Deals with the creation and registration of a <code>StandardServerPropertiesResourceInfo</code> instance in the
     * OSGi Service Registry. If an instance already exists in the service registry, then that instance is returned, and
     * a new instance is not created.
     * 
     * @param bundleContext the BundleContext to use to register the ServerPropertiesResourceInfo in the OSGi Service
     *        Registry
     * @return an existing or newly created ServerPropertiesResourceInfo
     */
    public static ServerPropertiesResourceInfo createAndRegisterServerPropertiesResourceInfo(BundleContext bundleContext) {
        ServerPropertiesResourceInfo serverPropertiesResourceInfo;

        ServiceReference serviceReference = bundleContext.getServiceReference(ServerPropertiesResourceInfo.class.getName());
        if (serviceReference == null) {
            serverPropertiesResourceInfo = new StandardServerPropertiesResourceInfo();
            bundleContext.registerService(ServerPropertiesResourceInfo.class.getName(), serverPropertiesResourceInfo, null);
        } else {
            serverPropertiesResourceInfo = (ServerPropertiesResourceInfo) bundleContext.getService(serviceReference);
        }

        return serverPropertiesResourceInfo;
    }

    /**
     * Private constructor to prevent creation of unregistered instances of this class.
     */
    private StandardServerPropertiesResourceInfo() {
        Package pkg = StandardServerPropertiesResourceInfo.class.getPackage();
        this.properties = new Properties();
        this.properties.put("SpringSource dm Server", pkg != null && pkg.getImplementationVersion() != null ? pkg.getImplementationVersion()
            : "Version information unavailable");
        this.properties.put("Embedded Tomcat", "Version 6.0.18");
        this.properties.put("System Architecture", System.getProperty("os.arch"));
        this.properties.put("Operating System", System.getProperty("os.name") + " - " + System.getProperty("os.version"));
        this.properties.put("Java Vendor", System.getProperty("java.vendor"));
        this.properties.put("Java Version", System.getProperty("java.version"));
        this.properties.put("Default Time Zone", System.getProperty("user.timezone"));
    }

    /**
     * {@inheritDoc}
     */
    public Properties getProperties() {
        return this.properties;
    }

}
