/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.management.proxy.serviceability;


import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import org.osgi.framework.BundleContext;
import org.osgi.framework.ServiceRegistration;

import com.springsource.server.serviceability.output.Destination;
import com.springsource.server.serviceability.output.DestinationCreationListener;
import com.springsource.server.serviceability.output.DestinationRegistry;

/**
 * <p>
 * A <code>DestinationCreationListener</code> implementation can be registered with a {@link DestinationRegistry} to
 * receive notifications for destinations created by the registry.
 * </p>
 *
 * <strong>Concurrent Semantics</strong><br />
 *
 * This class is Thread Safe
 *
 * @author Christopher Frost
 * @since 1.0
 * @see DestinationRegistry#addDestinationCreationListener(DestinationCreationListener)
 */
public class StandardDestinationCreationListener implements DestinationCreationListener {

    private final BundleContext bundleContext;
    
    private Map<String, ServiceRegistration> registeredServices = new ConcurrentHashMap<String, ServiceRegistration>();
    
    /**
     * Hold a reference to the <code>BundleContext</code> and registers itself with the 
     * destination registry so that it can be informed of any changes to the serviceability 
     * output destinations.   
     * 
     * @param bundleContext
     * @param destinationRegistry
     */
    public StandardDestinationCreationListener(BundleContext bundleContext, DestinationRegistry destinationRegistry) {
        this.bundleContext = bundleContext;
        destinationRegistry.addDestinationCreationListener(this);
    }
    
    /**
     * 
     * {@inheritDoc}
     */
    public void destinationCreated(String destinationIdentifier, Destination destination) {
        DestinationServiceInfo destinationServiceInfo = new StandardDestinationServiceInfo(destinationIdentifier, destination);
        ServiceRegistration registration = this.bundleContext.registerService(DestinationServiceInfo.class.getName(), destinationServiceInfo, null);
        registeredServices.put(destinationIdentifier, registration);
    }

	/**
	 * {@inheritDoc}
	 */
    public void destinationDestroyed(String destinationIdentifier) {
		ServiceRegistration registration = this.registeredServices.get(destinationIdentifier);
		if (registration != null) {
			registration.unregister();
		}
	}
}
