/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.framework;

import java.util.List;

import com.springsource.server.osgi.manifest.BundleManifest;
import com.springsource.server.osgi.manifest.Scoper;

/**
 * {@link ImportExpander} is an interface for expanding library and bundle imports into package imports and propagating
 * promoted imports across a collection of bundles. <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Implementations of this interface should be thread safe.
 * 
 * @author Glyn Normington
 * @author Andy Wilkinson
 * @since 1.0
 */
public interface ImportExpander {

    public interface ImportPromotionVector {
    }

    /**
     * Modifies the supplied {@link BundleManifest bundle manifests} by replacing all of the <code>Import-Library</code>
     * and <code>Import-Bundle</code> header entries with the equivalent <code>Import-Package</code> header entries and
     * propagating promoted imports across the supplied bundle manifests.
     * 
     * @param bundleManifests the manifests to perform <code>Import-Library</code> and <code>Import-Bundle</code>
     *        expansion upon.
     * @throws UnableToSatisfyDependenciesException if a manifest's dependencies cannot be satisfied
     * @throws ImportMergeException if there was a clash between some of the imports being merged
     */
    ImportPromotionVector expandImports(List<BundleManifest> bundleManifests) throws UnableToSatisfyDependenciesException, ImportMergeException;

    /**
     * Modifies the {@link BundleManifest} with the given bundle symbolic name in the given list of manifests by
     * replacing all of the <code>Import-Library</code> and <code>Import-Bundle</code> header entries with the
     * equivalent <code>Import-Package</code> header entries and propagating the promoted imports in the given
     * {@link ImportPromotionVector}.
     * 
     * @param bundleManifests
     * @param bundleSymbolicName
     * @param importPromotionVector
     * @param scoper the {@link Scoper} to apply to unscoped bundle symbolic names or <code>null</code> if not required
     * @return <code>true</code> if and only if the expansion could be performed without requiring the refresh to be
     *         escalated to a redeployment
     * @throws UnableToSatisfyDependenciesException
     * @throws ImportMergeException if there was a clash between some of the imports being merged
     */
    boolean expandOnRefresh(List<BundleManifest> bundleManifests, String bundleSymbolicName, ImportPromotionVector importPromotionVector,
        Scoper scoper) throws UnableToSatisfyDependenciesException, ImportMergeException;
}
