/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.framework;

import java.util.jar.Manifest;

import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.BundleException;
import org.osgi.service.packageadmin.PackageAdmin;

/**
 * <code>OsgiFramework</code> defines a basic abstraction for interacting with various OSGi implementations in a
 * server-independent manner. <p/>
 * 
 * Implementations <strong>must</strong> make themselves available in the OSGi service registry under both the
 * {@link BundleInstaller} and <code>OsgiFramework</code> class names.
 * 
 * @author Rob Harrop
 * @author Glyn Normington
 * @author Sam Brannen
 * @since 1.0
 */
public interface OsgiFramework {

    /**
     * Name of the {@link Manifest} entry defining the scope of a bundle.
     */
    public static String HEADER_MODULE_SCOPE = "Module-Scope";

    /**
     * Gets the {@link BundleInstaller} for the <code>OsgiFramework</code> instance.
     * 
     * @return the <code>BundleInstaller</code>.
     */
    BundleInstaller getBundleInstaller();

    /**
     * Stops the underlying OSGi framework.
     */
    void stop();

    /**
     * Gets the {@link BundleContext} of the system {@link Bundle}.
     * 
     * @return the system <code>BundleContext</code>.
     */
    BundleContext getBundleContext();

    /**
     * Sets a named framework property to the supplied value.
     * 
     * @param name the name of the framework property to set.
     * @param value the value for the framework property
     */
    void setFrameworkProperty(String name, String value);

    /**
     * Gets the value of a named framework property.
     * 
     * @param name the name of the framework property.
     * @return the framework property value.
     */
    String getFrameworkProperty(String name);

    /**
     * Determines {@link Bundle} for the supplied <code>Class</code>.
     * 
     * @param cls the class whose bundle is being determined
     * @return the class's bundle
     * @see PackageAdmin#getBundle(Class)
     */
    Bundle getClassBundle(Class<?> cls);

    /**
     * Gets the {@link ClassLoader} for the supplied {@link Bundle}.
     * 
     * @param bundle the <code>Bundle</code>.
     * @return the <code>ClassLoader</code>.
     */
    ClassLoader getBundleClassLoader(Bundle bundle);

    /**
     * Determines if the given {@link Bundle bundle} publishes (or <em>is expected to publish</em>) an application
     * context in the OSGi service registry.
     * 
     * @param bundle the bundle to check
     * @return <code>true</code> if the bundle is expected to publish an application context
     */
    boolean bundlePublishesApplicationContext(Bundle bundle);

    /**
     * Wait for an application context for the given bundle to be published in the OSGi service registry.
     * 
     * If the bundle is defined such that an application context will not be published, then this method returns
     * <code>true</code> immediately.
     * 
     * @param bundle the bundle to wait for
     * @param timeout the number of seconds to wait before timing out
     * @return <code>true</code> if and only if any expected application context was published before the timeout
     *         expired
     */
    boolean waitForBundleApplicationContext(Bundle bundle, int timeout);

    /**
     * Wait for an application context with the given bundle symbolic name to be published in the OSGi service registry.
     * 
     * Note: this only makes sense for bundles, such as those internal to the server, for which only a single version
     * is ever installed and started.
     * 
     * @param symbolicName the bundle symbolic name of the application context
     * @param timeout the number of seconds to wait before timing out
     * @return <code>true</code> if and only if the application context was published before the timeout expired
     */
    boolean waitForBundleApplicationContext(String symbolicName, int timeout);

    /**
     * Returns true if the class with the given name will be loaded via boot delegation by the underlying OSGi framework
     * 
     * @param className The class name
     * @return Whether the underlying framework considers the class to be boot delegated.
     */
    boolean isBootDelegated(String className);

    /**
     * Returns all of the supplied bundle's direct dependencies. A bundle's direct dependencies are the bundles to which
     * its package imports have been wired. Equivalent to calling <code>getDirectDependencies(bundle, false)</code>.
     * 
     * @param bundle The bundle for which the direct dependencies are required
     * @return the supplied bundle's direct dependencies
     */
    Bundle[] getDirectDependencies(Bundle bundle);

    /**
     * Returns all of the supplied bundle's direct dependencies, optionally including the fragments, if any, of each
     * direct dependency in the returned array. A bundle's direct dependencies are the bundles to which its package
     * imports have been wired.
     * 
     * @param bundle The bundle for which the direct dependencies are required
     * @param includeFragments <code>true</code> if fragments of bundles to which package imports have been wired
     *        should be included in the returned array
     * @return the supplied bundle's direct dependencies
     */
    Bundle[] getDirectDependencies(Bundle bundle, boolean includeFragments);

    /**
     * Refreshes the supplied {@link Bundle}. Reloads all contents, and refreshes the packages.
     * 
     * @param bundle the <code>Bundle</code> to refresh.
     * @throws BundleException if there is an error during refresh.
     * @see PackageAdmin#refreshPackages(Bundle[])
     */
    void refresh(Bundle bundle) throws BundleException;
}
