/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.framework;

import java.util.Dictionary;
import java.util.Enumeration;

import org.osgi.framework.Bundle;
import org.osgi.framework.Constants;
import org.osgi.framework.Version;
import org.springframework.util.StringUtils;

/**
 * Utility methods for working the OSGi framework and {@link Bundle Bundles}. <p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
public final class OsgiFrameworkUtils {

    private static final String CONTEXT_DIR = "/META-INF/spring/";

    private static final String CONTEXT_FILES = "*.xml";

    private static final String SPRING_CONTEXT_HEADER = "Spring-Context";

    /**
     * Gets the {@link Version} for the supplied {@link Bundle}.
     * 
     * @param bundle the <code>Bundle</code>.
     * @return the <code>Bundles</code> <code>Version</code>.
     */
    public static Version getBundleVersion(Bundle bundle) {
        String header = (String) bundle.getHeaders().get(Constants.BUNDLE_VERSION);
        if (StringUtils.hasText(header)) {
            return Version.parseVersion(header);
        } else {
            return Version.emptyVersion;
        }
    }

    /**
     * Gets the scope for the supplied {@link Bundle}.
     * 
     * @param bundle the <code>Bundle</code>.
     * @return the scope, or <code>null</code> if no scope is specified.
     */
    public static String getScope(Bundle bundle) {
        return (String) bundle.getHeaders().get(OsgiFramework.HEADER_MODULE_SCOPE);
    }

    /**
     * Queries whether the supplied {@link Bundle} is scoped.
     * 
     * @param bundle the <code>Bundle</code>.
     * @return <code>true</code> if the <code>Bundle</code> is scoped, otherwise <code>false</code>.
     */
    public static boolean isScoped(Bundle bundle) {
        String scope = getScope(bundle);
        return scope != null && scope.length() > 0;
    }

    /**
     * Queries whether both supplied {@link Bundle Bundles} are in the same scope.
     * 
     * @param a the first <code>Bundle</code>.
     * @param b the second <code>Bundle</code>.
     * @return <code>true</code> if both <code>Bundles</code> are in the same scope.
     */
    public static boolean sameScope(Bundle a, Bundle b) {
        String scopeA = getScope(a);
        if (scopeA == null || scopeA.length() == 0) {
            return false;
        } else {
            return scopeA.equals(getScope(b));
        }
    }

    /**
     * Queries whether the supplied {@link Bundle} is Spring-DM powered.
     * 
     * @param bundle the <code>Bundle</code>.
     * @return <code>true</code> if the <code>Bundle</code> is Spring-DM powered, otherwise <code>false</code>.
     */
    public static boolean isSpringOsgiPoweredBundle(Bundle bundle) {
        // look first for Spring-Context entry
        if (getSpringContextHeader(bundle.getHeaders()) != null) {
            return true;
        }

        // check the default locations if the manifest doesn't give any info
        Enumeration<?> defaultConfig = bundle.findEntries(CONTEXT_DIR, CONTEXT_FILES, false);
        return defaultConfig != null && defaultConfig.hasMoreElements();
    }

    private static String getSpringContextHeader(Dictionary<?, ?> headers) {
        Object header = null;
        if (headers != null) {
            header = headers.get(SPRING_CONTEXT_HEADER);
        }
        return header != null ? header.toString() : null;
    }
}
