/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.framework;

import org.osgi.framework.BundleEvent;
import org.osgi.framework.BundleListener;

import com.springsource.server.serviceability.logging.LogService;
import com.springsource.server.serviceability.logging.LogServiceFactory;


/**
 * Helper class for shutting down the Server
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * This class is <strong>thread-safe</strong>
 * 
 * @author Andy Wilkinson
 * @since 1.0
 */
public class Shutdown {

    private static final LogService LOG = LogServiceFactory.getLogService();

    private static volatile OsgiFramework osgiFramework;

    private static final Object shutdownLock = new Object();

    private static boolean shuttingDown = false;

    private static final Thread shutdownHook = new Thread() {

        public void run() {
            synchronized (shutdownLock) {
                if (!shuttingDown) {
                    LOG.log(OsgiFrameworkLogEvent.SHUTDOWN_INITIATED);
                    if (osgiFramework != null) {
                        osgiFramework.stop();
                    }
                }
            }
        }
    };

    static {
        Runtime.getRuntime().addShutdownHook(shutdownHook);
    }

    public static void setOsgiFramework(OsgiFramework framework) {
        osgiFramework = framework;
    }

    /**
     * Shuts down the Server in a graceful manner thereby minimising processing required during a subsequent restart.
     */
    public static void shutdown() {
        synchronized (shutdownLock) {
            if (!shuttingDown) {
                LOG.log(OsgiFrameworkLogEvent.SHUTDOWN_INITIATED);
                shuttingDown = true;
                Runtime.getRuntime().removeShutdownHook(shutdownHook);
                if (osgiFramework != null) {
                    osgiFramework.getBundleContext().addBundleListener(new BundleListener() {

                        public void bundleChanged(BundleEvent event) {
                            if (event.getBundle().getBundleId() == 0 && event.getType() == BundleEvent.STOPPED) {
                                System.exit(0);
                            }
                        }
                    });
                    osgiFramework.stop();
                } else {
                    System.exit(0);
                }
            }
        }
    }

    /**
     * Shuts down the Server immediately.
     */
    public static void immediateShutdown() {
        LOG.log(OsgiFrameworkLogEvent.IMMEDIATE_SHUTDOWN_INITIATED);
        Runtime.getRuntime().removeShutdownHook(shutdownHook);
        System.exit(0);
    }
}
