/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.framework.equinox;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

import org.eclipse.core.runtime.adaptor.LocationManager;
import org.eclipse.osgi.baseadaptor.BaseAdaptor;
import org.eclipse.osgi.baseadaptor.HookRegistry;
import org.eclipse.osgi.framework.adaptor.FrameworkAdaptor;
import org.eclipse.osgi.framework.internal.core.AbstractBundle;
import org.eclipse.osgi.framework.internal.core.BundleContextImpl;
import org.eclipse.osgi.framework.internal.core.Constants;
import org.eclipse.osgi.framework.internal.core.Framework;
import org.eclipse.osgi.framework.internal.core.FrameworkConsole;
import org.eclipse.osgi.framework.internal.core.FrameworkProperties;
import org.eclipse.osgi.framework.internal.core.OSGi;
import org.eclipse.osgi.framework.internal.core.PackageAdminImpl;
import org.eclipse.osgi.service.resolver.PlatformAdmin;
import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.BundleException;
import org.osgi.framework.InvalidSyntaxException;
import org.osgi.framework.ServiceReference;
import org.osgi.service.packageadmin.ExportedPackage;
import org.osgi.service.packageadmin.RequiredBundle;

import com.springsource.server.bootstrap.util.NetUtils;
import com.springsource.server.osgi.framework.BundleInstaller;
import com.springsource.server.osgi.framework.OsgiFrameworkLogEvent;
import com.springsource.server.osgi.framework.OsgiFrameworkUtils;
import com.springsource.server.osgi.framework.Shutdown;
import com.springsource.server.osgi.framework.support.AbstractOsgiFramework;
import com.springsource.server.osgi.framework.support.FilePathSanitiser;
import com.springsource.server.serviceability.logging.LogService;
import com.springsource.server.serviceability.logging.LogServiceFactory;

/**
 * Implementation of <code>OsgiFramework</code> using Equinox.
 * 
 * @author Rob Harrop
 * @author Sam Brannen
 * @author Andy Wilkinson
 * @since 1.0
 */
final public class EquinoxOsgiFramework extends AbstractOsgiFramework {

    /**
     * Equinox-specific property to control clean startup.
     */
    private static final String PROP_OSGI_CLEAN = "osgi.clean";

    /**
     * Equinox-specific property for setting the parent <code>ClassLoader</code>.
     */
    private static final String PROP_OSGI_PARENT_CLASSLOADER = "osgi.parentClassloader";

    private final EquinoxOsgiConfiguration configuration;

    private volatile BundleContext bundleContext;

    private volatile ServerOSGi osgi;

    private volatile PlatformAdmin serverAdmin;

    private volatile ServiceReference serverAdminReference;

    private volatile FrameworkConsole console;

    private volatile Thread consoleThread;

    private volatile EquinoxBootDelegationHelper bootDelegationHelper;
    
    private static final LogService LOGGER = LogServiceFactory.getLogService();

    /**
     * Creates a new <code>EquinoxOsgiFramework</code> with the supplied configuration.
     * 
     * @param configuration the configuration to use.
     */
    public EquinoxOsgiFramework(EquinoxOsgiConfiguration configuration) {
        super(configuration);
        this.configuration = configuration;
        setFrameworkProperty(PROP_OSGI_PARENT_CLASSLOADER, "fwk");
        setFrameworkProperty("osgi.context.bootdelegation", "false");

        setFrameworkProperty(HookRegistry.PROP_HOOK_CONFIGURATORS_INCLUDE, ServerHookConfigurator.class.getName());
        setFrameworkProperty(Constants.OSGI_COMPATIBILITY_BOOTDELEGATION, "false");
        setFrameworkProperty(Constants.OSGI_JAVA_PROFILE, configuration.getJavaProfilePath());
        setFrameworkProperty(Constants.OSGI_JAVA_PROFILE_BOOTDELEGATION, Constants.OSGI_BOOTDELEGATION_OVERRIDE);

        try {
            String frameworkLocation = FilePathSanitiser.sanitise(new File(OSGi.class.getProtectionDomain().getCodeSource().getLocation().toURI()).getCanonicalPath());
            setFrameworkProperty("osgi.framework", frameworkLocation);
        } catch (Exception e) {
            throw new IllegalStateException("Could not set the osgi framework location", e);
        }
    }

    /**
     * {@inheritDoc}
     */
    public BundleContext getBundleContext() {
        return this.bundleContext;
    }

    /**
     * {@inheritDoc}
     */
    public void refresh(Bundle bundle) throws BundleException {
        ClassLoader cl = getBundleClassLoader(bundle);
        List<Bundle> refreshBundles = new ArrayList<Bundle>();
        if (cl instanceof ServerBundleClassLoader) {
            ServerBundleClassLoader pbcl = (ServerBundleClassLoader) cl;
            if (pbcl.isInstrumented()) {
                Bundle[] dependencies = getDirectDependencies(bundle);
                for (Bundle dependency : dependencies) {
                    if (OsgiFrameworkUtils.sameScope(bundle, dependency)) {
                        dependency.update();
                        refreshBundles.add(dependency);
                    }
                }
            }
        }
        bundle.update();
        refreshBundles.add(bundle);
        Bundle[] toRefresh = refreshBundles.toArray(new Bundle[refreshBundles.size()]);
        getPackageAdmin().refreshPackages(toRefresh);
    }

    /**
     * {@inheritDoc}
     */
    @Override protected void doStart() {
        setFrameworkProperty(LocationManager.PROP_CONFIG_AREA, this.configuration.getConfigAreaPath());
        setFrameworkProperty(LocationManager.PROP_INSTALL_AREA, this.configuration.getInstallAreaPath());
        setFrameworkProperty(PROP_OSGI_CLEAN, Boolean.toString(this.configuration.isClean()));

        this.osgi = new ServerOSGi(new BaseAdaptor(null));

        this.bootDelegationHelper = new EquinoxBootDelegationHelper(this.osgi.getFramework().getProperty(Constants.OSGI_BOOTDELEGATION));

        this.osgi.launch();
        this.bundleContext = new BundleContextWrapper(this.osgi.getBundleContext());

        this.serverAdminReference = getBundleContext().getServiceReference(PlatformAdmin.class.getName());
        this.serverAdmin = (PlatformAdmin) getBundleContext().getService(this.serverAdminReference);
        
        Shutdown.setOsgiFramework(this);

        startConsoleIfRequired();
    }

    /**
     * {@inheritDoc}
     */
    @Override protected BundleInstaller createBundleInstaller() {
        return new EquinoxBundleInstaller(getBundleRepository(), getPackageAdmin(), getBundleContext(), this.serverAdmin);
    }

    /**
     * Starts the OSGi console if needed.
     */
    private void startConsoleIfRequired() {
        if (this.configuration.isRunConsole()) {
            int consolePort = this.configuration.getConsolePort();
            if (this.tracer.isInfoEnabled()) {
                if (consolePort > 0) {
                    this.tracer.traceInfo("Starting console on port: " + consolePort);
                } else {
                    this.tracer.traceInfo("Starting console");
                }
            }
            
            if (consolePort > 0) {
                if (NetUtils.isPortAvailable(consolePort)) {
                    this.console = new FrameworkConsole(this.osgi, consolePort, null);
                    LOGGER.log(OsgiFrameworkLogEvent.OSGI_CONSOLE_PORT, consolePort);
                } else {
                    LOGGER.log(OsgiFrameworkLogEvent.OSGI_CONSOLE_PORT_IN_USE, consolePort);
                    Shutdown.immediateShutdown();
                }
            } else {
                this.console = new FrameworkConsole(this.osgi, null);
            }
            
            this.consoleThread = new Thread(this.console, "osgi-console");
            this.consoleThread.start();
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override protected void doStop() {
        if (this.serverAdminReference != null) {
            getBundleContext().ungetService(this.serverAdminReference);
        }

        if (this.console != null) {
            this.console.shutdown();
        }
        this.osgi.shutdown();
    }

    /**
     * {@inheritDoc}
     */
    public String getFrameworkProperty(String name) {
        return FrameworkProperties.getProperty(name);
    }

    /**
     * {@inheritDoc}
     */
    public void setFrameworkProperty(String name, String value) {
        FrameworkProperties.setProperty(name, value);
    }

    /**
     * {@inheritDoc}
     */
    public ClassLoader getBundleClassLoader(Bundle bundle) {
        return EquinoxUtils.getBundleClassLoader(bundle, this.serverAdmin);
    }

    /**
     * {@inheritDoc}
     */
    public boolean isBootDelegated(String className) {
        if (this.bootDelegationHelper != null) {
            return this.bootDelegationHelper.isBootDelegated(className);
        }

        throw new IllegalStateException("OsgiFramework must have been started prior to querying its boot delegation");
    }

    /**
     * {@inheritDoc}
     */
    public Bundle[] getDirectDependencies(Bundle bundle, boolean includeFragments) {
        BundleContext bundleContext = getBundleContext();
        ServiceReference serviceRef = bundleContext.getServiceReference(PlatformAdmin.class.getName());
        try {
            PlatformAdmin serverAdmin = (PlatformAdmin) bundleContext.getService(serviceRef);
            return EquinoxUtils.getDirectDependencies(bundle, bundleContext, serverAdmin, includeFragments);
        } finally {
            bundleContext.ungetService(serviceRef);
        }
    }
    
    /**
     * {@inheritDoc}
     */
    public Bundle[] getDirectDependencies(Bundle bundle) {
        return getDirectDependencies(bundle, false);
    }

    private class ServerOSGi extends OSGi {

        private static final String SCOPE_KEY = "com.springsource.server.app.name";

        public ServerOSGi(FrameworkAdaptor adaptor) {
            super(adaptor);
        }

        public Framework getFramework() {
            return this.framework;
        }

        /**
         * {@inheritDoc}
         */
        @Override protected Framework createFramework(FrameworkAdaptor adaptor) {
            return new Framework(adaptor) {

                /**
                 * {@inheritDoc}
                 */
                @Override protected void initialize(FrameworkAdaptor adaptor) {
                    super.initialize(adaptor);
                    this.packageAdmin = new LockEnforcingPackageAdmin(this);
                }

                /**
                 * {@inheritDoc}
                 */
                @Override protected ServiceReference[] getServiceReferences(String clazz, String filterstring, BundleContextImpl context,
                    boolean allservices) throws InvalidSyntaxException {
                    ServiceReference[] refs = super.getServiceReferences(clazz, filterstring, context, allservices);
                    if (this.bundles != null && refs != null) {

                        String consumerScope = OsgiFrameworkUtils.getScope(context.getBundle());
                        List<ServiceReference> filteredReferences = new ArrayList<ServiceReference>(refs.length);
                        for (ServiceReference serviceReference : refs) {
                            String refScope = (String) serviceReference.getProperty(SCOPE_KEY);

                            if (refScope == null || refScope.equals(consumerScope)) {
                                filteredReferences.add(serviceReference);
                            }
                        }
                        return filteredReferences.toArray(new ServiceReference[filteredReferences.size()]);
                    } else {
                        return refs;
                    }

                }

            };
        }
    }

    /**
     * Extension to {@link PackageAdminImpl} that takes the {@link LockManager#getMainLock() main lock} before all
     * operations. <p/>
     * 
     * <strong>Concurrent Semantics</strong><br />
     * 
     * Threadsafe.
     * 
     * @author Rob Harrop
     * @since 1.0
     */
    private static class LockEnforcingPackageAdmin extends PackageAdminImpl {

        /**
         * @param framework
         */
        protected LockEnforcingPackageAdmin(Framework framework) {
            super(framework);
        }

        /**
         * {@inheritDoc}
         * @see org.osgi.service.packageadmin.PackageAdmin#getBundle(java.lang.Class)
         */
        @Override @SuppressWarnings("unchecked") public Bundle getBundle(Class clazz) {
            synchronized (LockManager.getMainLock()) {
                return super.getBundle(clazz);
            }
        }

        /**
         * {@inheritDoc}
         * @see org.osgi.service.packageadmin.PackageAdmin#getBundles(java.lang.String, java.lang.String)
         */
        @Override public Bundle[] getBundles(String symbolicName, String versionRange) {
            synchronized (LockManager.getMainLock()) {
                return super.getBundles(symbolicName, versionRange);
            }
        }

        /**
         * {@inheritDoc}
         * @see org.osgi.service.packageadmin.PackageAdmin#getBundleType(org.osgi.framework.Bundle)
         */
        @Override public int getBundleType(Bundle bundle) {
            synchronized (LockManager.getMainLock()) {
                return super.getBundleType(bundle);
            }
        }

        /**
         * {@inheritDoc}
         * @see org.osgi.service.packageadmin.PackageAdmin#getExportedPackage(java.lang.String)
         */
        @Override public ExportedPackage getExportedPackage(String name) {
            synchronized (LockManager.getMainLock()) {
                return super.getExportedPackage(name);
            }
        }

        /**
         * {@inheritDoc}
         * @see org.osgi.service.packageadmin.PackageAdmin#getExportedPackages(org.osgi.framework.Bundle)
         */
        @Override public ExportedPackage[] getExportedPackages(Bundle bundle) {
            synchronized (LockManager.getMainLock()) {
                return super.getExportedPackages(bundle);
            }
        }

        /**
         * {@inheritDoc}
         * @see org.osgi.service.packageadmin.PackageAdmin#getExportedPackages(java.lang.String)
         */
        @Override public ExportedPackage[] getExportedPackages(String name) {
            synchronized (LockManager.getMainLock()) {
                return super.getExportedPackages(name);
            }
        }

        /**
         * {@inheritDoc}
         * @see org.osgi.service.packageadmin.PackageAdmin#getFragments(org.osgi.framework.Bundle)
         */
        @Override public Bundle[] getFragments(Bundle bundle) {
            synchronized (LockManager.getMainLock()) {
                return super.getFragments(bundle);
            }
        }

        /**
         * {@inheritDoc}
         * @see org.osgi.service.packageadmin.PackageAdmin#getHosts(org.osgi.framework.Bundle)
         */
        @Override public Bundle[] getHosts(Bundle bundle) {
            synchronized (LockManager.getMainLock()) {
                return super.getHosts(bundle);
            }
        }

        /**
         * {@inheritDoc}
         * @see org.osgi.service.packageadmin.PackageAdmin#getRequiredBundles(java.lang.String)
         */
        @Override public RequiredBundle[] getRequiredBundles(String symbolicName) {
            synchronized (LockManager.getMainLock()) {
                return super.getRequiredBundles(symbolicName);
            }
        }

        /**
         * {@inheritDoc}
         * @see org.osgi.service.packageadmin.PackageAdmin#refreshPackages(org.osgi.framework.Bundle[])
         */
        @Override public void refreshPackages(Bundle[] bundles) {
            synchronized (LockManager.getMainLock()) {
                super.refreshPackages(bundles);
            }
        }

        /**
         * {@inheritDoc}
         * @see org.osgi.service.packageadmin.PackageAdmin#resolveBundles(org.osgi.framework.Bundle[])
         */
        @Override public boolean resolveBundles(Bundle[] bundles) {
            synchronized (LockManager.getMainLock()) {
                return super.resolveBundles(bundles);
            }
        }

        /** 
         * {@inheritDoc}
         */
        @Override
        protected synchronized void doResolveBundles(AbstractBundle[] bundles, boolean refreshPackages) {
            synchronized (LockManager.getMainLock()) {
                super.doResolveBundles(bundles, refreshPackages);
            }
        }
        
    }
}
