/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.framework.equinox;

import java.lang.reflect.Method;
import java.util.HashSet;
import java.util.Set;

import org.eclipse.osgi.framework.internal.core.BundleHost;
import org.eclipse.osgi.framework.internal.core.BundleLoader;
import org.eclipse.osgi.internal.baseadaptor.DefaultClassLoader;
import org.eclipse.osgi.service.resolver.BundleDescription;
import org.eclipse.osgi.service.resolver.ExportPackageDescription;
import org.eclipse.osgi.service.resolver.PlatformAdmin;
import org.eclipse.osgi.service.resolver.State;
import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;

import com.springsource.server.osgi.framework.OsgiFrameworkException;
import com.springsource.server.osgi.manifest.VersionRange;

/**
 * Utility methods for working with Equinox internals.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
final class EquinoxUtils {

    /**
     * Converts an Equinox {@link org.eclipse.osgi.service.resolver.VersionRange} to the corresponding internal
     * {@link VersionRange}.
     * 
     * @param vr the Equinox <code>VersionRange</code>.
     * @return the internal <code>VersionRange</code>.
     */
    public static VersionRange convertEquinoxVersionRange(org.eclipse.osgi.service.resolver.VersionRange vr) {
        return new VersionRange(vr.toString());
    }

    /**
     * Gets the {@link ClassLoader} for the supplied {@link Bundle}.
     * 
     * @param bundle the bundle.
     * @return the bundle <code>ClassLoader</code>.
     */
    public static ClassLoader getBundleClassLoader(Bundle bundle, PlatformAdmin serverAdmin) {
        ClassLoader classLoader = null;
        if (BundleHost.class.isAssignableFrom(bundle.getClass())) {
            BundleHost bundleHost = (BundleHost) bundle;

            Class<?>[] parmTypes = {};
            Method checkLoaderMethod;
            try {
                checkLoaderMethod = BundleHost.class.getDeclaredMethod("checkLoader", parmTypes);
                Object[] args = {};
                checkLoaderMethod.setAccessible(true);
                BundleLoader bundleLoader = (BundleLoader) checkLoaderMethod.invoke(bundleHost, args);

                if (bundleLoader == null) {
                    throw new IllegalStateException("Unable to access BundleLoader for bundle '" + bundle.getSymbolicName() + "'.");
                }

                Method createClassLoaderMethod = BundleLoader.class.getDeclaredMethod("createClassLoader", parmTypes);
                createClassLoaderMethod.setAccessible(true);

                classLoader = (DefaultClassLoader) createClassLoaderMethod.invoke(bundleLoader, args);
            } catch (Exception e) {
                String failureDescription = new ResolutionFailureDetective(serverAdmin).generateFailureDescription(bundle);
                throw new OsgiFrameworkException("Failed to get class loader for bundle '" + bundle + "' - possible resolution problem:\n"
                    + failureDescription, e);
            }
        }
        return classLoader;
    }

    /**
     * Gets all direct dependencies of the supplied {@link Bundle}.
     * 
     * @param bundle the <code>Bundle</code>.
     * @param bundleContext the {@link BundleContext} to use for service access - typically the system
     *        <code>BundleContext</code>.
     * @param serverAdmin the {@link PlatformAdmin} service.
     * @return the direct dependencies.
     */
    public static Bundle[] getDirectDependencies(Bundle bundle, BundleContext bundleContext, PlatformAdmin serverAdmin) {
        return getDirectDependencies(bundle, bundleContext, serverAdmin, false);
    }
    
    /**
     * Gets all direct dependencies of the supplied {@link Bundle}, optionally included fragments of the direct dependencies in the returned array.
     * 
     * @param bundle the <code>Bundle</code>.
     * @param bundleContext the {@link BundleContext} to use for service access - typically the system
     *        <code>BundleContext</code>.
     * @param serverAdmin the {@link PlatformAdmin} service.
     * @return the direct dependencies.
     */
    public static Bundle[] getDirectDependencies(Bundle bundle, BundleContext bundleContext, PlatformAdmin serverAdmin, boolean includeFragments) {
        State state = serverAdmin.getState(true);

        synchronized (LockManager.getMainLock()) {
            if (!state.isResolved()) {
                state.setResolver(serverAdmin.getResolver());
                state.resolve();
            }
        }

        ExportPackageDescription[] exportPackageDescriptions = serverAdmin.getStateHelper().getVisiblePackages(
            state.getBundle(bundle.getBundleId()));

        Set<Bundle> dependencies = new HashSet<Bundle>();

        for (ExportPackageDescription exportPackageDescription : exportPackageDescriptions) {
            BundleDescription bundleDescription = exportPackageDescription.getExporter();
            if (bundleDescription.getBundleId() != bundle.getBundleId()) {
                Bundle dependencyBundle = bundleContext.getBundle(bundleDescription.getBundleId());
                // Handle an uninstalled dependent bundle gracefully.
                if (dependencyBundle != null) {
                    dependencies.add(dependencyBundle);
                    if (includeFragments) {
                        BundleDescription[] fragmentDescriptions = bundleDescription.getFragments();
                        for (BundleDescription fragmentDescription : fragmentDescriptions) {
                            Bundle fragment = bundleContext.getBundle(fragmentDescription.getBundleId());
                            if (fragment != null) {
                                dependencies.add(fragment);
                            }
                        }
                    }
                }
            }
        }

        return dependencies.toArray(new Bundle[dependencies.size()]);
    }

    /**
     * Queries whether the supplied {@link Bundle} is the system bundle.
     * 
     * @param bundle the <code>Bundle</code>.
     * @return <code>true</code> if <code>bundle</code> is the system bundle, otherwise <code>false</code>.
     */
    public static boolean isSystemBundle(Bundle bundle) {
        return bundle != null && bundle.getBundleId() == 0;
    }
}
