/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.framework.equinox;

import java.io.File;
import java.io.IOException;
import java.net.JarURLConnection;
import java.net.URL;
import java.net.URLConnection;
import java.util.Enumeration;

import org.eclipse.osgi.baseadaptor.BaseData;
import org.eclipse.osgi.baseadaptor.bundlefile.BundleEntry;
import org.eclipse.osgi.baseadaptor.bundlefile.BundleFile;
import org.eclipse.osgi.baseadaptor.hooks.BundleFileWrapperFactoryHook;

/**
 * A {@link BundleFileWrapperFactoryHook} implementation that wraps {@link BundleFile BundleFiles} to ensure that all
 * returned resource {@link URL URLs} have a <code>file:</code> protocol, not a <code>bundleresource:</code>
 * protocol as is the Equinox default.
 * 
 * <strong>Concurrent Semantics</strong><br />
 * This class if <strong>thread-safe</strong>.
 * 
 * @author Andy Wilkinson
 * @since 1.0
 */
public class ServerBundleFileWrapperFactoryHook implements BundleFileWrapperFactoryHook {

    /**
     * {@inheritDoc}
     */
    public BundleFile wrapBundleFile(BundleFile bundleFile, Object content, BaseData data, boolean base) throws IOException {
        return new FileResourceEnforcingBundleFile(bundleFile);
    }

    /**
     * A concrete extension of {@link BundleFile} that ensures that all resource {@link URL URLs} returned have a
     * <code>file:</code> protocol, not a <code>bundleresource:</code> protocol as is the Equinox default.
     * <p>
     * <strong>Concurrent Semantics</strong><br />
     * As thread-safe as the encapsulated <code>BundleFile</code> instance.
     * 
     * @author Andy Wilkinson
     * @since 1.0
     */
    private class FileResourceEnforcingBundleFile extends BundleFile {

        private final BundleFile bundleFile;

        private FileResourceEnforcingBundleFile(BundleFile bundleFile) {
            this.bundleFile = bundleFile;
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public void close() throws IOException {
            this.bundleFile.close();

        }

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean containsDir(String dir) {
            return this.bundleFile.containsDir(dir);
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public BundleEntry getEntry(String path) {
            BundleEntry entry = this.bundleFile.getEntry(path);
            return entry;
        }

        /**
         * {@inheritDoc}
         */
        @Override
        @SuppressWarnings("unchecked")
        public Enumeration getEntryPaths(String path) {
            Enumeration paths = this.bundleFile.getEntryPaths(path);
            return paths;
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public File getFile(String path, boolean nativeCode) {
            return this.bundleFile.getFile(path, nativeCode);
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public void open() throws IOException {
            this.bundleFile.open();
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public File getBaseFile() {
            return this.bundleFile.getBaseFile();
        }

        /**
         * Locates the resource in the BundleFile identified by the supplied path and, if found, returns a
         * <code>file</code> protocol URL for the resource.
         * 
         * @return a <code>file</code> protocol URL for the resource
         */
        @Override
        public URL getResourceURL(String path, long hostBundleID, int index) {
            return doGetResourceURL(path);
        }

        /**
         * Locates the resource in the BundleFile identified by the supplied path and, if found, returns a
         * <code>file</code> protocol URL for the resource.
         * 
         * @return a <code>file</code> protocol URL for the resource
         */
        @Override
        public URL getResourceURL(String path, long hostBundleID) {
            return doGetResourceURL(path);
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public URL getResourceURL(String path, BaseData hostData, int index) {
            return doGetResourceURL(path);
        }

        private URL doGetResourceURL(String path) {
            BundleEntry entry = getEntry(path);
            if (entry != null) {
                return getLocalURLForEntry(entry);
            } else {
                return null;
            }
        }

        private URL getLocalURLForEntry(BundleEntry entry) {
            URL url = entry.getLocalURL();
            try {
                url.openConnection().setDefaultUseCaches(false);
                url.openConnection().setUseCaches(false);
            } catch (Exception e) {
                
            }
            if (!"jar".equals(url.getProtocol()) || doesJarEntryReallyExist(url)) {
                return url;
            } else {
                return null;
            }
        }

        private boolean doesJarEntryReallyExist(URL url) {
            boolean entryExists = false;
            try {
                URLConnection connection = url.openConnection();
                if (connection instanceof JarURLConnection) {
                    if (((JarURLConnection) connection).getJarEntry() != null) {
                        entryExists = true;
                    }
                }
            } catch (IOException ioe) {
                entryExists = false;
            }
            return entryExists;
        }
    }
}
