/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.framework.support.merge;

import java.util.List;

import com.springsource.server.osgi.framework.ImportMergeException;
import com.springsource.server.osgi.manifest.BundleManifest.ImportPackageHeader.PackageImport;

/**
 * {@link TrackedPackageImports} is a collection of package imports which supports merging. Merge conflicts result in an
 * {@link ImportMergeException} being thrown which can be tracked back to its sources for diagnosis of the reason for the
 * clash.
 * <p />
 * Various implementations of <code>TrackedPackageImports</code> are available and instance of these should be created
 * using the {@link TrackedPackageImportsFactory} interface.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Implementations of this class must be thread safe.
 * 
 * @author Glyn Normington
 * @since 1.0
 * @see TrackedPackageImportsFactory
 */
public interface TrackedPackageImports {

    /**
     * Merge the given package imports into this collection of package imports. If there is a conflict, issue
     * diagnostics and throw {@link ImportMergeException}.
     * <p />
     * A merge conflict is caused by disjoint version ranges or incompatible matching attributes.
     * 
     * @param importsToMerge the package imports to be merged
     * @throws ImportMergeException thrown if and only if there is a conflict
     */
    void merge(TrackedPackageImports importsToMerge) throws ImportMergeException;

    /**
     * Determine whether there are any merged imports.
     * 
     * @return <code>true</code> if and only if there are merged imports
     */
    boolean isEmpty();

    /**
     * Determine whether or not this {@link TrackedPackageImports} has equivalent merged imports to the given
     * {@link TrackedPackageImports}. The sources of the merged imports are disregarded in this comparison.
     * 
     * @param otherTrackedPackageImports
     * @return <code>true</code> if and only if this and the given {@link TrackedPackageImports} have equivalent
     *         merged imports
     */
    boolean isEquivalent(TrackedPackageImports otherTrackedPackageImports);

    /**
     * Get the merged package imports. If there has been a conflict, throw {@link ImportMergeException}.
     * 
     * @return the merged package imports
     * @throws ImportMergeException thrown if and only if there has been a conflict
     */
    List<PackageImport> getMergedImports() throws ImportMergeException;

    /**
     * Get any sources of the given package in this merged collection. Return <code>null</code> if there are no
     * sources in this collection.
     * 
     * @param pkg the name of the package whose sources are required
     * @return a description of the sources in this collection or <code>null</code> if there are no sources in this
     *         collection
     */
    String getSources(String pkg);

    /**
     * If the given package name was in the original collection, return the source of this {@link TrackedPackageImports}.
     * Otherwise return <code>null</code>.
     * 
     * @param pkg the name of the package whose source is required
     * @return a description of this source or <code>null</code> if the given package was not in the original
     *         collection
     */
    String getSource(String pkg);

}
