/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.manifest;

import java.util.jar.Attributes;

import com.springsource.server.osgi.manifest.BundleManifest.BundleActivationPolicy;

/**
 * A {@link Delegate} implementation for handling a manifest's <code>Bundle-ActivationPolicy</code> header.
 * <p />
 * <strong>Concurrent Semantics</strong><br />
 * This class is <strong>not</strong> thread-safe.
 * 
 * @author Andy Wilkinson
 * @since 1.0
 */
class BundleActivationPolicyDelegate implements Delegate {

    private static final String BUNDLE_ACTIVATION_POLICY_HEADER_NAME = "Bundle-ActivationPolicy";

    private static final BundleActivationPolicy DEFAULT_ACTIVATION_POLICY = null;

    private BundleActivationPolicy bundleActivationPolicy = DEFAULT_ACTIVATION_POLICY;

    BundleActivationPolicyDelegate(Attributes mainAttributes, BundleManifestFactory bundleManifestFactory) {
        String creationValue = mainAttributes.getValue(BUNDLE_ACTIVATION_POLICY_HEADER_NAME);
        if (creationValue != null) {
            this.bundleActivationPolicy = bundleManifestFactory.parseBundleActivationPolicy(creationValue);
        } else {
            this.bundleActivationPolicy = null;
        }
    }

    BundleActivationPolicy getBundleActivationPolicy() {
        return this.bundleActivationPolicy;
    }

    void setBundleActivationPolicy(BundleActivationPolicy bundleActivationPolicy) {
        this.bundleActivationPolicy = bundleActivationPolicy;
    }

    /** 
     * {@inheritDoc}
     */
    public void checkHeaderValidity() {  
    }
    
    @Override public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final BundleActivationPolicyDelegate other = (BundleActivationPolicyDelegate) obj;
        if (this.bundleActivationPolicy == null) {
            if (other.bundleActivationPolicy != null) {
                return false;
            }
        } else if (!this.bundleActivationPolicy.equals(other.bundleActivationPolicy)) {
            return false;
        }
        return true;
    }
    
    void write(Attributes outputAttributes) {
        if (this.bundleActivationPolicy != null) {
            checkHeaderValidity();
            String value = this.bundleActivationPolicy.toString();
            if (value != null) {
                outputAttributes.putValue(BUNDLE_ACTIVATION_POLICY_HEADER_NAME, value);
            } else {
                outputAttributes.remove(new Attributes.Name(BUNDLE_ACTIVATION_POLICY_HEADER_NAME));
            }
        }
    }
}
