/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.manifest;

import java.util.ArrayList;
import java.util.List;

import org.osgi.framework.Version;
import org.springframework.util.StringUtils;

import com.springsource.server.osgi.manifest.BundleManifest.ExportPackageHeader;
import com.springsource.server.osgi.manifest.BundleManifest.ExportPackageHeader.PackageExport;
import com.springsource.server.osgi.manifest.BundleManifest.ImportPackageHeader.PackageImport;

/**
 * Provides a number of helper methods for processing {@link BundleManifest BundleManifests}.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * Thread-safe
 * 
 * @author Andy Wilkinson
 * @since 1.0
 */
public class BundleManifestProcessor {

    private final BundleManifestFactory factory;

    private static final String VERSION_ATTRIBUTE_NAME = "version";

    private static final String BUNDLE_SYMBOLIC_NAME_ATTRIBUTE_NAME = "bundle-symbolic-name";

    private static final String BUNDLE_VERSION_ATTRIBUTE_NAME = "bundle-version";

    public BundleManifestProcessor(BundleManifestFactory factory) {
        this.factory = factory;
    }

    /**
     * Creates a {@link PackageImport} for each of the supplied {@link PackageExport PackageExports} and returns a
     * {@link Result} containing them. The <code>PackageImports</code> will be versioned with a range that matches the
     * corresponding <code>PackageExport</code>'s version and above. I.e an export with version x
     * will result in an import with ;version="x". The result also contains any warnings that were generated during
     * the processing of the manifests.
     * 
     * @param packageExports the <code>PackageExports</code> to process
     * @return a Result containing the package imports and any warnings
     */
    public PackageImport[] createPackageImportForEachPackageExport(List<PackageExport> packageExports, String bundleSymbolicName,
        String bundleVersion) {
        List<PackageImport> packageImports = new ArrayList<PackageImport>();

        packageImports = doCreatePackageImportForEachPackageExport(packageExports, bundleSymbolicName, bundleVersion);

        return packageImports.toArray(new PackageImport[packageImports.size()]);
    }

    /**
     * Creates a {@link PackageImport} for each package export found in the given {@link BundleManifest BundleManifests}
     * and returns a {@link Result} containing the package imports. The package imports will be versioned with a range
     * that matches the corresponding package export's version and above. I.e an export with version x
     * will result in an import with ;version="x". The result also contains any warnings that were generated
     * during the processing of the manifests.
     * 
     * @param bundleManifests the bundle manifests to process
     * @return a Result containing the package imports and any warnings
     */
    public Result<PackageImport[]> createPackageImportForEachPackageExport(BundleManifest[] bundleManifests) {

        List<PackageImport> allPackageImports = new ArrayList<PackageImport>();
        List<Warning> warnings = new ArrayList<Warning>();

        for (BundleManifest bundleManifest : bundleManifests) {
            ExportPackageHeader exportPackageHeader = bundleManifest.getExportPackageHeader();
            if (exportPackageHeader != null) {
                List<PackageExport> packageExports = exportPackageHeader.getPackageExports();
                List<PackageImport> packageImports = doCreatePackageImportForEachPackageExport(packageExports, null, null);
                
                for (PackageImport packageImport : packageImports) {
                    String name = packageImport.getPackageNames().get(0);
                    if (!containsImport(allPackageImports, name)) {
                        allPackageImports.add(packageImport);
                    } else {
                        warnings.add(new Warning(Code.DUPLICATE_PACKAGE_EXPORTS, name));
                    }
                }
            }
        }

        return new Result<PackageImport[]>(allPackageImports.toArray(new PackageImport[allPackageImports.size()]),
            warnings.toArray(new Warning[warnings.size()]));
    }

    private List<PackageImport> doCreatePackageImportForEachPackageExport(List<PackageExport> packageExports, String bundleSymbolicName, String version) {
        List<PackageImport> packageImports = new ArrayList<PackageImport>();
        for (PackageExport packageExport : packageExports) {
            for (String name : packageExport.getPackageNames()) {
                PackageImport existingImport = findImport(packageImports, name);
                Version exportVersion = packageExport.getVersion();
                if (existingImport != null) {                    
                    VersionRange existingRange = new VersionRange(existingImport.getAttributes().get(VERSION_ATTRIBUTE_NAME));
                    Version floor = exportVersion.compareTo(existingRange.getFloor()) < 0 ? exportVersion : existingRange.getFloor();
                    existingImport.getAttributes().put(VERSION_ATTRIBUTE_NAME, floor.toString());
                } else {
                    PackageImport packageImport = this.factory.createPackageImport();
                    packageImport.getPackageNames().add(name);
                    packageImport.getAttributes().put(VERSION_ATTRIBUTE_NAME, exportVersion.toString());
                    List<String> mandatory = packageExport.getMandatory();
                    if (mandatory != null) {
                        for (String attrName : mandatory) {
                            if (packageExport.getAttributes().containsKey(attrName)) {
                                packageImport.getAttributes().put(attrName, packageExport.getAttributes().get(attrName));
                            }
                        }
                    }
                    if (bundleSymbolicName != null) {
                        packageImport.getAttributes().put(BUNDLE_SYMBOLIC_NAME_ATTRIBUTE_NAME, bundleSymbolicName);
                    }
                    if (StringUtils.hasText(version)) {
                        packageImport.getAttributes().put(BUNDLE_VERSION_ATTRIBUTE_NAME, version);
                    }
                    packageImports.add(packageImport);
                }
            }
        }
        return packageImports;
    }

    private static boolean containsImport(List<PackageImport> imports, String packageName) {
        for (PackageImport packageImport : imports) {
            if (packageImport.getPackageNames().contains(packageName)) {
                return true;
            }
        }
        return false;
    }
    
    private static PackageImport findImport(List<PackageImport> imports, String packageName) {
        for (PackageImport packageImport : imports) {
            if (packageImport.getPackageNames().contains(packageName)) {
                return packageImport;
            }
        }
        return null;
    }

    /**
     * Encapsulates the output from bundle manifest processing processing along with any {@link Warning Warnings} that
     * were generated during that processing.
     * <p />
     * 
     * <strong>Concurrent Semantics</strong><br />
     * Thread-safe.
     * 
     * @author Andy Wilkinson
     * @since 1.0
     */
    public static class Result<T> {

        private final T output;

        private final Warning[] warnings;

        private Result(T output, Warning[] warnings) {
            this.output = output;
            this.warnings = warnings;
        }

        /**
         * Returns the output of this result
         * 
         * @return the output
         */
        public T getOutput() {
            return this.output;
        }

        /**
         * Returns any warnings generated while producing the result. If no warnings were generated an empty array will
         * be returned.
         * 
         * @return any generated warnings
         */
        public Warning[] getWarnings() {
            return this.warnings.clone();
        }

        /**
         * Returns whether this result was successful. A result is deemed to be successful if it does not contain any
         * warnings
         * 
         * @return <code>true</code> if the result is successful
         */
        public boolean success() {
            return this.warnings.length == 0;
        }
    }

    /**
     * A Warning describes a problem that was encountered during BundleManifestHelper processing
     * <p />
     * 
     * <strong>Concurrent Semantics</strong><br />
     * Thread-safe.
     * 
     * @author Andy Wilkinson
     * @since 1.0
     */
    public static class Warning {

        private final Code code;

        private final String reason;

        private Warning(Code code, String reason) {
            this.code = code;
            this.reason = reason;
        }

        /**
         * Returns the {@link Code} that identifies the nature of this warning
         * 
         * @return the warning's code
         */
        public Code getCode() {
            return this.code;
        }

        /**
         * Returns the reason for the generation of this warning
         * 
         * @return the warning's reason
         */
        public String getReason() {
            return this.reason;
        }

        /**
         * {@inheritDoc}
         */
        @Override public boolean equals(Object other) {
            if (other == null) {
                return false;
            }

            if (other == this) {
                return true;
            }

            if (!(other instanceof Warning)) {
                return false;
            }

            Warning that = (Warning) other;
            return this.getCode().equals(that.getCode()) && this.getReason().equals(that.getReason());
        }
    }

    /**
     * Defines a unique code for each of the warnings that may be generated as a result of bundle manifest processing.
     * <p />
     * <strong>Concurrent Semantics</strong><br />
     * Thread-safe.
     * 
     * @author Andy Wilkinson
     * @since 1.0
     */
    public enum Code {
        /**
         * More than one of the input {@link BundleManifest BundleManifests} exported the same package.
         */
        DUPLICATE_PACKAGE_EXPORTS;
    }
}
