/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.manifest;

import java.util.jar.Attributes;

import org.springframework.util.StringUtils;

import com.springsource.server.osgi.manifest.BundleManifest.BundleSymbolicName;
import com.springsource.server.osgi.manifest.parse.FatalParseException;

/**
 * A {@link Delegate} implementation for handling <code>Bundle-SymbolicName</code> declarations in an OSGi bundle
 * manifest.
 * 
 * <strong>Concurrent Semantics</strong><br />
 * This class is <strong>not</strong> thread-safe.
 * 
 * @author Andy Wilkinson
 * @since 1.0
 */
class BundleSymbolicNameDelegate implements Delegate {

    private static final String BUNDLE_SYMBOLIC_NAME_HEADER_NAME = "Bundle-SymbolicName";

    private final Attributes mainAttributes;

    private BundleSymbolicName bundleSymbolicName;

    BundleSymbolicNameDelegate(Attributes mainAttributes, BundleManifestFactory bundleManifestFactory) {
        this.mainAttributes = mainAttributes;
        String creationValue = this.mainAttributes.getValue(BUNDLE_SYMBOLIC_NAME_HEADER_NAME);
        if (creationValue != null) {
            this.bundleSymbolicName = bundleManifestFactory.parseBundleSymbolicName(creationValue);
        } else {
            this.bundleSymbolicName = null;
        }
    }

    BundleSymbolicName getBundleSymbolicName() {
        return this.bundleSymbolicName;
    }

    void setBundleSymbolicName(BundleSymbolicName bundleSymbolicName) {
        this.bundleSymbolicName = bundleSymbolicName;
    }

    /**
     * {@inheritDoc}
     */
    public void checkHeaderValidity() {
        if (this.bundleSymbolicName != null) {
            if (!StringUtils.hasText(this.bundleSymbolicName.getSymbolicName())) {
                throw new FatalParseException("The symbolic name must be specified");
            }
        }
    }

    @Override public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final BundleSymbolicNameDelegate other = (BundleSymbolicNameDelegate) obj;
        if (this.bundleSymbolicName == null) {
            if (other.bundleSymbolicName != null) {
                return false;
            }
        } else if (!this.bundleSymbolicName.equals(other.bundleSymbolicName)) {
            return false;
        }
        return true;
    }

    void write(Attributes outputAttributes) {
        if (this.bundleSymbolicName != null) {
            checkHeaderValidity();
            String value = this.bundleSymbolicName.toString();
            if (value != null) {
                outputAttributes.putValue(BUNDLE_SYMBOLIC_NAME_HEADER_NAME, value);
            } else {
                outputAttributes.remove(new Attributes.Name(BUNDLE_SYMBOLIC_NAME_HEADER_NAME));
            }
        }
    }
}
