/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.manifest;

import java.io.IOException;
import java.io.InputStream;
import java.io.Reader;
import java.util.jar.Manifest;

import com.springsource.server.osgi.manifest.parse.FatalParseException;

/**
 * Utility class to create a JRE {@link Manifest} from a {@link Reader} of the manifest contents.<p/>
 * 
 * @author Glyn Normington
 * @since 1.0
 * @see <a href="http://java.sun.com/j2se/1.5.0/docs/guide/jar/jar.html">JAR File Specification</a>
 */
public final class ManifestFactory {

    private final ManifestInputStream manifestInputStream;

    private final String source;

    public ManifestFactory(Reader reader) {
        this(reader, "unknown");
    }

    public ManifestFactory(Reader reader, String source) {
        this.manifestInputStream = new ManifestInputStream(reader);
        this.source = source;
    }

    public Manifest getManifest() {
        try {
            return new Manifest(this.manifestInputStream);
        } catch (IOException e) {
            throw new FatalParseException("Exception reading manifest. Source: '" + this.source + "'", e);
        }
    }

    private static class ManifestInputStream extends InputStream {

        private final Reader reader;

        private boolean returnedBlankLine = false;

        public ManifestInputStream(Reader reader) {
            this.reader = reader;
        }

        /*
         * Add a single blank line to the end of the manifest file to work around a bug in the JDK's manifest parser
         * which causes it to lose the last line in the manifest if it is not followed by a blank line.
         */
        @Override public int read() throws IOException {
            if (this.returnedBlankLine) {
                return this.reader.read();
            } else {
                int result = this.reader.read();

                if (result == -1) {
                    this.returnedBlankLine = true;
                    return '\n';
                } else {
                    return result;
                }
            }
        }
    }
}
