/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.manifest;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.osgi.framework.Version;
import org.springframework.util.StringUtils;

import com.springsource.server.osgi.manifest.BundleManifest.BundleSymbolicName;
import com.springsource.server.osgi.manifest.BundleManifest.DynamicImportPackageHeader;
import com.springsource.server.osgi.manifest.BundleManifest.ExportPackageHeader;
import com.springsource.server.osgi.manifest.BundleManifest.FragmentHostHeader;
import com.springsource.server.osgi.manifest.BundleManifest.ImportPackageHeader;
import com.springsource.server.osgi.manifest.BundleManifest.RequireBundleHeader;
import com.springsource.server.osgi.manifest.BundleManifest.DynamicImportPackageHeader.DynamicPackageImport;
import com.springsource.server.osgi.manifest.BundleManifest.ExportPackageHeader.PackageExport;
import com.springsource.server.osgi.manifest.BundleManifest.ImportPackageHeader.PackageImport;
import com.springsource.server.osgi.manifest.BundleManifest.RequireBundleHeader.RequireBundle;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * {@link Scoper} is a utility class for scoping a set of bundles. <p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * This class is not thread safe.
 * 
 * @author Glyn Normington
 * @since 1.0
 */
public class Scoper {

    private static final String BUNDLE_SYMBOLIC_NAME_ATTRIBUTE_NAME = "bundle-symbolic-name";

    public static final String SCOPE_SEPARATOR = "-";

    private static final int BUNDLE_MANIFEST_VERSION_FOR_OSGI_R4 = 2;

    private static final String VERSION_ATTRIBUTE_NAME = "version";

    private static final String VERSION_ATTRIBUTE_ALTERNATE_NAME = "specification-version";

    private static final String SCOPING_ATTRIBUTE_NAME = "module_scope";

    private static final Tracer TRACER = TracerFactory.createTracer(Scoper.class);

    private final List<BundleManifest> bundleManifests;

    private final String scopeName;

    private final Version scopeVersion;

    // The value to be used as a mandatory matching attribute value for scoping.
    private final String scopingAttributeValue;

    private final String scopingPrefix;

    // Map of unscoped package name to package version for all the packages exported by the bundles.
    private final Map<String, Version> exportedPackages = new HashMap<String, Version>();

    // Map of unscoped bundle symbolic name to bundle version for all the bundles.
    private final Map<String, Version> bundles = new HashMap<String, Version>();

    private final BundleManifestFactory bundleManifestFactory;

    /**
     * @param bundleManifests the manifests of the bundles to be scoped
     * @param scopeName the name of the scope to be created
     * @param scopeVersion the version of the scope to be created
     * @param bundleManifestFactory a factory for creating parts of a bundle manifest
     */
    public Scoper(List<BundleManifest> bundleManifests, String scopeName, Version scopeVersion, BundleManifestFactory bundleManifestFactory) {
        this.bundleManifests = bundleManifests;
        this.scopeName = scopeName;
        this.scopeVersion = scopeVersion;
        this.scopingAttributeValue = this.scopeName + SCOPE_SEPARATOR + versionToShortString(scopeVersion);
        this.scopingPrefix = this.scopingAttributeValue + SCOPE_SEPARATOR;
        this.bundleManifestFactory = bundleManifestFactory;
    }

    /**
     * Scope the bundles by transforming their metadata. This involves adding a mandatory matching attribute to exports
     * and any corresponding imports and adding a prefix to each bundle's bundle symbolic name and any corresponding
     * require bundle statements and matching attributes. The value of the mandatory matching attribute and the prefix
     * are uniquely determined by the application name and version.
     * 
     * Scoping also checks for ambiguities which could lead to unexpected wirings and throws an exception if these
     * checks fail. Specifically, each package exported by the bundles must be exported only once and no two bundles may
     * have the same bundle symbolic name.
     * 
     */
    public void scope() throws UnsupportedBundleManifestVersionException, DuplicateExportException, DuplicateBundleSymbolicNameException {
        scopeReferents();
        scopeReferences();
    }

    /**
     * Rescope the given bundle manifest.
     * 
     * @param bundleManifest the bundle manifest to be rescoped
     * @throws UnsupportedBundleManifestVersionException
     * @throws DuplicateExportException
     * @throws DuplicateBundleSymbolicNameException
     */
    public void rescope(BundleManifest bundleManifest) throws UnsupportedBundleManifestVersionException, DuplicateExportException,
        DuplicateBundleSymbolicNameException {
        scopeBundleReferents(bundleManifest, true);
        scopeBundleReferences(bundleManifest);
    }

    /**
     * Determine which packages are exported by the bundles and scope these exports and determine the bundles' symbolic
     * names and scope them.
     */
    private void scopeReferents() throws UnsupportedBundleManifestVersionException, DuplicateExportException, DuplicateBundleSymbolicNameException {
        for (BundleManifest bundleManifest : this.bundleManifests) {
            scopeBundleReferents(bundleManifest, false);
        }
    }

    /**
     * Scope the referents of the given bundle.
     * 
     * @param bundleManifest
     * @throws UnsupportedBundleManifestVersionException
     * @throws DuplicateExportException
     * @throws DuplicateBundleSymbolicNameException
     */
    private void scopeBundleReferents(BundleManifest bundleManifest, boolean allowDuplicates) throws UnsupportedBundleManifestVersionException,
        DuplicateExportException, DuplicateBundleSymbolicNameException {
        if (TRACER.isDebugEnabled()) {
            TRACER.traceDebug("Bundle manifest before scoping:\n" + bundleManifest);
        }
        // OSGi R4 features are essential for scoping.
        if (bundleManifest.getBundleManifestVersion() < BUNDLE_MANIFEST_VERSION_FOR_OSGI_R4) {

            throw new UnsupportedBundleManifestVersionException();
        }
        // Expand semicolon separated package lists to simplify scoping.
        bundleManifest.expandPackageLists();
        ExportPackageHeader exportPackageHeader = bundleManifest.getExportPackageHeader();
        if (exportPackageHeader != null) {
            for (PackageExport packageExport : exportPackageHeader.getPackageExports()) {
                scopePackageExport(packageExport, allowDuplicates);
            }
        }
        scopeBundleSymbolicName(bundleManifest, allowDuplicates);
        setModuleScope(bundleManifest);
    }

    private void setModuleScope(BundleManifest bundleManifest) {
        bundleManifest.setModuleScope(this.scopingAttributeValue);
    }

    /**
     * Scope the given package export.
     * 
     * @param packageExport the package export to be scoped
     * @param allowDuplicates true if and only if duplicated exports should be allowed
     */
    private void scopePackageExport(PackageExport packageExport, boolean allowDuplicates) throws DuplicateExportException,
        UnsupportedBundleManifestVersionException {
        packageExport.getAttributes().put(SCOPING_ATTRIBUTE_NAME, this.scopingAttributeValue);
        packageExport.getMandatory().add(SCOPING_ATTRIBUTE_NAME);

        Version packageVersion = packageExport.getVersion();

        for (String packageName : packageExport.getPackageNames()) {
            if (this.exportedPackages.containsKey(packageName)) {
                if (!allowDuplicates) {
                    diagnoseDuplicateExport(packageName);
                }
            } else {
                this.exportedPackages.put(packageName, packageVersion);
            }
        }
    }

    /**
     * @param packageName
     */
    private void diagnoseDuplicateExport(String packageName) throws DuplicateExportException, UnsupportedBundleManifestVersionException {
        StringBuffer exporters = new StringBuffer();
        boolean first = true;
        for (BundleManifest bundleManifest : this.bundleManifests) {
            // OSGi R4 features are essential for scoping.
            if (bundleManifest.getBundleManifestVersion() < BUNDLE_MANIFEST_VERSION_FOR_OSGI_R4) {
                throw new UnsupportedBundleManifestVersionException();
            }
            // Expand semicolon separated package lists to simplify scoping.
            bundleManifest.expandPackageLists();
            ExportPackageHeader exportPackageHeader = bundleManifest.getExportPackageHeader();
            if (exportPackageHeader != null) {
                for (PackageExport packageExport : exportPackageHeader.getPackageExports()) {
                    if (packageName.equals(packageExport.getPackageNames().get(0))) {
                        if (!first) {
                            exporters.append(", ");
                        }
                        first = false;
                        exporters.append(getUnscopedSymbolicName(bundleManifest));
                        exporters.append(" ");
                        exporters.append(bundleManifest.getBundleVersion());
                    }
                }
            }
        }

        throw new DuplicateExportException(packageName, exporters.toString());
    }

    /**
     * Get the bundle symbolic name of the given bundle manifest prior to scoping.
     * 
     * @param bundleManifest the bundle manifest
     * @return the unscoped bundle symbolic name
     */
    public static String getUnscopedSymbolicName(BundleManifest bundleManifest) {
        String symbolicName = null;
        BundleSymbolicName bundleSymbolicName = bundleManifest.getBundleSymbolicName();
        if (bundleSymbolicName != null) {
            symbolicName = bundleSymbolicName.getSymbolicName();
            String moduleScope = bundleManifest.getModuleScope();
            if (moduleScope != null) {
                String scopingPrefix = moduleScope + SCOPE_SEPARATOR;
                if (symbolicName.startsWith(scopingPrefix)) {
                    symbolicName = symbolicName.substring(scopingPrefix.length());
                }
            }
        }
        return symbolicName;
    }

    public String getUnscopedSymbolicName(String bundleSymbolicName) {
        String symbolicName = null;
        if (bundleSymbolicName != null) {
            symbolicName = bundleSymbolicName;
            if (symbolicName.startsWith(this.scopingPrefix)) {
                symbolicName = symbolicName.substring(this.scopingPrefix.length());
            }
        }
        return symbolicName;
    }

    public String getScopedSymbolicName(String bundleSymbolicName) {
        String symbolicName = getUnscopedSymbolicName(bundleSymbolicName);
        if (this.bundles.containsKey(symbolicName)) {
            return this.scopingPrefix + symbolicName;
        }
        return bundleSymbolicName;
    }

    /**
     * Scope the given bundle.
     * 
     * @param bundleManifest the bundle manifest to be scoped
     * @param allowDuplicates true if and only if duplicates are to be allowed
     */
    private void scopeBundleSymbolicName(BundleManifest bundleManifest, boolean allowDuplicates) throws DuplicateBundleSymbolicNameException {
        BundleSymbolicName bundleSymbolicName = bundleManifest.getBundleSymbolicName();
        if (bundleSymbolicName != null) {
            String symbolicName = getUnscopedSymbolicName(bundleManifest);
            if (this.bundles.containsKey(symbolicName)) {
                if (!allowDuplicates) {
                    throw new DuplicateBundleSymbolicNameException(symbolicName);
                }
            } else {
                Version version = stringToVersion(bundleManifest.getBundleVersion());
                this.bundles.put(symbolicName, version);
            }
            bundleSymbolicName.setSymbolicName(this.scopingPrefix + symbolicName);
        }
    }

    private Version stringToVersion(String versionString) {
        Version version;
        if (!StringUtils.hasText(versionString)) {
            version = Version.emptyVersion;
        } else {
            version = new Version(versionString);
        }
        return version;
    }

    /**
     * Scope the corresponding imports and dynamic imports and scope the corresponding require-bundle and imports and
     * dynamic imports.
     */
    private void scopeReferences() {
        for (BundleManifest bundleManifest : this.bundleManifests) {
            scopeBundleReferences(bundleManifest);
        }
    }

    /**
     * Scope the references of the given bundle.
     * 
     * @param bundleManifest
     */
    private void scopeBundleReferences(BundleManifest bundleManifest) {
        ImportPackageHeader importPackageHeader = bundleManifest.getImportPackageHeader();
        if (importPackageHeader != null) {
            for (PackageImport packageImport : importPackageHeader.getPackageImports()) {
                scopePackageImport(packageImport);
            }
        }
        DynamicImportPackageHeader dynamicImportPackageHeader = bundleManifest.getDynamicImportPackageHeader();
        if (dynamicImportPackageHeader != null) {
            List<DynamicPackageImport> resultDynamicPackageImports = new ArrayList<DynamicPackageImport>();
            for (DynamicPackageImport dynamicPackageImport : dynamicImportPackageHeader.getDynamicPackageImports()) {
                scopeDynamicPackageImport(dynamicPackageImport, resultDynamicPackageImports);
                resultDynamicPackageImports.add(dynamicPackageImport);
            }
            dynamicImportPackageHeader.setDynamicPackageImports(resultDynamicPackageImports);
        }
        scopeRequireBundle(bundleManifest);
        scopeFragmentHost(bundleManifest);

        if (TRACER.isDebugEnabled()) {
            TRACER.traceDebug("Bundle manifest after scoping:\n" + bundleManifest);
        }
    }

    /**
     * Scope the given package import.
     * 
     * @param packageImport the package import to be scoped
     */
    private void scopePackageImport(PackageImport packageImport) {
        for (String packageName : packageImport.getPackageNames()) {
            Version exportedVersion = this.exportedPackages.get(packageName);
            if (exportedVersion != null) {
                String importVersionRangeString = packageImport.getAttributes().get(VERSION_ATTRIBUTE_NAME);
                if (importVersionRangeString == null) {
                    importVersionRangeString = packageImport.getAttributes().get(VERSION_ATTRIBUTE_ALTERNATE_NAME);
                }
                VersionRange importVersionRange = new VersionRange(importVersionRangeString);
                if (importVersionRange.includes(exportedVersion)) {
                    packageImport.getAttributes().put(SCOPING_ATTRIBUTE_NAME, this.scopingAttributeValue);
                    if (packageImport.getAttributes().containsKey(BUNDLE_SYMBOLIC_NAME_ATTRIBUTE_NAME)) {
                        String symbolicName = packageImport.getAttributes().get(BUNDLE_SYMBOLIC_NAME_ATTRIBUTE_NAME);
                        if (this.bundles.containsKey(symbolicName)) {
                            packageImport.getAttributes().put(BUNDLE_SYMBOLIC_NAME_ATTRIBUTE_NAME, this.scopingPrefix + symbolicName);
                        }
                    }
                } else {
                    TRACER.traceWarning("Import of package '" + packageName + "' was not scoped to scope '" + this.scopeName + "' version '"
                        + this.scopeVersion + "' as its version range did not include the version of the scoped export of the package");
                }
            }
        }
    }

    /**
     * Scope the given dynamic package import and append the scoped form to the given list.
     * 
     * @param dynamicPackageImport the dynamic package import to be scoped
     * @param resultDynamicPackageImports the list to append to
     */
    private void scopeDynamicPackageImport(DynamicPackageImport dynamicPackageImport, List<DynamicPackageImport> resultDynamicPackageImports) {
        DynamicPackageImport scopedDynamicImport = this.bundleManifestFactory.createDynamicPackageImport();
        scopedDynamicImport.getWildcardPackageNames().addAll(dynamicPackageImport.getWildcardPackageNames());
        scopedDynamicImport.getAttributes().putAll(dynamicPackageImport.getAttributes());
        scopedDynamicImport.getAttributes().put(SCOPING_ATTRIBUTE_NAME, this.scopingAttributeValue);
        if (scopedDynamicImport.getAttributes().containsKey(BUNDLE_SYMBOLIC_NAME_ATTRIBUTE_NAME)) {
            String symbolicName = scopedDynamicImport.getAttributes().get(BUNDLE_SYMBOLIC_NAME_ATTRIBUTE_NAME);
            if (this.bundles.containsKey(symbolicName)) {
                scopedDynamicImport.getAttributes().put(BUNDLE_SYMBOLIC_NAME_ATTRIBUTE_NAME, this.scopingPrefix + symbolicName);
            }
        }
        resultDynamicPackageImports.add(scopedDynamicImport);
    }

    /**
     * Scope the given require bundle.
     * 
     * @param bundleManifest the bundle manifest to be scoped
     */
    private void scopeRequireBundle(BundleManifest bundleManifest) {
        RequireBundleHeader requireBundleHeader = bundleManifest.getRequireBundleHeader();
        if (requireBundleHeader != null) {
            List<RequireBundle> requiredBundles = requireBundleHeader.getRequiredBundles();
            for (RequireBundle requiredBundle : requiredBundles) {
                String requiredBundleSymbolicName = requiredBundle.getBundleSymbolicName();
                if (this.bundles.containsKey(requiredBundleSymbolicName)) {
                    Version version = this.bundles.get(requiredBundleSymbolicName);
                    VersionRange requiredVersionRange = new VersionRange(requiredBundle.getAttributes().get(VERSION_ATTRIBUTE_NAME));
                    if (requiredVersionRange.includes(version)) {
                        requiredBundle.setBundleSymbolicName(this.scopingPrefix + requiredBundleSymbolicName);
                    }
                }
            }
        }
    }

    /**
     * Scope any reference to one of the scoped bundles by a fragment host header.
     * 
     * @param bundleManifest the bundle manifest to be scoped for fragment host
     */
    private void scopeFragmentHost(BundleManifest bundleManifest) {
        FragmentHostHeader fragmentHostHeader = bundleManifest.getFragmentHostHeader();
        if (fragmentHostHeader != null) {
            String bundleSymbolicName = fragmentHostHeader.getBundleSymbolicName();
            if (this.bundles.containsKey(bundleSymbolicName)) {
                fragmentHostHeader.setBundleSymbolicName(this.scopingPrefix + bundleSymbolicName);
            }
        }
    }

    public class DuplicateExportException extends Exception {

        private static final long serialVersionUID = -6672058951941449763L;

        private final String packageName;

        private final String exporters;

        private DuplicateExportException(String packageName, String exporters) {
            this.packageName = packageName;
            this.exporters = exporters;
        }

        public String getPackageName() {
            return this.packageName;
        }

        public String getExporters() {
            return this.exporters;
        }
    }

    public class UnsupportedBundleManifestVersionException extends Exception {

        private static final long serialVersionUID = -282020071571817876L;

        UnsupportedBundleManifestVersionException() {
            super("Cannot scope a bundle which does not specify a bundle manifest version of at least " + BUNDLE_MANIFEST_VERSION_FOR_OSGI_R4);
        }

        /**
         * Returns the lowest bundle manifest version for which scoping is supported
         * 
         * @return the lowest bundle manifest version for which scoping is supported
         */
        public int getLowestSupportedVersion() {
            return BUNDLE_MANIFEST_VERSION_FOR_OSGI_R4;
        }
    }

    public class DuplicateBundleSymbolicNameException extends Exception {

        private static final long serialVersionUID = -4228236795055040322L;

        private final String bundleSymbolicName;

        public DuplicateBundleSymbolicNameException(String bundleSymbolicName) {
            this.bundleSymbolicName = bundleSymbolicName;
        }

        public String getBundleSymbolicName() {
            return this.bundleSymbolicName;
        }
    }

    private static String versionToShortString(Version version) {
        String result = version.toString();
        while (result.endsWith(".0")) {
            result = result.substring(0, result.length() - 2);
        }
        return result;
    }

}
