/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.manifest;

import java.io.IOException;
import java.io.OutputStream;
import java.io.Reader;
import java.io.StringReader;
import java.io.StringWriter;
import java.io.Writer;
import java.util.ArrayList;
import java.util.List;
import java.util.jar.Attributes;
import java.util.jar.Manifest;

import org.springframework.util.StringUtils;

import com.springsource.server.osgi.manifest.BundleManifest.DynamicImportPackageHeader.DynamicPackageImport;
import com.springsource.server.osgi.manifest.BundleManifest.ExportPackageHeader.PackageExport;
import com.springsource.server.osgi.manifest.BundleManifest.ImportPackageHeader.PackageImport;
import com.springsource.server.osgi.manifest.parse.FatalParseException;
import com.springsource.server.osgi.manifest.parse.OSGiIntegrationLogService;
import com.springsource.server.osgi.manifest.parse.ParserLogger;

/**
 * This is the implementation of the BundleManifest interface.
 * <p />
 * To support an additional header, provide an assist class, a setter, a getter, and update the write and isEquivalent
 * methods.
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * This class is not thread safe.
 * 
 * @author Glyn Normington
 * @author Christopher Frost
 * @author Andy Wilkinson
 * @since 1.0
 */
public class StandardBundleManifest implements BundleManifest {

    private static final String SERVER_FRAGMENT_AUTO_ATTACH = "Server-Fragment-AutoAttach";

    private static final String MANIFEST_VERSION_NAME = "Manifest-Version";

    private static final String DEFAULT_MANIFEST_VERSION = "1.0";

    private static String MODULE_SCOPE = "Module-Scope";

    private final BundleClassPathDelegate bundleClassPathDelegate;

    private final BundleDescriptionDelegate bundleDescriptionDelegate;

    private final BundleDocUrlDelegate bundleDocUrlDelegate;

    private final BundleUpdateLocationDelegate bundleUpdateLocationDelegate;

    private final BundleManifestVersionDelegate bundleManifestVersionDelegate;

    private final ImportPackageDelegate importPackageDelegate;

    private final BundleVersionDelegate bundleVersionDelegate;

    private final BundleSymbolicNameDelegate bundleSymbolicNameDelegate;

    private final BundleNameDelegate bundleNameDelegate;

    private final DynamicImportPackageDelegate dynamicImportPackageDelegate;

    private final RequireBundleDelegate requireBundleDelegate;

    private final ExportPackageDelegate exportPackageDelegate;

    private final FragmentHostDelegate fragmentHostDelegate;

    private final ImportLibraryDelegate importLibraryDelegate;

    private final ImportBundleDelegate importBundleDelegate;
    
    private final BundleActivationPolicyDelegate bundleActivationPolicyDelegate;

    protected final Manifest inputManifest;

    private final List<Delegate> delegates = new ArrayList<Delegate>(15);

    private final BundleManifestFactory bundleManifestFactory;

    private boolean packagesExpanded = false;

    public StandardBundleManifest(Reader manifest, ParserLogger logger) {
        this.inputManifest = new ManifestFactory(manifest).getManifest();
        if (getManifestVersion() == null) {
            setManifestVersion(DEFAULT_MANIFEST_VERSION);
        }

        if (logger == null) {
            logger = new OSGiIntegrationLogService();
        }

        this.bundleManifestFactory = new StandardBundleManifestFactory(logger);
        Attributes mainAttributes = this.inputManifest.getMainAttributes();

        this.delegates.add(this.bundleNameDelegate = new BundleNameDelegate(mainAttributes));
        this.delegates.add(this.bundleClassPathDelegate = new BundleClassPathDelegate(mainAttributes));
        this.delegates.add(this.bundleDescriptionDelegate = new BundleDescriptionDelegate(mainAttributes));
        this.delegates.add(this.bundleDocUrlDelegate = new BundleDocUrlDelegate(mainAttributes));
        this.delegates.add(this.bundleManifestVersionDelegate = new BundleManifestVersionDelegate(mainAttributes));
        this.delegates.add(this.importPackageDelegate = new ImportPackageDelegate(mainAttributes, this.bundleManifestFactory));
        this.delegates.add(this.dynamicImportPackageDelegate = new DynamicImportPackageDelegate(mainAttributes, this.bundleManifestFactory));
        this.delegates.add(this.bundleUpdateLocationDelegate = new BundleUpdateLocationDelegate(mainAttributes));
        this.delegates.add(this.bundleVersionDelegate = new BundleVersionDelegate(mainAttributes));
        this.delegates.add(this.bundleSymbolicNameDelegate = new BundleSymbolicNameDelegate(mainAttributes, this.bundleManifestFactory));
        this.delegates.add(this.requireBundleDelegate = new RequireBundleDelegate(mainAttributes, this.bundleManifestFactory));
        this.delegates.add(this.fragmentHostDelegate = new FragmentHostDelegate(mainAttributes, this.bundleManifestFactory));
        this.delegates.add(this.exportPackageDelegate = new ExportPackageDelegate(mainAttributes, this.bundleManifestFactory));
        this.delegates.add(this.importLibraryDelegate = new ImportLibraryDelegate(mainAttributes, this.bundleManifestFactory));
        this.delegates.add(this.importBundleDelegate = new ImportBundleDelegate(mainAttributes, this.bundleManifestFactory));
        this.delegates.add(this.bundleActivationPolicyDelegate = new BundleActivationPolicyDelegate(mainAttributes, this.bundleManifestFactory));

        for (Delegate delegate : this.delegates) {
            delegate.checkHeaderValidity();
        }

        if (this.bundleManifestFactory.isParseClean() != null) {
            BundleSymbolicName bunSymName = this.getBundleSymbolicName();
            String symName = bunSymName == null ? "Null symbolic name" : bunSymName.getSymbolicName();

            String headers = StringUtils.arrayToCommaDelimitedString(this.bundleManifestFactory.isParseClean());

            throw new FatalParseException("The manifest identified by Name '" + this.getBundleName() + "', Symbolic Name '" + symName
                + "', and Version '" + this.getBundleVersion() + "' has the following errors [" + headers + "]");
        }

    }

    /**
     * {@inheritDoc}
     */
    public void setManifestVersion(String string) {
        this.inputManifest.getMainAttributes().putValue(MANIFEST_VERSION_NAME, string);
    }

    /**
     * {@inheritDoc}
     */
    public String getManifestVersion() {
        return this.inputManifest.getMainAttributes().getValue(MANIFEST_VERSION_NAME);
    }

    /**
     * {@inheritDoc}
     */
    public void setModuleScope(String scope) {
        this.inputManifest.getMainAttributes().putValue(MODULE_SCOPE, scope);
    }

    /**
     * {@inheritDoc}
     */
    public String getModuleScope() {
        return this.inputManifest.getMainAttributes().getValue(MODULE_SCOPE);
    }

    /**
     * {@inheritDoc}
     */
    public boolean isFragmentAutoAttach() {
        String value = this.inputManifest.getMainAttributes().getValue(SERVER_FRAGMENT_AUTO_ATTACH);
        return value == null || "true".equals(value);
    }

    public StandardBundleManifest() {
        this(new StringReader(""), null);
    }

    /**
     * {@inheritDoc}
     */
    public Manifest toManifest() {
        for (Delegate delegate : this.delegates) {
            delegate.checkHeaderValidity();
        }

        Manifest outputManifest = new Manifest();
        Attributes outputAttributes = outputManifest.getMainAttributes();
        // Copy the input attributes
        outputAttributes.putAll(this.inputManifest.getMainAttributes());
        // Copy across any per-entry attributes
        outputManifest.getEntries().putAll(this.inputManifest.getEntries());

        // Overwrite the output attributes we care about skipping those that were never read or, more importantly,
        // written.
        this.bundleClassPathDelegate.write(outputAttributes);
        this.bundleDescriptionDelegate.write(outputAttributes);
        this.bundleDocUrlDelegate.write(outputAttributes);
        this.bundleManifestVersionDelegate.write(outputAttributes);
        this.importPackageDelegate.write(outputAttributes);
        this.bundleNameDelegate.write(outputAttributes);
        this.dynamicImportPackageDelegate.write(outputAttributes);
        this.bundleUpdateLocationDelegate.write(outputAttributes);
        this.bundleVersionDelegate.write(outputAttributes);
        this.bundleSymbolicNameDelegate.write(outputAttributes);
        this.requireBundleDelegate.write(outputAttributes);
        this.fragmentHostDelegate.write(outputAttributes);
        this.exportPackageDelegate.write(outputAttributes);
        this.importLibraryDelegate.write(outputAttributes);
        this.importBundleDelegate.write(outputAttributes);
        this.bundleActivationPolicyDelegate.write(outputAttributes);

        return outputManifest;
    }

    /**
     * {@inheritDoc}
     */
    public void write(Writer writer) {

        Manifest outputManifest = toManifest();

        // Write the manifest
        try {
            OutputStream manifestOutputStream = new ManifestOutputStream(writer);
            outputManifest.write(manifestOutputStream);
            manifestOutputStream.close();
        } catch (IOException e) {
            throw new FatalParseException("Exception writing manifest, Name[" + this.getBundleName() + "], Symbolic Name["
                + this.getBundleSymbolicName().getSymbolicName() + "], Version[" + this.getBundleVersion() + "]", e);
        }
    }

    /**
     * {@inheritDoc}
     */
    public List<String> getBundleClassPath() {
        return this.bundleClassPathDelegate.getBundleClassPath();
    }

    /**
     * {@inheritDoc}
     */
    public void setBundleClassPath(List<String> bundleClassPath) {
        this.bundleClassPathDelegate.setBundleClassPath(bundleClassPath);
    }

    /**
     * {@inheritDoc}
     */
    public String getBundleDescription() {
        return this.bundleDescriptionDelegate.getBundleDescription();
    }

    /**
     * {@inheritDoc}
     */
    public void setBundleDescription(String bundleDescription) {
        this.bundleDescriptionDelegate.setBundleDescription(bundleDescription);
    }

    /**
     * {@inheritDoc}
     */
    public String getBundleDocUrl() {
        return this.bundleDocUrlDelegate.getBundleDocURL();
    }

    /**
     * {@inheritDoc}
     */
    public void setBundleDocUrl(String bundleDocUrl) {
        this.bundleDocUrlDelegate.setBundleDocURL(bundleDocUrl);
    }

    /**
     * {@inheritDoc}
     */
    public String getBundleName() {
        return this.bundleNameDelegate.getBundleName();
    }

    /**
     * {@inheritDoc}
     */
    public void setBundleName(String bundleName) {
        this.bundleNameDelegate.setBundleName(bundleName);
    }

    /**
     * {@inheritDoc}
     */
    public BundleSymbolicName getBundleSymbolicName() {
        return this.bundleSymbolicNameDelegate.getBundleSymbolicName();
    }

    /**
     * {@inheritDoc}
     */
    public void setBundleSymbolicName(BundleSymbolicName bundleSymbolicName) {
        this.bundleSymbolicNameDelegate.setBundleSymbolicName(bundleSymbolicName);
    }

    /**
     * {@inheritDoc}
     */
    public String getBundleUpdateLocation() {
        return this.bundleUpdateLocationDelegate.getBundleUpdateLocation();
    }

    /**
     * {@inheritDoc}
     */
    public void setBundleUpdateLocation(String bundleUpdateLocation) {
        this.bundleUpdateLocationDelegate.setBundleUpdateLocation(bundleUpdateLocation);
    }

    /**
     * {@inheritDoc}
     */
    public String getBundleVersion() {
        return this.bundleVersionDelegate.getBundleVersion();
    }

    /**
     * {@inheritDoc}
     */
    public void setBundleVersion(String bundleVersion) {
        this.bundleVersionDelegate.setBundleVersion(bundleVersion);
    }

    /**
     * {@inheritDoc}
     */
    public DynamicImportPackageHeader getDynamicImportPackageHeader() {
        return this.dynamicImportPackageDelegate.getDynamicImportPackageHeader();
    }

    /**
     * {@inheritDoc}
     */
    public void setDynamicImportPackageHeader(DynamicImportPackageHeader dynamicImportPackageHeader) {
        this.dynamicImportPackageDelegate.setDynamicImportPackageHeader(dynamicImportPackageHeader);
    }

    /**
     * {@inheritDoc}
     */
    public ExportPackageHeader getExportPackageHeader() {
        return this.exportPackageDelegate.getExportPackageHeader();
    }

    /**
     * {@inheritDoc}
     */
    public void setExportPackageHeader(ExportPackageHeader exportPackageHeader) {
        this.exportPackageDelegate.setExportPackageHeader(exportPackageHeader);
    }

    /**
     * {@inheritDoc}
     */
    public FragmentHostHeader getFragmentHostHeader() {
        return this.fragmentHostDelegate.getFragmentHostHeader();
    }

    /**
     * {@inheritDoc}
     */
    public void setFragmentHostHeader(FragmentHostHeader fragmentHostHeader) {
        this.fragmentHostDelegate.setFragmentHostHeader(fragmentHostHeader);
    }

    /**
     * {@inheritDoc}
     */
    public ImportPackageHeader getImportPackageHeader() {
        return this.importPackageDelegate.getImportPackageHeader();
    }

    /**
     * {@inheritDoc}
     */
    public void setImportPackageHeader(ImportPackageHeader importPackageHeader) {
        this.importPackageDelegate.setImportPackageHeader(importPackageHeader);
    }

    /**
     * {@inheritDoc}
     */
    public RequireBundleHeader getRequireBundleHeader() {
        return this.requireBundleDelegate.getRequireBundleHeader();
    }

    /**
     * {@inheritDoc}
     */
    public void setRequireBundleHeader(RequireBundleHeader requireBundleHeader) {
        this.requireBundleDelegate.setRequireBundleHeader(requireBundleHeader);
    }

    /**
     * {@inheritDoc}
     */
    public ImportLibraryHeader getImportLibraryHeader() {
        return this.importLibraryDelegate.getImportLibraryHeader();
    }

    /**
     * {@inheritDoc}
     */
    public void setImportLibraryHeader(ImportLibraryHeader importLibraryHeader) {
        this.importLibraryDelegate.setImportLibraryHeader(importLibraryHeader);
    }

    /**
     * {@inheritDoc}
     */
    public ImportBundleHeader getImportBundleHeader() {
        return this.importBundleDelegate.getImportBundleHeader();
    }
    
    /** 
     * {@inheritDoc}
     */
    public void setImportBundleHeader(ImportBundleHeader importBundleHeader) {
        this.importBundleDelegate.setImportBundleHeader(importBundleHeader);
    }

    /**
     * {@inheritDoc}
     */
    public int getBundleManifestVersion() {
        return this.bundleManifestVersionDelegate.getBundleManifestVersion();
    }

    /** 
     * {@inheritDoc}
     */
    public void setBundleManifestVersion(int bundleManifestVersion) {
        this.bundleManifestVersionDelegate.setBundleManifestVersion(bundleManifestVersion);
    }
    
    /** 
     * {@inheritDoc}
     */
    public BundleActivationPolicy getBundleActivationPolicy() {
        return this.bundleActivationPolicyDelegate.getBundleActivationPolicy();
    }
    
    /** 
     * {@inheritDoc}
     */
    public void setBundleActivationPolicy(BundleActivationPolicy bundleActivationPolicy) {
        this.bundleActivationPolicyDelegate.setBundleActivationPolicy(bundleActivationPolicy);
    }

    public static class ManifestOutputStream extends OutputStream {

        private final Writer writer;

        public ManifestOutputStream(Writer writer) {
            super();
            this.writer = writer;
        }

        @Override
        public void write(int b) throws IOException {
            this.writer.write(b);
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public void close() throws IOException {
            this.writer.flush(); // may not be strictly necessary
            this.writer.close();
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String toString() {
        StringWriter outputManifest = new StringWriter();
        write(outputManifest);
        return outputManifest.toString();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public int hashCode() {
        throw new UnsupportedOperationException();
    }

    /**
     * {@inheritDoc}
     */
    public boolean isEquivalent(BundleManifest obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final StandardBundleManifest other = (StandardBundleManifest) obj;

        return this.bundleClassPathDelegate.equals(other.bundleClassPathDelegate) &&

        this.bundleDescriptionDelegate.equals(other.bundleDescriptionDelegate) &&

        this.bundleDocUrlDelegate.equals(other.bundleDocUrlDelegate) &&

        this.bundleManifestVersionDelegate.equals(other.bundleManifestVersionDelegate) &&

        this.bundleNameDelegate.equals(other.bundleNameDelegate) &&

        this.bundleSymbolicNameDelegate.equals(other.bundleSymbolicNameDelegate) &&

        this.bundleUpdateLocationDelegate.equals(other.bundleUpdateLocationDelegate) &&

        this.bundleVersionDelegate.equals(other.bundleVersionDelegate) &&

        this.dynamicImportPackageDelegate.equals(other.dynamicImportPackageDelegate) &&

        this.exportPackageDelegate.equals(other.exportPackageDelegate) &&

        this.fragmentHostDelegate.equals(other.fragmentHostDelegate) &&

        this.importPackageDelegate.equals(other.importPackageDelegate) &&

        this.requireBundleDelegate.equals(other.requireBundleDelegate) &&

        this.importLibraryDelegate.equals(other.importLibraryDelegate) &&

        this.importBundleDelegate.equals(other.importBundleDelegate) &&
        
        this.bundleActivationPolicyDelegate.equals(other.bundleActivationPolicyDelegate);

    }

    /**
     * {@inheritDoc}
     */
    public void expandPackageLists() {
        if (!this.packagesExpanded) {
            expandImportPackageLists();
            expandDynamicImportPackageLists();
            expandExportPackageLists();
            this.packagesExpanded = true;
        }
    }

    private void expandImportPackageLists() {
        ImportPackageHeader importPackageHeader = getImportPackageHeader();
        if (importPackageHeader != null) {
            List<PackageImport> expandedPackageImports = new ArrayList<PackageImport>(5);
            for (PackageImport packageImport : importPackageHeader.getPackageImports()) {
                if (packageImport.getPackageNames().size() > 1) {
                    for (String packageName : packageImport.getPackageNames()) {
                        PackageImport expandedPackageImport = this.bundleManifestFactory.createPackageImport();
                        expandedPackageImport.getPackageNames().add(packageName);
                        expandedPackageImport.getAttributes().putAll(packageImport.getAttributes());
                        expandedPackageImport.setResolution(packageImport.getResolution());
                        expandedPackageImports.add(expandedPackageImport);
                    }
                } else {
                    expandedPackageImports.add(packageImport);
                }
            }
            importPackageHeader.setPackageImports(expandedPackageImports);
        }
    }

    private void expandDynamicImportPackageLists() {
        DynamicImportPackageHeader dynamicImportPackageHeader = getDynamicImportPackageHeader();
        if (dynamicImportPackageHeader != null) {
            List<DynamicPackageImport> expandedDynamicPackageImports = new ArrayList<DynamicPackageImport>();
            for (DynamicPackageImport dynamicPackageImport : dynamicImportPackageHeader.getDynamicPackageImports()) {
                if (dynamicPackageImport.getWildcardPackageNames().size() > 1) {
                    for (String packageName : dynamicPackageImport.getWildcardPackageNames()) {
                        DynamicPackageImport expandedDynamicPackageImport = this.bundleManifestFactory.createDynamicPackageImport();
                        expandedDynamicPackageImport.getWildcardPackageNames().add(packageName);
                        expandedDynamicPackageImport.getAttributes().putAll(dynamicPackageImport.getAttributes());
                        expandedDynamicPackageImports.add(expandedDynamicPackageImport);
                    }
                } else {
                    expandedDynamicPackageImports.add(dynamicPackageImport);
                }
            }
            dynamicImportPackageHeader.setDynamicPackageImports(expandedDynamicPackageImports);
        }
    }

    private void expandExportPackageLists() {
        ExportPackageHeader exportPackageHeader = getExportPackageHeader();
        if (exportPackageHeader != null) {
            List<PackageExport> expandedPackageExports = new ArrayList<PackageExport>(5);
            for (PackageExport packageExport : exportPackageHeader.getPackageExports()) {
                if (packageExport.getPackageNames().size() > 1) {
                    for (String packageName : packageExport.getPackageNames()) {
                        PackageExport expandedPackageExport = this.bundleManifestFactory.createPackageExport();
                        expandedPackageExport.getPackageNames().add(packageName);
                        expandedPackageExport.getAttributes().putAll(packageExport.getAttributes());
                        expandedPackageExport.getUses().addAll(packageExport.getUses());
                        expandedPackageExport.getMandatory().addAll(packageExport.getMandatory());
                        expandedPackageExport.getInclude().addAll(packageExport.getInclude());
                        expandedPackageExport.getExclude().addAll(packageExport.getExclude());
                        expandedPackageExports.add(expandedPackageExport);
                    }
                } else {
                    expandedPackageExports.add(packageExport);
                }
            }
            exportPackageHeader.setPackageExports(expandedPackageExports);
        }
    }

}
