/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.manifest;

import java.io.Reader;
import java.util.List;

import com.springsource.server.osgi.manifest.BundleManifest.BundleActivationPolicy;
import com.springsource.server.osgi.manifest.BundleManifest.BundleSymbolicName;
import com.springsource.server.osgi.manifest.BundleManifest.DynamicImportPackageHeader;
import com.springsource.server.osgi.manifest.BundleManifest.ExportPackageHeader;
import com.springsource.server.osgi.manifest.BundleManifest.FragmentHostHeader;
import com.springsource.server.osgi.manifest.BundleManifest.ImportBundleHeader;
import com.springsource.server.osgi.manifest.BundleManifest.ImportLibraryHeader;
import com.springsource.server.osgi.manifest.BundleManifest.ImportPackageHeader;
import com.springsource.server.osgi.manifest.BundleManifest.RequireBundleHeader;
import com.springsource.server.osgi.manifest.BundleManifest.DynamicImportPackageHeader.DynamicPackageImport;
import com.springsource.server.osgi.manifest.BundleManifest.ExportPackageHeader.PackageExport;
import com.springsource.server.osgi.manifest.BundleManifest.ImportBundleHeader.BundleImport;
import com.springsource.server.osgi.manifest.BundleManifest.ImportLibraryHeader.LibraryImport;
import com.springsource.server.osgi.manifest.BundleManifest.ImportPackageHeader.PackageImport;
import com.springsource.server.osgi.manifest.BundleManifest.RequireBundleHeader.RequireBundle;
import com.springsource.server.osgi.manifest.parse.ParserLogger;

/**
 * <strong>Concurrent Semantics</strong><br />
 * 
 * This class is thread safe.
 * 
 * @author Glyn Normington
 * @author Andy Wilkinson
 * @since 1.0
 */
public class StandardBundleManifestFactory implements BundleManifestFactory {

    private final ParserLogger parserLogger;

    public StandardBundleManifestFactory(ParserLogger parserLogger) {
        this.parserLogger = parserLogger;
    }

    public String[] isParseClean() {
        return this.parserLogger.errorReports();
    }

    /**
     * {@inheritDoc}
     */
    public BundleManifest createBundleManifest() {
        return new StandardBundleManifest();
    }

    /**
     * {@inheritDoc}
     */
    public BundleManifest createBundleManifest(Reader manifestContents) {
        return new StandardBundleManifest(manifestContents, this.parserLogger);
    }

    // ####################################

    /**
     * {@inheritDoc}
     */
    public DynamicImportPackageHeader parseDynamicImportPackageHeader(String creationValue) {
        return new StandardDynamicImportPackageHeader(this.parserLogger, creationValue);
    }

    /**
     * {@inheritDoc}
     */
    public DynamicImportPackageHeader createDynamicImportPackageHeader(
        List<BundleManifest.DynamicImportPackageHeader.DynamicPackageImport> dynamicImports) {
        return new StandardDynamicImportPackageHeader(dynamicImports);
    }

    /**
     * {@inheritDoc}
     */
    public DynamicPackageImport createDynamicPackageImport() {
        return new StandardDynamicPackageImport();
    }

    // ####################################

    /**
     * {@inheritDoc}
     */
    public ExportPackageHeader parseExportPackageHeader(String creationValue) {
        return new StandardExportPackageHeader(this.parserLogger, creationValue);
    }

    /**
     * {@inheritDoc}
     */
    public ExportPackageHeader createExportPackageHeader(List<PackageExport> packageExports) {
        return new StandardExportPackageHeader(packageExports);
    }

    /**
     * {@inheritDoc}
     */
    public PackageExport createPackageExport() {
        return new StandardPackageExport();
    }

    // ####################################

    /**
     * {@inheritDoc}
     */
    public FragmentHostHeader parseFragmentHostHeader(String creationValue) {
        FragmentHostHeader fragmentHostHeader = new StandardFragmentHostHeader(this.parserLogger, creationValue);
        return fragmentHostHeader;
    }

    // ####################################

    /**
     * {@inheritDoc}
     */
    public ImportPackageHeader parseImportPackageHeader(String creationValue) {
        return new StandardImportPackageHeader(this.parserLogger, creationValue);
    }

    /**
     * {@inheritDoc}
     */
    public ImportPackageHeader createImportPackageHeader(List<PackageImport> packageImports) {
        return new StandardImportPackageHeader(packageImports);
    }

    /**
     * {@inheritDoc}
     */
    public PackageImport createPackageImport() {
        return new StandardPackageImport();
    }

    // ####################################

    /**
     * {@inheritDoc}
     */
    public RequireBundleHeader parseRequireBundleHeader(String creationValue) {
        return new StandardRequireBundleHeader(this.parserLogger, creationValue);
    }

    /**
     * {@inheritDoc}
     */
    public RequireBundle createRequireBundle(String bundleSymbolicName) {
        return new StandardRequireBundle(bundleSymbolicName);
    }

    /**
     * {@inheritDoc}
     */
    public LibraryImport createLibraryImport(String librarySymbolicName) {
        return new StandardLibraryImport(librarySymbolicName);
    }

    /**
     * {@inheritDoc}
     */
    public ImportLibraryHeader parseImportLibraryHeader(String creationValue) {
        return new StandardImportLibraryHeader(this.parserLogger, creationValue);
    }
    
    public ImportBundleHeader parseImportBundleHeader(String creationValue) {
        return new StandardImportBundleHeader(this.parserLogger, creationValue);
    }

    /**
     * {@inheritDoc}
     */
    public BundleSymbolicName createBundleSymbolicName(String bundleSymbolicName) {
        return new StandardBundleSymbolicName(bundleSymbolicName);
    }

    /**
     * {@inheritDoc}
     */
    public BundleSymbolicName parseBundleSymbolicName(String creationValue) {
        return new StandardBundleSymbolicName(this.parserLogger, creationValue);
    }

    /** 
     * {@inheritDoc}
     */
    public BundleActivationPolicy parseBundleActivationPolicy(String creationValue) {
        return new StandardBundleActivationPolicy(this.parserLogger, creationValue);
    }
    
    /**
     * {@inheritDoc}
     */
    public BundleImport createBundleImport(String bundleSymbolicName) {
        return new StandardBundleImport(bundleSymbolicName);
    }

    /** 
     * {@inheritDoc}
     */
    public ImportBundleHeader createImportBundleHeader(List<BundleImport> imports) {
        // TODO Auto-generated method stub
        return new StandardImportBundleHeader(imports);
    }
}
