/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.manifest;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.springframework.util.Assert;

import com.springsource.server.osgi.manifest.BundleManifest.BundleSymbolicName;
import com.springsource.server.osgi.manifest.parse.HeaderDeclaration;
import com.springsource.server.osgi.manifest.parse.HeaderParserFactory;
import com.springsource.server.osgi.manifest.parse.ParserLogger;

/**
 * <strong>Concurrent Semantics</strong><br />
 * This class is <strong>not</strong> thread-safe.
 * 
 * @author Andy Wilkinson
 * @since 1.0
 */
class StandardBundleSymbolicName implements BundleSymbolicName {

    private static final String EQUALS = "=";

    private static final String DIRECTIVE = ":=";

    private static final String SEPARATOR = ";";

    private static final String FRAGMENT_ATTACHMENT_DIRECTIVE = "fragment-attachment";
    
    private static final String SINGLETON_DIRECTIVE = "singleton";

    private String symbolicName;

    private boolean singleton = false;

    private FragmentAttachment fragmentAttachment = FragmentAttachment.ALWAYS;
    
    private Map<String, String> attributes;

    StandardBundleSymbolicName(String symbolicName) {
        this.symbolicName = symbolicName;
        this.attributes = new HashMap<String, String>();
    }

    StandardBundleSymbolicName(ParserLogger logger, String bundleSymbolicNameString) {
        HeaderDeclaration headerDeclaration = HeaderParserFactory.newHeaderParser(logger).parseBundleSymbolicName(bundleSymbolicNameString);
        List<String> names = headerDeclaration.getNames();
        Assert.isTrue(names.size() == 1);
        this.symbolicName = names.get(0);

        Map<String, String> directives = headerDeclaration.getDirectives();

        String singletonString = directives.get(SINGLETON_DIRECTIVE);
        if (singletonString != null) {
            this.singleton = Boolean.parseBoolean(singletonString);
        }

        String rawFragmentAttachment = directives.get(FRAGMENT_ATTACHMENT_DIRECTIVE);
        if (rawFragmentAttachment != null) {
            this.fragmentAttachment = FragmentAttachment.valueOf(rawFragmentAttachment.toUpperCase().replace("-", "_"));
        }
        
        this.attributes = headerDeclaration.getAttributes();
        if (this.attributes == null) {
            this.attributes = new HashMap<String, String>();
        }
    }

    /**
     * {@inheritDoc}
     */
    public Map<String, String> getAttributes() {
        return this.attributes;
    }

    /**
     * {@inheritDoc}
     */
    public FragmentAttachment getFragmentAttachment() {
        return this.fragmentAttachment;
    }

    /**
     * {@inheritDoc}
     */
    public boolean isSingleton() {
        return this.singleton;
    }

    /**
     * {@inheritDoc}
     */
    public String getSymbolicName() {
        return this.symbolicName;
    }

    /**
     * {@inheritDoc}
     */
    public void setSymbolicName(String symbolicName) {
        this.symbolicName = symbolicName;
    }

    @Override public String toString() {
        StringBuffer buffer = new StringBuffer();
        buffer.append(this.symbolicName);

        if (this.singleton) {
            appendDirective(SINGLETON_DIRECTIVE, Boolean.toString(this.singleton), buffer);
        }

        if (this.fragmentAttachment != null && !this.fragmentAttachment.equals(FragmentAttachment.ALWAYS)) {
            appendDirective(FRAGMENT_ATTACHMENT_DIRECTIVE, this.fragmentAttachment.toString().toLowerCase(), buffer);
        }
        
        for (String name : this.attributes.keySet()) {
            appendAttribute(name, this.attributes.get(name), buffer);
        }
        return buffer.toString();
    }

    private void appendDirective(String name, String value, StringBuffer buffer) {
        buffer.append(SEPARATOR).append(name).append(DIRECTIVE).append(value);
    }

    private void appendAttribute(String name, String value, StringBuffer buffer) {
        buffer.append(SEPARATOR).append(name).append(EQUALS).append(value);
    }

    @Override public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (!(obj instanceof StandardBundleSymbolicName)) {
            return false;
        }
        final StandardBundleSymbolicName other = (StandardBundleSymbolicName) obj;
        if (this.attributes == null) {
            if (other.attributes != null) {
                return false;
            }
        } else if (!this.attributes.equals(other.attributes)) {
            return false;
        }
        if (!this.symbolicName.equals(other.symbolicName)) {
            return false;
        }
        if (!this.singleton == other.singleton) {
            return false;
        }
        if (this.fragmentAttachment == null) {
            if (other.fragmentAttachment != null) {
                return false;
            }
        } else if (!this.fragmentAttachment.equals(other.fragmentAttachment)) {
            return false;
        }
        return true;
    }

    @Override public int hashCode() {
        return this.symbolicName.hashCode();
    }

}
