/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.manifest;

import org.osgi.framework.Version;

/**
 * Parses the <code>String</code> specification a range of {@link Version Versions} as defined in 3.2.5 of the OSGi
 * Service Server Core Specification.<p/>
 * 
 * The <code>VersionRange</code> can be queried to see if it includes a particular {@link Version} using
 * {@link #includes(Version)}.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * Implementation is immutable.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
public final class VersionRange {

    private static final Version MAX_VERSION = new Version(Integer.MAX_VALUE, Integer.MAX_VALUE, Integer.MAX_VALUE);

    private static final Version ZERO_VERSION = Version.emptyVersion;

    private static final char INCLUSIVE_LOWER = '[';

    private static final char INCLUSIVE_UPPER = ']';

    private static final char EXCLUSIVE_LOWER = '(';

    private static final char EXCLUSIVE_UPPER = ')';

    public static final VersionRange NATURAL_NUMBER_RANGE = new VersionRange(null);

    private final Version floor;

    private final Version ceiling;

    private final boolean floorInclusive;

    private final boolean ceilingInclusive;

    /**
     * Creates a <code>VersionRange</code> for the provided specification.
     * 
     * @param versionRange the <code>VersionRange</code> specification.
     */
    public VersionRange(String versionRange) {
        if (versionRange == null || versionRange.length() == 0) {
            this.floor = ZERO_VERSION;
            this.ceiling = MAX_VERSION;
            this.floorInclusive = true;
            this.ceilingInclusive = false;
            return;
        }

        char first = versionRange.charAt(0);
        if (first == INCLUSIVE_LOWER || first == EXCLUSIVE_LOWER) {
            char last = versionRange.charAt(versionRange.length() - 1);
            if (last == INCLUSIVE_UPPER || last == EXCLUSIVE_UPPER) {
                int comma = versionRange.indexOf(',');
                if (comma < 0) {
                    throw new IllegalArgumentException("Version range '" + versionRange + "' is invalid.");
                }
                this.floor = Version.parseVersion(versionRange.substring(1, comma).trim());
                this.floorInclusive = first == INCLUSIVE_LOWER;
                this.ceiling = Version.parseVersion(versionRange.substring(comma + 1, versionRange.length() - 1).trim());
                this.ceilingInclusive = last == INCLUSIVE_UPPER;
            } else {
                throw new IllegalArgumentException("Version range '" + versionRange + "' is invalid.");
            }
        } else {
            this.floor = Version.parseVersion(versionRange);
            this.floorInclusive = true;
            this.ceiling = MAX_VERSION;
            this.ceilingInclusive = false;
        }
    }

    /**
     * Creates a <code>VersionRange</code> encompassing all the natural numbers: <code>[0.0.0, )</code>.
     * 
     * @return a <code>VersionRange</code> encompassing all the natural numbers.
     */
    public static VersionRange naturalNumberRange() {
        return NATURAL_NUMBER_RANGE;
    }

    /**
     * Gets the floor of this <code>VersionRange</code>.
     * 
     * @return floor of this <code>VersionRange</code>.
     */
    public Version getFloor() {
        return this.floor;
    }

    /**
     * Gets the ceiling of this <code>VersionRange</code>.
     * 
     * @return ceiling of this <code>VersionRange</code>.
     */
    public Version getCeiling() {
        return this.ceiling;
    }

    /**
     * Indicates whether or not the floor of this <code>VersionRange</code> is inclusive.
     * 
     * @return <code>true</code> if the floor is inclusive; otherwise <code>false</code>.
     */
    public boolean isFloorInclusive() {
        return this.floorInclusive;
    }

    /**
     * Indicates whether or not the ceiling of this <code>VersionRange</code> is inclusive.
     * 
     * @return <code>true</code> if the ceiling is inclusive; otherwise <code>false</code>.
     */
    public boolean isCeilingInclusive() {
        return this.ceilingInclusive;
    }

    /**
     * Queries whether this <code>VersionRange</code> includes the supplied {@link Version}.
     * 
     * @param version the <code>Version</code> to check against.
     * @return <code>true</code> if the <code>Version</code> is included in this <code>VersionRange</code>;
     *         otherwise <code>false</code>.
     */
    public boolean includes(Version version) {
        int minCheck = this.floorInclusive ? 0 : 1;
        int maxCheck = this.ceilingInclusive ? 0 : -1;
        if (this.floor == null || this.ceiling == null) {
            throw new RuntimeException("ff");
        }
        return version.compareTo(this.floor) >= minCheck && version.compareTo(this.ceiling) <= maxCheck;
    }

    /**
     * {@inheritDoc}
     */
    @Override public boolean equals(Object object) {
        if (!(object instanceof VersionRange)) {
            return false;
        }
        VersionRange that = (VersionRange) object;
        return this.floorInclusive == that.floorInclusive && this.ceilingInclusive == that.ceilingInclusive && this.floor.equals(that.floor)
            && this.ceiling.equals(that.ceiling);
    }

    /**
     * {@inheritDoc}
     */
    @Override public int hashCode() {
        int result = 17;
        result = 37 * result + this.floor.hashCode();
        result = 37 * result + this.ceiling.hashCode();
        return result;
    }

    /**
     * {@inheritDoc}
     */
    @Override public String toString() {
        return new StringBuilder().append(this.floorInclusive ? INCLUSIVE_LOWER : EXCLUSIVE_LOWER).append(this.floor).append(", ").append(
            this.ceiling == MAX_VERSION ? "" : this.ceiling).append(this.ceilingInclusive ? INCLUSIVE_UPPER : EXCLUSIVE_UPPER).toString();
    }

    /**
     * Creates a <code>String</code> representation of this <code>VersionRange</code> that can be re-parsed.
     */
    public String toParseString() {
        if (this.ceiling == MAX_VERSION && this.floorInclusive) {
            return this.floor.toString();
        } else {
            return new StringBuilder().append(this.floorInclusive ? INCLUSIVE_LOWER : EXCLUSIVE_LOWER).append(this.floor).append(", ").append(
                this.ceiling).append(this.ceilingInclusive ? INCLUSIVE_UPPER : EXCLUSIVE_UPPER).toString();
        }
    }

    /**
     * Returns a <code>VersionRange</code> that is the intersection of the two supplied <code>VersionRanges</code>.
     * <code>null</code> is returned if the two ranges do not intersect.
     * 
     * @param rangeOne The first <code>VersionRange</code> for the intersection
     * @param rangeTwo The second <code>VersionRange</code> for the intersection
     * @return The intersection of the two <code>VersionRanges</code> or <code>null</code> if the ranges do not
     *         intersect
     */
    public static VersionRange intersection(VersionRange rangeOne, VersionRange rangeTwo) {

        Version floor;
        boolean floorInclusive;

        Version ceiling;
        boolean ceilingInclusive;

        int floorComparison = rangeOne.floor.compareTo(rangeTwo.floor);
        if (floorComparison < 0) {
            floor = rangeTwo.floor;
            floorInclusive = rangeTwo.floorInclusive;
        } else if (floorComparison > 0) {
            floor = rangeOne.floor;
            floorInclusive = rangeOne.floorInclusive;
        } else {
            floor = rangeOne.floor;
            floorInclusive = rangeOne.floorInclusive && rangeTwo.floorInclusive;
        }

        int ceilingComparison = rangeOne.ceiling.compareTo(rangeTwo.ceiling);
        if (ceilingComparison > 0) {
            ceiling = rangeTwo.ceiling;
            ceilingInclusive = rangeTwo.ceilingInclusive;
        } else if (ceilingComparison < 0) {
            ceiling = rangeOne.ceiling;
            ceilingInclusive = rangeOne.ceilingInclusive;
        } else {
            ceiling = rangeOne.ceiling;
            ceilingInclusive = rangeOne.ceilingInclusive && rangeTwo.ceilingInclusive;
        }

        int floorCeilingComparison = floor.compareTo(ceiling);
        if (floorCeilingComparison < 0 || floorCeilingComparison == 0 && floorInclusive && ceilingInclusive) {
            if (ceiling.equals(MAX_VERSION)) {
                return new VersionRange(floor.toString());
            } else {
                return new VersionRange((floorInclusive ? "[" : "(") + floor + "," + ceiling + (ceilingInclusive ? "]" : ")"));
            }
        }
        return null;
    }
}
