/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.manifest.parse;

import java.util.List;

import org.antlr.runtime.ANTLRStringStream;
import org.antlr.runtime.CommonTokenStream;
import org.antlr.runtime.RecognitionException;

/**
 * Implementation of {@link HeaderParser} that uses an ANTLR generated parser.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * This class is not thread safe.
 * 
 * @author Rob Harrop
 * @author Glyn Normington
 * @author Andy Wilkinson
 * @author Christopher Frost
 * @see OsgiHeaderParser
 * @since 1.0
 */
final class AntlrHeaderParser implements HeaderParser {

    private final ParserLogger logger;
    
    private final ManifestHeaderParserErrorHandler manifestParserErrorHandler;

    public AntlrHeaderParser(ParserLogger logger) {
        this.logger = logger;
        this.manifestParserErrorHandler = new ManifestHeaderParserErrorHandler(this.logger);
    }
    
    
    /**
     * Creates a new {@link OsgiHeaderParser} to parse the supplied header text.
     * 
     * @param header the header text to parse.
     * @return the new <code>OsgiHeaderParser</code>.
     */
    private StandardOsgiHeaderParser createParser(String header, String headerName) {
        ANTLRStringStream stream = new ANTLRStringStream(header);
        StandardOsgiHeaderLexer lexer = new StandardOsgiHeaderLexer(this.manifestParserErrorHandler, stream, headerName, header);
        CommonTokenStream tokenStream = new CommonTokenStream(lexer);
        StandardOsgiHeaderParser parser = new StandardOsgiHeaderParser(this.manifestParserErrorHandler, tokenStream, headerName, header);
        return parser;
    }

    /**
     * {@inheritDoc}
     */
    public List<HeaderDeclaration> parsePackageHeader(String header, String headerName) {
        StandardOsgiHeaderParser parser = createParser(header, headerName);
        try {
            parser.header();
        } catch (RecognitionException e) {
            String msg = "Error parsing bundle manifest, " + headerName + " at Char: " + e.charPositionInLine + " in [" + header + "]";
            this.logger.outputErrorMsg(e, msg);
            throw new BundleManifestParseException(msg, e);
        }
        return parser.getDeclarations();
    }

    /**
     * {@inheritDoc}
     */
    public List<HeaderDeclaration> parseDynamicImportPackageHeader(String header) {
        String headerName = "DynamicImport-Package";
        StandardOsgiHeaderParser parser = createParser(header, headerName);
        try {
            parser.dynHeader();
        } catch (RecognitionException e) {
            String msg = "Error parsing bundle manifest, " + headerName + " at Char: " + e.charPositionInLine + " in [" + header + "]";
            this.logger.outputErrorMsg(e, msg);
            throw new BundleManifestParseException(msg, e);
        }
        return parser.getDeclarations();
    }

    /**
     * {@inheritDoc}
     */
    public List<HeaderDeclaration> parseRequireBundleHeader(String header) {
        String headerName = "Require-Bundle";
        StandardOsgiHeaderParser parser = createParser(header, headerName);
        try {
            parser.requireBundleHeader();
        } catch (RecognitionException e) {
            String msg = "Error parsing bundle manifest, " + headerName + " at Char: " + e.charPositionInLine + " in [" + header + "]";
            this.logger.outputErrorMsg(e, msg);
            throw new BundleManifestParseException(msg, e);
        }
        return parser.getDeclarations();
    }

    /**
     * {@inheritDoc}
     */
    public HeaderDeclaration parseFragmentHostHeader(String header) {
        String headerName = "FragmentHost";
        StandardOsgiHeaderParser parser = createParser(header, headerName);
        try {
            parser.fragmentHostHeader();
        } catch (RecognitionException e) {
            String msg = "Error parsing bundle manifest, " + headerName + " at Char: " + e.charPositionInLine + " in [" + header + "]";
            this.logger.outputErrorMsg(e, msg);
            throw new BundleManifestParseException(msg, e);
        }
        return parser.getDeclarations().get(0);
    }

    /**
     * {@inheritDoc}
     */
    public HeaderDeclaration parseBundleSymbolicName(String header) {
        String headerName = "Bundle-SymbolicName";
        StandardOsgiHeaderParser parser = createParser(header, headerName);
        try {
            parser.bundleSymbolicNameHeader();
        } catch (RecognitionException e) {
            String msg = "Error parsing bundle manifest, " + headerName + " at Char: " + e.charPositionInLine + " in [" + header + "]";
            this.logger.outputErrorMsg(e, msg);
            throw new BundleManifestParseException(msg, e);
        }
        return parser.getDeclarations().get(0);
    }
    
    /**
     * {@inheritDoc}
     */
    public List<HeaderDeclaration> parseImportBundleHeader(String header) {
        String headerName = "Import-Bundle";
        StandardOsgiHeaderParser parser = createParser(header, headerName);
        try {
            parser.importBundleHeader();
        } catch (RecognitionException e) {
            String msg = "Error parsing bundle manifest, " + headerName + " at Char: " + e.charPositionInLine + " in [" + header + "]";
            this.logger.outputErrorMsg(e, msg);
            throw new BundleManifestParseException(msg, e);
        }
        return parser.getDeclarations();
    }
    
    /**
     * {@inheritDoc}
     */
    public List<HeaderDeclaration> parseImportLibraryHeader(String header) {
        String headerName = "Import-Library";
        StandardOsgiHeaderParser parser = createParser(header, headerName);
        try {
            parser.importLibraryHeader();
        } catch (RecognitionException e) {
            String msg = "Error parsing bundle manifest, " + headerName + " at Char: " + e.charPositionInLine + " in [" + header + "]";
            this.logger.outputErrorMsg(e, msg);
            throw new BundleManifestParseException(msg, e);
        }
        return parser.getDeclarations();
    }
    
    /**
     * {@inheritDoc}
     */
    public HeaderDeclaration parseLibrarySymbolicName(String header) {
        String headerName = "Library-SymbolicName";
        StandardOsgiHeaderParser parser = createParser(header, headerName);
        try {
            parser.librarySymbolicNameHeader();
        } catch (RecognitionException e) {
            String msg = "Error parsing bundle manifest, " + headerName + " at Char: " + e.charPositionInLine + " in [" + header + "]";
            this.logger.outputErrorMsg(e, msg);
            throw new BundleManifestParseException(msg, e);
        }
        return parser.getDeclarations().get(0);
    }

    /** 
     * {@inheritDoc}
     */
    public List<HeaderDeclaration> parseWebContextParamsHeader(String header) {
        String headerName = "Web-ContextParams";
        StandardOsgiHeaderParser parser = createParser(header, headerName);
        try {
            parser.webContextParamsHeader();
        } catch (RecognitionException e) {
            String msg = "Error parsing bundle manifest, " + headerName + " at Char: " + e.charPositionInLine + " in [" + header + "]";
            this.logger.outputErrorMsg(e, msg);
            throw new BundleManifestParseException(msg, e);
        }
        return parser.getDeclarations();
    }

    /** 
     * {@inheritDoc}
     */
    public List<HeaderDeclaration> parseWebFilterMappingsHeader(String header) {
        String headerName = "Web-FilterMappings";
        StandardOsgiHeaderParser parser = createParser(header, headerName);
        try {
            parser.webFilterMappingsHeader();
        } catch (RecognitionException e) {
            String msg = "Error parsing bundle manifest, " + headerName + " at Char: " + e.charPositionInLine + " in [" + header + "]";
            this.logger.outputErrorMsg(e, msg);
            throw new BundleManifestParseException(msg, e);
        }
        return parser.getDeclarations();
    }

    /** 
     * {@inheritDoc}
     */
    public List<HeaderDeclaration> parseWebServletsHeader(String header) {
        String headerName = "Web-Servlets";
        StandardOsgiHeaderParser parser = createParser(header, headerName);
        try {
            parser.webServletsHeader();
        } catch (RecognitionException e) {
            String msg = "Error parsing bundle manifest, " + headerName + " at Char: " + e.charPositionInLine + " in [" + header + "]";
            this.logger.outputErrorMsg(e, msg);
            throw new BundleManifestParseException(msg, e);
        }
        return parser.getDeclarations();
    }
    
    /** 
     * {@inheritDoc}
     */
    public List<HeaderDeclaration> parseWebErrorPagesHeader(String header) {
        String headerName = "Web-ErrorPages";
        StandardOsgiHeaderParser parser = createParser(header, headerName);
        try {
            parser.webErrorPagesHeader();
        } catch (RecognitionException e) {
            String msg = "Error parsing bundle manifest, " + headerName + " at Char: " + e.charPositionInLine + " in [" + header + "]";
            this.logger.outputErrorMsg(e, msg);
            throw new BundleManifestParseException(msg, e);
        }
        return parser.getDeclarations();
    }
    
    public HeaderDeclaration parseBundleActivationPolicy(String header) {
        String headerName = "Bundle-ActivationPolicy";
        StandardOsgiHeaderParser parser = createParser(header, headerName);
        try {
            parser.bundleActivationPolicyHeader();
        } catch (RecognitionException e) {
            String msg = "Error parsing bundle manifest, " + headerName + " at Char: " + e.charPositionInLine + " in [" + header + "]";
            this.logger.outputErrorMsg(e, msg);
            throw new BundleManifestParseException(msg, e);
        }
        return parser.getDeclarations().get(0);
    }
}
