/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.manifest.parse;

import java.util.ArrayList;
import java.util.List;

import org.antlr.runtime.EarlyExitException;
import org.antlr.runtime.FailedPredicateException;
import org.antlr.runtime.MismatchedNotSetException;
import org.antlr.runtime.MismatchedRangeException;
import org.antlr.runtime.MismatchedSetException;
import org.antlr.runtime.MismatchedTokenException;
import org.antlr.runtime.MismatchedTreeNodeException;
import org.antlr.runtime.NoViableAltException;
import org.antlr.runtime.RecognitionException;

/**
 * <p>
 * This provides some of the required logging functionality based 
 * on its knowledge of the Antlr parser. It still dosen't know what 
 * to actually do with log messages which is delegated to sub 
 * classes through the messageProcessed method. 
 * </p>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe
 * 
 * @author Christopher Frost
 * @since 1.0
 */
abstract public class AntlrLogMessageService implements ParserLogger {

    private final String subsystem;

    private volatile boolean used = false;

    private final List<String> messages = new ArrayList<String>();

    private static final int PARSER_ERROR_NUMBER_RANGE_START = 1000;

    private static final int ERROR_OFFSET = 0;

    private static final int WARN_OFFSET = 25;

    private static final int INFO_OFFSET = 50;

    /**
     * Constructor, takes the subsystem code for the subsystem that is using the parser
     * 
     * @param subSystem
     */
    protected AntlrLogMessageService(String subSystem) {
        this.subsystem = subSystem;
    }

    /**
     * This is where sub-classes can decide what to do with the caused exception
     * 
     * {@inheritDoc}
     */
    public abstract void messageProcessed(String eventCode, String message);

    /**
     * An unidentified parser related exception has happened causing an error
     * 
     * {@inheritDoc}
     */
    public void outputErrorMsg(Exception re, String item) {
        this.used = true;
        processEvent(re, "E", ERROR_OFFSET, item);
    }

    /**
     * An unidentified parser related exception has happened causing a warning
     * 
     * {@inheritDoc}
     */
    public void outputWarnMsg(Exception re, String item) {
        processEvent(re, "W", WARN_OFFSET, item);
    }

    /**
     * An unidentified parser related exception has happened causing a warning
     * 
     * {@inheritDoc}
     */
    public void outputInfoMsg(Exception re, String item) {
        processEvent(re, "I", INFO_OFFSET, item);
    }

    /**
     * 
     * {@inheritDoc}
     */
    public String[] errorReports() {
        if (this.used) {
            return this.messages.toArray(new String[this.messages.size()]);
        } else {
            return null;
        }
    }

    /**
     * ReSet the parser so that it can be reused and
     * 
     * {@inheritDoc}
     */
    public void resetErrorReport() {
        this.used = false;
    }

    /**
     * 
     * @param re
     * @param level
     * @param offset
     */
    private void processEvent(Exception re, String level, int offset, String itemMsg) {
        this.messages.add(itemMsg);
        messageProcessed(String.format("%1$s%2$d%3$s", this.subsystem, getErrorEventNumber(re) + offset, level), itemMsg);
    }

    /**
     * This method provides the main service of this class, that is to take a parser exception and map it to the correct
     * log event number as in the LogResource.properties files
     * 
     * @param re exception thrown by the Antlr Parser
     * @return The number of the log event mapped from the known Antler events
     */
    private int getErrorEventNumber(Exception re) {

        if (re instanceof MismatchedTokenException) {
            return PARSER_ERROR_NUMBER_RANGE_START + 2;
        } else if (re instanceof MismatchedTreeNodeException) {
            return PARSER_ERROR_NUMBER_RANGE_START + 3;
        } else if (re instanceof NoViableAltException) {
            return PARSER_ERROR_NUMBER_RANGE_START + 4;
        } else if (re instanceof EarlyExitException) {
            return PARSER_ERROR_NUMBER_RANGE_START + 5;
        } else if (re instanceof FailedPredicateException) {
            return PARSER_ERROR_NUMBER_RANGE_START + 6;
        } else if (re instanceof MismatchedRangeException) {
            return PARSER_ERROR_NUMBER_RANGE_START + 7;
        } else if (re instanceof MismatchedNotSetException) {// Need to test this way round as
                                                                // MismatchedNotSetException extends
                                                                // MismatchedSetException
            return PARSER_ERROR_NUMBER_RANGE_START + 9;
        } else if (re instanceof MismatchedSetException) {
            return PARSER_ERROR_NUMBER_RANGE_START + 8;
        } else if (re instanceof RecognitionException) {
            return PARSER_ERROR_NUMBER_RANGE_START + 1;
        } else {
            return PARSER_ERROR_NUMBER_RANGE_START + 0;
        }
    }

}
