/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.manifest.parse;

import java.util.List;

/**
 * Strategy for parsing OSGi manifest headers.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * Implementations need not be threadsafe.
 * 
 * @author Rob Harrop
 * @author Glyn Normington
 * @since 1.0
 */
public interface HeaderParser {

    /**
     * Parses the supplied import/export package header text and returns the list of corresponding
     * {@link HeaderDeclaration HeaderDeclarations}.
     * 
     * @param header the header text to parse.
     * @return the <code>HeaderDeclarations</code>.
     */
    List<HeaderDeclaration> parsePackageHeader(String header, String headerType);

    /**
     * Parses the supplied dynamic import header text and returns the list of corresponding
     * {@link HeaderDeclaration HeaderDeclarations}. Dynamic imports allow wildcarded package names.
     * 
     * @param header the header text to parse.
     * @return the <code>HeaderDeclarations</code>.
     */
    List<HeaderDeclaration> parseDynamicImportPackageHeader(String header);

    /**
     * Parses the supplied require bundle header text and returns the list of corresponding
     * {@link HeaderDeclaration HeaderDeclarations}.
     * 
     * @param header the header text to parse.
     * @return the <code>HeaderDeclarations</code>.
     */
    List<HeaderDeclaration> parseRequireBundleHeader(String header);

    /**
     * Parses the supplied fragment host header text and returns the corresponding
     * {@link HeaderDeclaration HeaderDeclaration}.
     * 
     * @param header the header text to parse.
     * @return the <code>HeaderDeclaration</code>.
     * 
     */
    HeaderDeclaration parseFragmentHostHeader(String header);

    /**
     * Parses the supplied bundle symbolic name header text and returns the corresponding
     * {@link HeaderDeclaration HeaderDeclaration}.
     * 
     * @param header the header text to parse.
     * @return the <code>HeaderDeclaration</code> for the Bundle-SymbolicName.
     */
    HeaderDeclaration parseBundleSymbolicName(String header);

    /**
     * Parses the supplied import bundle header text and returns the list of corresponding
     * {@link HeaderDeclaration HeaderDeclarations}.
     * 
     * @param header the header text to parse.
     * @return the <code>HeaderDeclarations</code>.
     */
    List<HeaderDeclaration> parseImportBundleHeader(String header);

    /**
     * Parses the supplied Import-Library header text and returns the list of corresponding
     * {@link HeaderDeclaration HeaderDeclarations}.
     * 
     * @param header the header text to parse.
     * @return the <code>HeaderDeclarations</code>.
     */
    List<HeaderDeclaration> parseImportLibraryHeader(String header);

    /**
     * Parses the supplied Library-SymbolicName header text and returns the list of corresponding
     * {@link HeaderDeclaration HeaderDeclarations}.
     * 
     * @param header the header text to parse.
     * @return the <code>HeaderDeclarations</code>.
     */
    HeaderDeclaration parseLibrarySymbolicName(String header);
    
    /**
     * Parses the supplied <code>Web-ContextParams</code> header text and returns the list
     * of corresponding {@link HeaderDeclaration HeaderDeclarations}.
     */
    List<HeaderDeclaration> parseWebContextParamsHeader(String header);
    
    /**
     * Parses the supplied <code>Web-FilterMappings</code> header text and returns the list
     * of corresponding {@link HeaderDeclaration HeaderDeclarations}.
     */
    List<HeaderDeclaration> parseWebFilterMappingsHeader(String header);
    
    /**
     * Parses the supplied <code>Web-Servlets</code> header text and returns the list
     * of corresponding {@link HeaderDeclaration HeaderDeclarations}.
     */
    List<HeaderDeclaration> parseWebServletsHeader(String header);
    
    /**
     * Parses the supplied <code>Web-ErrorPages</code> header text and returns the list
     * of corresponding {@link HeaderDeclaration HeaderDeclarations}.
     */
    List<HeaderDeclaration> parseWebErrorPagesHeader(String header);
    
    /**
     * Parses the supplied <code>Bundle-ActivationPolicy</code> header text and returns the
     * list of corresponding {@link HeaderDeclaration}.
     */
    HeaderDeclaration parseBundleActivationPolicy(String header);
}
