/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.provisioning.repository;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintWriter;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import org.osgi.framework.Version;

import com.springsource.server.osgi.manifest.VersionRange;
import com.springsource.server.osgi.provisioning.Definition;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * Maps some pair (K x Version) to <code>T</code>.
 * 
 * Locations for a given pair, (K x Version), are maintained in insertion order.
 */
abstract class CachingLocationTable<T extends Definition> {

    private static final long LAST_MODIFIED_PRECISION = 1000;

    private final Tracer tracer = TracerFactory.createTracer(getClass());

    /**
     * Can the same (K x Version) pair be mapped to multiple definitions?
     */
    private final boolean allowDuplicates;

    /**
     * b : (String x V) <-> T
     * 
     * map : String +> (Version +> P T)
     */
    private final Map<String, Map<Version, Set<T>>> map = new HashMap<String, Map<Version, Set<T>>>();

    private boolean cacheValid;

    private final long repositoryAge;

    private final File cacheFile;

    /**
     * @param allowDuplicates
     * @param cacheFile The location of the table's cache
     */
    public CachingLocationTable(boolean allowDuplicates, File cacheFile, long repositoryAge) {

        this.allowDuplicates = allowDuplicates;
        this.cacheFile = cacheFile;
        this.repositoryAge = repositoryAge;

        if (cacheFile.exists()) {
            this.cacheValid = populateTableFromCacheFile(this.cacheFile, this.repositoryAge);
        } else {
            this.cacheValid = false;
        }
    }

    /**
     * Record the supplied definition against the supplied K, Version pair.
     */
    public void recordDefinition(String key, Version version, T definition) throws DuplicateDefinitionException {
        recordDefinitionInternal(key, version, definition);
    }

    private void recordDefinitionInternal(String key, Version version, T definition) throws DuplicateDefinitionException {
        Map<Version, Set<T>> versionDefinitions = versionDefinitionsByKey(key);
        Set<T> definitions = versionDefinitions.get(version);
        if (definitions == null) {
            definitions = new LinkedHashSet<T>();
            versionDefinitions.put(version, definitions);
        } else if (!this.allowDuplicates && !definitions.isEmpty()) {
            T current = definitions.iterator().next();
            if (!current.equals(definition)) {
                if (!current.getLocation().isDirectory() && !definition.getLocation().isDirectory()) {
                    byte[] hash1 = getHash(current);
                    byte[] hash2 = getHash(definition);
                    if(Arrays.equals(hash1, hash2)) {
                        return;
                    }
                }
                throw new DuplicateDefinitionException(current);
            }
        }
        definitions.add(definition);
    }

    /**
     * Gets <code>(Version +> P T)</code> for the supplied <code>K</code>, that is, the set of definitions for each
     * <code>Version</code> of <code>K</code>.
     */
    private Map<Version, Set<T>> versionDefinitionsByKey(String key) {
        Map<Version, Set<T>> versionDefinitions = this.map.get(key);
        if (versionDefinitions == null) {
            versionDefinitions = new HashMap<Version, Set<T>>();
            this.map.put(key, versionDefinitions);
        }
        return versionDefinitions;
    }

    private byte[] getHash(T definition) {
        MessageDigest messageDigest;
        try {
            messageDigest = MessageDigest.getInstance("SHA-1");
        } catch (NoSuchAlgorithmException e) {
            throw new RuntimeException("Could not get a SHA-1 message digest to compute hash", e);
        }

        InputStream in = null;
        try {
            int length;
            byte[] b = new byte[4096];
            in = new FileInputStream(definition.getLocation());
            while ((length = in.read(b)) != -1) {
                messageDigest.update(b, 0, length);
            }
        } catch (FileNotFoundException e) {
            throw new RuntimeException("Unable to calculate hash of file", e);
        } catch (IOException e) {
            throw new RuntimeException("Unable to calculate hash of file", e);
        } finally {
            if (in != null) {
                try {
                    in.close();
                } catch (IOException e) {
                }
            }
        }

        return messageDigest.digest();
    }

    /**
     * Finds the definitions for <code>key</code> that are in the given <code>VersionRange</code>.
     */
    Set<? extends T> findKeyedDefinitionsInRange(String key, VersionRange versionRange) {
        Map<Version, Set<T>> versionDefinitions = versionDefinitionsByKey(key);
        Set<T> definitions = new HashSet<T>();
        if (!allowDuplicates) {
            Version highestVersion = null;
            Set<T> matches = null;
            for (Map.Entry<Version, Set<T>> entry : versionDefinitions.entrySet()) {
                if (versionRange.includes(entry.getKey())) {
                    if (highestVersion == null || entry.getKey().compareTo(highestVersion) > 0) {
                        matches = entry.getValue();
                        highestVersion = entry.getKey();
                    }
                }
            }
            if (matches != null) {
                definitions.addAll(matches);
            }
        } else {
            for (Map.Entry<Version, Set<T>> entry : versionDefinitions.entrySet()) {
                if (versionRange.includes(entry.getKey())) {
                    definitions.addAll(entry.getValue());
                }
            }
        }
        return definitions;
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        Set<Entry<String, Map<Version, Set<T>>>> entrySet = this.map.entrySet();
        for (Entry<String, Map<Version, Set<T>>> entry : entrySet) {
            sb.append(entry.getKey()).append('\n');
            for (Entry<Version, Set<T>> e : entry.getValue().entrySet()) {
                sb.append('\t').append(e.getKey()).append('\n');
                for (T definition : e.getValue()) {
                    sb.append("\t\t").append(definition).append("\n");
                }
            }
        }
        return sb.toString();
    }

    private boolean populateTableFromCacheFile(File cacheFile, long repositoryAge) {
        if (tracer.isInfoEnabled()) {
            tracer.traceInfo(String.format("Checking cache file '%s'", cacheFile));
        }
        try {
            BufferedReader reader = new BufferedReader(new FileReader(cacheFile));
            String line = reader.readLine();

            long openingCacheAge;

            if (line != null) {
                try {
                    openingCacheAge = Long.parseLong(line);

                    if (openingCacheAge != repositoryAge || Math.abs(repositoryAge - System.currentTimeMillis()) < LAST_MODIFIED_PRECISION) {
                        if (tracer.isInfoEnabled()) {
                            tracer.traceInfo(String.format("Cache file out of date '%s'", cacheFile));
                        }
                        cacheFile.delete();
                        return false;
                    }
                } catch (NumberFormatException nfe) {
                    cacheFile.delete();
                    return false;
                }
            } else {
                handleCorruptCache();
                return false;
            }

            line = reader.readLine();

            while (line != null) {
                String nextLine = reader.readLine();

                if (nextLine == null) {
                    try {
                        long closingCacheAge = Long.parseLong(line);

                        if (closingCacheAge != repositoryAge) {
                            handleCorruptCache();
                            return false;
                        }
                    } catch (NumberFormatException nfe) {
                        handleCorruptCache();
                        return false;
                    }
                } else {
                    String[] entryComponents = line.split("\\|");
                    if (entryComponents.length == 3) {
                        String key = entryComponents[0];
                        Version version = new Version(entryComponents[1]);
                        String path = entryComponents[2];
                        T bd = fromLocation(path);
                        if (bd != null) {
                            try {
                                recordDefinitionInternal(key, version, bd);
                            } catch (DuplicateDefinitionException dde) {
                                handleCorruptCache();
                            }
                        }
                    } else {
                        handleCorruptCache();
                    }
                }

                line = nextLine;
            }
        } catch (IOException ioe) {
            handleCorruptCache();
            return false;
        }
        if (tracer.isInfoEnabled()) {
            tracer.traceInfo(String.format("Restored from cache file '%s'", cacheFile));
        }
        return true;
    }

    private void handleCorruptCache() {
        if (tracer.isInfoEnabled()) {
            tracer.traceInfo(String.format("Cache corrupted - deleting '%s'", cacheFile));
        }
        this.cacheFile.delete();
        this.cacheValid = false;
    }

    private void addLocationToCache(String key, Version version, T definition, PrintWriter cacheWriter) {
        StringBuilder cacheEntryBuilder = new StringBuilder(key.toString());
        cacheEntryBuilder.append("|");
        cacheEntryBuilder.append(version.toString());
        cacheEntryBuilder.append("|");
        cacheEntryBuilder.append(toLocation(definition));
        cacheWriter.println(cacheEntryBuilder.toString());
        cacheWriter.flush();
    }

    protected abstract String toLocation(T definition);

    protected abstract T fromLocation(String location);

    public boolean isCacheValid() {
        return this.cacheValid;
    }

    public void updateCache() {
        try {
            PrintWriter cacheWriter = new PrintWriter(new FileWriter(this.cacheFile));

            cacheWriter.println(this.repositoryAge);

            Set<String> keys = this.map.keySet();

            for (String key : keys) {
                Map<Version, Set<T>> versionAndDefinitions = this.map.get(key);
                Set<Version> versions = versionAndDefinitions.keySet();
                for (Version version : versions) {
                    Set<T> definitions = versionAndDefinitions.get(version);
                    for (T definition : definitions) {
                        addLocationToCache(key, version, definition, cacheWriter);
                    }
                }
            }
            cacheWriter.println(this.repositoryAge);
            cacheWriter.flush();
            cacheWriter.close();

            this.cacheValid = true;
        } catch (IOException ioe) {
            throw new RuntimeException("An update to cache file " + this.cacheFile.getAbsolutePath() + " has failed.", ioe);
        }
    }

    public Set<T> getDefinitions() {
        Set<T> allDefinitions = new HashSet<T>();
        for (Map<Version, Set<T>> definitionsByVersion : map.values()) {
            for (Set<T> definitions : definitionsByVersion.values()) {
                allDefinitions.addAll(definitions);
            }
        }
        return allDefinitions;
    }
}
