/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.provisioning.repository;

import java.io.File;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.Iterator;
import java.util.Set;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.osgi.framework.Version;
import org.springframework.util.StringUtils;

import com.springsource.server.osgi.manifest.BundleManifest;
import com.springsource.server.osgi.manifest.VersionRange;
import com.springsource.server.osgi.manifest.BundleManifest.BundleSymbolicName;
import com.springsource.server.osgi.manifest.BundleManifest.ExportPackageHeader;
import com.springsource.server.osgi.manifest.BundleManifest.FragmentHostHeader;
import com.springsource.server.osgi.manifest.BundleManifest.ExportPackageHeader.PackageExport;
import com.springsource.server.osgi.provisioning.BundleDefinition;
import com.springsource.server.osgi.provisioning.BundleProvisioningException;
import com.springsource.server.osgi.provisioning.LibraryDefinition;
import com.springsource.server.osgi.provisioning.OsgiProvisioningLogEvent;
import com.springsource.server.osgi.provisioning.support.BundleDefinitionFactory;
import com.springsource.server.osgi.provisioning.support.LibraryDefinitionFactory;
import com.springsource.server.serviceability.logging.LogService;
import com.springsource.server.serviceability.logging.LogServiceFactory;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * Standard implementation of {@link BundleRepository} that scans directories for
 * {@link BundleDefinition BundleDefinitions} and {@link LibraryDefinition LibraryDefinitions}.<p/>
 * 
 * Directory scanning happens in a lazy fashion when the first <code>BundleDefinition</code> is requested.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * Implementation is thread safe.
 * 
 * @author Rob Harrop
 * @author Andy Wilkinson
 * @author Christopher Frost
 * @since 1.0
 */
class FileSystemBundleRepository implements BundleRepository {

    private static final Tracer TRACE = TracerFactory.createTracer(FileSystemBundleRepository.class);

    private static final LogService LOGGER = LogServiceFactory.getLogService();

    private final ReadWriteLock readWriteLock = new ReentrantReadWriteLock();

    private final BundleDefinitionTable definitionsBySymbolicName;

    private final BundleDefinitionTable definitionsByExportedPackage;

    private final BundleDefinitionTable definitionsByFragmentHost;

    private final LibraryDefinitionTable librariesBySymbolicName;

    private final FileSystemSearcher searcher;

    private volatile boolean initialised = false;
    
    private static final Pattern PROPERTY_PATTERN = Pattern.compile("(\\$\\{(([^\\}]+))\\})");

    /**
     * Creates a new <code>StandardBundleLocationRegistry</code> that scans the supplied the <code>searchPath</code>.
     * 
     * @param searchPath the path to scan.
     * @param cacheLocationPath the location to with the repository should write its cache
     * @param absoluteRoot An absolute path to be used to make the search path absolute if the supplied search path is
     *        relative
     */
    public FileSystemBundleRepository(String searchPath, String cacheLocationPath, String absoluteRoot) {
        String repositoryId = calculateRepositoryId(searchPath);
        this.searcher = createFileSystemSearcher(searchPath, absoluteRoot);

        long ageOfBundleDefinitionSpace = this.searcher.getAgeOfDefinitionSpace();

        if (!cacheLocationPath.endsWith(File.separator)) {
            cacheLocationPath += File.separator;
        }

        this.definitionsBySymbolicName = new BundleDefinitionTable(false, new File(cacheLocationPath + "BR" + repositoryId + "SN"),
            ageOfBundleDefinitionSpace);

        this.definitionsByExportedPackage = new BundleDefinitionTable(true, new File(cacheLocationPath + "BR" + repositoryId + "EP"),
            ageOfBundleDefinitionSpace);

        this.definitionsByFragmentHost = new BundleDefinitionTable(true, new File(cacheLocationPath + "BR" + repositoryId + "FH"),
            ageOfBundleDefinitionSpace);

        this.librariesBySymbolicName = new LibraryDefinitionTable(new File(cacheLocationPath + "BR" + repositoryId + "LB"),
            ageOfBundleDefinitionSpace);

    }

    /**
     * @param searchPath
     * @return
     */
    private static String calculateRepositoryId(String searchPath) {
        try {
            MessageDigest digest = MessageDigest.getInstance("SHA-1");
            digest.update(searchPath.getBytes());
            byte[] result = digest.digest();
            return HexEncoder.encodeHex(result);
        } catch (NoSuchAlgorithmException e) {
            throw new IllegalStateException("Unable to calculate repository ID for search path '" + searchPath + "'");
        }
    }

    /**
     * {@inheritDoc}
     */
    public BundleDefinition findBySymbolicName(String symbolicName, VersionRange versionRange) {
        if (TRACE.isInfoEnabled()) {
            TRACE.traceInfo(String.format("findBySymbolicName(%s, %s)", symbolicName, versionRange));
        }
        if (!this.definitionsBySymbolicName.isCacheValid()) {
            initialiseIfNecessary();
        }
        Lock readLock = this.readWriteLock.readLock();
        readLock.lock();
        try {
            Set<? extends BundleDefinition> candidateLocations = this.definitionsBySymbolicName.findKeyedDefinitionsInRange(symbolicName, versionRange);
            BundleDefinition found = pickDefinition(candidateLocations);
            if (TRACE.isInfoEnabled()) {
                TRACE.traceInfo(String.format("Found %s for symbolic name %s at version %s", found, symbolicName, versionRange));
            }
            return found;
        } finally {
            readLock.unlock();
        }
    }

    /**
     * {@inheritDoc}
     */
    public Set<? extends BundleDefinition> findByFragmentHost(String bundleSymbolicName, Version version) {
        if (TRACE.isInfoEnabled()) {
            TRACE.traceInfo(String.format("findByFragmentHost(%s, %s)", bundleSymbolicName, version));
        }
        if (!this.definitionsByFragmentHost.isCacheValid()) {
            initialiseIfNecessary();
        }
        Lock readLock = this.readWriteLock.readLock();
        readLock.lock();
        try {
            Set<? extends BundleDefinition> candidateLocations = this.definitionsByFragmentHost.findKeyedDefinitionsInRange(bundleSymbolicName,
                VersionRange.NATURAL_NUMBER_RANGE);
            Iterator<? extends BundleDefinition> iterator = candidateLocations.iterator();
            while (iterator.hasNext()) {
                BundleDefinition bundleDefinition = iterator.next();
                BundleManifest manifest = bundleDefinition.getManifest();
                FragmentHostHeader fragmentHost = manifest.getFragmentHostHeader();
                String bundleVersion = fragmentHost.getAttributes().get("bundle-version");
                if (StringUtils.hasText(bundleVersion)) {
                    VersionRange bundleVersionRange = new VersionRange(bundleVersion);
                    if (!bundleVersionRange.includes(version)) {
                        iterator.remove();
                    }
                }
            }
            return candidateLocations;
        } finally {
            readLock.unlock();
        }
    }

    /**
     * {@inheritDoc}
     */
    public Set<? extends BundleDefinition> findByExportedPackage(String packageName, VersionRange versionRange) {
        if (TRACE.isInfoEnabled()) {
            TRACE.traceInfo(String.format("findByExportedPackage(%s, %s)", packageName, versionRange));
        }
        if (!this.definitionsByExportedPackage.isCacheValid()) {
            initialiseIfNecessary();
        }
        Lock readLock = this.readWriteLock.readLock();
        readLock.lock();
        try {
            Set<? extends BundleDefinition> found = this.definitionsByExportedPackage.findKeyedDefinitionsInRange(packageName, versionRange);
            if (TRACE.isInfoEnabled()) {
                TRACE.traceInfo(String.format("Found %d match(es) for package %s at version %s", found.size(), packageName, versionRange));
            }
            return found;
        } finally {
            readLock.unlock();
        }
    }

    /**
     * {@inheritDoc}
     */
    public LibraryDefinition findLibrary(String libraryName, VersionRange versionRange) {
        if (TRACE.isInfoEnabled()) {
            TRACE.traceInfo(String.format("findLibrary(%s, %s)", libraryName, versionRange));
        }
        if (!this.librariesBySymbolicName.isCacheValid()) {
            initialiseIfNecessary();
        }
        Lock readLock = this.readWriteLock.readLock();
        readLock.lock();
        try {
            Set<? extends LibraryDefinition> candidateDefinitions = this.librariesBySymbolicName.findKeyedDefinitionsInRange(libraryName, versionRange);
            LibraryDefinition found = pickDefinition(candidateDefinitions);
            
            if(found == null){
                candidateDefinitions = this.librariesBySymbolicName.findKeyedDefinitionsInRange(libraryName + ".instrumented", versionRange);
                found = pickDefinition(candidateDefinitions);
                if( found != null){
                    LOGGER.log(OsgiProvisioningLogEvent.ALTERNATE_INSTRUMENTED_LIBRARY_FOUND, libraryName, versionRange.toString());
                }
            }
            
            if (TRACE.isInfoEnabled()) {
                TRACE.traceInfo(String.format("Found library %s for name %s at version %s", found, libraryName, versionRange));
            }
            return found;
        } finally {
            readLock.unlock();
        }
    }

    /**
     * Scans the configured directories if needed.
     */
    void initialiseIfNecessary() {
        Lock writeLock = this.readWriteLock.writeLock();
        writeLock.lock();
        try {
            if (!this.initialised) {
                try {
                    this.searcher.search(new FileSystemSearcher.SearchCallback() {

                        public void found(File file, boolean terminal) {
                            if (terminal) {
                                if (LibraryDefinitionFactory.isLibrary(file)) {
                                    registerLibraryDefinition(LibraryDefinitionFactory.createLibraryDefinition(file));
                                } else if (BundleDefinitionFactory.isBundle(file)) {
                                    registerBundleDefinition(BundleDefinitionFactory.createBundleDefinition(file));
                                }
                            }
                        }

                    });

                } catch (Exception e) {
                    throw new BundleProvisioningException("Error registering bundles from provider '" + this.searcher + "'", e);
                }
                this.initialised = true;
                updateCaches();
            }
        } finally {
            writeLock.unlock();
        }
    }

    /**
     * Extracts the OSGi metadata from the supplied <code>BundleDefinition</code> location and adds the
     * <code>BundleDefinition</code> to the repository.
     * 
     * @param definition the <code>BundleDefinition</code>.
     */
    void registerBundleDefinition(BundleDefinition definition) {
        if (TRACE.isInfoEnabled()) {
            TRACE.traceInfo("Registering Bundle Definition: '" + definition + "'");
        }
        try {
            parseBundleManifest(definition);
        } catch (Exception ex) {
            LOGGER.log(OsgiProvisioningLogEvent.MALFORMED_MANIFEST, definition.getLocation().getAbsolutePath(), ex.getMessage());
        }
    }

    /**
     * Registers the supplied {@link LibraryDefinition} in this repository.
     * 
     * @param definition the <code>LibraryDefinition</code>.
     */
    void registerLibraryDefinition(LibraryDefinition definition) {
        if (TRACE.isInfoEnabled()) {
            TRACE.traceInfo("Registering Library Definition: '" + definition + "'");
        }
        try {
            this.librariesBySymbolicName.recordDefinition(definition.getSymbolicName(), definition.getVersion(), definition);
        } catch (DuplicateDefinitionException dde) {
            LOGGER.log(OsgiProvisioningLogEvent.DUPLICATE_LIBRARY_DEFINITION, definition.getLocation().getAbsolutePath(),
                definition.getSymbolicName(), definition.getVersion(), dde.getExistingDefinition().getLocation().getAbsolutePath());
        }

        File installedLocation = definition.getLocation();
        if (installedLocation.isDirectory()) {
            File[] directoryEntries = installedLocation.listFiles();
            for (File file : directoryEntries) {
                if (file.isFile() && file.getName().endsWith(".jar")) {
                    BundleDefinition bundleDefinition = BundleDefinitionFactory.createBundleDefinition(file);
                    registerBundleDefinition(bundleDefinition);
                }
            }
        }
    }

    /**
     * Update all caches to reflect the current repository state.
     */
    void updateCaches() {
        this.definitionsBySymbolicName.updateCache();
        this.definitionsByExportedPackage.updateCache();
        this.librariesBySymbolicName.updateCache();
        this.definitionsByFragmentHost.updateCache();
    }

    /**
     * Picks a location for a definition from the supplied candidate set.
     * 
     * @param candidateDefinitions the candidate definitions.
     * @return the selected definition.
     */
    private <T> T pickDefinition(Set<? extends T> candidateDefinitions) {
        return candidateDefinitions == null || candidateDefinitions.isEmpty() ? null : candidateDefinitions.iterator().next();
    }

    /**
     * Parse the {@link BundleManifest} and register any symbolic name or exported packages against the supplied
     * <code>BundleDefinition</code> location.
     * 
     * @param definition the <code>BundleDefinition</code>.
     * @throws Exception if an error occurs.
     */
    private void parseBundleManifest(BundleDefinition definition) throws Exception {
        BundleManifest manifest = definition.getManifest();
        if (manifest == null) {
            return;
        }
        BundleSymbolicName bundleSymbolicName = manifest.getBundleSymbolicName();
        if (bundleSymbolicName != null) {
            String symbolicName = bundleSymbolicName.getSymbolicName();
            if (symbolicName != null && symbolicName.length() > 0) {
                String bundleVersion = manifest.getBundleVersion();
                Version version = Version.parseVersion(bundleVersion);
                try {
                    this.definitionsBySymbolicName.recordDefinition(symbolicName, version, definition);
                } catch (DuplicateDefinitionException dde) {
                    LOGGER.log(OsgiProvisioningLogEvent.DUPLICATE_BUNDLE_SYMBOLIC_NAME, definition.getLocation().getAbsolutePath(), symbolicName,
                        version, dde.getExistingDefinition().getLocation().getAbsolutePath());
                }
            }
        }

        if (manifest.isFragmentAutoAttach()) {
            FragmentHostHeader fragmentHost = manifest.getFragmentHostHeader();
            if (fragmentHost != null) {
                String symbolicName = fragmentHost.getBundleSymbolicName();
                this.definitionsByFragmentHost.recordDefinition(symbolicName, Version.emptyVersion, definition);
            }
        }

        processExportedPackages(definition);
    }

    /**
     * Parses any exported packages and registers them against the supplied {@link BundleDefinition}.
     * 
     * @param definition the <code>BundleDefinition</code>.
     * @throws Exception if an error occurs.
     */
    private void processExportedPackages(BundleDefinition definition) throws Exception {
        ExportPackageHeader header = definition.getManifest().getExportPackageHeader();
        if (header == null) {
            return;
        }
        for (PackageExport export : header.getPackageExports()) {
            Version version = export.getVersion();
            for (String pkg : export.getPackageNames()) {
                this.definitionsByExportedPackage.recordDefinition(pkg, version, definition);
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    public String toString() {
        return this.searcher.toString();
    }

    public void refresh() {
        Lock writeLock = this.readWriteLock.writeLock();
        try {
            writeLock.lock();
            this.initialised = false;
            initialiseIfNecessary();
        } finally {
            writeLock.unlock();
        }
    }
    
    private static FileSystemSearcher createFileSystemSearcher(String searchPath, String absoluteRoot) {
        return new AntPathMatchingFileSystemSearcher(convertToAntStylePath(expandProperties(searchPath)), absoluteRoot);
    }
    
    private static String convertToAntStylePath(String searchPath) {
        return searchPath.replaceAll("\\{[^\\}]+\\}", "*");
    }
    
    private static String expandProperties(String value) {
        Pattern regex = PROPERTY_PATTERN;
        StringBuffer buffer = new StringBuffer(value.length());
        Matcher matcher = regex.matcher(value);
        int propertyGroup = matcher.groupCount();
        String key, property = "";
        while (matcher.find()) {
            key = matcher.group(propertyGroup);
            property = "";
            if(key.contains("::")){
                String[] keyDefault = key.split("::");
                property = System.getProperty(keyDefault[0]);
                if(property == null){
                    property = keyDefault[1];
                }else{
                    property.replace('\\', '/');
                }
            }else{
                property = System.getProperty(matcher.group(propertyGroup)).replace('\\', '/');
            }
            matcher.appendReplacement(buffer, property);
        }
        matcher.appendTail(buffer);
        return buffer.toString();
    }

    /**
     * Hex encoding utility.
     * 
     * @author Andy Wilkinson
     * @since 1.0
     */
    private static class HexEncoder {

        /**
         * Used to build output as Hex
         */
        private static final char[] HEX_CHARS = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'a', 'b', 'c', 'd', 'e', 'f' };

        public static String encodeHex(byte[] bytes) {
            StringBuffer hexBuffer = new StringBuffer(bytes.length);
            for (byte b : bytes) {
                int unsignedByte = b & 0xFF;
                hexBuffer.append(HEX_CHARS[(unsignedByte & 0xF0) >>> 4]);
                hexBuffer.append(HEX_CHARS[(unsignedByte & 0xF)]);
            }
            return hexBuffer.toString();
        }

    }

    /** 
     * {@inheritDoc}
     */
    public Set<BundleDefinition> getBundles() {
        if (!this.definitionsBySymbolicName.isCacheValid()) {
            initialiseIfNecessary();
        }        
        Lock readLock = this.readWriteLock.readLock();
        readLock.lock();
        try {
            return this.definitionsBySymbolicName.getDefinitions();
        } finally {
            readLock.unlock();
        }
    }

    /** 
     * {@inheritDoc}
     */
    public Set<LibraryDefinition> getLibraries() {
        if (!this.librariesBySymbolicName.isCacheValid()) {
            initialiseIfNecessary();
        }        
        Lock readLock = this.readWriteLock.readLock();
        readLock.lock();
        try {
            return this.librariesBySymbolicName.getDefinitions();
        } finally {
            readLock.unlock();
        }
    }
}
