/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.provisioning.repository;

import java.io.File;

/**
 * A <code>FileSystemSearcher</code> is used to search the file system, typically for bundles and libraries.
 * 
 * <strong>Concurrent Semantics</strong><br /> 
 * Implementations need not be thread-safe.
 * 
 * @author Andy Wilkinson
 * @since 1.0
 */
interface FileSystemSearcher {

    /**
     * Gets the age of the search space. Can be used to determine if the contents of the search space have changed since
     * the last search
     * 
     * @return the age of the search space
     */
    long getAgeOfDefinitionSpace();

    /**
     * Returns the directory at the root of the search space.
     * @return the root of the search space
     */
    File getRootDir();

    /**
     * Searches the search space, notifying the given callback of matching files or directories that are found.
     * @param callback the <code>SearchCallback</code> to notify when a file or directory is found.
     */
    void search(SearchCallback callback);

    /**
     * Callback for searching against the search path. <p/>
     * 
     * <strong>Concurrent Semantics</strong><br />
     * 
     * Implementations need not be threadsafe.
     * 
     * @author Rob Harrop
     * @since 1.0
     */
    public interface SearchCallback {

        /**
         * Called when a file is found that matches the search pattern.
         * 
         * @param file the matching file.
         * @param terminal flag indicating whether the file matches against the terminal element of the search pattern.
         */
        void found(File file, boolean terminal);
    }
}
