/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.provisioning.repository;

import java.io.File;
import java.util.concurrent.atomic.AtomicLong;

import com.springsource.server.osgi.provisioning.BundleDefinition;
import com.springsource.server.osgi.provisioning.repository.SearchPattern.PatternElement;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * Helper class that recursively scans a directory, matching against a pattern to find bundle jars or directories.<p/>
 * 
 * A pattern can contain property placeholders and match elements. A property placeholder is defined using
 * <code>${&lt;name&gt;}</code> and is expanded to include the appropriate
 * {@link System#getProperty(String) system property} value. A match element is simply a marker defining a variable
 * section of the path. A root directory for the pattern is extracted from the leading static portion of the pattern.
 * Scanning proceeds from the root directory, and {@link BundleDefinition BundleDefinitions} are created for any jars or
 * directories that match against the supplied pattern. For example:<p/>
 * 
 * <code>
 * Pattern: ${user.dir}/../../ivy-cache/{org}/{name}/{bundle}.jar <p/>
 * </code>
 * 
 * This matches against <code>../../ivy-cache/org.aspectj/com.springsource.org.aspectj.runtime/1.6.0.m2/com.springsource.org.aspectj.runtime-1.6.0.m2.jar</code> but not
 * <code>../../ivy-cache/org.aspectj/com.springsource.org.aspectj.runtime/1.6.0.m2/com.springsource.org.aspectj.runtime-1.6.0.m2</code> (missing the trailing
 * <code>.jar</code>) or <code>../../ivy-cache/org.aspectj/aspectjrt</code> (not enough levels relative to the root dir). <p/>
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @author Andy Wilkinson
 * @since 1.0
 */
final class StandardFileSystemSearcher implements FileSystemSearcher {

    private static final Tracer TRACER  = TracerFactory.createTracer(StandardFileSystemSearcher.class);
    
    private final SearchPattern searchPattern;

    /**
     * Creates a new <code>FileSystemSearcher</code>.
     * 
     * @param searchPath the path to be searched
     * @param absoluteRoot an absolute path to be used if the search path is relative
     */
    StandardFileSystemSearcher(String searchPath, String absoluteRoot) {
        this.searchPattern = new SearchPattern(searchPath, absoluteRoot);
    }

    /**
     * Gets the age of the search space.
     * 
     * @return the age of the search space.
     */
    public long getAgeOfDefinitionSpace() {
        final AtomicLong age = new AtomicLong(this.searchPattern.getRootDir().lastModified());
        search(new SearchCallback() {

            public void found(File file, boolean terminal) {
                age.set(Math.max(age.get(), file.lastModified()));
            }

        });
        return age.get();
    }

    /**
     * Searches against the configured search path and calls back to the supplied {@link FileSystemSearcher.SearchCallback} for every
     * matched element. The <code>terminal</code> flag for the <code>SearchCallback</code> is set to
     * <code>true</code> when matching against a the terminal element of the search pattern (typically the actual file
     * or directory being searched for).
     * 
     * @param callback the <code>SearchCallback</code>.
     */
    public void search(SearchCallback callback) {
        search(this.searchPattern.getRootDir(), 0, callback);
    }

    private void search(File searchDir, int patternIndex, SearchCallback callback) {
        if (this.searchPattern.getPatternElementCount() == 0) {
            callback.found(searchDir, true);
            return;
        }
        PatternElement element = this.searchPattern.getPatternElement(patternIndex);
        boolean terminal = patternIndex == this.searchPattern.getPatternElementCount() - 1;
        File[] entries = searchDir.listFiles();

        for (File file : entries) {
            if (!element.matches(file)) {
                continue;
            }
            if(TRACER.isInfoEnabled()) {
                TRACER.traceInfo(String.format("Matched file %s against pattern %s.", file.getAbsolutePath(), this.searchPattern));
            }
            callback.found(file, terminal);
            if (!terminal && file.isDirectory()) {
                search(file, patternIndex + 1, callback);
            }
        }
    }

    @Override public String toString() {
        return String.format("%s [searchPattern=%s]", getClass().getSimpleName(), this.searchPattern);
    }
    
    public File getRootDir() {
        return this.searchPattern.getRootDir();
    }
}
