/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.provisioning.support;

import java.io.File;

import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

import com.springsource.server.osgi.provisioning.BundleDefinition;

/**
 * Factory for creating {@link BundleDefinition BundleDefinitions}. <p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
public final class BundleDefinitionFactory {

    private static final Tracer TRACER  = TracerFactory.createTracer(BundleDefinitionFactory.class);
    
    /**
     * Creates a {@link BundleDefinition} for the supplied {@link File}.
     * 
     * @param file the <code>File</code>.
     * @return the <code>BundleDefinition</code>.
     */
    public static BundleDefinition createBundleDefinition(File file) {
        if(TRACER.isDebugEnabled()) {
            TRACER.traceDebug("Creating bundle definition for: " + file.getAbsolutePath());
        }
        if (!file.exists()) {
            throw new IllegalArgumentException("File '" + file.getAbsolutePath() + "' does not exist.");
        }
        if (file.isDirectory()) {
            if(TRACER.isDebugEnabled()) {
                TRACER.traceDebug("Created directory bundle definition for: " + file.getAbsolutePath());
            }
            return new DirectoryBundleDefinition(file);
        } else {
            if(TRACER.isDebugEnabled()) {
                TRACER.traceDebug("Created jar bundle definition for: " + file.getAbsolutePath());
            }
            return new JarBundleDefinition(file);
        }
    }

    /**
     * Queries whether the supplied {@link File} is actually a bundle.
     * 
     * @param file the <code>File</code>.
     * @return <code>true</code> if the <code>File</code> is a bundle, otherwise <code>false</code>.
     */
    public static boolean isBundle(File file) {
        return JarBundleDefinition.isJarBundle(file) || DirectoryBundleDefinition.isDirectoryBundle(file);
    }
}
