/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.provisioning.support;

import java.io.File;
import java.io.Reader;
import java.util.ArrayList;
import java.util.List;
import java.util.jar.Attributes;
import java.util.jar.Manifest;

import com.springsource.server.osgi.manifest.BundleManifest;
import com.springsource.server.osgi.manifest.ManifestFactory;
import com.springsource.server.osgi.manifest.StandardBundleImport;
import com.springsource.server.osgi.manifest.StandardBundleManifestFactory;
import com.springsource.server.osgi.manifest.BundleManifest.ImportBundleHeader.BundleImport;
import com.springsource.server.osgi.manifest.parse.OSGiIntegrationLogService;
import com.springsource.server.osgi.provisioning.LibraryDefinition;

/**
 * Implementation of {@link LibraryDefinition} for exploded directory libraries. <p/>
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
final class DirectoryLibraryDefinition extends AbstractLibraryDefinition {

    /**
     * Creates a new <code>DirectoryLibraryDefinition</code>.
     * 
     * @param libraryDirectory the directory of the library.
     */
    public DirectoryLibraryDefinition(File libraryDirectory) {
        super(libraryDirectory, true);
    }

    /**
     * {@inheritDoc}
     */
    protected BundleImport[] locateLibraryBundles(File libraryDirectory) {
        File[] members = libraryDirectory.listFiles();
        List<BundleImport> bundles = new ArrayList<BundleImport>(members.length);
        for (File file : members) {
            if (file.getName().endsWith(".jar")) {
                Reader manifest = ManifestUtils.manifestReaderFromJar(file);
                BundleManifest bundleManifest = new StandardBundleManifestFactory(new OSGiIntegrationLogService()).createBundleManifest(manifest);
                String name = bundleManifest.getBundleSymbolicName().getSymbolicName();
                String version = bundleManifest.getBundleVersion();
                bundles.add(new StandardBundleImport(name, version, "mandatory", "bundle"));
            }
        }
        return bundles.toArray(new BundleImport[bundles.size()]);
    }
    

    /**
     * Queries whether the supplied file is an exploded directory library.
     * 
     * @param file the file.
     * @return <code>true</code> if the file is an exploded directory library, otherwise <code>false</code>.
     */
    public static boolean isDirectoryLibrary(File file) {
        Reader reader = ManifestUtils.manifestReaderFromExplodedDirectory(file);

        if (reader != null) {
            Manifest mf = new ManifestFactory(reader, file.getAbsolutePath()).getManifest();
            Attributes attrs = mf.getMainAttributes();
            return attrs.getValue(ATTR_LIBRARY_SYMBOLICNAME) != null && attrs.getValue(ATTR_LIBRARY_VERSION) != null;
        }

        return false;
    }

    /** 
     * {@inheritDoc}
     */
    @Override protected Reader manifestReader(File libraryLocation) {
        return ManifestUtils.manifestReaderFromExplodedDirectory(libraryLocation);
    }

}
