/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.provisioning.tools;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringReader;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import org.osgi.framework.Version;

import com.springsource.json.parser.AntlrJSONParser;
import com.springsource.json.parser.JSONParser;
import com.springsource.json.parser.ListNode;
import com.springsource.json.parser.Node;
import com.springsource.json.parser.ScalarNode;
import com.springsource.json.parser.ScalarNodeType;
import com.springsource.server.bootstrap.BootstrapConfigurationException;
import com.springsource.server.osgi.manifest.BundleManifest;
import com.springsource.server.osgi.manifest.BundleManifestFactory;
import com.springsource.server.osgi.manifest.BundleManifestProcessor;
import com.springsource.server.osgi.manifest.StandardBundleManifestFactory;
import com.springsource.server.osgi.manifest.VersionRange;
import com.springsource.server.osgi.manifest.BundleManifest.ExportPackageHeader;
import com.springsource.server.osgi.manifest.BundleManifest.ImportBundleHeader;
import com.springsource.server.osgi.manifest.BundleManifest.ImportLibraryHeader;
import com.springsource.server.osgi.manifest.BundleManifest.ImportPackageHeader;
import com.springsource.server.osgi.manifest.BundleManifest.RequireBundleHeader;
import com.springsource.server.osgi.manifest.BundleManifest.ExportPackageHeader.PackageExport;
import com.springsource.server.osgi.manifest.BundleManifest.ImportBundleHeader.BundleImport;
import com.springsource.server.osgi.manifest.BundleManifest.ImportLibraryHeader.LibraryImport;
import com.springsource.server.osgi.manifest.BundleManifest.ImportPackageHeader.PackageImport;
import com.springsource.server.osgi.manifest.BundleManifest.ImportPackageHeader.PackageImport.Resolution;
import com.springsource.server.osgi.manifest.BundleManifest.RequireBundleHeader.RequireBundle;
import com.springsource.server.osgi.manifest.BundleManifest.RequireBundleHeader.RequireBundle.Visibility;
import com.springsource.server.osgi.manifest.parse.HeaderDeclaration;
import com.springsource.server.osgi.manifest.parse.HeaderParserFactory;
import com.springsource.server.osgi.manifest.parse.OSGiIntegrationLogService;
import com.springsource.server.osgi.provisioning.BundleDefinition;
import com.springsource.server.osgi.provisioning.LibraryDefinition;
import com.springsource.server.osgi.provisioning.repository.BundleRepository;
import com.springsource.server.osgi.provisioning.repository.BundleRepositoryFactory;
import com.springsource.server.osgi.provisioning.tools.DependencyLocator.DependencyLocationException.ImportDescriptor;

/**
 * A helper class for locating a bundle's dependencies.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * The class is <strong>thread-safe</strong>
 * 
 * @author Andy Wilkinson
 * @since 1.0
 */
public final class DependencyLocator {

    private static final String SYSYEM_PACKAGES_PROPERTY = "org.osgi.framework.system.packages";

    private static final String BUNDLE_VERSION_ATTRIBUTE = "bundle-version";

    private static final String VERSION_ATTRIBUTE = "version";

    private static final String VERSION_ALTERNATE_ATTRIBUTE = "specification-version";

    private final BundleRepository bundleRepository;

    private static final List<String> SEARCH_PATHS_PATH = Arrays.asList(new String[] { "provisioning", "searchPaths" });

    private static final String[] DEFAULT_SEARCH_PATHS = new String[] { "repository/bundles/subsystems/{name}/{bundle}.jar",
        "repository/bundles/ext/{bundle}", "repository/libraries/ext/{library}", "repository/bundles/usr/{bundle}",
        "repository/libraries/usr/{library}" };

    private static final String[] LIB_SEARCH_PATH = new String[] { "lib/{bundle}.jar" };

    private static final String SERVER_CONFIG_PATH = File.separatorChar + "config" + File.separatorChar + "server.config";

    /**
     * Creates a new <code>DependencyLocator</code> that will search for dependencies within the Server instance
     * located at the supplied <code>serverHomePath</code>. To improve search performance, bundle and library
     * locations are cached. The cache files will be written to the directory identified by the supplied
     * <code>cacheDirectoryPath</code>.
     * 
     * @param serverHomePath The path to the server installation from within which dependencies are to be located
     * @param cacheDirectoryPath The path of the directory to which cache files should be written
     * @param javaVersion The version of Java to be used to determine the appropriate OSGi profile to use for defining
     *        system bundle exports
     * 
     * @throws IOException if a problem occurs loading and parsing the configuration of the Server instance.
     */
    public DependencyLocator(String serverHomePath, String cacheDirectoryPath, JavaVersion javaVersion) throws IOException {
        this(serverHomePath, null, cacheDirectoryPath, javaVersion);
    }

    /**
     * Creates a new <code>DependencyLocator</code> that will search for dependencies within the Server instance
     * located at the supplied <code>serverHomePath</code>. The supplied <code>additionalSearchPaths</code> will
     * also be included in the search. Each search path is used to locate bundles and libraries. To improve search
     * performance, bundle and library locations are cached. The cache files will be written to the directory identified
     * by the supplied <code>cacheDirectoryPath</code>.
     * 
     * @param serverHomePath The path to the server installation from within which dependencies are to be located
     * @param additionalSearchPaths The additional search paths to use to locate the bundles and libraries that can
     *        satisfy dependencies
     * @param cacheDirectoryPath The path of the directory to which cache files should be written
     * @param javaVersion The version of Java to be used to determine the appropriate OSGi profile to use for defining
     *        system bundle exports
     * 
     * @throws IOException if a problem occurs loading and parsing the configuration of the Server instance.
     */
    public DependencyLocator(String serverHomePath, String[] additionalSearchPaths, String cacheDirectoryPath, JavaVersion javaVersion) throws IOException {
        this.bundleRepository = new SystemPackageFilteringCompositeBundleRepository(serverHomePath, additionalSearchPaths, cacheDirectoryPath, javaVersion);
    }

    /**
     * Returns the {@link BundleRepository} that is being used by this <code>DependencyLocator</code> to locate
     * dependencies.
     * 
     * @return This <code>DependencyLocator</code>'s <code>BundleRepository</code>
     */
    public BundleRepository getBundleRepository() {
        return this.bundleRepository;
    }

    private static String[] readSearchPathsFromConfig(String configPath) {
        if (configPath != null) {
            File configFile = new File(configPath);
            if (configFile.exists()) {
                JSONParser parser = new AntlrJSONParser();
                try {
                    Node root = parser.parse(configFile.toURL());
                    return readStringsFromListNode(root, SEARCH_PATHS_PATH, DEFAULT_SEARCH_PATHS);
                } catch (Exception e) {
                    throw new RuntimeException("Config file '" + configPath + "' could not be parsed.", e);
                }
            } else {
                throw new IllegalArgumentException("Config file '" + configPath + "' does not exist.");
            }
        }
        return new String[0];
    }

    private static String[] readStringsFromListNode(Node root, List<String> pathToNode, String[] defaultValues) {
        if (root != null) {
            Node searchPathsNode = root.getNode(pathToNode);
            if (searchPathsNode != null) {
                if (searchPathsNode instanceof ListNode) {
                    List<String> searchPaths = new ArrayList<String>();
                    ListNode listNode = (ListNode) searchPathsNode;
                    List<Node> nodes = listNode.getNodes();
                    for (Node node : nodes) {
                        if (node instanceof ScalarNode && ScalarNodeType.String.equals(((ScalarNode) node).getType())) {
                            String path = ((ScalarNode) node).getValue();
                            searchPaths.add(path);
                        }
                    }
                    return searchPaths.toArray(new String[searchPaths.size()]);
                } else {
                    throw new BootstrapConfigurationException("The node at " + pathToNode + " must be a list node.");
                }
            } else {
                return defaultValues;
            }
        } else {
            return defaultValues;
        }
    }

    /**
     * Locates all of the dependencies defined in the supplied manifest. Dependencies are identified from the manifest's
     * <code>Import-Package</code>, <code>Import-Bundle</code> and <code>Import-Library</code> headers. The
     * dependencies are returned in the form of a <code>Map</code>. Each key in the <code>Map</code> is a
     * <code>File</code> which points to the location of a bundle that satisfies one or more of the manifest's
     * dependencies. Each <code>List&lt;String&gt;</code> contains all of the packages within a particular bundle
     * which are visible, i.e. they are imported by the supplied manifest. If a dependency is found to be satisfied by
     * the system bundle a <code>null</code> key may be included in the returned <code>Map</code> if the location of
     * the satisfying jar could not be determined, e.g. because the dependency is upon a standard JRE package.
     * <p>
     * Processing of the <code>Import-Package</code> header will, for each package, search for a bundle that exports
     * the package at the specified version and include its location in the returned <code>Map</code> along with an
     * entry in the associated <code>List</code> for the package. If no such bundle is found and the import does not
     * have a resolution of optional it will be added to a list of those which cannot be satisfied and included in a
     * <code>DependencyLocationException</code> thrown once processing is complete. If more than one such bundle is
     * found they will all be added to the Map.
     * <p>
     * Processing of the <code>Import-Library</code> header will, for each library, search for a library with the
     * required name and version. The location of each bundle that is in the library will be included in the returned
     * <code>Map</code> along with entries in the associated <code>Lists</code> for every package that is exported
     * from the library's bundles. If no library with the required name and version can be found, and the import does
     * not have a resolution of optional, the import of the library will be added to a list of those which cannot be
     * satisfied and included in a <code>DependencyLocationException</code> thrown once processing is complete.
     * <p>
     * If a single bundle satisfies more than one import its location will only be included once in the returned
     * <code>Map</code>.
     * <p>
     * If the manifest has no dependencies an empty <code>Map</code> is returned.
     * 
     * @param manifest
     * @return the locations of all of the given manifest's dependencies
     * @throws DependencyLocationException if any of the manifest's dependencies cannot be located
     */
    public Map<File, List<String>> locateDependencies(BundleManifest manifest) throws DependencyLocationException {

        List<ImportDescriptor> unsatisfiablePackageImports = new ArrayList<ImportDescriptor>();
        List<ImportDescriptor> unsatisfiableLibraryImports = new ArrayList<ImportDescriptor>();
        List<ImportDescriptor> unsatisfiableBundleImports = new ArrayList<ImportDescriptor>();
        List<ImportDescriptor> unsatisfiableRequireBundles = new ArrayList<ImportDescriptor>();

        Map<File, List<String>> dependencyLocations = new HashMap<File, List<String>>();

        ImportPackageHeader importPackageHeader = manifest.getImportPackageHeader();

        if (importPackageHeader != null) {
            processImportPackageHeader(importPackageHeader, dependencyLocations, unsatisfiablePackageImports);
        }

        ImportLibraryHeader importLibraryHeader = manifest.getImportLibraryHeader();
        if (importLibraryHeader != null) {
            processImportLibraryHeader(importLibraryHeader, dependencyLocations, unsatisfiableLibraryImports);
        }

        ImportBundleHeader importBundleHeader = manifest.getImportBundleHeader();
        if (importBundleHeader != null) {
            processImportBundleHeader(importBundleHeader, dependencyLocations, unsatisfiableBundleImports);
        }

        RequireBundleHeader requireBundleHeader = manifest.getRequireBundleHeader();
        if (requireBundleHeader != null) {
            List<String> packageNames = createListOfAllPackagesThatHaveAlreadyBeenSatisfied(dependencyLocations);
            processRequireBundleHeader(requireBundleHeader, dependencyLocations, unsatisfiableRequireBundles, packageNames, true);
        }

        throwDependencyLocationExceptionIfNecessary(unsatisfiablePackageImports, unsatisfiableBundleImports, unsatisfiableLibraryImports,
            unsatisfiableRequireBundles, dependencyLocations);

        return dependencyLocations;
    }

    private static List<String> createListOfAllPackagesThatHaveAlreadyBeenSatisfied(Map<File, List<String>> dependencyLocations) {
        List<String> packagesThatHaveBeenSatisfied = new ArrayList<String>();
        for (File location : dependencyLocations.keySet()) {
            packagesThatHaveBeenSatisfied.addAll(dependencyLocations.get(location));

        }
        return packagesThatHaveBeenSatisfied;
    }

    private void processImportPackageHeader(ImportPackageHeader importPackageHeader, Map<File, List<String>> dependencyLocations,
        List<ImportDescriptor> unsatisfiablePackageImports) {
        List<PackageImport> packageImports = importPackageHeader.getPackageImports();
        if (packageImports != null && packageImports.size() > 0) {
            processPackageImports(packageImports, dependencyLocations, unsatisfiablePackageImports);
        }
    }

    private void processImportBundleHeader(ImportBundleHeader importBundleHeader, Map<File, List<String>> dependencyLocations,
        List<ImportDescriptor> unsatisfiableBundleImports) {
        List<BundleImport> bundleImports = importBundleHeader.getImportedBundles();
        if (bundleImports != null && bundleImports.size() > 0) {
            processBundleImports(bundleImports, dependencyLocations, unsatisfiableBundleImports);
        }
    }

    private void processBundleImports(List<BundleImport> bundleImports, Map<File, List<String>> dependencyLocations,
        List<ImportDescriptor> unsatisfiableBundleImports) {
        for (BundleImport bundleImport : bundleImports) {
            processBundleImport(bundleImport, dependencyLocations, unsatisfiableBundleImports);
        }
    }

    private void processBundleImport(BundleImport bundleImport, Map<File, List<String>> dependencyLocations,
        List<ImportDescriptor> unsatisfiableBundleImports) {
        String symbolicName = bundleImport.getBundleSymbolicName();
        String versionRangeString = bundleImport.getVersionRange();
        VersionRange bundleVersionRange;
        if (versionRangeString == null) {
            bundleVersionRange = VersionRange.NATURAL_NUMBER_RANGE;
        } else {
            bundleVersionRange = new VersionRange(versionRangeString);
        }
        BundleDefinition bundleDefinition = this.bundleRepository.findBySymbolicName(symbolicName, bundleVersionRange);
        if (bundleDefinition == null) {
            unsatisfiableBundleImports.add(new ImportDescriptor(symbolicName, bundleVersionRange.toString(), bundleVersionRange.toParseString()));
        } else {
            registerDependencyLocationAndPackageNameForEveryExportedPackage(dependencyLocations, new StandardBundleManifestFactory(
                new OSGiIntegrationLogService()), bundleDefinition, null);
        }
    }

    private void processImportLibraryHeader(ImportLibraryHeader importLibraryHeader, Map<File, List<String>> dependencyLocations,
        List<ImportDescriptor> unsatisfiableLibraryImports) {
        List<LibraryImport> libraryImports = importLibraryHeader.getImportedLibraries();
        if (libraryImports != null && libraryImports.size() > 0) {
            processLibraryImports(libraryImports, dependencyLocations, unsatisfiableLibraryImports);
        }
    }

    private void processLibraryImports(List<LibraryImport> libraryImports, Map<File, List<String>> dependencyLocations,
        List<ImportDescriptor> unsatisfiableLibraryImports) {
        for (LibraryImport libraryImport : libraryImports) {
            String libraryName = libraryImport.getLibraryName();
            VersionRange versionRange = new VersionRange(libraryImport.getVersionRange());
            LibraryDefinition libraryDefinition = this.bundleRepository.findLibrary(libraryName, versionRange);

            if (libraryDefinition != null) {
                BundleImport[] bundleImports = libraryDefinition.getLibraryBundles();
                BundleManifestFactory bundleManifestFactory = new StandardBundleManifestFactory(new OSGiIntegrationLogService());
                for (int i = 0; i < bundleImports.length; i++) {
                    BundleImport libraryBundle = bundleImports[i];
                    String symbolicName = libraryBundle.getBundleSymbolicName();
                    String versionRangeString = libraryBundle.getVersionRange();
                    VersionRange bundleVersionRange;
                    if (versionRangeString == null) {
                        bundleVersionRange = VersionRange.NATURAL_NUMBER_RANGE;
                    } else {
                        bundleVersionRange = new VersionRange(versionRangeString);
                    }
                    BundleDefinition bundleDefinition = this.bundleRepository.findBySymbolicName(symbolicName, bundleVersionRange);
                    if (bundleDefinition == null) {
                        unsatisfiableLibraryImports.add(new ImportDescriptor(libraryName, versionRange.toString(), versionRange.toParseString()));
                    } else {
                        registerDependencyLocationAndPackageNameForEveryExportedPackage(dependencyLocations, bundleManifestFactory, bundleDefinition,
                            null);
                    }
                }
            } else if (LibraryImport.Resolution.MANDATORY.equals(libraryImport.getResolution())) {
                unsatisfiableLibraryImports.add(new ImportDescriptor(libraryName, versionRange.toString(), versionRange.toParseString()));
            }
        }
    }

    private void registerDependencyLocationAndPackageNameForEveryExportedPackage(Map<File, List<String>> dependencyLocations,
        BundleManifestFactory bundleManifestFactory, BundleDefinition bundleDefinition, List<String> packagesThatHaveAlreadyBeenSatisfied) {
        BundleManifestProcessor processor = new BundleManifestProcessor(bundleManifestFactory);
        PackageImport[] packageImports = processor.createPackageImportForEachPackageExport(new BundleManifest[] { bundleDefinition.getManifest() }).getOutput();
        for (PackageImport packageImport : packageImports) {
            for (String packageName : packageImport.getPackageNames()) {
                if (packagesThatHaveAlreadyBeenSatisfied == null || !packagesThatHaveAlreadyBeenSatisfied.contains(packageName)) {
                    registerPackageNameAgainstDependencyLocation(bundleDefinition.getLocation(), packageName, dependencyLocations);
                }
            }
        }
    }

    private void processPackageImports(List<PackageImport> packageImports, Map<File, List<String>> dependencyLocations,
        List<ImportDescriptor> unsatisfiablePackageImports) {
        for (PackageImport packageImport : packageImports) {
            String version = getVersionAttribute(packageImport.getAttributes());
            VersionRange versionRange;
            if (version != null) {
                versionRange = new VersionRange(version);
            } else {
                versionRange = VersionRange.NATURAL_NUMBER_RANGE;
            }
            List<String> packageNames = packageImport.getPackageNames();
            for (String packageName : packageNames) {
                Set<? extends BundleDefinition> bundleDefinitions = this.bundleRepository.findByExportedPackage(packageName, versionRange);
                if (bundleDefinitions.size() > 0) {
                    for (BundleDefinition bundleDefinition : bundleDefinitions) {
                        File location = bundleDefinition.getLocation();
                        registerPackageNameAgainstDependencyLocation(location, packageName, dependencyLocations);
                    }
                } else if (Resolution.MANDATORY.equals(packageImport.getResolution())) {
                    unsatisfiablePackageImports.add(new ImportDescriptor(packageName, versionRange.toString(), versionRange.toParseString()));
                }
            }
        }
    }

    /**
     * Get the version attribute from a given attribute map.
     * 
     * @param map the attribute map
     * @return the value of the version attribute
     */
    private static String getVersionAttribute(Map<String, String> map) {
        String version = map.get(VERSION_ATTRIBUTE);
        if (version == null) {
            version = map.get(VERSION_ALTERNATE_ATTRIBUTE);
        }
        return version;
    }

    private void processRequireBundleHeader(RequireBundleHeader requireBundleHeader, Map<File, List<String>> dependencyLocations,
        List<ImportDescriptor> unsatisfiableRequireBundles, List<String> packagesThatHaveAlreadyBeenSatisfied, boolean root) {
        List<RequireBundle> requiredBundles = requireBundleHeader.getRequiredBundles();
        BundleManifestFactory bundleManifestFactory = new StandardBundleManifestFactory(new OSGiIntegrationLogService());
        for (RequireBundle requiredBundle : requiredBundles) {
            if (root || Visibility.REEXPORT.equals(requiredBundle.getVisibility())) {
                String bundleSymbolicName = requiredBundle.getBundleSymbolicName();
                String bundleVersion = requiredBundle.getAttributes().get(BUNDLE_VERSION_ATTRIBUTE);
                VersionRange versionRange;
                if (bundleVersion != null) {
                    versionRange = new VersionRange(bundleVersion);
                } else {
                    versionRange = VersionRange.NATURAL_NUMBER_RANGE;
                }
                BundleDefinition bundleDefinition = this.bundleRepository.findBySymbolicName(bundleSymbolicName, versionRange);
                if (bundleDefinition != null) {
                    registerDependencyLocationAndPackageNameForEveryExportedPackage(dependencyLocations, bundleManifestFactory, bundleDefinition,
                        packagesThatHaveAlreadyBeenSatisfied);
                    RequireBundleHeader dependencysRequireBundleHeader = bundleDefinition.getManifest().getRequireBundleHeader();
                    if (dependencysRequireBundleHeader != null) {
                        processRequireBundleHeader(dependencysRequireBundleHeader, dependencyLocations, unsatisfiableRequireBundles,
                            packagesThatHaveAlreadyBeenSatisfied, false);
                    }
                } else if (com.springsource.server.osgi.manifest.BundleManifest.RequireBundleHeader.RequireBundle.Resolution.MANDATORY.equals(requiredBundle.getResolution())) {
                    unsatisfiableRequireBundles.add(new ImportDescriptor(requiredBundle.getBundleSymbolicName(), versionRange.toString(), versionRange.toParseString()));
                }
            }
        }
    }

    private void registerPackageNameAgainstDependencyLocation(File location, String packageName, Map<File, List<String>> dependencyLocations) {
        List<String> existingPackageNames = dependencyLocations.get(location);
        if (existingPackageNames == null) {
            existingPackageNames = new ArrayList<String>();
            dependencyLocations.put(location, existingPackageNames);
        }
        existingPackageNames.add(packageName);
    }

    private void throwDependencyLocationExceptionIfNecessary(List<ImportDescriptor> unsatisfiablePackageImports,
        List<ImportDescriptor> unsatisfiableBundleImports, List<ImportDescriptor> unsatisfiableLibraryImports,
        List<ImportDescriptor> unsatisfiableRequireBundles, Map<File, List<String>> satisfiedDependencies) {

        if (!unsatisfiableLibraryImports.isEmpty() || !unsatisfiablePackageImports.isEmpty() || !unsatisfiableRequireBundles.isEmpty()
            || !unsatisfiableBundleImports.isEmpty()) {
            throw new DependencyLocationException(toArray(unsatisfiablePackageImports), toArray(unsatisfiableBundleImports),
                toArray(unsatisfiableLibraryImports), toArray(unsatisfiableRequireBundles), satisfiedDependencies);
        }
    }

    private static ImportDescriptor[] toArray(List<ImportDescriptor> importDescriptors) {
        return importDescriptors.toArray(new ImportDescriptor[importDescriptors.size()]);
    }

    /**
     * A <code>DependencyLocationException</code> is thrown when one or more problems occurs when attempting to locate
     * all of a bundle's dependencies.
     * <p>
     * 
     * <strong>Concurrent Semantics</strong><br />
     * This class is <code>thread-safe</code>
     * 
     * @author Andy Wilkinson
     * @since 1.0
     */
    public static class DependencyLocationException extends RuntimeException {

        private static final long serialVersionUID = 7598582916642377631L;

        private final ImportDescriptor[] unsatisfiablePackageImports;

        private final ImportDescriptor[] unsatisfiableBundleImports;

        private final ImportDescriptor[] unsatisfiableLibraryImports;

        private final ImportDescriptor[] unsatisfiableRequireBundle;

        private final Map<File, List<String>> satisfiedDependencies;

        private DependencyLocationException(ImportDescriptor[] unsatisfiablePackageImports, ImportDescriptor[] unsatisfiableBundleImports,
            ImportDescriptor[] unsatisfiableLibraryImports, ImportDescriptor[] unsatisfiableRequireBundle,
            Map<File, List<String>> satisfiedDependencies) {
            super(createMessage(unsatisfiablePackageImports, unsatisfiableBundleImports, unsatisfiableLibraryImports, unsatisfiableRequireBundle));
            this.unsatisfiableLibraryImports = unsatisfiableLibraryImports;
            this.unsatisfiablePackageImports = unsatisfiablePackageImports;
            this.satisfiedDependencies = satisfiedDependencies;
            this.unsatisfiableRequireBundle = unsatisfiableRequireBundle;
            this.unsatisfiableBundleImports = unsatisfiableBundleImports;
        }

        private static String createMessage(ImportDescriptor[] unsatisfiablePackageImports, ImportDescriptor[] unsatisfiableBundleImports,
            ImportDescriptor[] unsatisfiableLibraryImports, ImportDescriptor[] unsatisfiableRequireBundle) {
            StringBuffer buffer = new StringBuffer("Dependency location failed. The following problems occurred:\n");
            buffer.append("    Unsatisfiable package imports:\n");
            appendImportDescriptors(unsatisfiablePackageImports, buffer);
            buffer.append("    Unsatisfiable bundle imports:\n");
            appendImportDescriptors(unsatisfiableBundleImports, buffer);
            buffer.append("    Unsatisfiable library imports:\n");
            appendImportDescriptors(unsatisfiableLibraryImports, buffer);
            buffer.append("    Unsatisfiable required bundles:\n");
            appendImportDescriptors(unsatisfiableRequireBundle, buffer);
            return buffer.toString();
        }

        private static void appendImportDescriptors(ImportDescriptor[] descriptors, StringBuffer buffer) {
            if (descriptors.length == 0) {
                buffer.append("        None.\n");
            } else {
                for (ImportDescriptor descriptor : descriptors) {
                    buffer.append("        ").append(descriptor.getName()).append(" with version ").append(descriptor.getVersion()).append("\n");
                }
            }
        }

        /**
         * Returns an array of all of the imported packages that could not be satisfied.
         * 
         * @return the manifest's unsatisfiable package imports
         */
        public ImportDescriptor[] getUnsatisfiablePackageImports() {
            return this.unsatisfiablePackageImports;
        }

        /**
         * Returns an array of all of the imported libraries that could not be satisfied.
         * 
         * @return the manifest's unsatisfiable library imports
         */
        public ImportDescriptor[] getUnsatisfiableLibraryImports() {
            return this.unsatisfiableLibraryImports;
        }

        /**
         * Returns an array of all of the required bundles that could not be satisfied.
         * 
         * @return the manifest's unsatisfiable required bundles
         */
        public ImportDescriptor[] getUnsatisfiableRequireBundle() {
            return unsatisfiableRequireBundle;
        }

        public ImportDescriptor[] getUnsatisfiableBundleImports() {
            return unsatisfiableBundleImports;
        }

        /**
         * Returns all of the dependencies that could be satisfied.
         * 
         * @return the manifest's satisfiable dependencies
         */
        public Map<File, List<String>> getSatisfiedDependencies() {
            return this.satisfiedDependencies;
        }

        /**
         * A descriptor of an import in a bundle manifest. An import can be of either a package or a library.
         * 
         * <strong>Concurrent Semantics</strong><br />
         * This class is <strong>thread-safe</strong>.
         * 
         * @author Andy Wilkinson
         * @since 1.0
         */
        public static class ImportDescriptor {

            private final String name;

            private final String version;

            private final String parseVersion;

            private ImportDescriptor(String name, String version, String parseVersion) {
                this.name = name;
                this.version = version;
                this.parseVersion = parseVersion;
            }

            public String getName() {
                return name;
            }

            public String getVersion() {
                return version;
            }

            public String getParseVersion() {
                return parseVersion;
            }
        }
    }

    private static class SystemPackageFilteringCompositeBundleRepository implements BundleRepository {

        private final Map<String, Version> systemPackages;

        private final BundleRepository mainRepository;

        private final BundleRepository systemPackageRepository;

        private final BundleManifest systemBundleManifest;

        private final Set<BundleDefinition> jreProvidedDependenciesDefinitions;

        private final BundleDefinition systemBundleDefinition;

        private static final String SYSTEM_BUNDLE_SYMBOLIC_NAME = "org.eclipse.osgi";

        SystemPackageFilteringCompositeBundleRepository(String serverHomePath, String[] additionalSearchPaths, String cacheDirectoryPath, JavaVersion javaVersion)
            throws IOException {

            String serverConfigPath = null;
            String serverProfilePath = null;

            if (serverHomePath != null) {
                serverConfigPath = serverHomePath + SERVER_CONFIG_PATH;
                if (javaVersion == JavaVersion.Java6) {
                    serverProfilePath = serverHomePath + File.separator + "lib" + File.separator + "java6-server.profile";
                } else {
                    serverProfilePath = serverHomePath + File.separator + "lib" + File.separator + "java5-server.profile";
                }

	            File serverProfile = new File(serverProfilePath);
	            if (!serverProfile.exists()) {
	                serverProfilePath = serverHomePath + File.separator + "lib" + File.separator + "server.profile";
	            }
            }
            

            String[] searchPathsFromConfig = readSearchPathsFromConfig(serverConfigPath);
            String[] searchPaths = new String[searchPathsFromConfig.length + additionalSearchPaths.length];
            System.arraycopy(searchPathsFromConfig, 0, searchPaths, 0, searchPathsFromConfig.length);
            System.arraycopy(additionalSearchPaths, 0, searchPaths, searchPathsFromConfig.length, additionalSearchPaths.length);

            this.mainRepository = BundleRepositoryFactory.createBundleRepository(searchPaths, cacheDirectoryPath, serverHomePath);
            this.systemPackageRepository = BundleRepositoryFactory.createBundleRepository(LIB_SEARCH_PATH, cacheDirectoryPath, serverHomePath);

            systemPackages = parseProfile(serverProfilePath);

            systemPackages.putAll(findExportsFromOsgiImplementationBundle(this.systemPackageRepository));

            systemBundleManifest = createManifestExportingPackages(systemPackages);

            jreProvidedDependenciesDefinitions = new HashSet<BundleDefinition>();
            systemBundleDefinition = new SystemBundleDefinition(systemBundleManifest, null);
            jreProvidedDependenciesDefinitions.add(systemBundleDefinition);
        }

        private Map<String, Version> findExportsFromOsgiImplementationBundle(BundleRepository bundleRepository) {
            Map<String, Version> exports = new HashMap<String, Version>();

            BundleDefinition definition = bundleRepository.findBySymbolicName(SYSTEM_BUNDLE_SYMBOLIC_NAME, VersionRange.NATURAL_NUMBER_RANGE);
            if (definition != null) {
                BundleManifest manifest = definition.getManifest();
                ExportPackageHeader exportPackageHeader = manifest.getExportPackageHeader();
                if (exportPackageHeader != null) {
                    for (PackageExport packageExport : exportPackageHeader.getPackageExports()) {
                        Version version = Version.parseVersion(packageExport.getAttributes().get(VERSION_ATTRIBUTE));
                        for (String name : packageExport.getPackageNames()) {
                            exports.put(name, version);
                        }
                    }
                }
            }

            return exports;
        }

        /**
         * {@inheritDoc}
         */
        public Set<? extends BundleDefinition> findByExportedPackage(String packageName, VersionRange versionRange) {
            Version version;
            if ((version = systemPackages.get(packageName)) != null) {
                if (versionRange.includes(version)) {
                    Set<? extends BundleDefinition> definitionsFromLib = this.systemPackageRepository.findByExportedPackage(packageName, versionRange);
                    if (definitionsFromLib.isEmpty()) {
                        return jreProvidedDependenciesDefinitions;
                    } else {
                        Set<BundleDefinition> systemBundleDefinitions = new HashSet<BundleDefinition>();
                        for (BundleDefinition definition : definitionsFromLib) {
                            systemBundleDefinitions.add(new SystemBundleDefinition(systemBundleManifest, definition.getLocation()));
                        }
                        return systemBundleDefinitions;
                    }
                }
            }
            return this.mainRepository.findByExportedPackage(packageName, versionRange);
        }

        /**
         * {@inheritDoc}
         */
        public Set<? extends BundleDefinition> findByFragmentHost(String bundleSymbolicName, Version version) {
            return this.mainRepository.findByFragmentHost(bundleSymbolicName, version);
        }

        /**
         * {@inheritDoc}
         */
        public BundleDefinition findBySymbolicName(String symbolicName, VersionRange versionRange) {
            return this.mainRepository.findBySymbolicName(symbolicName, versionRange);
        }

        /**
         * {@inheritDoc}
         */
        public LibraryDefinition findLibrary(String libraryName, VersionRange versionRange) {
            return this.mainRepository.findLibrary(libraryName, versionRange);
        }

        /**
         * {@inheritDoc}
         */
        public Set<? extends BundleDefinition> getBundles() {
            Set<BundleDefinition> bundles = new HashSet<BundleDefinition>();
            bundles.addAll(this.mainRepository.getBundles());
            // expose jre packages for content assist
            bundles.addAll(this.jreProvidedDependenciesDefinitions);
            return bundles;
        }

        /**
         * {@inheritDoc}
         */
        public Set<? extends LibraryDefinition> getLibraries() {
            return this.mainRepository.getLibraries();
        }

        /**
         * {@inheritDoc}
         */
        public void refresh() {
            this.mainRepository.refresh();
        }

        private static Map<String, Version> parseProfile(String serverProfilePath) throws IOException {
            Map<String, Version> systemPackages = new HashMap<String, Version>();
            if (serverProfilePath != null) {
                Properties properties = new Properties();
                properties.load(new FileInputStream(new File(serverProfilePath)));

                String systemPackagesString = properties.getProperty(SYSYEM_PACKAGES_PROPERTY);

                List<HeaderDeclaration> exportHeaders = HeaderParserFactory.newHeaderParser(new OSGiIntegrationLogService()).parsePackageHeader(
                    systemPackagesString, "Export-Package");

                for (HeaderDeclaration exportHeader : exportHeaders) {
                    String versionString = getVersionAttribute(exportHeader.getAttributes());
                    Version version = versionString != null ? new Version(versionString) : Version.emptyVersion;
                    for (String packageName : exportHeader.getNames()) {
                        systemPackages.put(packageName, version);
                    }
                }
            }
            return systemPackages;
        }

        private static BundleManifest createManifestExportingPackages(Map<String, Version> packages) {
            StringWriter manifestBuilder = new StringWriter();
            PrintWriter manifestWriter = new PrintWriter(manifestBuilder);
            manifestWriter.println("Bundle-ManifestVersion: 2");
            manifestWriter.println("Manifest-Version: 1.0");

            if (packages.size() > 0) {
                manifestWriter.println("Export-Package: ");

                Set<String> packageNamesSet = packages.keySet();
                Iterator<String> packageNames = packageNamesSet.iterator();

                while (packageNames.hasNext()) {
                    String packageName = packageNames.next();
                    Version version = packages.get(packageName);
                    if (packageNames.hasNext()) {
                        manifestWriter.println(" " + packageName + ";version=\"" + version.toString() + "\",");
                    } else {
                        manifestWriter.println(" " + packageName + ";version=\"" + version.toString() + "\"");
                    }
                }
            }
            return new StandardBundleManifestFactory(new OSGiIntegrationLogService()).createBundleManifest(new StringReader(
                manifestBuilder.toString()));
        }
    }

    static class SystemBundleDefinition implements BundleDefinition {

        private final BundleManifest manifest;

        private final File location;

        SystemBundleDefinition(BundleManifest manifest, File location) {
            this.manifest = manifest;
            this.location = location;
        }

        /**
         * {@inheritDoc}
         */
        public BundleManifest getManifest() {
            return this.manifest;
        }

        /**
         * {@inheritDoc}
         */
        public File getLocation() {
            return this.location;
        }
    }

    public enum JavaVersion {
        Java5, Java6;
    }
}
