/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.profile.core.internal;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.springframework.context.ApplicationContext;
import org.springframework.osgi.util.OsgiBundleUtils;

import com.springsource.server.config.core.ConfigNode;
import com.springsource.server.config.core.ConfigTree;
import com.springsource.server.config.core.ConfigTreeFactory;
import com.springsource.server.config.core.ListConfigNode;
import com.springsource.server.config.core.ScalarConfigNode;
import com.springsource.server.kernel.core.BundleDescriptor;
import com.springsource.server.kernel.core.BundleStarter;
import com.springsource.server.kernel.core.KernelException;
import com.springsource.server.kernel.core.Subsystem;
import com.springsource.server.kernel.core.SubsystemRegistry;
import com.springsource.server.management.proxy.profile.ProfileSubsystemInfo;
import com.springsource.server.osgi.framework.OsgiFramework;
import com.springsource.server.profile.core.FatalProfileException;
import com.springsource.server.profile.core.Profile;
import com.springsource.server.serviceability.logging.LogService;
import com.springsource.server.serviceability.logging.LogServiceFactory;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * This class gets the profile configuration and adds the subsystems specified by the configuration to the Server.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * This class is thread safe.
 * 
 * @author Glyn Normington
 * @since 1.0
 */
class ProfileConfiguration implements Profile {

    private static final String DEPLOYER_SUBSYSTEM_NAME = "com.springsource.server.deployer";

    private static final LogService LOGGER = LogServiceFactory.getLogService();

    private static final String PROFILE_CONFIG_NAME = "profile";

    private static final String PROFILE_NAME_KEY = "name";

    private static final String SUBSYSTEM_LIST_KEY = "subsystems";

    private static final String OPTIONAL_SUBSYSTEM_LIST_KEY = "optionalSubsystems";

    private static final List<String> SUBSYSTEM_LIST_PATH;

    private static final List<String> OPTIONAL_SUBSYSTEM_LIST_PATH;

    private final Tracer TRACE = TracerFactory.createTracer(ProfileConfiguration.class);
    
    private final SubsystemRegistry subsystemRegistry;

    private final OsgiFramework osgi;

    private volatile boolean started = false;

    private final ConfigTree configTree;

    private final Object monitor = new Object();

    private final String profileName;
    
    private final BundleStarter starter;

    static {
        SUBSYSTEM_LIST_PATH = new ArrayList<String>(1);
        SUBSYSTEM_LIST_PATH.add(SUBSYSTEM_LIST_KEY);
        OPTIONAL_SUBSYSTEM_LIST_PATH = new ArrayList<String>(1);
        OPTIONAL_SUBSYSTEM_LIST_PATH.add(OPTIONAL_SUBSYSTEM_LIST_KEY);
    }

    public ProfileConfiguration(ConfigTreeFactory configTreeFactory, BundleContext context, SubsystemRegistry subsystemRegistry, OsgiFramework osgi, BundleStarter starter) {
        this.osgi = osgi;
        this.subsystemRegistry = subsystemRegistry;
        this.configTree = configTreeFactory.createConfigTree(PROFILE_CONFIG_NAME, context);
        this.profileName = this.configTree.getStringValue(PROFILE_CONFIG_NAME + ":" + PROFILE_NAME_KEY);
        this.starter = starter;
        LOGGER.log(ProfileLogEvent.PROFILE_IDENTITY, this.profileName);

        // Add any subsystems required by the profile.
        for (ConfigNode n : ((ListConfigNode) this.configTree.getConfigPoint(PROFILE_CONFIG_NAME).getRootNode().getNode(SUBSYSTEM_LIST_PATH)).getNodes()) {
            addSubsystem((String) ((ScalarConfigNode) n).getValue(), context, true);
        }

        // Add any optional subsystems listed in the profile, but ignore any which cannot be installed.
        for (ConfigNode n : ((ListConfigNode) this.configTree.getConfigPoint(PROFILE_CONFIG_NAME).getRootNode().getNode(OPTIONAL_SUBSYSTEM_LIST_PATH)).getNodes()) {
            addSubsystem((String) ((ScalarConfigNode) n).getValue(), context, false);
        }

        LOGGER.log(ProfileLogEvent.PROFILE_INSTALLED, this.profileName);

    }

    /**
     * Install and start the subsystem with the given name and register a corresponding management service.
     * 
     * @param subsystemName the name of the subsystem to add
     * @param context a {@link BundleContext}
     * @param mandatory <code>true</code> if and only if absence of the subsystem should be treated as an error
     */
    private void addSubsystem(String subsystemName, BundleContext context, boolean mandatory) {
        Subsystem subsystem = null;
        try {
            subsystem = this.subsystemRegistry.installSubsystem(subsystemName);
        } catch (KernelException e) {
            if (mandatory) {
                throw new FatalProfileException("Required subsystem was not found", e);
            }
        }
        if (mandatory && subsystem == null) {
            throw new FatalProfileException("Required subsystem '" + subsystemName + "' was not found");
        }
        if (subsystem != null) {
            subsystem.start();
            context.registerService(ProfileSubsystemInfo.class.getName(), new StandardProfileSubsystemInfo(subsystem, this.profileName), null);
        } else {
            TRACE.traceWarning("Optional subsystem '%s' has not been installed", subsystemName);
        }
    }

    /**
     * {@inheritDoc}
     */
    public boolean waitUntilStarted(int timeout) {
        if (!this.started) {
            synchronized (this.monitor) {
                for (Subsystem subsystem : this.subsystemRegistry.getSubsystems()) {
                    if (DEPLOYER_SUBSYSTEM_NAME.equals(subsystem.getName())) {
                        continue;
                    }
                    for (BundleDescriptor bundleDescriptor : subsystem.getBundleDescriptors()) {
                        String bundleSymbolicName = bundleDescriptor.getSymbolicName();
                        /*
                         * Skip the deployer as this is the subsystem that drives waitUntilStarted and so will not have
                         * published its application context.
                         */
                        for (Bundle bundle : this.osgi.getBundleContext().getBundles()) {
                            if (bundleSymbolicName.equals(bundle.getSymbolicName())) {
                                if (!OsgiBundleUtils.isBundleActive(bundle)) {
                                    LOGGER.log(ProfileLogEvent.INACTIVE_BUNDLE, this.configTree.getStringValue(PROFILE_CONFIG_NAME + ":"
                                        + PROFILE_NAME_KEY), bundleSymbolicName);
                                } else {
                                    Future<ApplicationContext> future = this.starter.awaitStart(bundle);
                                    try {
                                        future.get(timeout, TimeUnit.SECONDS);
                                    } catch(TimeoutException ex) {
                                        LOGGER.log(ProfileLogEvent.PROFILE_TIMEOUT, this.configTree.getStringValue(PROFILE_CONFIG_NAME + ":"
                                            + PROFILE_NAME_KEY), bundleSymbolicName);
                                        return false;
                                    } catch (InterruptedException e) {
                                        LOGGER.log(ProfileLogEvent.PROFILE_TIMEOUT, this.configTree.getStringValue(PROFILE_CONFIG_NAME + ":"
                                            + PROFILE_NAME_KEY), bundleSymbolicName);
                                        return false;
                                    } catch (ExecutionException e) {
                                        LOGGER.log(ProfileLogEvent.PROFILE_FAILED, e.getCause(), this.configTree.getStringValue(PROFILE_CONFIG_NAME + ":"
                                            + PROFILE_NAME_KEY), bundleSymbolicName);
                                        return false;
                                    }
                                }
                                break;
                            }
                        }
                    }
                }
                this.started = true;
                LOGGER.log(ProfileLogEvent.PROFILE_STARTED, this.configTree.getStringValue(PROFILE_CONFIG_NAME + ":" + PROFILE_NAME_KEY));
            }
        }
        return true;
    }
}
