/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.serviceability.ffdc.aspects;

/**
 * Maintains state about which {@link Throwable} was last seen for the current thread.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * State is thread local.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
final class FFDCExceptionState {

    private static final State STATE = new State();

    /**
     * Record that the supplied {@link Throwable} has been seen.
     * 
     * @param t the <code>Throwable</code>.
     */
    public static void record(Throwable t) {
        STATE.set(t);
    }

    /**
     * Query whether the supplied {@link Throwable} or one of its causes has been seen.
     * 
     * @param t the <code>Throwable</code>
     * @return <code>true</code> if the <code>Throwable</code> has been seen, otherwise <code>false</code>.
     */
    public static boolean seen(Throwable t) {
        boolean seen = false;
        Throwable s = t;
        do {
            if (s.equals(STATE.get())) {
                seen = true;
                break;
            }
            s = s.getCause();
        } while (s != null);
        return seen;
    }

    private static class State extends ThreadLocal<Throwable> {

        /**
         * {@inheritDoc}
         */
        @Override public String toString() {
            return String.format("FFDC Exception State [lastSeen = '" + get() + "']");
        }

    }
}
