/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.serviceability.logging.internal;

import java.io.StringWriter;

import com.springsource.server.serviceability.logging.LogEvent;
import com.springsource.server.serviceability.output.AbstractEntry;
import com.springsource.server.serviceability.output.Level;

/**
 * A subclass of {@link AbstractEntry} that adds {@link LogEvent}-specific formatting.
 *
 * <strong>Concurrent Semantics</strong><br />
 * This class is <strong>thread-safe</strong>
 *
 * @author Andy Wilkinson
 * @since 1.0
 */
class LogEntry extends AbstractEntry {
      
    private final String text;
    private static final String SYSTEM_NAME = "SP";
    private final String logMessage;
    private final String entrySource;
    
    LogEntry(Level level, String message) {
        this(level, null, null, message);
    }
    
    /**
     * Creates a new <code>LogEntry</code> for the supplied {@link LogEvent}. When published the
     * entry will included the supplied message after having had the supplied inserts applied.
     * @param event The <code>LogEvent</code> 
     * @param mainMessage The message for the event
     * @param inserts The inserts to be applied to the message.
     */
    LogEntry(LogEvent event, String mainMessage, Object... inserts){
        this(event, mainMessage, null, inserts);
    }
    
    /**
     * Creates a new <code>LogEntry</code> for the supplied {@link LogEvent}. When published the
     * entry will included the supplied message after having had the supplied inserts applied. The
     * stack trace of the supplied <code>Throwable</code> will be appended to the message.
     * @param event The <code>LogEvent</code> 
     * @param mainMessage The message for the event
     * @param throwable A <code>Throwable</code>, the stack trace of which will be appended to the message.
     * @param inserts The inserts to be applied to the message.
     */
    LogEntry(LogEvent event, String mainMessage, Throwable throwable, Object... inserts) {
        this(event.getLevel(), throwable, event.getSubsystemName(), formatMessage(event, mainMessage, throwable, inserts));
    }
    
    private LogEntry(Level level, Throwable throwable, String entrySource, String message) {
        super(level, throwable);
        this.entrySource = entrySource;
        this.logMessage = message;
        this.text = String.format("%s %s %s", getFormattedTimestamp(), getFormattedThreadId(), this.logMessage);
    }
    
    final static String formatMessage(LogEvent event, String eventMessage, Throwable throwable, Object... inserts) {
        String eventMessageWithInserts;
        
        if (inserts != null && inserts.length > 0) {
            eventMessageWithInserts = String.format(eventMessage, inserts);
        } else {
            eventMessageWithInserts = eventMessage;
        }
        
        String outputMessage = String.format("<%s> %s", SYSTEM_NAME + event.getEventCode(), eventMessageWithInserts);

        if (throwable != null) {
            StringWriter stackWriter = new StringWriter();
            stackWriter.append(outputMessage);
            stackWriter.append(LINE_SEPARATOR);
            stackWriter.append(AbstractEntry.getThrowableStackTraceAsString(throwable));
            
            return stackWriter.toString();
        }

        return outputMessage;
    }

    /** 
     * {@inheritDoc}
     */
    public String getText() { 
        return this.text;
    }
    
    /** 
     * {@inheritDoc}
     */
    @Override
    public String getSource() {
        return entrySource;
    }
}
