/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.serviceability.logging.internal;

import java.io.IOException;

import com.springsource.server.serviceability.logging.LogServiceFactory;
import com.springsource.server.serviceability.output.Destination;
import com.springsource.server.serviceability.output.DestinationRegistry;
import com.springsource.server.serviceability.output.EntryPublisher;
import com.springsource.server.serviceability.output.EntryPublisherFactory;

/**
 * Bootstrap for the logging component. Initialises the {@link EntryPublisher EntryPublishers} and
 * {@link Destination Destinations} used by logging thereby making the {@link LogServiceFactory} ready for use by the
 * rest of the Server. <p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * This class is <strong>thread-safe</strong>.
 * 
 * @author Andy Wilkinson
 * @since 1.0
 */
public class LoggingBootstrap {

    private static EntryPublisher loggingEntryPublisher;

    private static EntryPublisher tracingEntryPublisher;

    private static final Object startedLock = new Object();

    private static boolean started;

    /**
     * Starts the logging component. This method should only be invoked once. Duplicate invocations will be ignored.
     */
    public static void start(String loggingDirectory, DestinationRegistry destinationRegistry,
        EntryPublisherFactory entryPublisherFactory) {
        synchronized (startedLock) {
            if (!started) {
                Destination console = destinationRegistry.getDestination(DestinationRegistry.CONSOLE_DESTINATION_NAME);

                Destination loggingDestination;

                try {
                    if (loggingDirectory != null) {
                        loggingDestination = destinationRegistry.createFileDestination("logging", "logging", loggingDirectory);
                    } else {
                        loggingDestination = destinationRegistry.createFileDestination("logging", "logging");
                    }
                } catch (IOException ioe) {
                    throw new RuntimeException("Unable to create logging destination with directory '" + loggingDirectory + "'", ioe);
                }

                // TODO Don't hard-code this name
                Destination tracingDestination = destinationRegistry.getDestination("trace");

                loggingEntryPublisher = entryPublisherFactory.createEntryPublisher(new Destination[] { console, loggingDestination });
                tracingEntryPublisher = entryPublisherFactory.createEntryPublisher(new Destination[] { tracingDestination });

                started = true;
            }
        }
    }

    /**
     * Returns the <code>EntryPublisher</code> to be used to publish translated log messages.
     * 
     * @return The <code>EntryPublisher</code> for the logging of translated messages
     */
    public static EntryPublisher getLoggingEntryPublisher() {
        return loggingEntryPublisher;
    }

    /**
     * Returns the <code>EntryPublisher</code> to be used to publish English messages to the Server's trace file(s).
     * 
     * @return The <code>EntryPublisher</code> for the tracing of English messages
     */
    public static EntryPublisher getTracingEntryPublisher() {
        return tracingEntryPublisher;
    }
}
