/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.serviceability.logging.internal;

import java.util.Locale;
import java.util.MissingResourceException;
import java.util.ResourceBundle;

import com.springsource.server.serviceability.logging.LogEventMessageResolver;
import com.springsource.server.serviceability.logging.LocaleResolver;
import com.springsource.server.serviceability.logging.LogEvent;

/**
 * LogEventMessageResolver which resolves messages from a {@link ResourceBundle} using
 * a {@link LocaleResolver} to determine the {@link Locale} in which the message should
 * be generated.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * This class is <strong> thread-safe</strong>.
 * 
 * @author Andy Wilkinson
 * @since 1.0
 */
public class ResourceBundleEventMessageResolver implements LogEventMessageResolver {

    private static final String RESOURCE_BUNDLE_BASENAME = "LogResource";

    private final LocaleResolver localeResolver = new StandardLocaleResolver();

    /**
     * {@inheritDoc}
     */
    public String resolveLogEventMessage(LogEvent event) {
        return resolveLogEventMessage(localeResolver.getLocale(), event);
    }

    /** 
     * {@inheritDoc}
     */
    public String resolveLogEventMessage(Locale locale, LogEvent event) {
        String message;
        try {
            String location = event.getSubsystemName() + RESOURCE_BUNDLE_BASENAME;
            ResourceBundle resBun = ResourceBundle.getBundle(location, locale, event.getClass().getClassLoader());
            message = resBun.getString(event.getEventCode());
        } catch (MissingResourceException e) {
            return null;
        }
        return message;
    }

}
