/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.serviceability.logging.internal;

import java.util.Locale;

import com.springsource.server.serviceability.logging.LogEventMessageResolver;
import com.springsource.server.serviceability.logging.LogEvent;
import com.springsource.server.serviceability.logging.LogService;
import com.springsource.server.serviceability.output.Entry;
import com.springsource.server.serviceability.output.EntryPublisher;
import com.springsource.server.serviceability.tracing.TraceEntry;

/**
 * Standard implementation of <code>LogService</code> which publishes <code>LogEvents</code> to both logging and
 * tracing.
 * 
 * @author Christopher Frost
 * @author Andy Wilkinson
 * @since 1.0
 */
public class StandardLogService implements LogService {

    private static final String UNKNOWN_SOURCE = "unknown";

    private LogEventMessageResolver eventMessageResolver;

    private EntryPublisher tracingEntryPublisher;

    private EntryPublisher loggingEntryPublisher;

    /**
     * Creates a new <code>StandardLogService</code> the will use the supplied <code>LogEventMessageResolver</code>
     * to resolve the messages of the <code>LogEvents</code> which it logs. The <code>LogEvents</code> will be
     * published to the supplied loggingEntryPublisher and to the supplied tracingEntryPublisher.
     * 
     * @param eventMessageResolver The resolver to use to resolve log event message.
     * @param loggingEntryPublisher The logging publisher to which events will be published.
     * @param tracingEntryPublisher The tracing publisher to which events will be published.
     */
    public StandardLogService(LogEventMessageResolver eventMessageResolver, EntryPublisher loggingEntryPublisher, EntryPublisher tracingEntryPublisher) {
        this.tracingEntryPublisher = tracingEntryPublisher;
        this.loggingEntryPublisher = loggingEntryPublisher;
        this.eventMessageResolver = eventMessageResolver;
    }

    /**
     * The entry point for a new SpringSource dm Server Log message
     * 
     * @param event - The unique log event that has happened and needs to be dealt with
     * @param inserts - The objects associated with the log event
     */
    public final void log(LogEvent event, Object... inserts) {
        log(event, null, inserts);
    }

    public final void log(LogEvent event, Throwable throwable, Object... inserts) {
        // Trace the log event
        if (tracingEntryPublisher != null) {
            String englishMessage = this.eventMessageResolver.resolveLogEventMessage(Locale.ENGLISH, event);
            if (englishMessage == null) {
                englishMessage = createMessageForMissingEntry(event, inserts);
            }
            String formattedMessage = LogEntry.formatMessage(event, englishMessage, throwable, inserts);
            Entry englishEntry = new TraceEntry(findSource(), formattedMessage, event.getLevel(), event.getLevel().toString());
            tracingEntryPublisher.publish(englishEntry);
        }

        // Log the log event
        if (loggingEntryPublisher != null) {
            String translatedMessage = this.eventMessageResolver.resolveLogEventMessage(event);
            if (translatedMessage == null) {
                translatedMessage = createMessageForMissingEntry(event, inserts);
            }
            Entry translatedEntry = new LogEntry(event, translatedMessage, throwable, inserts);
            loggingEntryPublisher.publish(translatedEntry);
        }
    }

    private static String findSource() {
        StackTraceElement[] stack = Thread.currentThread().getStackTrace();
        for (int i = stack.length - 2; i >= 0; i--) {
            String className = stack[i].getClassName();

            if (className != null && className.contains(StandardLogService.class.getName())) {
                return stack[i + 1].getClassName();
            }
        }
        return UNKNOWN_SOURCE;
    }

    private String createMessageForMissingEntry(LogEvent event, Object... inserts) {
        StringBuilder messageBuilder = new StringBuilder();
        messageBuilder.append("A message for key ");
        messageBuilder.append(event.getEventCode());
        messageBuilder.append(" could not be found.");

        if (inserts != null && inserts.length > 0) {
            messageBuilder.append(" The supplied inserts were: ");

            for (int i = 0; i < inserts.length; i++) {
                messageBuilder.append(inserts[i]);

                if (i < inserts.length - 1) {
                    messageBuilder.append(", ");
                }
            }
        }
        return messageBuilder.toString();
    }
}
