/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.serviceability.output;

import java.io.IOException;

/**
 * A DestinationRegistry provides the ability to create output {@link Destination Destinations} and also provides access
 * to <code>Destinations</code> that have been created.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * DestinationRegistry implementations <strong>must</strong> be thread safe.
 * 
 * @author Andy Wilkinson
 * @since 1.0
 */
public interface DestinationRegistry {

    public static final String CONSOLE_DESTINATION_NAME = "console";

    /**
     * Returns the Destination registered against the supplied identifier or null if no Destination has been registered
     * for the supplied identifier.
     * 
     * @param identifier The identifier of the Destination
     * @return The Destination identified by the supplied name
     */
    Destination getDestination(String identifier);

    /**
     * Creates a new destination that writes its output to a file with the given name in the default location.
     * 
     * @param identifier The identifier under which the Destination will be registered
     * @param filename The name of the file to which output will be written
     * 
     * @return the newly created Destination
     * @throws IOException if the output file cannot be created and/or accessed
     */
    Destination createFileDestination(String identifier, String filename) throws IOException;

    /**
     * Creates a new destination that writes its output to a file with the given name in the given location.
     * 
     * @param identifier The identifier under which the Destination will be registered
     * @param filename The name of the file to which output will be written
     * @param location The location for the destination's output file
     * 
     * @return the newly created Destination
     * @throws IOException if the output file cannot be created and/or accessed
     */
    Destination createFileDestination(String identifier, String filename, String location) throws IOException;

    /**
     * Adds a {@link DestinationCreationListener} to this registry that will then be notified of the creation of
     * {@link Destination Destinations} by the registry. When a listener is registered it will be notified of all
     * destinations created by the registry prior to its registration and will then be notified of each subsequent
     * registration as it occurs.
     * 
     * @param listener the listener to register with the registry
     */
    void addDestinationCreationListener(DestinationCreationListener listener);
}
