/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.serviceability.output;

/**
 * Trace levels are used to filter the information traced by the Server. Errors are always traced and so ERROR is the
 * lowest trace level defined. The highest trace level is VERBOSE which provides the finest detail. The default global
 * trace level is INFO which should provide sufficient context to diagnose most problems.
 * 
 * When a trace entry is created, the caller specifies the level of the trace entry. The trace subsystem filters the
 * trace entries that appear in the trace based on the subsystem trace level and the global trace level.
 * 
 * @author Christopher Frost
 * @since 1.0
 */
public enum Level {
    /**
     * Disables tracing.
     */
    NONE(0),
    /**
     * Errors include Server problems such as incorrect behaviour or corrupted internal state.
     */
    ERROR(1),

    /**
     * Warnings are 'soft' errors which don't necessarily imply the Server is broken.
     */
    WARN(2),

    /**
     * Informational traces record interesting Server events which provide more context for diagnosing errors and
     * warnings.
     */
    INFO(3),

    /**
     * Debug traces are relatively high volume and could impact the performance of an Server in production.
     */
    DEBUG(4),

    /**
     * Verbose traces may involve costly operations such as checking preconditions and invariants.
     */
    VERBOSE(5);

    private int levelValue;

    private Level(int i) {
        this.levelValue = i;
    }

    /**
     * Queries whether this level is less than or equal to the supplied {@link Level}.
     * 
     * @param check the <code>Level</code> to compare against.
     * @return <code>true</code> if this <code>Level</code> is less than or equal to the supplied <code>Level</code>,
     *         otherwise <code>false</code>.
     */
    public boolean isLessThanOrEqualTo(Level check) {
        return this.levelValue <= check.levelValue;
    }

}
