/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.serviceability.output.internal;

import java.io.BufferedWriter;
import java.io.File;
import java.io.IOException;
import java.io.Writer;
import java.util.ArrayList;

/**
 * A RollingOverFileWriter writes output to a file which will be rolled-over based on the rules encapsulated in a
 * RolloverPolicy. <p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * Instances of this class are <strong>not</strong> thread-safe.
 * 
 * @author Andy Wilkinson
 * @since 1.0
 */
class RollingOverFileWriter extends Writer {

    private File file;

    private Writer writer;
    
    private final RolloverPolicy rolloverPolicy;

    private final String location;

    private final String filename;

    private final String suffix;

    private final int maximumRolloverFiles;

    private final ArrayList<File> rolloverFiles = new ArrayList<File>();

    /**
     * Creates a new <code>RollingOverFileWriter</code> that will rollover based on the supplied
     * {@link RolloverPolicy}. The writer will retain up to the supplied number of rollover of files. Once this limit
     * is reached the oldest rollover file will be deleted to make way for a new one.
     * 
     * @param location The location to files should be written
     * @param filename The base filename to be used for the writer's files
     * @param suffix The suffix for the writer's files
     * @param rolloverPolicy The rollover policy to be used by the writer
     * @param maximumRolloverFiles The maximum number of files to be retained by the writer.
     * @throws IOException
     */
    RollingOverFileWriter(String location, String filename, String suffix, RolloverPolicy rolloverPolicy, int maximumRolloverFiles)
        throws IOException {
        this.location = location;
        this.filename = filename;
        this.suffix = suffix;
        this.file = createOutputFile();

        ensureLocationExists();

        this.writer = new BufferedWriter(new HeaderWritingFileWriter(this.file));
        this.rolloverPolicy = rolloverPolicy;
        this.maximumRolloverFiles = maximumRolloverFiles;

        prepopulateRolloverFilesList();
    }

    private void ensureLocationExists() {
        File locationDirectory = new File(location);
        locationDirectory.mkdirs();
    }

    private void prepopulateRolloverFilesList() {
        File rolloverFileDirectory = new File(location);
        File[] files = rolloverFileDirectory.listFiles();

        for (File file : files) {
            String name = file.getName();
            if (name.startsWith(filename) && name.endsWith(suffix) && !name.equals(filename + suffix)) {
                rolloverFiles.add(file);
            }
        }

        while (rolloverFiles.size() > maximumRolloverFiles) {
            rolloverFiles.remove(0).delete();
        }
    }

    /** 
     * {@inheritDoc}
     */
    @Override
    public void close() throws IOException {
        writer.close();
    }

    /** 
     * {@inheritDoc}
     */
    @Override
    public void flush() throws IOException {
        writer.flush();
    }

    private File createRolloverFile(String rolloverIdentifier) {
        StringBuilder filenameBuilder = new StringBuilder();
        filenameBuilder.append(location).append(File.separatorChar).append(filename).append(rolloverIdentifier).append(suffix);
        File rolloverFile = new File(filenameBuilder.toString());

        if (rolloverFile.exists()) {
            int duplicateDifferentiator = 0;
            while (rolloverFile.exists()) {
                filenameBuilder = new StringBuilder();
                filenameBuilder.append(location).append(File.separatorChar).append(filename).append(rolloverIdentifier).append('_').append(
                    duplicateDifferentiator++).append(suffix);
                rolloverFile = new File(filenameBuilder.toString());
            }
        }

        rolloverFiles.add(rolloverFile);

        if (rolloverFiles.size() > maximumRolloverFiles) {
            rolloverFiles.remove(0).delete();
        }

        return rolloverFile;
    }

    private File createOutputFile() throws IOException {
        StringBuilder filenameBuilder = new StringBuilder();
        filenameBuilder.append(location).append(File.separatorChar).append(filename).append(suffix);
        return new File(filenameBuilder.toString());
    }

    /** 
     * {@inheritDoc}
     */
    @Override
    public void write(char[] cbuf, int off, int len) throws IOException {

        if (rolloverPolicy.isRolloverRequired(file.length() + len)) {
            writer.flush();
            writer.close();

            File renameTarget = createRolloverFile(rolloverPolicy.getRolloverIdentifier());
            file.renameTo(renameTarget);

            file = createOutputFile();
            writer = new BufferedWriter(new HeaderWritingFileWriter(file));
        }

        writer.write(cbuf, off, len);
    }

    String getOutputFilePath() {
        return file.getAbsolutePath();
    }
}
