/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.serviceability.output.internal;

import org.osgi.framework.BundleContext;
import org.osgi.framework.ServiceRegistration;

import com.springsource.server.serviceability.output.DestinationRegistry;
import com.springsource.server.serviceability.output.EntryFilterFactory;
import com.springsource.server.serviceability.output.EntryPublisherFactory;

/**
 * This class provides bootstrap functionality for the serviceability output component.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * This class is thread-safe.
 * 
 * @author Andy Wilkinson
 * @since 1.0
 */
public class ServiceabilityOutputBootstrap {

    private static ServiceRegistration destinationRegistryRegistration;

    private static ServiceRegistration entryPublisherFactoryRegistration;

    private static ServiceRegistration entryFilterFactoryRegistration;

    private static StandardDestinationRegistry destinationRegistry;

    private static EntryPublisherFactory entryPublisherFactory;

    private static EntryFilterFactory entryFilterFactory;

    private static boolean started = false;
    
    private static final Object startedLock = new Object();

    private static boolean servicesPublished = false;

    private static final Object servicesPublishedLock = new Object();

    /**
     * Starts the serviceability output component. This method should only be invoked once.
     * Duplicate invocations will be ignored.
     */
    public static void start() {
        synchronized (startedLock) {
            if (!started) {
                destinationRegistry = new StandardDestinationRegistry();
                entryPublisherFactory = new StandardEntryPublisherFactory();
                entryFilterFactory = new StandardEntryFilterFactory();

                started = true;
            }
        }
    }

    /**
     * Publishes the serviceability output component's services to the OSGi Service Registry
     * available from the supplied <code>BundleContext</code>. Three services for use by
     * clients are published: a {@link DestinationRegistry}, an {@link EntryPublisherFactory},
     * and an {@link EntryFilterFactory}.
     * <p>
     * This method should only be called after {@link #start} has been called. Subsequent invocations
     * without a call to {@link #retractServices()} will have no effect.
     *  
     * @param context The <code>BundleContext</code> to be used to publish the component's services
     * @see #start
     */
    public static void publishServices(BundleContext context) {
        synchronized(startedLock) {
            if (!started) {
                throw new IllegalStateException("Start must have been called prior to service publication.");
            }
        }
        synchronized(servicesPublishedLock) {
            if (!servicesPublished) {
                destinationRegistryRegistration = context.registerService(DestinationRegistry.class.getName(), destinationRegistry, null);
                entryPublisherFactoryRegistration = context.registerService(EntryPublisherFactory.class.getName(), entryPublisherFactory, null);
                entryFilterFactoryRegistration = context.registerService(EntryFilterFactory.class.getName(), entryFilterFactory, null);
                
                servicesPublished = true;
            }
        }
    }

    /**
     * Retracts the service's published by a call to {@link #publishServices(BundleContext)}.
     * <p>
     * This method should only be called once. Duplicate invocations will have no effect.
     */
    public static void retractServices() {
        synchronized(servicesPublishedLock) {
            if (servicesPublished) {               
                destinationRegistryRegistration.unregister();
                entryPublisherFactoryRegistration.unregister();
                entryFilterFactoryRegistration.unregister();
                destinationRegistry.clearListeners();
                servicesPublished = false;
            }
        }
    }

    /**
     * Returns the serviceability output component's {@link DestinationRegistry} created
     * during a call to {@link #start}. If <code>start</code> has not be invoked an
     * IllegalStateException is thrown.
     * @throws IllegalStateException if <code>start</code> has not been called
     * @return the <code>DestinationRegistry</code>
     */
    public static DestinationRegistry getDestinationRegistry() {
        checkStarted();
        return destinationRegistry;
    }

    /**
     * Returns the serviceability output component's {@link EntryPublisherFactory} created
     * during a call to {@link #start}. If <code>start</code> has not be invoked an
     * IllegalStateException is thrown.
     * @throws IllegalStateException if <code>start</code> has not been called
     * @return the <code>EntryPublisherFactory</code>
     */
    public static EntryPublisherFactory getEntryPublisherFactory() {
        checkStarted();
        return entryPublisherFactory;
    }

    /**
     * Returns the serviceability output component's {@link EntryFilterFactory} created
     * during a call to {@link #start}. If <code>start</code> has not be invoked an
     * IllegalStateException is thrown.
     * @throws IllegalStateException if <code>start</code> has not been called
     * @return the <code>EntryFilterFactory</code>
     */
    public static EntryFilterFactory getEntryFilterFactory() {
        checkStarted();
        return entryFilterFactory;
    }
    
    private static void checkStarted() {
        synchronized (startedLock) {
            if (!started) {
                throw new IllegalStateException("The serviceability output component has not been started.");
            }           
        }
    }
}
