/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.serviceability.output.internal;

import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.locks.ReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;

import com.springsource.server.serviceability.output.Entry;
import com.springsource.server.serviceability.output.EntryFilter;
import com.springsource.server.serviceability.output.Level;

/**
 * Standard implementation of {@link EntryFilter}
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * This class is <strong>thread-safe</strong>
 * 
 * @author Andy Wilkinson
 * @since 1.0
 */
public class StandardEntryFilter implements EntryFilter {

    private EntryFilterTreeNode root = new EntryFilterTreeNode(null);

    private final ReadWriteLock lock = new ReentrantReadWriteLock();

    private final Map<String, Level> levels;

    StandardEntryFilter() {
        this(null);
    }

    public StandardEntryFilter(Map<String, Level> levels) {
        this.levels = levels;
        initialize();
    }

    /**
     * {@inheritDoc}
     */
    public boolean filter(Entry entry) {
        return this.filter(entry.getSource(), entry.getLevel());
    }

    /**
     * {@inheritDoc}
     */
    public boolean filter(String source, Level level) {
        
        if (level == null) {
            return true;
        }
        
        Level matchedLevel = null;

        if (source != null) {
            try {
                this.lock.readLock().lock();
                EntryFilterTreeNode matchedNode = this.root;
                EntryFilterTreeNode mostSpecificEverythingFromHereDownMatch = null;

                int index = source.indexOf('.');
                if (index == -1) {
                    index = source.length();
                }
                int lastIndex = 0;

                while (true) {

                    // Check to see if there's a match everything from here down
                    // node, i.e. a '*' and remember this as the best match.
                    EntryFilterTreeNode starNode = matchedNode.getChild("*");

                    if (starNode != null) {
                        mostSpecificEverythingFromHereDownMatch = starNode;
                    }

                    matchedNode = matchedNode.getChild(source.substring(lastIndex, index));

                    if (matchedNode == null) {
                        matchedNode = mostSpecificEverythingFromHereDownMatch;
                        break;
                    }

                    if (index == source.length()) {
                        break;
                    }
                    lastIndex = index + 1;
                    index = source.indexOf('.', lastIndex);
                    if (index == -1) {
                        index = source.length();
                    }
                }

                if (matchedNode != null) {
                    matchedLevel = matchedNode.getLevel();
                }
            } finally {
                this.lock.readLock().unlock();
            }
        }

        if (matchedLevel == null) {
            return false;
        }

        return level.isLessThanOrEqualTo(matchedLevel);
    }

    private void parseEntry(String identifier, Level level) {

        String[] identifierComponents = identifier.split("\\.");

        EntryFilterTreeNode currentNode = this.root;

        for (String component : identifierComponents) {
            EntryFilterTreeNode nodeForComponent = currentNode.getChild(component);

            if (nodeForComponent == null) {
                nodeForComponent = new EntryFilterTreeNode(component);
                currentNode.addChild(nodeForComponent);
            }

            currentNode = nodeForComponent;
        }

        currentNode.setLevel(level);
    }

    private class EntryFilterTreeNode {

        private final String name;

        private Level level;

        private final Map<String, EntryFilterTreeNode> children = new HashMap<String, EntryFilterTreeNode>();

        public EntryFilterTreeNode(String name) {
            this.name = name;
        }

        public EntryFilterTreeNode getChild(String name) {
            return this.children.get(name);
        }

        public void setLevel(Level level) {
            this.level = level;
        }

        public Level getLevel() {
            return this.level;
        }

        public String getName() {
            return this.name;
        }

        public void addChild(EntryFilterTreeNode node) {
            this.children.put(node.getName(), node);
        }

        @Override public String toString() {
            return this.name + " " + this.level + " " + this.children;
        }
    }

    private void parse(Map<String, Level> specificLevels) {
        if (specificLevels != null) {
            for (String identifier : specificLevels.keySet()) {
                parseEntry(identifier, specificLevels.get(identifier));
            }
        }
    }

    private void initialize() {
        try {
            this.lock.writeLock().lock();
            this.root = new EntryFilterTreeNode(null);
            parse(this.levels);
        } finally {
            this.lock.writeLock().unlock();
        }
    }

    /**
     * 
     * {@inheritDoc}
     */
    @Override public String toString() {
        return this.levels.toString();
    }
}
