/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.serviceability.output.internal;

import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;

/**
 * Standard implementation of {@link RolloverPolicy}
 * <strong>Concurrent Semantics</strong><br />
 * This class is <strong>thread-safe</strong>
 * @author Andy Wilkinson
 * @since 1.0
 */
class StandardRolloverPolicy implements RolloverPolicy {
    
    private final long maximumFileSize;
    private final RolloverTimeInterval rolloverTimeInterval;
    private final SimpleDateFormat DATE_FORMAT = new SimpleDateFormat("-yyyy-MM-dd_HH-mm-ss");
    private volatile Calendar startTime;
    
    StandardRolloverPolicy(long maximumFileSize) {
        this(maximumFileSize, RolloverTimeInterval.NEVER);
    }
    
    StandardRolloverPolicy(RolloverTimeInterval timeInterval) {
        this(-1, timeInterval);
        
    }
    
    StandardRolloverPolicy(long maximumFileSize, RolloverTimeInterval timeInterval) {
        this.maximumFileSize = maximumFileSize;
        this.rolloverTimeInterval = timeInterval;   
        this.startTime = Calendar.getInstance();
    }

    /** 
     * {@inheritDoc}
     */
    public String getRolloverIdentifier() {
        return DATE_FORMAT.format(new Date());
    }

    /** 
     * {@inheritDoc}
     */
    public boolean isRolloverRequired(long numberOfCharsOutput) {
        boolean rolloverRequired;
        
        if (maximumFileSize > 0 && numberOfCharsOutput >= maximumFileSize) {
            rolloverRequired = true;
        } else {
            rolloverRequired = isTimeBasedRolloverRequired();
        }
        
        if (rolloverRequired) {
            startTime = Calendar.getInstance();
        }
        
        return rolloverRequired;
    } 
    
    private boolean isTimeBasedRolloverRequired() {
        Calendar currentTime = Calendar.getInstance();
        long start;
        long current;
        
        switch (rolloverTimeInterval) {
            case SECONDS: {
                start = startTime.get(Calendar.SECOND);
                current = currentTime.get(Calendar.SECOND);
                break;
            }
            case MINUTES: {
                start = startTime.get(Calendar.MINUTE);
                current = currentTime.get(Calendar.MINUTE);
                break;
            }
            case HOURS: {
                start = startTime.get(Calendar.HOUR);
                current = currentTime.get(Calendar.HOUR);
                break;
            }
            case DAYS: 
            {
                start = startTime.get(Calendar.DAY_OF_MONTH);
                current = currentTime.get(Calendar.DAY_OF_MONTH);
                break;
            }
            case WEEKS: {
                start = startTime.get(Calendar.WEEK_OF_YEAR);
                current = currentTime.get(Calendar.WEEK_OF_YEAR);
                break;
            }
            case MONTHS: {
                start = startTime.get(Calendar.MONTH);
                current = currentTime.get(Calendar.MONTH);
                break;
            }
            case NEVER: 
            default: {
                return false;    
            }                   
        }
        
        return start != current;
    }
}
