/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.serviceability.tracing;

import java.io.StringWriter;

import com.springsource.server.serviceability.output.AbstractEntry;
import com.springsource.server.serviceability.output.Destination;
import com.springsource.server.serviceability.output.EntryPublisher;
import com.springsource.server.serviceability.output.Level;

/**
 * A TraceEntry represents a single entry in the Server's trace. A TraceEntry is published to one or more
 * {@link Destination Destinations} via an {@link EntryPublisher}.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * TraceEntry is immutable and therefore thread-safe.
 * 
 * @author Damilola Senbanjo
 * @author Sam Brannen
 * @author Christopher Frost
 * @author Andy Wilkinson
 * @since 1.0
 */
public class TraceEntry extends AbstractEntry {

    /**
     * The format string used for formatting trace entries.
     */
    private static final String FORMAT_STRING = "%s %s %s %s %s";

    private final String message;

    public TraceEntry(String className, String methodName, int lineNumber, String traceMessage, Throwable throwable, Level level) {
        this(className, methodName, lineNumber, traceMessage, throwable, level, level.toString());
    }

    /**
     * Creates a new <code>TraceEntry</code>.
     * 
     * @param className the name of the class being traced.
     * @param methodName the method that triggered trace.
     * @param lineNumber the line number that trace was triggered at.
     * @param traceMessage the trace message itself.
     * @param throwable A Throwable, the stack trace of which is to be included with the entry
     */
    public TraceEntry(String className, String methodName, int lineNumber, String traceMessage, Throwable throwable, Level level,
        String levelIdentifier) {
        super(level, throwable, buildSource(className, methodName, lineNumber), levelIdentifier);

        String formattedMessage = String.format(FORMAT_STRING, super.getFormattedTimestamp(), super.getFormattedThreadId(),
            super.getFormattedSource(), super.getFormattedLevelIdentifier(), traceMessage);

        if (throwable != null) {
            StringWriter stackWriter = new StringWriter();
            stackWriter.append(formattedMessage);
            stackWriter.append(LINE_SEPARATOR);
            stackWriter.append(AbstractEntry.getThrowableStackTraceAsString(throwable));

            formattedMessage = stackWriter.toString();
        }

        this.message = formattedMessage;
    }

    public TraceEntry(String className, String traceMessage, Level level, String levelIdentifier) {
        this(className, null, -1, traceMessage, null, level, levelIdentifier);
    }
    
    public TraceEntry(String className, String traceMessage, Throwable throwable, Level level, String levelIdentifier) {
        this(className, null, -1, traceMessage, throwable, level, levelIdentifier);
    }

    private static String buildSource(String className, String methodName, int lineNumber) {
        StringBuilder builder = new StringBuilder(className);

        if (methodName != null) {
            builder.append(".").append(methodName);
        }

        if (lineNumber > -1) {
            builder.append("(").append(lineNumber).append(")");
        }

        return builder.toString();
    }

    public String getText() {
        return this.message;
    }
}
