/* 
 * This file is part of the SpringSource Application Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource Application Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource Application Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource Application Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.serviceability.tracing.aspects;

import org.aspectj.lang.JoinPoint;

import com.springsource.server.serviceability.output.Level;
import com.springsource.server.serviceability.tracing.EntryExitTraceLevel;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * Adds entry/exit tracing to engine classes.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * Implementation is stateless.
 * 
 * @author Rob Harrop
 * @since 1.0
 */

public final aspect EntryExitTraceAspect pertypewithin(com.springsource.server..* && !com.springsource.server.serviceability.tracing..* && !com.springsource.server.serviceability.ffdc..*) {

    private volatile Tracer tracer;

    pointcut annotatedOperation() : execution (@EntryExitTraceLevel * *(..));

    pointcut duringEntryExitTracing() : cflow(within(EntryExitTraceAspect) && adviceexecution());
    
    pointcut excludedFromStandardTrace() : duringEntryExitTracing() || annotatedOperation();

    pointcut infoLevelTracedMethodExecution() : execution(public * *(..)) && !excludedFromStandardTrace();

    pointcut debugLevelTracedMethodExecution() : execution(!public * *(..)) && !excludedFromStandardTrace();

    pointcut customLevelTracedMethodExecution(Level level) :
        execution ( @EntryExitTraceLevel !@EntryExitTraceLevel(Level.NONE)* *(..)) &&
        @annotation(EntryExitTraceLevel(level)) &&
        !duringEntryExitTracing();

    //
    // Info tracing
    //
    
    before() : infoLevelTracedMethodExecution() {
        Tracer t = tracer(thisJoinPointStaticPart);
        t.traceEntry(Level.INFO, "%s", thisJoinPointStaticPart.getSignature());
    }

    after() returning : infoLevelTracedMethodExecution() {
        Tracer t = tracer(thisJoinPointStaticPart);
        t.traceExit(Level.INFO, "%s", thisJoinPointStaticPart.getSignature());
    }

    after() throwing(Throwable throwable) : infoLevelTracedMethodExecution() {
        Tracer t = tracer(thisJoinPointStaticPart);
        t.traceExit(Level.INFO, "%s", throwable, thisJoinPointStaticPart.getSignature());
    }

    //
    // Debug tracing
    //
    
    before() : debugLevelTracedMethodExecution() {
        Tracer t = tracer(thisJoinPointStaticPart);
        t.traceEntry(Level.DEBUG, "%s", thisJoinPointStaticPart.getSignature());
    }

    after() returning : debugLevelTracedMethodExecution() {
        Tracer t = tracer(thisJoinPointStaticPart);
        t.traceExit(Level.DEBUG, "%s", thisJoinPointStaticPart.getSignature());
    }

    after() throwing(Throwable throwable) : debugLevelTracedMethodExecution() {
        Tracer t = tracer(thisJoinPointStaticPart);
        t.traceExit(Level.DEBUG, "%s", throwable, thisJoinPointStaticPart.getSignature());
    }
    
    //
    // Custom tracing levels
    //
    
    before(Level level) : customLevelTracedMethodExecution(level) {
        Tracer t = tracer(thisJoinPointStaticPart);
        t.traceEntry(level, "%s", thisJoinPointStaticPart.getSignature());
    }

    after(Level level) returning : customLevelTracedMethodExecution(level) {
        Tracer t = tracer(thisJoinPointStaticPart);
        t.traceExit(level, "%s", thisJoinPointStaticPart.getSignature());
    }

    after(Level level) throwing(Throwable throwable) : customLevelTracedMethodExecution(level) {
        Tracer t = tracer(thisJoinPointStaticPart);
        t.traceExit(level, "%s", throwable, thisJoinPointStaticPart.getSignature());
    }
    
    private Tracer tracer(JoinPoint.StaticPart sp) {
        if (this.tracer == null) {
            Class<?> type = sp.getSignature().getDeclaringType();
            this.tracer = TracerFactory.createTracer(type);
        }
        return this.tracer;
    }
}
