/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.serviceability.tracing.internal;

import java.io.PrintStream;
import java.util.Locale;

import com.springsource.server.serviceability.output.Entry;
import com.springsource.server.serviceability.output.Level;
import com.springsource.server.serviceability.tracing.TraceEntry;

/**
 * A TracingPrintStreamWrapper wraps a PrintStream instance and logs all data that
 * is written to it. The data is encapsulated in a
 * {@link com.springsource.server.serviceability.output.Entry} instance and published
 * using a {@link com.springsource.server.serviceability.output.EntryPublisher}. 
 * <p />
 *
 * <strong>Concurrent Semantics</strong><br />
 * TracingPrintStreamWrapper is thread-safe.
 *
 * @author Andy Wilkinson
 * @since 1.0
 */
class TracingPrintStreamWrapper extends PrintStream {

    private final ThreadLocal<StringBuilder> entryBuilders;
    private final EntryPublisherAccessor accessor;
    private final String sourceName;
    private final Level outputLevel;
    
    /**
     * Creates a new TracingPrintStreamWrapper for the given PrintStream. Data written
     * to the stream is published via the given EntryPublisher with each published Entry
     * having the supplied source name and the supplied Level.
     * @param printStream The PrintStream instance to wrap
     * @param accessor The accessor that will be used to obtain the entry publisher via which entries will be published
     * @param sourceName The source name to be of the created entries
     * @param outputLevel The Level of the created entries
     */
    TracingPrintStreamWrapper(PrintStream printStream, EntryPublisherAccessor accessor, String sourceName, Level outputLevel) {
        super(printStream);
        
        entryBuilders = new ThreadLocal<StringBuilder>() {
            public StringBuilder initialValue() {
                return new StringBuilder();
            }
        };
        this.accessor = accessor;
        this.sourceName = sourceName;
        this.outputLevel = outputLevel;
    }
    
    @Override
    public PrintStream append(char c) {
        if (c == '\n' || c == '\r') {
            createEntryAndPublish(entryBuilders.get());
        } else {
            entryBuilders.get().append(c);
        }
        return this;
    }
    
    private boolean internalAppend(char c) {
        if (c == '\n' || c == '\r') {
            createEntryAndPublish(entryBuilders.get());
            return true;
        } else {
            entryBuilders.get().append(c);
            return false;
        }
    }

    @Override
    public PrintStream append(CharSequence csq, int start, int end) {
        
        for (int i = start; i < end; i++) {
            boolean publishedEntry = internalAppend(csq.charAt(i));
            if (publishedEntry && i < (end - 1)) {
                char c = csq.charAt(i + 1);
                if (c == '\n' || c == '\r') {
                    i++;
                }
            }
        }

        return this;
    }

    @Override
    public PrintStream append(CharSequence csq) {
        return append(csq, 0, csq.length());
    }

    @Override
    public boolean checkError() {
        flush();
        return false;
    }

    @Override
    public void close() {
        
    }

    @Override
    public void flush() {
        accessor.getCurrentEntryPublisher().flush();
    }

    @Override
    public PrintStream format(Locale l, String format, Object... args) {
        print(String.format(l, format, args));
        return this;
    }

    @Override
    public PrintStream format(String format, Object... args) {
        print(String.format(format, args));
        return this;
    }

    @Override
    public void print(boolean b) {
        entryBuilders.get().append(b);
    }

    @Override
    public void print(char c) {
        append(c);
    }

    @Override
    public void print(char[] s) {
        append(new String(s));
    }

    @Override
    public void print(double d) {
        entryBuilders.get().append(d);
    }

    @Override
    public void print(float f) {
        entryBuilders.get().append(f);
    }

    @Override
    public void print(int i) {
        entryBuilders.get().append(i);
    }

    @Override
    public void print(long l) {
        entryBuilders.get().append(l);
    }

    @Override
    public void print(Object obj) {
        if (obj == null) {
            entryBuilders.get().append(obj);
        } else {
            print(obj.toString().toCharArray());
        }
    }

    @Override
    public void print(String s) {
        append(s);
    }

    @Override
    public PrintStream printf(Locale l, String format, Object... args) {
        return format(l, format, args);
    }

    @Override
    public PrintStream printf(String format, Object... args) {
        return format(format, args);
    }

    @Override
    public void println() {
        createEntryAndPublish(entryBuilders.get());
    }

    @Override
    public void println(boolean x) {
        print(x);
        createEntryAndPublish(entryBuilders.get());
    }

    @Override
    public void println(char x) {
        print(x);
        createEntryAndPublish(entryBuilders.get());
    }

    @Override
    public void println(char[] x) {
        print(x);
        createEntryAndPublish(entryBuilders.get());
    }

    @Override
    public void println(double x) {
        print(x);
        createEntryAndPublish(entryBuilders.get());
    }

    @Override
    public void println(float x) {
        print(x);
        createEntryAndPublish(entryBuilders.get());
    }

    @Override
    public void println(int x) {
        print(x);
        createEntryAndPublish(entryBuilders.get());
    }

    @Override
    public void println(long x) {
        print(x);
        createEntryAndPublish(entryBuilders.get());
    }

    @Override
    public void println(Object x) {
        print(x);
        createEntryAndPublish(entryBuilders.get());
    }

    @Override
    public void println(String x) {
        print(x);
        createEntryAndPublish(entryBuilders.get());
    }

    @Override
    protected void setError() {
        
    }

    @Override
    public void write(byte[] buf, int off, int len) {
        byte[] outputBytes = new byte[len];
        System.arraycopy(buf, off, outputBytes, 0, len);
        print(new String(outputBytes));
    }

    @Override
    public void write(int b) {
        if (b == '\n' || b == '\r') {
            createEntryAndPublish(entryBuilders.get());
        } else {
            entryBuilders.get().append(new String(new byte[] {(byte)b}));
        }
    }   
    
    private void createEntryAndPublish(final StringBuilder stringBuilder) {
    	Entry entry = new TraceEntry(sourceName, stringBuilder.toString(), outputLevel, outputLevel.toString());        
    	accessor.getCurrentEntryPublisher().publish(entry);
        entryBuilders.remove();
    }
}
