/* 
 * This file is part of the SpringSource Application Platform.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource Application Platform is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource Application Platform is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource Application Platform. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.serviceability.tracing.slf4j;

import org.slf4j.Marker;
import org.slf4j.helpers.MessageFormatter;
import org.slf4j.spi.LocationAwareLogger;

import com.springsource.server.serviceability.output.Level;
import com.springsource.server.serviceability.tracing.Tracer;

/**
 * An adapter class for using a Platform {@link Tracer} as an SLF4J {@link LocationAwareLogger}.<p/>
 * <strong>Concurrent Semantics</strong><br />
 * This class is <strong>thread-safe</strong>.
 * 
 * @author Andy Wilkinson
 * @since 1.0
 */
class TracerToLoggerAdapter implements LocationAwareLogger {

    private final String name;

    private final Tracer tracer;

    /**
     * Create a new adapter for the class identified by the supplied name
     * 
     * @param name The name of the tracer, typically the name of the class that it will trace
     * @param tracer The {@link Tracer} to adapt into a {@link LocationAwareLogger}
     */
    TracerToLoggerAdapter(String name, Tracer tracer) {
        this.name = name;
        this.tracer = tracer;
    }

    /**
     * {@inheritDoc}
     */
    public void debug(String msg) {
        trace(Level.DEBUG, msg);
    }

    /**
     * {@inheritDoc}
     */
    public void debug(String format, Object arg) {
        trace(Level.DEBUG, MessageFormatter.format(format, arg));
    }

    /**
     * {@inheritDoc}
     */
    public void debug(String format, Object[] argArray) {
        trace(Level.DEBUG, MessageFormatter.arrayFormat(format, argArray));
    }

    /**
     * {@inheritDoc}
     */
    public void debug(String msg, Throwable t) {
        trace(Level.DEBUG, msg, t);
    }

    /**
     * {@inheritDoc}
     */
    public void debug(String format, Object arg1, Object arg2) {
        trace(Level.DEBUG, MessageFormatter.format(format, arg1, arg2));
    }

    /**
     * {@inheritDoc}
     */
    public void error(String msg) {
        trace(Level.ERROR, msg);
    }

    /**
     * {@inheritDoc}
     */
    public void error(String format, Object arg) {
        trace(Level.ERROR, MessageFormatter.format(format, arg));
    }

    /**
     * {@inheritDoc}
     */
    public void error(String format, Object[] argArray) {
        trace(Level.ERROR, MessageFormatter.arrayFormat(format, argArray));
    }

    /**
     * {@inheritDoc}
     */
    public void error(String msg, Throwable t) {
        trace(Level.ERROR, msg, t);
    }

    /**
     * {@inheritDoc}
     */
    public void error(String format, Object arg1, Object arg2) {
        trace(Level.ERROR, MessageFormatter.format(format, arg1, arg2));
    }

    /**
     * {@inheritDoc}
     */
    public String getName() {
        return name;
    }

    /**
     * {@inheritDoc}
     */
    public void info(String msg) {
        trace(Level.INFO, msg);
    }

    /**
     * {@inheritDoc}
     */
    public void info(String format, Object arg) {
        trace(Level.INFO, MessageFormatter.format(format, arg));
    }

    /**
     * {@inheritDoc}
     */
    public void info(String format, Object[] argArray) {
        trace(Level.INFO, MessageFormatter.arrayFormat(format, argArray));
    }

    /**
     * {@inheritDoc}
     */
    public void info(String msg, Throwable t) {
        trace(Level.INFO, msg, t);
    }

    /**
     * {@inheritDoc}
     */
    public void info(String format, Object arg1, Object arg2) {
        trace(Level.INFO, MessageFormatter.format(format, arg1, arg2));
    }

    /**
     * {@inheritDoc}
     */
    public boolean isDebugEnabled() {
        return tracer.isDebugEnabled();
    }

    /**
     * {@inheritDoc}
     */
    public boolean isErrorEnabled() {
        return tracer.isErrorEnabled();
    }

    /**
     * {@inheritDoc}
     */
    public boolean isInfoEnabled() {
        return tracer.isInfoEnabled();
    }

    /**
     * {@inheritDoc}
     */
    public boolean isTraceEnabled() {
        return tracer.isVerboseEnabled();
    }

    /**
     * {@inheritDoc}
     */
    public boolean isWarnEnabled() {
        return tracer.isWarningEnabled();
    }

    /**
     * {@inheritDoc}
     */
    public void trace(String msg) {
        trace(Level.VERBOSE, msg);
    }

    /**
     * {@inheritDoc}
     */
    public void trace(String format, Object arg) {
        trace(Level.VERBOSE, MessageFormatter.format(format, arg));
    }

    /**
     * {@inheritDoc}
     */
    public void trace(String format, Object[] argArray) {
        trace(Level.VERBOSE, MessageFormatter.arrayFormat(format, argArray));
    }

    /**
     * {@inheritDoc}
     */
    public void trace(String msg, Throwable t) {
        trace(Level.VERBOSE, msg, t);
    }

    /**
     * {@inheritDoc}
     */
    public void trace(String format, Object arg1, Object arg2) {
        trace(Level.VERBOSE, MessageFormatter.format(format, arg1, arg2));
    }

    /**
     * {@inheritDoc}
     */
    public void warn(String msg) {
        trace(Level.WARN, msg);
    }

    /**
     * {@inheritDoc}
     */
    public void warn(String format, Object arg) {
        trace(Level.WARN, MessageFormatter.format(format, arg));
    }

    /**
     * {@inheritDoc}
     */
    public void warn(String format, Object[] argArray) {
        trace(Level.WARN, MessageFormatter.arrayFormat(format, argArray));
    }

    /**
     * {@inheritDoc}
     */
    public void warn(String msg, Throwable t) {
        trace(Level.WARN, msg, t);
    }

    /**
     * {@inheritDoc}
     */
    public void warn(String format, Object arg1, Object arg2) {
        trace(Level.WARN, MessageFormatter.format(format, arg1, arg2));
    }

    /**
     * {@inheritDoc}
     */
    public void log(Marker marker, String fqcn, int slf4jLevel, String message, Throwable t) {
        Level level;

        switch (slf4jLevel) {
            case LocationAwareLogger.DEBUG_INT:
                level = Level.DEBUG;
                break;
            case LocationAwareLogger.ERROR_INT:
                level = Level.ERROR;
                break;
            case LocationAwareLogger.INFO_INT:
                level = Level.INFO;
                break;
            case LocationAwareLogger.TRACE_INT:
                level = Level.VERBOSE;
                break;
            case LocationAwareLogger.WARN_INT:
                level = Level.WARN;
                break;
            default:
                level = Level.VERBOSE;
                break;
        }

        trace(level, message, t);
    }

    /**
     * {@inheritDoc}
     */
    public void debug(Marker marker, String msg) {
        debug(msg);

    }

    /**
     * {@inheritDoc}
     */
    public void debug(Marker marker, String format, Object arg) {
        debug(format, arg);
    }

    /**
     * {@inheritDoc}
     */
    public void debug(Marker marker, String format, Object[] argArray) {
        debug(format, argArray);

    }

    /**
     * {@inheritDoc}
     */
    public void debug(Marker marker, String msg, Throwable t) {
        debug(msg, t);
    }

    /**
     * {@inheritDoc}
     */
    /**
     * {@inheritDoc}
     */
    public void debug(Marker marker, String format, Object arg1, Object arg2) {
        debug(format, arg1, arg2);
    }

    /**
     * {@inheritDoc}
     */
    public void error(Marker marker, String msg) {
        error(msg);
    }

    /**
     * {@inheritDoc}
     */
    public void error(Marker marker, String format, Object arg) {
        error(format, arg);
    }

    /**
     * {@inheritDoc}
     */
    public void error(Marker marker, String format, Object[] argArray) {
        error(format, argArray);
    }

    /**
     * {@inheritDoc}
     */
    public void error(Marker marker, String msg, Throwable t) {
        error(msg, t);
    }

    /**
     * {@inheritDoc}
     */
    public void error(Marker marker, String format, Object arg1, Object arg2) {
        error(format, arg1, arg2);
    }

    /**
     * {@inheritDoc}
     */
    public void info(Marker marker, String msg) {
        info(msg);
    }

    /**
     * {@inheritDoc}
     */
    public void info(Marker marker, String format, Object arg) {
        info(format, arg);
    }

    /**
     * {@inheritDoc}
     */
    public void info(Marker marker, String format, Object[] argArray) {
        info(format, argArray);
    }

    /**
     * {@inheritDoc}
     */
    public void info(Marker marker, String msg, Throwable t) {
        info(msg, t);
    }

    /**
     * {@inheritDoc}
     */
    public void info(Marker marker, String format, Object arg1, Object arg2) {
        info(format, arg1, arg2);
    }

    /**
     * {@inheritDoc}
     */
    public boolean isDebugEnabled(Marker marker) {
        return isDebugEnabled();
    }

    /**
     * {@inheritDoc}
     */
    public boolean isErrorEnabled(Marker marker) {
        return isErrorEnabled();
    }

    /**
     * {@inheritDoc}
     */
    public boolean isInfoEnabled(Marker marker) {
        return isInfoEnabled();
    }

    /**
     * {@inheritDoc}
     */
    public boolean isTraceEnabled(Marker marker) {
        return isTraceEnabled();
    }

    /**
     * {@inheritDoc}
     */
    public boolean isWarnEnabled(Marker marker) {
        return isWarnEnabled();
    }

    /**
     * {@inheritDoc}
     */
    public void trace(Marker marker, String msg) {
        trace(msg);
    }

    /**
     * {@inheritDoc}
     */
    public void trace(Marker marker, String format, Object arg) {
        trace(format, arg);
    }

    /**
     * {@inheritDoc}
     */
    public void trace(Marker marker, String format, Object[] argArray) {
        trace(format, argArray);
    }

    /**
     * {@inheritDoc}
     */
    public void trace(Marker marker, String msg, Throwable t) {
        trace(msg, t);
    }

    /**
     * {@inheritDoc}
     */
    public void trace(Marker marker, String format, Object arg1, Object arg2) {
        trace(format, arg1, arg2);
    }

    /**
     * {@inheritDoc}
     */
    public void warn(Marker marker, String msg) {
        warn(msg);
    }

    /**
     * {@inheritDoc}
     */
    public void warn(Marker marker, String format, Object arg) {
        warn(format, arg);
    }

    /**
     * {@inheritDoc}
     */
    public void warn(Marker marker, String format, Object[] argArray) {
        warn(format, argArray);
    }

    /**
     * {@inheritDoc}
     */
    public void warn(Marker marker, String msg, Throwable t) {
        warn(msg, t);
    }

    /**
     * {@inheritDoc}
     */
    public void warn(Marker marker, String format, Object arg1, Object arg2) {
        warn(format, arg1, arg2);
    }

    private void trace(Level level, String message) {
        trace(level, message, null);
    }

    private void trace(Level level, String message, Throwable t) {
        tracer.trace(level, message, t);
    }
}
