/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.servlet.admin.service.impl;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.osgi.framework.BundleContext;
import org.springframework.web.multipart.MultipartFile;

import com.springsource.server.deployer.core.ApplicationDeployer;
import com.springsource.server.deployer.core.DeploymentException;
import com.springsource.server.deployer.core.DeploymentIdentity;
import com.springsource.server.deployer.core.ApplicationDeployer.DeploymentOptions;
import com.springsource.server.io.core.work.WorkAreaManager;
import com.springsource.server.kernel.core.Subsystem;
import com.springsource.server.kernel.core.SubsystemRegistry;
import com.springsource.server.management.proxy.application.ServerApplicationInfo;
import com.springsource.server.management.proxy.application.StandardServerApplicationInfoAdaptor;
import com.springsource.server.management.proxy.core.ManagementUtil;
import com.springsource.server.management.proxy.server.ServerPropertiesResourceInfo;
import com.springsource.server.management.proxy.serviceability.DestinationServiceInfo;
import com.springsource.server.servlet.admin.service.ApplicationManagerService;

import com.springsource.server.bootstrap.BootstrapConfiguration;

/**
 * Standard implementation of {@link ApplicationManagerService}.
 * 
 * @author Christopher Frost
 * @author Sam Brannen
 * @since 1.0
 */
public class StandardApplicationManagerService implements ApplicationManagerService {

    private static final String ADMIN_APPLICATION_NAME = "server.admin.web";

    private static final Log LOGGER = LogFactory.getLog(StandardApplicationManagerService.class);

    private final ApplicationDeployer applicationDeployer;
    
    private final BootstrapConfiguration bootstrapConfiguration;

    private final BundleContext bundleContext;

    private final File stagingDir;

    /**
     * @param applicationDeployer
     * @param subsystemRegistry
     * @param workAreaManager
     * @param bundleContext
     */
    public StandardApplicationManagerService(ApplicationDeployer applicationDeployer, SubsystemRegistry subsystemRegistry,
        WorkAreaManager workAreaManager, BootstrapConfiguration bootstrapConfiguration, BundleContext bundleContext) {
        this.applicationDeployer = applicationDeployer;
        this.bundleContext = bundleContext;
        Subsystem subsystem = subsystemRegistry.getSubsystemByClass(applicationDeployer.getClass());
        this.stagingDir = workAreaManager.getWorkDirectory(subsystem).newChild("upload").createDirectory().toFile();
        this.bootstrapConfiguration = bootstrapConfiguration;
    }

    /**
     * {@inheritDoc}
     */
    public List<ServerApplicationInfo> findAppInfoForDeployedApplications() {
        List<ServerApplicationInfo> formattedServerApplicationInfos = new ArrayList<ServerApplicationInfo>();
        List<ServerApplicationInfo> unformattedServerApplicationInfos = ManagementUtil.listAll(this.bundleContext, ServerApplicationInfo.class);
        for (ServerApplicationInfo unformattedServerApplicationInfo : unformattedServerApplicationInfos) {
            formattedServerApplicationInfos.add(new StandardServerApplicationInfoAdaptor(unformattedServerApplicationInfo));
        }
        return formattedServerApplicationInfos;
    }

    /**
     * {@inheritDoc}
     */
    public String deploy(MultipartFile multipartFile) {
        File dest = new File(this.stagingDir.getAbsolutePath(), multipartFile.getOriginalFilename());
        try {
            dest.createNewFile();
            multipartFile.transferTo(dest);
            try {
                this.applicationDeployer.deploy(dest.toURI(), new DeploymentOptions(true, true));
                return "Application deployed";
            } catch (DeploymentException e) {
                dest.delete();
                return "Deployment Error: " + e.getMessage();
            }
        } catch (IOException e) {
            String msg = "An error occurred while uploading a file to the deployment directory [" + dest.getAbsolutePath() + "]";
            LOGGER.warn(msg);
            return msg;
        }
    }

    /**
     * {@inheritDoc}
     */
    public String undeploy(String path) {
        if (path.contains(ADMIN_APPLICATION_NAME)) {
            LOGGER.warn("There has been an attempt to remove the admin console from the application, which has been ignored.");
            return "The server admin console cannot remove itself!";
        } else {
            URI uri;
			try {
				uri = new URI(path);
			} catch (URISyntaxException e) {
	            return "Unable to process undeployment URI: " + e.getMessage();
			}
            try {
                DeploymentIdentity deploymentIdentity = this.applicationDeployer.getDeploymentIdentity(uri);
                if (deploymentIdentity != null) {
                    this.applicationDeployer.undeploy(deploymentIdentity);
                } else {
                	new File(uri.getPath()).delete();
                }
                return "Application undeployed";
            } catch (DeploymentException e) {
            	new File(uri.getPath()).delete();
                return "Undeployment Error: " + e.getMessage();
            } 
        }
    }

    /**
     * {@inheritDoc}
     */
    public Map<String, String> getServerProperties() {
        List<ServerPropertiesResourceInfo> props = ManagementUtil.listAll(this.bundleContext, ServerPropertiesResourceInfo.class);
        Map<String, String> fields = new TreeMap<String, String>();
        
        if(props.size() >= 1){ // There must be at least one info object returned, if there is more than one then something strange is happening but it dosn't actually stop us doing our job here.
            ServerPropertiesResourceInfo properties = props.get(0);
            for (Object key : properties.getProperties().keySet()) {
                fields.put(key.toString(), properties.getProperties().getProperty(key.toString()));
            }
        }
        
        fields.put("Pickup Directory", this.bootstrapConfiguration.getPickUpDirectory());
        
        /*
        if(reply.size() != 1){
            LOGGER.warn("There should only ever be one ServerPropertiesResourceInfo object registered for each instance of the server, this is not the case [" + reply.size() + "]. Reported server properties may be incorrect.");
        }
        */
        
        return fields;
    }
    
    /**
     * {@inheritDoc}
     */
    public Map<String, String> getDestinations() {
        Map<String, String> fields = new TreeMap<String, String>();
        List<DestinationServiceInfo> destinations = ManagementUtil.listAll(this.bundleContext, DestinationServiceInfo.class);
        
        for(DestinationServiceInfo destinationServiceInfo : destinations){
            if(!destinationServiceInfo.getName().equals("console")){
                fields.put(destinationServiceInfo.getName(), destinationServiceInfo.getDestinationString());
            }
        }
        
        /*
        if(reply.size() != 1){
            LOGGER.warn("There should only ever be one ServerPropertiesResourceInfo object registered for each instance of the server, this is not the case [" + reply.size() + "]. Reported server properties may be incorrect.");
        }
        */
        
        return fields;
        
    }

}
