/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.servlet.core;

import javax.servlet.Servlet;

/**
 * <code>DynamicContext</code> is an extension of {@link Context}, which defines additional functionality for
 * creating and deploying <em>dynamic</em> contexts within the {@link ServletContainer}.
 * 
 * <p>
 * Note: {@link #addServlet(String, Servlet)} and {@link #addServletMapping(String, String)} are intended solely for
 * dynamic creation of contexts (i.e., for contexts that are created programmatically as opposed to as part of a WAR
 * deployment).
 * </p>
 * 
 * <p>
 * Concrete implementations must provide reasonable implementations of:
 * </p>
 * <ul>
 * <li>{@link Object#equals(Object)}</li>
 * <li>{@link Object#hashCode()}</li>
 * <li>{@link Object#toString()}</li>
 * </ul>
 * 
 * @author Sam Brannen
 * @author Rob Harrop
 * @since 1.0
 * @see ServletContainer
 */
public interface DynamicContext extends Context {

    /**
     * Adds the supplied, preconfigured {@link Servlet} to this web application context and registers it under the
     * provided <code>name</code>.
     * 
     * @param name the name of the servlet to add to this context.
     * @param servlet the servlet to add to this context.
     * @see #addServletMapping(String, String)
     */
    void addServlet(String name, Servlet servlet);

    /**
     * Adds a new servlet mapping to this web application context for the named servlet under the provided
     * <code>mappingPath</code>.
     * 
     * @param servletName the name of the servlet to be mapped.
     * @param mappingPath the path under which the named servlet should be mapped.
     * @see #addServlet(String, Servlet)
     */
    void addServletMapping(String servletName, String mappingPath);

    // XXX Consider adding support for Servlet Filters.
    // XXX Consider adding support for init-params for Servlets.
    // XXX Consider adding support for init-params for Servlet contexts.

}
