/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.servlet.tomcat;

import java.io.IOException;

import javax.servlet.Servlet;
import javax.servlet.ServletConfig;
import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServlet;

import com.springsource.server.serviceability.output.Level;
import com.springsource.server.serviceability.tracing.EntryExitTraceLevel;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * ProxyServlet serves as a proxy for servlets which are programmatically created and dynamically added to a web
 * application context in contrast to typical deployment via a <em>web.xml</em> deployment descriptor.
 * <p>
 * <u>Usage</u>:
 * </p>
 * <ul>
 * <li>Instantiate and configure a target servlet.</li>
 * <li>Configure a ProxyServlet in the web application context and pass it an <code>init-param</code> with
 * {@link ProxyServlet#INIT_PARAM_PROXY_SERVLET} as the name and a unique servlet name as the value.</li>
 * <li>Set the target servlet instance as an attribute in the {@link ServletContext} under the unique servlet name
 * which you passed to the proxy servlet in the previous step.</li>
 * </ul>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * If a ProxyServlet is correctly {@link #init() initialized} (i.e., once and only once, prior to servicing any
 * requests), it is considered to be thread-safe in a Servlet environment, assuming the proxied servlet is itself
 * thread-safe.
 * 
 * @author Rob Harrop
 * @author Sam Brannen
 * @since 1.0
 */
public class ProxyServlet extends HttpServlet {

    public static final String INIT_PARAM_PROXY_SERVLET = "proxy.servlet";

    private static final Tracer TRACER = TracerFactory.createTracer(ProxyServlet.class);

    private static final long serialVersionUID = -1094847028391655868L;

    private volatile Servlet target;

    @Override
    public void init(ServletConfig config) throws ServletException {
        super.init(config);
        final String servletName = config.getInitParameter(INIT_PARAM_PROXY_SERVLET);
        this.target = (Servlet) config.getServletContext().getAttribute(servletName);
    }

    @Override
    @EntryExitTraceLevel(Level.DEBUG)
    public void service(ServletRequest request, ServletResponse response) throws ServletException, IOException {
        if (this.target == null) {
            String msg = "Configuration error: ProxyServlet must be configured with a target servlet.";
            TRACER.traceError(msg);
            throw new ServletException(msg);
        } else {
            this.target.service(request, response);
        }
    }

}
