/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.servlet.tomcat;

import javax.servlet.ServletContext;

import org.springframework.core.style.ToStringCreator;
import org.springframework.util.Assert;

import com.springsource.server.servlet.core.Context;

/**
 * Tomcat-based implementation of the {@link Context} interface.
 * 
 * @author Sam Brannen
 * @author Rob Harrop
 * @since 1.0
 * @see Context
 * @see TomcatServletContainer
 */
public class TomcatContext implements Context {

    private final org.apache.catalina.Context catalinaContext;

    /**
     * Creates a new TomcatContext with the supplied Catalina {@link org.apache.catalina.Context context}.
     * 
     * @param catalinaContext the underlying Catalina context to use; must not be <code>null</code>
     */
    public TomcatContext(final org.apache.catalina.Context catalinaContext) {
        Assert.notNull(catalinaContext, "'catalinaContext' must not be null");
        this.catalinaContext = catalinaContext;
    }

    /**
     * {@inheritDoc}
     */
    public String getPath() {
        return this.catalinaContext.getPath();
    }

    /**
     * {@inheritDoc}
     */
    public String getContextRoot() {
        return this.catalinaContext.getDocBase();
    }

    /**
     * {@inheritDoc}
     */
    public ServletContext getServletContext() {
        return this.catalinaContext.getServletContext();
    }

    /**
     * Gets the Catalina {@link org.apache.catalina.Context context} to which this context delegates. For internal use
     * only!
     * 
     * @return the Catalina context
     */
    protected org.apache.catalina.Context getCatalinaContext() {
        return this.catalinaContext;
    }

    /**
     * Delegates to {@link org.apache.catalina.Context#getReloadable()}.
     * 
     * @see Context#isReloadable()
     */
    public boolean isReloadable() {
        return this.catalinaContext.getReloadable();
    }

    /**
     * {@inheritDoc}
     */
    public boolean reload() {
        boolean reloaded = false;
        if (isReloadable() && this.catalinaContext.getAvailable()) {
            try {
                this.catalinaContext.reload();
                reloaded = true;
            } catch (IllegalStateException ise) {
                reloaded = false;
            }
        }
        return reloaded;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result + ((getPath() == null) ? 0 : getPath().hashCode());
        return result;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        TomcatContext that = (TomcatContext) obj;
        if (getPath() == null) {
            if (that.getPath() != null) {
                return false;
            }
        } else if (!getPath().equals(that.getPath())) {
            return false;
        }
        return true;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String toString() {
        return new ToStringCreator(this)//
        .append("path", getPath())//
        .append("contextRoot", getContextRoot())//
        .append("reloadable", isReloadable())//
        .toString();
    }

}
