/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.servlet.tomcat;

import static com.springsource.server.io.core.FileSystemUtils.createDirectoryIfNecessary;
import static com.springsource.server.servlet.core.ServletContainerLogEvent.CREATING_CONNECTOR;
import static com.springsource.server.servlet.core.ServletContainerLogEvent.CREATING_CONTEXT;
import static com.springsource.server.servlet.core.ServletContainerLogEvent.FAILED_TO_CREATE_CONTEXT;
import static com.springsource.server.servlet.core.ServletContainerLogEvent.FAILED_TO_DEPLOY_WAR;
import static com.springsource.server.servlet.core.ServletContainerLogEvent.FAILED_TO_START_CONTAINER;
import static com.springsource.server.servlet.core.ServletContainerLogEvent.FAILED_TO_START_CONTEXT_DUE_TO_EXISING_CONTEXT_PATH;
import static com.springsource.server.servlet.core.ServletContainerLogEvent.FAILED_TO_START_CONTEXT_DUE_TO_UNKNOWN_ERROR;
import static com.springsource.server.servlet.core.ServletContainerLogEvent.FAILED_TO_STOP_CONTAINER;
import static com.springsource.server.servlet.core.ServletContainerLogEvent.PORT_IN_USE;
import static com.springsource.server.servlet.core.ServletContainerLogEvent.REMOVING_CONTEXT;
import static com.springsource.server.servlet.core.ServletContainerLogEvent.STARTING_CONTAINER;
import static com.springsource.server.servlet.core.ServletContainerLogEvent.STARTING_CONTEXT;
import static com.springsource.server.servlet.core.ServletContainerLogEvent.STOPPING_CONTAINER;
import static com.springsource.server.servlet.core.util.ServletContainerUtils.convertContextPath;
import static com.springsource.server.servlet.core.util.ServletContainerUtils.getDisplaySuitableContextPath;
import static com.springsource.server.servlet.core.util.ServletContainerUtils.getRawContextPath;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.net.URL;
import java.net.URLConnection;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.concurrent.atomic.AtomicBoolean;

import javax.servlet.ServletContext;

import org.apache.catalina.Cluster;
import org.apache.catalina.Container;
import org.apache.catalina.Engine;
import org.apache.catalina.Executor;
import org.apache.catalina.Host;
import org.apache.catalina.Lifecycle;
import org.apache.catalina.LifecycleException;
import org.apache.catalina.LifecycleListener;
import org.apache.catalina.Loader;
import org.apache.catalina.Realm;
import org.apache.catalina.Valve;
import org.apache.catalina.connector.Connector;
import org.apache.catalina.core.ContainerBase;
import org.apache.catalina.core.StandardContext;
import org.apache.catalina.core.StandardHost;
import org.apache.catalina.ha.CatalinaCluster;
import org.apache.catalina.ha.ClusterDeployer;
import org.apache.catalina.ha.ClusterListener;
import org.apache.catalina.ha.ClusterManager;
import org.apache.catalina.realm.MemoryRealm;
import org.apache.catalina.startup.ContextConfig;
import org.apache.catalina.startup.Embedded;
import org.apache.catalina.tribes.Channel;
import org.apache.catalina.tribes.ChannelInterceptor;
import org.apache.catalina.tribes.ChannelReceiver;
import org.apache.catalina.tribes.ChannelSender;
import org.apache.catalina.tribes.ManagedChannel;
import org.apache.catalina.tribes.MembershipService;
import org.apache.catalina.tribes.transport.MultiPointSender;
import org.apache.catalina.tribes.transport.ReplicationTransmitter;
import org.apache.catalina.util.LifecycleSupport;
import org.apache.catalina.valves.AccessLogValve;
import org.apache.catalina.valves.ErrorReportValve;
import org.apache.catalina.valves.Constants.AccessLog;
import org.apache.coyote.ProtocolHandler;
import org.apache.naming.resources.DirContextURLStreamHandler;
import org.apache.naming.resources.FileDirContext;
import org.apache.naming.resources.WARDirContext;
import org.apache.tomcat.util.IntrospectionUtils;
import org.osgi.framework.Bundle;
import org.osgi.framework.ServiceRegistration;
import org.osgi.service.url.AbstractURLStreamHandlerService;
import org.osgi.service.url.URLConstants;
import org.springframework.util.Assert;
import org.springframework.util.ClassUtils;
import org.springframework.util.ReflectionUtils;
import org.springframework.util.StringUtils;

import com.springsource.server.bootstrap.util.NetUtils;
import com.springsource.server.io.core.FileSystemUtils;
import com.springsource.server.osgi.framework.OsgiFramework;
import com.springsource.server.osgi.framework.Shutdown;
import com.springsource.server.serviceability.ffdc.NonNull;
import com.springsource.server.serviceability.logging.LogService;
import com.springsource.server.serviceability.logging.LogServiceFactory;
import com.springsource.server.serviceability.output.Level;
import com.springsource.server.serviceability.tracing.EntryExitTraceLevel;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;
import com.springsource.server.servlet.core.Context;
import com.springsource.server.servlet.core.DynamicContext;
import com.springsource.server.servlet.core.ServletContainer;
import com.springsource.server.servlet.core.ServletContainerException;
import com.springsource.server.servlet.tomcat.internal.ApplicationNameTrackingValve;
import com.springsource.server.servlet.tomcat.internal.ServerErrorReportValve;
import com.springsource.server.servlet.tomcat.internal.StandardTomcatServletContainerConfiguration;
import com.springsource.server.servlet.tomcat.loader.OsgiWebappLoader;
import com.springsource.server.servlet.tomcat.loader.WebappClassLoader;
import com.springsource.server.servlet.tomcat.loader.WebappLoader;

/**
 * Tomcat-based implementation of the {@link ServletContainer} interface, which uses an {@link Embedded embedded Tomcat
 * engine} as the underlying Servlet Engine.
 * 
 * @author Sam Brannen
 * @author Rob Harrop
 * @since 1.0
 * @see TomcatDynamicContext
 */
public class TomcatServletContainer implements ServletContainer, Lifecycle {

    private static final LogService LOG_SERVICE = LogServiceFactory.getLogService();

    private static final Tracer TRACER = TracerFactory.createTracer(TomcatServletContainer.class);

    private final StandardTomcatServletContainerConfiguration config;

    private final Executor executor;

    protected final Embedded embedded;

    protected final Host host;

    private final Object contextsMonitor = new Object();

    private final AtomicBoolean started = new AtomicBoolean();

    private final ServiceRegistration urlHandlerServiceRegistration;

    private final LifecycleSupport lifecycle = new LifecycleSupport(this);

    /**
     * Constructs a new TomcatServletContainer with the supplied {@link TomcatServletContainerConfiguration
     * configuration}, {@link ApplicationNameTrackingValve}, {@link Executor}, and {@link OsgiFramework}.
     * 
     * @param config the configuration parameters to use to create this servlet container.
     * @param applicationNameTrackingValve the <em>application name</em> tracking valve
     * @param executor the Catalina {@link Executor} to use for processing incoming requests.
     * @param osgiFramework the {@link OsgiFramework} to use for registering services
     */
    public TomcatServletContainer(@NonNull StandardTomcatServletContainerConfiguration config,
        @NonNull ApplicationNameTrackingValve applicationNameTrackingValve, @NonNull Executor executor, @NonNull OsgiFramework osgiFramework) {

        Assert.hasText(config.getName(), "Failed to instantiate the ServletContainer: 'name' must not be null or empty");
        Assert.hasText(config.getConfigDir(), "Failed to instantiate the ServletContainer: 'config' directory must not be null or empty");

        this.config = config;

        this.executor = executor;
        try {
            this.executor.start();
        } catch (LifecycleException e) {
            throw new ServletContainerException("Unable to start executor.", e);
        }

        TRACER.traceDebug("Starting ServletContainer with configuration: %s.", this.toString());

        // Create an embedded servlet engine
        this.embedded = createEmbedded(config.getConfigDir());
        Engine engine = this.embedded.createEngine();
        engine.setName("springsource.server.catalina");
        engine.setDefaultHost(config.getHostName());
        engine.setJvmRoute(config.getJvmRoute());
        if (hasClusterConfiguration()) {
            engine.setCluster(configureCluster());
        }

        // Create a default virtual host
        StandardHost standardHost = createAndConfigureHost(engine, config.getHostName(), config.getApplicationDir(), config.getWorkDir());
        addValvesToHost(standardHost, createErrorReportValve(), createLoggingValve(null), applicationNameTrackingValve);
        this.host = standardHost;

        // Install the assembled container hierarchy
        this.embedded.addEngine(engine);
        configureListeners();
        configureConnectors();
        this.lifecycle.fireLifecycleEvent(INIT_EVENT, null);

        this.urlHandlerServiceRegistration = registerUrlStreamHandlerServiceForJndiUrls(osgiFramework);
    }

    private void shutdownIfPortNotAvailable(int port) {
        if (port > 0 && !NetUtils.isPortAvailable(port)) {
            LOG_SERVICE.log(PORT_IN_USE, port);
            Shutdown.immediateShutdown();
        }
    }

    /**
     * Configures {@link LifecycleListener lifecycle listeners} based on properties retrieved from
     * {@link TomcatServletContainerConfiguration#getListeners()}.
     */
    protected void configureListeners() {
        Set<String> excludes = new HashSet<String>();
        excludes.add("className");
        excludes.add("enabled");

        for (Map<String, String> properties : getConfiguration().getListeners()) {
            TRACER.traceVerbose("Processing LifecycleListener properties: %s.", properties);
            if (Boolean.parseBoolean(properties.get("enabled"))) {
                String className = properties.get("className");

                try {
                    TRACER.traceInfo("Creating LifecycleListener of type [%s].", className);
                    LifecycleListener listener = (LifecycleListener) ClassUtils.forName(className).newInstance();

                    for (String name : properties.keySet()) {
                        if (!excludes.contains(name)) {
                            setProperty(listener, name, properties.get(name));
                        }
                    }

                    addLifecycleListener(listener);
                } catch (Exception e) {
                    TRACER.traceWarning("Unable to create LifecycleListener of type [%s].", className);
                }
            }
        }
    }

    /**
     * Configures {@link Connector connectors} based on properties retrieved from
     * {@link TomcatServletContainerConfiguration#getConnectors()}.
     */
    protected void configureConnectors() {

        Set<String> excludes = new HashSet<String>();
        excludes.add("enabled");
        excludes.add("executor");
        excludes.add("protocol");

        for (Map<String, String> properties : getConfiguration().getConnectors()) {
            TRACER.traceVerbose("Processing Connector properties: %s.", properties);
            if (Boolean.parseBoolean(properties.get("enabled"))) {
                TRACER.traceInfo("Creating Connector based on properties: %s.", properties);
                String strPort = properties.get("port");
                String protocol = properties.get("protocol");
                String scheme = properties.get("scheme");
                if (!StringUtils.hasText(scheme)) {
                    scheme = "http";
                }
                int port = -1;
                try {
                    port = Integer.parseInt(strPort);
                } catch (Exception e) {
                    /* ignore */
                }
                if (port > 0) {
                    shutdownIfPortNotAvailable(port);
                    LOG_SERVICE.log(CREATING_CONNECTOR, protocol, scheme, port);
                    try {
                        Connector connector = new Connector(protocol);
                        for (String name : properties.keySet()) {
                            if (!excludes.contains(name)) {
                                setProperty(connector, name, properties.get(name));
                            }
                        }

                        configureExecutorForConnector(connector);
                        this.embedded.addConnector(connector);
                    } catch (Exception e) {
                        TRACER.traceWarning("Unable to create Connector for protocol [%s] and port [%s].", protocol, strPort);
                    }
                }
            }
        }
    }

    private boolean hasClusterConfiguration() {
        return config.getCluster().containsKey("className");
    }

    @SuppressWarnings("unchecked")
    private Cluster configureCluster() {
        Map<String, ? extends Object> clusterConfig = config.getCluster();
        TRACER.traceVerbose("Configuring cluster based on properties: %s", clusterConfig);

        try {
            Set<String> reservedKeys = new HashSet<String>();
            reservedKeys.add("className");
            reservedKeys.add("manager");
            reservedKeys.add("channel");
            reservedKeys.add("valves");
            reservedKeys.add("deployer");
            reservedKeys.add("clusterListeners");

            String clusterClassName = (String) clusterConfig.get("className");
            Cluster cluster = (Cluster) ClassUtils.forName(clusterClassName).newInstance();
            setProperties((Map<String, String>) clusterConfig, reservedKeys, cluster);

            configureManager((Map<String, String>) clusterConfig.get("manager"), cluster);
            configureChannel((Map<String, Object>) clusterConfig.get("channel"), cluster);
            configureValves((List<Map<String, String>>) clusterConfig.get("valves"), cluster);
            if (cluster instanceof CatalinaCluster) {
                configureDeployer((Map<String, String>) clusterConfig.get("deployer"), (CatalinaCluster) cluster);
                configureClusterListeners((List<Map<String, String>>) clusterConfig.get("clusterListeners"), (CatalinaCluster) cluster);
            }

            return cluster;
        } catch (Exception e) {
            e.printStackTrace();
            TRACER.traceWarning("Unable to create Cluster.");
        }
        return null;
    }

    private void configureManager(Map<String, String> managerConfig, Cluster cluster) throws LinkageError, Exception {
        if (managerConfig == null) {
            return;
        }

        Set<String> reservedKeys = new HashSet<String>();
        reservedKeys.add("className");

        String managerClassName = managerConfig.get("className");
        ClusterManager manager = (ClusterManager) ClassUtils.forName(managerClassName).newInstance();
        setProperties(managerConfig, reservedKeys, manager);
        callMethod(cluster, "setManagerTemplate", manager, ClusterManager.class);
    }

    @SuppressWarnings("unchecked")
    private void configureChannel(Map<String, ? extends Object> channelConfig, Cluster cluster) throws LinkageError, Exception {
        if (channelConfig == null) {
            return;
        }

        Set<String> reservedKeys = new HashSet<String>();
        reservedKeys.add("className");
        reservedKeys.add("membership");
        reservedKeys.add("receiver");
        reservedKeys.add("sender");
        reservedKeys.add("interceptors");

        String channelClassName = (String) channelConfig.get("className");
        Channel channel = (Channel) ClassUtils.forName(channelClassName).newInstance();
        setProperties((Map<String, String>) channelConfig, reservedKeys, channel);
        if (channel instanceof ManagedChannel) {
            configureMembership((Map<String, String>) channelConfig.get("membership"), (ManagedChannel) channel);
            configureReceiver((Map<String, String>) channelConfig.get("receiver"), (ManagedChannel) channel);
            configureSender((Map<String, Object>) channelConfig.get("sender"), (ManagedChannel) channel);
        }
        configureInterceptors((List<Map<String, String>>) channelConfig.get("interceptors"), channel);
        callMethod(cluster, "setChannel", channel, Channel.class);
    }

    private void configureMembership(Map<String, String> membershipConfig, ManagedChannel channel) throws LinkageError, Exception {
        if (membershipConfig == null) {
            return;
        }

        Set<String> reservedKeys = new HashSet<String>();
        reservedKeys.add("className");

        String membershipClassName = membershipConfig.get("className");
        MembershipService membership = (MembershipService) ClassUtils.forName(membershipClassName).newInstance();
        setProperties(membershipConfig, reservedKeys, membership);
        callMethod(channel, "setMembershipService", membership, MembershipService.class);
    }

    private void configureReceiver(Map<String, String> receiverConfig, ManagedChannel channel) throws LinkageError, Exception {
        if (receiverConfig == null) {
            return;
        }

        Set<String> reservedKeys = new HashSet<String>();
        reservedKeys.add("className");

        String receiverClassName = receiverConfig.get("className");
        ChannelReceiver receiver = (ChannelReceiver) ClassUtils.forName(receiverClassName).newInstance();
        setProperties(receiverConfig, reservedKeys, receiver);
        callMethod(channel, "setChannelReceiver", receiver, ChannelReceiver.class);
    }

    @SuppressWarnings("unchecked")
    private void configureSender(Map<String, ? extends Object> senderConfig, ManagedChannel channel) throws LinkageError, Exception {
        if (senderConfig == null) {
            return;
        }

        Set<String> reservedKeys = new HashSet<String>();
        reservedKeys.add("className");
        reservedKeys.add("transport");

        String senderClassName = (String) senderConfig.get("className");
        ChannelSender sender = (ChannelSender) ClassUtils.forName(senderClassName).newInstance();
        setProperties((Map<String, String>) senderConfig, reservedKeys, sender);
        if (sender instanceof ReplicationTransmitter) {
            configureTransport((Map<String, String>) senderConfig.get("transport"), (ReplicationTransmitter) sender);
        }
        callMethod(channel, "setChannelSender", sender, ChannelSender.class);
    }

    private void configureTransport(Map<String, String> transportConfig, ReplicationTransmitter sender) throws LinkageError, Exception {
        if (transportConfig == null) {
            return;
        }

        Set<String> reservedKeys = new HashSet<String>();
        reservedKeys.add("className");

        String transportClassName = transportConfig.get("className");
        MultiPointSender transport = (MultiPointSender) ClassUtils.forName(transportClassName).newInstance();
        setProperties(transportConfig, reservedKeys, transport);
        callMethod(sender, "setTransport", transport, MultiPointSender.class);
    }

    private void configureInterceptors(List<Map<String, String>> interceptorsConfig, Channel channel) throws LinkageError, Exception {
        if (interceptorsConfig == null) {
            return;
        }
        Set<String> reservedKeys = new HashSet<String>();
        reservedKeys.add("className");

        for (Map<String, String> interceptorConfig : interceptorsConfig) {
            String interceptorClassName = interceptorConfig.get("className");
            ChannelInterceptor interceptor = (ChannelInterceptor) ClassUtils.forName(interceptorClassName).newInstance();
            setProperties(interceptorConfig, reservedKeys, interceptor);
            callMethod(channel, "addInterceptor", interceptor, ChannelInterceptor.class);
        }
    }

    private void configureValves(List<Map<String, String>> valvesConfig, Cluster cluster) throws LinkageError, Exception {
        if (valvesConfig == null) {
            return;
        }
        Set<String> reservedKeys = new HashSet<String>();
        reservedKeys.add("className");

        for (Map<String, String> valveConfig : valvesConfig) {
            String valveClassName = valveConfig.get("className");
            Valve valve = (Valve) ClassUtils.forName(valveClassName).newInstance();
            setProperties(valveConfig, reservedKeys, valve);
            callMethod(cluster, "addValve", valve, Valve.class);
        }
    }

    private void configureDeployer(Map<String, String> deployerConfig, CatalinaCluster cluster) throws LinkageError, Exception {
        if (deployerConfig == null) {
            return;
        }

        Set<String> reservedKeys = new HashSet<String>();
        reservedKeys.add("className");

        String deployerClassName = deployerConfig.get("className");
        ClusterDeployer deployer = (ClusterDeployer) ClassUtils.forName(deployerClassName).newInstance();
        setProperties(deployerConfig, reservedKeys, deployer);
        callMethod(cluster, "setClusterDeployer", deployer, ClusterDeployer.class);
    }

    private void configureClusterListeners(List<Map<String, String>> clusterListenersConfig, CatalinaCluster cluster) throws LinkageError, Exception {
        if (clusterListenersConfig == null) {
            return;
        }

        Set<String> reservedKeys = new HashSet<String>();
        reservedKeys.add("className");

        for (Map<String, String> clusterListenerConfig : clusterListenersConfig) {
            String listenerClassName = clusterListenerConfig.get("className");
            ClusterListener listener = (ClusterListener) ClassUtils.forName(listenerClassName).newInstance();
            setProperties(clusterListenerConfig, reservedKeys, listener);
            callMethod(cluster, "addClusterListener", listener, ClusterListener.class);
        }
    }

    private Object callMethod(Object instance, String method, Object argument, Class<?> argumentType) throws Exception {
        return IntrospectionUtils.callMethod1(instance, method, argument, argumentType.getCanonicalName(),
            Thread.currentThread().getContextClassLoader());
    }

    private void setProperties(Map<String, String> config, Set<String> reservedKeys, Object instance) {
        for (Map.Entry<String, String> entry : config.entrySet()) {
            if (!reservedKeys.contains(entry.getKey())) {
                setProperty(instance, entry.getKey(), entry.getValue());
            }
        }
    }

    private void setProperty(Object instance, String name, String value) {
        TRACER.traceDebug("Setting property %s for object of type %s", name, instance.getClass().getCanonicalName());
        if (!IntrospectionUtils.setProperty(instance, name, value)) {
            TRACER.traceError("Unable to set property %s for object of type %s", name, instance.getClass().getCanonicalName());
        }
    }

    /**
     * Attempts to configure the {@link Executor} for the supplied {@link Connector connector}. If the supplied
     * connector's {@link ProtocolHandler} does not have a <code>setExecutor(Executor)</code> method, this method will
     * have no effect.
     */
    protected void configureExecutorForConnector(Connector connector) {
        ProtocolHandler protocolHandler = connector.getProtocolHandler();
        try {
            Method method = protocolHandler.getClass().getDeclaredMethod("setExecutor", java.util.concurrent.Executor.class);
            if (method != null) {
                ReflectionUtils.makeAccessible(method);
                ReflectionUtils.invokeMethod(method, protocolHandler, new Object[] { this.executor });
                TRACER.traceInfo("Set Executor for ProtocolHandler [%s].", protocolHandler);
            }
        } catch (Exception e) {
            TRACER.traceInfo("Unable to set Executor for ProtocolHandler [%s].", protocolHandler);
        }
    }

    private ServiceRegistration registerUrlStreamHandlerServiceForJndiUrls(OsgiFramework osgiFramework) {
        Properties properties = new Properties();
        properties.put(URLConstants.URL_HANDLER_PROTOCOL, "jndi");
        final DirContextURLStreamHandler handler = new DirContextURLStreamHandler();
        return osgiFramework.getBundleContext().registerService("org.osgi.service.url.URLStreamHandlerService",
            new AbstractURLStreamHandlerService() {

                @Override
                public URLConnection openConnection(URL u) throws IOException {
                    return new URL(null, u.toExternalForm(), handler).openConnection();
                }
            }, properties);
    }

    /**
     * Create the {@link Realm security realm} for the embedded servlet engine.
     */
    protected MemoryRealm createSecurityRealm() {
        MemoryRealm memoryRealm = new MemoryRealm();
        memoryRealm.setPathname(getConfiguration().getTomcatUsersXml());
        return memoryRealm;
    }

    /**
     * Creates a new {@link Loader}. The type created is dependent on whether or not the supplied {@link Bundle bundle}
     * is <code>null</code>.
     * 
     * <p>
     * If the supplied {@link Bundle bundle} is non-null, an {@link OsgiWebappLoader} will be created using the supplied
     * <code>bundle</code> and <code>webBundle</code>; otherwise, a new {@link WebappLoader} will be created which uses
     * a {@link WebappClassLoader} as the underlying class loader, and the <code>ClassLoader</code> for this
     * <code>ServletContainer</code> will be used as the parent class loader for the <code>WebappClassLoader</code>.
     * </p>
     * 
     * @param contextBundle the <em>context</em> bundle for which a loader should be created; may be <code>null</code>
     * @param webBundle the <em>web</em> bundle for which a loader should be created; may be <code>null</code>
     * @return a new <code>Loader</code>
     */
    protected Loader createLoader(Bundle contextBundle, Bundle webBundle) {
        if (contextBundle == null) {
            ClassLoader classLoader = getClass().getClassLoader();
            TRACER.traceVerbose("Setting parent ClassLoader for the WebappLoader to [%s].", classLoader);
            WebappLoader webappLoader = new WebappLoader(classLoader);
            webappLoader.setLoaderClass(WebappClassLoader.class.getName());
            webappLoader.setDelegate(false);
            return webappLoader;
        } else {
            return new OsgiWebappLoader(contextBundle, webBundle);
        }
    }

    /**
     * Creates an {@link Embedded embedded} Tomcat servlet engine with the supplied <code>name</code> and
     * <code>path</code>.
     * 
     * @param path the absolute path to the embedded servlet engine's <em>home</em> directory
     * @return a newly configured and instantiated embedded servlet engine
     */
    protected Embedded createEmbedded(final String path) {
        Assert.notNull(path, "'path' must not be null");
        TRACER.traceDebug("Creating embedded servlet engine with home directory [%s].", path);
        Embedded embedded = new Embedded(createSecurityRealm());
        embedded.setCatalinaBase(path);
        embedded.setCatalinaHome(path);
        embedded.addExecutor(this.executor);
        return embedded;
    }

    /**
     * Creates a virtual host with the supplied <code>hostName</code>, <code>appBase</code>, and <code>workDir</code>.
     * 
     * @param engine the servlet engine to which the host should be added
     * @param hostName the host name for the virtual host to create
     * @param appBase the absolute path to the virtual host's <em>application base</em> directory in the local file
     *        system; must <strong>not</strong> be <code>null</code> or <em>empty</em>
     * @param workDir the absolute path to the virtual host's <em>working</em> directory in the local file system; must
     *        <strong>not</strong> be <code>null</code> or <em>empty</em>
     * @return a fully configured and registered virtual host
     */
    protected StandardHost createAndConfigureHost(final Engine engine, final String hostName, final String appBase, final String workDir) {
        Assert.notNull(engine, "'engine' must not be null");
        Assert.hasText(hostName, "'hostName' must not be null or empty");
        Assert.hasText(appBase, "'appBase' must not be null or empty");
        Assert.hasText(workDir, "'workDir' must not be null or empty");
        TRACER.traceDebug("Deploying virtual host [%s] with application directory [%s] and working directory [%s].", hostName, appBase, workDir);

        StandardHost host = (StandardHost) this.embedded.createHost(hostName, appBase);
        host.setWorkDir(workDir);
        host.setUnpackWARs(true);
        host.setDeployOnStartup(false);
        host.setLiveDeploy(false);
        host.setAutoDeploy(false);
        host.setXmlValidation(false);
        host.setXmlNamespaceAware(false);
        // Set ErrorReportValve class name to null, so that we can add our own.
        host.setErrorReportValveClass(null);

        engine.addChild(host);
        return host;
    }

    /**
     * Adds the supplied <code>valves</code> to the supplied <code>host</code>.
     */
    protected void addValvesToHost(final StandardHost host, Valve... valves) {
        Assert.notNull(host, "'host' must not be null");
        for (Valve valve : valves) {
            if (valve != null) {
                TRACER.traceInfo("Adding valve [%s] to host [%s].", valve.getInfo(), host);
                host.addValve(valve);
            }
        }
    }

    /**
     * Creates an {@link ErrorReportValve error report valve}.
     */
    protected Valve createErrorReportValve() {
        return new ServerErrorReportValve();
    }

    /**
     * Creates an {@link AccessLogValve access logging valve}.
     * 
     * @param prefix text to use to construct the prefix for the log file
     */
    @EntryExitTraceLevel(Level.DEBUG)
    protected Valve createLoggingValve(String prefix) {
        StringBuilder prefixStringBuilder = new StringBuilder(getName()).append(".");
        if (prefix != null) {
            prefixStringBuilder.append(StringUtils.hasText(prefix) ? prefix : "_").append(".");
        }

        String format = "short".equals(getConfiguration().getAccessLogFormat()) ? AccessLog.COMMON_ALIAS : AccessLog.COMBINED_ALIAS;
        AccessLogValve accessLogValve = new AccessLogValve();
        accessLogValve.setDirectory(getConfiguration().getAccessLogDir());
        accessLogValve.setPrefix(prefixStringBuilder.toString());
        accessLogValve.setSuffix(".log");
        accessLogValve.setPattern(format);
        accessLogValve.setResolveHosts(false);
        return accessLogValve;
    }

    /**
     * {@inheritDoc}
     */
    @EntryExitTraceLevel(Level.VERBOSE)
    public StandardTomcatServletContainerConfiguration getConfiguration() {
        return this.config;
    }

    /**
     * {@inheritDoc}
     */
    public String getName() {
        return getConfiguration().getName();
    }

    /**
     * {@inheritDoc}
     */
    public void start() throws ServletContainerException {
        if (this.started.compareAndSet(false, true)) {
            try {
                // Notify our interested LifecycleListeners
                this.lifecycle.fireLifecycleEvent(BEFORE_START_EVENT, null);
                this.lifecycle.fireLifecycleEvent(START_EVENT, null);

                LOG_SERVICE.log(STARTING_CONTAINER);
                this.embedded.start();

                // Notify our interested LifecycleListeners
                this.lifecycle.fireLifecycleEvent(AFTER_START_EVENT, null);
            } catch (LifecycleException e) {
                LOG_SERVICE.log(FAILED_TO_START_CONTAINER, e);
                throw new ServletContainerException("Failed to start the TomcatServletContainer", e);
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    public void stop() throws ServletContainerException {
        if (this.started.compareAndSet(true, false)) {
            this.urlHandlerServiceRegistration.unregister();
            try {
                // Notify our interested LifecycleListeners
                this.lifecycle.fireLifecycleEvent(BEFORE_STOP_EVENT, null);
                this.lifecycle.fireLifecycleEvent(STOP_EVENT, null);

                LOG_SERVICE.log(STOPPING_CONTAINER);
                this.embedded.stop();

                // Notify our interested LifecycleListeners
                this.lifecycle.fireLifecycleEvent(AFTER_STOP_EVENT, null);
            } catch (LifecycleException e) {
                LOG_SERVICE.log(FAILED_TO_STOP_CONTAINER, e);
                throw new ServletContainerException("Failed to stop the TomcatServletContainer", e);
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    @EntryExitTraceLevel(Level.VERBOSE)
    public boolean isStarted() {
        return this.started.get();
    }

    /**
     * Asserts that the container has been started and otherwise throws an {@link IllegalStateException}.
     * 
     * @see #isStarted()
     */
    @EntryExitTraceLevel(Level.VERBOSE)
    protected void assertContainerIsStarted() {
        Assert.state(isStarted(), "This action cannot be performed on a ServletContainer that has not been started.");
    }

    /**
     * {@inheritDoc}
     */
    public void addLifecycleListener(LifecycleListener listener) {
        this.lifecycle.addLifecycleListener(listener);
    }

    /**
     * {@inheritDoc}
     */
    public LifecycleListener[] findLifecycleListeners() {
        return this.lifecycle.findLifecycleListeners();
    }

    /**
     * {@inheritDoc}
     */
    public void removeLifecycleListener(LifecycleListener listener) {
        this.lifecycle.removeLifecycleListener(listener);
    }

    /**
     * {@inheritDoc}
     */
    public DynamicContext createContext(String contextPath) throws ServletContainerException, IllegalStateException {
        return createContext(contextPath, (File) null);
    }

    /**
     * {@inheritDoc}
     */
    @EntryExitTraceLevel(Level.VERBOSE)
    public DynamicContext createContext(String contextPath, Bundle contextBundle, Bundle webBundle) throws ServletContainerException,
        IllegalStateException {
        return createContext(contextPath, null, contextBundle, webBundle);
    }

    /**
     * {@inheritDoc}
     */
    @EntryExitTraceLevel(Level.VERBOSE)
    public DynamicContext createContext(String contextPath, File contextRoot) throws ServletContainerException, IllegalStateException {
        return createContext(contextPath, contextRoot, (Bundle) null, (Bundle) null);
    }

    /**
     * {@inheritDoc}
     */
    public DynamicContext createContext(String contextPath, File contextRoot, Bundle contextBundle, Bundle webBundle)
        throws ServletContainerException, IllegalStateException {

        assertContainerIsStarted();

        String contextRootPath = contextRoot == null ? null : FileSystemUtils.convertToCanonicalPath(contextRoot.toURI());
        String displaySuitableContextPath = getDisplaySuitableContextPath(contextPath);
        try {
            contextPath = convertContextPath(contextPath);
            contextRootPath = determineAbsolutePathToContextRoot(contextRootPath, contextPath);
            createDirectoryIfNecessary(contextRootPath);
            LOG_SERVICE.log(CREATING_CONTEXT, displaySuitableContextPath);

            StandardContext catalinaContext = (StandardContext) this.embedded.createContext(contextPath, contextRootPath);
            catalinaContext.setLoader(createLoader(contextBundle, webBundle));
            catalinaContext.setPrivileged(false);
            catalinaContext.setReloadable(false);
            // Note: do NOT set the name of a context, as the name of a context must be its path!
            // catalinaContext.setName(???);
            String rawContextPath = getRawContextPath(contextPath);

            if (getConfiguration().perApplicationLogging()) {
                catalinaContext.addValve(createLoggingValve(rawContextPath.equals("") ? "ROOT" : rawContextPath));
            }

            setContextDefaults(catalinaContext);

            if (webBundle != null) {
                initializeContextResources(catalinaContext);
            }

            return new TomcatDynamicContext(catalinaContext);
        } catch (Exception e) {
            LOG_SERVICE.log(FAILED_TO_CREATE_CONTEXT, e, displaySuitableContextPath);
            throw e instanceof ServletContainerException ? (ServletContainerException) e : new ServletContainerException(
                "Failed to create web application context [" + displaySuitableContextPath + "] from directory [" + contextRootPath + "]", e);
        }
    }

    /**
     * Sets the configuration defaults for the supplied context.
     */
    @EntryExitTraceLevel(Level.VERBOSE)
    private void setContextDefaults(StandardContext context) {
        for (LifecycleListener lifecycleListener : ((Lifecycle) context).findLifecycleListeners()) {
            if (lifecycleListener instanceof ContextConfig) {
                ContextConfig contextConfig = (ContextConfig) lifecycleListener;
                contextConfig.setDefaultContextXml(getConfiguration().getDefaultContextXml());
                contextConfig.setDefaultWebXml(getConfiguration().getDefaultWebXml());
            }
        }
    }

    /**
     * Initializes the resources of the supplied {@link org.apache.catalina.Context context} after it has been created
     * but prior to adding it to a {@link Host}. Calling this method is necessary in order to have a fully functional
     * {@link ServletContext} for the supplied context before it is completely deployed. Note that the same (or a
     * similar) process will occur again, once the context is added to a host.
     */
    @EntryExitTraceLevel(Level.VERBOSE)
    private void initializeContextResources(org.apache.catalina.Context catalinaContext) {

        if (catalinaContext instanceof StandardContext) {
            try {
                StandardContext standardContext = (StandardContext) catalinaContext;
                Method getBasePathMethod = standardContext.getClass().getDeclaredMethod("getBasePath");
                getBasePathMethod.setAccessible(true);
                String basePath = (String) getBasePathMethod.invoke(standardContext);
                String docBase = catalinaContext.getDocBase();

                try {
                    if (docBase != null && docBase.endsWith(".war") && !new File(basePath).isDirectory()) {
                        standardContext.setResources(new WARDirContext());
                    } else {
                        standardContext.setResources(new FileDirContext());
                    }
                    Field parentField = ContainerBase.class.getDeclaredField("parent");
                    parentField.setAccessible(true);
                    parentField.set(standardContext, this.host);
                    standardContext.resourcesStart();

                } catch (IllegalArgumentException e) {
                    TRACER.traceWarning("Could not initialize context's resources", e);
                }
            } catch (Exception e) {
                TRACER.traceWarning("Could not initialize context's resources", e);
            }
        }
    }

    /**
     * Determines the absolute path to the supplied <code>contextRootPath</code>.
     * 
     * <ul>
     * <li>If the <code>contextRootPath</code> is absolute, it will be returned &quot;as is&quot;.</li>
     * <li>If the <code>contextRootPath</code> is <code>null</code>, the supplied <code>contextPath</code> will be
     * appended to the configured <em>application directory</em>, and the resulting absolute path will be returned.</li>
     * <li>Otherwise, the (relative) <code>contextRootPath</code> will be appended to the configured
     * <em>application directory</em>, and the resulting absolute path will be returned.</li>
     * </ul>
     * 
     * @param contextRootPath the context root path, possibly relative, absolute, or <code>null</code>
     * @param contextPath the context path; never <code>null</code>
     * @return the absolute path to the context root
     */
    @EntryExitTraceLevel(Level.VERBOSE)
    protected String determineAbsolutePathToContextRoot(String contextRootPath, @NonNull String contextPath) {
        contextPath = convertContextPath(contextPath);

        if (contextRootPath != null) {
            File file = new File(contextRootPath);
            if (!file.isAbsolute()) {
                file = new File(getConfiguration().getApplicationDir(), contextRootPath);
                contextRootPath = file.getAbsolutePath();
            }
        } else {
            File file = new File(getConfiguration().getApplicationDir(), contextPath);
            contextRootPath = file.getAbsolutePath();
        }

        return contextRootPath;
    }

    /**
     * {@inheritDoc}
     */
    @EntryExitTraceLevel(Level.VERBOSE)
    public boolean addContext(final DynamicContext context) throws ServletContainerException, IllegalStateException {
        return addContext(context, (Bundle) null);
    }

    /**
     * {@inheritDoc}
     */
    public boolean addContext(final DynamicContext context, Bundle webBundle) throws ServletContainerException, IllegalStateException {
        boolean added = true;
        assertContainerIsStarted();
        Assert.notNull(context, "'context' must not be null");
        if (context instanceof TomcatContext) {
            String contextPath = context.getPath();
            String displaySuitableContextPath = getDisplaySuitableContextPath(contextPath);
            LOG_SERVICE.log(STARTING_CONTEXT, displaySuitableContextPath);

            org.apache.catalina.Context catalinaContext = ((TomcatContext) context).getCatalinaContext();
            synchronized (this.contextsMonitor) {
                if (this.host.findChild(contextPath) != null) {
                    LOG_SERVICE.log(FAILED_TO_START_CONTEXT_DUE_TO_EXISING_CONTEXT_PATH, displaySuitableContextPath);
                    added = false;
                } else {
                    this.host.addChild(catalinaContext);
                    if (!catalinaContext.getAvailable()) {
                        LOG_SERVICE.log(FAILED_TO_START_CONTEXT_DUE_TO_UNKNOWN_ERROR, displaySuitableContextPath);
                        added = false;
                        // Context start-up failed, so remove any lingering remains.
                        removeContext(contextPath);
                    }
                }
            }
        } else {
            String msg = "Unsupported Context type [" + context.getClass().getName() + "]; only [" + TomcatContext.class.getName()
                + "] is supported.";
            TRACER.traceError("Failed to add context [%s] to the ServletContainer: %s", context, msg);
            throw new IllegalArgumentException("Failed to add context [" + context + "] to the ServletContainer: " + msg);
        }
        return added;
    }

    /**
     * {@inheritDoc}
     */
    public DynamicContext getContext(final String contextPath) throws ServletContainerException {
        DynamicContext context = null;
        try {
            Container container = this.host.findChild(convertContextPath(contextPath));
            if (container instanceof org.apache.catalina.Context) {
                context = new TomcatDynamicContext((org.apache.catalina.Context) container);
            }
        } catch (Exception e) {
            String displaySuitableContextPath = getDisplaySuitableContextPath(contextPath);
            TRACER.traceError("Failed to retrieve context [%s] from the ServletContainer.", e, displaySuitableContextPath);
            throw new ServletContainerException("Failed to retrieve context [" + displaySuitableContextPath + "] from the ServletContainer", e);
        }
        return context;
    }

    /**
     * {@inheritDoc}
     */
    public Set<Context> getContexts() throws ServletContainerException {
        try {
            Set<Context> contexts = new HashSet<Context>();
            Container[] children = this.host.findChildren();
            for (int i = 0; i < children.length; i++) {
                Container container = children[i];
                if (container instanceof org.apache.catalina.Context) {
                    contexts.add(new TomcatContext((org.apache.catalina.Context) container));
                }
            }
            return contexts;
        } catch (Exception e) {
            TRACER.traceError("Failed to retrieve contexts from the ServletContainer.", e);
            throw new ServletContainerException("Failed to retrieve contexts from the ServletContainer", e);
        }
    }

    /**
     * {@inheritDoc}
     */
    public boolean removeContext(String contextPath) throws ServletContainerException, IllegalStateException {
        assertContainerIsStarted();
        contextPath = convertContextPath(contextPath);
        String displaySuitableContextPath = getDisplaySuitableContextPath(contextPath);
        LOG_SERVICE.log(REMOVING_CONTEXT, displaySuitableContextPath);
        boolean removed = true;
        try {
            synchronized (this.contextsMonitor) {
                Container container = this.host.findChild(contextPath);
                if (container instanceof org.apache.catalina.Context) {
                    org.apache.catalina.Context context = (org.apache.catalina.Context) container;
                    ClassLoader classLoader = context.getLoader().getClassLoader();
                    try {
                        this.embedded.removeContext(context);
                    } catch (Exception e) {
                        TRACER.traceError("Encountered an error while trying to remove web application '%s'.", e, displaySuitableContextPath);
                        removed = false;
                    }
                    try {
                        ((ContainerBase) context).destroy();
                    } catch (Exception e) {
                        TRACER.traceError("Encountered an error while trying to remove web application '%s'.", e, displaySuitableContextPath);
                        removed = false;
                    }

                    if (classLoader != null) {
                        DirContextURLStreamHandler.unbind(classLoader);
                    }
                    cleanUpDeploymentArtifacts(contextPath);
                } else {
                    TRACER.traceWarning("'%s' does not refer to a managed web application.", displaySuitableContextPath);
                }
            }
        } catch (ServletContainerException e) {
            throw e;
        } catch (Exception e) {
            TRACER.traceError("Failed to remove context [%s] from the ServletContainer.", e, displaySuitableContextPath);
            throw new ServletContainerException("Failed to remove context [" + displaySuitableContextPath + "] from the ServletContainer", e);
        }
        return removed;
    }

    /**
     * Cleans up any lingering deployment artifacts for the web application with the supplied <code>contextPath</code>.
     * Should only be called in conjunction with {@link #removeContext(String) removal} of a context.
     * 
     * @param contextPath the context path for the application which should be cleaned
     */
    @EntryExitTraceLevel(Level.DEBUG)
    private void cleanUpDeploymentArtifacts(String contextPath) {
        try {
            final String docBase = getDocBase(contextPath);
            if (!"".equals(docBase) && !"/".equals(docBase)) {
                FileSystemUtils.deleteRecursively(new File(getConfiguration().getApplicationDir(), docBase + ".war"));
                FileSystemUtils.deleteRecursively(new File(getConfiguration().getApplicationDir(), docBase + ".WAR"));
                FileSystemUtils.deleteRecursively(new File(getConfiguration().getApplicationDir(), docBase));
                FileSystemUtils.deleteRecursively(new File(getConfiguration().getWorkDir(), docBase));
            }
        } catch (Exception e) {
            /* ignore */
        }
    }

    /**
     * Given a context path, get the config file name.
     */
    @EntryExitTraceLevel(Level.VERBOSE)
    protected String getDocBase(String path) {
        if (path == null || path.equals("")) {
            path = "ROOT";
        } else if (path.startsWith("/")) {
            path = path.substring(1);
        }
        return path;
    }

    /**
     * {@inheritDoc}
     */
    @EntryExitTraceLevel(Level.VERBOSE)
    public DynamicContext deployWar(String contextPath, File war) throws ServletContainerException, IllegalStateException {
        return deployWar(contextPath, war, (Bundle) null, (Bundle) null);
    }

    /**
     * {@inheritDoc}
     */
    public DynamicContext deployWar(final String contextPath, final File war, final Bundle contextBundle, final Bundle webBundle)
        throws ServletContainerException, IllegalStateException {
        assertContainerIsStarted();
        Assert.notNull(war, "'war' must not be null");
        String displaySuitableContextPath = getDisplaySuitableContextPath(contextPath);
        try {
            String canonicalPath = FileSystemUtils.convertToCanonicalPath(war.getPath());
            File contextRoot = new File(canonicalPath);
            DynamicContext context = createContext(contextPath, contextRoot, contextBundle, webBundle);
            if (!addContext(context, webBundle)) {
                throw new ServletContainerException("Failed to deploy WAR [" + displaySuitableContextPath + "] from resource [" + war
                    + "]: a web application with that context path already exists.");
            }
            return context;
        } catch (ServletContainerException sce) {
            LOG_SERVICE.log(FAILED_TO_DEPLOY_WAR, sce, displaySuitableContextPath);
            throw sce;
        } catch (Exception e) {
            LOG_SERVICE.log(FAILED_TO_DEPLOY_WAR, e, displaySuitableContextPath);
            throw new ServletContainerException("Failed to deploy WAR [" + displaySuitableContextPath + "] from resource [" + war + "].", e);
        }
    }

    /**
     * {@inheritDoc}
     */
    public String[] getPackageNamesOfDependenciesThatContainApplicationExposedTypes() {

        return new String[] {//
        // Servlet, JSP, and JSTL:
            "javax.servlet",//
            "javax.servlet.http",//
            "javax.servlet.resources",//
            "javax.servlet.jsp", //
            "javax.servlet.jsp.el",//
            "javax.servlet.jsp.resources",//
            "javax.servlet.jsp.tagext",//
            "javax.servlet.jsp.jstl.core",//
            "javax.servlet.jsp.jstl.fmt",//
            "javax.servlet.jsp.jstl.sql",//
            "javax.servlet.jsp.jstl.tlv",//
            // Apache EL Implementation:
            "org.apache.el",//
            "org.apache.el.lang",//
            "org.apache.el.parser",//
            "org.apache.el.util",//
            // Common Annotations:
            "javax.annotation",//
            "javax.annotation.security",//
            "javax.ejb",//
            "javax.el",//
            "javax.persistence",//
            "javax.xml.ws",//
            // Management:
            "javax.management",//
            "javax.management.loading",//
            "javax.management.modelmbean",//
            "javax.management.monitor",//
            "javax.management.openmbean",//
            "javax.management.relation",//
            "javax.management.remote",//
            "javax.management.remote.rmi",//
            "javax.management.timer",//
            // Naming:
            "javax.naming",//
            "javax.naming.directory",//
            "javax.naming.event",//
            "javax.naming.ldap",//
            "javax.naming.spi",//
            // Net:
            "javax.net",//
            "javax.net.ssl",//
            // RMI:
            "javax.rmi",//
            "javax.rmi.CORBA",//
            "javax.rmi.ssl",//
        };
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String toString() {
        return new StringBuilder()//
        .append("Name [").append(getName()).append("], ")//
        .append("Hostname [").append(getConfiguration().getHostName()).append("], ")//
        .append("Config Directory [").append(getConfiguration().getConfigDir()).append("], ")//
        .append("Application Directory [").append(getConfiguration().getApplicationDir()).append("], ")//
        .append("Work Directory [").append(getConfiguration().getWorkDir()).append("], ")//
        .append("Access Logs Directory [").append(getConfiguration().getAccessLogDir()).append("].")//
        .append("Access Log Format [").append(getConfiguration().getAccessLogFormat()).append("].")//
        .toString();
    }

}
