/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.servlet.tomcat.internal;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Writer;

import org.apache.catalina.connector.Request;
import org.apache.catalina.connector.Response;
import org.apache.catalina.util.RequestUtil;
import org.apache.catalina.valves.ErrorReportValve;
import org.springframework.util.FileCopyUtils;
import org.springframework.util.StringUtils;

/**
 * <code>ServerErrorReportValve</code> is a custom extension of {@link ErrorReportValve} which generates error
 * reports using SpringSource styles.
 * <p />
 * 
 * @author Sam Brannen
 * @since 1.0
 */
public class ServerErrorReportValve extends ErrorReportValve {

    /**
     * The descriptive information related to this implementation.
     */
    private static final String info = "com.springsource.server.servlet.tomcat.internal.ServerErrorReportValve/1.0";

    private String serverName = "SpringSource dm Server";

    private String serverVersion = "1.0";

    public String getServerName() {
        return this.serverName;
    }

    public void setServerName(String serverName) {
        this.serverName = serverName;
    }

    public String getServerVersion() {
        return this.serverVersion;
    }

    public void setServerVersion(String serverVersion) {
        this.serverVersion = serverVersion;
    }

    @Override
    public String getInfo() {
        return info;
    }

    protected String getServerInfo() {
        return getServerName() + " / " + getServerVersion();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void report(Request request, Response response, Throwable throwable) {

        // Do nothing on non-HTTP responses
        int statusCode = response.getStatus();

        // Do nothing on a 1xx, 2xx and 3xx status
        // Do nothing if anything has been written already
        if (statusCode < 400 || response.getContentCount() > 0) {
            return;
        }

        String message = RequestUtil.filter(response.getMessage());
        if (message == null) {
            message = "";
        }

        // Do nothing if there is no report for the specified status code
        String report = null;
        try {
            report = sm.getString("http." + statusCode, message);
        } catch (Throwable t) {
            ;
        }
        if (report == null) {
            return;
        }

        StringBuffer sb = new StringBuffer(
            "<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\" \"http://www.w3.org/TR/html4/loose.dtd\">");

        sb.append("<html><head>");
        sb.append("<title>").append(getPageTitle()).append("</title>");
        sb.append("<style><!--\n").append(CSS).append("\n--></style> ");
        sb.append("</head><body class=\"main\">");
        sb.append("<div id=\"page\">");

        appendHeader(sb);

        sb.append("<div id=\"container\">");
        sb.append("<div id=\"content-no-nav\">");

        sb.append("<h1>").append(sm.getString("errorReportValve.statusHeader", "" + statusCode, message)).append("</h1>");

        OddEven oddEven = new OddEven();
        sb.append("<table class=\"bordered-table\">");

        // --- TYPE ---
        sb.append("<tr><th class=\"sublevel1-" + oddEven + "\">type:&nbsp;</th>");
        sb.append("<td class=\"sublevel1-" + oddEven + "\">");
        if (throwable != null) {
            sb.append(sm.getString("errorReportValve.exceptionReport"));
        } else {
            sb.append(sm.getString("errorReportValve.statusReport"));
        }
        sb.append("</td>");
        sb.append("</tr>");
        oddEven.toggle();

        // --- MESSAGE ---
        if (StringUtils.hasText(message)) {
            sb.append("<tr><th class=\"sublevel1-" + oddEven + "\">").append(sm.getString("errorReportValve.message")).append(":&nbsp;</th>");
            sb.append("<td class=\"sublevel1-" + oddEven + "\">");
            sb.append("<u>").append(message).append("</u></td>");
            sb.append("</tr>");
            oddEven.toggle();
        }

        // --- DESCRIPTION ---
        sb.append("<tr><th class=\"sublevel1-" + oddEven + "\">").append(sm.getString("errorReportValve.description")).append(":&nbsp;</th>");
        sb.append("<td class=\"sublevel1-" + oddEven + "\">");
        sb.append("<u>").append(report).append("</u></td>");
        sb.append("</tr>");
        oddEven.toggle();

        // --- EXCEPTION ---
        if (throwable != null) {

            String stackTrace = getPartialServletStackTrace(throwable);
            sb.append("<tr><th class=\"sublevel1-" + oddEven + "\">").append(sm.getString("errorReportValve.exception")).append(":&nbsp;</th>");
            sb.append("<td class=\"sublevel1-" + oddEven + "\">");
            sb.append("<pre>").append(RequestUtil.filter(stackTrace)).append("</pre></td>");
            sb.append("</tr>");
            oddEven.toggle();

            int loops = 0;
            Throwable rootCause = throwable.getCause();
            while (rootCause != null && loops < 10) {
                stackTrace = getPartialServletStackTrace(rootCause);
                sb.append("<tr><th class=\"sublevel1-" + oddEven + "\">").append(sm.getString("errorReportValve.rootCause")).append(":&nbsp;</th>");
                sb.append("<td class=\"sublevel1-" + oddEven + "\">");
                sb.append("<pre>").append(RequestUtil.filter(stackTrace)).append("</pre></td>");
                sb.append("</tr>");
                // In case root cause is somehow heavily nested
                rootCause = rootCause.getCause();
                loops++;
                oddEven.toggle();
            }

            sb.append("<tr><th class=\"sublevel1-" + oddEven + "\">").append(sm.getString("errorReportValve.note")).append(":&nbsp;</th>");
            sb.append("<td class=\"sublevel1-" + oddEven + "\">");
            sb.append("<u>").append(sm.getString("errorReportValve.rootCauseInLogs", getServerName())).append("</u></td>");
            sb.append("</tr>");
            oddEven.toggle();
        }

        sb.append("</table>");

        sb.append("</div>"); // content
        sb.append("</div>"); // container
        appendFooter(sb);
        sb.append("</div>"); // page
        sb.append("</div></body></html>");

        try {
            try {
                response.setContentType("text/html");
                response.setCharacterEncoding("utf-8");
            } catch (Throwable t) {
                if (this.container.getLogger().isDebugEnabled()) {
                    this.container.getLogger().debug("status.setContentType", t);
                }
            }
            Writer writer = response.getReporter();
            if (writer != null) {
                // If writer is null, it's an indication that the response has
                // been hard committed already, which should never happen
                writer.write(sb.toString());
            }
        } catch (IOException e) {
            ;
        } catch (IllegalStateException e) {
            ;
        }
    }

    private void appendHeader(StringBuffer sb) {
        sb.append("<div id=\"header\">");
        sb.append("<div id=\"name-and-company\">");
        sb.append("<div id='site-name'>").append(getPageTitle()).append("</div>");
        sb.append("<div id='company-name'></div>");
        sb.append("</div>");
        sb.append("</div>");
    }

    private String getPageTitle() {
        String errorReportString = sm.getString("errorReportValve.errorReport");
        return getServerName() + (errorReportString != null && errorReportString.trim().length() > 0 ? " - " + errorReportString : "");
    }

    private void appendFooter(StringBuffer sb) {
        sb.append("<div id=\"footer-wrapper\">");
        sb.append("<div id=\"footer-left\"><a href=\"http://www.springsource.com\" title=\"SpringSource\" target=\"_blank\">SpringSource</a></div>");
        sb.append("<div id=\"footer-right\"></div> ");
        sb.append("</div>");
    }

    private static final String CSS;

    static {
        StringBuilder stringBuilder = new StringBuilder();
        try {
            InputStream inputStream = ServerErrorReportValve.class.getResourceAsStream("main.css");
            stringBuilder.append(FileCopyUtils.copyToString(new InputStreamReader(inputStream)));
            inputStream = ServerErrorReportValve.class.getResourceAsStream("colors.css");
            stringBuilder.append("\n").append(FileCopyUtils.copyToString(new InputStreamReader(inputStream)));
        } catch (IOException e) {
            e.printStackTrace();
        }
        CSS = stringBuilder.toString();
    }

    private static class OddEven {

        private boolean odd = true;

        public void toggle() {
            this.odd = !this.odd;
        }

        public String toString() {
            return this.odd ? "odd" : "even";
        }
    }
}
