/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.servlet.tomcat.internal;

import static com.springsource.server.io.core.FileSystemUtils.convertToCanonicalPath;
import static com.springsource.server.io.core.FileSystemUtils.createDirectoryIfNecessary;

import java.io.File;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.osgi.framework.BundleContext;
import org.springframework.core.style.ToStringCreator;
import org.springframework.util.Assert;
import org.springframework.util.StringUtils;

import com.springsource.server.bootstrap.BootstrapConfiguration;
import com.springsource.server.config.core.ConfigNode;
import com.springsource.server.config.core.ConfigTree;
import com.springsource.server.config.core.ConfigTreeFactory;
import com.springsource.server.config.core.ListConfigNode;
import com.springsource.server.config.core.MapConfigNode;
import com.springsource.server.config.core.ScalarConfigNode;
import com.springsource.server.io.core.PathReference;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;
import com.springsource.server.servlet.tomcat.TomcatServletContainerConfiguration;

/**
 * Standard implementation of the {@link TomcatServletContainerConfiguration} interface. <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Immutable and therefore thread-safe.
 * 
 * @author Sam Brannen
 * @author Ben Hale
 * @since 1.0
 */
public class StandardTomcatServletContainerConfiguration implements TomcatServletContainerConfiguration {

    /* Logging */
    private static final Tracer TRACER = TracerFactory.createTracer(StandardTomcatServletContainerConfiguration.class);

    /* Config tree location constants */
    private static final String CONFIG_NAME = "servletContainer";

    private static final String CONFIG_DIR = "configDir";

    private static final String HOST_NAME = "hostName";

    private static final String CONNECTORS = "connectors";

    private static final String LISTENERS = "listeners";

    private static final String JVM_ROUTE = "jvmRoute";

    private static final String LOGS = "logs";

    private static final String[] ACCESS_LOGS_DIR = { LOGS, "accessLogDir" };

    private static final String[] ACCESS_LOGS_FORMAT = { LOGS, "accessLogFormat" };

    private static final String[] PER_APPLICATION_LOGGING = { LOGS, "perApplicationLogging" };

    private static final String THREAD_POOL = "threadPool";

    private static final String[] THREAD_POOL_MIN_SIZE = { THREAD_POOL, "minSize" };

    private static final String[] THREAD_POOL_MAX_SIZE = { THREAD_POOL, "maxSize" };

    private static final String[] THREAD_POOL_KEEP_ALIVE_PERIOD = { THREAD_POOL, "keepAlivePeriod" };

    private static final String CLUSTER = "cluster";

    /* Config tree default constants */

    public static final String DEFAULT_NAME = "SpringSource-dm-Server";

    public static final String DEFAULT_CONFIG_DIR = "config" + File.separator + "servlet";

    public static final String DEFAULT_HOST_NAME = "localhost";

    public static final String DEFAULT_APPLICATION_DIR = "webapps";

    public static final String DEFAULT_WORK_DIR = "work";

    public static final String DEFAULT_ACCESS_LOGS_DIR = "serviceability/logs/access";

    public static final String DEFAULT_ACCESS_LOGS_FORMAT = "long";

    /* Config values */
    private final String name = DEFAULT_NAME;

    private final String applicationDir;

    private final String configDir;

    private final String workDir;

    private final String hostName;

    private final String jvmRoute;

    private final List<Map<String, String>> connectors;

    private final List<Map<String, String>> listeners;

    private final boolean perApplicationLogging;

    private final String accessLogDir;

    private final String accessLogFormat;

    private final int threadPoolMinSize;

    private final int threadPoolMaxSize;

    private final int threadPoolKeepAlivePeriod;

    private final Map<String, ? extends Object> cluster;

    /**
     * {@inheritDoc}
     */
    public String getName() {
        return this.name;
    }

    /**
     * {@inheritDoc}
     */
    public String getApplicationDir() {
        return this.applicationDir;
    }

    /**
     * {@inheritDoc}
     */
    public String getConfigDir() {
        return this.configDir;
    }

    /**
     * {@inheritDoc}
     */
    public String getWorkDir() {
        return this.workDir;
    }

    /**
     * {@inheritDoc}
     */
    public String getHostName() {
        return this.hostName;
    }

    /**
     * {@inheritDoc}
     */
    public String getJvmRoute() {
        return this.jvmRoute;
    }

    /**
     * {@inheritDoc}
     */
    public List<Map<String, String>> getConnectors() {
        return this.connectors;
    }

    /**
     * {@inheritDoc}
     */
    public List<Map<String, String>> getListeners() {
        return this.listeners;
    }

    /**
     * {@inheritDoc}
     */
    public boolean perApplicationLogging() {
        return this.perApplicationLogging;
    }

    /**
     * {@inheritDoc}
     */
    public String getAccessLogDir() {
        return this.accessLogDir;
    }

    /**
     * {@inheritDoc}
     */
    public String getAccessLogFormat() {
        return this.accessLogFormat;
    }

    /**
     * {@inheritDoc}
     */
    public int getThreadPoolMinSize() {
        return this.threadPoolMinSize;
    }

    /**
     * {@inheritDoc}
     */
    public int getThreadPoolMaxSize() {
        return this.threadPoolMaxSize;
    }

    /**
     * {@inheritDoc}
     */
    public int getThreadPoolKeepAlivePeriod() {
        return this.threadPoolKeepAlivePeriod;
    }

    /**
     * {@inheritDoc}
     */
    public Map<String, ? extends Object> getCluster() {
        return this.cluster;
    }

    /**
     * {@inheritDoc}
     */
    public String getTomcatUsersXml() {
        return getConfigDir() + File.separator + "tomcat-users.xml";
    }

    /**
     * {@inheritDoc}
     */
    public String getDefaultWebXml() {
        return getConfigDir() + File.separator + "web.xml";
    }

    /**
     * {@inheritDoc}
     */
    public String getDefaultContextXml() {
        // Must not be an absolute path!
        return "context.xml";
    }

    /**
     * Constructs a new <code>StandardTomcatServletContainerConfiguration</code> using the supplied configuration
     * parameters. This constructor performs validation and defaulting as specified in
     * {@link TomcatServletContainerConfiguration}. In addition, the supplied <code>configDir</code> must not be
     * <code>null</code> or <em>empty</em> unless a non-null {@link BootstrapConfiguration} is supplied.
     * 
     * @param configDir the path to the servlet container's <em>config</em> directory; may be <code>null</code>
     * @param applicationDir the path to the servlet container's <em>application</em> directory
     * @param workDir the path to the <em>work</em> directory
     * @param hostName the host name to use
     * @param jvmRoute the JVM route, typically used in conjunction with load balancing
     * @param connectors the list of connectors to configure
     * @param listeners the list of listeners to configure
     * @param accessLogDir the path to the <em>access log</em> directory
     * @param accessLogFormat the format to use for formatting <em>access log</em>
     * @param perApplicationLogging whether to use per-application logging
     * @param threadPoolMinSize the minimum size of the threadPool
     * @param threadPoolMaxSize the maximum size of the threadPool
     * @param threadPoolKeepAlivePeriod the period to keep a thread idle before removing it
     * @param cluster the cluster configuration
     * @param bootstrapConfiguration bootstrap configuration to use to infer the home directory if an explicit home
     *        directory is not supplied; may be <code>null</code>
     * @throws IllegalArgumentException if the supplied <code>configDir</code> is <code>null</code> or <em>empty</em>
     *         and if the supplied <code>bootstrapConfiguration</code> is <code>null</code>.
     */
    public StandardTomcatServletContainerConfiguration(String configDir, String applicationDir, String workDir, String hostName, String jvmRoute,
        List<Map<String, String>> connectors, List<Map<String, String>> listeners, String accessLogDir, String accessLogFormat,
        boolean perApplicationLogging, int threadPoolMinSize, int threadPoolMaxSize, int threadPoolKeepAlivePeriod,
        Map<String, ? extends Object> cluster, BootstrapConfiguration bootstrapConfiguration) {

        this.configDir = determineConfigDir(configDir, bootstrapConfiguration);
        this.applicationDir = determineApplicationDir(applicationDir, bootstrapConfiguration);
        this.workDir = determineWorkDir(workDir, bootstrapConfiguration);
        this.hostName = determineHostName(hostName);
        this.connectors = Collections.synchronizedList(connectors == null ? Collections.<Map<String, String>> emptyList() : connectors);
        this.listeners = Collections.synchronizedList(listeners == null ? Collections.<Map<String, String>> emptyList() : listeners);
        this.jvmRoute = jvmRoute;
        this.accessLogDir = determineAccessLogDir(accessLogDir, bootstrapConfiguration);
        this.accessLogFormat = determineAccessLogFormat(accessLogFormat);
        this.perApplicationLogging = perApplicationLogging;
        this.threadPoolMinSize = threadPoolMinSize;
        this.threadPoolMaxSize = threadPoolMaxSize;
        this.threadPoolKeepAlivePeriod = threadPoolKeepAlivePeriod;
        this.cluster = Collections.synchronizedMap(cluster == null ? Collections.<String, Object> emptyMap() : cluster);
    }

    /**
     * Constructs a new <code>StandardTomcatServletContainerConfiguration</code> by parsing the required configuration
     * parameters from the supplied {@link ConfigTreeFactory}, {@link BootstrapConfiguration}, and {@link BundleContext}
     * .
     */
    public StandardTomcatServletContainerConfiguration(ConfigTreeFactory configTreeFactory, BootstrapConfiguration bootstrapConfiguration,
        PathReference workDir, BundleContext context) {
        this(configTreeFactory.createConfigTree(CONFIG_NAME, context), bootstrapConfiguration, workDir);
    }

    @SuppressWarnings("unchecked")
    StandardTomcatServletContainerConfiguration(ConfigTree configTree, BootstrapConfiguration bootstrapConfiguration, PathReference workDir) {
        this(configTree.getStringValue(CONFIG_NAME, CONFIG_DIR), //
            null, //
            workDir != null ? workDir.getCanonicalPath() : null, //
            configTree.getStringValue(CONFIG_NAME, HOST_NAME), //
            configTree.getStringValue(CONFIG_NAME, JVM_ROUTE), //
            (List<Map<String, String>>) parseListConfigNode(configTree.getConfigPoint(CONFIG_NAME).getNode(CONNECTORS)),//
            (List<Map<String, String>>) parseListConfigNode(configTree.getConfigPoint(CONFIG_NAME).getNode(LISTENERS)),//
            configTree.getStringValue(CONFIG_NAME, ACCESS_LOGS_DIR), //
            configTree.getStringValue(CONFIG_NAME, ACCESS_LOGS_FORMAT), //
            configTree.getBooleanValue(CONFIG_NAME, PER_APPLICATION_LOGGING), //
            configTree.getIntValue(CONFIG_NAME, THREAD_POOL_MIN_SIZE), //
            configTree.getIntValue(CONFIG_NAME, THREAD_POOL_MAX_SIZE), //
            configTree.getIntValue(CONFIG_NAME, THREAD_POOL_KEEP_ALIVE_PERIOD), //
            parseMapConfigNode(configTree.getConfigPoint(CONFIG_NAME).getNode(CLUSTER)),//
            bootstrapConfiguration);
    }

    protected static List<? extends Object> parseListConfigNode(ConfigNode node) {
        if (node == null) {
            return null;
        }
        Assert.isInstanceOf(ListConfigNode.class, node, "Cannot parse ListConfigNode");

        List<Object> list = new ArrayList<Object>();
        for (ConfigNode child : ((ListConfigNode) node).getNodes()) {
            if (child instanceof ListConfigNode) {
                list.add(parseListConfigNode(child));
            } else if (child instanceof MapConfigNode) {
                list.add(parseMapConfigNode(child));
            } else {
                list.add(parseScalarConfigNode(child));
            }
        }
        return list;
    }

    protected static Map<String, ? extends Object> parseMapConfigNode(ConfigNode node) {
        if (node == null) {
            return null;
        }
        Assert.isInstanceOf(MapConfigNode.class, node, "Cannot parse MapConfigNode");

        Map<String, Object> map = new HashMap<String, Object>();
        for (String key : ((MapConfigNode) node).getKeys()) {
            ConfigNode child = node.getNode(key);
            if (child instanceof ListConfigNode) {
                map.put(key, parseListConfigNode(child));
            } else if (child instanceof MapConfigNode) {
                map.put(key, parseMapConfigNode(child));
            } else {
                map.put(key, parseScalarConfigNode(child));
            }
        }
        return map;
    }

    protected static String parseScalarConfigNode(ConfigNode node) {
        if (node == null) {
            return null;
        }

        Assert.isInstanceOf(ScalarConfigNode.class, node, "Cannot parse ScalarConfigNode");
        return ((ScalarConfigNode) node).getValue().toString();
    }

    private static boolean isRelativePath(String path) {
        return !new File(path).isAbsolute();
    }

    private static boolean isEmpty(String str) {
        return !StringUtils.hasText(str);
    }

    private String determineHostName(String hostName) {
        TRACER.traceVerbose("ServletContainer host name configured as '%s'", hostName);
        if (isEmpty(hostName)) {
            hostName = DEFAULT_HOST_NAME;
        }
        return hostName;
    }

    private String determineConfigDir(String configDir, BootstrapConfiguration bootstrapConfiguration) {
        TRACER.traceVerbose("ServletContainer config directory configured as '%s'", configDir);
        // We require either a non-null bootstrapConfiguration or a non-empty configDir
        if (bootstrapConfiguration == null && isEmpty(configDir)) {
            throw new IllegalArgumentException(
                "The ServletContainer's config directory must not be null or empty: either supply a valid config directory or provide a non-null BootstrapConfiguration.");
        }
        // Use the default config directory?
        if (isEmpty(configDir)) {
            configDir = DEFAULT_CONFIG_DIR;
        }
        // Relative paths should be relative to the configured server's home directory if there is one.
        if (bootstrapConfiguration != null && isRelativePath(configDir)) {
            configDir = bootstrapConfiguration.getHomeDirectory() + File.separatorChar + configDir;
        }
        return convertToCanonicalPath(configDir, true);
    }

    private String determineApplicationDir(String applicationDir, BootstrapConfiguration bootstrapConfiguration) {
        TRACER.traceVerbose("ServletContainer application directory configured as '%s'", applicationDir);
        if (isEmpty(applicationDir)) {
            applicationDir = System.getProperty("java.io.tmpdir", DEFAULT_APPLICATION_DIR);
        }
        // Relative paths should be relative to the configured server's home directory if there is one.
        if (isRelativePath(applicationDir)) {
            applicationDir = bootstrapConfiguration.getHomeDirectory() + File.separatorChar + applicationDir;
        }
        return createDirectoryIfNecessary(applicationDir);
    }

    private String determineWorkDir(String workDir, BootstrapConfiguration bootstrapConfiguration) {
        TRACER.traceVerbose("ServletContainer work directory configured as '%s'", workDir);
        if (isEmpty(workDir)) {
            workDir = DEFAULT_WORK_DIR;
        }
        // Relative paths should be relative to the configured server's home directory if there is one.
        if (isRelativePath(workDir)) {
            workDir = bootstrapConfiguration.getHomeDirectory() + File.separatorChar + workDir;
        }
        return createDirectoryIfNecessary(workDir);
    }

    private String determineAccessLogDir(String accessLogDir, BootstrapConfiguration bootstrapConfiguration) {
        TRACER.traceVerbose("ServletContainer access log directory configured as '%s'", accessLogDir);
        if (isEmpty(accessLogDir)) {
            accessLogDir = DEFAULT_ACCESS_LOGS_DIR;
        }
        // Relative paths should be relative to the configured server's home directory if there is one.
        if (isRelativePath(accessLogDir)) {
            accessLogDir = bootstrapConfiguration.getHomeDirectory() + File.separatorChar + accessLogDir;
        }
        return createDirectoryIfNecessary(accessLogDir);
    }

    private String determineAccessLogFormat(String accessLogFormat) {
        TRACER.traceVerbose("ServletContainer access log format configured as '%s'", accessLogFormat);
        if (isEmpty(accessLogFormat)) {
            accessLogFormat = DEFAULT_ACCESS_LOGS_FORMAT;
        }
        return accessLogFormat;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String toString() {
        return new ToStringCreator(this)//
        .append("name", this.name)//
        .append("applicationDir", this.applicationDir)//
        .append("configDir", this.configDir)//
        .append("workDir", this.workDir)//
        .append("hostName", this.hostName)//
        .append("jvmRoute", this.jvmRoute)//
        .append("perApplicationLogging", this.perApplicationLogging)//
        .append("accessLogDir", this.accessLogDir)//
        .append("accessLogFormat", this.accessLogFormat)//
        .append("threadPoolMinSize", this.threadPoolMinSize)//
        .append("threadPoolMaxSize", this.threadPoolMaxSize)//
        .append("threadPoolKeepAlivePeriod", this.threadPoolKeepAlivePeriod)//
        .toString();
    }

}
