/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.servlet.tomcat.loader;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;

import javax.management.MBeanRegistration;
import javax.management.MBeanServer;
import javax.management.ObjectName;
import javax.servlet.ServletContext;

import org.apache.catalina.Container;
import org.apache.catalina.Context;
import org.apache.catalina.Engine;
import org.apache.catalina.Globals;
import org.apache.catalina.Lifecycle;
import org.apache.catalina.LifecycleException;
import org.apache.catalina.LifecycleListener;
import org.apache.catalina.Loader;
import org.apache.catalina.core.StandardContext;
import org.apache.catalina.loader.Constants;
import org.apache.catalina.util.LifecycleSupport;
import org.apache.catalina.util.StringManager;
import org.apache.tomcat.util.modeler.Registry;
import org.osgi.framework.Bundle;
import org.springframework.util.Assert;
import org.springframework.util.ClassUtils;

import com.springsource.server.serviceability.output.Level;
import com.springsource.server.serviceability.tracing.EntryExitTraceLevel;

/**
 * TODO Document OsgiWebappLoader
 * 
 * @author Craig R. McClanahan
 * @author Remy Maucherat
 * @author Sam Brannen
 * @author Andy Wilkinson
 */
public class OsgiWebappLoader implements Loader, Lifecycle, PropertyChangeListener, MBeanRegistration {

    /*
     * Implementation Notes:
     * 
     * Note that 'delegate' is not supported in OsgiWebappClassLoader; thus, we need to determine how we _handle_ calls
     * to setDelegate() in this class (e.g., throw an UnsupportedOperationException).
     */

    private static org.apache.juli.logging.Log log = org.apache.juli.logging.LogFactory.getLog(OsgiWebappLoader.class);

    /**
     * The OSGi {@link Bundle bundle} which will back the {@link ClassLoader} we will create.
     */
    private volatile Bundle bundle;

    /**
     * The time stamp which is used to track when the bundle's {@link Bundle#getLastModified() last modified} time stamp
     * was last checked.
     */
    private long bundleModificationCheckTimestamp;

    private Object bundleModificationLock = new Object();

    /**
     * The class loader being managed by this Loader.
     */
    private ClassLoader classLoader = null;

    /**
     * The "follow standard delegation model" flag that will be used to configure our ClassLoader.
     */
    private boolean delegate = false;

    /**
     * The Container with which this Loader has been associated.
     */
    private Container container = null;

    /**
     * The descriptive information about this Loader implementation.
     */
    private static final String INFO = "org.springframework.osgi.web.servlet.tomcat.loader.OsgiWebappLoader/1.0";

    /**
     * The lifecycle event support for this Loader.
     */
    protected LifecycleSupport lifecycle = new LifecycleSupport(this);

    /**
     * The reloadable flag for this Loader.
     */
    private boolean reloadable = false;

    /**
     * The string manager for this package.
     */
    protected static final StringManager sm = StringManager.getManager(Constants.Package);

    /**
     * Has this Loader been started?
     */
    private boolean started = false;

    /**
     * The property change support for this Loader.
     */
    protected PropertyChangeSupport support = new PropertyChangeSupport(this);

    private ObjectName objectName;

    private ObjectName controller;

    private boolean initialized = false;

    private volatile Bundle webBundle;

    // -------------------------------------------------------------------------
    // --- Constructors
    // -------------------------------------------------------------------------

    /**
     * TODO Document OsgiWebappLoader(Bundle)
     * 
     * @param bundle
     */
    public OsgiWebappLoader(Bundle bundle, Bundle webBundle) {
        Assert.notNull(bundle, "'bundle' must not be null");
        this.bundle = bundle;
        this.bundleModificationCheckTimestamp = this.bundle.getLastModified();
        this.webBundle = webBundle;
    }

    // -------------------------------------------------------------------------
    // --- OsgiWebappLoader-specific implementation
    // -------------------------------------------------------------------------

    /**
     * @return
     * @see #getClassLoaderName()
     */
    private ClassLoader createClassLoader() {
        return new OsgiWebappClassLoader(this.bundle, this.webBundle);
    }

    // -------------------------------------------------------------------------
    // --- Loader
    // -------------------------------------------------------------------------

    /**
     * Add a property change listener to this component.
     * 
     * @param listener The listener to add
     */
    public void addPropertyChangeListener(PropertyChangeListener listener) {
        this.support.addPropertyChangeListener(listener);
    }

    /**
     * @throws UnsupportedOperationException
     */
    public void addRepository(String repository) {
        throw new UnsupportedOperationException(ClassUtils.getShortName(getClass()) + " does not support addRepository(String)");
    }

    /**
     * Execute a periodic task, such as reloading, etc. This method will be invoked inside the class loading context of
     * this container. Unexpected throwables will be caught and logged.
     */
    @EntryExitTraceLevel(Level.DEBUG)
    public void backgroundProcess() {
        if (this.reloadable && modified()) {
            try {
                Thread.currentThread().setContextClassLoader(OsgiWebappLoader.class.getClassLoader());
                if (this.container instanceof StandardContext) {
                    ((StandardContext) this.container).reload();
                }
            } finally {
                if (this.container.getLoader() != null) {
                    Thread.currentThread().setContextClassLoader(this.container.getLoader().getClassLoader());
                }
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    public String[] findRepositories() {
        // TODO Determine if we need to implement findRepositories() on top of our OsgiWebappClassLoader
        return new String[0];
    }

    /**
     * {@inheritDoc}
     */
    @EntryExitTraceLevel(Level.DEBUG)
    public ClassLoader getClassLoader() {
        return this.classLoader;
    }

    /**
     * Return the Container with which this Logger has been associated.
     */
    public Container getContainer() {
        return this.container;
    }

    /**
     * Set the Container with which this Logger has been associated.
     * 
     * @param container The associated Container
     */
    public void setContainer(Container container) {

        // unregister from the old Container (if any)
        if (this.container != null && this.container instanceof Context) {
            ((Context) this.container).removePropertyChangeListener(this);
        }

        // Process this property change
        Container oldContainer = this.container;
        this.container = container;
        this.support.firePropertyChange("container", oldContainer, this.container);

        // Register with the new Container (if any)
        if (this.container != null && this.container instanceof Context) {
            setReloadable(((Context) this.container).getReloadable());
            ((Context) this.container).addPropertyChangeListener(this);
        }
    }

    /**
     * Return the "follow standard delegation model" flag used to configure our ClassLoader.
     */
    public boolean getDelegate() {
        return this.delegate;
    }

    /**
     * Set the "follow standard delegation model" flag used to configure our ClassLoader.
     * 
     * @param delegate The new flag
     */
    public void setDelegate(boolean delegate) {
        boolean oldDelegate = this.delegate;
        this.delegate = delegate;
        this.support.firePropertyChange("delegate", new Boolean(oldDelegate), new Boolean(this.delegate));
    }

    /**
     * Return descriptive information about this Loader implementation and the corresponding version number, in the
     * format <code>&lt;description&gt;/&lt;version&gt;</code>.
     */
    public String getInfo() {
        return INFO;
    }

    /**
     * {@inheritDoc}
     */
    public boolean modified() {

        synchronized (bundleModificationLock) {
            // If the current last modification time stamp is newer than the time stamp from the last time we checked,
            // there's been change!
            final long lastModified = this.bundle.getLastModified();
            if (lastModified > this.bundleModificationCheckTimestamp) {
                this.bundleModificationCheckTimestamp = lastModified;
                return true;
            }
        }

        return false;
    }

    /**
     * {@inheritDoc}
     */
    public void removePropertyChangeListener(PropertyChangeListener listener) {
        this.support.removePropertyChangeListener(listener);
    }

    /**
     * Return the reloadable flag for this Loader.
     */
    public boolean getReloadable() {
        return this.reloadable;
    }

    /**
     * Set the reloadable flag for this Loader.
     * 
     * @param reloadable The new reloadable flag
     */
    public void setReloadable(boolean reloadable) {
        // Process this property change
        boolean oldReloadable = this.reloadable;
        this.reloadable = reloadable;
        this.support.firePropertyChange("reloadable", new Boolean(oldReloadable), new Boolean(this.reloadable));
    }

    // -------------------------------------------------------------------------
    // --- Lifecycle
    // -------------------------------------------------------------------------

    /**
     * {@inheritDoc}
     */
    public void addLifecycleListener(LifecycleListener listener) {
        this.lifecycle.addLifecycleListener(listener);
    }

    /**
     * {@inheritDoc}
     */
    public LifecycleListener[] findLifecycleListeners() {
        return this.lifecycle.findLifecycleListeners();
    }

    /**
     * {@inheritDoc}
     */
    public void removeLifecycleListener(LifecycleListener listener) {
        this.lifecycle.removeLifecycleListener(listener);
    }

    /**
     * Ensures that the grandparent {@link Container} of the supplied {@link StandardContext context} is an
     * {@link Engine}.
     * 
     * @throws IllegalStateException if the grandparent is not an <code>Engine</code>
     */
    private void ensureGrandparentIsAnEngine(StandardContext standardContext) {
        if (!(standardContext.getParent().getParent() instanceof Engine)) {
            throw new IllegalStateException("Grandparent of [" + standardContext + "] is not an instanceof [" + Engine.class.getName() + "].");
        }
    }

    protected String getCatalinaContextPath(StandardContext context) {
        String contextPath = context.getPath();
        if (contextPath.equals("")) {
            contextPath = "/";
        }
        return contextPath;
    }

    protected void init() {
        this.initialized = true;

        if (this.objectName == null) {
            // not registered yet - stand-alone or API
            if (this.container instanceof StandardContext) {
                // Register ourself. The container must be a webapp
                try {
                    StandardContext ctx = (StandardContext) this.container;
                    ensureGrandparentIsAnEngine(ctx);
                    this.objectName = new ObjectName(ctx.getEngineName() + ":type=Loader,path=" + getCatalinaContextPath(ctx) + ",host="
                        + ctx.getParent().getName());
                    Registry.getRegistry(null, null).registerComponent(this, this.objectName, null);
                    this.controller = this.objectName;
                } catch (Exception e) {
                    log.error("Error registering loader", e);
                }
            }
        }

        if (this.container == null) {
            // JMX created the loader
            // ???
        }
    }

    protected void destroy() {
        if (this.controller == this.objectName) {
            // Self-registration, undo it
            Registry.getRegistry(null, null).unregisterComponent(this.objectName);
            this.objectName = null;
        }
        this.initialized = false;
    }

    /**
     * {@inheritDoc}
     */
    public void start() throws LifecycleException {
        // Validate and update our current component state
        if (!this.initialized) {
            init();
        }
        if (this.started) {
            throw new LifecycleException(sm.getString("webappLoader.alreadyStarted"));
        }
        if (log.isDebugEnabled()) {
            log.debug(sm.getString("webappLoader.starting"));
        }
        this.lifecycle.fireLifecycleEvent(START_EVENT, null);
        this.started = true;

        if (this.container.getResources() == null) {
            log.info("No resources for " + this.container);
            return;
        }

        // Construct a class loader based on our current repositories list
        try {

            this.classLoader = createClassLoader();
            if (this.classLoader instanceof Lifecycle) {
                ((Lifecycle) this.classLoader).start();
            }

            StandardContext ctx = (StandardContext) this.container;
            ensureGrandparentIsAnEngine(ctx);
            ObjectName classLoaderObjectName = new ObjectName(ctx.getEngineName() + ":type=OsgiWebappClassLoader,path=" + getCatalinaContextPath(ctx)
                + ",host=" + ctx.getParent().getName());
            Registry.getRegistry(null, null).registerComponent(this.classLoader, classLoaderObjectName, null);

        } catch (Throwable t) {
            log.error("LifecycleException ", t);
            throw new LifecycleException("start: ", t);
        }
    }

    /**
     * {@inheritDoc}
     */
    public void stop() throws LifecycleException {

        // Validate and update our current component state
        if (!this.started) {
            throw new LifecycleException(sm.getString("webappLoader.notStarted"));
        }
        if (log.isDebugEnabled()) {
            log.debug(sm.getString("webappLoader.stopping"));
        }
        this.lifecycle.fireLifecycleEvent(STOP_EVENT, null);
        this.started = false;

        // Remove context attributes as appropriate
        if (this.container instanceof Context) {
            ServletContext servletContext = ((Context) this.container).getServletContext();
            servletContext.removeAttribute(Globals.CLASS_PATH_ATTR);
        }

        // Throw away our current class loader
        if (this.classLoader instanceof Lifecycle) {
            ((Lifecycle) this.classLoader).stop();
        }

        try {
            StandardContext ctx = (StandardContext) this.container;
            ensureGrandparentIsAnEngine(ctx);
            ObjectName classLoaderObjectName = new ObjectName(ctx.getEngineName() + ":type=OsgiWebappClassLoader,path=" + getCatalinaContextPath(ctx)
                + ",host=" + ctx.getParent().getName());
            Registry.getRegistry(null, null).unregisterComponent(classLoaderObjectName);
        } catch (Throwable t) {
            log.error("LifecycleException ", t);
        }

        this.classLoader = null;
        this.bundle = null;
        this.webBundle = null;
        destroy();
    }

    // -------------------------------------------------------------------------
    // --- PropertyChangeListener
    // -------------------------------------------------------------------------

    /**
     * {@inheritDoc}
     */
    public void propertyChange(PropertyChangeEvent event) {

        // Validate the source of this event
        if (!(event.getSource() instanceof Context)) {
            return;
        }

        // Process a relevant property change
        if (event.getPropertyName().equals("reloadable")) {
            try {
                setReloadable(((Boolean) event.getNewValue()).booleanValue());
            } catch (Exception e) {
                log.error(sm.getString("webappLoader.reloadable", event.getNewValue().toString()));
            }
        }
    }

    // -------------------------------------------------------------------------
    // --- MBeanRegistration
    // -------------------------------------------------------------------------

    /**
     * {@inheritDoc}
     */
    public void postDeregister() {
        /* no-op */
    }

    /**
     * {@inheritDoc}
     */
    public void postRegister(Boolean registrationDone) {
        /* no-op */
    }

    /**
     * {@inheritDoc}
     */
    public void preDeregister() throws Exception {
        /* no-op */
    }

    /**
     * {@inheritDoc}
     */
    public ObjectName preRegister(MBeanServer mBeanServer, ObjectName objectName) throws Exception {
        this.objectName = objectName;
        return objectName;
    }

    // -------------------------------------------------------------------------

}
