/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.servlet.tomcat.loader;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FilePermission;
import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.security.AccessControlException;
import java.security.AccessController;
import java.security.CodeSource;
import java.security.Permission;
import java.security.PermissionCollection;
import java.security.Policy;
import java.security.PrivilegedAction;
import java.sql.Driver;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;
import java.util.jar.Attributes;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;
import java.util.jar.Manifest;
import java.util.jar.Attributes.Name;

import javax.naming.NameClassPair;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;
import javax.naming.directory.DirContext;

import org.apache.catalina.Lifecycle;
import org.apache.catalina.LifecycleException;
import org.apache.catalina.LifecycleListener;
import org.apache.catalina.loader.Constants;
import org.apache.catalina.loader.Reloader;
import org.apache.catalina.loader.ResourceEntry;
import org.apache.catalina.util.StringManager;
import org.apache.naming.JndiPermission;
import org.apache.naming.resources.Resource;
import org.apache.naming.resources.ResourceAttributes;
import org.apache.tomcat.util.IntrospectionUtils;

/**
 * Specialized web application class loader which is a custom extension of {@link WebappClassLoader} that overrides
 * various methods (e.g., {@link #loadClass(String, boolean)} and {@link #delegateToSystemClassLoader(String)}) to
 * modify search algorithms for classes and resources by taking into account that the class loader is being used in an
 * OSGi environment.
 * <p>
 * This class loader is a full reimplementation of the <code>URLClassLoader</code> from the JDK. It is designed to be
 * fully compatible with a normal <code>URLClassLoader</code>, although its internal behavior may be completely
 * different.
 * <p>
 * <strong>IMPLEMENTATION NOTE</strong> - This class loader faithfully follows the delegation model recommended in the
 * specification. The system class loader will be queried first, then the local repositories, and only then delegation
 * to the parent class loader will occur. This allows the web application to override any shared class except the
 * classes from J2SE. Special handling is provided from the JAXP XML parser interfaces, the JNDI interfaces, and the
 * classes from the servlet API, which are never loaded from the webapp repository.
 * <p>
 * <strong>IMPLEMENTATION NOTE</strong> - Due to limitations in Jasper compilation technology, any repository which
 * contains classes from the servlet API will be ignored by the class loader.
 * <p>
 * <strong>IMPLEMENTATION NOTE</strong> - The class loader generates source URLs which include the full JAR URL when a
 * class is loaded from a JAR file, which allows setting security permission at the class level, even when a class is
 * contained inside a JAR.
 * <p>
 * <strong>IMPLEMENTATION NOTE</strong> - Local repositories are searched in the order they are added via the initial
 * constructor and/or any subsequent calls to <code>addRepository()</code> or <code>addJar()</code>.
 * <p>
 * <strong>IMPLEMENTATION NOTE</strong> - No check for sealing violations or security is made unless a security manager
 * is present.
 * 
 * @author Remy Maucherat
 * @author Craig R. McClanahan
 * @author Sam Brannen
 */
public class WebappClassLoader extends URLClassLoader implements Reloader, Lifecycle {

    protected static org.apache.juli.logging.Log log = org.apache.juli.logging.LogFactory.getLog(WebappClassLoader.class);

    public static final boolean ENABLE_CLEAR_REFERENCES = Boolean.valueOf(
        System.getProperty("org.apache.catalina.loader.WebappClassLoader.ENABLE_CLEAR_REFERENCES", "true")).booleanValue();

    protected class PrivilegedFindResource implements PrivilegedAction<Object> {

        protected File file;

        protected String path;

        PrivilegedFindResource(File file, String path) {
            this.file = file;
            this.path = path;
        }

        public Object run() {
            return findResourceInternal(this.file, this.path);
        }
    }

    // ------------------------------------------------------- Static Variables

    /**
     * The set of trigger classes that will cause a proposed repository not to be added if this class is visible to the
     * class loader that loaded this factory class. Typically, trigger classes will be listed for components that have
     * been integrated into the JDK for later versions, but where the corresponding JAR files are required to run on
     * earlier versions.
     */
    protected static final String[] triggers = {

    "javax.servlet.Servlet" // Servlet API

    };

    /**
     * Set of package names which are not allowed to be loaded from a webapp class loader without delegating first.
     */
    protected static final String[] packageTriggers = {};

    /**
     * The string manager for this package.
     */
    protected static final StringManager sm = StringManager.getManager(Constants.Package);

    /**
     * Use anti JAR locking code, which does URL rerouting when accessing resources.
     */
    boolean antiJARLocking = false;

    // ----------------------------------------------------------- Constructors

    /**
     * Construct a new ClassLoader with no defined repositories and no parent ClassLoader.
     */
    public WebappClassLoader() {

        super(new URL[0]);
        this.parent = getParent();
        this.system = getSystemClassLoader();
        this.securityManager = System.getSecurityManager();

        if (this.securityManager != null) {
            refreshPolicy();
        }
    }

    /**
     * Construct a new ClassLoader with no defined repositories and no parent ClassLoader.
     */
    public WebappClassLoader(ClassLoader parent) {

        super(new URL[0], parent);

        this.parent = getParent();

        this.system = getSystemClassLoader();
        this.securityManager = System.getSecurityManager();

        if (this.securityManager != null) {
            refreshPolicy();
        }
    }

    // ----------------------------------------------------- Instance Variables

    /**
     * Associated directory context giving access to the resources in this webapp.
     */
    protected DirContext resources = null;

    /**
     * The cache of ResourceEntry for classes and resources we have loaded, keyed by resource name.
     */
    protected HashMap<String, ResourceEntry> resourceEntries = new HashMap<String, ResourceEntry>();

    /**
     * The list of not found resources.
     */
    protected HashMap<String, String> notFoundResources = new HashMap<String, String>();

    /**
     * Should this class loader delegate to the parent class loader <strong>before</strong> searching its own
     * repositories (i.e. the usual Java2 delegation model)? If set to <code>false</code>, this class loader will
     * search its own repositories first, and delegate to the parent only if the class or resource is not found locally.
     */
    protected boolean delegate = false;

    /**
     * Last time a JAR was accessed.
     */
    protected long lastJarAccessed = 0L;

    /**
     * The list of local repositories, in the order they should be searched for locally loaded classes or resources.
     */
    protected String[] repositories = new String[0];

    /**
     * Repositories URLs, used to cache the result of getURLs.
     */
    protected URL[] repositoryURLs = null;

    /**
     * Repositories translated as path in the work directory (for Jasper originally), but which is used to generate fake
     * URLs should getURLs be called.
     */
    protected File[] files = new File[0];

    /**
     * The list of JARs, in the order they should be searched for locally loaded classes or resources.
     */
    protected JarFile[] jarFiles = new JarFile[0];

    /**
     * The list of JARs, in the order they should be searched for locally loaded classes or resources.
     */
    protected File[] jarRealFiles = new File[0];

    /**
     * The path which will be monitored for added Jar files.
     */
    protected String jarPath = null;

    /**
     * The list of JARs, in the order they should be searched for locally loaded classes or resources.
     */
    protected String[] jarNames = new String[0];

    /**
     * The list of JARs last modified dates, in the order they should be searched for locally loaded classes or
     * resources.
     */
    protected long[] lastModifiedDates = new long[0];

    /**
     * The list of resources which should be checked when checking for modifications.
     */
    protected String[] paths = new String[0];

    /**
     * A list of read File and Jndi Permission's required if this loader is for a web application context.
     */
    protected ArrayList<Permission> permissionList = new ArrayList<Permission>();

    /**
     * Path where resources loaded from JARs will be extracted.
     */
    protected File loaderDir = null;

    /**
     * The PermissionCollection for each CodeSource for a web application context.
     */
    protected HashMap<String, PermissionCollection> loaderPC = new HashMap<String, PermissionCollection>();

    /**
     * Instance of the SecurityManager installed.
     */
    protected SecurityManager securityManager = null;

    /**
     * The parent class loader.
     */
    protected ClassLoader parent = null;

    /**
     * The system class loader.
     */
    protected ClassLoader system = null;

    /**
     * Has this component been started?
     */
    protected boolean started = false;

    /**
     * Has external repositories.
     */
    protected boolean hasExternalRepositories = false;

    /**
     * need conversion for properties files
     */
    protected boolean needConvert = false;

    /**
     * All permission.
     */
    protected Permission allPermission = new java.security.AllPermission();

    // ------------------------------------------------------------- Properties

    /**
     * Get associated resources.
     */
    public DirContext getResources() {
        return this.resources;
    }

    /**
     * Set associated resources.
     */
    public void setResources(DirContext resources) {
        this.resources = resources;
    }

    /**
     * Return the "delegate first" flag for this class loader.
     */
    public boolean getDelegate() {
        return this.delegate;
    }

    /**
     * Set the "delegate first" flag for this class loader.
     * 
     * @param delegate The new "delegate first" flag
     */
    public void setDelegate(boolean delegate) {
        this.delegate = delegate;
    }

    /**
     * @return Returns the antiJARLocking.
     */
    public boolean getAntiJARLocking() {
        return this.antiJARLocking;
    }

    /**
     * @param antiJARLocking The antiJARLocking to set.
     */
    public void setAntiJARLocking(boolean antiJARLocking) {
        this.antiJARLocking = antiJARLocking;
    }

    /**
     * If there is a Java SecurityManager create a read FilePermission or JndiPermission for the file directory path.
     * 
     * @param path file directory path
     */
    public void addPermission(String path) {
        if (path == null) {
            return;
        }

        if (this.securityManager != null) {
            Permission permission = null;
            if (path.startsWith("jndi:") || path.startsWith("jar:jndi:")) {
                if (!path.endsWith("/")) {
                    path = path + "/";
                }
                permission = new JndiPermission(path + "*");
                addPermission(permission);
            } else {
                if (!path.endsWith(File.separator)) {
                    permission = new FilePermission(path, "read");
                    addPermission(permission);
                    path = path + File.separator;
                }
                permission = new FilePermission(path + "-", "read");
                addPermission(permission);
            }
        }
    }

    /**
     * If there is a Java SecurityManager create a read FilePermission or JndiPermission for URL.
     * 
     * @param url URL for a file or directory on local system
     */
    public void addPermission(URL url) {
        if (url != null) {
            addPermission(url.toString());
        }
    }

    /**
     * If there is a Java SecurityManager create a Permission.
     * 
     * @param permission The permission
     */
    public void addPermission(Permission permission) {
        if (this.securityManager != null && permission != null) {
            this.permissionList.add(permission);
        }
    }

    /**
     * Return the JAR path.
     */
    public String getJarPath() {
        return this.jarPath;
    }

    /**
     * Change the Jar path.
     */
    public void setJarPath(String jarPath) {
        this.jarPath = jarPath;
    }

    /**
     * Change the work directory.
     */
    public void setWorkDir(File workDir) {
        this.loaderDir = new File(workDir, "loader");
    }

    /**
     * Utility method for use in subclasses. Must be called before Lifecycle methods to have any effect.
     */
    protected void setParentClassLoader(ClassLoader pcl) {
        this.parent = pcl;
    }

    // ------------------------------------------------------- Reloader Methods

    /**
     * Add a new repository to the set of places this ClassLoader can look for classes to be loaded.
     * 
     * @param repository Name of a source of classes to be loaded, such as a directory pathname, a JAR file pathname, or
     *        a ZIP file pathname
     * 
     * @exception IllegalArgumentException if the specified repository is invalid or does not exist
     */
    public void addRepository(String repository) {

        // Ignore any of the standard repositories, as they are set up using
        // either addJar or addRepository
        if (repository.startsWith("/WEB-INF/lib") || repository.startsWith("/WEB-INF/classes")) {
            return;
        }

        // Add this repository to our underlying class loader
        try {
            URL url = new URL(repository);
            super.addURL(url);
            this.hasExternalRepositories = true;
            this.repositoryURLs = null;
        } catch (MalformedURLException e) {
            IllegalArgumentException iae = new IllegalArgumentException("Invalid repository: " + repository);
            iae.initCause(e);
            throw iae;
        }
    }

    /**
     * Add a new repository to the set of places this ClassLoader can look for classes to be loaded.
     * 
     * @param repository Name of a source of classes to be loaded, such as a directory pathname, a JAR file pathname, or
     *        a ZIP file pathname
     * 
     * @exception IllegalArgumentException if the specified repository is invalid or does not exist
     */
    // BEGIN - SpringSource Modifications
    // + public
    public synchronized void addRepository(String repository, File file) {
        // END - SpringSource Modifications

        // Note : There should be only one (of course), but I think we should
        // keep this a bit generic

        if (repository == null) {
            return;
        }

        if (log.isDebugEnabled()) {
            log.debug("addRepository(" + repository + ")");
        }

        int i;

        // Add this repository to our internal list
        String[] result = new String[this.repositories.length + 1];
        for (i = 0; i < this.repositories.length; i++) {
            result[i] = this.repositories[i];
        }
        result[this.repositories.length] = repository;
        this.repositories = result;

        // Add the file to the list
        File[] result2 = new File[this.files.length + 1];
        for (i = 0; i < this.files.length; i++) {
            result2[i] = this.files[i];
        }
        result2[this.files.length] = file;
        this.files = result2;
    }

    // BEGIN - SpringSource Modifications
    // + public
    public synchronized void addJar(String jar, JarFile jarFile, File file) throws IOException {
        // END - SpringSource Modifications

        if (jar == null) {
            return;
        }
        if (jarFile == null) {
            return;
        }
        if (file == null) {
            return;
        }

        if (log.isDebugEnabled()) {
            log.debug("addJar(" + jar + ")");
        }

        int i;

        if (this.jarPath != null && jar.startsWith(this.jarPath)) {

            String jarName = jar.substring(this.jarPath.length());
            while (jarName.startsWith("/")) {
                jarName = jarName.substring(1);
            }

            String[] result = new String[this.jarNames.length + 1];
            for (i = 0; i < this.jarNames.length; i++) {
                result[i] = this.jarNames[i];
            }
            result[this.jarNames.length] = jarName;
            this.jarNames = result;
        }

        try {

            // Register the JAR for tracking
            long lastModified = ((ResourceAttributes) this.resources.getAttributes(jar)).getLastModified();

            String[] result = new String[this.paths.length + 1];
            for (i = 0; i < this.paths.length; i++) {
                result[i] = this.paths[i];
            }
            result[this.paths.length] = jar;
            this.paths = result;

            long[] result3 = new long[this.lastModifiedDates.length + 1];
            for (i = 0; i < this.lastModifiedDates.length; i++) {
                result3[i] = this.lastModifiedDates[i];
            }
            result3[this.lastModifiedDates.length] = lastModified;
            this.lastModifiedDates = result3;

        } catch (NamingException e) {
            // Ignore
        }

        // If the JAR currently contains invalid classes, don't actually use it
        // for classloading
        if (!validateJarFile(file)) {
            return;
        }

        JarFile[] result2 = new JarFile[this.jarFiles.length + 1];
        for (i = 0; i < this.jarFiles.length; i++) {
            result2[i] = this.jarFiles[i];
        }
        result2[this.jarFiles.length] = jarFile;
        this.jarFiles = result2;

        // Add the file to the list
        File[] result4 = new File[this.jarRealFiles.length + 1];
        for (i = 0; i < this.jarRealFiles.length; i++) {
            result4[i] = this.jarRealFiles[i];
        }
        result4[this.jarRealFiles.length] = file;
        this.jarRealFiles = result4;
    }

    /**
     * Return a String array of the current repositories for this class loader. If there are no repositories, a
     * zero-length array is returned.For security reason, returns a clone of the Array (since String are immutable).
     */
    public String[] findRepositories() {
        return (String[]) this.repositories.clone();
    }

    /**
     * Have one or more classes or resources been modified so that a reload is appropriate?
     */
    public boolean modified() {

        if (log.isDebugEnabled()) {
            log.debug("modified()");
        }

        // Checking for modified loaded resources
        int length = this.paths.length;

        // A rare race condition can occur in the updates of the two arrays
        // It's totally ok if the latest class added is not checked (it will
        // be checked the next time
        int length2 = this.lastModifiedDates.length;
        if (length > length2) {
            length = length2;
        }

        for (int i = 0; i < length; i++) {
            try {
                long lastModified = ((ResourceAttributes) this.resources.getAttributes(this.paths[i])).getLastModified();
                if (lastModified != this.lastModifiedDates[i]) {
                    if (log.isDebugEnabled()) {
                        log.debug("  Resource '" + this.paths[i] + "' was modified; Date is now: " + new java.util.Date(lastModified) + " Was: "
                            + new java.util.Date(this.lastModifiedDates[i]));
                    }
                    return true;
                }
            } catch (NamingException e) {
                log.error("    Resource '" + this.paths[i] + "' is missing");
                return true;
            }
        }

        length = this.jarNames.length;

        // Check if JARs have been added or removed
        if (getJarPath() != null) {

            try {
                NamingEnumeration<?> enumeration = this.resources.listBindings(getJarPath());
                int i = 0;
                while (enumeration.hasMoreElements() && i < length) {
                    NameClassPair ncPair = (NameClassPair) enumeration.nextElement();
                    String name = ncPair.getName();
                    // Ignore non JARs present in the lib folder
                    if (!name.endsWith(".jar")) {
                        continue;
                    }
                    if (!name.equals(this.jarNames[i])) {
                        // Missing JAR
                        log.info("    Additional JARs have been added : '" + name + "'");
                        return true;
                    }
                    i++;
                }
                if (enumeration.hasMoreElements()) {
                    while (enumeration.hasMoreElements()) {
                        NameClassPair ncPair = (NameClassPair) enumeration.nextElement();
                        String name = ncPair.getName();
                        // Additional non-JAR files are allowed
                        if (name.endsWith(".jar")) {
                            // There was more JARs
                            log.info("    Additional JARs have been added");
                            return true;
                        }
                    }
                } else if (i < this.jarNames.length) {
                    // There was less JARs
                    log.info("    Additional JARs have been added");
                    return true;
                }
            } catch (NamingException e) {
                if (log.isDebugEnabled()) {
                    log.debug("    Failed tracking modifications of '" + getJarPath() + "'");
                }
            } catch (ClassCastException e) {
                log.error("    Failed tracking modifications of '" + getJarPath() + "' : " + e.getMessage());
            }

        }

        // No classes have been modified
        return false;
    }

    /**
     * Render a String representation of this object.
     */
    public String toString() {

        StringBuffer sb = new StringBuffer("WebappClassLoader\r\n");
        sb.append("  delegate: ");
        sb.append(this.delegate);
        sb.append("\r\n");
        sb.append("  repositories:\r\n");
        if (this.repositories != null) {
            for (int i = 0; i < this.repositories.length; i++) {
                sb.append("    ");
                sb.append(this.repositories[i]);
                sb.append("\r\n");
            }
        }
        if (this.parent != null) {
            sb.append("----------> Parent Classloader:\r\n");
            sb.append(this.parent.toString());
            sb.append("\r\n");
        }
        return sb.toString();

    }

    // ---------------------------------------------------- ClassLoader Methods

    /**
     * Add the specified URL to the classloader.
     */
    protected void addURL(URL url) {
        super.addURL(url);
        this.hasExternalRepositories = true;
        this.repositoryURLs = null;
    }

    /**
     * Find the specified class in our local repositories, if possible. If not found, throw
     * <code>ClassNotFoundException</code>.
     * 
     * @param name Name of the class to be loaded
     * 
     * @exception ClassNotFoundException if the class was not found
     */
    public Class<?> findClass(String name) throws ClassNotFoundException {

        if (log.isDebugEnabled()) {
            log.debug("    findClass(" + name + ")");
        }

        // Cannot load anything from local repositories if class loader is stopped
        if (!this.started) {
            throw new ClassNotFoundException(name);
        }

        // (1) Permission to define this class when using a SecurityManager
        if (this.securityManager != null) {
            int i = name.lastIndexOf('.');
            if (i >= 0) {
                try {
                    if (log.isTraceEnabled()) {
                        log.trace("      securityManager.checkPackageDefinition");
                    }
                    this.securityManager.checkPackageDefinition(name.substring(0, i));
                } catch (Exception se) {
                    if (log.isTraceEnabled()) {
                        log.trace("      -->Exception-->ClassNotFoundException", se);
                    }
                    throw new ClassNotFoundException(name, se);
                }
            }
        }

        // Ask our superclass to locate this class, if possible
        // (throws ClassNotFoundException if it is not found)
        Class<?> clazz = null;
        try {
            if (log.isTraceEnabled()) {
                log.trace("      findClassInternal(" + name + ")");
            }
            try {
                clazz = findClassInternal(name);
            } catch (ClassNotFoundException cnfe) {
                if (!this.hasExternalRepositories) {
                    throw cnfe;
                }
            } catch (AccessControlException ace) {
                throw new ClassNotFoundException(name, ace);
            } catch (RuntimeException e) {
                if (log.isTraceEnabled()) {
                    log.trace("      -->RuntimeException Rethrown", e);
                }
                throw e;
            }
            if (clazz == null && this.hasExternalRepositories) {
                try {
                    synchronized (this) {
                        clazz = super.findClass(name);
                    }
                } catch (AccessControlException ace) {
                    throw new ClassNotFoundException(name, ace);
                } catch (RuntimeException e) {
                    if (log.isTraceEnabled()) {
                        log.trace("      -->RuntimeException Rethrown", e);
                    }
                    throw e;
                }
            }
            if (clazz == null) {
                if (log.isDebugEnabled()) {
                    log.debug("    --> Returning ClassNotFoundException");
                }
                throw new ClassNotFoundException(name);
            }
        } catch (ClassNotFoundException e) {
            if (log.isTraceEnabled()) {
                log.trace("    --> Passing on ClassNotFoundException");
            }
            throw e;
        }

        // Return the class we have located
        if (log.isTraceEnabled()) {
            log.debug("      Returning class " + clazz);
        }
        if (log.isTraceEnabled() && clazz != null) {
            log.debug("      Loaded by " + clazz.getClassLoader());
        }
        return clazz;

    }

    /**
     * Find the specified resource in our local repository, and return a <code>URL</code> refering to it, or
     * <code>null</code> if this resource cannot be found.
     * 
     * @param name Name of the resource to be found
     */
    public URL findResource(final String name) {

        if (log.isDebugEnabled()) {
            log.debug("    findResource(" + name + ")");
        }

        URL url = null;

        ResourceEntry entry = (ResourceEntry) this.resourceEntries.get(name);
        if (entry == null) {
            entry = findResourceInternal(name, name);
        }
        if (entry != null) {
            url = entry.source;
        }

        if (url == null && this.hasExternalRepositories) {
            url = super.findResource(name);
        }

        if (log.isDebugEnabled()) {
            if (url != null) {
                log.debug("    --> Returning '" + url.toString() + "'");
            } else {
                log.debug("    --> Resource not found, returning null");
            }
        }
        return url;

    }

    /**
     * Return an enumeration of <code>URLs</code> representing all of the resources with the given name. If no
     * resources with this name are found, return an empty enumeration.
     * 
     * @param name Name of the resources to be found
     * 
     * @exception IOException if an input/output error occurs
     */
    public Enumeration<URL> findResources(String name) throws IOException {

        if (log.isDebugEnabled()) {
            log.debug("    findResources(" + name + ")");
        }

        Vector<URL> result = new Vector<URL>();

        int jarFilesLength = this.jarFiles.length;
        int repositoriesLength = this.repositories.length;

        int i;

        // Looking at the repositories
        for (i = 0; i < repositoriesLength; i++) {
            try {
                String fullPath = this.repositories[i] + name;
                this.resources.lookup(fullPath);
                // Note : Not getting an exception here means the resource was
                // found
                try {
                    result.addElement(getURI(new File(this.files[i], name)));
                } catch (MalformedURLException e) {
                    // Ignore
                }
            } catch (NamingException e) {
            }
        }

        // Looking at the JAR files
        synchronized (this.jarFiles) {
            if (openJARs()) {
                for (i = 0; i < jarFilesLength; i++) {
                    JarEntry jarEntry = this.jarFiles[i].getJarEntry(name);
                    if (jarEntry != null) {
                        try {
                            String jarFakeUrl = getURI(this.jarRealFiles[i]).toString();
                            jarFakeUrl = "jar:" + jarFakeUrl + "!/" + name;
                            result.addElement(new URL(jarFakeUrl));
                        } catch (MalformedURLException e) {
                            // Ignore
                        }
                    }
                }
            }
        }

        // Adding the results of a call to the superclass
        if (this.hasExternalRepositories) {
            Enumeration<URL> otherResourcePaths = super.findResources(name);
            while (otherResourcePaths.hasMoreElements()) {
                result.addElement(otherResourcePaths.nextElement());
            }
        }

        return result.elements();
    }

    /**
     * Find the resource with the given name. A resource is some data (images, audio, text, etc.) that can be accessed
     * by class code in a way that is independent of the location of the code. The name of a resource is a "/"-separated
     * path name that identifies the resource. If the resource cannot be found, return <code>null</code>.
     * <p>
     * This method searches according to the following algorithm, returning as soon as it finds the appropriate URL. If
     * the resource cannot be found, returns <code>null</code>.
     * <ul>
     * <li>If the <code>delegate</code> property is set to <code>true</code>, call the <code>getResource()</code>
     * method of the parent class loader, if any.</li>
     * <li>Call <code>findResource()</code> to find this resource in our locally defined repositories.</li>
     * <li>Call the <code>getResource()</code> method of the parent class loader, if any.</li>
     * </ul>
     * 
     * @param name Name of the resource to return a URL for
     */
    public URL getResource(String name) {

        if (log.isDebugEnabled()) {
            log.debug("getResource(" + name + ")");
        }
        URL url = null;

        // (1) Delegate to parent if requested
        if (this.delegate) {
            if (log.isDebugEnabled()) {
                log.debug("  Delegating to parent classloader " + this.parent);
            }
            ClassLoader loader = this.parent;
            if (loader == null) {
                loader = this.system;
            }
            url = loader.getResource(name);
            if (url != null) {
                if (log.isDebugEnabled()) {
                    log.debug("  --> Returning '" + url.toString() + "'");
                }
                return url;
            }
        }

        // (2) Search local repositories
        url = findResource(name);
        if (url != null) {
            // Locating the repository for special handling in the case
            // of a JAR
            if (this.antiJARLocking) {
                ResourceEntry entry = (ResourceEntry) this.resourceEntries.get(name);
                try {
                    String repository = entry.codeBase.toString();
                    if (repository.endsWith(".jar") && !name.endsWith(".class")) {
                        // Copy binary content to the work directory if not present
                        File resourceFile = new File(this.loaderDir, name);
                        url = getURI(resourceFile);
                    }
                } catch (Exception e) {
                    // Ignore
                }
            }
            if (log.isDebugEnabled()) {
                log.debug("  --> Returning '" + url.toString() + "'");
            }
            return url;
        }

        // (3) Delegate to parent unconditionally if not already attempted
        if (!this.delegate) {
            ClassLoader loader = this.parent;
            if (loader == null) {
                loader = this.system;
            }
            url = loader.getResource(name);
            if (url != null) {
                if (log.isDebugEnabled()) {
                    log.debug("  --> Returning '" + url.toString() + "'");
                }
                return url;
            }
        }

        // (4) Resource was not found
        if (log.isDebugEnabled()) {
            log.debug("  --> Resource not found, returning null");
        }
        return null;
    }

    /**
     * Find the resource with the given name, and return an input stream that can be used for reading it. The search
     * order is as described for <code>getResource()</code>, after checking to see if the resource data has been
     * previously cached. If the resource cannot be found, return <code>null</code>.
     * 
     * @param name Name of the resource to return an input stream for
     */
    public InputStream getResourceAsStream(String name) {

        if (log.isDebugEnabled()) {
            log.debug("getResourceAsStream(" + name + ")");
        }
        InputStream stream = null;

        // (0) Check for a cached copy of this resource
        stream = findLoadedResource(name);
        if (stream != null) {
            if (log.isDebugEnabled()) {
                log.debug("  --> Returning stream from cache");
            }
            return stream;
        }

        // (1) Delegate to parent if requested
        if (this.delegate) {
            if (log.isDebugEnabled()) {
                log.debug("  Delegating to parent classloader " + this.parent);
            }
            ClassLoader loader = this.parent;
            if (loader == null) {
                loader = this.system;
            }
            stream = loader.getResourceAsStream(name);
            if (stream != null) {
                // FIXME - cache???
                if (log.isDebugEnabled()) {
                    log.debug("  --> Returning stream from parent");
                }
                return stream;
            }
        }

        // (2) Search local repositories
        if (log.isDebugEnabled()) {
            log.debug("  Searching local repositories");
        }
        URL url = findResource(name);
        if (url != null) {
            // FIXME - cache???
            if (log.isDebugEnabled()) {
                log.debug("  --> Returning stream from local");
            }
            stream = findLoadedResource(name);
            try {
                if (this.hasExternalRepositories && stream == null) {
                    stream = url.openStream();
                }
            } catch (IOException e) {
                ; // Ignore
            }
            if (stream != null) {
                return stream;
            }
        }

        // (3) Delegate to parent unconditionally
        if (!this.delegate) {
            if (log.isDebugEnabled()) {
                log.debug("  Delegating to parent classloader unconditionally " + this.parent);
            }
            ClassLoader loader = this.parent;
            if (loader == null) {
                loader = this.system;
            }
            stream = loader.getResourceAsStream(name);
            if (stream != null) {
                // FIXME - cache???
                if (log.isDebugEnabled()) {
                    log.debug("  --> Returning stream from parent");
                }
                return stream;
            }
        }

        // (4) Resource was not found
        if (log.isDebugEnabled()) {
            log.debug("  --> Resource not found, returning null");
        }
        return null;
    }

    /**
     * Load the class with the specified name. This method searches for classes in the same manner as
     * <code>loadClass(String, boolean)</code> with <code>false</code> as the second argument.
     * 
     * @param name Name of the class to be loaded
     * 
     * @exception ClassNotFoundException if the class was not found
     */
    public Class<?> loadClass(String name) throws ClassNotFoundException {
        return loadClass(name, false);
    }

    /**
     * Load the class with the specified name, searching using the following algorithm until it finds and returns the
     * class. If the class cannot be found, returns <code>ClassNotFoundException</code>.
     * <ul>
     * <li>Call <code>findLoadedClass(String)</code> to check if the class has already been loaded. If it has, the
     * same <code>Class</code> object is returned.</li>
     * <li>If the <code>delegate</code> property is set to <code>true</code>, call the <code>loadClass()</code>
     * method of the parent class loader, if any.</li>
     * <li>Call <code>findClass()</code> to find this class in our locally defined repositories.</li>
     * <li>Call the <code>loadClass()</code> method of our parent class loader, if any.</li>
     * </ul>
     * If the class was found using the above steps, and the <code>resolve</code> flag is <code>true</code>, this
     * method will then call <code>resolveClass(Class)</code> on the resulting Class object.
     * 
     * @param name Name of the class to be loaded
     * @param resolve If <code>true</code> then resolve the class
     * 
     * @exception ClassNotFoundException if the class was not found
     */
    public Class<?> loadClass(String name, boolean resolve) throws ClassNotFoundException {

        if (log.isDebugEnabled()) {
            log.debug("loadClass(" + name + ", " + resolve + ")");
        }
        Class<?> clazz = null;

        // Log access to stopped classloader
        if (!this.started) {
            try {
                throw new IllegalStateException();
            } catch (IllegalStateException e) {
                log.info(sm.getString("webappClassLoader.stopped", name), e);
            }
        }

        // (0) Check our previously loaded local class cache
        clazz = findLoadedClass0(name);
        if (clazz != null) {
            if (log.isDebugEnabled()) {
                log.debug("  Returning class from cache");
            }
            if (resolve) {
                resolveClass(clazz);
            }
            return clazz;
        }

        // (0.1) Check our previously loaded class cache
        clazz = findLoadedClass(name);
        if (clazz != null) {
            if (log.isDebugEnabled()) {
                log.debug("  Returning class from cache");
            }
            if (resolve) {
                resolveClass(clazz);
            }
            return clazz;
        }

        // (0.2) Try loading the class with the system class loader, to prevent
        // the webapp from overriding J2SE classes
        // BEGIN - SpringSource Modifications
        if (delegateToSystemClassLoader(name)) {
            // END - SpringSource Modifications
            try {
                clazz = this.system.loadClass(name);
                if (clazz != null) {
                    if (resolve) {
                        resolveClass(clazz);
                    }
                    return clazz;
                }
            } catch (ClassNotFoundException e) {
                // Ignore
            }
            // BEGIN - SpringSource Modifications
        }
        // END - SpringSource Modifications

        // (0.5) Permission to access this class when using a SecurityManager
        if (this.securityManager != null) {
            int i = name.lastIndexOf('.');
            if (i >= 0) {
                try {
                    this.securityManager.checkPackageAccess(name.substring(0, i));
                } catch (SecurityException se) {
                    String error = "Security Violation, attempt to use " + "Restricted Class: " + name;
                    log.info(error, se);
                    throw new ClassNotFoundException(error, se);
                }
            }
        }

        boolean delegateLoad = this.delegate || filter(name);

        // (1) Delegate to our parent if requested
        if (delegateLoad) {
            if (log.isDebugEnabled()) {
                log.debug("  Delegating to parent classloader1 " + this.parent);
            }
            ClassLoader loader = this.parent;
            if (loader == null) {
                loader = this.system;
            }
            try {
                clazz = loader.loadClass(name);
                if (clazz != null) {
                    if (log.isDebugEnabled()) {
                        log.debug("  Loading class from parent");
                    }
                    if (resolve) {
                        resolveClass(clazz);
                    }
                    return clazz;
                }
            } catch (ClassNotFoundException e) {
                ;
            }
        }

        // (2) Search local repositories
        if (log.isDebugEnabled()) {
            log.debug("  Searching local repositories");
        }
        try {
            clazz = findClass(name);
            if (clazz != null) {
                if (log.isDebugEnabled()) {
                    log.debug("  Loading class from local repository");
                }
                if (resolve) {
                    resolveClass(clazz);
                }
                return clazz;
            }
        } catch (ClassNotFoundException e) {
            ;
        }

        // (3) Delegate to parent unconditionally
        if (!delegateLoad) {
            if (log.isDebugEnabled()) {
                log.debug("  Delegating to parent classloader at end: " + this.parent);
            }
            ClassLoader loader = this.parent;
            if (loader == null) {
                loader = this.system;
            }
            try {
                clazz = loader.loadClass(name);
                if (clazz != null) {
                    if (log.isDebugEnabled()) {
                        log.debug("  Loading class from parent");
                    }
                    if (resolve) {
                        resolveClass(clazz);
                    }
                    return clazz;
                }
            } catch (ClassNotFoundException e) {
                ;
            }
        }

        throw new ClassNotFoundException(name);
    }

    /**
     * Get the Permissions for a CodeSource. If this instance of WebappClassLoader is for a web application context, add
     * read FilePermission or JndiPermissions for the base directory (if unpacked), the context URL, and jar file
     * resources.
     * 
     * @param codeSource where the code was loaded from
     * @return PermissionCollection for CodeSource
     */
    protected PermissionCollection getPermissions(CodeSource codeSource) {

        String codeUrl = codeSource.getLocation().toString();
        PermissionCollection pc;
        if ((pc = (PermissionCollection) this.loaderPC.get(codeUrl)) == null) {
            pc = super.getPermissions(codeSource);
            if (pc != null) {
                Iterator<Permission> perms = this.permissionList.iterator();
                while (perms.hasNext()) {
                    Permission p = (Permission) perms.next();
                    pc.add(p);
                }
                this.loaderPC.put(codeUrl, pc);
            }
        }
        return pc;
    }

    /**
     * Returns the search path of URLs for loading classes and resources. This includes the original list of URLs
     * specified to the constructor, along with any URLs subsequently appended by the addURL() method.
     * 
     * @return the search path of URLs for loading classes and resources.
     */
    public URL[] getURLs() {

        if (this.repositoryURLs != null) {
            return this.repositoryURLs;
        }

        URL[] external = super.getURLs();

        int filesLength = this.files.length;
        int jarFilesLength = this.jarRealFiles.length;
        int length = filesLength + jarFilesLength + external.length;
        int i;

        try {

            URL[] urls = new URL[length];
            for (i = 0; i < length; i++) {
                if (i < filesLength) {
                    urls[i] = getURL(this.files[i], true);
                } else if (i < filesLength + jarFilesLength) {
                    urls[i] = getURL(this.jarRealFiles[i - filesLength], true);
                } else {
                    urls[i] = external[i - filesLength - jarFilesLength];
                }
            }

            this.repositoryURLs = urls;

        } catch (MalformedURLException e) {
            this.repositoryURLs = new URL[0];
        }

        return this.repositoryURLs;
    }

    // ------------------------------------------------------ Lifecycle Methods

    /**
     * Add a lifecycle event listener to this component.
     * 
     * @param listener The listener to add
     */
    public void addLifecycleListener(LifecycleListener listener) {
    }

    /**
     * Get the lifecycle listeners associated with this lifecycle. If this Lifecycle has no listeners registered, a
     * zero-length array is returned.
     */
    public LifecycleListener[] findLifecycleListeners() {
        return new LifecycleListener[0];
    }

    /**
     * Remove a lifecycle event listener from this component.
     * 
     * @param listener The listener to remove
     */
    public void removeLifecycleListener(LifecycleListener listener) {
    }

    /**
     * Start the class loader.
     * 
     * @exception LifecycleException if a lifecycle error occurs
     */
    public void start() throws LifecycleException {

        this.started = true;
        String encoding = null;
        try {
            encoding = System.getProperty("file.encoding");
        } catch (Exception e) {
            return;
        }
        if (encoding.indexOf("EBCDIC") != -1) {
            this.needConvert = true;
        }
    }

    /**
     * Stop the class loader.
     * 
     * @exception LifecycleException if a lifecycle error occurs
     */
    public void stop() throws LifecycleException {

        // Clearing references should be done before setting started to
        // false, due to possible side effects
        clearReferences();

        this.started = false;

        int length = this.files.length;
        for (int i = 0; i < length; i++) {
            this.files[i] = null;
        }

        length = this.jarFiles.length;
        for (int i = 0; i < length; i++) {
            try {
                if (this.jarFiles[i] != null) {
                    this.jarFiles[i].close();
                }
            } catch (IOException e) {
                // Ignore
            }
            this.jarFiles[i] = null;
        }

        this.notFoundResources.clear();
        this.resourceEntries.clear();
        this.resources = null;
        this.repositories = null;
        this.repositoryURLs = null;
        this.files = null;
        this.jarFiles = null;
        this.jarRealFiles = null;
        this.jarPath = null;
        this.jarNames = null;
        this.lastModifiedDates = null;
        this.paths = null;
        this.hasExternalRepositories = false;
        this.parent = null;

        this.permissionList.clear();
        this.loaderPC.clear();

        if (this.loaderDir != null) {
            deleteDir(this.loaderDir);
        }
    }

    /**
     * Used to periodically signal to the classloader to release JAR resources.
     */
    public void closeJARs(boolean force) {
        if (this.jarFiles.length > 0) {
            synchronized (this.jarFiles) {
                if (force || System.currentTimeMillis() > this.lastJarAccessed + 90000) {
                    for (int i = 0; i < this.jarFiles.length; i++) {
                        try {
                            if (this.jarFiles[i] != null) {
                                this.jarFiles[i].close();
                                this.jarFiles[i] = null;
                            }
                        } catch (IOException e) {
                            if (log.isDebugEnabled()) {
                                log.debug("Failed to close JAR", e);
                            }
                        }
                    }
                }
            }
        }
    }

    // ------------------------------------------------------ Protected Methods

    /**
     * Clear references.
     */
    @SuppressWarnings("unchecked") protected void clearReferences() {

        // Unregister any JDBC drivers loaded by this classloader
        Enumeration<?> drivers = DriverManager.getDrivers();
        while (drivers.hasMoreElements()) {
            Driver driver = (Driver) drivers.nextElement();
            if (driver.getClass().getClassLoader() == this) {
                try {
                    DriverManager.deregisterDriver(driver);
                } catch (SQLException e) {
                    log.warn("SQL driver deregistration failed", e);
                }
            }
        }

        // Null out any static or final fields from loaded classes,
        // as a workaround for apparent garbage collection bugs
        if (ENABLE_CLEAR_REFERENCES) {
            Iterator<ResourceEntry> loadedClasses = ((HashMap<String, ResourceEntry>) this.resourceEntries.clone()).values().iterator();
            while (loadedClasses.hasNext()) {
                ResourceEntry entry = (ResourceEntry) loadedClasses.next();
                if (entry.loadedClass != null) {
                    Class<?> clazz = entry.loadedClass;
                    try {
                        Field[] fields = clazz.getDeclaredFields();
                        for (int i = 0; i < fields.length; i++) {
                            Field field = fields[i];
                            int mods = field.getModifiers();
                            if (field.getType().isPrimitive() || field.getName().indexOf("$") != -1) {
                                continue;
                            }
                            if (Modifier.isStatic(mods)) {
                                try {
                                    field.setAccessible(true);
                                    if (Modifier.isFinal(mods)) {
                                        if (!(field.getType().getName().startsWith("java.") || field.getType().getName().startsWith("javax."))) {
                                            nullInstance(field.get(null));
                                        }
                                    } else {
                                        field.set(null, null);
                                        if (log.isDebugEnabled()) {
                                            log.debug("Set field " + field.getName() + " to null in class " + clazz.getName());
                                        }
                                    }
                                } catch (Throwable t) {
                                    if (log.isDebugEnabled()) {
                                        log.debug("Could not set field " + field.getName() + " to null in class " + clazz.getName(), t);
                                    }
                                }
                            }
                        }
                    } catch (Throwable t) {
                        if (log.isDebugEnabled()) {
                            log.debug("Could not clean fields for class " + clazz.getName(), t);
                        }
                    }
                }
            }
        }

        // Clear the IntrospectionUtils cache.
        IntrospectionUtils.clear();

        // Clear the classloader reference in common-logging
        org.apache.juli.logging.LogFactory.release(this);

        // Clear the classloader reference in the VM's bean introspector
        java.beans.Introspector.flushCaches();
    }

    protected void nullInstance(Object instance) {
        if (instance == null) {
            return;
        }
        Field[] fields = instance.getClass().getDeclaredFields();
        for (int i = 0; i < fields.length; i++) {
            Field field = fields[i];
            int mods = field.getModifiers();
            if (field.getType().isPrimitive() || field.getName().indexOf("$") != -1) {
                continue;
            }
            try {
                field.setAccessible(true);
                if (Modifier.isStatic(mods) && Modifier.isFinal(mods)) {
                    // Doing something recursively is too risky
                    continue;
                } else {
                    Object value = field.get(instance);
                    if (null != value) {
                        Class<? extends Object> valueClass = value.getClass();
                        if (!loadedByThisOrChild(valueClass)) {
                            if (log.isDebugEnabled()) {
                                log.debug("Not setting field " + field.getName() + " to null in object of class " + instance.getClass().getName()
                                    + " because the referenced object was of type " + valueClass.getName()
                                    + " which was not loaded by this WebappClassLoader.");
                            }
                        } else {
                            field.set(instance, null);
                            if (log.isDebugEnabled()) {
                                log.debug("Set field " + field.getName() + " to null in class " + instance.getClass().getName());
                            }
                        }
                    }
                }
            } catch (Throwable t) {
                if (log.isDebugEnabled()) {
                    log.debug("Could not set field " + field.getName() + " to null in object instance of class " + instance.getClass().getName(), t);
                }
            }
        }
    }

    /**
     * Determine whether a class was loaded by this class loader or one of its child class loaders.
     */
    protected boolean loadedByThisOrChild(Class<? extends Object> clazz) {
        boolean result = false;
        for (ClassLoader classLoader = clazz.getClassLoader(); null != classLoader; classLoader = classLoader.getParent()) {
            if (classLoader.equals(this)) {
                result = true;
                break;
            }
        }
        return result;
    }

    /**
     * Used to periodically signal to the classloader to release JAR resources.
     */
    protected boolean openJARs() {
        if (this.started && this.jarFiles.length > 0) {
            this.lastJarAccessed = System.currentTimeMillis();
            if (this.jarFiles[0] == null) {
                for (int i = 0; i < this.jarFiles.length; i++) {
                    try {
                        this.jarFiles[i] = new JarFile(this.jarRealFiles[i]);
                    } catch (IOException e) {
                        if (log.isDebugEnabled()) {
                            log.debug("Failed to open JAR", e);
                        }
                        return false;
                    }
                }
            }
        }
        return true;
    }

    /**
     * Find specified class in local repositories.
     * 
     * @return the loaded class, or null if the class isn't found
     */
    protected Class<?> findClassInternal(String name) throws ClassNotFoundException {

        if (!validate(name)) {
            throw new ClassNotFoundException(name);
        }

        String tempPath = name.replace('.', '/');
        String classPath = tempPath + ".class";

        ResourceEntry entry = null;

        entry = findResourceInternal(name, classPath);

        if (entry == null) {
            throw new ClassNotFoundException(name);
        }

        Class<?> clazz = entry.loadedClass;
        if (clazz != null) {
            return clazz;
        }

        synchronized (this) {
            if (entry.binaryContent == null && entry.loadedClass == null) {
                throw new ClassNotFoundException(name);
            }

            // Looking up the package
            String packageName = null;
            int pos = name.lastIndexOf('.');
            if (pos != -1) {
                packageName = name.substring(0, pos);
            }

            Package pkg = null;

            if (packageName != null) {
                pkg = getPackage(packageName);
                // Define the package (if null)
                if (pkg == null) {
                    try {
                        if (entry.manifest == null) {
                            definePackage(packageName, null, null, null, null, null, null, null);
                        } else {
                            definePackage(packageName, entry.manifest, entry.codeBase);
                        }
                    } catch (IllegalArgumentException e) {
                        // Ignore: normal error due to dual definition of package
                    }
                    pkg = getPackage(packageName);
                }
            }

            if (this.securityManager != null) {

                // Checking sealing
                if (pkg != null) {
                    boolean sealCheck = true;
                    if (pkg.isSealed()) {
                        sealCheck = pkg.isSealed(entry.codeBase);
                    } else {
                        sealCheck = entry.manifest == null || !isPackageSealed(packageName, entry.manifest);
                    }
                    if (!sealCheck) {
                        throw new SecurityException("Sealing violation loading " + name + " : Package " + packageName + " is sealed.");
                    }
                }

            }

            if (entry.loadedClass == null) {
                clazz = defineClass(name, entry.binaryContent, 0, entry.binaryContent.length, new CodeSource(entry.codeBase, entry.certificates));
                entry.loadedClass = clazz;
                entry.binaryContent = null;
                entry.source = null;
                entry.codeBase = null;
                entry.manifest = null;
                entry.certificates = null;
            } else {
                clazz = entry.loadedClass;
            }
        }

        return clazz;
    }

    /**
     * Find specified resource in local repositories. This block will execute under an AccessControl.doPrivilege block.
     * 
     * @return the loaded resource, or null if the resource isn't found
     */
    protected ResourceEntry findResourceInternal(File file, String path) {
        ResourceEntry entry = new ResourceEntry();
        try {
            entry.source = getURI(new File(file, path));
            entry.codeBase = getURL(new File(file, path), false);
        } catch (MalformedURLException e) {
            return null;
        }
        return entry;
    }

    /**
     * Find specified resource in local repositories.
     * 
     * @return the loaded resource, or null if the resource isn't found
     */
    protected ResourceEntry findResourceInternal(String name, String path) {

        if (!this.started) {
            log.info(sm.getString("webappClassLoader.stopped", name));
            return null;
        }

        if (name == null || path == null) {
            return null;
        }

        ResourceEntry entry = (ResourceEntry) this.resourceEntries.get(name);
        if (entry != null) {
            return entry;
        }

        int contentLength = -1;
        InputStream binaryStream = null;

        int jarFilesLength = this.jarFiles.length;
        int repositoriesLength = this.repositories.length;

        int i;

        Resource resource = null;

        boolean fileNeedConvert = false;

        for (i = 0; entry == null && i < repositoriesLength; i++) {
            try {

                String fullPath = this.repositories[i] + path;

                Object lookupResult = this.resources.lookup(fullPath);
                if (lookupResult instanceof Resource) {
                    resource = (Resource) lookupResult;
                }

                // Note : Not getting an exception here means the resource was
                // found
                if (this.securityManager != null) {
                    PrivilegedAction<?> dp = new PrivilegedFindResource(this.files[i], path);
                    entry = (ResourceEntry) AccessController.doPrivileged(dp);
                } else {
                    entry = findResourceInternal(this.files[i], path);
                }

                ResourceAttributes attributes = (ResourceAttributes) this.resources.getAttributes(fullPath);
                contentLength = (int) attributes.getContentLength();
                entry.lastModified = attributes.getLastModified();

                if (resource != null) {

                    try {
                        binaryStream = resource.streamContent();
                    } catch (IOException e) {
                        return null;
                    }

                    if (this.needConvert) {
                        if (path.endsWith(".properties")) {
                            fileNeedConvert = true;
                        }
                    }

                    // Register the full path for modification checking
                    // Note: Only syncing on a 'constant' object is needed
                    synchronized (this.allPermission) {

                        int j;

                        long[] result2 = new long[this.lastModifiedDates.length + 1];
                        for (j = 0; j < this.lastModifiedDates.length; j++) {
                            result2[j] = this.lastModifiedDates[j];
                        }
                        result2[this.lastModifiedDates.length] = entry.lastModified;
                        this.lastModifiedDates = result2;

                        String[] result = new String[this.paths.length + 1];
                        for (j = 0; j < this.paths.length; j++) {
                            result[j] = this.paths[j];
                        }
                        result[this.paths.length] = fullPath;
                        this.paths = result;
                    }
                }

            } catch (NamingException e) {
            }
        }

        if (entry == null && this.notFoundResources.containsKey(name)) {
            return null;
        }

        JarEntry jarEntry = null;

        synchronized (this.jarFiles) {

            if (!openJARs()) {
                return null;
            }
            for (i = 0; entry == null && i < jarFilesLength; i++) {

                jarEntry = this.jarFiles[i].getJarEntry(path);

                if (jarEntry != null) {

                    entry = new ResourceEntry();
                    try {
                        entry.codeBase = getURL(this.jarRealFiles[i], false);
                        String jarFakeUrl = getURI(this.jarRealFiles[i]).toString();
                        jarFakeUrl = "jar:" + jarFakeUrl + "!/" + path;
                        entry.source = new URL(jarFakeUrl);
                        entry.lastModified = this.jarRealFiles[i].lastModified();
                    } catch (MalformedURLException e) {
                        return null;
                    }
                    contentLength = (int) jarEntry.getSize();
                    try {
                        entry.manifest = this.jarFiles[i].getManifest();
                        binaryStream = this.jarFiles[i].getInputStream(jarEntry);
                    } catch (IOException e) {
                        return null;
                    }

                    // Extract resources contained in JAR to the workdir
                    if (this.antiJARLocking && !path.endsWith(".class")) {
                        byte[] buf = new byte[1024];
                        File resourceFile = new File(this.loaderDir, jarEntry.getName());
                        if (!resourceFile.exists()) {
                            Enumeration<?> entries = this.jarFiles[i].entries();
                            while (entries.hasMoreElements()) {
                                JarEntry jarEntry2 = (JarEntry) entries.nextElement();
                                if (!jarEntry2.isDirectory() && !jarEntry2.getName().endsWith(".class")) {
                                    resourceFile = new File(this.loaderDir, jarEntry2.getName());
                                    resourceFile.getParentFile().mkdirs();
                                    FileOutputStream os = null;
                                    InputStream is = null;
                                    try {
                                        is = this.jarFiles[i].getInputStream(jarEntry2);
                                        os = new FileOutputStream(resourceFile);
                                        while (true) {
                                            int n = is.read(buf);
                                            if (n <= 0) {
                                                break;
                                            }
                                            os.write(buf, 0, n);
                                        }
                                    } catch (IOException e) {
                                        // Ignore
                                    } finally {
                                        try {
                                            if (is != null) {
                                                is.close();
                                            }
                                        } catch (IOException e) {
                                        }
                                        try {
                                            if (os != null) {
                                                os.close();
                                            }
                                        } catch (IOException e) {
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }

            if (entry == null) {
                synchronized (this.notFoundResources) {
                    this.notFoundResources.put(name, name);
                }
                return null;
            }

            if (binaryStream != null) {

                byte[] binaryContent = new byte[contentLength];

                int pos = 0;
                try {

                    while (true) {
                        int n = binaryStream.read(binaryContent, pos, binaryContent.length - pos);
                        if (n <= 0) {
                            break;
                        }
                        pos += n;
                    }
                    binaryStream.close();
                } catch (IOException e) {
                    e.printStackTrace();
                    return null;
                } catch (Exception e) {
                    e.printStackTrace();
                    return null;
                }

                if (fileNeedConvert) {
                    String str = new String(binaryContent, 0, pos);
                    try {
                        binaryContent = str.getBytes("UTF-8");
                    } catch (Exception e) {
                        return null;
                    }
                }
                entry.binaryContent = binaryContent;

                // The certificates are only available after the JarEntry
                // associated input stream has been fully read
                if (jarEntry != null) {
                    entry.certificates = jarEntry.getCertificates();
                }
            }
        }

        // Add the entry in the local resource repository
        synchronized (this.resourceEntries) {
            // Ensures that all the threads which may be in a race to load
            // a particular class all end up with the same ResourceEntry
            // instance
            ResourceEntry entry2 = (ResourceEntry) this.resourceEntries.get(name);
            if (entry2 == null) {
                this.resourceEntries.put(name, entry);
            } else {
                entry = entry2;
            }
        }

        return entry;
    }

    /**
     * Returns true if the specified package name is sealed according to the given manifest.
     */
    protected boolean isPackageSealed(String name, Manifest man) {

        String path = name.replace('.', '/') + '/';
        Attributes attr = man.getAttributes(path);
        String sealed = null;
        if (attr != null) {
            sealed = attr.getValue(Name.SEALED);
        }
        if (sealed == null) {
            if ((attr = man.getMainAttributes()) != null) {
                sealed = attr.getValue(Name.SEALED);
            }
        }
        return "true".equalsIgnoreCase(sealed);
    }

    /**
     * Finds the resource with the given name if it has previously been loaded and cached by this class loader, and
     * return an input stream to the resource data. If this resource has not been cached, return <code>null</code>.
     * 
     * @param name Name of the resource to return
     */
    protected InputStream findLoadedResource(String name) {

        ResourceEntry entry = (ResourceEntry) this.resourceEntries.get(name);
        if (entry != null) {
            if (entry.binaryContent != null) {
                return new ByteArrayInputStream(entry.binaryContent);
            }
        }
        return null;
    }

    /**
     * Finds the class with the given name if it has previously been loaded and cached by this class loader, and return
     * the Class object. If this class has not been cached, return <code>null</code>.
     * 
     * @param name Name of the resource to return
     */
    protected Class<?> findLoadedClass0(String name) {

        ResourceEntry entry = (ResourceEntry) this.resourceEntries.get(name);
        if (entry != null) {
            return entry.loadedClass;
        }
        return null; // FIXME - findLoadedResource()
    }

    /**
     * Refresh the system policy file, to pick up eventual changes.
     */
    protected void refreshPolicy() {

        try {
            // The policy file may have been modified to adjust
            // permissions, so we're reloading it when loading or
            // reloading a Context
            Policy policy = Policy.getPolicy();
            policy.refresh();
        } catch (AccessControlException e) {
            // Some policy files may restrict this, even for the core,
            // so this exception is ignored
        }
    }

    /**
     * Filter classes.
     * 
     * @param name class name
     * @return true if the class should be filtered
     */
    protected boolean filter(String name) {

        if (name == null) {
            return false;
        }

        // Looking up the package
        String packageName = null;
        int pos = name.lastIndexOf('.');
        if (pos != -1) {
            packageName = name.substring(0, pos);
        } else {
            return false;
        }

        for (int i = 0; i < packageTriggers.length; i++) {
            if (packageName.startsWith(packageTriggers[i])) {
                return true;
            }
        }

        return false;
    }

    /**
     * Validate a classname. As per SRV.9.7.2, we must restrict loading of classes from J2SE (java.*) and classes of the
     * servlet API (javax.servlet.*). That should enhance robustness and prevent a number of user error (where an older
     * version of servlet.jar would be present in /WEB-INF/lib).
     * 
     * @param name class name
     * @return true if the name is valid
     */
    protected boolean validate(String name) {

        if (name == null) {
            return false;
        }
        if (name.startsWith("java.")) {
            return false;
        }

        return true;
    }

    /**
     * Check the specified JAR file, and return <code>true</code> if it does not contain any of the trigger classes.
     * 
     * @param jarfile The JAR file to be checked
     * 
     * @exception IOException if an input/output error occurs
     */
    protected boolean validateJarFile(File jarfile) throws IOException {

        if (triggers == null) {
            return true;
        }
        JarFile jarFile = new JarFile(jarfile);
        for (int i = 0; i < triggers.length; i++) {
            Class<?> clazz = null;
            try {
                if (this.parent != null) {
                    clazz = this.parent.loadClass(triggers[i]);
                } else {
                    clazz = Class.forName(triggers[i]);
                }
            } catch (Throwable t) {
                clazz = null;
            }
            if (clazz == null) {
                continue;
            }
            String name = triggers[i].replace('.', '/') + ".class";
            if (log.isDebugEnabled()) {
                log.debug(" Checking for " + name);
            }
            JarEntry jarEntry = jarFile.getJarEntry(name);
            if (jarEntry != null) {
                log.info("validateJarFile(" + jarfile + ") - jar not loaded. See Servlet Spec 2.3, " + "section 9.7.2. Offending class: " + name);
                jarFile.close();
                return false;
            }
        }
        jarFile.close();
        return true;
    }

    /**
     * Get URL.
     */
    protected URL getURL(File file, boolean encoded) throws MalformedURLException {

        File realFile = file;
        try {
            realFile = realFile.getCanonicalFile();
        } catch (IOException e) {
            // Ignore
        }
        if (encoded) {
            return getURI(realFile);
        } else {
            return realFile.toURL();
        }
    }

    /**
     * Get URL.
     */
    protected URL getURI(File file) throws MalformedURLException {

        File realFile = file;
        try {
            realFile = realFile.getCanonicalFile();
        } catch (IOException e) {
            // Ignore
        }
        return realFile.toURI().toURL();
    }

    /**
     * Delete the specified directory, including all of its contents and subdirectories recursively.
     * 
     * @param dir File object representing the directory to be deleted
     */
    protected static void deleteDir(File dir) {

        String files[] = dir.list();
        if (files == null) {
            files = new String[0];
        }
        for (int i = 0; i < files.length; i++) {
            File file = new File(dir, files[i]);
            if (file.isDirectory()) {
                deleteDir(file);
            } else {
                file.delete();
            }
        }
        dir.delete();
    }

    // -------------------------------------------------------------------------
    // --- SpringSource Modifications ------------------------------------------
    // -------------------------------------------------------------------------

    /**
     * Determines whether or not this class loader should delegate to the system class loader to load the {@link Class}
     * with the supplied <code>name</code>, based on configured <em>include</em> and <em>exclude</em> package
     * filters.
     * 
     * @param name the name of the class the check
     * @return <code>true</code> if this class loader should delegate to the system class loader to load the named
     *         class.
     */
    protected boolean delegateToSystemClassLoader(String name) {
        List<String> includeFilters = Arrays.asList("java.", "javax.accessibility", "javax.activity", "javax.crypto", "javax.crypto.interfaces",
            "javax.crypto.spec", "javax.imageio", "javax.imageio.event", "javax.imageio.metadata", "javax.imageio.plugins.bmp",
            "javax.imageio.plugins.jpeg", "javax.imageio.spi", "javax.imageio.stream", "javax.management", "javax.management.loading",
            "javax.management.modelmbean", "javax.management.monitor", "javax.management.openmbean", "javax.management.relation",
            "javax.management.remote", "javax.management.remote.rmi", "javax.management.timer", "javax.naming", "javax.naming.directory",
            "javax.naming.event", "javax.naming.ldap", "javax.naming.spi", "javax.net", "javax.net.ssl", "javax.print", "javax.print.attribute",
            "javax.print.attribute.standard", "javax.print.event", "javax.rmi", "javax.rmi.CORBA", "javax.rmi.ssl", "javax.security.auth",
            "javax.security.auth.callback", "javax.security.auth.kerberos", "javax.security.auth.login", "javax.security.auth.spi",
            "javax.security.auth.x500", "javax.security.cert", "javax.security.sasl", "javax.sound.midi", "javax.sound.midi.spi",
            "javax.sound.sampled", "javax.sound.sampled.spi", "javax.sql", "javax.sql.rowset", "javax.sql.rowset.serial", "javax.sql.rowset.spi",
            "javax.swing", "javax.swing.border", "javax.swing.colorchooser", "javax.swing.event", "javax.swing.filechooser", "javax.swing.plaf",
            "javax.swing.plaf.basic", "javax.swing.plaf.metal", "javax.swing.plaf.multi", "javax.swing.plaf.synth", "javax.swing.table",
            "javax.swing.text", "javax.swing.text.html", "javax.swing.text.html.parser", "javax.swing.text.rtf", "javax.swing.tree",
            "javax.swing.undo", "javax.xml", "javax.xml.datatype", "javax.xml.namespace", "javax.xml.parsers", "javax.xml.transform",
            "javax.xml.transform.dom", "javax.xml.transform.sax", "javax.xml.transform.stream", "javax.xml.validation", "javax.xml.xpath",
            "org.ietf.jgss", "org.omg.CORBA", "org.omg.CORBA_2_3", "org.omg.CORBA_2_3.portable", "org.omg.CORBA.DynAnyPackage",
            "org.omg.CORBA.ORBPackage", "org.omg.CORBA.portable", "org.omg.CORBA.TypeCodePackage", "org.omg.CosNaming",
            "org.omg.CosNaming.NamingContextExtPackage", "org.omg.CosNaming.NamingContextPackage", "org.omg.Dynamic", "org.omg.DynamicAny",
            "org.omg.DynamicAny.DynAnyFactoryPackage", "org.omg.DynamicAny.DynAnyPackage", "org.omg.IOP", "org.omg.IOP.CodecFactoryPackage",
            "org.omg.IOP.CodecPackage", "org.omg.Messaging", "org.omg.PortableInterceptor", "org.omg.PortableInterceptor.ORBInitInfoPackage",
            "org.omg.PortableServer", "org.omg.PortableServer.CurrentPackage", "org.omg.PortableServer.POAManagerPackage",
            "org.omg.PortableServer.POAPackage", "org.omg.PortableServer.portable", "org.omg.PortableServer.ServantLocatorPackage",
            "org.omg.SendingContext", "org.omg.stub.java.rmi", "org.w3c.dom", "org.w3c.dom.bootstrap", "org.w3c.dom.events", "org.w3c.dom.ls",
            "org.xml.sax", "org.xml.sax.ext", "org.xml.sax.helpers");

        List<String> excludeFilters = Collections.<String> emptyList();
        // Arrays.asList("javax.el", "javax.servlet", "org.apache.catalina", "org.apache.jasper");

        boolean include = false;
        boolean exclude = false;
        if (name != null) {

            for (String filter : excludeFilters) {
                exclude = name.startsWith(filter);
                if (exclude) {
                    break;
                }
            }

            if (!exclude) {
                for (String filter : includeFilters) {
                    include = name.startsWith(filter);
                    if (include) {
                        break;
                    }
                }
            }
        } else {
            include = true;
        }
        return include && !exclude;
    }

}
