/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.system.core;

import java.lang.management.GarbageCollectorMXBean;
import java.lang.management.ManagementFactory;
import java.lang.management.MemoryMXBean;
import java.lang.management.MemoryPoolMXBean;
import java.lang.management.MemoryUsage;
import java.lang.management.OperatingSystemMXBean;
import java.lang.management.RuntimeMXBean;
import java.util.Dictionary;
import java.util.Enumeration;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import org.eclipse.osgi.service.resolver.BundleDescription;
import org.eclipse.osgi.service.resolver.ExportPackageDescription;
import org.eclipse.osgi.service.resolver.PlatformAdmin;
import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.Constants;
import org.osgi.framework.ServiceReference;

import com.springsource.server.kernel.core.BundleDescriptor;
import com.springsource.server.kernel.core.Subsystem;
import com.springsource.server.kernel.core.SubsystemRegistry;
import com.springsource.server.osgi.framework.OsgiFramework;
import com.springsource.server.serviceability.ffdc.DumpExtension;
import com.springsource.server.serviceability.ffdc.DumpWriter.ListWriter;
import com.springsource.server.serviceability.ffdc.DumpWriter.MapWriter;
import com.springsource.server.serviceability.ffdc.DumpWriter.SectionWriter;
import com.sun.management.GcInfo;

/**
 * Implementation of {@link DumpExtension} which contributes System-specific information to the system subsystem dump
 * files. <p/>
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * Not thread-safe.
 * 
 * @author Damilola Senbanjo
 * @author Rob Harrop
 * @author Sam Brannen
 * @since 1.0
 */
public class SystemDumpExtension implements DumpExtension {

    private final OsgiFramework osgiFramework;

    private final SubsystemRegistry subsystemRegistry;

    /**
     * Constructs a new SystemDumpExtension
     * 
     * @param osgiFramework the {@link OsgiFramework}
     */
    public SystemDumpExtension(SubsystemRegistry subsystemRegistry, OsgiFramework osgiFramework) {
        this.osgiFramework = osgiFramework;
        this.subsystemRegistry = subsystemRegistry;
    }

    /**
     * Dumps the operating system-specific information to the {@link SectionWriter} using the Java
     * {@link ManagementFactory} {@link OperatingSystemMXBean} and {@link com.sun.management.OperatingSystemMXBean}
     * 
     * @param sectionWriter the {@link SectionWriter} being written to
     */
    public void dumpOperatingSystemInfo(SectionWriter sectionWriter) {
        OperatingSystemMXBean operatingSystemMXBean = ManagementFactory.getOperatingSystemMXBean();
        com.sun.management.OperatingSystemMXBean sunOs = (com.sun.management.OperatingSystemMXBean) operatingSystemMXBean;

        sectionWriter//
        .scalar("OS Name", operatingSystemMXBean.getName())//
        .scalar("Architecture", operatingSystemMXBean.getArch())//
        .scalar("Version", operatingSystemMXBean.getVersion())//
        .scalar("Processors", operatingSystemMXBean.getAvailableProcessors())//
        .scalar("Committed Virtual Memory Size", sunOs.getCommittedVirtualMemorySize())//
        .scalar("Free Physical Memory", sunOs.getFreePhysicalMemorySize())//
        .scalar("Free Swap Space", sunOs.getFreeSwapSpaceSize())//
        .scalar("Process CPU Time", sunOs.getProcessCpuTime())//
        .scalar("Total Physical Memory Size", sunOs.getTotalPhysicalMemorySize())//
        .scalar("Total Swap Space Size", sunOs.getTotalSwapSpaceSize());
    }

    /**
     * Dumps the Java-related system properties to the {@link SectionWriter} using Java's {@link System} class
     * 
     * @param sectionWriter the {@link SectionWriter} being written to
     */
    public void dumpJavaSystemProperties(SectionWriter sectionWriter) {
        MapWriter<SectionWriter> map = sectionWriter.map("Environment Variables");
        for (Map.Entry<String, String> environment : System.getenv().entrySet()) {
            map.scalar(environment.getKey(), environment.getValue());
        }
    }

    /**
     * Dumps information about the OSGi environment, including installed bundles and their import and export packages,
     * to the {@link SectionWriter}
     * 
     * @param sectionWriter the {@link SectionWriter} being written to
     */
    public void dumpEquinoxInfo(SectionWriter sectionWriter) {
        BundleContext bc = this.osgiFramework.getBundleContext();
        ServiceReference serviceRef = bc.getServiceReference(PlatformAdmin.class.getName());
        PlatformAdmin serverAdmin = (PlatformAdmin) bc.getService(serviceRef);

        try {
            MapWriter<SectionWriter> equinoxInformation = sectionWriter.map("Equinox Information");

            equinoxInformation.scalar("Boot delegation", bc.getProperty(Constants.FRAMEWORK_BOOTDELEGATION));
            equinoxInformation.scalar("Execution Environment", bc.getProperty(Constants.FRAMEWORK_EXECUTIONENVIRONMENT));
            equinoxInformation.scalar("System Packages", bc.getProperty(Constants.FRAMEWORK_SYSTEMPACKAGES));
            equinoxInformation.scalar("Language", bc.getProperty(Constants.FRAMEWORK_LANGUAGE));

            ListWriter<MapWriter<SectionWriter>> bundlesWriter = equinoxInformation.list("Bundles");

            Bundle[] bundles = bc.getBundles();

            for (Bundle bundle : bundles) {
                MapWriter<ListWriter<MapWriter<SectionWriter>>> bundleWriter = bundlesWriter.map();
                dumpBundleInformation(bundleWriter, bundle, serverAdmin);
                bundleWriter.end();
            }

            equinoxInformation.scalar("Vendor", bc.getProperty(Constants.FRAMEWORK_VENDOR));
            equinoxInformation.scalar("Version", bc.getProperty(Constants.FRAMEWORK_VERSION));
            equinoxInformation.end();

        } finally {
            bc.ungetService(serviceRef);
        }
    }

    private void dumpBundleInformation(MapWriter<ListWriter<MapWriter<SectionWriter>>> bundleWriter, Bundle bundle, PlatformAdmin serverAdmin) {
        bundleWriter.scalar("ID", bundle.getBundleId());
        bundleWriter.scalar("Bundle symbolic name", bundle.getSymbolicName());
        bundleWriter.scalar("Bundle Version", bundle.getHeaders().get("Bundle-Version"));
        bundleWriter.scalar("Location", bundle.getLocation());
        bundleWriter.scalar("State", getState(bundle));
        bundleWriter.scalar("Services in use", bundle.getServicesInUse());
        bundleWriter.scalar("Registered services", bundle.getRegisteredServices());

        BundleDescription bundleDescription = serverAdmin.getState(false).getBundle(bundle.getBundleId());

        ListWriter<MapWriter<ListWriter<MapWriter<SectionWriter>>>> exportedPackagesWriter = bundleWriter.list("Exported packages");

        ExportPackageDescription[] exportedPackages = bundleDescription.getExportPackages();

        if (exportedPackages != null) {
            for (ExportPackageDescription exportedPackage : exportedPackages) {
                MapWriter<ListWriter<MapWriter<ListWriter<MapWriter<SectionWriter>>>>> exportedPackageWriter = exportedPackagesWriter.map();
                dumpExportedPackage(exportedPackageWriter, exportedPackage);
                exportedPackageWriter.end();
            }
        }

        exportedPackagesWriter.end();

        ListWriter<MapWriter<ListWriter<MapWriter<SectionWriter>>>> visiblePackagesWriter = bundleWriter.list("Visible packages");

        ExportPackageDescription[] visiblePackages = serverAdmin.getStateHelper().getVisiblePackages(bundleDescription);

        if (visiblePackages != null) {
            for (ExportPackageDescription visiblePackage : visiblePackages) {
                MapWriter<ListWriter<MapWriter<ListWriter<MapWriter<SectionWriter>>>>> visiblePackageWriter = visiblePackagesWriter.map();
                dumpVisiblePackage(visiblePackageWriter, visiblePackage);
                visiblePackageWriter.end();
            }
        }

        visiblePackagesWriter.end();

        Dictionary<?, ?> headers = bundle.getHeaders();

        MapWriter<MapWriter<ListWriter<MapWriter<SectionWriter>>>> manifestWriter = bundleWriter.map("Manifest");

        Enumeration<?> keys = headers.keys();
        while (keys.hasMoreElements()) {
            Object key = keys.nextElement();
            Object value = headers.get(key);
            manifestWriter.scalar((String) key, value);
        }

        manifestWriter.end();
    }

    private void dumpExportedPackage(MapWriter<ListWriter<MapWriter<ListWriter<MapWriter<SectionWriter>>>>> exportedPackageWriter,
        ExportPackageDescription exportedPackage) {
        exportedPackageWriter.scalar("Package name", exportedPackage.getName());
        exportedPackageWriter.scalar("Version", exportedPackage.getVersion());
    }

    private void dumpVisiblePackage(MapWriter<ListWriter<MapWriter<ListWriter<MapWriter<SectionWriter>>>>> visiblePackageWriter,
        ExportPackageDescription visiblePackage) {
        visiblePackageWriter.scalar("Package name", visiblePackage.getName());
        visiblePackageWriter.scalar("Version", visiblePackage.getVersion());
        visiblePackageWriter.scalar("Supplier bundle ID", visiblePackage.getSupplier().getBundleId());
    }

    /**
     * Dumps Java virtual machine implementation information to the {@link SectionWriter} using the Java
     * {@link ManagementFactory} {@link RuntimeMXBean}
     * 
     * @param sectionWriter the {@link SectionWriter} being written to
     */
    public void dumpJavaVMInfo(SectionWriter sectionWriter) {
        sectionWriter//
        .scalar("VM Name", ManagementFactory.getRuntimeMXBean().getVmName())//
        .scalar("VM Vendor", ManagementFactory.getRuntimeMXBean().getVmVendor())//
        .scalar("VM Version", ManagementFactory.getRuntimeMXBean().getVmVersion())//
        .scalar("MSpec Version", ManagementFactory.getRuntimeMXBean().getManagementSpecVersion());
    }

    /**
     * Dumps the Java virtual machine specification information to the {@link SectionWriter} using the Java
     * {@link ManagementFactory} {@link RuntimeMXBean}
     * 
     * @param sectionWriter the {@link SectionWriter} being written to
     */
    public void dumpSpecInfo(SectionWriter sectionWriter) {
        sectionWriter//
        .scalar("Spec Name", ManagementFactory.getRuntimeMXBean().getSpecName())//
        .scalar("Spec Vendor", ManagementFactory.getRuntimeMXBean().getSpecVendor())//
        .scalar("Spec Version", ManagementFactory.getRuntimeMXBean().getSpecVersion());
    }

    /**
     * Dumps the general runtime information of the Java virtual machine to the {@link SectionWriter} using the Java
     * {@link ManagementFactory} {@link RuntimeMXBean}
     * 
     * @param section the {@link SectionWriter} being written to
     */
    public void dumpGeneralRuntimeInfo(SectionWriter section) {
        ListWriter<SectionWriter> sectionList = section.list("General Runtime Information");
        sectionList.map()//
        .scalar("Input Arguments", ManagementFactory.getRuntimeMXBean().getInputArguments())//
        .scalar("Name", ManagementFactory.getRuntimeMXBean().getName())//
        .scalar("Start Time", ManagementFactory.getRuntimeMXBean().getStartTime())//
        .scalar("Uptime", ManagementFactory.getRuntimeMXBean().getUptime());

        Map<String, String> props = new TreeMap<String, String>(ManagementFactory.getRuntimeMXBean().getSystemProperties());
        MapWriter<SectionWriter> map = section.map("System Properties");
        for (Map.Entry<String, String> entry : props.entrySet()) {
            map.scalar(entry.getKey(), entry.getValue());
        }
    }

    /**
     * Dumps information regarding the class path to the {@link SectionWriter} using the Java {@link ManagementFactory}
     * {@link RuntimeMXBean}
     * 
     * @param sectionWriter the {@link SectionWriter} being written to
     */
    public void dumpClasspathInfo(SectionWriter sectionWriter) {
        sectionWriter//
        .scalar("Boot Classpath", ManagementFactory.getRuntimeMXBean().getBootClassPath())//
        .scalar("Classpath", ManagementFactory.getRuntimeMXBean().getClassPath())//
        .scalar("Library Classpath", ManagementFactory.getRuntimeMXBean().getLibraryPath());
    }

    /**
     * Dumps the memory usage information to the {@link SectionWriter} using Java's {@link ManagementFactory}
     * {@link MemoryMXBean}
     * 
     * @param sectionWriter the {@link SectionWriter} being written to
     */
    public void dumpMemoryInfo(SectionWriter sectionWriter) {
        sectionWriter//
        .scalar("Heap Memory Usage", ManagementFactory.getMemoryMXBean().getHeapMemoryUsage().toString())//
        .scalar("Non Heap Usage", ManagementFactory.getMemoryMXBean().getNonHeapMemoryUsage().toString());

        List<MemoryPoolMXBean> pools = ManagementFactory.getMemoryPoolMXBeans();
        MapWriter<SectionWriter> poolsMap = sectionWriter.map("Memory Pools");

        for (MemoryPoolMXBean pool : pools) {
            MemoryUsage peak = pool.getPeakUsage();
            poolsMap.map(pool.getName())//
            .scalar("Used", peak.getUsed())//
            .scalar("Committed", peak.getCommitted())//
            .scalar("Init", peak.getInit())//
            .scalar("Max", peak.getMax())//
            .end();
        }
        poolsMap.end();
    }

    /**
     * Dumps the garbage collection information to the {@link SectionWriter} using Java's {@link ManagementFactory}
     * {@link GarbageCollectorMXBean} and {@link com.sun.management.GarbageCollectorMXBean}
     * 
     * @param sectionWriter the {@link SectionWriter} being written to
     */
    public void dumpGarbageCollectionInfo(SectionWriter sectionWriter) {
        List<GarbageCollectorMXBean> gmxbeans = ManagementFactory.getGarbageCollectorMXBeans();
        ListWriter<SectionWriter> list = sectionWriter.list("Garbage Collection Information");

        for (GarbageCollectorMXBean bean : gmxbeans) {
            list.map()//
            .scalar(bean.getName(), bean.getCollectionCount())//
            .scalar(bean.getName(), bean.getCollectionTime());

            com.sun.management.GarbageCollectorMXBean sunbean = (com.sun.management.GarbageCollectorMXBean) bean;
            GcInfo lastGcInfo = sunbean.getLastGcInfo();
            if (lastGcInfo != null) {
                MapWriter<ListWriter<ListWriter<SectionWriter>>> gcMap = list.list().map();
                gcMap//
                .scalar("Start time", lastGcInfo.getStartTime())//
                .scalar("Duration", lastGcInfo.getDuration())//
                .scalar("End time", lastGcInfo.getEndTime())//
                .scalar("ID", lastGcInfo.getId());
                gcMap.list("Memory usage before").map()//
                .scalar("Tenured Gen", lastGcInfo.getMemoryUsageBeforeGc().get("Tenured Gen"))//
                .scalar("Eden Space", lastGcInfo.getMemoryUsageBeforeGc().get("Eden Space"))//
                .scalar("Perm Gen", lastGcInfo.getMemoryUsageBeforeGc().get("Perm Gen"))//
                .scalar("Survivor Space", lastGcInfo.getMemoryUsageBeforeGc().get("Survivor Space"))//
                .scalar("Code Cache", lastGcInfo.getMemoryUsageBeforeGc().get("Code Cache"))//
                .scalar("Perm Gen [shared-ro]", lastGcInfo.getMemoryUsageBeforeGc().get("Perm Gen [shared-ro]"))//
                .scalar("Perm Gen [shared-rw]", lastGcInfo.getMemoryUsageBeforeGc().get("Perm Gen [shared-rw]"));
                gcMap.list("Memory usage after").map()//
                .scalar("Tenured Gen", lastGcInfo.getMemoryUsageAfterGc().get("Tenured Gen"))//
                .scalar("Eden Space", lastGcInfo.getMemoryUsageAfterGc().get("Eden Space"))//
                .scalar("Perm Gen", lastGcInfo.getMemoryUsageAfterGc().get("Perm Gen"))//
                .scalar("Survivor Space", lastGcInfo.getMemoryUsageAfterGc().get("Survivor Space"))//
                .scalar("Code Cache", lastGcInfo.getMemoryUsageAfterGc().get("Code Cache"))//
                .scalar("Perm Gen [shared-ro]", lastGcInfo.getMemoryUsageAfterGc().get("Perm Gen [shared-ro]"))//
                .scalar("Perm Gen [shared-rw]", lastGcInfo.getMemoryUsageAfterGc().get("Perm Gen [shared-rw]"));
                gcMap.end();
            }
        }
        list.end();
    }

    /**
     * Dumps information about the Subsystem state
     * 
     * @param sw the {@link SectionWriter} being written to
     */
    public void dumpKernelInformation(SectionWriter sw) {
        ListWriter<SectionWriter> subsystemList = sw.list("Subsystem registry");

        for (Subsystem subsystem : this.subsystemRegistry.getSubsystems()) {
            MapWriter<ListWriter<SectionWriter>> subsystemMap = subsystemList.map();

            if (subsystem.getBundleDescriptors() != null) {
                subsystemMap//
                .scalar("name", subsystem.getName())//
                .scalar("state", subsystem.getState());
                ListWriter<MapWriter<ListWriter<SectionWriter>>> mapList = subsystemMap.list("subsystem bundles");

                for (BundleDescriptor bundleDescriptor : subsystem.getBundleDescriptors()) {
                    mapList.scalar(bundleDescriptor.getSymbolicName());
                }
            }
        }
    }

    /**
     * Retrieves the bundle state
     * 
     * @param b the OSGi bundle
     * @return the OSGi bundle state
     */
    public static String getState(Bundle b) {
        switch (b.getState()) {
            case Bundle.ACTIVE:
                return "ACTIVE";
            case Bundle.INSTALLED:
                return "INSTALLED";
            case Bundle.RESOLVED:
                return "RESOLVED";
            case Bundle.STARTING:
                return "STARTING";
            case Bundle.STOPPING:
                return "";
            case Bundle.UNINSTALLED:
                return "UNINSTALLED";
            default:
                return "UNKNOWN";
        }
    }

    /**
     * {@inheritDoc}
     */
    public void dump(SectionWriter sw) {
        dumpOperatingSystemInfo(sw);
        dumpEquinoxInfo(sw);
        dumpJavaSystemProperties(sw);
        dumpJavaVMInfo(sw);
        dumpSpecInfo(sw);
        dumpGeneralRuntimeInfo(sw);
        dumpClasspathInfo(sw);
        dumpMemoryInfo(sw);
        dumpGarbageCollectionInfo(sw);
        dumpKernelInformation(sw);
    }

    /**
     * {@inheritDoc}
     */
    public String getName() {
        return this.getClass().getCanonicalName().replace(".core.SystemDumpExtension", "");
    }

}
