/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.test.framework;

import java.lang.annotation.Documented;
import java.lang.annotation.ElementType;
import java.lang.annotation.Inherited;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;

/**
 * <p>
 * <code>BundleConfiguration</code> defines class-level metadata which denotes how OSGi bundles should be loaded and
 * configured for testing scenarios.
 * </p>
 * 
 * @author Sam Brannen
 * @since 1.0
 */
@Retention(RetentionPolicy.RUNTIME)//
@Target(ElementType.TYPE)//
@Inherited//
@Documented//
public @interface BundleConfiguration {

    /**
     * The names of OSGi bundles to load.
     */
    String[] names() default {};

    /**
     * <p>
     * Whether or not {@link #names() bundle names} from superclasses should be <em>inherited</em>.
     * </p>
     * <p>
     * The default value is <code>true</code>, which means that an annotated class will <em>inherit</em> the bundle
     * names defined by an annotated superclass. Specifically, the bundle names for an annotated class will be appended
     * to the list of bundle names defined by an annotated superclass. Thus, subclasses have the option of
     * <em>extending</em> the list of bundle names. In the following example, the OSGi bundles for
     * <code>ExtendedTest</code> will be loaded from &quot;com.springsource.server.kernel.core&quot; <strong>and</strong>
     * &quot;com.springsource.server.kernel.logging&quot;, in that order.
     * </p>
     * 
     * <pre class="code">
     * {@link BundleConfiguration @BundleConfiguration}(names={&quot;com.springsource.server.kernel.core&quot;})
     * public class BaseTest {
     *     // ...
     * }
     * {@link BundleConfiguration @BundleConfiguration}(names={&quot;com.springsource.server.kernel.logging&quot;})
     * public class ExtendedTest extends BaseTest {
     *     // ...
     * }
     * </pre>
     * 
     * <p>
     * If <code>inheritNames</code> is set to <code>false</code>, the bundle names for the annotated class will
     * <em>shadow</em> and effectively replace any bundle names defined by a superclass.
     * </p>
     */
    boolean inheritNames() default true;

}
