/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.test.framework;

import java.io.File;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicReference;

import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.BundleException;
import org.osgi.framework.ServiceReference;
import org.osgi.framework.ServiceRegistration;
import org.springframework.osgi.context.event.OsgiBundleApplicationContextListener;
import org.springframework.osgi.extender.support.ApplicationContextConfiguration;
import org.springframework.osgi.extender.support.internal.ConfigUtils;
import org.springframework.osgi.util.OsgiBundleUtils;

import com.springsource.server.osgi.framework.OsgiFramework;
import com.springsource.server.osgi.framework.OsgiFrameworkException;

/**
 * TODO: Document <p/> <strong>Concurrent Semantics</strong><br/> TODO: Document concurrent semantics of
 * 
 * @author Rob Harrop
 * @author Glyn Normington
 * @since 1.0
 */
public class OsgiBootstrap {

    private static final String DEPLOYER_CORE_SYMBOLIC_NAME = "com.springsource.server.deployer.core";

    private final OsgiFramework osgi;

    private final String testClassName;

    private Bundle testFrameworkBundle;

    private Bundle testBundle;

    private ServiceRegistration listenerRegistration;

    /**
     * @param osgi
     */
    public OsgiBootstrap(OsgiFramework osgi, String testClassName) {
        this.osgi = osgi;
        this.testClassName = testClassName;
    }

    public Class<?> installTestBundles(String... additionalBundles) throws Exception {
        this.osgi.getBundleInstaller().installBundle("com.springsource.org.junit");
        this.osgi.getBundleInstaller().installBundle("com.springsource.server.osgi.framework");
        this.testFrameworkBundle = this.osgi.getBundleInstaller().installBundle("com.springsource.server.test.framework");

        AtomicReference<String> failureRef = registerListener();

        // Ensure that the subsystems defined by the profile manager have been installed. The deployer ensures this.
        waitForBundleApplicationContext(DEPLOYER_CORE_SYMBOLIC_NAME, failureRef);

        startBundles();
        waitForSpringOsgiBundles(failureRef);

        if (failureRef.get() != null) {
            throw new RuntimeException(failureRef.get());
        }

        ServiceReference deployerReference = this.osgi.getBundleContext().getServiceReference(
            "com.springsource.server.deployer.core.ApplicationDeployer");
        Object deployer = this.osgi.getBundleContext().getService(deployerReference);

        boolean recoveryComplete = (Boolean) deployer.getClass().getMethod("awaitRecovery", long.class, TimeUnit.class).invoke(deployer, 30,
            TimeUnit.SECONDS);

        if (!recoveryComplete) {
            throw new Exception("Deployer recovery did not complete within timeout period.");
        }

        this.osgi.getBundleContext().ungetService(deployerReference);

        for (String additionalBundle : additionalBundles) {
            this.osgi.getBundleInstaller().installBundle(additionalBundle);
        }

        this.testBundle = installTestBundle();

        startBundles();

        return this.testBundle.loadClass(this.testClassName);
    }

    /**
     * Post-condition: all started Spring DM modules that are going to have published their application contexts.
     * 
     * @param failureRef
     * 
     * @throws InterruptedException
     */
    private void waitForSpringOsgiBundles(AtomicReference<String> failureRef) throws InterruptedException {
        Bundle[] bundles = this.osgi.getBundleContext().getBundles();
        for (Bundle bundle : bundles) {
            if (OsgiBundleUtils.isBundleActive(bundle) && new ApplicationContextConfiguration(bundle).isSpringPoweredBundle()
                && ConfigUtils.getPublishContext(bundle.getHeaders()) && !bundle.getSymbolicName().equals(DEPLOYER_CORE_SYMBOLIC_NAME)) {
                waitForBundleApplicationContext(bundle.getSymbolicName(), failureRef);
            }
        }
    }

    private void waitForBundleApplicationContext(String symbolicName, AtomicReference<String> failureRef) throws InterruptedException {
        int waitCount = 5;
        int count = 0;
        while (count++ < waitCount) {
            if (failureRef.get() != null) {
                return;
            }
            if (this.osgi.waitForBundleApplicationContext(symbolicName, 30)) {
                return;
            }
        }
        System.out.println("OsgiBootstrap.waitForBundleApplicationContext: Bundle '" + symbolicName
            + "' failed to publish its application context. Terminating...");
        System.exit(1);
    }

    private void startBundles() throws Exception {
        Bundle[] bundles = this.osgi.getBundleContext().getBundles();
        for (Bundle bundle : bundles) {
            if (!isActive(bundle) && !isFragment(bundle)) {
                try {
                    bundle.start();
                } catch (Exception ex) {
                    throw new Exception("Failed to start bundle: " + bundle, ex);
                }
            }
        }
    }

    private boolean isFragment(Bundle bundle) {
        String value = (String) bundle.getHeaders().get("Fragment-Host");
        return value != null && value.length() > 0;
    }

    public void stop() {
        if (this.listenerRegistration != null) {
            this.listenerRegistration.unregister();
        }
        this.osgi.stop();
    }

    private boolean isActive(Bundle bundle) {
        return bundle.getState() == Bundle.ACTIVE;
    }

    /**
     * Installs the test bundle from the default location.
     * 
     * Note: this is a special case resolution for tests only. For the purpose of removing circles some test packages
     * need to sit outside their logical subsystems. This approach allows for explicit resolution of the output
     * directory for the executing project.
     */
    private Bundle installTestBundle() throws Exception {
        File file = new File("./target/classes");
        Bundle bundle = null;
        try {
            bundle = this.osgi.getBundleInstaller().installBundleFromLocation(file);
        } catch (OsgiFrameworkException ofe) {
            ofe.printStackTrace();
            throw new BundleException("Test bundle not found in [" + file + "] check output folder definition.", ofe);
        }
        return bundle;
    }

    /**
     * @return
     */
    public Bundle getTestFrameworkBundle() {
        return this.testFrameworkBundle;
    }

    /**
     * @return
     */
    public Bundle getTestBundle() {
        return this.testBundle;
    }

    /**
     * @return
     */
    public OsgiFramework getOsgiFramework() {
        return this.osgi;
    }

    private AtomicReference<String> registerListener() throws Exception {
        AtomicReference<String> failureRef = new AtomicReference<String>();

        Class<?> listenerClass = this.testFrameworkBundle.loadClass(Listener.class.getName());
        Object listener = listenerClass.getDeclaredConstructor(AtomicReference.class).newInstance(failureRef);
        BundleContext bundleContext = this.osgi.getBundleContext();
        this.listenerRegistration = bundleContext.registerService(OsgiBundleApplicationContextListener.class.getName(), listener, null);
        return failureRef;
    }
}
