/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.test.framework.junit4;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import org.junit.Assume.AssumptionViolatedException;
import org.junit.internal.runners.MethodRoadie;
import org.junit.runner.Description;
import org.junit.runner.notification.Failure;
import org.junit.runner.notification.RunNotifier;
import org.osgi.framework.Bundle;

/**
 * Custom {@link MethodRoadie} for use in an OSGi integration testing environment.
 * 
 * @author Sam Brannen
 * @author Rob Harrop
 * @since 1.0
 */
public class EngineMethodRoadie {

    private final RunNotifier notifier;

    private final EngineTestMethod testMethod;

    private final Description description;

    private final Object testInstance;

    private Throwable testException;

    private final Class<? extends Exception> assumptionViolatedExceptionClass;

    /**
     * @param notifier
     * @param testMethod
     * @param description
     * @param testInstance
     * @param testException
     */
    @SuppressWarnings("unchecked")
    public EngineMethodRoadie(RunNotifier notifier, EngineTestMethod testMethod, Description description, Object testInstance,
        Bundle testFrameworkBundle) {
        this.notifier = notifier;
        this.testMethod = testMethod;
        this.description = description;
        this.testInstance = testInstance;

        try {
            this.assumptionViolatedExceptionClass = testFrameworkBundle.loadClass(AssumptionViolatedException.class.getName());
        } catch (ClassNotFoundException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * Runs the <em>test</em>, including notification of events to the {@link RunNotifier} and
     * {@link TestContextManager} as well as proper handling of {@link org.junit.Ignore @Ignore},
     * {@link org.junit.Test#expected() expected exceptions}, {@link org.junit.Test#timeout() test timeouts}, and
     * {@link org.junit.Assume.AssumptionViolatedException assumptions}.
     */
    public void run() {

        if (getTestMethod().isIgnored()) {
            getNotifier().fireTestIgnored(getDescription());
            return;
        }
        getNotifier().fireTestStarted(getDescription());
        try {
            final long junitTimeout = getTestMethod().getTimeout();

            if (junitTimeout > 0) {
                runWithTimeout(junitTimeout);
            } else {
                runBeforesThenTestThenAfters();
            }
        } finally {
            getNotifier().fireTestFinished(getDescription());
        }
    }

    /**
     * <p>
     * Runs the following methods on the test instance, guaranteeing that {@link #runAfters() @After methods} will have
     * a chance to execute:
     * </p>
     * <ul>
     * <li>{@link #runBefores() @Before methods}</li>
     * <li>The supplied, {@link Runnable} <code>test</code></li>
     * <li>{@link #runAfters() @After methods}</li>
     * </ul>
     * 
     * @param test the runnable test.
     */
    protected void runBeforesThenTestThenAfters() {
        try {
            runBefores();
            runTestMethod();
        } catch (final FailedBefore e) {
        } catch (final Exception e) {
            throw new RuntimeException("test should never throw an exception to this level");
        } finally {
            runAfters();
        }
    }

    /**
     * Calls {@link TestContextManager#beforeTestMethod(Object,Method)} and then runs
     * {@link org.junit.Before @Before methods}, registering failures and throwing {@link FailedBefore} exceptions as
     * necessary.
     * 
     * @throws FailedBefore If an error occurs while executing a <em>before</em> method.
     */
    protected void runBefores() throws FailedBefore {
        try {
            final List<Method> befores = getTestMethod().getBefores();
            for (final Method before : befores) {
                before.invoke(getTestInstance());
            }
        } catch (final InvocationTargetException e) {
            addFailure(e.getTargetException());
            throw new FailedBefore();
        } catch (final Throwable t) {
            addFailure(t);
            throw new FailedBefore();
        }
    }

    /**
     * Runs {@link org.junit.After @After methods}, registering failures as necessary, and then calls
     * {@link TestContextManager#afterTestMethod(Object,Method,Throwable)}.
     */
    protected void runAfters() {
        final List<Method> afters = getTestMethod().getAfters();
        for (final Method after : afters) {
            try {
                after.invoke(getTestInstance());
            } catch (final InvocationTargetException e) {
                addFailure(e.getTargetException());
            } catch (final Throwable t) {
                addFailure(t); // Untested, but seems impossible
            }
        }

    }

    /**
     * Runs the test method on the test instance with the specified <code>timeout</code>.
     * 
     * @param timeout the timeout in milliseconds.
     * @see #runWithRepetitions(Runnable)
     * @see #runTestMethod()
     */
    protected void runWithTimeout(final long timeout) {

        final ExecutorService service = Executors.newSingleThreadExecutor();
        final Callable<Object> callable = new Callable<Object>() {

            public Object call() throws Exception {

                runBeforesThenTestThenAfters();
                return null;
            }
        };
        final Future<Object> result = service.submit(callable);
        service.shutdown();
        try {
            boolean terminated = service.awaitTermination(timeout, TimeUnit.MILLISECONDS);
            if (!terminated) {
                service.shutdownNow();
            }
            // throws the exception if one occurred during the invocation
            result.get(0, TimeUnit.MILLISECONDS);
        } catch (final TimeoutException e) {
            addFailure(new Exception(String.format("test timed out after %d milliseconds", timeout)));
        } catch (final Exception e) {
            addFailure(e);
        }

    }

    /**
     * Runs the test method on the test instance, processing exceptions (both expected and unexpected), assumptions, and
     * registering failures as necessary.
     */
    protected void runTestMethod() {
        this.testException = null;
        try {
            getTestMethod().invoke(getTestInstance());
            if (getTestMethod().expectsException()) {
                addFailure(new AssertionError("Expected exception: " + getTestMethod().getExpectedException().getName()));
            }
        } catch (InvocationTargetException e) {
            this.testException = e.getTargetException();

            // if (this.testException instanceof AssumptionViolatedException) {
            if (this.assumptionViolatedExceptionClass.isInstance(this.testException)) {
                return;
            } else if (!getTestMethod().expectsException()) {
                addFailure(this.testException);
            } else if (getTestMethod().isUnexpected(this.testException)) {
                final String message = "Unexpected exception, expected<" + getTestMethod().getExpectedException().getName() + "> but was<"
                    + this.testException.getClass().getName() + ">";
                addFailure(new Exception(message, this.testException));
            }
        } catch (final Throwable t) {
            addFailure(t);
        } finally {
            if (this.testException != null) {
                System.out.println("Test method [" + getTestMethod().getMethod() + "] threw exception [" + this.testException + "].");
            }

        }
    }

    /**
     * Fires a failure for the supplied <code>exception</code> with the {@link #getNotifier() RunNotifier}.
     * 
     * @param exception the exception upon which to base the failure.
     */
    protected void addFailure(final Throwable exception) {
        getNotifier().fireTestFailure(new Failure(getDescription(), exception));
    }

    /**
     * Gets the test instance.
     * 
     * @return the test instance.
     */
    protected final Object getTestInstance() {
        return this.testInstance;
    }

    protected RunNotifier getNotifier() {
        return this.notifier;
    }

    protected EngineTestMethod getTestMethod() {
        return this.testMethod;
    }

    protected Description getDescription() {
        return this.description;
    }

    /**
     * Marker exception to signal that an exception was encountered while executing an {@link org.junit.Before @Before}
     * method.
     */
    @SuppressWarnings("serial")
    private static class FailedBefore extends Exception {
    }
}
