/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.test.framework.junit4;

import java.lang.annotation.Annotation;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.junit.After;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import org.junit.internal.runners.TestClass;
import org.osgi.framework.Bundle;

/**
 * Custom {@link TestClass} for use in an OSGi integration testing environment.
 * 
 * @author Rob Harrop
 * @author Sam Brannen
 * @since 1.0
 */
public class EngineTestClass {

    private final Class<?> testClass;

    private final Class<Annotation> testAnnotationClass;

    private final Class<Annotation> beforeClassAnnotationClass;

    private final Class<Annotation> afterClassAnnotationClass;

    private final Class<Annotation> beforeAnnotationClass;

    private final Class<Annotation> afterAnnotationClass;

    @SuppressWarnings("unchecked")
    public EngineTestClass(Class<?> klass, Bundle testFrameworkBundle) throws Exception {
        this.testClass = klass;

        this.testAnnotationClass = testFrameworkBundle.loadClass(Test.class.getName());
        this.beforeClassAnnotationClass = testFrameworkBundle.loadClass(BeforeClass.class.getName());
        this.afterClassAnnotationClass = testFrameworkBundle.loadClass(AfterClass.class.getName());
        this.beforeAnnotationClass = testFrameworkBundle.loadClass(Before.class.getName());
        this.afterAnnotationClass = testFrameworkBundle.loadClass(After.class.getName());
    }

    List<Method> getTestMethods() {
        return getAnnotatedMethods(this.testAnnotationClass);
    }

    List<Method> getBefores() {
        return getAnnotatedMethods(this.beforeClassAnnotationClass);
    }

    List<Method> getAfters() {
        return getAnnotatedMethods(this.afterClassAnnotationClass);
    }

    List<Method> getMethodBefores() {
        return getAnnotatedMethods(this.beforeAnnotationClass);
    }

    List<Method> getMethodAfters() {
        return getAnnotatedMethods(this.afterAnnotationClass);
    }

    public List<Method> getAnnotatedMethods(Class<? extends Annotation> annotationClass) {
        List<Method> results = new ArrayList<Method>();
        for (Class<?> eachClass : getSuperClasses(testClass)) {
            Method[] methods = eachClass.getDeclaredMethods();
            for (Method eachMethod : methods) {
                Annotation annotation = eachMethod.getAnnotation(annotationClass);
                if (annotation != null && !isShadowed(eachMethod, results))
                    results.add(eachMethod);
            }
        }
        if (runsTopToBottom(annotationClass))
            Collections.reverse(results);
        return results;
    }

    private boolean runsTopToBottom(Class<? extends Annotation> annotation) {
        return annotation.equals(this.beforeAnnotationClass) || annotation.equals(this.beforeClassAnnotationClass);
    }

    private boolean isShadowed(Method method, List<Method> results) {
        for (Method each : results) {
            if (isShadowed(method, each))
                return true;
        }
        return false;
    }

    private boolean isShadowed(Method current, Method previous) {
        if (!previous.getName().equals(current.getName()))
            return false;
        if (previous.getParameterTypes().length != current.getParameterTypes().length)
            return false;
        for (int i = 0; i < previous.getParameterTypes().length; i++) {
            if (!previous.getParameterTypes()[i].equals(current.getParameterTypes()[i]))
                return false;
        }
        return true;
    }

    private List<Class<?>> getSuperClasses(Class<?> testClass) {
        ArrayList<Class<?>> results = new ArrayList<Class<?>>();
        Class<?> current = testClass;
        while (current != null) {
            results.add(current);
            current = current.getSuperclass();
        }
        return results;
    }

    public Constructor<?> getConstructor() throws SecurityException, NoSuchMethodException {
        return testClass.getConstructor();
    }

    public Class<?> getJavaClass() {
        return testClass;
    }

    public String getName() {
        return testClass.getName();
    }
}
