/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.test.framework.junit4;

import java.lang.annotation.Annotation;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.List;

import org.junit.After;
import org.junit.Before;
import org.junit.Ignore;
import org.junit.Test;
import org.junit.Test.None;
import org.junit.internal.runners.TestClass;
import org.junit.internal.runners.TestMethod;
import org.osgi.framework.Bundle;
import org.springframework.core.annotation.AnnotationUtils;

/**
 * Custom {@link TestMethod} for use in an OSGi integration testing environment.
 * 
 * @author Sam Brannen
 * @author Rob Harrop
 * @since 1.0
 */
public class EngineTestMethod {

    private final Method method;

    private final EngineTestClass testClass;

    private final Class<Annotation> testAnnotationClass;

    private final Class<Annotation> ignoreAnnotationClass;

    private final Class<None> noneClass;

    /**
     * Constructs a test method for the supplied {@link Method method} and {@link TestClass test class}.
     * 
     * @param method The test method.
     * @param testClass The test class.
     */
    @SuppressWarnings("unchecked")
    public EngineTestMethod(final Method method, final EngineTestClass testClass, Bundle testFrameworkBundle) {
        this.method = method;
        this.testClass = testClass;
        try {
            this.testAnnotationClass = testFrameworkBundle.loadClass(Test.class.getName());
            this.ignoreAnnotationClass = testFrameworkBundle.loadClass(Ignore.class.getName());
            this.noneClass = testFrameworkBundle.loadClass(None.class.getName());
        } catch (ClassNotFoundException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * Determines if this test method is {@link Test#expected() expected} to throw an exception.
     * 
     * @return <code>true</code> if this test method should throw an exception.
     */
    public boolean expectsException() {
        return getExpectedException() != null;
    }

    /**
     * Gets the {@link After @After} methods for this test method.
     * 
     * @return The <em>after</em> methods.
     */
    public List<Method> getAfters() {
        return getTestClass().getMethodAfters();
    }

    /**
     * Gets the {@link Before @Before} methods for this test method.
     * 
     * @return The <em>before</em> methods.
     */
    public List<Method> getBefores() {
        return getTestClass().getMethodBefores();
    }

    /**
     * <p>
     * Gets the <code>exception</code> that this test method is expected to throw.
     * </p>
     * <p>
     * Supports both Spring's {@link ExpectedException @ExpectedException(...)} and JUnit's
     * {@link Test#expected() @Test(expected=...)} annotations, but not both simultaneously.
     * </p>
     * 
     * @return The expected exception, or <code>null</code> if none was specified.
     * @throws IllegalStateException if both types of configuration are used simultaneously.
     */
    @SuppressWarnings("unchecked")
    public Class<? extends Throwable> getExpectedException() throws IllegalStateException {

        final Annotation testAnnotation = getMethod().getAnnotation(this.testAnnotationClass);
        final Class<? extends Throwable> expectedAttribute = (Class<? extends Throwable>) AnnotationUtils.getValue(testAnnotation, "expected");

        Class<? extends Throwable> expectedException = testAnnotation != null && expectedAttribute != this.noneClass ? expectedAttribute : null;
        return expectedException;
    }

    /**
     * Gets the actual {@link Method method} referenced by this test method.
     * 
     * @return The test method.
     */
    public final Method getMethod() {
        return this.method;
    }

    /**
     * Gets the {@link EngineTestClass test class} for this test method.
     * 
     * @return The test class.
     */
    public final EngineTestClass getTestClass() {
        return this.testClass;
    }

    /**
     * <p>
     * Gets the configured <code>timeout</code> for this test method.
     * </p>
     * <p>
     * Supports JUnit's {@link Test#timeout() @Test(timeout=...)} annotation.
     * </p>
     * 
     * @return The timeout, or <code>0</code> if none was specified.
     */
    public long getTimeout() {
        final Annotation testAnnotation = getMethod().getAnnotation(this.testAnnotationClass);
        final Long timeoutAttribute = (Long) AnnotationUtils.getValue(testAnnotation, "timeout");
        final long timeout = testAnnotation != null && timeoutAttribute > 0 ? timeoutAttribute : 0;
        return timeout;
    }

    /**
     * Convenience method for {@link Method#invoke(Object,Object...) invoking} the method associated with this test
     * method. Throws exceptions consistent with {@link Method#invoke(Object,Object...) Method.invoke()}.
     * 
     * @param testInstance The test instance upon which to invoke the method.
     */
    public void invoke(final Object testInstance) throws IllegalArgumentException, IllegalAccessException, InvocationTargetException {
        getMethod().invoke(testInstance);
    }

    /**
     * Determines if this test method should be ignored.
     * 
     * @return <code>true</code> if this test method should be ignored.
     * @see ProfileValueUtils#isTestEnabledInThisEnvironment(ProfileValueSource, Method)
     */
    public boolean isIgnored() {
        return getMethod().isAnnotationPresent(this.ignoreAnnotationClass);
    }

    /**
     * Determines if this test method {@link Test#expected() expects} exceptions of the type of the supplied
     * <code>exception</code> to be thrown.
     * 
     * @param exception The thrown exception.
     * @return <code>true</code> if the supplied exception was of an expected type.
     */
    public boolean isUnexpected(final Throwable exception) {
        return !getExpectedException().isAssignableFrom(exception.getClass());
    }
}
