/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.test.framework.junit4;

import java.io.File;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;

import org.junit.Test;
import org.junit.internal.runners.InitializationError;
import org.junit.internal.runners.JUnit4ClassRunner;
import org.junit.runner.Description;
import org.junit.runner.notification.RunNotifier;
import org.osgi.framework.BundleContext;
import org.osgi.framework.BundleException;

import com.springsource.server.osgi.framework.OsgiFramework;
import com.springsource.server.osgi.framework.OsgiLauncher;
import com.springsource.server.osgi.framework.equinox.EquinoxOsgiConfiguration;
import com.springsource.server.osgi.framework.equinox.EquinoxOsgiLauncher;
import com.springsource.server.serviceability.output.internal.ServiceabilityOutputBootstrap;
import com.springsource.server.serviceability.tracing.internal.TracingBootstrap;
import com.springsource.server.test.framework.BundleConfiguration;
import com.springsource.server.test.framework.BundleInstaller;
import com.springsource.server.test.framework.BundleRepositorySearchPaths;
import com.springsource.server.test.framework.OsgiBootstrap;

/**
 * <p>
 * OsgiIntegrationTestJUnit4ClassRunner is a custom implementation of {@link JUnit4ClassRunner} which provides
 * functionality to standard JUnit 4.4+ tests for loading, testing, and interacting with OSGi bundles.
 * </p>
 * <u>Supported Annotations</u>:
 * <ul>
 * <li>{@link BundleConfiguration @BundleConfiguration}</li>
 * <li>{@link com.springsource.server.test.framework.OsgiFramework @OsgiFramework}</li>
 * <li>{@link BundleInstaller @BundleInstaller}</li>
 * <li>{@link BundleRepositorySearchPaths @BundleRepositorySearchPaths}</li>
 * </ul>
 * 
 * @author Sam Brannen
 * @author Rob Harrop
 * @since 1.0
 */
class OsgiIntegrationTestJUnit4ClassRunner extends JUnit4ClassRunner {

    private OsgiFramework osgi;

    protected OsgiBootstrap bootstrap;

    private String[] additionalBundleLocations;

    private EngineTestClass testClass;

    /**
     * Constructs a new <code>OsgiIntegrationTestJUnit4ClassRunner</code>.
     * 
     * @param clazz the Class object corresponding to the test class to be run.
     * @throws BundleException
     */
    public OsgiIntegrationTestJUnit4ClassRunner(final Class<?> clazz) throws InitializationError {
        super(clazz);
    }

    /**
     * @param testClass
     * @return
     */
    private String[] readBundleConfiguration(Class<?> testClass) {
        BundleConfiguration bc = testClass.getAnnotation(BundleConfiguration.class);
        String[] value = (bc != null ? bc.names() : null);
        return (value != null ? value : new String[0]);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void run(RunNotifier notifier) {
        try {
            try {
                this.osgi = createOsgiFramework();
                BundleContext bundleContext = this.osgi.getBundleContext();
                ServiceabilityOutputBootstrap.publishServices(bundleContext);
                TracingBootstrap.publishServices(bundleContext);
                this.bootstrap = new OsgiBootstrap(osgi, getTestClass().getName());
                this.additionalBundleLocations = readBundleConfiguration(getTestClass().getJavaClass());
                preconfigureEnvironment(getTestClass().getJavaClass());
                preInstallTestBundles();
                Class<?> testClass = this.bootstrap.installTestBundles(this.additionalBundleLocations);
                validateConfiguration();
                this.testClass = new EngineTestClass(testClass, this.bootstrap.getTestFrameworkBundle());
            } catch (InitializationError e) {
                System.out.println("Unable to start JUnit Test because:");
                for (Throwable t : e.getCauses()) {
                    System.out.println("\t* " + t.getMessage());
                }
                notifier.pleaseStop();
                return;
            } catch (Exception e) {
                notifier.testAborted(getDescription(), e);
                notifier.pleaseStop();
            }
            if (this.testClass != null) {
                doRun(notifier);
            }
            if (this.bootstrap != null) {
                this.bootstrap.stop();
            }
        } finally {
            ServiceabilityOutputBootstrap.retractServices();
            TracingBootstrap.retractServices();
        }
    }

    /**
     * Preconfigures the run time environment
     * 
     * @param testClass the test class to preconfigure
     */
    protected void preconfigureEnvironment(Class<?> testClass) {
    }

    /**
     * Convenience method for logging and then setting a system property.
     * 
     * @see System#setProperty(String, String)
     */
    protected void setSystemProperty(String key, String value) {
        System.out.println("OsgiIntegrationTestJUnit4ClassRunner: setting system property [" + key + "] to [" + value + "].");
        System.setProperty(key, value);
    }

    protected void preInstallTestBundles() throws Exception {
        /* no-op */
    }

    /**
     * Performs basic validation of the test environment set up
     * 
     * @throws InitializationError
     */
    private void validateConfiguration() throws InitializationError {
        List<Throwable> errors = new ArrayList<Throwable>();

        try {
            this.bootstrap.getTestBundle().loadClass(Test.class.getName());
        } catch (ClassNotFoundException e) {
            errors.add(new RuntimeException("Unable to load JUnit @Test class in OSGi. "
                + "Does your test bundle have org.junit in the import packages?"));
        }

        try {
            this.bootstrap.getTestBundle().loadClass(BundleConfiguration.class.getName());
        } catch (ClassNotFoundException e) {
            errors.add(new RuntimeException("Unable to access test framework classes. "
                + "Does your test bundle have com.springsource.server.test.framework in the import packages?"));
        }
        if (!errors.isEmpty()) {
            throw new InitializationError(errors);
        }
    }

    /**
     * @param notifier
     */
    private void doRun(final RunNotifier notifier) {
        new EngineClassRoadie(notifier, this.testClass, getDescription(), new Runnable() {

            public void run() {
                runMethods(notifier);
            }

        }).runProtected();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void runMethods(RunNotifier notifier) {
        List<Method> methods = this.testClass.getTestMethods();
        for (Method method : methods) {
            invokeTestMethod(method, notifier);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void invokeTestMethod(Method method, RunNotifier notifier) {
        Description description = Description.createTestDescription(method.getDeclaringClass(), method.getName());

        Object testInstance;
        try {
            testInstance = createTest();
        } catch (InvocationTargetException ex) {
            notifier.testAborted(description, ex.getCause());
            return;
        } catch (Exception ex) {
            notifier.testAborted(description, ex);
            return;
        }

        EngineTestMethod testMethod = new EngineTestMethod(method, this.testClass, this.bootstrap.getTestFrameworkBundle());
        new EngineMethodRoadie(notifier, testMethod, description, testInstance, this.bootstrap.getTestFrameworkBundle()).run();
    }

    protected String[] getSearchPaths() {

        BundleRepositorySearchPaths searchPaths = getTestClass().getJavaClass().getAnnotation(BundleRepositorySearchPaths.class);
        if (searchPaths != null && searchPaths.value() != null) {
            return searchPaths.value();
        } else {
            return new String[] { "{subsystem}/{bundle}/target/classes",
                "${ivy.cache::ivy-cache}/repository/{org}/{name}/{version}/{bundle}.jar",
                "${ivy.cache::ivy-cache}/repository/{org}/{name}/{version}/{library}.libd" };
        }
    }

    /**
     * @return
     */
    private OsgiFramework createOsgiFramework() {
        String[] searchPath = getSearchPaths();

        EquinoxOsgiConfiguration config = new EquinoxOsgiConfiguration("./target/", "./target/install", new File("../../").getAbsolutePath(),
            searchPath);
        config.setClean(true);
        config.setRunConsole(true);
        config.setConsolePort(2401);
        config.setConfigAreaPath("target/equinox-config");
        OsgiLauncher launcher = new EquinoxOsgiLauncher(config);
        return launcher.start();
    }

    /**
     * TODO Document createTest().
     * 
     * @return a new test instance
     * @see JUnit4ClassRunner#createTest()
     */
    @Override
    protected Object createTest() throws Exception {
        Class<?> testClass = this.testClass.getJavaClass();
        Object testInstance = testClass.newInstance();
        injectFields(testInstance);
        return testInstance;
    }

    /**
     * @param testInstance
     */
    private void injectFields(Object testInstance) throws Exception {
        InjectionHandler injectionHandler = new InjectionHandler(this.bootstrap.getTestFrameworkBundle());
        injectionHandler.registerInjectionValue(com.springsource.server.test.framework.OsgiFramework.class, this.osgi);
        injectionHandler.registerInjectionValue(BundleInstaller.class, this.osgi.getBundleInstaller());
        registerAdditionalInjectionValues(injectionHandler);
        injectionHandler.injectFieldValues(testInstance);
    }

    /**
     * @param injectionHandler
     * @throws Exception
     */
    protected void registerAdditionalInjectionValues(InjectionHandler injectionHandler) throws Exception {
        /* no-op */
    }
}
