/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.test.framework.junit4;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import org.junit.internal.runners.InitializationError;
import org.osgi.framework.BundleContext;
import org.osgi.framework.ServiceReference;

import com.springsource.server.bootstrap.BootstrapConfiguration;
import com.springsource.server.kernel.bootstrap.KernelBootstrap;
import com.springsource.server.kernel.core.SubsystemRegistry;
import com.springsource.server.serviceability.logging.internal.LoggingBootstrap;
import com.springsource.server.serviceability.output.DestinationRegistry;
import com.springsource.server.serviceability.output.EntryPublisherFactory;
import com.springsource.server.serviceability.output.internal.ServiceabilityOutputBootstrap;
import com.springsource.server.serviceability.tracing.internal.TracingBootstrap;

/**
 * <p>
 * <code>ServerOsgiIntegrationTestJUnit4ClassRunner</code> is a custom extension of
 * {@link OsgiIntegrationTestJUnit4ClassRunner} which provides additional integration testing functionality specific to
 * the <em>SpringSource dm Server (AP)</em>.
 * </p>
 * <u>Supported Annotations</u>:
 * <ul>
 * <li>All annotations supported by {@link OsgiIntegrationTestJUnit4ClassRunner} plus... </li>
 * <li>{@link SubsystemRegistry @SubsystemRegistry}</li>
 * </ul>
 * 
 * @author Rob Harrop
 * @author Sam Brannen
 * @since 1.0
 */
public class ServerOsgiIntegrationTestJUnit4ClassRunner extends OsgiIntegrationTestJUnit4ClassRunner {

    private final BootstrapConfiguration bootstrapConfiguration;

    /**
     * @param clazz
     * @throws InitializationError
     */
    public ServerOsgiIntegrationTestJUnit4ClassRunner(Class<?> clazz) throws InitializationError {
        super(clazz);
        this.bootstrapConfiguration = new IntegrationTestBootstrapConfiguration();
        ServiceabilityOutputBootstrap.start();
        DestinationRegistry destinationRegistry = ServiceabilityOutputBootstrap.getDestinationRegistry();
        EntryPublisherFactory entryPublisherFactory = ServiceabilityOutputBootstrap.getEntryPublisherFactory();       
        TracingBootstrap.start(ServiceabilityOutputBootstrap.getDestinationRegistry(), ServiceabilityOutputBootstrap.getEntryFilterFactory(),
            ServiceabilityOutputBootstrap.getEntryPublisherFactory(), bootstrapConfiguration.getTraceLevels(), bootstrapConfiguration.getTraceDirectory());
        LoggingBootstrap.start(bootstrapConfiguration.getLogsDirectory(), destinationRegistry, entryPublisherFactory);
    }

    /**
     * Delegates to {@link OsgiIntegrationTestJUnit4ClassRunner#preconfigureEnvironment(Class)}; sets the
     * <code>java.io.tmpdir</code>.
     */
    @Override
    protected void preconfigureEnvironment(Class<?> testClass) {
        super.preconfigureEnvironment(testClass);
        String targetDir = new File("target").getAbsolutePath();
        setSystemProperty("java.io.tmpdir", targetDir);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void preInstallTestBundles() throws Exception {
        new KernelBootstrap(this.bootstrap.getOsgiFramework(), new IntegrationTestBootstrapConfiguration()).start();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void registerAdditionalInjectionValues(InjectionHandler injectionHandler) throws Exception {
        BundleContext bundleContext = this.bootstrap.getTestBundle().getBundleContext();
        ServiceReference serviceReference = bundleContext.getServiceReference(SubsystemRegistry.class.getName());
        Object subsystemRegistry = bundleContext.getService(serviceReference);
        injectionHandler.registerInjectionValue(com.springsource.server.test.framework.SubsystemRegistry.class, subsystemRegistry);
    }

    private class IntegrationTestBootstrapConfiguration implements BootstrapConfiguration {

        private final File targetDir = new File("target");

        /**
         * {@inheritDoc}
         */
        public String[] getConfigPaths() {
            File currentDir = new File(".");
            File integrationTestConfigDir = new File("../../com.springsource.server.test/integration-test-config");
            try {
                return new String[] {currentDir.getCanonicalPath(), integrationTestConfigDir.getCanonicalPath()};
            } catch (IOException ioe) {
                throw new RuntimeException(ioe);
            }
        }

        /**
         * {@inheritDoc}
         */
        public String getDumpDirectory() {
            return targetDir.getAbsolutePath();
        }

        /**
         * {@inheritDoc}
         */
        public String getHomeDirectory() {
            return targetDir.getAbsolutePath();
        }

        /**
         * {@inheritDoc}
         */
        public String getLogsDirectory() {
            return targetDir.getAbsolutePath();
        }

        /**
         * {@inheritDoc}
         */
        public String getPickUpDirectory() {
            return new File(targetDir, "pickup").getAbsolutePath();
        }

        /**
         * {@inheritDoc}
         */
        public Map<String, String> getTraceLevels() {
            Map<String, String> levels = new HashMap<String, String>();
            levels.put("*", "info");
            return levels;
        }

        /**
         * {@inheritDoc}
         */
        public String getSubsystemsDirectory() {
            return null;
        }

        /**
         * {@inheritDoc}
         */
        public String getTraceDirectory() {
            return targetDir.getAbsolutePath();
        }

        /**
         * {@inheritDoc}
         */
        public String getWorkDirectory() {
            return targetDir.getAbsolutePath();
        }
    }
}
