/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.web.core;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.jar.JarEntry;
import java.util.jar.JarInputStream;

import org.osgi.framework.BundleContext;
import org.osgi.framework.ServiceReference;
import org.osgi.service.packageadmin.ExportedPackage;
import org.osgi.service.packageadmin.PackageAdmin;

import com.springsource.server.deployer.core.DeploymentContext;
import com.springsource.server.deployer.core.DeploymentException;
import com.springsource.server.deployer.core.FatalDeploymentException;
import com.springsource.server.deployer.core.Module;
import com.springsource.server.deployer.core.DeploymentArtefact.Entry;
import com.springsource.server.kernel.core.BundleStarter;
import com.springsource.server.osgi.framework.OsgiFramework;
import com.springsource.server.osgi.manifest.BundleManifestFactory;
import com.springsource.server.osgi.manifest.BundleManifest.ImportPackageHeader.PackageImport;
import com.springsource.server.serviceability.ffdc.NonNull;
import com.springsource.server.serviceability.output.Level;
import com.springsource.server.serviceability.tracing.EntryExitTraceLevel;
import com.springsource.server.servlet.core.DeployedWebApplicationNamesRegistry;
import com.springsource.server.servlet.core.ServletContainer;

/**
 * <code>WarModulePersonality</code> provides deployment of free-standing '.war' files.
 * 
 * <p/>
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * This class is thread safe providing the supplied {@link ServletContainer} is thread safe.
 * 
 * @author Glyn Normington
 * @author Sam Brannen
 * @since 1.0
 */
public class WarModulePersonality extends WebModulePersonality {

    public static final String PERSONALITY = "WAR";

    private static final String WAR_EXTENSION = ".war";

    private static final String WEB_INF_CLASSES_PATH = "/WEB-INF/classes/";

    private static final String WEB_INF_LIB_PATH = "/WEB-INF/lib/";

    private final OsgiFramework osgi;

    public WarModulePersonality(@NonNull OsgiFramework osgi, @NonNull BundleManifestFactory bundleManifestFactory,
        @NonNull ServletContainer servletContainer, @NonNull Module module,
        @NonNull DeployedWebApplicationNamesRegistry deployedWebApplicationNamesRegistry, @NonNull DeploymentContext deploymentContext,
        @NonNull BundleStarter bundleStarter) {
        super(bundleManifestFactory, servletContainer, module, deployedWebApplicationNamesRegistry, deploymentContext, bundleStarter);
        this.osgi = osgi;
    }

    /**
     * @return the default context path based on the name of the WAR file.
     */
    @Override
    protected String getDefaultContextPath() {
        String warName = getModuleData().getFilename();
        return warName.substring(0, warName.length() - WAR_EXTENSION.length());
    }

    /**
     * Always returns &quot;{@link #PERSONALITY WAR}&quot;.
     */
    @Override
    public String getPersonality() {
        return PERSONALITY;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void transform() throws DeploymentException {
        transformBundleManifest();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void addPersonalitySpecificPackageImports(List<PackageImport> imports) {
        addImportsForDependencies(imports, new String[] { "com.springsource.server.web.dm" });
        addImportsForSystemBundleExports(imports);
    }

    protected void addImportsForSystemBundleExports(List<PackageImport> imports) {
        BundleContext bundleContext = this.osgi.getBundleContext();
        ServiceReference ref = bundleContext.getServiceReference(PackageAdmin.class.getName());
        Set<String> selfProvidedPackages = getSelfProvidedPackages();
        Set<ExportedPackage> exportedPackages = new HashSet<ExportedPackage>();

        for (ExportedPackage systemExport : ((PackageAdmin) bundleContext.getService(ref)).getExportedPackages(bundleContext.getBundle())) {
            String packageName = systemExport.getName();
            if (!isExcludedSystemBundleExport(selfProvidedPackages, packageName) && !isAlreadyImported(imports, packageName)) {

                Set<ExportedPackage> duplicatePackages = new HashSet<ExportedPackage>();
                ExportedPackage systemExportWithLowestVersion = null;
                for (ExportedPackage trackedPackage : exportedPackages) {
                    if (trackedPackage.getName().equals(packageName)) {
                        duplicatePackages.add(trackedPackage);
                        if (trackedPackage.getVersion().compareTo(systemExport.getVersion()) < 0) {
                            systemExportWithLowestVersion = trackedPackage;
                        }
                    }
                }

                if (systemExportWithLowestVersion == null) {
                    systemExportWithLowestVersion = systemExport;
                }
                exportedPackages.removeAll(duplicatePackages);
                exportedPackages.add(systemExportWithLowestVersion);
            }
        }
        bundleContext.ungetService(ref);

        for (ExportedPackage systemExport : exportedPackages) {
            String packageName = systemExport.getName();
            PackageImport systemPackageImport = this.bundleManifestFactory.createPackageImport();
            systemPackageImport.getPackageNames().add(packageName);
            systemPackageImport.getAttributes().put("version", constructStrictVersionRange(systemExport));
            imports.add(systemPackageImport);
        }
    }

    /**
     * {@inheritDoc}
     */
    protected File getWebRoot() {
        return getModuleData().getFile();
    }

    private String constructStrictVersionRange(ExportedPackage systemExport) {
        final String dotZero = ".0";
        StringBuffer packageVersionRange = new StringBuffer(32).append("[");
        String packageVersion = systemExport.getVersion().toString();
        while (packageVersion.endsWith(dotZero)) {
            packageVersion = packageVersion.substring(0, packageVersion.length() - dotZero.length());
        }
        packageVersionRange.append(packageVersion).append(",").append(packageVersion).append("]");
        return packageVersionRange.toString();
    }

    /**
     * Determines whether the supplied package name should be excluded when auto-importing packages exported by the
     * system bundle.
     */
    private boolean isExcludedSystemBundleExport(Set<String> selfProvidedPackages, String packageName) {
        return packageName.startsWith("org.eclipse") || packageName.startsWith("com.springsource") || selfProvidedPackages.contains(packageName);
    }

    private Set<String> getSelfProvidedPackages() {
        Set<String> selfProvidedPackages = new HashSet<String>();
        getPackagesFromLib(selfProvidedPackages);
        getPackagesFromClasses(selfProvidedPackages);
        return selfProvidedPackages;
    }

    private void getPackagesFromLib(Set<String> packages) {
        if (this.module != null) {
            List<Entry> libJars = getLibJars(this.module.getModuleData().getEntry(WEB_INF_LIB_PATH));
            for (Entry libJar : libJars) {
                try {
                    JarInputStream input = new JarInputStream(libJar.getInputStream());
                    JarEntry jarEntry;
                    while ((jarEntry = input.getNextJarEntry()) != null) {
                        if (jarEntry.isDirectory()) {
                            packages.add(getPackageName(jarEntry.getName(), ""));
                        }
                    }
                } catch (IOException e) {
                    throw new FatalDeploymentException("Unable to deploy module " + this.module.getModuleData().getFilename(), e);
                }
            }
        }
    }

    private List<Entry> getLibJars(Entry libPath) {
        List<Entry> libJars = new ArrayList<Entry>();
        for (Entry child : libPath.children()) {
            if (child.getName().endsWith(".jar")) {
                libJars.add(child);
            }
        }
        return libJars;
    }

    private void getPackagesFromClasses(Set<String> packages) {
        if (this.module != null) {
            getChildPackages(packages, this.module.getModuleData().getEntry(WEB_INF_CLASSES_PATH));
        }
    }

    @EntryExitTraceLevel(Level.DEBUG)
    private void getChildPackages(Set<String> packages, Entry entry) {
        if (entry.isDirectory()) {
            packages.add(getPackageName(entry.getName(), WEB_INF_CLASSES_PATH));
            for (Entry child : entry.children()) {
                getChildPackages(packages, child);
            }
        }
    }

    @EntryExitTraceLevel(Level.DEBUG)
    private String getPackageName(String path, String prefix) {
        String work = path;
        if (work.endsWith("/")) {
            work = work.substring(0, work.length() - 1);
        }
        return work.replace(prefix, "").replace('/', '.');
    }

}
