/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.web.core;

import static com.springsource.server.web.dm.ServerOsgiBundleXmlWebApplicationContext.BUNDLE_CONTEXT_ATTRIBUTE;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

import javax.servlet.ServletContext;

import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.springframework.core.Conventions;
import org.springframework.osgi.util.OsgiStringUtils;
import org.springframework.util.StringUtils;

import com.springsource.server.deployer.core.DeploymentContext;
import com.springsource.server.deployer.core.DeploymentException;
import com.springsource.server.deployer.core.Module;
import com.springsource.server.deployer.core.ZeroModulePersonality;
import com.springsource.server.deployer.core.DeploymentArtefact.Entry;
import com.springsource.server.kernel.core.BundleStarter;
import com.springsource.server.osgi.manifest.BundleManifest;
import com.springsource.server.osgi.manifest.BundleManifestFactory;
import com.springsource.server.osgi.manifest.BundleManifest.ImportPackageHeader;
import com.springsource.server.osgi.manifest.BundleManifest.ImportPackageHeader.PackageImport;
import com.springsource.server.serviceability.ffdc.NonNull;
import com.springsource.server.serviceability.output.Level;
import com.springsource.server.serviceability.tracing.EntryExitTraceLevel;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;
import com.springsource.server.servlet.core.DeployedWebApplicationNamesRegistry;
import com.springsource.server.servlet.core.DynamicContext;
import com.springsource.server.servlet.core.ServletContainer;
import com.springsource.server.web.core.internal.ApplicationContextRetrievingAndPublishingCallback;
import com.springsource.server.web.core.internal.webxml.WebXmlBuilder;
import com.springsource.server.web.dm.ServerOsgiBundleXmlWebApplicationContext;

/**
 * {@link WebModulePersonality} provides deployment of Web modules.
 * 
 * <p/> <strong>Concurrent Semantics</strong><br/>
 * 
 * This class is thread safe providing the supplied {@link ServletContainer} is thread safe.
 * 
 * @author Sam Brannen
 * @author Andy Wilkinson
 * @author Glyn Normington
 * @since 1.0
 */
public class WebModulePersonality extends ZeroModulePersonality {

    public static final String PERSONALITY = "Web";

    public static final String DYNAMIC_CONTEXT_ATTRIBUTE = Conventions.getQualifiedAttributeName(WebModulePersonality.class, "dynamicContext");

    public static final String WEB_APPLICATION_ID_ATTRIBUTE = "SpringSource-Web-Application-ID";

    private static final String ROOT_PATH = ".";

    private static final String BUNDLE_PATH_SEPARATOR = "/";

    private static final String JAR_EXTENSION = ".jar";

    private static final String WEB_INF_PATH = "WEB-INF";

    private static final String WEB_INF_ABSOLUTE_PATH = BUNDLE_PATH_SEPARATOR + WEB_INF_PATH;

    private static final String WEB_INF_CLASSES_PATH = WEB_INF_PATH + BUNDLE_PATH_SEPARATOR + "classes";

    private static final String WEB_INF_CLASSES_ABSOLUTE_PATH = BUNDLE_PATH_SEPARATOR + WEB_INF_CLASSES_PATH;

    private static final String WEB_INF_LIB_PATH = WEB_INF_PATH + BUNDLE_PATH_SEPARATOR + "lib";

    private static final String WEB_INF_LIB_ABSOLUTE_PATH = BUNDLE_PATH_SEPARATOR + WEB_INF_LIB_PATH;

    private static final String MODULE_INF_PATH = "MODULE-INF";

    private static final String MODULE_INF_ABSOLUTE_PATH = BUNDLE_PATH_SEPARATOR + MODULE_INF_PATH;

    private static final Tracer TRACER = TracerFactory.createTracer(WebModulePersonality.class);

    protected final BundleManifestFactory bundleManifestFactory;

    private final ServletContainer servletContainer;

    private final DeploymentContext deploymentContext;

    private final WebXmlBuilder webXmlBuilder;

    private String personalityIdentifier;

    private final ApplicationContextRetrievingAndPublishingCallback applicationContextRetrievingAndPublishingCallback;

    private final DeployedWebApplicationNamesRegistry deployedWebApplicationNamesRegistry;

    public WebModulePersonality(@NonNull BundleManifestFactory bundleManifestFactory, @NonNull ServletContainer servletContainer,
        @NonNull Module module, @NonNull DeployedWebApplicationNamesRegistry deployedWebApplicationNamesRegistry,
        @NonNull DeploymentContext deploymentContext, @NonNull BundleStarter bundleStarter) {
        super(module);
        this.bundleManifestFactory = bundleManifestFactory;
        this.servletContainer = servletContainer;
        this.deployedWebApplicationNamesRegistry = deployedWebApplicationNamesRegistry;
        this.deploymentContext = deploymentContext;
        this.webXmlBuilder = new WebXmlBuilder();
        this.applicationContextRetrievingAndPublishingCallback = new ApplicationContextRetrievingAndPublishingCallback(bundleStarter);
    }

    /**
     * Gets the default <em>context path</em> based on the filename of the deployment artifact. Can be overridden by
     * subclasses.
     * 
     * @return the default <em>context path</em>
     * @see #getContextPath()
     */
    protected String getDefaultContextPath() {
        String fileName = getModuleData().getFilename();
        return fileName.lastIndexOf(".") == -1 ? fileName : fileName.substring(0, fileName.lastIndexOf("."));
    }

    /**
     * Gets the <em>context path</em> as specified in the web module via the
     * {@value #WEB_CONTEXT_PATH "Web-ContextPath"} bundle manifest entry.
     * 
     * @return the context path from the bundle manifest, or <code>null</code> if not found.
     * @see #WEB_CONTEXT_PATH
     * @see #getContextPath()
     */
    protected String getContextPathFromBundleManifest() {
        return (String) this.module.getBundle().getHeaders().get(WebManifestHeaders.CONTEXT_PATH);
    }

    /**
     * Determine the <em>context path</em> for the web module.
     * <p>
     * If the web module contains metadata which specifies a desired context path, that configured value should be used;
     * otherwise, a default context path should be inferred.
     * </p>
     * 
     * @return the context path for the web module
     * @see #getContextPathFromBundleManifest()
     * @see #getDefaultContextPath()
     */
    protected String getContextPath() {
        String contextPath = getContextPathFromBundleManifest();
        if (!StringUtils.hasText(contextPath)) {
            contextPath = getDefaultContextPath();
        }
        if (!contextPath.startsWith("/")) {
            contextPath = "/" + contextPath;
        }
        return contextPath;
    }

    /**
     * Gets the unique ID for the web application based on the application's name and version.
     */
    protected String getWebApplicationId() {
        return this.deploymentContext.getApplicationSymbolicName() + "-" + this.deploymentContext.getApplicationVersion();
    }

    /**
     * Gets the absolute path to the root of the web application which will be passed to the {@link ServletContainer} to
     * {@link ServletContainer#createContext(String, File, Bundle, Bundle) create} the context.
     * 
     * @throws DeploymentException
     */
    protected File getWebRoot() throws DeploymentException {
        Entry moduleInfEntry = getModuleData().getEntry(MODULE_INF_ABSOLUTE_PATH);
        if (!moduleInfEntry.exists()) {
            TRACER.traceError("Web module [%s] uses invalid format: '%s' path must exist.", getModuleData().getFile(), MODULE_INF_ABSOLUTE_PATH);
            throw new DeploymentException("Web module [" + getModuleData().getFile() + "] uses invalid format: '" + MODULE_INF_ABSOLUTE_PATH
                + "' path must exist.");
        }
        return new File(getModuleData().getFile(), moduleInfEntry.getFilename());
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void transform() throws DeploymentException {
        transformWebModuleIntoWarFormat();
        generateWebXmlForWebModule();
        transformBundleManifest();
    }

    /**
     * Transforms a <em>web module</em> into a standard Java EE WAR format.
     * 
     * @throws DeploymentException
     */
    protected void transformWebModuleIntoWarFormat() throws DeploymentException {
        if (!getModuleData().getEntry(MODULE_INF_ABSOLUTE_PATH).exists()) {
            if (getModuleData().getEntry(WEB_INF_ABSOLUTE_PATH).exists()) {
                TRACER.traceError("Web module [%s] uses invalid format: '%s' must reside under '%s'.", getModuleData().getFile(), WEB_INF_PATH,
                    MODULE_INF_ABSOLUTE_PATH);
                throw new DeploymentException("Web module [" + getModuleData().getFile() + "] uses invalid format: '" + WEB_INF_PATH
                    + "' must reside under '" + MODULE_INF_ABSOLUTE_PATH + "'.");
            } else {
                getModuleData().getEntry("/").getDeploymentArtefact().createDirectoryEntry(MODULE_INF_ABSOLUTE_PATH);
            }
        }
    }

    /**
     * Generates <code>web.xml</code> for the <em>web module</em>.
     */
    protected void generateWebXmlForWebModule() {
        this.webXmlBuilder.generateWebXml(super.module);
    }

    @EntryExitTraceLevel(Level.DEBUG)
    protected boolean isAlreadyImported(List<PackageImport> imports, String packageName) {
        for (PackageImport packageImport : imports) {
            if (packageImport.getPackageNames().contains(packageName)) {
                return true;
            }
        }
        return false;
    }

    @EntryExitTraceLevel(Level.DEBUG)
    protected void addImportsForDependencies(List<PackageImport> imports, String[] dependencyPackages) {
        for (String packageName : dependencyPackages) {
            if (!isAlreadyImported(imports, packageName)) {
                PackageImport packageImport = this.bundleManifestFactory.createPackageImport();
                packageImport.getPackageNames().add(packageName);
                imports.add(packageImport);
            }
        }
    }

    /**
     * <em>Template method</em> which transforms the supplied {@link BundleManifest}.
     * 
     * @see #addContainerSpecificPackageImports(List)
     * @see #addPersonalitySpecificPackageImports(List)
     * @see #getBundleClassPathUpdatedWithWarLocations(List)
     */
    protected final void transformBundleManifest() throws DeploymentException {

        BundleManifest bundleManifest = getModuleData().getBundleManifest();

        List<PackageImport> imports;
        ImportPackageHeader importPackageHeader = bundleManifest.getImportPackageHeader();
        if (importPackageHeader == null) {
            imports = new ArrayList<PackageImport>(1);
            bundleManifest.setImportPackageHeader(this.bundleManifestFactory.createImportPackageHeader(imports));
        } else {
            imports = importPackageHeader.getPackageImports();
        }

        addContainerSpecificPackageImports(imports);
        addPersonalitySpecificPackageImports(imports);

        bundleManifest.setBundleClassPath(getBundleClassPathUpdatedWithWarLocations(bundleManifest.getBundleClassPath()));
    }

    /**
     * Adds package names specific to the {@link ServletContainer} to the supplied list of packages to be auto-imported
     * for a bundle.
     * 
     * @param imports the list of packages to which to add personality-specific packages
     * 
     * @see ServletContainer#getPackageNamesOfDependenciesThatContainApplicationExposedTypes()
     */
    protected void addContainerSpecificPackageImports(List<PackageImport> imports) {
        addImportsForDependencies(imports, this.servletContainer.getPackageNamesOfDependenciesThatContainApplicationExposedTypes());
    }

    /**
     * Adds package names specific to this module personality to the supplied list of packages to be auto-imported for a
     * bundle.
     * 
     * @param imports the list of packages to which to add personality-specific packages
     */
    protected void addPersonalitySpecificPackageImports(List<PackageImport> imports) {
        addImportsForDependencies(imports, new String[] { "com.springsource.server.web.dm", "org.springframework.web.servlet" });
    }

    /**
     * Augments the supplied bundle class path. Specifically, the resulting bundle class path should contain the root
     * (for resource loading only), /WEB_INF/classes if that directory exists, and all the jars in /WEB_INF/lib if that
     * directory exists.
     * 
     * @param initialBundleClassPath the initial bundle class path
     */
    private List<String> getBundleClassPathUpdatedWithWarLocations(List<String> initialBundleClassPath) throws DeploymentException {

        List<String> updatedBundleClassPath = new ArrayList<String>(2);

        updatedBundleClassPath.add(ROOT_PATH);

        String webInfPrefix = "";
        String webInfAbsPathPrefix = "";
        Entry moduleInfEntry = getModuleData().getEntry(MODULE_INF_ABSOLUTE_PATH);
        if (moduleInfEntry.exists()) {
            updatedBundleClassPath.add(MODULE_INF_PATH);
            webInfPrefix = MODULE_INF_PATH + BUNDLE_PATH_SEPARATOR;
            webInfAbsPathPrefix = MODULE_INF_ABSOLUTE_PATH;
        }

        if (getModuleData().getEntry(webInfAbsPathPrefix + WEB_INF_CLASSES_ABSOLUTE_PATH).exists()) {
            updatedBundleClassPath.add(webInfPrefix + WEB_INF_CLASSES_PATH);
        }

        Entry webInfLib = getModuleData().getEntry(webInfAbsPathPrefix + WEB_INF_LIB_ABSOLUTE_PATH);
        if (webInfLib.exists()) {
            for (Entry entry : webInfLib.getDeploymentArtefact().getContents()) {
                String libName = entry.getName();
                if (libName.toLowerCase().endsWith(JAR_EXTENSION)) {
                    updatedBundleClassPath.add(webInfPrefix + WEB_INF_LIB_PATH + libName);
                }
            }
        }

        if (initialBundleClassPath != null && initialBundleClassPath.size() > 0) {
            if (initialBundleClassPath.get(0).equals(ROOT_PATH)) {
                initialBundleClassPath.remove(0);
            }
            updatedBundleClassPath.addAll(initialBundleClassPath);
        }

        return updatedBundleClassPath;
    }

    /**
     * {@link ServletContainer#createContext(String, java.io.File, Bundle, Bundle) Creates} a {@link DynamicContext} for
     * the managed web module and stores the context as a module data attribute under the
     * {@link #DYNAMIC_CONTEXT_ATTRIBUTE} key.
     * 
     * Publishes the {@link BundleContext} of the managed module's {@link Bundle web bundle} via the
     * {@link ServletContext} of the <code>DynamicContext</code> under the attribute name
     * {@link ServerOsgiBundleXmlWebApplicationContext#BUNDLE_CONTEXT_ATTRIBUTE}.
     * 
     * Publishes the {@link #getWebApplicationId() web application ID} of the managed module via the
     * {@link ServletContext} of the <code>DynamicContext</code> under the attribute name
     * {@link #WEB_APPLICATION_ID_ATTRIBUTE}.
     * 
     * @see #getContextPath()
     * @see #getWebRoot()
     * @see #getWebApplicationId()
     */
    @Override
    public void onStarting() throws DeploymentException {
        super.onStarting();

        DynamicContext context = this.servletContainer.createContext(getContextPath(), getWebRoot(), this.module.getContext().getBundle(),
            this.module.getBundle());
        this.module.getModuleData().setAttribute(DYNAMIC_CONTEXT_ATTRIBUTE, context);

        TRACER.traceDebug("Setting BundleContext as ServletContext attribute '%s' for web bundle '%s'.", BUNDLE_CONTEXT_ATTRIBUTE,
            OsgiStringUtils.nullSafeSymbolicName(this.module.getBundle()));
        context.getServletContext().setAttribute(BUNDLE_CONTEXT_ATTRIBUTE, this.module.getBundle().getBundleContext());
        context.getServletContext().setAttribute(WEB_APPLICATION_ID_ATTRIBUTE, getWebApplicationId());
    }

    /**
     * Delegates to {@link ApplicationContextRetrievingAndPublishingCallback} for pre-deployment processing of the
     * module and then {@link ServletContainer#addContext(DynamicContext, Bundle) adds} the {@link DynamicContext}
     * created in {@link #onStarting()} to the {@link ServletContainer} thus effectively deploying the web application.
     */
    @Override
    public void deploy() throws DeploymentException {
        super.deploy();

        final DynamicContext context = (DynamicContext) this.module.getModuleData().getAttribute(DYNAMIC_CONTEXT_ATTRIBUTE);

        this.applicationContextRetrievingAndPublishingCallback.preContextDeployment(this.module, context);

        if (!this.servletContainer.addContext(context, this.module.getBundle())) {
            throw new DeploymentException("Failed to start web application [" + context.getPath() + "] for web bundle [" + getModuleData().getFile()
                + "].");
        }

        this.personalityIdentifier = getContextPath();
        this.deployedWebApplicationNamesRegistry.registerWebApplication(getContextPath(), getWebApplicationId());
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void start() throws DeploymentException {
        super.start();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void stop() throws DeploymentException {
        super.stop();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void undeploy() throws DeploymentException {
        try {
            super.undeploy();
            this.servletContainer.removeContext(getContextPath());
        } finally {
            this.deployedWebApplicationNamesRegistry.unregisterWebApplication(getContextPath());
        }
    }

    /**
     * Always returns &quot;{@link #PERSONALITY Web}&quot;.
     */
    @Override
    public String getPersonality() {
        return PERSONALITY;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String getPersonalityIdentifier() {
        return this.personalityIdentifier;
    }

}
