/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.web.core.internal.webxml;

import org.springframework.util.StringUtils;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * Object representation of <code>web.xml</code>'s <code>&lt;error-page&gt;</code> element.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Immutable and therefore thread-safe.
 * 
 * @author Sam Brannen
 * @since 1.0
 * @see DeploymentDescriptor
 */
class ErrorPage implements DeploymentDescriptorElement {

    protected static final String ERROR_PAGE_ELEMENT = "error-page";

    protected static final String EXCEPTION_TYPE_ELEMENT = "exception-type";

    protected static final String ERROR_CODE_ELEMENT = "error-code";

    protected static final String LOCATION_ELEMENT = "location";

    private final String errorCode;

    private final String exceptionType;

    private final String location;

    /**
     * Creates a new <code>ErrorPage</code> element with the supplied error code, exception type, and location.
     */
    private ErrorPage(String errorCode, String exceptionType, String location) {
        this.errorCode = errorCode;
        this.exceptionType = exceptionType;
        this.location = location;
    }

    /**
     * Creates a new <code>ErrorPage</code> element with the supplied error code and location.
     */
    public static final ErrorPage createForErrorCode(String errorCode, String location) {
        return new ErrorPage(errorCode, null, location);
    }

    /**
     * Creates a new <code>ErrorPage</code> element with the supplied exception type and location.
     */
    public static final ErrorPage createForExceptionType(String exceptionType, String location) {
        return new ErrorPage(null, exceptionType, location);
    }

    /**
     * {@inheritDoc}
     */
    public void appendToDocument(Document doc, Element root) {
        Element errorPageElement = doc.createElement(ERROR_PAGE_ELEMENT);
        if (StringUtils.hasText(getErrorCode())) {
            WebXmlBuilder.appendTextElement(doc, errorPageElement, ERROR_CODE_ELEMENT, getErrorCode());
        } else if (StringUtils.hasText(getExceptionType())) {
            WebXmlBuilder.appendTextElement(doc, errorPageElement, EXCEPTION_TYPE_ELEMENT, getExceptionType());
        }
        WebXmlBuilder.appendTextElement(doc, errorPageElement, LOCATION_ELEMENT, getLocation());
        root.appendChild(errorPageElement);
    }

    /**
     * Gets the error code for this error page, possibly <code>null</code>.
     */
    public String getErrorCode() {
        return errorCode;
    }

    /**
     * Gets the exception for this error page, possibly <code>null</code>.
     */
    public String getExceptionType() {
        return this.exceptionType;
    }

    /**
     * Gets the location for this error page.
     */
    public String getLocation() {
        return location;
    }

}
