/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.web.core.internal.webxml;

import java.util.ArrayList;
import java.util.List;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * Object representation of <code>web.xml</code>'s <code>&lt;filter&gt;</code> element.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Immutable and therefore thread-safe.
 * 
 * @author Sam Brannen
 * @since 1.0
 * @see DeploymentDescriptor
 * @see FilterMapping
 */
class Filter implements DeploymentDescriptorElement {

    protected static final String FILTER_ELEMENT = "filter";

    protected static final String FILTER_NAME_ELEMENT = "filter-name";

    protected static final String FILTER_CLASS_ELEMENT = "filter-class";

    private final String name;

    private final String className;

    private final List<InitParam> initParams = new ArrayList<InitParam>();

    private final List<FilterMapping> mappings = new ArrayList<FilterMapping>();

    /**
     * Creates a new <code>Filter</code> element with the supplied name and class name.
     */
    public Filter(String name, String className) {
        this.name = name;
        this.className = className;
    }

    /**
     * {@inheritDoc}
     */
    public void appendToDocument(Document doc, Element root) {
        Element filterElement = doc.createElement(Filter.FILTER_ELEMENT);
        WebXmlBuilder.appendTextElement(doc, filterElement, Filter.FILTER_NAME_ELEMENT, getName());
        WebXmlBuilder.appendTextElement(doc, filterElement, Filter.FILTER_CLASS_ELEMENT, getClassName());
        for (InitParam initParam : this.initParams) {
            initParam.appendToDocument(doc, filterElement);
        }
        root.appendChild(filterElement);

        for (FilterMapping filterMapping : getMappings()) {
            filterMapping.appendToDocument(doc, root);
        }
    }

    /**
     * Gets the name of this filter definition
     */
    public String getName() {
        return this.name;
    }

    /**
     * Gets the class name for this filter definition.
     */
    public String getClassName() {
        return this.className;
    }

    /**
     * Gets the list of {@link InitParam init-params} for this <code>Filter</code>.
     */
    public List<InitParam> getInitParams() {
        return this.initParams;
    }

    /**
     * Adds an {@link InitParam} with the supplied name and value to this filter definition.
     * 
     * @return the new <code>InitParam</code> as a convenience
     */
    public InitParam addInitParam(String name, String value) {
        InitParam initParam = new InitParam(name, value);
        getInitParams().add(initParam);
        return initParam;
    }

    /**
     * Gets the list of {@link FilterMapping filter mappings} for this filter definition.
     */
    public List<FilterMapping> getMappings() {
        return this.mappings;
    }

    /**
     * Adds a {@link FilterMapping} with the supplied URL pattern to this filter definition.
     * 
     * @return the new <code>FilterMapping</code> as a convenience.
     */
    public FilterMapping addMapping(String urlPattern) {
        FilterMapping filterMapping = new FilterMapping(this, urlPattern);
        getMappings().add(filterMapping);
        return filterMapping;
    }

    /**
     * Adds a {@link FilterMapping} with the supplied URL pattern and dispatcher mappings to this filter definition.
     * 
     * @return the new <code>FilterMapping</code> as a convenience.
     */
    public FilterMapping addMapping(String urlPattern, List<String> dispatcherMappings) {
        FilterMapping filterMapping = new FilterMapping(this, urlPattern);
        filterMapping.getDispatcherMappings().addAll(dispatcherMappings);
        getMappings().add(filterMapping);
        return filterMapping;
    }

}
