/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.web.core.internal.webxml;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * Object representation of <code>web.xml</code>'s <code>&lt;listener&gt;</code> element.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Immutable and therefore thread-safe.
 * 
 * @author Sam Brannen
 * @since 1.0
 * @see DeploymentDescriptor
 */
class ServletContextListener implements DeploymentDescriptorElement {

    protected static final String LISTENER_ELEMENT = "listener";

    protected static final String LISTENER_CLASS_ELEMENT = "listener-class";

    private final String className;

    /**
     * Creates a new <code>ServletContextListener</code> with the supplied class name.
     */
    public ServletContextListener(String className) {
        this.className = className;
    }

    /**
     * {@inheritDoc}
     */
    public void appendToDocument(Document doc, Element root) {
        Element listenerElement = doc.createElement(LISTENER_ELEMENT);
        WebXmlBuilder.appendTextElement(doc, listenerElement, LISTENER_CLASS_ELEMENT, getClassName());
        root.appendChild(listenerElement);
    }

    /**
     * Gets the class name of this ServletContextListener element.
     */
    public String getClassName() {
        return this.className;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result + ((this.className == null) ? 0 : this.className.hashCode());
        return result;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        ServletContextListener other = (ServletContextListener) obj;
        if (this.className == null) {
            if (other.className != null) {
                return false;
            }
        } else if (!this.className.equals(other.className)) {
            return false;
        }
        return true;
    }

}
