/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.web.core.internal.webxml;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.jar.Attributes;

import org.springframework.util.StringUtils;

import com.springsource.server.deployer.core.Module;
import com.springsource.server.osgi.manifest.BundleManifest;
import com.springsource.server.osgi.manifest.parse.HeaderDeclaration;
import com.springsource.server.osgi.manifest.parse.HeaderParser;
import com.springsource.server.osgi.manifest.parse.HeaderParserFactory;
import com.springsource.server.osgi.manifest.parse.OSGiIntegrationLogService;
import com.springsource.server.serviceability.output.Level;
import com.springsource.server.serviceability.tracing.EntryExitTraceLevel;
import com.springsource.server.web.core.WebManifestHeaders;
import com.springsource.server.web.core.internal.ApplicationContextRetrievingAndPublishingCallback;

/**
 * Standard implementation of the {@link WebModuleDeploymentDescriptorBuilder} interface.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Immutable and therefore thread-safe.
 * 
 * @author Sam Brannen
 * @since 1.0
 */
class StandardWebModuleDeploymentDescriptorBuilder implements WebModuleDeploymentDescriptorBuilder {

    public static final String DELEGATING_FILTER_PROXY_CLASS_NAME = org.springframework.web.filter.DelegatingFilterProxy.class.getName();

    public static final String DISPATCHER_SERVLET_CLASS_NAME = org.springframework.web.servlet.DispatcherServlet.class.getName();

    public static final String CONTEXT_ATTRIBUTE = "contextAttribute";

    public static final String TARGET_FILTER_LIFECYCLE = "targetFilterLifecycle";

    public static final String DISPATCHER_SERVLET_SUFFIX = "-DispatcherServlet";

    private static final String URL_PATTERNS_DIRECTIVE = "url-patterns";

    private final HeaderParser headerParser = HeaderParserFactory.newHeaderParser(new OSGiIntegrationLogService());

    /**
     * {@inheritDoc}
     */
    public DeploymentDescriptor buildDeploymentDescriptor(Module module) {

        BundleManifest bundleManifest = module.getModuleData().getBundleManifest();
        Attributes attributes = bundleManifest.toManifest().getMainAttributes();

        String bundleSymbolicName = bundleManifest.getBundleSymbolicName().getSymbolicName();
        String dispatcherServletName = bundleSymbolicName + DISPATCHER_SERVLET_SUFFIX;
        String contextAttribute = bundleSymbolicName + ApplicationContextRetrievingAndPublishingCallback.APPLICATION_CONTEXT_SUFFIX;

        DeploymentDescriptor deploymentDescriptor = new DeploymentDescriptor();

        if (isSpringDmPoweredModule(module)) {
            deploymentDescriptor.getServlets().add(retrieveDispatcherServlet(attributes, dispatcherServletName, contextAttribute));
            deploymentDescriptor.setFilters(retrieveDelegatingFilterProxies(attributes, contextAttribute));
        }

        return deploymentDescriptor;
    }

    /**
     * Determines if the supplied {@link Module} is Spring-DM powered.
     */
    protected boolean isSpringDmPoweredModule(Module module) {
        // ApplicationContextConfiguration config = new ApplicationContextConfiguration(module.getBundle());
        // return config.isSpringPoweredBundle();

        // Currently all Web Modules are required to be Spring-DM powered...
        return true;
    }

    protected List<String> parseUrlPatterns(HeaderDeclaration declaration) {
        return parseCommaDelimitedString(declaration.getDirectives().get(URL_PATTERNS_DIRECTIVE));
    }

    @EntryExitTraceLevel(Level.DEBUG)
    protected List<String> parseCommaDelimitedString(String string) {
        List<String> list = new ArrayList<String>();
        if (StringUtils.hasText(string)) {
            String[] urlPatterns = string.split(",");
            for (int i = 0; i < urlPatterns.length; i++) {
                urlPatterns[i] = urlPatterns[i].trim();
                list.add(urlPatterns[i]);
            }
        }
        return list;
    }

    protected List<Filter> retrieveDelegatingFilterProxies(Attributes attributes, String contextAttribute) {
        final List<Filter> filters = new ArrayList<Filter>();
        String bundleHeader = attributes.getValue(WebManifestHeaders.FILTER_MAPPINGS);
        if (StringUtils.hasText(bundleHeader)) {
            List<HeaderDeclaration> declarations = this.headerParser.parseWebFilterMappingsHeader(bundleHeader);
            for (HeaderDeclaration declaration : declarations) {
                String filterName = declaration.getNames().get(0);
                Filter filter = new Filter(filterName, DELEGATING_FILTER_PROXY_CLASS_NAME);
                filter.addInitParam(CONTEXT_ATTRIBUTE, contextAttribute);

                String targetFilterLifecycle = declaration.getDirectives().get(TARGET_FILTER_LIFECYCLE);
                targetFilterLifecycle = StringUtils.hasText(targetFilterLifecycle) ? targetFilterLifecycle.trim() : Boolean.TRUE.toString();
                filter.addInitParam(TARGET_FILTER_LIFECYCLE, targetFilterLifecycle);

                List<String> dispatcherMappings = parseCommaDelimitedString(declaration.getDirectives().get(FilterMapping.DISPATCHER_ELEMENT));
                for (String urlPattern : parseUrlPatterns(declaration)) {
                    filter.addMapping(urlPattern, dispatcherMappings);
                }

                filters.add(filter);
            }
        }
        return filters;
    }

    protected Servlet retrieveDispatcherServlet(Attributes attributes, String dispatcherServletName, String contextAttribute) {
        List<String> dispatcherServletUrlPatterns = retrieveDispatcherServletUrlPatterns(attributes);
        Servlet dispatcherServlet = new Servlet(dispatcherServletName, DISPATCHER_SERVLET_CLASS_NAME, 2);
        dispatcherServlet.addInitParam(CONTEXT_ATTRIBUTE, contextAttribute);
        dispatcherServlet.addMappings(dispatcherServletUrlPatterns);
        return dispatcherServlet;
    }

    protected List<String> retrieveDispatcherServletUrlPatterns(Attributes attributes) {
        List<String> list = parseCommaDelimitedString(attributes.getValue(WebManifestHeaders.DISPATCHER_SERVLET_URL_PATTERNS));
        if (list.isEmpty()) {
            return Arrays.asList(WebManifestHeaders.DEFAULT_DISPATCHER_SERVLET_URL_PATTERNS);
        } else {
            return list;
        }
    }

}
