/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.bootstrap;

import java.util.Map;

/**
 * BootstrapConfiguration provides access to the configuration that's required to bootstrap and launch the Server. More
 * detailed runtime configuration is obtained from the config subsystem.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * Implementations of BootstrapConfiguration must be thread-safe.
 * 
 * @author Andy Wilkinson
 * @author Sam Brannen
 * @since 1.0
 */
public interface BootstrapConfiguration {
    
    static final String SERVER_HOME_PROPERTY = "com.springsource.server.home";
    
    static final String SERVER_CONFIGDIR_PROPERTY = "com.springsource.server.configDir";

    /**
     * <p>
     * The Server's <em>home</em> directory serves as the root directory for the entire system (henceforth referred to
     * as the &lt;server-home&gt;). Each subsystem should create a unique sub-directory within the &lt;server-home&gt;
     * to ensure that its temporary data is isolated from that of other subsystems. In addition, the home directory is
     * typically the parent of the {@link #getWorkDirectory() work} directory, {@link #getLogsDirectory() logs}
     * directory, etc.
     * </p>
     * <p>
     * The Server must have read/write access to this directory.
     * </p>
     * 
     * @return The <strong>absolute path</strong> to the configured home directory for the Server, never
     *         <code>null</code>.
     */
    String getHomeDirectory();

    /**
     * <p>
     * The Server's <em>work</em> directory is a location for subsystems to store temporary data. Each subsystem
     * should create a unique sub-directory to ensure that its temporary data is isolated from that of other subsystems.
     * </p>
     * <p>
     * By default the work directory is located at &lt;server-home&gt;/work.
     * </p>
     * <p>
     * The Server must have read/write access to this directory.
     * </p>
     * 
     * @return The <strong>absolute path</strong> to the configured work directory for the Server, or <code>null</code>
     *         if not available.
     */
    String getWorkDirectory();

    /**
     * <p>
     * The <em>logs</em> directory is the location to which the Server will write its serviceability log files.
     * </p>
     * <p>
     * By default the logs directory is located at &lt;server-home&gt;/serviceability/logs.
     * </p>
     * <p>
     * The Server must have read/write access to this directory.
     * </p>
     * 
     * @return The <strong>absolute path</strong> to the configured logs directory for the Server, or <code>null</code>
     *         if not available.
     */
    String getLogsDirectory();

    /**
     * <p>
     * The <em>trace</em> directory is the location to which the Server will write its serviceability trace files.
     * </p>
     * <p>
     * By default the trace directory is located at &lt;server-home&gt;/serviceability/trace.
     * </p>
     * <p>
     * The Server must have read/write access to this directory.
     * </p>
     * 
     * @return The <strong>absolute path</strong> to the configured trace directory for the Server, or
     *         <code>null</code> if not available.
     */
    String getTraceDirectory();

    /**
     * <p>
     * The <em>dump</em> directory is the location to which the Server will write its serviceability first-failure
     * data capture (FFDC) dumps.
     * </p>
     * <p>
     * By default the logs directory is located at &lt;server-home&gt;/serviceability/dumps.
     * </p>
     * <p>
     * The Server must have read/write access to this directory.
     * </p>
     * 
     * @return The <strong>absolute path</strong> to the configured dump directory for the Server, or <code>null</code>
     *         if not available.
     */
    String getDumpDirectory();

    /**
     * <p>
     * The <em>config</em> path is the set of directories, from which the Server will read each subsystem's configuration.
     * </p>
     * <p>
     * By default the config path is comprised of &lt;server-home&gt;/config.
     * </p>
     * <p>
     * The Server must have read access to each of the directories in this path.
     * </p>
     * 
     * @return The configured <em>config</em> path for the Server, never <code>null</code>.
     */
    String[] getConfigPaths();

    /**
     * <p>
     * The <em>subsystems</em> directory is the location that contains the Server's subsystems.
     * </p>
     * <p>
     * The Server's subsystems are located at &lt;server-home&gt;/bundles/subsystems.
     * </p>
     * 
     * @return The directory containing the Server's subsystems, or <code>null</code> if not available.
     */
    String getSubsystemsDirectory();

    /**
     * <p>
     * The Server's <em>pick-up</em> directory is a location for users to drop files for detection and consumption by
     * the Server. This includes the hot deployment directory.
     * </p>
     * <p>
     * By default the pickup directory is located at &lt;server-home&gt;/pickup.
     * </p>
     * <p>
     * The Server must have read access to this directory.
     * </p>
     * 
     * @return The <strong>absolute path</strong> to the configured pick-up directory for the Server, or
     *         <code>null</code> if not available.
     */
    String getPickUpDirectory();
    
    /**
     * Returns a Map of the classes and packages with configured trace levels. 
     * @return the map of classes and packages with configured trace levels
     */
    Map<String, String> getTraceLevels();

}
