/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.common.env;

/**
 * Identifies the vendor of a JVM implementation. The vendor of the currently running JVM can be accessed from
 * {@link #current()}. When no vendor can be determined, the vendor is identified as {@link #UNKNOWN}. <p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
public enum JvmVendor {
    /**
     * Sun Microsystems.
     */
    SUN("Sun"),

    /**
     * Apple.
     */
    APPLE("Apple"),

    /**
     * IBM.
     */
    IBM("IBM"),

    /**
     * BEA Systems.
     */
    BEA("BEA"),

    /**
     * Unknown.
     */
    UNKNOWN("unknown");

    private static final JvmVendor current;

    private final String identifer;

    static {
        String vmVendor = System.getProperty("java.vm.vendor");
        JvmVendor foundVendor = null;
        for (JvmVendor vendor : values()) {
            if (vmVendor.contains(vendor.identifer)) {
                foundVendor = vendor;
                break;
            }
        }
        current = foundVendor == null ? UNKNOWN : foundVendor;
    }

    private JvmVendor(String identifer) {
        this.identifer = identifer;
    }

    /**
     * Gets the identifier for the vendor of the currently running JVM.
     * 
     * @return the identifier for the JVM vendor.
     */
    public static JvmVendor current() {
        return current;
    }

    /**
     * Indicates whether this <code>JvmVendor</code> is one of the supplied list of vendors.
     * 
     * @param jvmVendors the vendors to check against.
     * @return <code>true</code> if this <code>JvmVendor</code> is in the supplied list, otherwise
     *         <code>false</code>.
     */
    public boolean isOneOf(JvmVendor... jvmVendors) {
        for (JvmVendor jvmVendor : jvmVendors) {
            if (this.equals(jvmVendor)) {
                return true;
            }
        }
        return false;
    }
}
