/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.common.math;

import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * 
 * These methods do simple conversions of between different number types and byte[] and back again These methods work on
 * representations of bytes only, they don't understand values out of ranger for a byte
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Christopher Frost
 * @since 1.0
 */
public class Bases {

    private static final Tracer TRACER = TracerFactory.createTracer(Bases.class);

    /**
     * Converts a positive int into a byte[] of 4 bytes without encoding in to 2's company Does not change the real
     * values of the number, only it's representation
     */
    public static byte[] unsignedIntToByteArray(int integer) {
        byte[] bytes = new byte[4];

        bytes[0] = (byte) (integer >>> 24);
        bytes[1] = (byte) (integer >>> 16);
        bytes[2] = (byte) (integer >>> 8);
        bytes[3] = (byte) integer;

        return bytes;
    }

    /**
     * Converts a byte array of 4 bytes in to an int assuming the byte array is written without 2's company Does not
     * change the real values of the number, only it's representation
     */
    public static int unsignedByteArrayToInt(byte... bytes) {
        return (bytes[0] << 24 & 0xFF000000) + (bytes[1] << 16 & 0xFF0000) + (bytes[2] << 8 & 0xFF00) + (bytes[3] & 0xFF);
    }

    /**
     * Converts a positive short into a byte[] of 2 bytes without encoding in to 2's company Does not change the real
     * values of the number, only it's representation
     */
    public static byte[] unsignedShortToByteArray(short number) {
        byte[] bytes = new byte[2];

        bytes[0] = (byte) (number >>> 8);
        bytes[1] = (byte) number;

        return bytes;
    }

    /**
     * Converts a byte array of 2 bytes in to a short assuming the byte array is written without 2's company Does not
     * change the real values of the number, only it's representation
     */
    public static short unsignedByteArrayToShort(byte... bytes) {
        return (short) ((bytes[0] << 8 & 0xFF00) + (bytes[1] & 0xFF));
    }

    /**
     * Takes an int in the range 0 - 255 and converts it to the java signed version in the range -128 to 127 Does not
     * change the real values of the number, only its representation
     * 
     * @param b
     * @return the signed version in the range -128 to 127
     */
    public static int applyTwosComplementToByte(int b) {
        if (b < 128 && b > -1) {
            return b;
        } else if (b > 127 && b < 256) {
            return b - 256;
        }
        TRACER.traceError("Can not convert byte as it is out of range or already in two's compliment form", b);
        return b;
    }

    /**
     * Takes an int in the range -128 to 127 and converts it to it's unsigned value in the range 0 - 255 Does not change
     * the real values of the number, only it's representation
     * 
     * @param b
     * @return the unsigned value in the range 0 - 255
     */
    public static int removeTwosComplementFromByte(int b) {
        if (b < 128 && b > -1) {
            return b;
        } else if (b > -129 && b < 0) {
            return b + 256;
        }
        TRACER.traceError("Can not convert byte as it is out of range or not in two's compliment form", b);
        return b;
    }

}
