/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.common.math;

import java.util.Set;

/**
 * Provide a mathematical relation which is notionally a set of ordered pairs. It represents a relationship between the
 * members of two sets: the domain and the range.
 * 
 * @author Glyn Normington
 * 
 * @param <X> the type of elements in the domain of the relation
 * @param <Y> the type of elements in the range of the relation
 */
public interface ConcurrentRelation<X, Y> {

    /**
     * Add the given pair (x, y) to the relation. It is not an error if the pair is already present.
     * 
     * @param x
     * @param y
     * @return true if and only if the relation was updated
     */
    boolean add(X x, Y y);

    /**
     * Remove the given pair (x, y) from the relation. It is not an error if the pair is not present.
     * 
     * @param x
     * @param y
     * @return true if and only if the relation was updated
     */
    boolean remove(X x, Y y);

    /**
     * Return true if and only if the given pair (x, y) is present in the relation.
     * 
     * @param x
     * @param y
     * @return true if the pair is present in the relation
     */
    boolean contains(X x, Y y);

    /**
     * Return the domain of the relation.
     * 
     * @return a set comprising the domain
     */
    Set<X> dom();

    /**
     * Return the range of the relation.
     * 
     * @return a set comprising the range
     */
    Set<Y> ran();

    /**
     * Return the relational image of a set of X's.
     * 
     * @param xset a set of X's
     * @return the set of Y's which relate to X's in xset
     */
    Set<Y> relationalImage(Set<X> xset);

    /**
     * Remove all the pairs (x, y) where x is in the given set of X's.
     * 
     * @param xset a set of X's
     */
    void domSubtract(Set<X> xset);
}
