/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.concurrent.core;

import java.lang.management.ManagementFactory;
import java.lang.management.ThreadInfo;
import java.lang.management.ThreadMXBean;
import java.util.HashMap;
import java.util.Map;

import com.springsource.server.concurrent.core.accessor.ThreadLocalAccessor;
import com.springsource.server.concurrent.core.accessor.ThreadLocalAccessorFactory;
import com.springsource.server.concurrent.core.accessor.ThreadLocalData;
import com.springsource.server.concurrent.core.monitor.DeadlockAnalyser;
import com.springsource.server.concurrent.core.monitor.DeadlockAnalyser.Deadlock;
import com.springsource.server.serviceability.ffdc.DumpExtension;
import com.springsource.server.serviceability.ffdc.DumpWriter;
import com.springsource.server.serviceability.ffdc.Dumpable;
import com.springsource.server.serviceability.ffdc.DumpWriter.ListWriter;
import com.springsource.server.serviceability.ffdc.DumpWriter.MapWriter;
import com.springsource.server.serviceability.ffdc.DumpWriter.SectionWriter;

/**
 * Dump extension for the concurrent subsystem.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
final class ConcurrentDumpExtension implements DumpExtension {

    private static final int THREAD_LIST_TRIES = 12;

    private static final int ACTIVE_COUNT_TOLERANCE = 2;

    private final ThreadManager threadManager;

    private final ThreadLocalAccessor threadLocalAccessor = ThreadLocalAccessorFactory.newThreadLocalAccessor();

    private final DeadlockAnalyser deadlockAnalyser = new DeadlockAnalyser();

    private final ThreadMXBean threadBean;
    
    private final Dumpable applicationContextDependencyMonitor;

    /**
     * Creates a new <code>ConcurrentDumpExtension</code>.
     * 
     * @param threadManager the {@link ThreadManager}.
     */
    public ConcurrentDumpExtension(ThreadManager threadManager, Dumpable applicationContextDependencyMonitor) {
        this.threadManager = threadManager;
        this.threadBean = ManagementFactory.getThreadMXBean();
        this.applicationContextDependencyMonitor = applicationContextDependencyMonitor;
    }

    /**
     * {@inheritDoc}
     */
    public void dump(SectionWriter mainSection) {
        long[] threadIds = this.threadBean.getAllThreadIds();
        ThreadInfo[] threadInfos = this.threadBean.getThreadInfo(threadIds, 32);
        Map<Long, Thread> threads = findThreads(threadInfos);
        ListWriter<SectionWriter> list = mainSection.list("threads");
        for (ThreadInfo threadInfo : threadInfos) {
            MapWriter<ListWriter<SectionWriter>> threadWriter = list.map();
            Thread thread = threads.get(threadInfo.getThreadId());

            dumpBaseInformation(threadWriter, threadInfo, thread);
            dumpThreadLocals(threadWriter, thread);
            threadWriter.end();
        }
        list.end();
        dumpDeadlocks(mainSection);
        this.applicationContextDependencyMonitor.dump(mainSection);
        mainSection.end();
    }

    /**
     * Dump out all deadlocks.
     * 
     * @param mainSection the {@link DumpWriter} to dump to.
     */
    private void dumpDeadlocks(SectionWriter mainSection) {
        Deadlock[] deadlocks = this.deadlockAnalyser.findDeadlocks();
        ListWriter<SectionWriter> deadlockList = mainSection.list("deadlocks");
        if (deadlocks.length > 0) {
            for (Deadlock deadlock : deadlocks) {
                deadlockList.scalar(deadlock.toString());
            }
        }
        deadlockList.end();
    }

    /**
     * Dump out all thread local state.
     * 
     * @param threadWriter the {@link SectionWriter} to dump to.
     * @param thread the {@link Thread} to dump for.
     */
    private void dumpThreadLocals(MapWriter<ListWriter<SectionWriter>> threadWriter, Thread thread) {
        ThreadLocalData[] threadLocals = this.threadLocalAccessor.threadLocals(thread);
        ListWriter<MapWriter<ListWriter<SectionWriter>>> list = threadWriter.list("thread-locals");
        for (ThreadLocalData threadLocalData : threadLocals) {
            ThreadLocal<?> tl = threadLocalData.getThreadLocal();
            if (tl == null) {
                continue;
            }
            try {
                list.map().scalar("name", tl.getClass().getName()).scalar("type", tl.getClass().getSimpleName()).scalar("value",
                    new DeadlockSafeToString(threadLocalData.get())).end();
            } catch (NoClassDefFoundError e) {
                list.map().scalar("error", e.getMessage());
            }
        }
        list.end();
    }

    /**
     * Dump out all base thread information.
     * 
     * @param threadWriter the {@link MapWriter} to dump to.
     * @param threadInfo the {@link ThreadInfo}.
     * @param thread the {@link Thread}.
     */
    private void dumpBaseInformation(MapWriter<ListWriter<SectionWriter>> threadWriter, ThreadInfo threadInfo, Thread thread) {
        threadWriter.scalar("thread-id", threadInfo.getThreadId());
        threadWriter.scalar("name", threadInfo.getThreadName());
        threadWriter.scalar("state", threadInfo.getThreadState());
        threadWriter.scalar("name", threadInfo.getThreadName());
        threadWriter.scalar("blocked-time", threadInfo.getBlockedTime());
        threadWriter.scalar("blocked-count", threadInfo.getBlockedCount());
        threadWriter.scalar("waited-time", threadInfo.getWaitedTime());
        threadWriter.scalar("waited-count", threadInfo.getWaitedCount());
        threadWriter.scalar("is-managed", this.threadManager.isEngineManagedThread(thread));
        threadWriter.scalar("lock-owner-id", threadInfo.getLockOwnerId());
        threadWriter.scalar("lock-owner-name", threadInfo.getLockOwnerName());
        threadWriter.scalar("lock-name", threadInfo.getLockName());
        threadWriter.list("stack-trace").scalar(threadInfo.getStackTrace());
        threadWriter.scalar("is-triggering-thread", thread.equals(Thread.currentThread()));

    }

    /**
     * Create a mapping <code>(ThreadId +> Thread)</code> for all the supplied {@link ThreadInfo} instances.
     * 
     * @param threadInfos the {@link ThreadInfo} instances.
     * @return the mapping.
     */
    private Map<Long, Thread> findThreads(ThreadInfo[] threadInfos) {
        ThreadGroup main = Thread.currentThread().getThreadGroup();
        while (main.getParent() != null) {
            main = main.getParent();
        }
        Thread[] threads = new Thread[main.activeCount() * ACTIVE_COUNT_TOLERANCE];
        int threadCount = Integer.MAX_VALUE;
        int tries = 0;

        while (tries++ < THREAD_LIST_TRIES && (threadCount = main.enumerate(threads)) >= threads.length) {
            threads = new Thread[threads.length * ACTIVE_COUNT_TOLERANCE];
        }

        Map<Long, Thread> results = new HashMap<Long, Thread>(threadCount);
        for (int x = 0; x < threadCount; x++) {
            Thread thread = threads[x];
            results.put(thread.getId(), thread);
        }
        return results;
    }

    /**
     * {@inheritDoc}
     */
    public String getName() {
        return this.getClass().getCanonicalName().replace(".core.ConcurrentDumpExtension", "");
    }
    
    private class DeadlockSafeToString {
    	
    	private final Object object;
    	
    	private DeadlockSafeToString(Object object) {
    		this.object = object;
    	}
    	
    	public String toString() {
    		if (object instanceof Character || object instanceof Number || object instanceof String || object instanceof Boolean) {
    		    return object.toString();
    		} else if (object != null) {
    			return object.getClass().getName() + "@" + System.identityHashCode(object);
    		} else {
    			return "null";
    		}
    	}
    }
}
