/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.concurrent.core;

import java.lang.ref.Reference;
import java.util.Set;

import org.osgi.framework.ServiceRegistration;

import com.google.common.base.FinalizableWeakReference;
import com.google.common.base.ReferenceType;
import com.google.common.collect.ReferenceMap;
import com.springsource.server.common.math.ConcurrentHashSet;

/**
 * Simple wrapper around {@link ReferenceMap} with weak keys and strong values, that unregisters a
 * {@link ServiceRegistration} when the keys are collected.<p/>
 * 
 * Obvious caveat: <strong>do not</strong> register the key values directly with the service registry - they will never
 * be collected.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
public final class ConcurrentWeakServiceRegistrationTracker<K, V> {

    private final ReferenceMap<K, V> map = new ReferenceMap<K, V>(ReferenceType.WEAK, ReferenceType.STRONG);

    private final Set<Reference<K>> references = new ConcurrentHashSet<Reference<K>>();

    /**
     * Records the supplied key/value pair and tracks the supplied {@link ServiceRegistration} for un-registration when
     * the key is collected.
     * 
     * @param key the key.
     * @param value the value.
     * @param registration the <code>ServiceRegistration</code>.
     */
    public void record(K key, V value, final ServiceRegistration registration) {
        this.map.put(key, value);
        this.references.add(new FinalizableWeakReference<K>(key) {

            public void finalizeReferent() {
                ConcurrentWeakServiceRegistrationTracker.this.references.remove(this);
                if (registration != null) {
                    try {
                        registration.unregister();
                    } catch (IllegalStateException ex) {
                        // service already unregistered
                    }
                }
            }

        });
    }

    /**
     * Retrieves the value for the supplied key.
     * 
     * @param key the key.
     * @return the value, or <code>null</code> if the value doesn't exist.
     */
    public V retrieve(K key) {
        return this.map.get(key);
    }

    /**
     * Gets all recorded keys.
     * 
     * @return the recorded keys.
     */
    public Set<K> recordedKeys() {
        return this.map.keySet();
    }
}
