/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.concurrent.core;

import java.util.concurrent.BlockingQueue;
import java.util.concurrent.RejectedExecutionHandler;
import java.util.concurrent.TimeUnit;

import com.springsource.server.kernel.core.Subsystem;

/**
 * Factory for creating managed {@link ServerExecutorService} and {@link ServerScheduledExecutorService} instances.<p/>
 * 
 * When creating an {@link ServerExecutorService} with an unbounded queue, the <code>maxSize</code> parameter is ignored
 * because no more than <code>coreSize</code> threads are created.
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * Implementations must be threadsafe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
public interface ExecutorManager {

    /**
     * Creates a new {@link ServerExecutorService} with a fixed-size thread pool and a bounded work queue.
     * 
     * @param name the name of the new <code>ServerExecutorService</code>.
     * @param owner the {@link Subsystem} that owns the new <code>ServerExecutorService</code>.
     * @param coreSize the core size of the thread pool. This is the steady state thread count. Core threads are
     *        <strong>not</strong> pre-initialised.
     * @param maxSize the maximum number of threads in the thread pool.
     * @param keepAlivePeriod when the number of threads is greater than the core, this is the maximum time that excess
     *        idle threads will wait for new tasks before terminating.
     * @param keepAliveUnit the unit for <code>keepAlivePeriod</code>.
     * @param queue the work queue.
     * @return the new <code>ServerExecutorService</code>.
     */
    ServerExecutorService newExecutorService(String name, Subsystem owner, int coreSize, int maxSize, long keepAlivePeriod, TimeUnit keepAliveUnit,
        BlockingQueue<Runnable> queue);

    /**
     * Creates a new {@link ServerExecutorService} with a fixed-size thread pool and a bounded work queue.
     * 
     * @param name the name of the new <code>ServerExecutorService</code>.
     * @param owner the {@link Subsystem} that owns the new <code>ServerExecutorService</code>.
     * @param coreSize the core size of the thread pool. This is the steady state thread count. Core threads are
     *        <strong>not</strong> pre-initialised.
     * @param maxSize the maximum number of threads in the thread pool.
     * @param keepAlivePeriod when the number of threads is greater than the core, this is the maximum time that excess
     *        idle threads will wait for new tasks before terminating.
     * @param keepAliveUnit the unit for <code>keepAlivePeriod</code>.
     * @param queue the work queue
     * @param rejectedExecutionHandler the {@link RejectedExecutionHandler} to use.
     * @return the new <code>ServerExecutorService</code>.
     */
    ServerExecutorService newExecutorService(String name, Subsystem owner, int coreSize, int maxSize, long keepAlivePeriod, TimeUnit keepAliveUnit,
        BlockingQueue<Runnable> queue, RejectedExecutionHandler rejectedExecutionHandler);

    /**
     * Creates a new {@link ServerScheduledExecutorService} with a fixed-size thread pool.
     * 
     * @param name the name of the new <code>ServerScheduledExecutorService</code>.
     * @param owner the {@link Subsystem} that owns the new <code>ServerScheduledExecutorService</code>.
     * @param coreSize the size of the thread pool.
     * @return the new <code>ServerScheduledExecutorService</code>.
     */
    ServerScheduledExecutorService newScheduledExecutorService(String name, Subsystem owner, int coreSize);

}
