/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.concurrent.core;

import java.util.concurrent.BlockingQueue;
import java.util.concurrent.RejectedExecutionHandler;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

import org.springframework.osgi.service.importer.ImportedOsgiServiceProxy;
import org.springframework.osgi.service.importer.ServiceProxyDestroyedException;

import com.springsource.server.serviceability.output.Level;
import com.springsource.server.serviceability.tracing.EntryExitTraceLevel;
import com.springsource.server.serviceability.tracing.TracingService;

/**
 * Extension of {@link ThreadPoolExecutor} that enforces thread creation via a {@link ThreadManager} and ensures that
 * {@link Thread} local state is cleared after execution of a task.
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * Thread-safe.
 * 
 * @author Rob Harrop
 * @author Christopher Frost
 * @since 1.0
 */
final class ServerThreadPoolExecutor extends ThreadPoolExecutor implements ServerExecutorService {

    static final String DEFAULT_POOL_NAME_PREFIX = "ServerThreadPool";

    private static final AtomicInteger POOL_NAME_COUNTER = new AtomicInteger(0);

    private final ApplicationNameAccessor applicationNameAccessor;

    private final ExecutorServiceDelegate delegate = new ExecutorServiceDelegate();

    private final String poolName;

    /**
     * Creates a new <code>ServerThreadPoolExecutor</code>.
     * 
     * @param corePoolSize the core size of the pool.
     * @param maximumPoolSize the maximum size of the pool.
     * @param keepAliveTime the thread keep alive time.
     * @param unit the {@link TimeUnit} for the keep alive time.
     * @param workQueue the work queue.
     * @param handler the rejected execution handler.
     */
    public ServerThreadPoolExecutor(int corePoolSize, int maximumPoolSize, long keepAliveTime, TimeUnit unit, ThreadManager threadManager,
        BlockingQueue<Runnable> workQueue, RejectedExecutionHandler handler, TracingService tracingService) {
        this(corePoolSize, maximumPoolSize, keepAliveTime, unit, threadManager, workQueue, newPoolName(), handler, tracingService);
    }

    /**
     * Creates a new <code>ServerThreadPoolExecutor</code>.
     * 
     * @param corePoolSize the core size of the pool.
     * @param maximumPoolSize the maximum size of the pool.
     * @param keepAliveTime the thread keep alive time.
     * @param unit the {@link TimeUnit} for the keep alive time.
     * @param workQueue the work queue.
     */
    public ServerThreadPoolExecutor(int corePoolSize, int maximumPoolSize, long keepAliveTime, TimeUnit unit, ThreadManager threadManager,
        BlockingQueue<Runnable> workQueue, TracingService tracingService) {
        this(corePoolSize, maximumPoolSize, keepAliveTime, unit, threadManager, workQueue, newPoolName(), tracingService);
    }

    /**
     * Creates a new <code>ServerThreadPoolExecutor</code>.
     * 
     * @param corePoolSize the core size of the pool.
     * @param maximumPoolSize the maximum size of the pool.
     * @param keepAliveTime the thread keep alive time.
     * @param unit the {@link TimeUnit} for the keep alive time.
     * @param workQueue the work queue.
     * @param poolName the name of the thread pool.
     * @param handler the rejected execution handler.
     */
    public ServerThreadPoolExecutor(int corePoolSize, int maximumPoolSize, long keepAliveTime, TimeUnit unit, ThreadManager threadManager,
        BlockingQueue<Runnable> workQueue, String poolName, RejectedExecutionHandler handler, TracingService tracingService) {
        super(corePoolSize, maximumPoolSize, keepAliveTime, unit, workQueue, new ThreadManagerPoolThreadFactory(poolName, threadManager), handler);
        this.poolName = poolName;
        this.applicationNameAccessor = new ApplicationNameAccessor(tracingService);
    }

    /**
     * Creates a new <code>ServerThreadPoolExecutor</code>.
     * 
     * @param corePoolSize the core size of the pool.
     * @param maximumPoolSize the maximum size of the pool.
     * @param keepAliveTime the thread keep alive time.
     * @param unit the {@link TimeUnit} for the keep alive time.
     * @param workQueue the work queue.
     * @param poolName the name of the thread pool.
     */
    public ServerThreadPoolExecutor(int corePoolSize, int maximumPoolSize, long keepAliveTime, TimeUnit unit, ThreadManager threadManager,
        BlockingQueue<Runnable> workQueue, String poolName, TracingService tracingService) {
        super(corePoolSize, maximumPoolSize, keepAliveTime, unit, workQueue, new ThreadManagerPoolThreadFactory(poolName, threadManager));
        this.poolName = poolName;
        this.applicationNameAccessor = new ApplicationNameAccessor(tracingService);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void afterExecute(Runnable r, Throwable t) {
        super.afterExecute(r, t);
        this.delegate.clearThreadLocals();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    @EntryExitTraceLevel(Level.DEBUG)
    public void execute(Runnable command) {
        final Runnable decorated = this.delegate.decorateForStatisticsTracking(command);
        final String applicationName = this.applicationNameAccessor.getCurrentApplicationName();
        super.execute(new Runnable() {

            @EntryExitTraceLevel(Level.DEBUG)
            public void run() {
                ServerThreadPoolExecutor.this.applicationNameAccessor.setCurrentApplicationName(applicationName);
                decorated.run();
                ServerThreadPoolExecutor.this.applicationNameAccessor.setCurrentApplicationName(null);
            }

        });
    }

    /**
     * Gets an estimate of the average amount of time spent processing successful tasks.
     * 
     * @return the estimate of time spent.
     * @see ThreadPoolExecutor#getCompletedTaskCount()
     */
    public long getAverageExecutionTime() {
        return this.delegate.getAverageExecutionTime(getCompletedTaskCount());
    }

    /**
     * Gets an estimate of the total amount of time spent processing successful tasks.
     * 
     * @return the estimate of time spent.
     * @see ThreadPoolExecutor#getCompletedTaskCount()
     */
    public long getExecutionTime() {
        return this.delegate.getExecutionTime();
    }

    /**
     * Get the unique name of the Pool used in this executor service
     * 
     * {@inheritDoc}
     */
    public String getPoolName() {
        return this.poolName;
    }

    private static String newPoolName() {
        return String.format("%1$s-%2$d", DEFAULT_POOL_NAME_PREFIX, POOL_NAME_COUNTER.getAndIncrement());
    }

    /**
     * Wrapper around {@link TracingService} that handles the service proxy disappearing.<p/>
     * 
     * <strong>Concurrent Semantics</strong><br />
     * 
     * Threadsafe.
     * 
     * @author Rob Harrop
     * @since 1.0
     */
    private static class ApplicationNameAccessor {

        private final TracingService tracingService;

        /**
         * Creates a new <code>ApplicationNameAccessor</code>.
         * 
         * @param tracingService the wrapped {@link TracingService}.
         */
        public ApplicationNameAccessor(TracingService tracingService) {
            this.tracingService = tracingService;
        }

        /**
         * @see TracingService#getCurrentApplicationName()
         */
        @EntryExitTraceLevel(Level.DEBUG)
        public String getCurrentApplicationName() {
            if (isLive()) {
                try {
                    return this.tracingService.getCurrentApplicationName();
                } catch (ServiceProxyDestroyedException ex) {
                }
            }
            return null;
        }

        /**
         * @see TracingService#setCurrentApplicationName(String)
         */
        @EntryExitTraceLevel(Level.DEBUG)
        public void setCurrentApplicationName(String applicationName) {
            if (isLive()) {
                try {
                    this.tracingService.setCurrentApplicationName(applicationName);
                } catch (ServiceProxyDestroyedException ex) {
                }
            }
        }

        private boolean isLive() {
            if (this.tracingService instanceof ImportedOsgiServiceProxy) {
                return ((ImportedOsgiServiceProxy) this.tracingService).getServiceReference().getBundle() != null;
            } else {
                return true;
            }
        }
    }
}
