/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.concurrent.core;

/**
 * {@link Ticker} provides a heartbeat interface for tracking unanticipated delays. The heartbeat is configurable using
 * a policy such as {@link StandardTicker.ExponentialHeartBeatPolicy} which lengthens the heartbeat interval
 * exponentially over time until it reaches a fixed upper bound.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Implementations of this interface must be thread safe.
 * 
 * @author Glyn Normington
 * @since 1.0
 */
public interface Ticker {

    /**
     * Cancel this Ticker.
     *
     * @return <code>true</code> if and only this Ticker has ticked at least once
     */
    boolean cancel();

    /**
     * {@link HeartBeatPolicy} is an interface for configuring heartbeat intervals.
     * <p />
     * 
     * <strong>Concurrent Semantics</strong><br />
     * 
     * Implementations of this interface should be thread safe to avoid potential abuse.
     */
    public static interface HeartBeatPolicy {

        /**
         * Return the heartbeat interval and update the next interval according to the policy.
         * 
         * @return the heartbeat interval in milliseconds
         */
        long getNextHeartBeatIntervalMillis();
    }

}
