/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.concurrent.core.accessor;

import java.lang.ref.Reference;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.locks.ReadWriteLock;

import org.springframework.util.ReflectionUtils;
import org.springframework.util.StringUtils;

import com.springsource.server.concurrent.core.diagnostics.FatalConcurrentException;

/**
 * {@link ThreadLocalAccessor} implementation for the Sun JVM.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
final class SunThreadLocalAccessor implements ThreadLocalAccessor {

    private static Class<?> CLASS_THREAD_LOCAL_MAP = classForName("java.lang.ThreadLocal$ThreadLocalMap");

    private static Class<?> CLASS_THREAD_LOCAL_MAP_ENTRY = classForName("java.lang.ThreadLocal$ThreadLocalMap$Entry");

    /**
     * Method <code>ThreadLocal.set(Object)</code>.
     */
    private static final Method METHOD_SET = getDeclaredMethod(CLASS_THREAD_LOCAL_MAP, "set", ThreadLocal.class, Object.class);

    private static final Method METHOD_REMOVE = getDeclaredMethod(CLASS_THREAD_LOCAL_MAP, "remove", ThreadLocal.class);

    /**
     * Constructor <code>ThreadLocalMap(ThreadLocal, Object)</code>.
     */
    private static final Constructor<?> CONSTRUCTOR_THREAD_LOCAL_MAP = getDeclaredConstructor(CLASS_THREAD_LOCAL_MAP, ThreadLocal.class, Object.class);

    /**
     * Field <code>Thread.threadLocals</code>.
     */
    private static final Field FIELD_THREAD_LOCALS = getDeclaredField(Thread.class, "threadLocals");

    /**
     * Field <code>Thread.inheritableThreadLocals</code>.
     */
    private static final Field FIELD_INHERITABLE_THREAD_LOCALS = getDeclaredField(Thread.class, "inheritableThreadLocals");

    /**
     * Field <code>Reference.referent</code>.
     */
    private static final Field FIELD_REFERENT = getDeclaredField(Reference.class, "referent");

    /**
     * Field <code>ThreadLocalMap.Entry.value</code>.
     */
    private static final Field FIELD_VALUE = getDeclaredField(CLASS_THREAD_LOCAL_MAP_ENTRY, "value");

    /**
     * Field <code>ThreadLocalMap.table</code>.
     */
    private static final Field FIELD_TABLE = getDeclaredField(CLASS_THREAD_LOCAL_MAP, "table");

    /**
     * {@inheritDoc}
     */
    public ThreadLocalData[] threadLocals(Thread thread) {
        try {
            List<ThreadLocalData> threadLocals = new ArrayList<ThreadLocalData>();
            gatherThreadLocals(thread, FIELD_THREAD_LOCALS, threadLocals);
            gatherThreadLocals(thread, FIELD_INHERITABLE_THREAD_LOCALS, threadLocals);
            return threadLocals.toArray(new ThreadLocalData[threadLocals.size()]);
        } catch (IllegalAccessException e) {
            throw new FatalConcurrentException("Unable to access ThreadLocal state.", e);
        }
    }

    /**
     * Gathers the <code>ThreadLocals</code> for a given thread.
     * 
     * @param thread the <code>Thread</code>.
     * @param fieldThreadLocals the <code>Field</code> to access the <code>ThreadLocalMap</code> from.
     * @param threadLocals the <code>List</code> to add <code>ThreadLocals</code> to.
     * @throws IllegalAccessException
     */
    private void gatherThreadLocals(Thread thread, Field fieldThreadLocals, List<ThreadLocalData> threadLocals) throws IllegalAccessException {
        Object threadLocalMap = getThreadLocalMap(thread, fieldThreadLocals);

        if (threadLocalMap != null) {
            Object[] table = (Object[]) FIELD_TABLE.get(threadLocalMap);
            for (Object entry : table) {
                if (entry != null) {
                    threadLocals.add(new ThreadLocalMapEntryData(entry));
                }
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    public <T> void setThreadLocalValue(ThreadLocal<T> threadLocal, T value, Thread thread) {
        Object threadLocalMap = getThreadLocalMapForThreadLocal(threadLocal, thread);
        if (threadLocalMap == null) {
            threadLocalMap = createThreadLocalMap(threadLocal, value, thread);
        }
        try {
            METHOD_SET.invoke(threadLocalMap, threadLocal, value);
        } catch (IllegalArgumentException e) {
            throw new FatalConcurrentException("Unexpected error calling ThreadLocalMap.set() for Thread '" + thread + "'", e);
        } catch (IllegalAccessException e) {
            throw new FatalConcurrentException("Unable to access ThreadLocalMap.set() for Thread '" + thread + "'", e);
        } catch (InvocationTargetException e) {
            throw new FatalConcurrentException("Error setting ThreadLocal on Thread '" + thread + "'", e.getCause());
        }
    }

    /**
     * {@inheritDoc}
     */
    public void removeThreadLocal(ThreadLocal<?> threadLocal, Thread thread) {
        Object threadLocalMap = getThreadLocalMapForThreadLocal(threadLocal, thread);
        if (threadLocalMap != null) {
            try {
                METHOD_REMOVE.invoke(threadLocalMap, threadLocal);
            } catch (IllegalArgumentException e) {
                throw new FatalConcurrentException("Unexpected error removing ThreadLocal from Thread '" + thread + "'", e);
            } catch (IllegalAccessException e) {
                throw new FatalConcurrentException("Access error trying to remove ThreadLocal from Thread '" + thread + "'", e);
            } catch (InvocationTargetException e) {
                throw new FatalConcurrentException("Error removing ThreadLocal from '" + thread + "'", e.getCause());
            }
        }
    }

    /**
     * Determines the <code>ThreadLocalMap</code> for the given <code>ThreadLocal</code>.
     */
    private Object getThreadLocalMapForThreadLocal(ThreadLocal<?> threadLocal, Thread thread) {
        Field mapField = threadLocal instanceof InheritableThreadLocal ? FIELD_INHERITABLE_THREAD_LOCALS : FIELD_THREAD_LOCALS;
        Object threadLocalMap = getThreadLocalMap(thread, mapField);
        return threadLocalMap;
    }

    /**
     * Creates a <code>ThreadLocalMap</code> instance on the supplied <code>Thread</code>
     */
    private <T> Object createThreadLocalMap(ThreadLocal<T> threadLocal, T value, Thread thread) {
        try {
            Object threadLocalMap = CONSTRUCTOR_THREAD_LOCAL_MAP.newInstance(threadLocal, value);
            FIELD_THREAD_LOCALS.set(thread, threadLocalMap);
            return threadLocalMap;
        } catch (IllegalArgumentException e) {
            throw new FatalConcurrentException("Unexpected error creating ThreadLocalMap for Thread '" + thread + "'", e);
        } catch (InstantiationException e) {
            throw new FatalConcurrentException("Unable to create ThreadLocalMap for Thread '" + thread + "'", e);
        } catch (IllegalAccessException e) {
            throw new FatalConcurrentException("Access error trying to create ThreadLocalMap for '" + thread + "'", e);
        } catch (InvocationTargetException e) {
            throw new FatalConcurrentException("Error creating ThreadLocalMap on Thread '" + thread + "'", e.getCause());
        }
    }

    /**
     * Gets the <code>ThreadLocalMap</code> instance for the supplied <code>Thread</code>
     */
    private Object getThreadLocalMap(Thread thread, Field threadLocalsField) {
        try {
            return threadLocalsField.get(thread);
        } catch (IllegalArgumentException e) {
            throw new FatalConcurrentException("Unable to get ThreadLocalMap for Thread '" + thread + "' from field '" + threadLocalsField.getName()
                + "'", e);
        } catch (IllegalAccessException e) {
            throw new FatalConcurrentException("Unable to access ThreadLocalMap field for Thread '" + thread + "'", e);
        }
    }

    private static Field getDeclaredField(Class<?> from, String name) {
        try {
            Field f = from.getDeclaredField(name);
            f.setAccessible(true);
            return f;
        } catch (SecurityException e) {
            throw new FatalConcurrentException(String.format("Unable to access field '%s' on type '%s'.", name, from.getName()), e);
        } catch (NoSuchFieldException e) {
            throw new FatalConcurrentException(String.format("Field '%s' is not a declared field on type '%s'.", name, from.getName()), e);
        }
    }

    private static Method getDeclaredMethod(Class<?> from, String name, Class<?>... parameterTypes) {
        Method m = ReflectionUtils.findMethod(from, name, parameterTypes);
        if (m == null) {
            throw new FatalConcurrentException("Unable to find method '" + name + "' on class '" + from + "' with types '"
                + StringUtils.arrayToCommaDelimitedString(parameterTypes) + "'.");
        }
        ReflectionUtils.makeAccessible(m);
        return m;
    }

    private static Constructor<?> getDeclaredConstructor(Class<?> from, Class<?>... parameterTypes) {
        try {
            Constructor<?> ctor = from.getDeclaredConstructor(parameterTypes);
            ctor.setAccessible(true);
            return ctor;
        } catch (SecurityException e) {
            throw new FatalConcurrentException(String.format("Unable to access constructor (%s) on type '%s'.",
                StringUtils.arrayToCommaDelimitedString(parameterTypes), from.getName()), e);
        } catch (NoSuchMethodException e) {
            throw new FatalConcurrentException("Unexpected error - cannot locate ThreadLocalMap constructor.", e);
        }
    }

    private static Class<?> classForName(String name) {
        try {
            return Class.forName(name);
        } catch (ClassNotFoundException e) {
            throw new FatalConcurrentException("Unable to load class '" + name + "'.", e);
        }
    }

    /**
     * Implementation of {@link ThreadLocalData} that delegates directly to the <code>ThreadLocalMap.Entry</code>.
     */
    private final class ThreadLocalMapEntryData implements ThreadLocalData {

        private final ThreadLocal<?> threadLocal;

        private final Object entry;

        private final ReadWriteLock lock;

        /**
         * Creates a new <code>ThreadLocalMapEntryData</code> from the supplied entry.
         */
        public ThreadLocalMapEntryData(Object entry) {
            this.entry = entry;
            try {
                this.threadLocal = (ThreadLocal<?>) FIELD_REFERENT.get(entry);
                if (this.threadLocal instanceof ConcurrentThreadLocal) {
                    this.lock = ((ConcurrentThreadLocal<?>) this.threadLocal).lock();
                } else {
                    this.lock = null;
                }
            } catch (IllegalArgumentException e) {
                throw new FatalConcurrentException("Unable to get ThreadLocal from entry '" + entry + "'", e);
            } catch (IllegalAccessException e) {
                throw new FatalConcurrentException("Unable to access ThreadLocal in entry '" + entry + "'", e);
            }
        }

        /** 
         * {@inheritDoc}
         */
        public ThreadLocal<?> getThreadLocal() {
            return this.threadLocal;
        }

        /**
         * {@inheritDoc}
         */
        public Object get() {
            lockRead();
            try {
                return FIELD_VALUE.get(this.entry);
            } catch (IllegalArgumentException e) {
                throw new FatalConcurrentException("Unable to get value from entry '" + this.entry + "'", e);
            } catch (IllegalAccessException e) {
                throw new FatalConcurrentException("Unable to access value in entry '" + this.entry + "'", e);
            } finally {
                unlockRead();
            }
        }

        /**
         * {@inheritDoc}
         */
        public void set(Object value) {
            lockWrite();
            try {
                FIELD_VALUE.set(this.entry, value);
            } catch (IllegalArgumentException e) {
                throw new FatalConcurrentException("Unable to set value from entry '" + this.entry + "'", e);
            } catch (IllegalAccessException e) {
                throw new FatalConcurrentException("Unable to access value in entry '" + this.entry + "'", e);
            } finally {
                unlockWrite();
            }
        }

        private void lockRead() {
            if (this.lock != null) {
                this.lock.readLock().lock();
            }
        }

        private void unlockRead() {
            if (this.lock != null) {
                this.lock.readLock().unlock();
            }
        }

        private void lockWrite() {
            if (this.lock != null) {
                this.lock.writeLock().lock();
            }
        }

        private void unlockWrite() {
            if (this.lock != null) {
                this.lock.writeLock().unlock();
            }
        }
    }
}
