/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.concurrent.core.accessor;

/**
 * Provides access to all {@link ThreadLocal ThreadLocals} bound to the current thread.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * <dl>
 * <dt>Reading</dt>
 * <dd>The value of a <code>ThreadLocal</code> value for <code>Thread</code> <code>a</code>, is <strong>not</strong>
 * guaranteed to accurately reflect the current value when read from any <code>Thread</code> other than <code>a</code>.</dd>
 * <dt>Writing</dt>
 * <dd>Writing a value to a <code>ThreadLocal</code> for <code>Thread</code> <code>a</code> from
 * <code>Thread</code> <code>b</code> (where <code>a != b</code>) does <strong>not</strong> guarantee visibility
 * of that value from any <code>Thread</code> other than <code>b</code>.</dd>
 * </dl>
 * 
 * When using a <code>ThreadLocalAccessor</code> to access all <code>ThreadLocals</code> for a given
 * <code>Thread</code> attempt to do so from within that <code>Thread</code>. When this is not possible, provide
 * higher-level access to the <code>ThreadLocal</code> and introduce a synchronization or other mechanism for
 * preserving the happens-before relationship.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
public interface ThreadLocalAccessor {

    /**
     * Gets the {@link ThreadLocal ThreadLocals} for the supplied thread.
     * 
     * @param thread the <code>Thread</code> to get the <code>ThreadLocals</code> for.
     * @return the <code>ThreadLocals</code>; never <code>null</code>.
     */
    ThreadLocalData[] threadLocals(Thread thread);

    /**
     * Sets the supplied {@link ThreadLocal} on the supplied {@link Thread} to <code>value</code>.
     * 
     * @param <T>
     * @param threadLocal the <code>ThreadLocal</code> to set.
     * @param value the value to set.
     * @param thread the <code>Thread</code> to set the value on.
     */
    <T> void setThreadLocalValue(ThreadLocal<T> threadLocal, T value, Thread thread);

    /**
     * Removes state for the supplied {@link ThreadLocal} from the supplied {@link Thread}.
     * 
     * @param threadLocal the <code>ThreadLocal</code>.
     * @param thread the <code>Thread</code>.
     */
    void removeThreadLocal(ThreadLocal<?> threadLocal, Thread thread);
}
