/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.concurrent.core.config;

import org.osgi.framework.BundleContext;
import org.springframework.beans.factory.FactoryBean;
import org.springframework.osgi.context.BundleContextAware;

import com.springsource.server.concurrent.core.ExecutorManager;
import com.springsource.server.concurrent.core.diagnostics.FatalConcurrentException;
import com.springsource.server.kernel.core.Subsystem;
import com.springsource.server.kernel.core.SubsystemRegistry;
import com.springsource.server.serviceability.ffdc.NonNull;

/**
 * Base {@link FactoryBean} for creating services using the {@link ExecutorManager}<p/>
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
abstract class AbstractExecutorServiceFactoryBean implements FactoryBean, BundleContextAware {

    private final ExecutorManager executorManager;

    private final SubsystemRegistry subsystemRegistry;

    private BundleContext bundleContext;

    /**
     * @param executorManager
     */
    protected AbstractExecutorServiceFactoryBean(@NonNull ExecutorManager executorManager, @NonNull SubsystemRegistry registry) {
        this.executorManager = executorManager;
        this.subsystemRegistry = registry;
    }

    /**
     * {@inheritDoc}
     */
    public boolean isSingleton() {
        return true;
    }

    public final Object getObject() {
        if (this.bundleContext == null) {
            throw new FatalConcurrentException("BundleContext has not been injected. Are you running in Spring DM?");
        }
        Subsystem subsystem = this.subsystemRegistry.getSubsystemByBundle(this.bundleContext.getBundle());
        return doGetObject(subsystem);
    }

    public void setBundleContext(BundleContext bundleContext) {
        this.bundleContext = bundleContext;
    }

    /**
     * Gets the configured {@link ExecutorManager}.
     * 
     * @return the <code>ExecutorManager</code>.
     */
    protected final ExecutorManager getExecutorManager() {
        return this.executorManager;
    }

    /**
     * Gets the configured {@link SubsystemRegistry}.
     * 
     * @return the <code>SubsystemRegistry</code>.
     */
    protected final SubsystemRegistry getSubsystemRegistry() {
        return this.subsystemRegistry;
    }

    /**
     * @param subsystem The <code>Subsystem</code> creating the object
     * @return the object created by this factory
     */
    protected abstract Object doGetObject(Subsystem subsystem);
}
