/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.concurrent.core.config;

import java.util.concurrent.BlockingQueue;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.RejectedExecutionHandler;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import org.springframework.beans.factory.FactoryBean;

import com.springsource.server.concurrent.core.ExecutorManager;
import com.springsource.server.kernel.core.Subsystem;
import com.springsource.server.kernel.core.SubsystemRegistry;
import com.springsource.server.serviceability.ffdc.Assert;

/**
 * {@link FactoryBean} for creating managed {@link ExecutorService} instances using the {@link ExecutorManager}.<p/>
 * 
 * The default configuration creates an {@link ExecutorService} with a fixed-size thread pool of one thread.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
public final class ExecutorServiceFactoryBean extends AbstractExecutorServiceFactoryBean {

    public static final int DEFAULT_CORE_SIZE = 1;

    public static final int DEFAULT_MAX_SIZE = 1;

    public static final long DEFAULT_KEEP_ALIVE_TIME = 60L;

    public static final TimeUnit DEFAULT_KEEP_ALIVE_UNIT = TimeUnit.SECONDS;

    public static final RejectedExecutionHandler DEFAULT_REJECTED_EXECUTION_HANDLER = new ThreadPoolExecutor.AbortPolicy();

    private volatile int coreSize = DEFAULT_CORE_SIZE;

    private volatile int maxSize = DEFAULT_MAX_SIZE;

    private volatile long keepAliveTime = DEFAULT_KEEP_ALIVE_TIME;

    private volatile TimeUnit keepAliveUnit = DEFAULT_KEEP_ALIVE_UNIT;

    private volatile RejectedExecutionHandler rejectedExecutionHandler = DEFAULT_REJECTED_EXECUTION_HANDLER;

    private final String name;

    private final BlockingQueue<Runnable> queue;

    /**
     * Creates a new {@link ExecutorServiceFactoryBean}.
     * 
     * @param executorManager the {@link ExecutorManager}.
     * @param queue the work queue.
     */
    public ExecutorServiceFactoryBean(ExecutorManager executorManager, SubsystemRegistry registry, String name, BlockingQueue<Runnable> queue) {
        super(executorManager, registry);
        Assert.notNull(name, "ExecutorService name cannot be null");
        Assert.notNull(queue, "ExecutorService queue cannot be null");
        this.name = name;
        this.queue = queue;
    }

    /**
     * Sets the core size.
     * 
     * @param coreSize core size.
     */
    public void setCoreSize(int coreSize) {
        this.coreSize = coreSize;
    }

    /**
     * Sets the max size.
     * 
     * @param maxSize max size.
     */
    public void setMaxSize(int maxSize) {
        this.maxSize = maxSize;
    }

    /**
     * Sets the keep alive time.
     * 
     * @param keepAliveTime keep alive time.
     */
    public void setKeepAliveTime(long keepAliveTime) {
        this.keepAliveTime = keepAliveTime;
    }

    /**
     * Sets the keep alive unit.
     * 
     * @param keepAliveUnit keep alive unit.
     */
    public void setKeepAliveUnit(TimeUnit keepAliveUnit) {
        this.keepAliveUnit = keepAliveUnit;
    }

    /**
     * Sets the {@link RejectedExecutionHandler}.
     * 
     * @param rejectedExecutionHandler the <code>RejectedExecutionHandler</code>.
     */
    public void setRejectedExecutionHandler(RejectedExecutionHandler rejectedExecutionHandler) {
        this.rejectedExecutionHandler = rejectedExecutionHandler;
    }

    /**
     * {@inheritDoc}
     */
    @Override protected Object doGetObject(Subsystem owner) {
        return getExecutorManager().newExecutorService(this.name, owner, this.coreSize, this.maxSize, this.keepAliveTime, this.keepAliveUnit,
            this.queue, this.rejectedExecutionHandler);
    }

    /**
     * {@inheritDoc}
     */
    public Class<?> getObjectType() {
        return ExecutorService.class;
    }

}
