/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.concurrent.core.monitor;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.concurrent.Callable;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ScheduledExecutorService;

import org.osgi.framework.Bundle;
import org.springframework.context.ApplicationContext;
import org.springframework.osgi.context.event.OsgiBundleApplicationContextEvent;
import org.springframework.osgi.context.event.OsgiBundleApplicationContextListener;
import org.springframework.osgi.context.event.OsgiBundleContextRefreshedEvent;
import org.springframework.osgi.extender.event.BootstrappingDependencyEvent;
import org.springframework.osgi.service.importer.OsgiServiceDependency;
import org.springframework.osgi.service.importer.event.OsgiServiceDependencyEvent;
import org.springframework.osgi.service.importer.event.OsgiServiceDependencyWaitEndedEvent;
import org.springframework.osgi.service.importer.event.OsgiServiceDependencyWaitStartingEvent;
import org.springframework.osgi.service.importer.event.OsgiServiceDependencyWaitTimedOutEvent;

import com.springsource.server.concurrent.core.Ticker;
import com.springsource.server.concurrent.core.diagnostics.ConcurrentLogEvent;
import com.springsource.server.concurrent.core.scheduling.StandardTicker;
import com.springsource.server.osgi.framework.OsgiFrameworkUtils;
import com.springsource.server.serviceability.ffdc.Dumpable;
import com.springsource.server.serviceability.ffdc.DumpWriter.ListWriter;
import com.springsource.server.serviceability.ffdc.DumpWriter.MapWriter;
import com.springsource.server.serviceability.ffdc.DumpWriter.SectionWriter;
import com.springsource.server.serviceability.logging.LogService;
import com.springsource.server.serviceability.logging.LogServiceFactory;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * {@link ApplicationContextDependencyMonitor} is a class that tracks the satisfaction of service dependencies needed
 * during the creation of application contexts and issues log messages for delayed service dependencies. <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * This class is thread safe.
 * 
 * @author Glyn Normington
 * @since 1.0
 */
final class ApplicationContextDependencyMonitor implements OsgiBundleApplicationContextListener, Dumpable {

    private static final int MAXIMUM_WARNING_INTERVAL = 60 * 1000;

    private static final int WARNING_INTERVAL_INCREASE_RATE_PERCENT = 200;

    private static final int INITIAL_WARNING_INTERVAL = 2 * 1000;

    private static final int SLOW_WARNING_INTERVAL = 5 * 60 * 1000;

    private static final Tracer TRACER = TracerFactory.createTracer(ApplicationContextDependencyMonitor.class);

    private static final LogService LOGGER = LogServiceFactory.getLogService();

    private final ScheduledExecutorService scheduledExecutorService;

    private final Map<ApplicationContext, Map<OsgiServiceDependency, Ticker>> tickers = new HashMap<ApplicationContext, Map<OsgiServiceDependency, Ticker>>();

    private final Map<ApplicationContext, Bundle> applicationContextBundle = new ConcurrentHashMap<ApplicationContext, Bundle>();

    private final Object monitor = new Object();

    /**
     * Construct a {@link ApplicationContextDependencyMonitor} which uses the given {@link ScheduledExecutorService} to
     * schedule its warning messages.
     * 
     * @param scheduledExecutorService the {@link ScheduledExecutorService} for scheduling warning messages
     */
    ApplicationContextDependencyMonitor(ScheduledExecutorService scheduledExecutorService) {
        this.scheduledExecutorService = scheduledExecutorService;
    }

    /**
     * {@inheritDoc}
     */
    public void onOsgiApplicationEvent(OsgiBundleApplicationContextEvent event) {
        synchronized (this.monitor) {
            if (event instanceof BootstrappingDependencyEvent) {
                ApplicationContext applicationContext = event.getApplicationContext();
                OsgiServiceDependencyEvent dependencyEvent = ((BootstrappingDependencyEvent) event).getDependencyEvent();
                final OsgiServiceDependency serviceDependency = dependencyEvent.getServiceDependency();
                if (dependencyEvent instanceof OsgiServiceDependencyWaitStartingEvent) {
                    addServiceDependencyTicker(applicationContext, serviceDependency, event.getBundle());
                } else if (dependencyEvent instanceof OsgiServiceDependencyWaitTimedOutEvent) {
                    removeServiceDependencyTicker(applicationContext, serviceDependency, event.getBundle(), false);
                } else if (dependencyEvent instanceof OsgiServiceDependencyWaitEndedEvent) {
                    removeServiceDependencyTicker(applicationContext, serviceDependency, event.getBundle(), true);
                }
            } else if (event instanceof OsgiBundleContextRefreshedEvent) {
                ApplicationContext applicationContext = event.getApplicationContext();
                removeAnyServiceDependencyTickers(applicationContext, event.getBundle());
                this.applicationContextBundle.remove(applicationContext);
            }
        }
    }

    /**
     * Cancel and remove any service dependency tickers associated with the given application context and given
     * associated bundle.
     * 
     * @param applicationContext the application context whose service tickers are to be canceled and removed.
     * @param bundle the {@link Bundle} associated with the given application context
     */
    private void removeAnyServiceDependencyTickers(ApplicationContext applicationContext, Bundle bundle) {
        Map<OsgiServiceDependency, Ticker> serviceDependencyTickers = getAnyServiceDependencyTickers(applicationContext);
        if (serviceDependencyTickers != null) {
            for (OsgiServiceDependency serviceDependency : serviceDependencyTickers.keySet()) {
                boolean hasTicked = serviceDependencyTickers.get(serviceDependency).cancel();
                serviceDependencyTickers.remove(serviceDependency);
                if (hasTicked && bundle.getState() != Bundle.UNINSTALLED) {
                    LOGGER.log(ConcurrentLogEvent.APPLICATION_CONTEXT_DEPENDENCY_SATISFIED, serviceDependency.getBeanName(),
                        bundle.getSymbolicName(), OsgiFrameworkUtils.getBundleVersion(bundle), serviceDependency.getServiceFilter());
                }
            }
            this.tickers.remove(applicationContext);
            this.applicationContextBundle.remove(applicationContext);
        }
    }

    /**
     * Remove any service dependency ticker for the given application context, given associated bundle, and given
     * service dependency.
     * 
     * @param applicationContext the partially constructed application context which needed the service dependency
     * @param serviceDependency the service dependency
     * @param bundle the {@link Bundle} associated with the given application context
     * @param <code>true</code> if the service dependency was satisfied and <code>false</code> if it timed out
     */
    private void removeServiceDependencyTicker(ApplicationContext applicationContext, OsgiServiceDependency serviceDependency, Bundle bundle,
        boolean satisfied) {
        TRACER.traceInfo("Service dependency '%s' %s", serviceDependency, satisfied ? "was satisfied" : "timed out");
        Map<OsgiServiceDependency, Ticker> serviceDependencyTickers = getAnyServiceDependencyTickers(applicationContext);
        if (serviceDependencyTickers != null) {
            Ticker ticker = serviceDependencyTickers.get(serviceDependency);
            if (ticker != null) {
                boolean hasTicked = ticker.cancel();
                serviceDependencyTickers.remove(serviceDependency);
                if (hasTicked) {
                    LOGGER.log(satisfied ? ConcurrentLogEvent.APPLICATION_CONTEXT_DEPENDENCY_SATISFIED
                        : serviceDependency.isMandatory() ? ConcurrentLogEvent.APPLICATION_CONTEXT_MANDATORY_DEPENDENCY_TIMED_OUT
                            : ConcurrentLogEvent.APPLICATION_CONTEXT_OPTIONAL_DEPENDENCY_TIMED_OUT, serviceDependency.getBeanName(),
                        bundle.getSymbolicName(), OsgiFrameworkUtils.getBundleVersion(bundle), serviceDependency.getServiceFilter());
                }
            }
        }
    }

    /**
     * Add a service dependency ticker for the given application context, given associated bundle, and given service
     * dependency.
     * 
     * @param applicationContext the partially constructed application context which needs the service dependency
     * @param serviceDependency the service dependency
     * @param bundle the {@link Bundle} associated with the given application context
     */
    private void addServiceDependencyTicker(ApplicationContext applicationContext, final OsgiServiceDependency serviceDependency, final Bundle bundle) {
        Map<OsgiServiceDependency, Ticker> serviceDependencyTickers = getServiceDependencyTickers(applicationContext);
        if (serviceDependencyTickers.containsKey(serviceDependency)) {
            TRACER.traceWarning("Service dependency '%s' already being waited upon", serviceDependency);
        } else {
            this.applicationContextBundle.put(applicationContext, bundle);
            // Services which are flagged as likely to be slow to be published are given a longer initial warning
            // interval.
            boolean slowService = serviceDependency.getServiceFilter().toString().contains("(com.springsource.server.slowservice=true)");
            serviceDependencyTickers.put(serviceDependency, StandardTicker.createExponentialTicker(slowService ? SLOW_WARNING_INTERVAL
                : INITIAL_WARNING_INTERVAL, WARNING_INTERVAL_INCREASE_RATE_PERCENT, slowService ? SLOW_WARNING_INTERVAL : MAXIMUM_WARNING_INTERVAL,
                new Callable<Void>() {

                    public Void call() throws Exception {
                        synchronized (ApplicationContextDependencyMonitor.this.monitor) {
                            if (bundle.getState() != Bundle.UNINSTALLED) {
                                ApplicationContextDependencyMonitor.LOGGER.log(
                                    serviceDependency.isMandatory() ? ConcurrentLogEvent.APPLICATION_CONTEXT_MANDATORY_DEPENDENCY_DELAYED
                                        : ConcurrentLogEvent.APPLICATION_CONTEXT_OPTIONAL_DEPENDENCY_DELAYED, serviceDependency.getBeanName(),
                                    bundle.getSymbolicName(), OsgiFrameworkUtils.getBundleVersion(bundle), serviceDependency.getServiceFilter());
                            } else {
                                Iterator<ApplicationContext> it = ApplicationContextDependencyMonitor.this.applicationContextBundle.keySet().iterator();
                                while (it.hasNext()) {
                                    ApplicationContext ac = it.next();
                                    if (ApplicationContextDependencyMonitor.this.applicationContextBundle.get(ac) == bundle) {
                                        removeAnyServiceDependencyTickers(ac, bundle);
                                        it.remove();
                                    }
                                }
                            }
                            return null;
                        }
                    }

                }, this.scheduledExecutorService));
        }
    }

    /**
     * Get the possibly empty map of service dependency tickers for the given application context.
     * 
     * @param applicationContext the application context whose service dependencies are required
     * @return a map of service dependency tickers
     */
    protected Map<OsgiServiceDependency, Ticker> getServiceDependencyTickers(ApplicationContext applicationContext) {
        Map<OsgiServiceDependency, Ticker> applicationContextTickers = getAnyServiceDependencyTickers(applicationContext);
        if (applicationContextTickers == null) {
            // Use a concurrent hash map to avoid ConcurrentModificationException while iterating and removing elements
            applicationContextTickers = new ConcurrentHashMap<OsgiServiceDependency, Ticker>();
            this.tickers.put(applicationContext, applicationContextTickers);
        }
        return applicationContextTickers;
    }

    /**
     * Get the possibly null map of service dependency tickers for the given application context.
     * 
     * @param applicationContext the application context whose service dependencies are required
     * @return a map of service dependency tickers or <code>null</code> if there is no such map
     */
    private Map<OsgiServiceDependency, Ticker> getAnyServiceDependencyTickers(ApplicationContext applicationContext) {
        return this.tickers.get(applicationContext);
    }

    /**
     * {@inheritDoc}
     */
    public void dump(SectionWriter sw) {
        synchronized (this.monitor) {
            ListWriter<SectionWriter> list = sw.list("application context service dependency monitoring");
            for (ApplicationContext applicationContext : this.tickers.keySet()) {
                MapWriter<ListWriter<SectionWriter>> appCtxWriter = list.map();
                appCtxWriter.scalar("application context", applicationContext.getClass().getName() + "@"
                    + Integer.toHexString(System.identityHashCode(applicationContext)));
                Bundle bundle = this.applicationContextBundle.get(applicationContext);
                appCtxWriter.scalar("bundle", bundle.getClass().getName() + "@" + Integer.toHexString(System.identityHashCode(bundle)) + " "
                    + bundle.getSymbolicName() + " version " + OsgiFrameworkUtils.getBundleVersion(bundle));
                ListWriter<MapWriter<ListWriter<SectionWriter>>> depWriter = appCtxWriter.list("service dependencies");
                Map<OsgiServiceDependency, Ticker> sdTickers = getAnyServiceDependencyTickers(applicationContext);
                if (sdTickers != null) {
                    for (OsgiServiceDependency osgiServiceDependency : sdTickers.keySet()) {
                        depWriter.scalar(osgiServiceDependency.getBeanName() + " " + osgiServiceDependency.getServiceFilter());
                    }
                }
                depWriter.end();
                appCtxWriter.end();
            }
            list.end();
        }
    }

}
