/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.concurrent.core.monitor;

import java.util.HashSet;
import java.util.Set;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;

import com.springsource.server.common.math.Sets;
import com.springsource.server.concurrent.core.diagnostics.ConcurrentLogEvent;
import com.springsource.server.concurrent.core.monitor.DeadlockAnalyser.Deadlock;
import com.springsource.server.serviceability.ffdc.FFDCService;
import com.springsource.server.serviceability.logging.LogService;
import com.springsource.server.serviceability.logging.LogServiceFactory;
import com.springsource.server.serviceability.output.Level;
import com.springsource.server.serviceability.tracing.EntryExitTraceLevel;

/**
 * Monitors all running {@link Thread Threads} and triggers a dump when a deadlock is detected.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
final class DeadlockMonitor {

    private static final int PERIOD = 10;

    private static final TimeUnit UNIT = TimeUnit.SECONDS;

    private final ScheduledExecutorService executorService;

    private final FFDCService ffdcService;

    private volatile ScheduledFuture<?> future;

    /**
     * Creates a new <code>DeadlockMonitor</code>.
     * 
     * @param executorService the <code>ScheduledExecutorService</code>
     * @param ffdcService the <code>FFDCService</code> to trigger a dump.
     */
    public DeadlockMonitor(ScheduledExecutorService executorService, FFDCService ffdcService) {
        this.executorService = executorService;
        this.ffdcService = ffdcService;
    }

    /**
     * Starts the deadlock monitor.
     */
    public void start() {
        this.future = this.executorService.scheduleAtFixedRate(new DeadlockMonitorTask(this.ffdcService), PERIOD, PERIOD, UNIT);
    }

    /**
     * Stops the deadlock monitor.
     */
    public void stop() {
        if (this.future != null) {
            this.future.cancel(true);
        }
    }

    /**
     * Task for monitoring threads for deadlocks.<p/>
     */
    private static class DeadlockMonitorTask implements Runnable {

        private final DeadlockAnalyser analyser = new DeadlockAnalyser();

        private final LogService log = LogServiceFactory.getLogService();

        private final Set<Deadlock> lastSeenDeadlocks = new HashSet<Deadlock>();

        private final Object monitor = new Object();

        private final FFDCService ffdcService;

        /**
         * Creates a new <code>DeadlockMonitorTask</code>.
         * 
         * @param ffdcService the <code>FFDCService</code> to use.
         */
        public DeadlockMonitorTask(FFDCService ffdcService) {
            this.ffdcService = ffdcService;
        }

        /**
         * {@inheritDoc}
         */
        @EntryExitTraceLevel(Level.DEBUG)
        public void run() {
            synchronized (this.monitor) {
                Deadlock[] deadlocks = this.analyser.findDeadlocks();
                if (deadlocks != null && deadlocks.length > 0) {
                    Set<Deadlock> asSet = Sets.asSet(deadlocks);
                    if (!asSet.equals(this.lastSeenDeadlocks)) {
                        this.log.log(ConcurrentLogEvent.DEADLOCK_DETECTED);
                        this.ffdcService.generateDump(null);
                        this.lastSeenDeadlocks.clear();
                        this.lastSeenDeadlocks.addAll(asSet);
                    }
                } else {
                    this.lastSeenDeadlocks.clear();
                }
            }
        }

    }
}
