/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.concurrent.core.queue;

import java.util.Collection;
import java.util.Iterator;
import java.util.concurrent.TimeUnit;

import com.springsource.server.serviceability.output.Level;
import com.springsource.server.serviceability.tracing.EntryExitTraceLevel;

import edu.emory.mathcs.backport.java.util.concurrent.BlockingDeque;

/**
 * Implementation of {@link com.springsource.server.concurrent.core.queue.BlockingDeque} that is backed by a <a
 * href="http://backport-jsr166.sourceforge.net/">backport-util-concurrent</a> {@link BlockingDeque}.<p/>
 * 
 * Subclasses <strong>must</strong> supply the <code>BlockingDeque</code> implementation during construction.
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe if the wrapped <code>BlockingDeque</code> is threadsafe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
/**
 * TODO Document BackportConcurrentBackedBlockingDeque
 * <p />
 *
 * <strong>Concurrent Semantics</strong><br />
 *
 * TODO Document concurrent semantics of BackportConcurrentBackedBlockingDeque
 *
 * @author Andy Wilkinson
 * @since TODO Document 'since'
 */
abstract class BackportConcurrentBackedBlockingDeque<E> implements com.springsource.server.concurrent.core.queue.BlockingDeque<E> {

    private final BlockingDeque delegate;

    /**
     * @param delegate
     */
    public BackportConcurrentBackedBlockingDeque(BlockingDeque delegate) {
        this.delegate = delegate;
    }

    /** 
     * {@inheritDoc}
     */
    public boolean add(E e) {
        return this.delegate.add(e);
    }
   
    /** 
     * {@inheritDoc}
     */
    @SuppressWarnings("unchecked") public boolean addAll(Collection<? extends E> c) {
        return this.delegate.addAll(c);
    }

    /** 
     * {@inheritDoc}
     */
    public void clear() {
        this.delegate.clear();
    }

    /** 
     * {@inheritDoc}
     */
    public boolean contains(Object o) {
        return this.delegate.contains(o);
    }

    /** 
     * {@inheritDoc}
     */
    @SuppressWarnings("unchecked") public boolean containsAll(Collection<?> c) {
        return this.delegate.containsAll(c);
    }

    /**
     * @return an iterator that descends the items on this queue
     */
    @SuppressWarnings("unchecked") public Iterator<E> descendingIterator() {
        return this.delegate.descendingIterator();
    }

    /** 
     * {@inheritDoc}
     */
    public int drainTo(Collection<? super E> c, int maxElements) {
        return this.delegate.drainTo(c, maxElements);
    }

    /** 
     * {@inheritDoc}
     */
    public int drainTo(Collection<? super E> c) {
        return this.delegate.drainTo(c);
    }

    /** 
     * {@inheritDoc}
     */
    @SuppressWarnings("unchecked") public E element() {
        return (E) this.delegate.element();
    }

    /**
     * {@inheritDoc}
     * 
     * @see java.util.Collection#equals(java.lang.Object)
     */
    @Override public boolean equals(Object o) {
        return this.delegate.equals(o);
    }

    /**
     * {@inheritDoc}
     * 
     * @see java.util.Collection#hashCode()
     */
    @Override public int hashCode() {
        return this.delegate.hashCode();
    }

    /**
     * {@inheritDoc}
     * 
     * @see java.util.Collection#isEmpty()
     */
    public boolean isEmpty() {
        return this.delegate.isEmpty();
    }

    /**
     * {@inheritDoc}
     * 
     * @see edu.emory.mathcs.backport.java.util.concurrent.BlockingDeque#iterator()
     */
    @SuppressWarnings("unchecked") public Iterator<E> iterator() {
        return this.delegate.iterator();
    }

    /**
     * @see edu.emory.mathcs.backport.java.util.concurrent.BlockingDeque#offer(java.lang.Object, long,
     *      edu.emory.mathcs.backport.java.util.concurrent.TimeUnit)
     */
    public boolean offer(E e, long timeout, TimeUnit unit) throws InterruptedException {
        return this.delegate.offer(e, timeout, edu.emory.mathcs.backport.java.util.concurrent.TimeUnit.valueOf(unit.name()));
    }

    /**
     * @see edu.emory.mathcs.backport.java.util.concurrent.BlockingDeque#offer(java.lang.Object)
     */
    public boolean offer(E e) {
        return this.delegate.offer(e);
    }

    /**
     * @see edu.emory.mathcs.backport.java.util.concurrent.BlockingDeque#peek()
     */
    @SuppressWarnings("unchecked") public E peek() {
        return (E) this.delegate.peek();
    }

    /**
     * @see edu.emory.mathcs.backport.java.util.concurrent.BlockingDeque#poll()
     */
    @EntryExitTraceLevel(Level.VERBOSE)
    @SuppressWarnings("unchecked") public E poll() {
        return (E) this.delegate.poll();
    }

    /**
     * @see edu.emory.mathcs.backport.java.util.concurrent.BlockingDeque#poll(long,
     *      edu.emory.mathcs.backport.java.util.concurrent.TimeUnit)
     */
    @EntryExitTraceLevel(Level.VERBOSE)
    @SuppressWarnings("unchecked") public E poll(long timeout, TimeUnit unit) throws InterruptedException {
        return (E) this.delegate.poll(timeout, edu.emory.mathcs.backport.java.util.concurrent.TimeUnit.valueOf(unit.name()));
    }

    /**
     * @see edu.emory.mathcs.backport.java.util.Deque#pollLast()
     */
    @EntryExitTraceLevel(Level.VERBOSE)
    @SuppressWarnings("unchecked") public E pollLast() {
        return (E) this.delegate.pollLast();
    }

    /**
     * @see edu.emory.mathcs.backport.java.util.Deque#pop()
     */
    @SuppressWarnings("unchecked") public E pop() {
        return (E) this.delegate.pop();
    }

    /**
     * @see edu.emory.mathcs.backport.java.util.concurrent.BlockingDeque#push(java.lang.Object)
     */
    public void push(E e) {
        this.delegate.push(e);
    }

    /**
     * @see edu.emory.mathcs.backport.java.util.concurrent.BlockingDeque#put(java.lang.Object)
     */
    public void put(E e) throws InterruptedException {
        this.delegate.put(e);
    }

    /**
     * @see edu.emory.mathcs.backport.java.util.concurrent.BlockingQueue#remainingCapacity()
     */
    public int remainingCapacity() {
        return this.delegate.remainingCapacity();
    }

    /**
     * @see edu.emory.mathcs.backport.java.util.concurrent.BlockingDeque#remove()
     */
    @SuppressWarnings("unchecked") public E remove() {
        return (E) this.delegate.remove();
    }

    /**
     * @see edu.emory.mathcs.backport.java.util.concurrent.BlockingDeque#remove(java.lang.Object)
     */
    public boolean remove(Object o) {
        return this.delegate.remove(o);
    }

    /**
     * @see java.util.Collection#removeAll(java.util.Collection)
     */
    @SuppressWarnings("unchecked") public boolean removeAll(Collection c) {
        return this.delegate.removeAll(c);
    }

    /**
     * @see edu.emory.mathcs.backport.java.util.concurrent.BlockingDeque#size()
     */
    public int size() {
        return this.delegate.size();
    }

    /**
     * @see edu.emory.mathcs.backport.java.util.concurrent.BlockingDeque#take()
     */
    @SuppressWarnings("unchecked") public E take() throws InterruptedException {
        return (E) this.delegate.take();
    }

    /**
     * {@inheritDoc}
     */
    public Object[] toArray() {
        return this.delegate.toArray();
    }

    /**
     * {@inheritDoc}
     */
    @SuppressWarnings("unchecked") public <T> T[] toArray(T[] a) {
        return (T[]) this.delegate.toArray(a);
    }

    /**
     * {@inheritDoc}
     */
    @SuppressWarnings("unchecked") public boolean retainAll(Collection<?> c) {
        return this.delegate.retainAll(c);
    }

}
