/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.concurrent.core.queue;

import java.util.concurrent.BlockingQueue;

import org.osgi.framework.BundleContext;
import org.osgi.framework.ServiceRegistration;

import com.springsource.server.concurrent.core.ConcurrentWeakServiceRegistrationTracker;
import com.springsource.server.management.proxy.concurrent.WorkQueueInfo;

/**
 * Standard implementation of {@link WorkQueueFactory}. <p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
final class StandardWorkQueueFactory implements WorkQueueFactory {

    /**
     * Tracks all known queues.
     */
    private final ConcurrentWeakServiceRegistrationTracker<WorkStealingQueue, Boolean> knownQueues = new ConcurrentWeakServiceRegistrationTracker<WorkStealingQueue, Boolean>();

    private final BundleContext bundleContext;

    public StandardWorkQueueFactory(BundleContext bundleContext) {
        this.bundleContext = bundleContext;
    }

    /**
     * {@inheritDoc}
     */
    public BlockingQueue<Runnable> newStealingWorkQueue() {
        BlockingDeque<Runnable> queue = new LinkedBlockingDeque<Runnable>(Integer.MAX_VALUE);
        return newStealingQueue(queue, "Blocking Queue");
    }

    /**
     * {@inheritDoc}
     */
    public BlockingQueue<Runnable> newStealingWorkQueue(int capacity) {
        BlockingDeque<Runnable> queue = new LinkedBlockingDeque<Runnable>(capacity);
        return newStealingQueue(queue, "Blocking Queue with set capacity");
    }

    /**
     * Creates a new {@link WorkStealingQueue} backed by the supplied delegate.
     * 
     * @param delegate the backing {@link BlockingDeque}.
     * @return
     */
    private WorkStealingQueue newStealingQueue(BlockingDeque<Runnable> delegate, String type) {
        WorkStealingQueue queue = new WorkStealingQueue(delegate, this.knownQueues.recordedKeys());
        ServiceRegistration sr = this.bundleContext.registerService(WorkQueueInfo.class.getName(), new StandardWorkQueueInfo(queue, type), null);
        this.knownQueues.record(queue, Boolean.TRUE, sr);
        return queue;
    }

}
