/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.config.core;

import java.util.Set;

import com.springsource.server.serviceability.ffdc.FatalServerException;

/**
 * A ConfigTree is an abstract syntax tree representation of a configuration file.
 * 
 * Implementations of this interface must be thread-safe.
 * 
 * @author Glyn Normington
 * @author Andy Wilkinson
 * @author Sam Brannen
 * @since 1.0
 */
public interface ConfigTree {

    /**
     * Get the names of the configuration points defined by this config. tree.
     * 
     * @return a set of config. point names
     */
    Set<String> getConfigPointNames();

    /**
     * Get a configuration point with the given name or return null if there is no such config. point in this config.
     * tree.
     * 
     * @param configPointName the config. point required
     * @return a config. point or null if the config. point was not found
     */
    ConfigPoint getConfigPoint(String configPointName);

    /**
     * Validate the contents of this config. tree.
     * 
     * @return <code>true</code> if the config. tree is valid or if the config. tree does not have a schema
     */
    boolean validate();

    /**
     * Navigate the config. tree and return the integer located using the given navigation path. A navigation path is a
     * string beginning with the config. point name followed by a colon and then a comma separated list of configuration
     * keys, for example: "serverSettings:tcp,hostname".
     * 
     * @param navigationPath the path of the required integer
     * @return the integer located at the given path
     * @throws FatalServerException if the path is not the path of an integer in the config. tree
     */
    int getIntValue(String navigationPath);

    /**
     * Navigate the config. tree and return the integer located using the given navigation path. This method has the
     * same functionality of {@link #getIntValue(String)} except that the caller is not required to have any knowledge
     * of the navigation syntax.
     * 
     * @param configPointName the config. point required
     * @param pathComponents the navigation components that define the path of the required integer
     * @return the integer located at the given path
     * @throws FatalServerException if the path is not the path of an integer in the config. tree
     */
    int getIntValue(String configPointName, String... pathComponents);

    /**
     * Navigate the config. tree and return the string located using the given navigation path. A navigation path is a
     * string beginning with the config. point name followed by a colon and then a comma separated list of configuration
     * keys, for example: "serverSettings:tcp,port".
     * 
     * @param navigationPath the path of the required string
     * @return the string located at the given path
     * @throws FatalServerException if the path is not the path of a string in the config. tree
     */
    String getStringValue(String navigationPath);

    /**
     * Navigate the config. tree and return the string located using the given navigation path. This method has the same
     * functionality of {@link #getStringValue(String)} except that the caller is not required to have any knowledge of
     * the navigation syntax.
     * 
     * @param configPointName the config. point required
     * @param pathComponents the navigation components that define the path of the required String
     * @return the string located at the given path
     * @throws FatalServerException if the path is not the path of a string in the config. tree
     */
    String getStringValue(String configPointName, String... pathComponents);

    /**
     * Navigate the config. tree and return the {@link Boolean} located using the given navigation path. A navigation
     * path is a string beginning with the config. point name followed by a colon and then a comma separated list of
     * configuration keys, for example: "serverSettings:tcp,port".
     * 
     * @param navigationPath the path of the required string
     * @return the string located at the given path
     * @throws FatalServerException if the path is not the path of a string in the config. tree
     */
    Boolean getBooleanValue(String navigationPath);

    /**
     * Navigate the config. tree and return the {@link Boolean} located using the given navigation path. This method has
     * the same functionality of {@link #getStringValue(String)} except that the caller is not required to have any
     * knowledge of the navigation syntax.
     * 
     * @param configPointName the config. point required
     * @param pathComponents the navigation components that define the path of the required String
     * @return the string located at the given path
     * @throws FatalServerException if the path is not the path of a string in the config. tree
     */
    Boolean getBooleanValue(String configPointName, String... pathComponents);

}
