/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.config.core.internal;

import java.util.ArrayList;
import java.util.List;

import com.springsource.server.config.core.ConfigNode;
import com.springsource.server.config.core.ConfigPoint;
import com.springsource.server.config.core.ConfigTree;
import com.springsource.server.config.core.FatalConfigurationException;
import com.springsource.server.config.core.ScalarConfigNode;
import com.springsource.server.serviceability.ffdc.FatalServerException;
import com.springsource.server.serviceability.ffdc.NonNull;

/**
 * ConfigNavigator provides methods for navigating a ConfigTree and returning values from within the tree.
 * 
 * <p>
 * This class is thread-safe (as a result of ConfigTree being thread-safe)
 * </p>
 * 
 * @author Andy Wilkinson
 * @author Sam Brannen
 * @since 1.0
 */
final class ConfigNavigator {

    private final ConfigTree configTree;

    /**
     * Creates a new ConfigNavigator for navigation of the supplied ConfigTree
     * 
     * @param tree The tree to navigate
     */
    public ConfigNavigator(ConfigTree tree) {
        this.configTree = tree;
    }

    /**
     * Navigates the tree and returns, as a String, the value located using the supplied navigation path.
     * 
     * @param configPointNavigationPath The path with which the String is located
     * @return The String located using the supplied path
     * @throws FatalServerException if the navigation path is malformed, does not identify a node, or the node it
     *         identifies is not a ScalarConfigNode
     */
    public String getString(@NonNull String configPointNavigationPath) throws FatalServerException {
        return getString(configPointNavigationPath, getNode(configPointNavigationPath));
    }

    /**
     * Navigates the tree and returns, as a String, the value located using the supplied navigation path. This method
     * has the same functionality as {@link #getString(String)} except that the caller is not required to have explicit
     * knowledge of the navigation syntax.
     * 
     * @param configPointName the config. point required
     * @param pathComponents the navigation components that define the path of the required String
     * @return The String located using the supplied path
     * @throws FatalServerException if the navigation path is malformed, does not identify a node, or the node it
     *         identifies is not a ScalarConfigNode
     */
    public String getString(@NonNull String configPointName, @NonNull String... pathComponents) {
        return getString(buildConfigPointNavigationPath(configPointName, pathComponents));
    }

    private String getString(String configPointNavigationPath, ConfigNode node) {
        if (node instanceof ScalarConfigNode) {
            return ((ScalarConfigNode) node).getValue();
        } else {
            throw new FatalConfigurationException("The ConfigNode " + node + " is not of type ScalarConfigNode", null);
        }
    }

    /**
     * Navigates the tree and returns, as an int, the value located using the supplied navigation path.
     * 
     * @param configPointNavigationPath The path with which the int is located
     * @return The int located using the supplied path
     * @throws FatalServerException if the navigation path is malformed, does not identify a node, or the node it
     *         identifies is not a ScalarConfigNode containing an int value
     */
    public int getInt(@NonNull String configPointNavigationPath) throws FatalServerException {
        return getInt(configPointNavigationPath, getNode(configPointNavigationPath));
    }

    Boolean getBoolean(@NonNull String navigationPath) throws FatalServerException {
        return getBoolean(navigationPath, getNode(navigationPath));
    }
    
    Boolean getBoolean(@NonNull String configPointName, @NonNull String... pathComponents) {
        return getBoolean(buildConfigPointNavigationPath(configPointName, pathComponents));
    }

    private Boolean getBoolean(String configPointNavigationPath, ConfigNode node) {
        if (node instanceof ScalarConfigNode) {
            return ((ScalarConfigNode) node).getValue();
        } else {
            throw new FatalConfigurationException("The ConfigNode " + node + " is not of type ScalarConfigNode", null);
        }
    }

    /**
     * Navigates the tree and returns, as an int, the value located using the supplied navigation path. This method has
     * the same functionality as {@link #getInt(String)} except that the caller is not required to have explicit
     * knowledge of the navigation syntax.
     * 
     * @param configPointName the config. point required
     * @param pathComponents the navigation components that define the path of the required integer
     * @return the integer located at the given path
     * @throws FatalServerException if the navigation path is malformed, does not identify a node, or the node it
     *         identifies is not a ScalarConfigNode containing an int value
     */
    public int getInt(@NonNull String configPointName, @NonNull String... pathComponents) {
        return getInt(buildConfigPointNavigationPath(configPointName, pathComponents));
    }

    private int getInt(String configPointNavigationPath, ConfigNode node) {
        if (node instanceof ScalarConfigNode) {
            final String stringValue = ((ScalarConfigNode) node).getValue();
            try {
                return Integer.parseInt(stringValue);
            } catch (final NumberFormatException nfe) {
                throw new FatalConfigurationException("The value '" + stringValue + "' at '" + configPointNavigationPath
                    + "' could not be parsed to an int", nfe);
            }
        } else {
            throw new FatalConfigurationException("The ConfigNode " + node + " is not of type ScalarConfigNode", null);
        }
    }

    private String buildConfigPointNavigationPath(String configPointName, String... pathComponents) {
        StringBuilder navigationPath = new StringBuilder(configPointName).append(":");
        for (int i = 0; i < pathComponents.length; i++) {
            navigationPath.append(pathComponents[i]);
            if (i < pathComponents.length - 1) {
                navigationPath.append(",");
            }
        }
        return navigationPath.toString();
    }

    private ConfigNode getNode(String configPointNavigationPath) throws FatalServerException {

        int colonIndex = configPointNavigationPath.indexOf(':');
        if (colonIndex == -1) {
            throw new FatalConfigurationException("The navigation path must begin with 'configPointName:' to identify a ConfigPoint in the tree");
        }

        String configPointName = configPointNavigationPath.substring(0, colonIndex);
        String navigationPath = configPointNavigationPath.substring(colonIndex + 1);
        String[] pathComponents = navigationPath.split(",");

        ConfigPoint configPoint = this.configTree.getConfigPoint(configPointName);
        if (configPoint == null) {
            throw new FatalConfigurationException("The config point '" + configPointName + "' identified by navigation path '"
                + configPointNavigationPath + "' does not exist");
        }

        List<String> path = new ArrayList<String>();
        for (final String eachComponent : pathComponents) {
            path.add(eachComponent.trim());
        }

        ConfigNode node = configPoint.getNode(path);
        if (node == null) {
            throw new FatalConfigurationException("The node '" + pathComponents + "' under ConfigPoint '" + configPointName
                + "' identified by navigation path '" + configPointNavigationPath + "' does not exist");
        }
        return node;
    }

}
