/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.config.core.internal;

import java.util.List;
import java.util.Set;

import com.springsource.json.parser.ListNode;
import com.springsource.json.parser.MapNode;
import com.springsource.json.parser.Node;
import com.springsource.json.parser.ScalarNode;
import com.springsource.server.config.core.ConfigNode;
import com.springsource.server.config.core.ConfigPoint;
import com.springsource.server.config.core.FatalConfigurationException;

/**
 * Factory for creating {@link ConfigNode} instances from a {@link MapNode}<p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
final class JsonConfigNodeFactory {

    /**
     * Creates a {@link ConfigNode} that maps to the supplied JSON {@link MapNode}.
     * 
     * @param parent the parent of the <code>ConfigNode</code>.
     * @param rootNode the JSON <code>MapNode<?/code>.
     * @return the created <code>ConfigNode</code>.
     */
    public static ConfigNode createConfigNode(ConfigPoint parent, MapNode rootNode) {
        return createConfigNodeInternal(parent, rootNode);
    }

    /**
     * Creates the correct {@link ConfigNode} for supplied JSON {@link Node}.
     * 
     * @param parent the parent <code>ConfigNode</code>.
     * @param node the JSON <code>Node</code>.
     * @return the created <code>ConfigNode</code>.
     */
    private static ConfigNode createConfigNodeInternal(ConfigNode parent, Node node) {
        if (node instanceof MapNode) {
            return createMapNode(parent, (MapNode) node);
        } else if (node instanceof ListNode) {
            return createListNode(parent, (ListNode) node);
        } else if (node instanceof ScalarNode) {
            return createScalarNode(parent, (ScalarNode) node);
        } else {
            throw new FatalConfigurationException("Unknown node type '" + node.getClass().getName() + "'");
        }
    }

    /**
     * Creates a {@link StandardListConfigNode} for the supplied {@link ListNode}.
     */
    private static ConfigNode createListNode(ConfigNode parent, ListNode node) {
        StandardListConfigNode configNode = new StandardListConfigNode(parent);
        List<Node> nodes = node.getNodes();
        for (Node child : nodes) {
            configNode.addNode(createConfigNodeInternal(configNode, child));
        }
        return configNode;
    }

    /**
     * Creates a {@link StandardScalarConfigNode} for the supplied {@link ScalarNode}.
     */
    private static ConfigNode createScalarNode(ConfigNode parent, ScalarNode node) {
        return new StandardScalarConfigNode(parent, node.getValue());
    }

    /**
     * Creates a {@link StandardMapConfigNode} for the supplied {@link MapNode}.
     */
    private static ConfigNode createMapNode(ConfigNode parent, MapNode node) {
        StandardMapConfigNode configNode = new StandardMapConfigNode(parent);
        Set<String> keys = node.getKeys();
        for (String key : keys) {
            Node child = node.getNode(key);
            configNode.setNode(key, createConfigNodeInternal(configNode, child));
        }
        return configNode;
    }
}
