/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.config.core.internal;

import java.net.URL;
import java.util.Collections;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;

import org.osgi.framework.BundleContext;

import com.springsource.json.parser.AntlrJSONParser;
import com.springsource.json.parser.JSONParseException;
import com.springsource.json.parser.JSONParser;
import com.springsource.json.parser.MapNode;
import com.springsource.json.parser.Node;
import com.springsource.server.config.core.ConfigPoint;
import com.springsource.server.config.core.ConfigTree;
import com.springsource.server.config.core.FatalConfigurationException;
import com.springsource.server.config.core.internal.validate.ConfigValidator;
import com.springsource.server.management.proxy.configuration.ConfigPointInfo;
import com.springsource.server.serviceability.ffdc.FatalServerException;
import com.springsource.server.serviceability.ffdc.NonNull;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * JSON-based implementation of {@link ConfigTree}.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @author Sam Brannen
 * @since 1.0
 */
public final class JsonConfigTree implements ConfigTree {

    private static final Tracer TRACER = TracerFactory.createTracer(JsonConfigTree.class);

    private final ConcurrentMap<String, ConfigPoint> configPoints = new ConcurrentHashMap<String, ConfigPoint>();

    private final ConcurrentMap<String, ConfigValidator> configPointValidators = new ConcurrentHashMap<String, ConfigValidator>();

    private final ConfigNavigator navigator;

    private final String fileName;

    /**
     * Creates a new <code>JsonConfigTree</code> from the JSON data at the supplied {@link URL}.
     * 
     * @param url the <code>URL</code> of the JSON data.
     */
    public JsonConfigTree(URL url) {
        this.fileName = url.getFile();
        JSONParser parser = new AntlrJSONParser(new ServerParserLogger());
        try {
            Node rootNode = parser.parse(url);
            if (!(rootNode instanceof MapNode)) {
                throw new FatalConfigurationException("Root config node is required to be a JSON object.");
            }
            parseConfigPoints((MapNode) rootNode);
        } catch (JSONParseException e) {
            throw new FatalConfigurationException("The syntax of config file '" + url + "' is invalid.", e);
        }

        this.navigator = new ConfigNavigator(this);
    }

    private void parseConfigPoints(MapNode rootNode) {
        Set<String> keys = rootNode.getKeys();
        for (String key : keys) {
            Node configPointNode = rootNode.getNode(key);
            if (configPointNode instanceof MapNode) {
                this.configPoints.put(key, new JsonConfigPoint(key, (MapNode) configPointNode));
            } else {
                throw new FatalConfigurationException("Config point '" + key + "' is not a valid JSON map node.");
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    public Set<String> getConfigPointNames() {
        return Collections.unmodifiableSet(this.configPoints.keySet());
    }

    /**
     * {@inheritDoc}
     */
    public ConfigPoint getConfigPoint(@NonNull String configPointName) {
        return this.configPoints.get(configPointName);
    }

    /**
     * {@inheritDoc}
     */
    public int getIntValue(@NonNull String navigationPath) throws FatalServerException {
        return this.navigator.getInt(navigationPath);
    }

    /**
     * {@inheritDoc}
     */
    public int getIntValue(@NonNull String configPointName, @NonNull String... pathComponents) {
        return this.navigator.getInt(configPointName, pathComponents);
    }

    /**
     * {@inheritDoc}
     */
    public String getStringValue(@NonNull String navigationPath) throws FatalServerException {
        return this.navigator.getString(navigationPath);
    }

    /**
     * {@inheritDoc}
     */
    public String getStringValue(@NonNull String configPointName, @NonNull String... pathComponents) {
        return this.navigator.getString(configPointName, pathComponents);
    }

    /**
     * {@inheritDoc}
     */
    public Boolean getBooleanValue(@NonNull String navigationPath) throws FatalServerException {
        return this.navigator.getBoolean(navigationPath);
    }

    /**
     * {@inheritDoc}
     */
    public Boolean getBooleanValue(@NonNull String configPointName, @NonNull String... pathComponents) {
        return this.navigator.getBoolean(configPointName, pathComponents);
    }

    /**
     * {@inheritDoc}
     */
    public boolean validate() {
        boolean validity = true;

        for (String cpn : this.configPoints.keySet()) {
            if (this.configPointValidators.containsKey(cpn)) {
                validity = validity && this.configPointValidators.get(cpn).validate(this.configPoints.get(cpn).getRootNode());
            } else {
                TRACER.traceWarning("Missing configuration point schema '%s'", cpn);
            }
        }

        return validity;
    }

    /**
     * Sets the {@link ConfigValidator} for the {@link ConfigPoint} identified by <code>configPointName</code>.
     * 
     * @param configPointName the name of the <code>ConfigPoint</code>.
     * @param cpValidator the validator.
     */
    void setConfigPointValidator(String configPointName, ConfigValidator cpValidator) {
        if (this.configPointValidators.putIfAbsent(configPointName, cpValidator) != null) {
            throw new FatalConfigurationException("Config. point validator already set", null);
        }
    }

    /**
     * Publishes the {@link ConfigPointInfo} for this <code>JsonConfigTree</code>.
     * 
     * @param configPointName the config point name,
     * @param context the {@link BundleContext}.
     */
    void registerInfoObject(String configPointName, BundleContext context) {
        ConfigPointInfo configPointInfo = new StandardConfigPointInfo(this.configPoints.get(configPointName), this.fileName);
        context.registerService(ConfigPointInfo.class.getName(), configPointInfo, null);
    }
}
