/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.config.core.internal;

import com.springsource.json.parser.JSONParserLogger;
import com.springsource.json.parser.ParserMessage;
import com.springsource.server.config.core.diagnostics.ConfigLogEvent;
import com.springsource.server.serviceability.logging.LogEvent;
import com.springsource.server.serviceability.logging.LogService;
import com.springsource.server.serviceability.logging.LogServiceFactory;
import com.springsource.server.serviceability.output.Level;

/**
 * {@link JSONParserLogger} that directs syntax errors to the server log. <p/>
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @see LogService
 * @since 1.0
 */
final class ServerParserLogger implements JSONParserLogger {

    private static final int LOG_CODE_OFFSET = 1000;

    private static final LogService LOG = LogServiceFactory.getLogService();

    /**
     * {@inheritDoc}
     */
    public void error(ParserMessage message, int line, int col, Object... inserts) {
        error(message, line, col, null, inserts);
    }

    /**
     * {@inheritDoc}
     */
    public void error(ParserMessage message, int line, int col, Throwable throwable, Object... inserts) {
        LOG.log(translate(message, Level.ERROR), throwable, packageInserts(line, col, inserts));
    }

    /**
     * {@inheritDoc}
     */
    public void info(ParserMessage message, int line, int col, Object... inserts) {
        LOG.log(translate(message, Level.INFO), packageInserts(line, col, inserts));
    }

    /**
     * {@inheritDoc}
     */
    public void warn(ParserMessage message, int line, int col, Object... inserts) {
        LOG.log(translate(message, Level.WARN), packageInserts(line, col, inserts));
    }

    /**
     * Package the inserts for a {@link LogEvent}. The line and column number appear as the first two inserts, with the
     * inserts supplied by the JSON parser coming after that.
     */
    private Object[] packageInserts(int line, int col, Object[] inserts) {
        int size = 2 + (inserts == null ? 0 : inserts.length);
        Object[] result = new Object[size];
        result[0] = line;
        result[1] = col;
        if (inserts != null) {
            for (int x = 0; x < inserts.length; x++) {
                result[x + 2] = inserts[x];
            }
        }
        return result;
    }

    /**
     * Translates a {@link ParserMessage} into a server {@link LogEvent}.
     */
    private LogEvent translate(ParserMessage message, Level level) {
        return ConfigLogEvent.createDynamic(message.getCode() + LOG_CODE_OFFSET, level);
    }
}
