/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.config.core.internal;

import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.springsource.server.config.core.ConfigNode;
import com.springsource.server.config.core.FatalConfigurationException;
import com.springsource.server.config.core.MapConfigNode;

/**
 * Standard implementation of {@link MapConfigNode}. <p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Glyn Normington
 * @since 1.0
 */
public final class StandardMapConfigNode extends AbstractConfigNode implements MapConfigNode {

    private final Map<String, ConfigNode> map = Collections.synchronizedMap(new LinkedHashMap<String, ConfigNode>());

    /**
     * Creates a new <code>StandardMapConfigNode</code>.
     * 
     * @param parent the parent {@link ConfigNode}.
     */
    public StandardMapConfigNode(ConfigNode parent) {
        super(parent);
    }

    /**
     * {@inheritDoc}
     */
    public ConfigNode getNode(String key) {
        return this.map.get(key);
    }

    public void setNode(String key, ConfigNode value) {
        if (value.getParent() != this) {
            throw new FatalConfigurationException("Cannot set node of a map node which is not a child of the map node", null);
        }
        this.map.put(key, value);
    }

    /**
     * {@inheritDoc}
     */
    public ConfigNode getNode(List<String> path) {
        ConfigNode result = null;
        if (!path.isEmpty()) {
            String head = head(path);
            if (this.map.containsKey(head)) {
                result = this.map.get(head).getNode(tail(path));
            }
        } else {
            result = this;
        }
        return result;
    }

    /**
     * {@inheritDoc}
     */
    public boolean addNode(List<String> path, ConfigNode newNode) {
        boolean result = false;
        if (!path.isEmpty()) {
            String head = head(path);
            List<String> tail = tail(path);
            if (tail.isEmpty()) {
                if (!this.map.containsKey(head)) {
                    this.map.put(head, newNode);
                    result = true;
                }
            } else {
                if (!this.map.containsKey(head)) {
                    this.map.put(head, new StandardMapConfigNode(this));
                }
                if (this.map.get(head) instanceof MapConfigNode) {
                    MapConfigNode child = (MapConfigNode) this.map.get(head);
                    result = child.addNode(tail, newNode);
                }
            }
        }
        return result;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String toString() {
        StringBuffer result = new StringBuffer();
        result.append('{');
        boolean first = true;
        for (String k : this.map.keySet()) {
            if (first == false) {
                result.append(", ");
            }
            result.append(k);
            result.append(" : ");
            result.append(this.map.get(k).toString());
            first = false;
        }
        result.append('}');
        return result.toString();
    }

    /**
     * {@inheritDoc}
     */
    public Set<String> getKeys() {
        return Collections.unmodifiableSet(this.map.keySet());
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected String doGetPath(ConfigNode childNode) {
        StringBuffer path = new StringBuffer();
        path.append(getPath());
        if (getParent().getParent() != null) {
            path.append(',');
        }
        for (String k : this.map.keySet()) {
            if (this.map.get(k) == childNode) {
                path.append(k);
                return path.toString();
            }
        }
        throw new FatalConfigurationException("Child node not found in parent map", null);
    }
}
