/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.config.core.internal.validate;

import com.springsource.server.config.core.ConfigNode;
import com.springsource.server.config.core.ListConfigNode;
import com.springsource.server.config.core.diagnostics.ConfigLogEvent;
import com.springsource.server.serviceability.logging.LogService;
import com.springsource.server.serviceability.logging.LogServiceFactory;

/**
 * {@link StandardListConfigNodeValidator} is a standard implementation of {@link ConfigValidator} for lists.
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * This class is thread safe provided the list element validator passed on the constructor is also thread safe.
 * 
 * @author Glyn Normington
 * @since 1.0
 */
final class StandardListConfigNodeValidator implements ConfigValidator {

    private static final LogService LOGGER = LogServiceFactory.getLogService();

    private final ConfigValidator listElementValidator;

    /**
     * Create a list validator which uses the given validator to validate each of the elements of the list.
     *  
     * @param listElementValidator the validator to be applied to list elements
     */
    public StandardListConfigNodeValidator(ConfigValidator listElementValidator) {
        this.listElementValidator = listElementValidator;
    }

    /**
     * {@inheritDoc}
     */
    public boolean validate(ConfigNode configNode) {
        boolean result = true;
        try {
            for (ConfigNode n : ((ListConfigNode) configNode).getNodes()) {
                if (!this.listElementValidator.validate(n)) {
                    result = false; // but continue validating any remaining elements
                }
            }
        } catch (ClassCastException e) {
            LOGGER.log(ConfigLogEvent.INVALID_LIST, configNode.getPath());
            result = false;
        }
        return result;
    }
}
