/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.config.core.internal.validate;

import com.springsource.server.common.math.Range;
import com.springsource.server.common.math.Version;
import com.springsource.server.config.core.ConfigNode;
import com.springsource.server.config.core.ScalarConfigNode;
import com.springsource.server.config.core.diagnostics.ConfigLogEvent;
import com.springsource.server.serviceability.logging.LogService;
import com.springsource.server.serviceability.logging.LogServiceFactory;

/**
 * {@link StandardVersionRangeConfigNodeValidator} is a standard implementation of {@link StandardScalarConfigNodeValidator} for version ranges.
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * This class is thread safe.
 * 
 * @author Glyn Normington
 * @since 1.0
 */
public final class StandardVersionRangeConfigNodeValidator extends StandardScalarConfigNodeValidator {

    private static final LogService LOGGER = LogServiceFactory.getLogService();

    private final Range<Version> versionRange;

    /**
     * Create a validator for the given version range.
     * 
     * @param versionRange the version range to be validated against
     */
    public StandardVersionRangeConfigNodeValidator(Range<Version> versionRange) {
        this.versionRange = versionRange;
    }

    /** 
     * {@inheritDoc}
     */
    @Override public boolean validate(ConfigNode configNode) {
        ScalarConfigNode scalarConfigNode;
        try {
            scalarConfigNode = (ScalarConfigNode) configNode;
        } catch (ClassCastException e) {
            LOGGER.log(ConfigLogEvent.INVALID_SCALAR, configNode.getPath());
            return false;
        }
        try {
            if (!this.versionRange.contains(new Version((String) scalarConfigNode.getValue()))) {
                LOGGER.log(ConfigLogEvent.VERSION_OUT_OF_RANGE, scalarConfigNode.getPath(), scalarConfigNode.getValue());
                return false;
            }
        } catch (NumberFormatException e) {
            LOGGER.log(ConfigLogEvent.INVALID_VERSION, scalarConfigNode.getPath(), scalarConfigNode.getValue());
            return false;
        }
        return true;
    }

    public Range<Version> getVersionRange() {
        return this.versionRange;
    }

}
