/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.control.core.internal;

import static com.springsource.server.control.core.internal.Constants.CONTROL_EXPORT_DOMAIN;
import static com.springsource.server.control.core.internal.Constants.CONTROL_SERVICE_FILTER_STRING;
import static com.springsource.server.control.core.internal.Constants.CONTROL_SERVICE_PROPERTY_KEY_PREFIX;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import javax.management.JMException;
import javax.management.MBeanException;
import javax.management.MBeanServer;
import javax.management.ObjectName;
import javax.management.OperationsException;
import javax.management.modelmbean.InvalidTargetObjectTypeException;
import javax.management.modelmbean.ModelMBean;
import javax.management.modelmbean.ModelMBeanInfo;
import javax.management.modelmbean.RequiredModelMBean;

import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.Constants;
import org.osgi.framework.InvalidSyntaxException;
import org.osgi.framework.ServiceEvent;
import org.osgi.framework.ServiceListener;
import org.osgi.framework.ServiceReference;

import com.springsource.server.control.core.diagnostics.ControlLogEvent;
import com.springsource.server.control.core.diagnostics.FatalControlException;
import com.springsource.server.serviceability.logging.LogService;
import com.springsource.server.serviceability.logging.LogServiceFactory;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * Imports controls that have been published to the OSGi service registry
 * and registers them as MBeans with the running MBean server.
 *
 * <strong>Concurrent Semantics</strong><br />
 * Thread-safe.
 *
 * @author Andy Wilkinson
 * @since 1.0
 */
public class MBeanRegisteringControlImporter {
    
    private static final LogService LOG = LogServiceFactory.getLogService();
    private static final Tracer TRACE = TracerFactory.createTracer(MBeanRegisteringControlImporter.class);
        
    private final BundleContext bundleContext;
    private final ModelMBeanInfoAssembler modelMBeanInfoAssembler;
    private final MBeanServer mBeanServer;   
    private final ServiceListener controlServiceListener;
    private final Map<ServiceReference, List<ObjectName>> registeredMBeans = new ConcurrentHashMap<ServiceReference, List<ObjectName>>();
    
    public MBeanRegisteringControlImporter(BundleContext bundleContext, ModelMBeanInfoAssembler modelMBeanInfoAssembler, MBeanServerLocator mBeanServerLocator) {
        this.bundleContext = bundleContext;
        this.modelMBeanInfoAssembler = modelMBeanInfoAssembler;   
        this.mBeanServer = mBeanServerLocator.locateDefaultMBeanServer();
        
        try {  
            this.controlServiceListener = new ControlServiceListener();
            bundleContext.addServiceListener(controlServiceListener, CONTROL_SERVICE_FILTER_STRING);
            ServiceReference[] serviceReferences = bundleContext.getServiceReferences(null, CONTROL_SERVICE_FILTER_STRING);
            if (serviceReferences != null) {
                for (ServiceReference serviceReference: serviceReferences) {
                    controlExported(serviceReference);
                }
            }
        } catch (InvalidSyntaxException ise) {
            throw new FatalControlException("Unable to filter service registrations", ise);
        }
    }
    
    public void shutdown() {
        if (controlServiceListener != null) {
            bundleContext.removeServiceListener(controlServiceListener);
        }       
    }
    
    private void controlExported(ServiceReference serviceReference) {
        List<Class<?>> registrationClasses = loadRegistrationClasses(serviceReference);
               
        String[] propertyKeys = serviceReference.getPropertyKeys();
        
        Hashtable<String, String> properties = new Hashtable<String, String>();
        
        for (String key : propertyKeys) {
            if (key.startsWith(CONTROL_SERVICE_PROPERTY_KEY_PREFIX)) {
                properties.put(key.substring(CONTROL_SERVICE_PROPERTY_KEY_PREFIX.length()), (String)serviceReference.getProperty(key));
            }
        }
        
        Object service = bundleContext.getService(serviceReference); 
        
        List<ObjectName> objectNames = new ArrayList<ObjectName>();
        
        for (Class<?> registrationClass : registrationClasses) {   
             try {
                 ObjectName mBeanObjectName = createAndRegisterMBean(registrationClass, service, properties);
                 objectNames.add(mBeanObjectName);
                 TRACE.traceInfo("MBean %s registered for service %s defined by %s", mBeanObjectName, service, registrationClass.getName());
             } catch (Exception e) {
                 LOG.log(ControlLogEvent.MBEAN_REGISTER_FAILURE, e, service, registrationClass); 
             } 
        }
        
        this.registeredMBeans.put(serviceReference, objectNames);
    }
    
    private List<Class<?>> loadRegistrationClasses(ServiceReference serviceReference) {
        String[] registrationClassNames = (String[])serviceReference.getProperty(Constants.OBJECTCLASS);            
        Bundle owningBundle = serviceReference.getBundle();
        
        List<Class<?>> registrationClasses = new ArrayList<Class<?>>();
        
        for (String registrationClassName : registrationClassNames) {
            try {
                Class<?> clazz = owningBundle.loadClass(registrationClassName);
                registrationClasses.add(clazz);
            } catch (ClassNotFoundException cnfe) {
                LOG.log(ControlLogEvent.CLASS_NOT_FOUND_LOADING_SERVICE_REGISTRATION_CLASS, registrationClassName, serviceReference, owningBundle.getSymbolicName());
            }
        }
        
        return registrationClasses;
    }
    
    private void controlRetracted(ServiceReference serviceReference) {                  
        List<ObjectName> mBeanNames = this.registeredMBeans.get(serviceReference);
    
        for (ObjectName mBeanName: mBeanNames) {
            try {
                TRACE.traceInfo("Unregistering MBean %s", mBeanName);
                this.mBeanServer.unregisterMBean(mBeanName);
            } catch (Exception e) {
                LOG.log(ControlLogEvent.MBEAN_UNREGISTER_FAILURE, e, mBeanName);
            }
        }
    }
    
    private ObjectName createAndRegisterMBean(Class<?> definition, Object service, Hashtable<String, String> properties) throws MBeanException, OperationsException, JMException, InvalidTargetObjectTypeException {
        ModelMBean modelMBean = createModelMBean(definition, service);
        modelMBean.setManagedResource(service, "ObjectReference");
                       
        return registerMBean(modelMBean, CONTROL_EXPORT_DOMAIN, properties);
    }
    
    private ObjectName registerMBean(ModelMBean mBean, String domain, Hashtable<String, String> properties) throws MBeanException,OperationsException {
        ObjectName objectName = new ObjectName(domain, properties);
        this.mBeanServer.registerMBean(mBean, objectName);
        return objectName;
    }
    
    private ModelMBean createModelMBean(Class<?> definition, Object service) throws JMException {
        ModelMBean modelMBean = new RequiredModelMBean();
        modelMBean.setModelMBeanInfo(createModelMBeanInfo(definition, service.getClass().getName()));
        return modelMBean;
    }
    
    private ModelMBeanInfo createModelMBeanInfo(Class<?> definition, String mBeanClassName) {
        return this.modelMBeanInfoAssembler.assembleModelMBeanInfo(definition, mBeanClassName);
    }
    
    private class ControlServiceListener implements ServiceListener {
        public void serviceChanged(ServiceEvent event) {           
            ServiceReference reference = event.getServiceReference();
            
            if (event.getType() == ServiceEvent.REGISTERED) {
                controlExported(reference);
            } else if (event.getType() == ServiceEvent.UNREGISTERING) {
                controlRetracted(reference);
            }
        }  
    }
}
