/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.control.core.internal;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;

import javax.management.modelmbean.ModelMBeanAttributeInfo;
import javax.management.modelmbean.ModelMBeanConstructorInfo;
import javax.management.modelmbean.ModelMBeanInfo;
import javax.management.modelmbean.ModelMBeanInfoSupport;
import javax.management.modelmbean.ModelMBeanNotificationInfo;
import javax.management.modelmbean.ModelMBeanOperationInfo;
import javax.management.openmbean.CompositeData;
import javax.management.openmbean.TabularData;

/**
 * Standard implementation of {@link ModelMBeanInfoAssembler} that uses reflection to translate the definition
 * {@link Class type} into a {@link ModelMBeanInfo}.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Andy Wilkinson
 * @since 1.0
 */
public class ReflectionBasedModelMBeanInfoAssembler implements ModelMBeanInfoAssembler {

    /**
     * {@inheritDoc}
     */
    public ModelMBeanInfo assembleModelMBeanInfo(Class<?> definition, String className) {

        ModelMBeanAttributeInfo[] attributes = new ModelMBeanAttributeInfo[0];
        ModelMBeanConstructorInfo[] constructors = new ModelMBeanConstructorInfo[0];
        ModelMBeanNotificationInfo[] notifications = new ModelMBeanNotificationInfo[0];

        ModelMBeanOperationInfo[] operations = assembleOperations(definition);

        ModelMBeanInfo info = new ModelMBeanInfoSupport(className, null, attributes, constructors, operations, notifications);
        return info;
    }

    /**
     * Creates {@link ModelMBeanOperationInfo} instances for all valid control methods on the supplied
     * {@link Class definition}.
     * 
     * @param definition the control definition interface.
     * @return the <code>ModelMBeanOperationInfo</code> instances for the supplied definition.
     */
    private ModelMBeanOperationInfo[] assembleOperations(Class<?> definition) {
        List<ModelMBeanOperationInfo> methodInfos = new ArrayList<ModelMBeanOperationInfo>();

        Method[] methods = definition.getMethods();

        for (Method method : methods) {
            Class<?> returnType = method.getReturnType();
            if (returnType == null || returnType.isPrimitive() || isOpenType(returnType)) {
                methodInfos.add(new ModelMBeanOperationInfo("method", method));
            }
        }

        return methodInfos.toArray(new ModelMBeanOperationInfo[0]);
    }

    /**
     * Queries whether or not the supplied type is a JMX Open Type ({@link CompositeData} or {@link TabularData}).
     */
    private boolean isOpenType(Class<?> returnType) {
        return CompositeData.class.isAssignableFrom(returnType) || TabularData.class.isAssignableFrom(returnType);
    }
}
