/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.control.core.internal;

import static com.springsource.server.control.core.internal.Constants.CONTROL_IDENTIFIER;
import static com.springsource.server.control.core.internal.Constants.CONTROL_SERVICE_PROPERTY_KEY_PREFIX;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.ServiceRegistration;

import com.springsource.server.control.core.ControlExporter;
import com.springsource.server.control.core.ControlHandle;
import com.springsource.server.osgi.framework.OsgiFramework;
import com.springsource.server.serviceability.ffdc.NonNull;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * A <code>ControlExporter</code> that exports controls
 * by publishing them as services in the OSGi service registry. 
 *
 * <strong>Concurrent Semantics</strong><br />
 * Thread-safe.
 *
 * @author Andy Wilkinson
 * @since 1.0
 */
public class ServiceRegistryControlExporter implements ControlExporter {
    
    private List<ControlHandle> controlHandles = new ArrayList<ControlHandle>();
    private final Object registrationLock = new Object();
    private static final Tracer TRACE = TracerFactory.createTracer(ServiceRegistryControlExporter.class);
    private final OsgiFramework osgiFramework;
    
    public ServiceRegistryControlExporter(OsgiFramework osgiFramework) {
        this.osgiFramework = osgiFramework;
    }
    
    /** 
     * {@inheritDoc}
     */
    public ControlHandle exportControl(@NonNull Object control, @NonNull Class<?> controlInterface) {
        return exportControl(control, controlInterface, null);
    }

    /** 
     * {@inheritDoc}
     */
    public ControlHandle exportControl(@NonNull Object control, @NonNull Class<?> controlInterface, Map<String, String> controlPropertiesMap) {
        synchronized(registrationLock) {
            Bundle owningBundle = osgiFramework.getClassBundle(controlInterface);
            BundleContext controlBundleContext = owningBundle.getBundleContext();
            
            Properties properties = convertToProperties(controlPropertiesMap, CONTROL_SERVICE_PROPERTY_KEY_PREFIX);

            if (properties.getProperty(CONTROL_SERVICE_PROPERTY_KEY_PREFIX + "type") == null) {
                properties.setProperty(CONTROL_SERVICE_PROPERTY_KEY_PREFIX + "type", determineControlType(controlInterface));
            }
            properties.setProperty(CONTROL_IDENTIFIER, "true");
            ControlHandle controlHandle = new ServiceRegistrationControlHandle(controlBundleContext.registerService(controlInterface.getName(), control, properties));
            controlHandles.add(controlHandle);
            return controlHandle;
        }
    }
    
    private String determineControlType(Class<?> controlInterface) {
        return controlInterface.getSimpleName();
    }
    
    private Properties convertToProperties(Map<String, String> properties, String keyPrefix) {
        Properties convertedAndPrefixed = new Properties();       
        if (properties != null) 
        {
            for (String key: properties.keySet()) {
                String value = properties.get(key);
                convertedAndPrefixed.put(CONTROL_SERVICE_PROPERTY_KEY_PREFIX + key, value);
            }
        }
        return convertedAndPrefixed;
    }
    
    public void retractControls() {
        synchronized(registrationLock) {
            for (ControlHandle controlHandle: controlHandles) {
                controlHandle.retractControl();
            }
        }
    }
    
    /**
     * A <code>ControlHandle</code> for a control that has been exported
     * to the OSGi service registry, i.e. the handle is really a
     * {@link ServiceRegistration}.
     * <p />
     *
     * <strong>Concurrent Semantics</strong><br />
     * Thread-safe.
     *
     * @author Andy Wilkinson
     * @since 1.0
     */
    private class ServiceRegistrationControlHandle implements ControlHandle {
        private ServiceRegistration controlRegistration;
        
        public ServiceRegistrationControlHandle(ServiceRegistration controlRegistration) {
            this.controlRegistration = controlRegistration;
        }
        /** 
         * {@inheritDoc}
         */
        public void retractControl() {
            synchronized(registrationLock) {
                try {
                    controlRegistration.unregister(); 
                } catch (IllegalStateException ise) {
                    if (TRACE.isDebugEnabled()) TRACE.traceDebug("Control '%s' already unregistered", controlRegistration);
                }
            }
        }        
    }
}
