/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.deployer.app;

import java.util.jar.Attributes;

import org.springframework.util.StringUtils;

import com.springsource.server.common.math.Version;
import com.springsource.server.deployer.core.DeployerLogEvent;
import com.springsource.server.deployer.core.DeploymentException;
import com.springsource.server.deployer.core.parse.DeployerBundleManifest;
import com.springsource.server.serviceability.logging.LogService;
import com.springsource.server.serviceability.logging.LogServiceFactory;

/**
 * Parses the {@link MultiBundleApplicationDescriptor OSGi application descriptor}
 * contained in a JAR manifest.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * Implementation is entirely stateless and as such is suitable for concurrent
 * access.
 * 
 * @author Glyn Normington
 * @author Andy Wilkinson
 * @since 1.0
 */
final class MultiBundleApplicationDescriptorParser {

	private static final String DEFAULT_APPLICATION_VERSION = "0";

	/**
	 * The name of the '<code>application-symbolicname</code>' header.
	 */
	private static final String SYMBOLIC_NAME_ATTRIBUTE = "application-symbolicname";

	private static final String SYMBOLIC_NAME_REGEX = "[-_0-9a-zA-Z]+(\\.[-_0-9a-zA-Z]+)*";

	/**
	 * The name of the '<code>application-name</code>' header.
	 */
	private static final String NAME_ATTRIBUTE = "application-name";

	/**
	 * The name of the '<code>application-description</code>' header.
	 */
	private static final String DESCRIPTION_ATTRIBUTE = "application-description";

	/**
	 * The name of the '<code>application-version</code>' header.
	 */
	private static final String VERSION_ATTRIBUTE = "application-version";

	private static final LogService LOGGER = LogServiceFactory.getLogService();

	/**
	 * Creates a new {@link MultiBundleApplicationDescriptor} by parsing the supplied manifest.
	 * 
	 * @param manifest
	 *            the <code>DeployerBundleManifest</code> to parse.
	 * @return the resultant {@link  MultiBundleApplicationDescriptor}.
	 * @throws DeploymentException
	 */
	public MultiBundleApplicationDescriptor parse(DeployerBundleManifest manifest)
			throws DeploymentException {
		Attributes attributes = manifest.getMainAttributes();
		return parseApplicationElement(attributes);
	}

	private MultiBundleApplicationDescriptor parseApplicationElement(
			Attributes attributes) throws DeploymentException {
		String symbolicName = attributes.getValue(SYMBOLIC_NAME_ATTRIBUTE);
		if (!StringUtils.hasText(symbolicName)) {
			LOGGER.log(DeployerLogEvent.APP_SYMBOLIC_NAME_MISSING, attributes
					.entrySet());
			throw new DeploymentException(
					"OSGi application symbolic name missing");
		}
		if (!symbolicName.matches(SYMBOLIC_NAME_REGEX)) {
			LOGGER.log(DeployerLogEvent.APP_SYMBOLIC_NAME_BAD_SYNTAX,
					attributes.entrySet());
			throw new DeploymentException("OSGi application symbolic name '"
					+ symbolicName + "' has invalid syntax");

		}
		String versionString = attributes.getValue(VERSION_ATTRIBUTE);
		if (!StringUtils.hasText(versionString)) {
			versionString = DEFAULT_APPLICATION_VERSION;
		}
		Version version;
		try {
			version = new Version(versionString);
		} catch (NumberFormatException e) {
			LOGGER.log(DeployerLogEvent.APP_VERSION_INVALID, versionString,
					attributes.entrySet());
			throw new DeploymentException("OSGi application version invalid");
		}
		return new MultiBundleApplicationDescriptor(symbolicName, version, attributes
				.getValue(NAME_ATTRIBUTE), attributes
				.getValue(DESCRIPTION_ATTRIBUTE));
	}

}
