/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.deployer.app.legacy;

import java.net.URI;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import com.springsource.server.common.math.Version;
import com.springsource.server.deployer.core.CompletableDeploymentContext;
import com.springsource.server.deployer.core.DeployerLogEvent;
import com.springsource.server.deployer.core.DeploymentContextFactory;
import com.springsource.server.deployer.core.DeploymentException;
import com.springsource.server.deployer.core.Module;
import com.springsource.server.deployer.core.ModuleDeployerSet;
import com.springsource.server.deployer.core.ModuleFactory;
import com.springsource.server.deployer.core.ModuleInfo;
import com.springsource.server.deployer.core.ApplicationDeployer.DeploymentOptions;
import com.springsource.server.deployer.core.internal.AbstractInformedApplicationDeployable;
import com.springsource.server.deployer.core.parse.DeployerBundleManifest;
import com.springsource.server.io.core.PathReference;
import com.springsource.server.osgi.framework.BundleInstaller;
import com.springsource.server.osgi.framework.ImportMergeException;
import com.springsource.server.osgi.framework.OsgiFramework;
import com.springsource.server.osgi.framework.UnableToSatisfyDependenciesException;
import com.springsource.server.osgi.manifest.BundleManifest;
import com.springsource.server.osgi.manifest.BundleManifest.BundleSymbolicName;
import com.springsource.server.serviceability.ffdc.Assert;
import com.springsource.server.serviceability.logging.LogService;
import com.springsource.server.serviceability.logging.LogServiceFactory;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;
import com.springsource.server.serviceability.tracing.TracingService;

/**
 * A {@link SingleBundleApplication} provides deployer support for .war files and such like as opposed to multi-bundle
 * OSGi Applications. <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * This class is not thread safe and so depends on the synchronization provided by StandardApplicationDeployer.
 * 
 * @author Glyn Normington
 * @author Andy Wilkinson
 * @since 1.0
 */
public class SingleBundleApplication extends AbstractInformedApplicationDeployable {
    
    private static final LogService LOGGER = LogServiceFactory.getLogService();

    private static final Tracer TRACER = TracerFactory.createTracer(SingleBundleApplication.class);

    private final Module module;

    private final BundleInstaller bundleInstaller;

    private final DeploymentContextFactory deploymentContextFactory;

    private final ModuleDeployerSet moduleDeployers;

    private final CompletableDeploymentContext deploymentContext;

    public SingleBundleApplication(OsgiFramework osgi, URI location, URI sourceLocation, DeploymentOptions deploymentOptions,
        ModuleDeployerSet moduleDeployers, ModuleFactory moduleFactory, DeploymentContextFactory deploymentContextFactory,
        TracingService tracingService) throws DeploymentException {
        super(sourceLocation, Type.SINGLE_BUNDLE_APPLICATION, deploymentOptions, osgi, tracingService);
        // Set the application's symbolic name temporarily to the filename, so that we have a reasonable name for the module's work area.
        setName(new PathReference(location).getName());
        this.module = moduleFactory.create(this, location);
        BundleManifest bundleManifest = this.module.getModuleData().getBundleManifest();

        /*
         * Use the bundle's symbolic name, if any, and bundle version, which defaults to 0, as the application's
         * symbolic name and version.
         */
        BundleSymbolicName bundleSymbolicName = bundleManifest.getBundleSymbolicName();
        if (bundleSymbolicName != null) {
            setName(bundleSymbolicName.getSymbolicName());
        }
        setVersion(new Version(bundleManifest.getBundleVersion()));

        super.setApplicationManifest(((DeployerBundleManifest) bundleManifest));
        this.module.refreshApplicationTraceName();
        this.bundleInstaller = osgi.getBundleInstaller();
        this.deploymentContextFactory = deploymentContextFactory;
        this.deploymentContext = this.deploymentContextFactory.create(this);
        this.moduleDeployers = moduleDeployers;
        this.moduleDeployers.determinePersonality(this.deploymentContext, module);
    }

    /**
     * @throws DeploymentException
     * @throws UnableToSatisfyDependenciesException {@inheritDoc}
     */
    public void deploy() throws DeploymentException, UnableToSatisfyDependenciesException {
        try {
            List<BundleManifest> bundleManifestList = new ArrayList<BundleManifest>(1);
            bundleManifestList.add(this.module.getModuleData().getBundleManifest());
            this.bundleInstaller.getImportExpander().expandImports(bundleManifestList);
            module.transform();
            module.install();
            this.bundleInstaller.satisfyConstraints(module.getBundle());
            super.deploy();
            module.deploy();
            module.checkDeployed();
            super.deployed();
            this.deploymentContext.completed();
        } catch (DeploymentException e) {
            undeployAfterFailedDeploy();
            throw e;
        } catch (UnableToSatisfyDependenciesException utsde) {
            undeployAfterFailedDeploy();
            throw utsde;
        } catch (ImportMergeException e) {
            undeployAfterFailedDeploy();
            LOGGER.log(DeployerLogEvent.CONFLICTING_PACKAGE_EXPORTS, this.getName(), this.getVersion(), e.getConflictingPackageName(), e.getSources());
            throw new DeploymentException("Clashing exports of package '" + e.getConflictingPackageName() + "' while expanding imports: "
                + e.getSources(), e);
        } catch (RuntimeException e) {
            undeployAfterFailedDeploy();
            throw e;
        }

    }

    private void undeployAfterFailedDeploy() {
        try {
            undeploy();
        } catch (Exception ex) {
            TRACER.traceWarning("ApplicationUndeploying failed deployment for application '%s' at version '%s' was not successful.", this.getName(),
                this.getVersion());
        }
    }

    /**
     * {@inheritDoc}
     */
    public void start() throws DeploymentException {
        module.start();
    }

    /**
     * {@inheritDoc}
     */
    public void stop() throws DeploymentException {
        module.stop();

    }

    /**
     * {@inheritDoc}
     */
    public void undeploy() throws DeploymentException {
        module.undeploy();
        super.undeploy();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Set<ModuleInfo> getModuleInfo() {
        Set<ModuleInfo> result = new HashSet<ModuleInfo>(1);
        result.add(this.module);
        return result;
    }

    /**
     * {@inheritDoc}
     */
    public boolean refresh(String symbolicName, ModuleDeployerSet moduleDeployerSet, CompletableDeploymentContext deploymentContext)
        throws DeploymentException {
        Assert.isNull(moduleDeployerSet, "non-null moduleDeployerSet");
        Assert.isNull(deploymentContext, "non-null deploymentContext");
        if (symbolicName == null) {
            symbolicName = this.module.getUnscopedSymbolicName();
        }
        return this.module.refresh(symbolicName, this.moduleDeployers, this.deploymentContextFactory.create(this));
    }

    /**
     * {@inheritDoc}
     */
    public void updateResource(String symbolicName, URI resourceLocation, String targetPath) throws DeploymentException {
        module.updateResource(symbolicName, resourceLocation, targetPath);
    }

    /**
     * {@inheritDoc}
     */
    public void deleteResource(String symbolicName, String targetPath) throws DeploymentException {
        module.deleteResource(symbolicName, targetPath);
    }

}
