/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.deployer.app.spring;

import java.net.URL;

import org.springframework.context.ApplicationContext;
import org.springframework.osgi.context.support.OsgiBundleXmlApplicationContext;

/**
 * Extension of {@link OsgiBundleXmlApplicationContext} that enforces correct thread context classloader propagation.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * As threadsafe as <code>OsgiBundleXmlApplicationContext</code>.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
public class GenericModuleOsgiBundleXmlApplicationContext extends OsgiBundleXmlApplicationContext {

    /**
     * Creates a <code>GenericModuleOsgiBundleXmlApplicationContext</code> with no parent.
     */
    public GenericModuleOsgiBundleXmlApplicationContext() {
        super();
    }

    /**
     * Creates a <code>GenericModuleOsgiBundleXmlApplicationContext</code> with the supplied parent.
     * 
     * @param parent the parent {@link ApplicationContext}.
     */
    public GenericModuleOsgiBundleXmlApplicationContext(ApplicationContext parent) {
        super(parent);
    }

    /**
     * Creates a <code>GenericModuleOsgiBundleXmlApplicationContext</code> with the supplied parent and config
     * locations.
     * 
     * @param configLocations the config locations.
     * @param parent the parent {@link ApplicationContext}.
     */
    public GenericModuleOsgiBundleXmlApplicationContext(String[] configLocations, ApplicationContext parent) {
        super(configLocations, parent);
    }

    /**
     * Creates a <code>GenericModuleOsgiBundleXmlApplicationContext</code> with the supplied config locations.
     * 
     * @param configLocations the config locations.
     */
    public GenericModuleOsgiBundleXmlApplicationContext(String[] configLocations) {
        super(configLocations);
    }

    /**
     * Returns the current thread's context classloader rather than the bean classloader. If no thread context
     * classloader is set, then the bean classloader is returned.
     * 
     * @return the context class loader to use.
     */
    protected final ClassLoader getContextClassLoader() {
        return new ChainedClassLoader(new ClassLoader[] { getClassLoader(), Thread.currentThread().getContextClassLoader() });
    }

    public class ChainedClassLoader extends ClassLoader {

        private final ClassLoader[] loaders;

        public ChainedClassLoader(ClassLoader[] loaders) {
            this.loaders = (ClassLoader[]) loaders.clone();
        }

        public URL getResource(String name) {
            URL url = null;
            for (int i = 0; i < loaders.length; i++) {
                ClassLoader loader = loaders[i];
                url = loader.getResource(name);
                if (url != null)
                    return url;
            }
            return url;
        }

        public Class<?> loadClass(String name) throws ClassNotFoundException {
            Class<?> clazz = null;
            for (int i = 0; i < loaders.length; i++) {
                ClassLoader loader = loaders[i];
                try {
                    clazz = loader.loadClass(name);
                    return clazz;
                } catch (ClassNotFoundException e) {
                    // keep moving through the class loaders
                }
            }
            throw new ClassNotFoundException(name);
        }
    }
}
