/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.deployer.controls;

import javax.management.openmbean.CompositeData;

import com.springsource.server.deployer.core.DeploymentException;
import com.springsource.server.deployer.core.DeploymentIdentity;

/**
 * Definition of the Deployer control used to allow the Server to support remote deployment.
 * 
 * <strong>Concurrent Semantics</strong><br />
 * Implementations must be thread-safe.
 * 
 * @author Andy Wilkinson
 * @author Glyn Normington
 * @since 1.0
 */
public interface Deployer {

    /**
     * Deploy an application located at the given URI.
     * 
     * @param uri the URI string of the location of the application
     * @return the {@link DeploymentIdentity} of the deployed application as a {@link CompositeData} object.
     * @throws DeploymentException
     */
    CompositeData deploy(String uri) throws DeploymentException;

    /**
     * Deploy an application located at the given URI.
     * 
     * @param uri the URI string of the location of the application.
     * @param recoverable whether or not the application should be recovered on warm restart.
     * @return the {@link DeploymentIdentity} of the deployed application as a {@link CompositeData} object.
     * @throws DeploymentException
     */
    CompositeData deploy(String uri, boolean recoverable) throws DeploymentException;

    /**
     * Undeploy an application with a given symbolic name and version.
     * 
     * @param applicationSymbolicName the symbolic name of the application
     * @param version the version of the application in string form
     * @throws DeploymentException
     */
    void undeploy(String applicationSymbolicName, String version) throws DeploymentException;

    /**
     * Refresh a single module of the application which was deployed from the given URI.
     * 
     * @param uri the URI string from which the application was deployed
     * @param symbolicName the bundle symbolic name of the module to be refreshed
     * @throws DeploymentException if the refresh failed
     */
    void refresh(String uri, String symbolicName) throws DeploymentException;

    /**
     * Update files or directories of the module of the single-bundle application with the given symbolic name and
     * version. The application must have been deployed with the non-recoverable deployment option.
     * <p />
     * The given location refers to a file or directory to be copied recursively over the file or directory at a given
     * path relative to the root directory of the module.
     * <p />
     * <b>WARNING</b>
     * <p />
     * The intended use of this method is to update static resources which are loaded directly from the exploded form of
     * the module in the deployer's work area. This method does <b>not</b> update or refresh the given module and so
     * the updated resources <b>cannot</b> be accessed using the module's class loader. The updated resources will be
     * <b>overwritten</b> if the given module is refreshed or redeployed.
     * 
     * @param applicationSymbolicName the symbolic name of the application
     * @param version the version of the application in string form
     * @param resourceLocation the location of the file or directory to be overlaid on the file or directory at the
     *        given target path of the deployed module
     * @param targetPath the relative URI path of the location of the file to be updated within the deployed module
     * @throws DeploymentException if the update failed
     */
    void updateResource(String applicationSymbolicName, String version, String resourceLocation, String targetPath) throws DeploymentException;

    /**
     * Update files or directories of a module of the multi-bundle application with the given symbolic name and version.
     * The application must have been deployed with the non-recoverable deployment option.
     * <p />
     * A location is given which refers to a file or directory to be copied recursively over the file or directory at a
     * given path relative to the root directory of the module.
     * <p />
     * A <code>null</code> location may be used to delete the file or directory at the given path relative to the root
     * directory of the module. In this case any non-existent directories of the given path are ignored.
     * <p />
     * <b>WARNING</b>
     * <p />
     * The intended use of this method is to update static resources which are loaded directly from the exploded form of
     * the module in the deployer's work area. This method does <b>not</b> update or refresh the given module and so
     * the updated resources <b>cannot</b> be accessed using the module's class loader. The updated resources will be
     * <b>overwritten</b> if the given module is refreshed or redeployed.
     * 
     * @param applicationSymbolicName the symbolic name of the application
     * @param version the version of the application in string form
     * @param symbolicName the bundle symbolic name of the module to be updated
     * @param resourceLocation the location of the file or directory to be overlaid on the file or directory at the
     *        given target path of the deployed module or <code>null</code> if the file or directory at the given
     *        target path is to be deleted
     * @param targetPath the relative URI path of the location of the file to be updated within the deployed module
     * @throws DeploymentException if the update failed
     */
    void updateResource(String applicationSymbolicName, String version, String symbolicName, String resourceLocation, String targetPath)
        throws DeploymentException;

    /**
     * Delete a file or directory at the given path relative to the root directory of the module of the single-bundle
     * application with the given symbolic name and version. The application must have been deployed with the
     * non-recoverable deployment option.
     * <p />
     * <b>WARNING</b>
     * <p />
     * The intended use of this method is to delete static resources which are loaded directly from the exploded form of
     * the module in the deployer's work area. This method does <b>not</b> update or refresh the given module and so
     * the updated resources <b>may still</b> be accessed using the module's class loader.
     * 
     * @param applicationSymbolicName the symbolic name of the application
     * @param version the version of the application in string form
     * @param targetPath the relative URI path of the location of the file to be deleted within the deployed module
     * @throws DeploymentException if the deletion failed
     */
    void deleteResource(String applicationSymbolicName, String version, String targetPath) throws DeploymentException;

    /**
     * Delete a file or directory at the given path relative to the root directory of a module of the multi-bundle
     * application with the given symbolic name and version. The application must have been deployed with the
     * non-recoverable deployment option.
     * <p />
     * <b>WARNING</b>
     * <p />
     * The intended use of this method is to delete static resources which are loaded directly from the exploded form of
     * the module in the deployer's work area. This method does <b>not</b> update or refresh the given module and so
     * the updated resources <b>may still</b> be accessed using the module's class loader.
     * 
     * @param applicationSymbolicName the symbolic name of the application
     * @param version the version of the application in string form
     * @param symbolicName the bundle symbolic name of the module to be updated
     * @param targetPath the relative URI path of the location of the file to be deleted within the deployed module
     * @throws DeploymentException if the deletion failed
     */
    void deleteResource(String applicationSymbolicName, String version, String symbolicName, String targetPath) throws DeploymentException;

}
