/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.deployer.core;

import java.net.URI;

import com.springsource.server.osgi.framework.UnableToSatisfyDependenciesException;

/**
 * A {@link Deployable} provides the deployment lifecycle for artefacts such as applications, libraries, single module
 * applications, and modules of applications. <p/>
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * Implementations of this interface need not be thread safe. It is the responsibility of callers to synchronize access.
 * 
 * @author Glyn Normington
 * @since 1.0
 */
public interface Deployable {

    /**
     * Start and deploy this artefact.
     * 
     * @throws DeploymentException
     * @throws UnableToSatisfyDependenciesException
     */
    void deploy() throws DeploymentException, UnableToSatisfyDependenciesException;

    /**
     * Start this artefact. If it is already started, this has no effect.
     * 
     * @throws DeploymentException
     */
    void start() throws DeploymentException;

    /**
     * Stop this artefact. If it is already stopped, this has no effect.
     * 
     * @throws DeploymentException
     */
    void stop() throws DeploymentException;

    /**
     * Undeploy and stop this artefact.
     * 
     * @throws DeploymentException
     */
    void undeploy() throws DeploymentException;

    /**
     * Refresh the given module of this artefact. If the module cannot be refreshed, for example because the set of
     * exported packages has changed, return <code>false</code>.
     * <p />
     * To refresh an artefact with a single module, the given module may be <code>null</code>.
     * 
     * @param symbolicName the bundle symbolic name of the module to be refreshed, or <code>null</code> for a single module artefact
     * @return <code>true</code> if and only if the refresh was successful
     * @throws DeploymentException
     */
    boolean refresh(String symbolicName, ModuleDeployerSet moduleDeployerSet, CompletableDeploymentContext deploymentContext)
        throws DeploymentException;

    /**
     * Update some resources of the given module of this artefact. The application must have been deployed with the
     * non-recoverable deployment option.
     * <p />
     * <b>WARNING</b>
     * <p />
     * The intended use of this method is to update static resources which are loaded directly from the exploded form of
     * the module in the deployer's work area. This method does <b>not</b> update or refresh the given module and so
     * the updated resources <b>cannot</b> be accessed using the module's class loader. The updated resources will be
     * <b>overwritten</b> if the given module is refreshed or redeployed.
     * 
     * @param symbolicName the bundle symbolic name of the module to be updated (ignored for single-bundle applications)
     * @param resourceLocation the location of the file to be overlaid on the deployed version of the given module
     * @param targetPath the relative URI path of the location of the file to be updated within the given module
     * @throws DeploymentException 
     */
    void updateResource(String symbolicName, URI resourceLocation, String targetPath) throws DeploymentException;

    /**
     * Delete the resource at the given relative target path of the given module of this artefact. The application must
     * have been deployed with the non-recoverable deployment option.
     * <p />
     * <b>WARNING</b>
     * <p />
     * The intended use of this method is to delete static resources which are loaded directly from the exploded form of
     * the module in the deployer's work area. This method does <b>not</b> update or refresh the given module and so
     * the updated resources <b>may still</b> be accessed using the module's class loader.
     * 
     * @param symbolicName the bundle symbolic name of the module to be deleted (ignored for single-bundle applications)
     * @param targetPath the relative URI path of the location of the file or directory to be deleted within the given module
     * @throws DeploymentException
     */
    void deleteResource(String symbolicName, String targetPath) throws DeploymentException;

}
