/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.deployer.core;

import java.net.URI;
import java.util.Date;
import java.util.Set;

import com.springsource.server.common.math.Version;
import com.springsource.server.deployer.core.ApplicationDeployer.DeploymentOptions;

/**
 * {@link DeployableInfo} exposes management information about a particular {@link Deployable}.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Implementations of this interface must be thread safe.
 * 
 * @author Glyn Normington
 * @since 1.0
 */
public interface DeployableInfo {

    enum Type {
        MULTI_BUNDLE_APPLICATION, SINGLE_BUNDLE_APPLICATION, LIBRARY
    }

    /**
     * Get the type of this {@link Deployable}.
     * 
     * @return the {@link Type} of this {@link Deployable}
     */
    Type getType();

    /**
     * Get the name of this {@link Deployable}.
     * 
     * Note: the name of a library is not available until it has been deployed.
     * 
     * @return the name of this {@link Deployable}
     */
    String getName();

    /**
     * Get the version of this {@link Deployable}.
     * 
     * Note: the version of a library is not available until it has been deployed.
     * 
     * @return the version of this {@link Deployable}
     */
    Version getVersion();

    /**
     * Get the time when this {@link Deployable} was deployed.
     * 
     * Note: if the server is restarted, the current deployer recovery scheme redeploys applications and updates their
     * deploy times accordingly.
     * 
     * @return the time of last deployment as a number of milliseconds since midnight, January 1, 1970 UTC or null if this
     *         {@link Deployable} has not yet been deployed
     */
    Date getDeployTime();
    
    /**
     * Return the source URI that this {@link Deployable} came from.
     * 
     * @return the source URI of this {@link Deployable}
     */
    URI getSourceUri();
    
    
    /**
     * Get the {@link DeploymentOptions} associated with this {@link Deployable}.
     * 
     * @return this <code>Deployable</code>'s <code>DeploymentOptions</code>
     */
    DeploymentOptions getDeploymentOptions();

    /**
     * Get the {@link ModuleInfo ModuleInfos} of all the modules of this {@link Deployable}.
     * 
     * @return this {@link Deployable Deployable's} {@link ModuleInfo ModuleInfos}
     */
    Set<ModuleInfo> getModuleInfo();
}
