/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.deployer.core;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.InputStream;
import java.net.URI;
import java.util.Set;

import com.springsource.server.io.core.PathReference;

/**
 * TODO Document DeploymentArtefact
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Implementations <strong>must</strong> be threadsafe.
 * 
 * @author Glyn Normington
 * @author Rob Harrop
 * @author Sam Brannen
 * @since 1.0
 */
public interface DeploymentArtefact {

    /**
     * Return a URI referring to this deployment artefact.
     */
    URI getURI();

    /**
     * Return a File containing this deployment artefact. The underlying file must not be modified.
     */
    File getFile();

    /**
     * Return a filename for this deployment artefact which is typically the last part of the path: for example,
     * "myfile.txt".
     */
    String getFilename();

    /**
     * Return a description for this deployment artefact for diagnostics.
     */
    String getDescription();

    /**
     * Gets the <code>Entry</code> for the supplied path. Paths <code>must</code> start with a <code>/</code>
     * (e.g., <code>/META-INF/spring/foo-context.xml</code>).
     * 
     * @param path the path of the entry.
     * @return the entry, never <code>null</code>.
     */
    Entry getEntry(String path);

    /**
     * Gets the contents of the deployment artefact as a set of entries.
     * 
     * @return the set of entries corresponding to the contents of this deployment artefact
     */
    Set<Entry> getContents();

    /**
     * Creates a new directory entry with the supplied <code>path</code> within this deployment artefact. Paths
     * <code>must</code> start with a <code>/</code> (e.g., <code>/META-INF/spring/foo-context.xml</code>).
     * 
     * @param path the directory path to create
     * @return the created directory entry
     * @throws IllegalArgumentException if the supplied <code>path</code> does not start with a <code>/</code>
     */
    Entry createDirectoryEntry(String path) throws IllegalArgumentException;

    /**
     * Copies the file or directory at the given source reference over the file or directory identified by the given
     * path.
     * 
     * @param source the file or directory to be copied from
     * @param path the file or directory to be copied to
     */
    void copy(PathReference source, String path);

    /**
     * Describes an entry in a deployment artefact. This entry may or may not exist, and may or may not be a directory
     * entry. Directory entries can contain nested child entries.<p/>
     * 
     * <strong>Concurrent Semantics</strong><br />
     * 
     * Implementations <code>must</code> be thread-safe.
     * 
     * @author Rob Harrop
     * @since 1.0
     */
    public interface Entry {

        /**
         * Gets the name of this entry.
         * 
         * @return the entry name.
         */
        String getName();

        /**
         * Gets the filename for this entry which is typically the last part of the path: for example, "myfile.txt".
         * 
         * @return the entry's filename.
         */
        String getFilename();

        /**
         * Creates a {@link BufferedReader} for this entry.
         * 
         * @return the <code>BufferedReader</code>.
         * @throws FatalDeploymentException if this entry does not exist or is a directory entry.
         */
        BufferedReader getReader() throws FatalDeploymentException;

        /**
         * Creates an {@link InputStream} for this entry.
         * 
         * @return the <code>InputStream</code>.
         * @throws FatalDeploymentException if this entry does not exist or is a directory entry.
         */
        InputStream getInputStream() throws FatalDeploymentException;

        /**
         * Creates a {@link BufferedWriter} for this entry, creating the entry if it does not exist.
         * 
         * @param append flag indicating whether or not to append to existing content.
         * @return the <code>BufferedWriter</code>.
         * @throws FatalDeploymentException if this is a directory entry.
         */
        BufferedWriter getWriter(boolean append) throws FatalDeploymentException;

        /**
         * Gets the child entries for this entry.
         * 
         * @return the child entries.
         */
        Set<Entry> children();

        /**
         * Queries whether or not this entry exists in the deployment artefact.
         * 
         * @return <code>true</code> if the artefact exists, otherwise <code>false</code>.
         */
        boolean exists();

        /**
         * Queries whether or not this entry is a directory entry.
         * 
         * @return <code>true</code> if the directory is an entry, otherwise <code>false</code>.
         */
        boolean isDirectory();

        /**
         * Deletes this entry. If this entry is a directory, the directory and all of its contents will be recursively
         * deleted.
         * 
         * @return <code>true</code> if this entry was successfully deleted, <code>false</code> otherwise
         * @see #exists()
         */
        boolean delete();

        /**
         * Moves this entry to the supplied <code>destination</code> within the enclosing {@link DeploymentArtefact}
         * and returns a reference to the entry at its new location. If this entry refers to a directory, the directory
         * and its contents will be recursively moved to the destination.
         * 
         * @param destination the destination to which this entry should be moved
         * @return the moved entry, never <code>null</code>.
         * @throws IllegalArgumentException if the supplied destination does not start with a <code>/</code>, or if
         *         this entry does not exist
         */
        Entry move(String destination) throws IllegalArgumentException;

        /**
         * Moves the contents of this entry -- which must be a directory -- to the supplied <code>destination</code>
         * directory within the enclosing {@link DeploymentArtefact} and returns a reference to the new directory entry.
         * The contents of this entry will be recursively moved to the destination. After completing the move, this
         * entry will be deleted.
         * 
         * @param destination the destination to which the contents of this entry should be moved
         * @return the new directory entry, never <code>null</code>.
         * @throws IllegalArgumentException if the supplied destination does not start with a <code>/</code>, or if
         *         this entry does not exist
         * @see #moveContents(String, boolean)
         */
        Entry moveContents(String destination) throws IllegalArgumentException;

        /**
         * Moves the contents of this entry -- which must be a directory -- to the supplied <code>destination</code>
         * directory within the enclosing {@link DeploymentArtefact} and returns a reference to the new directory entry.
         * The contents of this entry will be recursively moved to the destination. If the <code>deleteSource</code>
         * argument is set to <code>true</code>, the source directory will be deleted; otherwise it will remain
         * intact.
         * 
         * @param destination the destination to which the contents of this entry should be moved
         * @param deleteSource if <code>true</code>, the source directory will be deleted
         * @return the new directory entry, never <code>null</code>.
         * @throws IllegalArgumentException if the supplied destination does not start with a <code>/</code>, or if
         *         this entry does not exist
         * @see #moveContents(String)
         */
        Entry moveContents(String destination, boolean deleteSource) throws IllegalArgumentException;

        /**
         * Gets a non-directory entry as a deployment artefact.
         * 
         * @return the deployment artefact corresponding to this entry or <code>null</code> if this entry is a
         *         directory
         * @throws DeploymentException Thrown if an error occurs getting the <code>DeploymentArtefact</code>.
         */
        DeploymentArtefact getDeploymentArtefact() throws DeploymentException;
    }

}
