/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.deployer.core;

import org.osgi.framework.Bundle;

import com.springsource.server.osgi.framework.UnableToSatisfyDependenciesException;

/**
 * A {@link Module} represents a unit of application packaging.
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * Implementations of this interface need not be thread safe. It is the responsibility of callers to synchronize access.
 * 
 * @author Glyn Normington
 * @since 1.0
 */
public interface Module extends ModulePersonality, ModuleInfo {

    /**
     * Get the module content data.
     * 
     * @return the module data
     */
    ModuleData getModuleData();

    /**
     * Set the personality of this module.
     * 
     * @param modulePersonality
     */
    void setPersonality(ModulePersonality modulePersonality);

    /**
     * Determine whether this module has a personality.
     * 
     * @return true if and only if this module has a personality
     */
    boolean hasPersonality();
    
    /**
     * Set the context module of this module.
     * 
     * @param context the context module of this module
     */
    void setContext(Module context);
    
    /**
     * Get the context module of this module.
     * 
     * @return the context module of this module
     */
    Module getContext();

    /**
     * Install this module as a bundle in OSGi.
     * 
     * @throws UnableToSatisfyDependenciesException 
     */
    void install() throws UnableToSatisfyDependenciesException;
    
    /**
     * Uninstall this module's bundle. Must not be called after the module has been deployed or otherwise started.
     */
    void uninstall();

    /**
     * Gets the {@link Bundle} for this module.
     * 
     * @return the <code>Bundle</code>, may be <code>null</code> if the
     *         module is not yet installed.
     * 
     * TODO: it may be possible to delete this method when the context bundle
     * has been synthesised.
     */
    Bundle getBundle();

    /**
     * Set the {@link ModuleScoper} for this module to use for re-scoping itself.
     * 
     * @param moduleScoper the {@link ModuleScoper} to be set
     */
    void setScoper(ModuleScoper moduleScoper);
    
    /**
     * Set the {@link ModuleImportExpander} for this module to use for expanding its imports during refresh.
     * 
     * @param moduleImportExpander the {@link ModuleImportExpander} to be set
     */
    void setImportExpander(ModuleImportExpander moduleImportExpander);

    /**
     * Check the results of a previous deploy call.
     * 
     * @throws DeploymentException
     */
    void checkDeployed() throws DeploymentException;
    
    /**
     * Gets the maximum time that this module may take to be deployed.
     * 
     * @return the maximum time in seconds
     */
    int getDeploymentTimeoutSeconds();
    
    /**
     * Refresh the application trace name from the informed deployable associated with this module.
     */
    void refreshApplicationTraceName();

}
