/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.deployer.core;

import java.io.File;
import java.net.URI;

import org.springframework.core.AttributeAccessor;

import com.springsource.server.deployer.core.DeploymentArtefact.Entry;
import com.springsource.server.osgi.manifest.BundleManifest;

/**
 * Provides an interface for manipulating the contents of an application module..<p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Implementations <strong>must</strong> be threadsafe.
 * 
 * @author Rob Harrop
 * @author Glyn Normington
 * @author Andy Wilkinson
 * @author Sam Brannen
 * @since 1.0
 */
public interface ModuleData extends AttributeAccessor {

    /**
     * Return the module's bundle manifest. Any modifications subsequently made to the bundle manifest will be recorded
     * when replaceManifest is called.
     * 
     * @return the bundle manifest abstract syntax tree
     */
    BundleManifest getBundleManifest();

    /**
     * Record the bundle manifest including any changes which have been made to it in the manifest file.
     */
    void replaceManifest();

    /**
     * Refresh the contents from the original source.
     * 
     * @throws DeploymentException 
     */
    void refreshContents() throws DeploymentException;

    /**
     * Remove staging files associated with the module.
     */
    void cleanup();

    /**
     * Return the filename for this module which is the last part of the path: for example, "myapp.jar".
     */
    String getFilename();

    /**
     * Rename the module to have to given filename.
     * 
     * @param newFilename the new filename of the module
     */
    void rename(String newFilename);

    /**
     * Gets the <code>Entry</code> in this module for the supplied path. Paths <code>must</code> start with a
     * <code>/</code> (e.g., <code>/META-INF/spring/foo-context.xml</code>).
     * 
     * @param path the path of the entry.
     * @return the entry, never <code>null</code>.
     */
    Entry getEntry(String path);

    /**
     * Return a File containing this module.
     */
    File getFile();

    /**
     * Update some resources of this module.
     * 
     * @param resourceLocation the location of the file to be overlaid on this module
     * @param targetPath the relative URI path of the location of the file to be updated within this module
     */
    void updateResource(URI resourceLocation, String targetPath);

    /**
     * Delete the resource at the given relative target path of this module.
     * 
     * @param targetPath the relative URI path of the location of the file or directory to be deleted within this module
     */
    void deleteResource(String targetPath);

}
