/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.deployer.core;

/**
 * A {@link ModulePersonality} provides the personality-specific behaviour of a {@link Module} or default behaviour for
 * modules devoid of personality. <p/>
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * Implementations of this interface need not be thread safe. It is the responsibility of callers to synchronize access.
 * 
 * @author Glyn Normington
 * @since 1.0
 */
public interface ModulePersonality extends Deployable {

    /**
     * Transform the bundle manifest of this module in preparation for metadata rewriting prior to deployment. For
     * example, a standard set of package imports could be added so the module can access the APIs of its container.
     * This may be a no-op if no changes are required.
     * 
     * A post-condition is that the Module-Type header is set to the appropriate personality.
     * 
     * @throws DeploymentException 
     */
    void transform() throws DeploymentException;

    /**
     * Get the personality of this module.
     * 
     * @return the personality of this module or <code>null</code> if this module is devoid of personality
     */
    String getPersonality();

    /**
     * Get an identifier of this module which uniquely identifies the module among all modules of the same personality.
     * 
     * @return the personality identifier of this module or <code>null</code> if this module is devoid of personality
     *         or the module has not yet been deployed
     */
    String getPersonalityIdentifier();

    /**
     * Perform any processing before this module is started.
     * 
     * @throws DeploymentException 
     */
    void onStarting() throws DeploymentException;
    
    /**
     * Perform any processing after this module is started.
     * 
     * @throws DeploymentException 
     */
    void onStarted() throws DeploymentException;
    
    /**
     * Perform any processing after before this module is stopped.
     * 
     * @throws DeploymentException 
     */
    void onStopping() throws DeploymentException;

    /**
     * Perform any processing after this module is stopped.
     * 
     * @throws DeploymentException 
     */
    void onStopped() throws DeploymentException;
    
    /**
     * Perform any processing after this module is unresolved.
     */
    void onUnresolved();

}
