/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.deployer.core.event;

import java.util.Set;

import org.osgi.framework.Bundle;

import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * {@link ApplicationDeploymentEventNotifier} is used by the deployer to notify all
 * {@link DeploymentListener ApplicationDeploymentListeners} of an application deployment event.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * This class is thread safe.
 * 
 * @author Glyn Normington
 * @since 1.0
 */
class ApplicationDeploymentEventNotifier implements DeploymentListener {

    private static final Tracer TRACER = TracerFactory.createTracer(ApplicationDeploymentEventNotifier.class);

    private final Set<DeploymentListener> listeners;

    /**
     * Construct an {@link ApplicationDeploymentEventNotifier} for the given set of listeners. The given set will be
     * modified externally to this class as listeners come and go.
     * 
     * @param listeners the listeners to be notified
     */
    ApplicationDeploymentEventNotifier(Set<DeploymentListener> listeners) {
        this.listeners = listeners;
    }

    /**
     * {@inheritDoc}
     */
    public void onEvent(ApplicationDeploymentEvent event) {
        Bundle bundle = null;
        if (event instanceof ApplicationBundleDeploymentEvent) {
            bundle = ((ApplicationBundleDeploymentEvent) event).getBundle();
        }
        
        if (bundle == null) {
            TRACER.traceInfo("Delivering '%s' for application '%s' version '%s' to application deployment listeners", event.getClass().getName(),
                event.getApplicationSymbolicName(), event.getApplicationVersion());
        } else {
            TRACER.traceInfo("Delivering '%s' for bundle '%s' of application '%s' version '%s' to application deployment listeners",
           
                event.getClass().getName(), bundle.getSymbolicName(), event.getApplicationSymbolicName(), event.getApplicationVersion());
        }
        
        for (DeploymentListener listener : listeners) {
            try {
                listener.onEvent(event);
            } catch (RuntimeException e) {
                // Trace and ignore the exception.
                TRACER.traceError("Application deployment listener '%s' threw exception", e, listener);
            }
        }
        
        if (bundle == null) {
            TRACER.traceInfo("Delivered '%s' for application '%s' version '%s' to application deployment listeners", event.getClass().getName(),
                event.getApplicationSymbolicName(), event.getApplicationVersion());
        } else {
            TRACER.traceInfo("Delivered '%s' for bundle '%s' of application '%s' version '%s' to application deployment listeners",
                event.getClass().getName(), bundle.getSymbolicName(), event.getApplicationSymbolicName(), event.getApplicationVersion());
        }
    }

}
