/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.deployer.core.internal;

import java.net.URI;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;

import org.osgi.framework.ServiceRegistration;

import com.springsource.server.deployer.core.Deployable;
import com.springsource.server.deployer.core.DeploymentException;
import com.springsource.server.deployer.core.ModuleInfo;
import com.springsource.server.deployer.core.ApplicationDeployer.DeploymentOptions;
import com.springsource.server.deployer.core.parse.DeployerBundleManifest;
import com.springsource.server.management.proxy.application.ServerApplicationInfo;
import com.springsource.server.management.proxy.application.ServerModuleInfo;
import com.springsource.server.osgi.framework.OsgiFramework;
import com.springsource.server.osgi.framework.UnableToSatisfyDependenciesException;
import com.springsource.server.serviceability.output.Level;
import com.springsource.server.serviceability.tracing.TracingService;

/**
 * <code>AbstractInformedApplicationDeployable</code> is a base class for application {@link Deployable Deployables}.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * This class is thread safe.
 * 
 * @author Andy Wilkinson
 * @since 1.0
 */
public abstract class AbstractInformedApplicationDeployable extends AbstractInformedDeployable {

    private volatile ServiceRegistration serverApplicationInfoServiceRegistration;
    
    private volatile DeployerBundleManifest applicationManifest;
    
    private final TracingService tracingService;
    
    private static final String APPLICATION_TRACE_LEVELS_MANIFEST_HEADER_NAME = "Application-TraceLevels";

    public AbstractInformedApplicationDeployable(URI sourceLocation, Type type, DeploymentOptions deploymentOptions, OsgiFramework osgi, TracingService tracingService) {
        super(sourceLocation, type, deploymentOptions, osgi);
        assert type != Type.LIBRARY;
        this.tracingService = tracingService;
    }
    
    public void setApplicationManifest(DeployerBundleManifest manifest) {
        this.applicationManifest = manifest;
    }

    /**
     * {@inheritDoc}
     */
    public void deploy() throws UnableToSatisfyDependenciesException, DeploymentException {
        super.deploy();
        
        if (applicationManifest != null) {
            String value = applicationManifest.getMainAttributes().getValue(APPLICATION_TRACE_LEVELS_MANIFEST_HEADER_NAME); 
            
            Map<String, Level> configuredLevels = new HashMap<String, Level>();
            
            if (value != null) {
                String[] entries = value.split(",");
                if (entries != null) {
                    for (String entry : entries) {
                        int index = entry.indexOf('=');
                        if (index > 0) {
                            String name = entry.substring(0, index);
                            String level = entry.substring(index + 1);
                            configuredLevels.put(name, Level.valueOf(level.toUpperCase(Locale.ENGLISH)));
                        }
                    }
                }
            }  
            
            tracingService.applicationDeployed(getApplicationName(), configuredLevels);
        }
    }
    
    public void deployed() {
        this.serverApplicationInfoServiceRegistration = this.osgi.getBundleContext().registerService(ServerApplicationInfo.class.getName(),
            new ServerApplicationInfo() {

                public Date getDeployTime() {
                    return AbstractInformedApplicationDeployable.this.getDeployTime();
                }

                public String getName() {
                    return AbstractInformedApplicationDeployable.this.getName();
                }

                public String getVersion() {
                    return AbstractInformedApplicationDeployable.this.getVersion().toString();
                }

                public String getType() {
                    return AbstractInformedApplicationDeployable.this.getType().toString();
                }

                public List<ServerModuleInfo> getServerModuleInfo() {
                    Set<ModuleInfo> moduleInfos = AbstractInformedApplicationDeployable.this.getModuleInfo();
                    List<ServerModuleInfo> serverModuleInfos = new ArrayList<ServerModuleInfo>(moduleInfos.size());
                    for (final ModuleInfo moduleInfo : moduleInfos) {
                        serverModuleInfos.add(new ServerModuleInfo() {

                            public String getPersonality() {
                                return moduleInfo.getPersonality();
                            }

                            public String getPersonalityIdentifier() {
                                return moduleInfo.getPersonalityIdentifier();
                            }

                            public String getBundleSymbolicName() {
                                return moduleInfo.getUnscopedSymbolicName();
                            }

                            public int compareTo(ServerModuleInfo other) {
                                return this.getBundleSymbolicName().compareTo(other.getBundleSymbolicName());
                            }

                        });
                    }
                    return serverModuleInfos;
                }

                public String getURI() {
                    return AbstractInformedApplicationDeployable.this.getSourceUri().toString();
                }

            }, null);
    }
    
    /** 
     * {@inheritDoc}
     */
    public void undeploy() throws DeploymentException {
        super.undeploy();
        if (this.serverApplicationInfoServiceRegistration != null) {
            this.serverApplicationInfoServiceRegistration.unregister();
        }
    }
    
    public String getApplicationName() {
        return getName() + "-" + getVersion().toString();
    }
}
