/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.deployer.core.internal;

import java.net.URI;
import java.util.Date;
import java.util.HashSet;
import java.util.Set;

import org.osgi.framework.ServiceRegistration;

import com.springsource.server.common.math.Version;
import com.springsource.server.deployer.core.Deployable;
import com.springsource.server.deployer.core.DeploymentException;
import com.springsource.server.deployer.core.InformedDeployable;
import com.springsource.server.deployer.core.ModuleInfo;
import com.springsource.server.deployer.core.ApplicationDeployer.DeploymentOptions;
import com.springsource.server.osgi.framework.OsgiFramework;
import com.springsource.server.osgi.framework.UnableToSatisfyDependenciesException;

/**
 * {@link AbstractInformedDeployable} is a base class for top-level {@link Deployable Deployables}.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * This class is thread safe.
 * 
 * @author Glyn Normington
 * @since 1.0
 */
public abstract class AbstractInformedDeployable implements InformedDeployable {

    private static final String NULL_STRING = null;

    private static final HashSet<ModuleInfo> EMPTY_MODULE_INFO_SET = new HashSet<ModuleInfo>();

    private volatile Date deployTime = null;

    private volatile String name;
    
    private final URI sourceUri;

    private final Type type;
    
    private final DeploymentOptions deploymentOptions;

    private volatile Version version = new Version(NULL_STRING);

    protected final OsgiFramework osgi;

    private ServiceRegistration serverApplicationInfoServiceRegistration;

    public AbstractInformedDeployable(URI sourceLocation, Type type, DeploymentOptions deploymentOptions, OsgiFramework osgi) {
        this.sourceUri = sourceLocation;
        this.type = type;
        this.deploymentOptions = deploymentOptions;
        this.osgi = osgi;
    }

    /**
     * @throws DeploymentException 
     * @throws UnableToSatisfyDependenciesException 
     * {@inheritDoc}
     */
    public void deploy() throws DeploymentException, UnableToSatisfyDependenciesException {
        this.deployTime = new Date(System.currentTimeMillis());
    }

    /**
     * {@inheritDoc}
     */
    public void undeploy() throws DeploymentException {
        if (this.serverApplicationInfoServiceRegistration != null) {
            this.serverApplicationInfoServiceRegistration.unregister();
        }
    }

    /**
     * {@inheritDoc}
     */
    public final Date getDeployTime() {
        return this.deployTime;
    }

    /**
     * 
     * Set the name of this {@link Deployable}.
     * 
     * @param name the name of this {@link Deployable}
     */
    public void setName(String name) {
        this.name = name;
    }

    /**
     * {@inheritDoc}
     */
    public final String getName() {
        return this.name;
    }

    /**
     * {@inheritDoc}
     */
    public final Type getType() {
        return this.type;
    }

    /**
     * 
     * Set the version of this {@link Deployable}.
     * 
     * @param version the version of this {@link Deployable}
     */
    public void setVersion(Version version) {
        this.version = version;
    }

    /**
     * {@inheritDoc}
     */
    public final Version getVersion() {
        return this.version;
    }

    /**
     * {@inheritDoc}
     */
    public Set<ModuleInfo> getModuleInfo() {
        return EMPTY_MODULE_INFO_SET;
    }
    
    /** 
     * {@inheritDoc}
     */
    public URI getSourceUri() {
        return sourceUri;
    }

    /** 
     * {@inheritDoc}
     */
    public final DeploymentOptions getDeploymentOptions() {
        return this.deploymentOptions;
    }

}
