/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.deployer.core.internal;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Reader;
import java.io.Writer;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.LinkedHashMap;
import java.util.Map;

import com.springsource.server.deployer.core.FatalDeploymentException;
import com.springsource.server.deployer.core.ApplicationDeployer.DeploymentOptions;
import com.springsource.server.io.core.PathReference;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * {@link DeployerRecoveryLog} maintains the deployer's recoverable state across Server restarts.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * This class is thread safe.
 * 
 * @author Glyn Normington
 * @since 1.0
 */
final class DeployerRecoveryLog {

    private static final String REDEPLOY_FILE_NAME = "deployed";

    private static final String REDEPLOY_COMPRESSION_FILE_NAME = "deployed.compress";

    private static final int INITIAL_REDEPLOY_DATA_SIZE = 32 * 1024;

    private static final int COMPRESSION_THRESHOLD = 10;

    private static final int COMMAND_LENGTH = 1;

    private static final String DEPLOY_UNOWNED_URI_COMMAND = "+";

    private static final String DEPLOY_OWNED_URI_COMMAND = ">";

    private static final String DEPLOY_NON_RECOVERABLE_URI_COMMAND = "!";

    private static final String UNDEPLOY_URI_COMMAND = "-";

    private static final String URI_SEPARATOR = ";";

    private static final Tracer TRACER = TracerFactory.createTracer(DeployerRecoveryLog.class);

    private static final DeploymentOptions UNOWNED = new DeploymentOptions(true, false);

    private static final DeploymentOptions OWNED = new DeploymentOptions(true, true);
    
    private static final DeploymentOptions NON_RECOVERABLE = new DeploymentOptions(false, false);

    private final PathReference redeployDataset;

    private final PathReference redeployCompressionDataset;
    
    private final long redeployFileLastModified;

    DeployerRecoveryLog(DeployArea deployArea) {
        this.redeployDataset = deployArea.createFile(REDEPLOY_FILE_NAME);
        this.redeployFileLastModified = this.redeployDataset.toFile().lastModified();
        this.redeployCompressionDataset = deployArea.createFile(REDEPLOY_COMPRESSION_FILE_NAME);

        // Recover from a crash during compression
        if (!this.redeployDataset.exists() && this.redeployCompressionDataset.exists()) {
            this.redeployCompressionDataset.copy(this.redeployDataset);
            this.redeployCompressionDataset.delete();
        }
    }

    /**
     * Get the URIs that need to be recovered along with their deployment options.
     * 
     * @return a map of URI to deployment options
     */
    public Map<URI, DeploymentOptions> getRecoveryState() {
        Map<URI, DeploymentOptions> redeploySet = new LinkedHashMap<URI, DeploymentOptions>(20);

        StringBuffer redeployData = new StringBuffer(INITIAL_REDEPLOY_DATA_SIZE);
        Reader redeployDataReader;
        try {
            redeployDataReader = new BufferedReader(new FileReader(redeployDataset.toFile()));
            try {
                char[] chars = new char[INITIAL_REDEPLOY_DATA_SIZE];
                int numRead;
                while (-1 != (numRead = redeployDataReader.read(chars))) {
                    redeployData.append(String.valueOf(chars, 0, numRead));
                }

            } catch (IOException e) {
                TRACER.traceError("Problem reading redeploy dataset", e);
            } finally {
                try {
                    redeployDataReader.close();
                } catch (IOException e) {
                    TRACER.traceError("Problem closing redeploy dataset", e);
                }
            }
        } catch (FileNotFoundException e) {
            // Ignore - this is acceptable if there are no deployed applications
        }

        int recordCount = 0;
        int undeployCount = 0;

        for (String uriCommandString : redeployData.toString().split(URI_SEPARATOR)) {
            recordCount++;
            // Skip zero length command strings as there will typically be one
            // at the end of the dataset.
            if (uriCommandString.length() >= COMMAND_LENGTH) {
                String uriCommand = uriCommandString.substring(0, COMMAND_LENGTH);
                String uriString = uriCommandString.substring(1);
                try {
                    URI uri = new URI(uriString);
                    if (DEPLOY_UNOWNED_URI_COMMAND.equals(uriCommand)) {
                        redeploySet.put(uri, UNOWNED);
                    } else if (DEPLOY_OWNED_URI_COMMAND.equals(uriCommand)) {
                        redeploySet.put(uri, OWNED);
                    } else if (DEPLOY_NON_RECOVERABLE_URI_COMMAND.equals(uriCommand)) {
                        redeploySet.put(uri, NON_RECOVERABLE);
                    } else if (UNDEPLOY_URI_COMMAND.equals(uriCommand)) {
                        undeployCount++;
                        redeploySet.remove(uri);
                    } else {
                        TRACER.traceError("Invalid URI command string '%s' read from redeploy dataset", uriCommandString);
                        // skip and carry on
                    }
                } catch (URISyntaxException e) {
                    TRACER.traceError("Invalid URI command string '%s' read from redeploy dataset", e, uriCommandString);
                    // skip and carry on
                }
            }
        }

        // If there is a significant amount of wasted space in the redeploy
        // dataset, rewrite it.
        if (COMPRESSION_THRESHOLD * undeployCount > recordCount) {
            rewriteRedeploySet(redeploySet);
        }

        return redeploySet;
    }

    /**
     * Write the given set of URIs to the redeploy dataset. To avoid corruption if a crash occurs, write to a redeploy
     * compression file and then switch this for the redeploy dataset.
     * 
     * @param redeploySet the URIs to be written
     */
    private void rewriteRedeploySet(Map<URI, DeploymentOptions> redeploySet) {
        this.redeployCompressionDataset.delete();
        Writer redeployDataWriter;
        try {
            redeployDataWriter = new BufferedWriter(new FileWriter(this.redeployCompressionDataset.toFile()));
            for (URI uri : redeploySet.keySet()) {
                recordUriCommand(redeployDataWriter, uri, getCommandString(redeploySet.get(uri)));
            }
            redeployDataWriter.close();
        } catch (IOException e) {
            TRACER.traceWarning("Problem while rewriting redeploy dataset", e);
            // Return without replacing the redeploy dataset.
        }
        // Now switch the files
        this.redeployDataset.delete();
        this.redeployCompressionDataset.copy(this.redeployDataset);
        this.redeployCompressionDataset.delete();
    }

    /**
     * Add the given location and deployment options to the recovery state.
     * 
     * @param location
     * @param deploymentOptions
     */
    public void add(URI location, DeploymentOptions deploymentOptions) {
        recordUriCommand(location, getCommandString(deploymentOptions));
    }

    /**
     * Return the command string for tagging the type of a log record based on the given deployment options.
     * 
     * @param deploymentOptions
     * @return the command string
     */
    private String getCommandString(DeploymentOptions deploymentOptions) {
        return deploymentOptions.getRecoverable() ? (deploymentOptions.getDeployerOwned() ? DEPLOY_OWNED_URI_COMMAND : DEPLOY_UNOWNED_URI_COMMAND)
            : DEPLOY_NON_RECOVERABLE_URI_COMMAND;
    }

    /**
     * Remove the given location and associated deployment options from the recovery state.
     * 
     * @param location
     */
    public void remove(URI location) {
        recordUriCommand(location, UNDEPLOY_URI_COMMAND);
    }

    private void recordUriCommand(URI uri, String command) {
        Writer writer;
        try {
            writer = new FileWriter(this.redeployDataset.toFile(), true);
            recordUriCommand(writer, uri, command);
            writer.close();
        } catch (IOException e) {
            throw new FatalDeploymentException("Failed to record (un)deployment", e);
        }

    }

    private static void recordUriCommand(Writer writer, URI uri, String command) throws IOException {
        writer.write(command);
        writer.write(uri.toString());
        writer.write(URI_SEPARATOR);
    }

    /**
     * Get the last modified time of the deployer's recovery file. Any applications in the pickup directory with a later last modified time will need
     * to be redeployed.
     * 
     * @return the last modified time of the deployer's recovery file
     */
    public long getRedeployFileLastModified() {
        return redeployFileLastModified;
    }
    
    /**
     * Update the last modified time of the deployer's recovery file.
     */
    public boolean setRedeployFileLastModified() {
        return this.redeployDataset.touch();
    }

}
