/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.deployer.core.module;

import java.util.concurrent.ExecutorService;

import org.osgi.framework.BundleEvent;
import org.osgi.framework.BundleListener;
import org.osgi.framework.SynchronousBundleListener;

import com.springsource.server.deployer.core.DeploymentException;
import com.springsource.server.deployer.core.FatalDeploymentException;
import com.springsource.server.deployer.core.Module;
import com.springsource.server.deployer.core.ModuleRegistry;
import com.springsource.server.osgi.framework.OsgiFramework;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * {@link DeployerBundleListener} listens synchronously for certain bundle events and asynchronously for other events
 * and notifies modules of events on their bundles.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * This class is thread safe.
 * 
 * @author Glyn Normington
 * @since 1.0
 */
class DeployerBundleListener {

    private static final Tracer TRACER = TracerFactory.createTracer(DeployerBundleListener.class);

    private final ExecutorService asyncDeliveryService;

    private static String bundleEventAsString(BundleEvent event) {
        switch (event.getType()) {
            case (BundleEvent.INSTALLED):
                return "INSTALLED";
            case (BundleEvent.LAZY_ACTIVATION):
                return "LAZY_ACTIVATION";
            case (BundleEvent.RESOLVED):
                return "RESOLVED";
            case (BundleEvent.STARTED):
                return "STARTED";
            case (BundleEvent.STARTING):
                return "STARTING";
            case (BundleEvent.STOPPED):
                return "STOPPED";
            case (BundleEvent.STOPPING):
                return "STOPPING";
            case (BundleEvent.UNINSTALLED):
                return "UNINSTALLED";
            case (BundleEvent.UNRESOLVED):
                return "UNRESOLVED";
            case (BundleEvent.UPDATED):
                return "UPDATED";
            default:
                return "???";
        }
    }

    public DeployerBundleListener(OsgiFramework osgi, final ModuleRegistry moduleRegistry, ExecutorService asyncDeliveryService) {
        this.asyncDeliveryService = asyncDeliveryService;
        osgi.getBundleContext().addBundleListener(new SynchronousBundleListener() {

            private volatile boolean shuttingDown = false;

            public void bundleChanged(BundleEvent event) {
                if (event.getBundle().getBundleId() == 0 && event.getType() == BundleEvent.STOPPING) {
                    this.shuttingDown = true;
                } else if (!this.shuttingDown) {
                    Module module = moduleRegistry.getModule(event.getBundle());
                    if (module != null) {
                        try {
                            switch (event.getType()) {
                                case BundleEvent.STARTING:
                                    TRACER.traceInfo("Delivering synchronous bundle event '%s' for bundle with symbolic name '%s' and module '%s'",
                                        bundleEventAsString(event), event.getBundle().getSymbolicName(), module);
                                    module.onStarting();
                                    break;
                                case BundleEvent.STOPPING:
                                    TRACER.traceInfo("Delivering synchronous bundle event '%s' for bundle with symbolic name '%s' and module '%s'",
                                        bundleEventAsString(event), event.getBundle().getSymbolicName(), module);
                                    module.onStopping();
                                    break;
                                case BundleEvent.STOPPED:
                                    TRACER.traceInfo("Delivering synchronous bundle event '%s' for bundle with symbolic name '%s' and module '%s'",
                                        bundleEventAsString(event), event.getBundle().getSymbolicName(), module);
                                    module.onStopped();
                                    break;
                                case BundleEvent.UNRESOLVED:
                                    TRACER.traceInfo("Delivering synchronous bundle event '%s' for bundle with symbolic name '%s' and module '%s'",
                                        bundleEventAsString(event), event.getBundle().getSymbolicName(), module);
                                    module.onUnresolved();
                                    break;
                                default:
                                    break;
                            }
                        } catch (DeploymentException e) {
                            throw new FatalDeploymentException("Synchronous listener detected exception", e);
                        }
                    } else {
                        TRACER.traceInfo("Synchronous bundle event '%s' for bundle with symbolic name '%s' and no associated module",
                            bundleEventAsString(event), event.getBundle().getSymbolicName());
                    }
                }
            }
        });
        osgi.getBundleContext().addBundleListener(new BundleListener() {

            private volatile boolean shuttingDown = false;

            public void bundleChanged(BundleEvent event) {
                if (event.getBundle().getBundleId() == 0 && event.getType() == BundleEvent.STOPPING) {
                    this.shuttingDown = true;
                } else if (!this.shuttingDown) {
                    TRACER.traceDebug("Asynchronous bundle event '%s' for bundle with symbolic name '%s'", bundleEventAsString(event),
                        event.getBundle().getSymbolicName());
                    final Module module = moduleRegistry.getModule(event.getBundle());
                    if (module != null) {
                        switch (event.getType()) {
                            case BundleEvent.STARTED:
                                TRACER.traceInfo("Delivering asynchronous bundle event '%s' for bundle with symbolic name '%s' and module '%s'",
                                    bundleEventAsString(event), event.getBundle().getSymbolicName(), module);
                                DeployerBundleListener.this.asyncDeliveryService.execute(new Runnable() {

                                    public void run() {
                                        try {
                                            module.onStarted();
                                        } catch (DeploymentException e) {
                                            DeployerBundleListener.TRACER.traceError(
                                                "Asynchronous delivery executor detected exception for module '%s'", e, module);
                                            // throw new FatalDeploymentException("Asynchronous listener detected exception", e);
                                        }
                                    }
                                });
                                break;
                            default:
                                break;
                        }
                    } else {
                        TRACER.traceDebug("Asynchronous bundle event '%s' for bundle with symbolic name '%s' and no associated module",
                            bundleEventAsString(event), event.getBundle().getSymbolicName());
                    }
                }
            }
        });
    }

}
