/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.deployer.core.module;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Reader;
import java.io.StringReader;
import java.net.URI;

import org.springframework.core.AttributeAccessorSupport;

import com.springsource.server.deployer.core.DeploymentArtefact;
import com.springsource.server.deployer.core.DeploymentException;
import com.springsource.server.deployer.core.FatalDeploymentException;
import com.springsource.server.deployer.core.InformedDeployable;
import com.springsource.server.deployer.core.ModuleData;
import com.springsource.server.deployer.core.DeploymentArtefact.Entry;
import com.springsource.server.deployer.core.internal.ArtefactNotFoundException;
import com.springsource.server.deployer.core.internal.DeployArea;
import com.springsource.server.deployer.core.internal.ExplodedDirectoryDeploymentArtefact;
import com.springsource.server.deployer.core.parse.DeployerBundleManifest;
import com.springsource.server.io.core.PathReference;
import com.springsource.server.osgi.manifest.BundleManifest;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * Standard implementation of {@link ModuleData}.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Glyn Normington
 * @author Andy Wilkinson
 * @author Sam Brannen
 * @since 1.0
 */
final class StandardModuleData extends AttributeAccessorSupport implements ModuleData {

    private static final String URI_PATH_SEPARATOR = "/";

    private static final String DEFAULT_MANIFEST = "";

    private static final long serialVersionUID = -6613990042596630147L;

    private static final String BUNDLE_MANIFEST_PATH = "META-INF/MANIFEST.MF";

    private static final Tracer TRACER = TracerFactory.createTracer(StandardModuleData.class);

    private final Object monitor = new Object();

    private final DeployArea deployArea;

    private final URI sourceLocation;

    private DeploymentArtefact deploymentArtefact;

    private PathReference explodedBundle;

    private DeployerBundleManifest bundleManifest;

    private PathReference manifestPath;

    public StandardModuleData(InformedDeployable informedDeployable, URI bundleLocation, DeployArea deployArea) throws DeploymentException {
        this.deployArea = deployArea.createWorkArea(informedDeployable.getName() + "-" + informedDeployable.getVersion().toString(), false);
        this.sourceLocation = bundleLocation;
        synchronized (this.monitor) {
            this.explodedBundle = this.deployArea.copyToDeployArea(this.sourceLocation);
            this.deploymentArtefact = new ExplodedDirectoryDeploymentArtefact(this.explodedBundle.toURI());
            this.manifestPath = this.explodedBundle.newChild(BUNDLE_MANIFEST_PATH);
            refreshManifest();
        }
    }

    /**
     * {@inheritDoc}
     */
    public void refreshContents() throws DeploymentException {
        synchronized (this.monitor) {
            this.deployArea.deleteArtefact(this.explodedBundle);
            this.explodedBundle = this.deployArea.copyToDeployArea(this.sourceLocation);
            this.deploymentArtefact = new ExplodedDirectoryDeploymentArtefact(this.explodedBundle.toURI());
            this.manifestPath = this.explodedBundle.newChild(BUNDLE_MANIFEST_PATH);
            refreshManifest();
        }
    }

    /**
     * Read the manifest from file. If no manifest is found, provide a default.
     * 
     * @throws DeploymentException
     */
    private void refreshManifest() throws DeploymentException {
        try {
            Reader manifestReader;
            try {
                manifestReader = new FileReader(this.manifestPath.getAbsolutePath());
            } catch (FileNotFoundException ignore) {
                TRACER.traceInfo("Deployment artefact '%s' had no manifest: default provided", this.deploymentArtefact);
                manifestReader = new StringReader(DEFAULT_MANIFEST);
            }
            this.bundleManifest = new DeployerBundleManifest(manifestReader);
            manifestReader.close();
        } catch (IOException ioe) {
            cleanup();
            throw new FatalDeploymentException("Failed to close manifest reader for deployment artefact '" + this.deploymentArtefact + "'", ioe);
        }
    }

    /**
     * {@inheritDoc}
     */
    public void rename(String newFilename) {
        synchronized (this.monitor) {
            this.explodedBundle = this.deployArea.renameArtefact(this.explodedBundle, newFilename);
            try {
                this.deploymentArtefact = new ExplodedDirectoryDeploymentArtefact(this.explodedBundle.toURI());
            } catch (ArtefactNotFoundException e) {
                throw new FatalDeploymentException("Renamed artefact does not exist", e);
            }
            this.manifestPath = this.explodedBundle.newChild(BUNDLE_MANIFEST_PATH);
        }
    }

    /**
     * {@inheritDoc}
     */
    public void cleanup() {
        synchronized (this.monitor) {
            this.deployArea.clean();
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override public String toString() {
        synchronized (this.monitor) {
            return this.deploymentArtefact.toString();
        }
    }

    /**
     * {@inheritDoc}
     */
    public BundleManifest getBundleManifest() {
        synchronized (this.monitor) {
            return this.bundleManifest;
        }
    }

    /**
     * {@inheritDoc}
     */
    public void replaceManifest() {
        synchronized (this.monitor) {
            try {
                TRACER.traceInfo("Bundle manifest for module '%s' transformed from: \n%s\nto: \n%s", this, this.manifestPath.fileContents(),
                    this.bundleManifest);
            } catch (IOException e) {
                TRACER.traceInfo("Bundle manifest for module '%s' set to:\n%s", this, this.bundleManifest);
            }
            this.manifestPath.delete();
            try {
                FileWriter manifestWriter;
                try {
                    manifestWriter = new FileWriter(this.manifestPath.getCanonicalPath());
                } catch (IOException ignore) {
                    // Ensure META-INF exists and try again.
                    new File(this.manifestPath.getParent().getCanonicalPath()).mkdir();
                    manifestWriter = new FileWriter(this.manifestPath.getCanonicalPath());
                }
                this.bundleManifest.write(manifestWriter);
                manifestWriter.close();
            } catch (IOException e) {
                throw new FatalDeploymentException("Unable to write updated manifest of bundle for module '" + this + "'.", e);
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    public Entry getEntry(String path) {
        synchronized (this.monitor) {
            return this.deploymentArtefact.getEntry(path);
        }
    }

    /**
     * {@inheritDoc}
     */
    public File getFile() {
        synchronized (this.monitor) {
            return this.deploymentArtefact.getFile();
        }
    }

    /**
     * {@inheritDoc}
     */
    public String getFilename() {
        synchronized (this.monitor) {
            return this.deploymentArtefact.getFilename();
        }
    }

    public int hashCode() {
        return this.sourceLocation.hashCode();
    }
    
    public boolean equals(Object other) {
        if(other == this) {
            return true;
        }
        if(other instanceof StandardModuleData) {
            StandardModuleData that = (StandardModuleData) other;
            return super.equals(that) && this.sourceLocation.equals(that.sourceLocation);
        }
        return false;
    }

    /** 
     * {@inheritDoc}
     */
    public void updateResource(URI resourceLocation, String targetPath) {
        this.deploymentArtefact.copy(new PathReference(resourceLocation), URI_PATH_SEPARATOR + targetPath);
    }
    
    /** 
     * {@inheritDoc}
     */
    public void deleteResource(String targetPath) {
        this.deploymentArtefact.getEntry(URI_PATH_SEPARATOR + targetPath).delete();
    }

}
