/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.deployer.lib;

import java.io.File;
import java.net.URI;

import com.springsource.server.common.math.Version;
import com.springsource.server.deployer.core.CompletableDeploymentContext;
import com.springsource.server.deployer.core.DeploymentException;
import com.springsource.server.deployer.core.ModuleDeployerSet;
import com.springsource.server.deployer.core.ApplicationDeployer.DeploymentOptions;
import com.springsource.server.deployer.core.internal.AbstractInformedDeployable;
import com.springsource.server.osgi.framework.BundleInstaller;
import com.springsource.server.osgi.framework.OsgiFramework;
import com.springsource.server.osgi.framework.UnableToSatisfyDependenciesException;
import com.springsource.server.osgi.provisioning.LibraryDefinition;
import com.springsource.server.osgi.provisioning.support.LibraryDefinitionFactory;

/**
 * This class represents a deployable library or, equivalently, library descriptor.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * This class is thread safe provided the given <code>BundleInstaller</code> is thread safe.
 * 
 * @author Glyn Normington
 * @since 1.0
 */
final class Library extends AbstractInformedDeployable {

    private final BundleInstaller bundleInstaller;

    private final File libraryFile;

    private LibraryDefinition libraryDefinition;

    public Library(BundleInstaller bundleInstaller, URI location, URI sourceLocation, DeploymentOptions deploymentOptions, OsgiFramework osgi) {
        super(sourceLocation, Type.LIBRARY, deploymentOptions, osgi);
        this.bundleInstaller = bundleInstaller;
        this.libraryFile = new File(location);
        this.libraryDefinition = LibraryDefinitionFactory.createLibraryDefinition(libraryFile);
        setName(this.libraryDefinition.getSymbolicName());
        setVersion(new Version(this.libraryDefinition.getVersion())); // convert from OSGi framework version to server version wrapper
    }

    /** 
     * @throws UnableToSatisfyDependenciesException 
     * {@inheritDoc}
     */
    public void deploy() throws DeploymentException, UnableToSatisfyDependenciesException {
        this.bundleInstaller.getBundleRepository().addLibrary(this.libraryFile);
        super.deploy();
    }

    /** 
     * {@inheritDoc}
     */
    public void start() {
        throw new UnsupportedOperationException("Library start is not supported");
    }

    /** 
     * {@inheritDoc}
     */
    public void stop() {
        throw new UnsupportedOperationException("Library stop is not supported");
    }

    /** 
     * {@inheritDoc}
     */
    public void undeploy() throws DeploymentException {
        super.undeploy();
        throw new UnsupportedOperationException("Library undeploy is not supported");
    }

    /**
     * {@inheritDoc}
     */
    @Override public String toString() {
        return String.format("%s - %s, %s, %s", getClass().getSimpleName(), this.libraryDefinition == null ? "" : this.libraryDefinition.getSymbolicName(),
            this.libraryDefinition == null ? "" : this.libraryDefinition.getVersion().toString(), this.libraryFile.getAbsolutePath());
    }

    /** 
     * {@inheritDoc}
     */
    public boolean refresh(String symbolicName, ModuleDeployerSet moduleDeployerSet, CompletableDeploymentContext deploymentContext) {
        // Library refresh is not supported.
        throw new UnsupportedOperationException();
    }

    /** 
     * {@inheritDoc}
     */
    public void deleteResource(String symbolicName, String targetPath) {
        throw new UnsupportedOperationException("Library resource deletion is not supported");
    }

    /** 
     * {@inheritDoc}
     */
    public void updateResource(String symbolicName, URI resourceLocation, String targetPath) {
        throw new UnsupportedOperationException("Library resource update is not supported");
    }

}
