/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.deployer.hot;

import java.io.File;
import java.util.concurrent.TimeUnit;

import com.springsource.server.bootstrap.BootstrapConfiguration;
import com.springsource.server.concurrent.core.ThreadManager;
import com.springsource.server.deployer.core.ApplicationDeployer;
import com.springsource.server.deployer.core.FatalDeploymentException;
import com.springsource.server.io.core.PathReference;
import com.springsource.server.io.core.fs.FileSystemWatcher;
import com.springsource.server.serviceability.ffdc.NonNull;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * Handles hot deployment of application artefacts in the Server.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @author Glyn Normington
 * @since 1.0
 */
public final class HotDeployer {

    private static final String EXCLUDE_PATTERN = "\\.DS_Store";

    private static final String DEFAULT_PICKUP_DIRECTORY = "." + File.separator + "target" + File.separator + "pickup" + File.separator;

    private static final Tracer TRACER = TracerFactory.createTracer(HotDeployer.class);

    private final ApplicationDeployer applicationDeployer;
    /**
     * The {@link FileSystemWatcher} used to monitor the pickup location.
     */
    private final FileSystemWatcher watcher;

    private final File pickupDir;

    /**
     * Creates a new <code>HotDeployer</code>.
     * 
     * @param deployer the {@link ApplicationDeployer} to deploy to.
     * @param threadManager the {@link ThreadManager}.
     * @param configuration the {@link BootstrapConfiguration} containing the work directory config.
     */
    public HotDeployer(@NonNull ApplicationDeployer deployer, @NonNull ThreadManager threadManager, @NonNull BootstrapConfiguration configuration) {
        this.applicationDeployer = deployer;
        this.pickupDir = createHotDeployDir(configuration);
        this.watcher = new FileSystemWatcher(this.pickupDir, threadManager, EXCLUDE_PATTERN);
        this.watcher.addListener(new HotDeploymentFileSystemListener(deployer));
    }

    /**
     * Creates the hot deploy directory.
     * 
     * @param configuration the {@link BootstrapConfiguration}, defines the location of the work directory.
     * @return the {@link PathReference} to the deploy directory.
     */
    private File createHotDeployDir(BootstrapConfiguration configuration) {
        String pickUpDirectory = configuration.getPickUpDirectory();
        if (pickUpDirectory == null) {
            pickUpDirectory = DEFAULT_PICKUP_DIRECTORY;
        }
        PathReference dir = new PathReference(pickUpDirectory);

        if (dir.isFile()) {
            TRACER.traceDebug("Deleting stray file from hot deployment directory location '%s'.", dir.getAbsolutePath());
            dir.delete();
        }
        if (!dir.exists()) {
            TRACER.traceInfo("Creating hot deployment directory at '%s'.", dir.getAbsolutePath());
            dir.createDirectory();
        } else {
            TRACER.traceInfo("Using hot deployment directory at '%s'.", dir.getAbsolutePath());
        }
        return dir.toFile();
    }

    /**
     * Instructs the hot deployer to begin watching the pickup directory for deployment events.
     */
    public void start() {
        while (!awaitDeployerRecovery()) {
            TRACER.traceWarning("Deployer recovery not complete - pickup directory monitoring delayed.");
        }
        this.watcher.start();
        TRACER.traceInfo("Hot deployer started watching");
    }

    /**
     * Instructs the hot deployer to cease monitoring the pickup directory.
     */
    public void stop() {
        this.watcher.stop();
    }

    private boolean awaitDeployerRecovery() {
        try {
            return this.applicationDeployer.awaitRecovery(60, TimeUnit.SECONDS);
        } catch (InterruptedException e) {
            throw new FatalDeploymentException("Unable to wait for recovery.", e);
        }
    }
    /**
     * {@inheritDoc}
     */
    public String toString() {
        return String.format("Hot Deployer [pickupDir = %s]", this.pickupDir.getAbsolutePath());
    }
}
