/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.deployer.hot;

import java.io.File;
import java.net.URI;
import java.net.URISyntaxException;

import com.springsource.server.deployer.core.ApplicationDeployer;
import com.springsource.server.deployer.core.DeployerLogEvent;
import com.springsource.server.deployer.core.DeploymentException;
import com.springsource.server.deployer.core.DeploymentIdentity;
import com.springsource.server.deployer.core.FatalDeploymentException;
import com.springsource.server.deployer.core.ApplicationDeployer.DeploymentOptions;
import com.springsource.server.io.core.PathReference;
import com.springsource.server.io.core.fs.FileSystemEvent;
import com.springsource.server.io.core.fs.FileSystemListener;
import com.springsource.server.serviceability.ffdc.NonNull;
import com.springsource.server.serviceability.logging.LogEvent;
import com.springsource.server.serviceability.logging.LogService;
import com.springsource.server.serviceability.logging.LogServiceFactory;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * {@link FileSystemListener} that monitors a pickup directory for file system events. When a file is created it is
 * passed to the {@link ApplicationDeployer} for deployment. When a file is modified, it is re-deployed. When a file is
 * deleted, the application is undeployed.
 * <p />
 * The <code>ApplicationDeployer</code> is responsible for recovering the files deployed via this route and is given
 * ownership of these files so that undeploying one of them, e.g. by name and version, will delete the corresponding
 * file in the pickup directory.
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @author Glyn Normington
 * @author Andy Wilkinson
 * @since 1.0
 */
final class HotDeploymentFileSystemListener implements FileSystemListener {

    private static final Tracer TRACER = TracerFactory.createTracer(HotDeploymentFileSystemListener.class);

    private static final LogService LOG = LogServiceFactory.getLogService();

    private final ApplicationDeployer deployer;

    /**
     * Creates a new <code>HotDeploymentFileSystemListener</code>.
     * 
     * @param deployer the {@link ApplicationDeployer} to deploy to.
     */
    public HotDeploymentFileSystemListener(@NonNull ApplicationDeployer deployer) {
        this.deployer = deployer;
    }

    /**
     * {@inheritDoc}
     * 
     * Reacts to changes in the pickup directory and calls the {@link ApplicationDeployer} as appropriate.
     */
    public void onChange(String path, FileSystemEvent event) {
        PathReference sourceArtefact = new PathReference(path);
        LOG.log(DeployerLogEvent.HOT_DEPLOY_PROCESSING_FILE, event, sourceArtefact.getName());
        try {
            if (event == FileSystemEvent.CREATED) {
                TRACER.traceInfo("ApplicationDeploying file '%s'.", path);
                deploy(path);
            } else if (event == FileSystemEvent.MODIFIED) {
                TRACER.traceInfo("Redeploying file '%s'.", path);
                deploy(path);
            } else if (event == FileSystemEvent.DELETED) {
                TRACER.traceInfo("ApplicationUndeploying file '%s'.", path);
                undeploy(path);
            } else if (event == FileSystemEvent.INITIAL) {
                TRACER.traceInfo("ApplicationConditionallyDeploying file '%s'.", path);
                deployIfNotDeployed(path);
            }
        } catch (Exception ex) {
            determineFailureAndLogMessage(event, path, ex);
        }
    }

    /**
     * Determines the {@link LogEvent} that corresponds the {@link FileSystemEvent}.
     */
    private void determineFailureAndLogMessage(FileSystemEvent event, String path, Exception ex) {
        switch (event) {
            case CREATED:
            case INITIAL:
                LOG.log(DeployerLogEvent.HOT_DEPLOY_FAILED, path);
                break;
            case MODIFIED:
                if (ex instanceof DeploymentException) {
                    LOG.log(DeployerLogEvent.HOT_REDEPLOY_FAILED, path);
                } else {
                    LOG.log(DeployerLogEvent.HOT_REDEPLOY_FAILED, ex, path);
                }
                break;
            default:
                if (ex instanceof DeploymentException) {
                    LOG.log(DeployerLogEvent.HOT_UNDEPLOY_FAILED, path);
                } else {
                    LOG.log(DeployerLogEvent.HOT_UNDEPLOY_FAILED, ex, path);
                }
        }
    }

    /**
     * Undeploys the application that corresponds to the supplied source artefact URI.
     * 
     * @param sourceArtefact the source artefact URI string
     * @throws DeploymentException
     */
    private void undeploy(String sourceArtefact) throws DeploymentException {
        DeploymentIdentity deploymentIdentity = getDeploymentIdentity(sourceArtefact);
        if (deploymentIdentity != null) {
            this.deployer.undeploy(deploymentIdentity);
        }
    }

    /**
     * Get the {@link DeploymentIdentity} of the given artefact. Return <code>null</code> if the given artefact is not
     * currently deployed.
     * 
     * @param sourceArtefact the source artefact URI string
     * @return the <code>DeploymentIdentity</code> of the given artefact or <code>null</code>
     */
    private DeploymentIdentity getDeploymentIdentity(String sourceArtefact) {
        return this.deployer.getDeploymentIdentity(getDefinitiveUri(sourceArtefact));
    }
    
    /**
     * Determine whether or not the given artefact is already deployed. Return <code>true</code> if the given artefact at its file's last modified time
     * is already deployed.
     * 
     * @param sourceArtefact the source artefact URI string
     * @return <code>true</code> if and only if the given artefact at its file's last modified time is already deployed
     */
    private boolean isDeployed(String sourceArtefact) {
        return this.deployer.isDeployed(getDefinitiveUri(sourceArtefact));
    }

    /**
     * Converts a string URI to a URI with a predictable format, particularly in the case where the string URI ends in a
     * file separator.
     * 
     * @param sourceArtefact the URI string
     * @return
     */
    private URI getDefinitiveUri(String sourceArtefact) {
        URI baseUri = new File(sourceArtefact).toURI();
        if (sourceArtefact.endsWith(File.separator) && !baseUri.toString().endsWith("/")) {
            try {
                baseUri = new URI(baseUri.toString() + "/");
            } catch (URISyntaxException e) {
                new FatalDeploymentException("Unexpected URI syntax problem.", e);
            }
        }
        return baseUri;
    }

    /**
     * Deploys the application at the supplied PathReference.
     * 
     * @param sourceArtefact the source artefact URI string
     * @throws DeploymentException
     */
    private void deploy(String sourceArtefact) throws DeploymentException {
        this.deployer.deploy(getDefinitiveUri(sourceArtefact), new DeploymentOptions(true, true));
    }

    /**
     * Deploys the application at the supplied PathReference if it is not already deployed.
     * 
     * @param sourceArtefact the source artefact URI string
     * @throws DeploymentException
     */
    private void deployIfNotDeployed(String sourceArtefact) throws DeploymentException {
        if (!isDeployed(sourceArtefact)) {
            deploy(sourceArtefact);
        } else {
            LOG.log(DeployerLogEvent.HOT_DEPLOY_SKIPPED, new PathReference(sourceArtefact).getName());
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override public String toString() {
        return String.format("Hot Deploy Listener");
    }
}
