/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.ffdc.core.internal;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.io.Writer;

import org.springframework.util.ObjectUtils;

import com.springsource.json.writer.JSONArray;
import com.springsource.json.writer.JSONObject;
import com.springsource.server.serviceability.ffdc.DumpWriter;
import com.springsource.server.serviceability.output.Level;
import com.springsource.server.serviceability.tracing.EntryExitTraceLevel;

/**
 * 
 * JSONDumpWriter formats the dump using the {@link JSONObject}
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Not threadsafe.
 * 
 * @author Damilola Senbanjo
 * @author Rob Harrop
 * @since 1.0
 */

public class JSONDumpWriter implements DumpWriter {

    final JSONObject root;

    final JSONObject main;

    /**
     * Constructs a new {@link JSONDumpWriter} containing two JSONObjects. The root JSONObject encompasses the main one.
     */
    public JSONDumpWriter() {
        this.root = new JSONObject();
        this.main = new JSONObject();
        this.root.put("dump", this.main);
    }

    /**
     * {@inheritDoc}
     */
    @EntryExitTraceLevel(Level.VERBOSE)
    public void write(Writer writer) {
        this.root.write(writer, 3);
    }

    /**
     * {@inheritDoc}
     */
    @EntryExitTraceLevel(Level.VERBOSE)
    public SectionWriter startSection(String name) {
        JSONObject section = new JSONObject();

        if (name == null) {
            throw new FatalDumpException("Please provide a valid section name");
        } else {
            this.main.put(name, section);
        }

        return new SectionWriterImpl(section);
    }

    /**
     * 
     * Static inner class SectionWriterImpl is an implementation of the {@link SectionWriter}. This class is
     * responsible for beginning each new section.
     * <p />
     * 
     * <strong>Concurrent Semantics</strong><br />
     * 
     * Not threadsafe.
     * 
     * @author Damilola Senbanjo
     * @since 1.0
     */
    protected static class SectionWriterImpl implements SectionWriter {

        private final JSONObject section;

        public SectionWriterImpl(JSONObject section) {
            this.section = section;
        }

        /**
         * {@inheritDoc}
         */
        @EntryExitTraceLevel(Level.VERBOSE)
        public SectionWriter scalar(String key, Object value) {
            if (value instanceof Throwable) {
                StringWriter sw = new StringWriter();
                ((Throwable) value).printStackTrace(new PrintWriter(sw));
                value = sw.toString();
            }

            if (key == null) {
                throw new FatalDumpException("Please provide valid information to write");
            } else {
                this.section.accumulate(key, ObjectUtils.nullSafeToString(value));
            }
            return this;
        }

        /**
         * {@inheritDoc}
         */
        @EntryExitTraceLevel(Level.VERBOSE)
        public ListWriter<SectionWriter> list(String name) {
            JSONArray array = new JSONArray();

            if (name == null) {
                throw new FatalDumpException("Please provide a valid list name");
            } else {
                this.section.put(name, array);
            }
            return new ListWriterImpl<SectionWriter>(array, this);
        }

        /**
         * {@inheritDoc}
         */
        @EntryExitTraceLevel(Level.VERBOSE)
        public MapWriter<SectionWriter> map(String name) {
            JSONObject object = new JSONObject();

            if (name == null) {
                throw new FatalDumpException("Please provide a valid map name");
            } else {
                this.section.put(name, object);
            }
            return new MapWriterImpl<SectionWriter>(object, this);
        }

        /**
         * {@inheritDoc}
         */
        @EntryExitTraceLevel(Level.VERBOSE)
        public void end() {

        }
    }

    /**
     * 
     * Static inner class ObjectWriterImpl is an implementation of the {@link MapWriter}
     * <p />
     * 
     * <strong>Concurrent Semantics</strong><br />
     * 
     * Not threadsafe.
     * 
     * @author Damilola Senbanjo
     * @since 1.0
     */
    protected static class MapWriterImpl<T> implements MapWriter<T> {

        private final JSONObject object;

        private final T parent;

        public MapWriterImpl(JSONObject object, T parent) {
            this.object = object;
            this.parent = parent;
        }

        /**
         * {@inheritDoc}
         */
        @EntryExitTraceLevel(Level.VERBOSE)
        public MapWriter<T> scalar(String name, Object val) {

            if (val instanceof Throwable) {
                StringWriter sw = new StringWriter();
                ((Throwable) val).printStackTrace(new PrintWriter(sw));
                val = sw.toString();
            }

            if (name == null) {
                throw new FatalDumpException("Please provide valid information to write");
            } else {
                this.object.put(name, ObjectUtils.nullSafeToString(val));
            }
            return this;
        }

        /**
         * {@inheritDoc}
         */
        @EntryExitTraceLevel(Level.VERBOSE)
        public ListWriter<MapWriter<T>> list(String name) {
            JSONArray array = new JSONArray();

            if (name == null) {
                throw new FatalDumpException("Please provide a valid list name");
            } else {
                this.object.put(name, array);
            }
            return new ListWriterImpl<MapWriter<T>>(array, this);
        }

        /**
         * {@inheritDoc}
         */
        @EntryExitTraceLevel(Level.VERBOSE)
        public T end() {
            return this.parent;
        }

        /**
         * {@inheritDoc}
         */
        @EntryExitTraceLevel(Level.VERBOSE)
        public MapWriter<MapWriter<T>> map(String key) {
            JSONObject newObject = new JSONObject();

            if (key == null) {
                throw new FatalDumpException("Map name cannot be null");
            } else {
                this.object.put(key, newObject);
            }
            return new MapWriterImpl<MapWriter<T>>(newObject, this);

        }
    }

    /**
     * 
     * Static inner class ListWriterImpl is an implementation of the {@link ListWriter} interface
     * <p />
     * 
     * <strong>Concurrent Semantics</strong><br />
     * 
     * Not threadsafe.
     * 
     * @author Damilola Senbanjo
     * @since 1.0
     */
    protected static class ListWriterImpl<T> implements ListWriter<T> {

        private final JSONArray array;

        private final T parent;

        public ListWriterImpl(JSONArray array, T parent) {
            this.array = array;
            this.parent = parent;
        }

        /**
         * {@inheritDoc}
         */
        @EntryExitTraceLevel(Level.VERBOSE)
        public ListWriter<T> scalar(Object value) {

            if (value instanceof Throwable) {
                StringWriter sw = new StringWriter();
                ((Throwable) value).printStackTrace(new PrintWriter(sw));
                value = sw.toString();
            }

            if (value != null) {
                this.array.put(value);
            }
            return this;
        }

        /**
         * {@inheritDoc}
         */
        @EntryExitTraceLevel(Level.VERBOSE)
        public MapWriter<ListWriter<T>> map() {
            JSONObject object = new JSONObject();
            this.array.put(object);
            return new MapWriterImpl<ListWriter<T>>(object, this);
        }

        /**
         * {@inheritDoc}
         */
        @EntryExitTraceLevel(Level.VERBOSE)
        public T end() {
            return this.parent;
        }

        /**
         * {@inheritDoc}
         */
        @EntryExitTraceLevel(Level.VERBOSE)
        public ListWriter<ListWriter<T>> list() {
            JSONArray newArray = new JSONArray();
            this.array.put(newArray);
            return new ListWriterImpl<ListWriter<T>>(newArray, this);
        }
    }

}
