/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.ffdc.core.internal;

import static com.springsource.server.ffdc.core.internal.FFDCLogEvent.DUMP_FILE_WRITTEN;
import static com.springsource.server.ffdc.core.internal.FFDCLogEvent.FAILED_TO_LOCATE_DUMP_DIRECTORY;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Set;

import com.springsource.server.bootstrap.BootstrapConfiguration;
import com.springsource.server.serviceability.ffdc.DumpExtension;
import com.springsource.server.serviceability.ffdc.DumpGenerator;
import com.springsource.server.serviceability.ffdc.DumpWriter;
import com.springsource.server.serviceability.ffdc.NonNull;
import com.springsource.server.serviceability.ffdc.DumpWriter.SectionWriter;
import com.springsource.server.serviceability.logging.LogService;
import com.springsource.server.serviceability.logging.LogServiceFactory;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * Standard implementation of the {@link DumpGenerator} which creates and writes to a dump file
 * 
 * <p/>
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * Not thread-safe.
 * 
 * @author Damilola Senbanjo
 * @author Sam Brannen
 * @since 1.0
 */
public class StandardDumpGenerator implements DumpGenerator {

    private final DateFormat fileNameDateFormat = new SimpleDateFormat("yyyy-MM-dd_HH-mm-ss");

    private final DateFormat dateFormat = DateFormat.getDateInstance();

    private final DateFormat timeFormat = DateFormat.getTimeInstance(DateFormat.LONG);

    private static final LogService LOG_SERVICE = LogServiceFactory.getLogService();

    private final String dumpDir;

    private final Set<DumpExtension> dumpExtensions;

    /**
     * Constructs a new implementation of {@link DumpGenerator} which takes a subsystem registry as one of its argument
     * 
     * @param bootstrapConfiguration configuration required to launch the Server
     * @param dumpExtensions set of dump extensions
     */
    public StandardDumpGenerator(BootstrapConfiguration bootstrapConfiguration, Set<DumpExtension> dumpExtensions) {
        this(dumpExtensions, lookupDumpDir(bootstrapConfiguration));
    }

    /**
     * Constructs a new implementation of a {@link DumpGenerator} which takes a dump directory as one of its arguments
     * 
     * @param dumpExtensions set of dump extensions
     * @param dumpDir the directory to which the dump file will be saved
     */
    public StandardDumpGenerator(Set<DumpExtension> dumpExtensions, String dumpDir) {
        this.dumpDir = dumpDir;
        this.dumpExtensions = dumpExtensions;
    }

    /**
     * Finds the dump directory for the given bootstrap configuration
     * 
     * @param bootstrapConfiguration configuration required to launch the Server
     * @return the bootstrap configuration's dump directory
     */
    private static String lookupDumpDir(BootstrapConfiguration bootstrapConfiguration) {

        if (bootstrapConfiguration.getDumpDirectory() == null) {
            LOG_SERVICE.log(FAILED_TO_LOCATE_DUMP_DIRECTORY, bootstrapConfiguration.getDumpDirectory());
            throw new FatalDumpException("Dump directory is null");
        }

        File dir = new File(bootstrapConfiguration.getDumpDirectory());

        if (!dir.exists() && !dir.mkdirs()) {
            throw new FatalDumpException("Dump directory [" + dir + "] does not exist and could not be created.");
        }

        try {
            return dir.getCanonicalPath();
        } catch (IOException e) {
            LOG_SERVICE.log(FAILED_TO_LOCATE_DUMP_DIRECTORY, e, dir);
            throw new FatalDumpException("Failed to get canonical path for dump directory [" + dir + "]", e);
        }
    }

    /**
     * Protected method used by the {@link StandardDumpGeneratorTests}'s inner class DumpGeneratorDummy to access the
     * date
     * 
     * @return new date
     */
    protected Date crashDate() {
        return new Date();
    }

    /**
     * Writes the supplied {@link Throwable cause} to the provided {@link DumpWriter}.
     * 
     * <p>
     * If the supplied cause is <code>null</code>, calling this method will not write to the dump writer.
     * </p>
     * 
     * @param writer the DumpWriter to which to write; may not be <code>null</code>
     * @param cause the cause of the need to generate a dump; may be <code>null</code>
     */
    public void writeCause(@NonNull DumpWriter writer, Throwable cause) {
        if (cause != null) {
            writer.startSection("Exception")//
            .scalar("Cause", cause);
            Tracer tracer = TracerFactory.createTracer(getClass());
            tracer.traceError("Exception trace - ", cause);
        }
    }

    /**
     * Gets the directory where the dump file will be saved
     * 
     * @return dump file directory
     */
    private String getDumpDirectory() {
        try {
            if (this.dumpDir == null) {
                throw new FatalDumpException("Unspecified dump directory - FFDC failed");
            } else if (this.dumpDir.endsWith(File.separator)) {
                return this.dumpDir;
            } else {
                return this.dumpDir + File.separator;
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
        return "";
    }

    /**
     * {@inheritDoc}
     */
    public final void createDumpFile(Throwable cause) {
        try {
            Date crashDate = crashDate();
            String dumpFilename = getDumpDirectory() + fileNameDateFormat.format(crashDate) + ".dump";
            Writer writer = createWriter(dumpFilename);

            JSONDumpWriter jsonWriter = new JSONDumpWriter();
            writeFileHeader(jsonWriter, crashDate);
            writeCause(jsonWriter, cause);
            writeExtensions(jsonWriter);
            jsonWriter.write(writer);

            writer.flush();
            writer.close();
            LOG_SERVICE.log(DUMP_FILE_WRITTEN, cause, dumpFilename);

        } catch (IOException ioe) {
            try {
                Tracer tracer = TracerFactory.createTracer(getClass());
                tracer.traceError("Failed to create dump file.", ioe);
            } catch (Exception e) {
                // printing original cause of failure.
                ioe.printStackTrace();
            }
        }
    }

    /**
     * Creates a writer from a file
     * 
     * @param dumpFilename name of the file to be written to
     * @return a {@link Writer}
     * @throws IOException
     */
    protected Writer createWriter(String dumpFilename) throws IOException {
        return new FileWriter(dumpFilename);
    }

    /**
     * This method writes the file header, including the date, time, thread name, version of the server, and number of
     * contributing subsystems.
     * 
     * @param dumpWriter the {@link DumpWriter} to write to
     * @param crashDate the crash date/time-stamp
     */
    protected void writeFileHeader(DumpWriter dumpWriter, Date crashDate) throws IOException {
        dumpWriter.startSection("File Header")//
        .scalar("Date", this.dateFormat.format(crashDate))//
        .scalar("Time", this.timeFormat.format(crashDate.getTime()))//
        .scalar("Thread Name", Thread.currentThread().getName())//
        .scalar("Server Version", "1.0.0")//
        .scalar("No of contributing subsystems", this.dumpExtensions.size())//
        .end();
    }

    /**
     * Writes the dump extensions to the dump file
     * 
     * @param dw the {@link DumpWriter} to write to
     */
    protected void writeExtensions(DumpWriter dw) {
        for (DumpExtension extension : this.dumpExtensions) {
            String subsystemName = extension.getName();
            SectionWriter sw = dw.startSection("Subsystem '" + subsystemName + "'");
            try {
                extension.dump(sw);
            } catch (Throwable t) {
                try {
                    String msg = "Exception encountered during dumping of subsystem '" + subsystemName + "'";

                    Tracer tracer = TracerFactory.createTracer(getClass());
                    tracer.traceError(msg, t);

                    dw.startSection(msg) //
                    .scalar("Exception message", t.getMessage())//
                    .scalar("Exception cause", t);
                } catch (Exception e) {
                    // printing original cause of failure.
                    t.printStackTrace();
                }
            }
        }
    }

}
