/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.ffdc.core.internal;

import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.ServiceRegistration;
import org.springframework.beans.factory.DisposableBean;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.context.ApplicationContext;
import org.springframework.osgi.context.BundleContextAware;
import org.springframework.osgi.context.ConfigurableOsgiBundleApplicationContext;
import org.springframework.osgi.context.event.OsgiBundleApplicationContextEvent;
import org.springframework.osgi.context.event.OsgiBundleApplicationContextListener;
import org.springframework.osgi.context.event.OsgiBundleContextFailedEvent;

import com.springsource.server.kernel.core.Subsystem;
import com.springsource.server.kernel.core.SubsystemRegistry;
import com.springsource.server.serviceability.ffdc.FFDCService;
import com.springsource.server.serviceability.logging.LogService;
import com.springsource.server.serviceability.logging.LogServiceFactory;

/**
 * Monitors subsystem bundles for failures during {@link ApplicationContext} creation. When a failure is detected, a
 * dump is triggered. <p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
final class SubsystemMonitor implements BundleContextAware, InitializingBean, DisposableBean {

    private final LogService LOG = LogServiceFactory.getLogService();

    private final SubsystemRegistry subsystemRegistry;

    private final FFDCService ffdcService;

    private volatile BundleContext bundleContext;

    private volatile ServiceRegistration listenerRegistration;

    /**
     * Creates a new <code>SubsystemMonitor</code>.
     * 
     * @param subsystemRegistry the {@link SubsystemRegistry}.
     * @param ffdcService the {@link FFDCService}.
     */
    public SubsystemMonitor(SubsystemRegistry subsystemRegistry, FFDCService ffdcService) {
        this.subsystemRegistry = subsystemRegistry;
        this.ffdcService = ffdcService;
    }

    public static Bundle getOsgiBundle(OsgiBundleApplicationContextEvent event) {
        Bundle bundle = null;
        ApplicationContext applicationContext = event.getApplicationContext();
        if (applicationContext instanceof ConfigurableOsgiBundleApplicationContext) {
            ConfigurableOsgiBundleApplicationContext osgiBundleApplicationContext = (ConfigurableOsgiBundleApplicationContext) applicationContext;
            bundle = osgiBundleApplicationContext.getBundle();
        }
        return bundle;
    }

    /**
     * {@inheritDoc}
     */
    public void setBundleContext(BundleContext bundleContext) {
        this.bundleContext = bundleContext;
    }

    /**
     * {@inheritDoc}
     */
    public void afterPropertiesSet() throws Exception {
        this.listenerRegistration = this.bundleContext.registerService(OsgiBundleApplicationContextListener.class.getName(),
            new SubsystemFailureListener(), null);
    }

    /**
     * {@inheritDoc}
     */
    public void destroy() throws Exception {
        if (this.listenerRegistration != null) {
            this.listenerRegistration.unregister();
        }
    }

    /**
     * {@link OsgiBundleApplicationContextListener} for subsystem bundle failure detection.<p/>
     * 
     * <strong>Concurrent Semantics</strong><br />
     * 
     * Threadsafe.
     * 
     * @author Rob Harrop
     * @since 1.0
     */
    final class SubsystemFailureListener implements OsgiBundleApplicationContextListener {

        /**
         * {@inheritDoc}
         */
        public void onOsgiApplicationEvent(OsgiBundleApplicationContextEvent event) {
            if (event instanceof OsgiBundleContextFailedEvent) {
                OsgiBundleContextFailedEvent failedEvent = (OsgiBundleContextFailedEvent) event;
                Bundle bundle = getOsgiBundle(event);
                Subsystem subsystem = SubsystemMonitor.this.subsystemRegistry.getSubsystemByBundle(bundle);
                if (subsystem != null) {
                    SubsystemMonitor.this.LOG.log(FFDCLogEvent.SUBSYSTEM_BUNDLE_FAILED, failedEvent.getFailureCause(), bundle.getSymbolicName(),
                        subsystem.getName());
                    SubsystemMonitor.this.ffdcService.generateDump(failedEvent.getFailureCause());
                }
            }
        }

    }
}
