/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.io.core;

import java.io.File;
import java.io.IOException;
import java.net.URI;

import com.springsource.server.serviceability.ffdc.Assert;
import com.springsource.server.serviceability.ffdc.NonNull;
import com.springsource.server.serviceability.output.Level;
import com.springsource.server.serviceability.tracing.EntryExitTraceLevel;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * Utility methods for dealing with the file system.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Thread-safe.
 * 
 * @author Rob Harrop
 * @author Sam Brannen
 * @author Glyn Normington
 * @since 1.0
 */
public final class FileSystemUtils {

    private static final Tracer TRACER = TracerFactory.createTracer(FileSystemUtils.class);

    private static final String URI_FILE_SCHEME = "file";

    private FileSystemUtils() {
        /* no-op */
    }

    /**
     * Alias for {@link #convertToCanonicalPath(String, boolean) convertToCanonicalPath(path, false) }.
     * 
     * @see #convertToCanonicalPath(String, boolean)
     */
    public static String convertToCanonicalPath(String path) throws FatalIOException {
        return convertToCanonicalPath(path, false);
    }

    /**
     * Converts the supplied <code>path</code> to a {@link File#getCanonicalPath() canonical path}, wrapping any
     * {@link IOException IOExceptions} in a {@link FatalIOException}.
     * 
     * <p>
     * If <code>verifyPathIsDirectory</code> is <code>true</code> and the supplied path is not a directory, a
     * {@link FatalIOException} will be thrown.
     * </p>
     * 
     * @param path the path to convert; may be <code>null</code>
     * @param verifyPathIsDirectory <code>true</code> if the path should only refer to a directory
     * @return the canonical path or <code>null</code> if the supplied path was <code>null</code>
     */
    public static String convertToCanonicalPath(String path, boolean verifyPathIsDirectory) throws FatalIOException {
        if (path == null) {
            return null;
        }
        try {
            File dir = new File(path);
            if (verifyPathIsDirectory && !dir.isDirectory()) {
                throw new FatalIOException("The supplied path [" + path + "] is not a directory.");
            }
            return dir.getCanonicalPath();
        } catch (IOException e) {
            throw new FatalIOException("Could not determine the canonical path for [" + path + "]", e);
        }
    }

    /**
     * Converts the supplied {@link URI} to a {@link File#getCanonicalPath() canonical path}, wrapping any
     * {@link IOException IOExceptions} in a {@link FatalIOException}.
     * 
     * @param uri the path to convert; may be <code>null</code>
     * @return the canonical path or <code>null</code> if the supplied <code>URI</code> was <code>null</code>
     */
    public static String convertToCanonicalPath(URI uri) throws FatalIOException {
        if (uri == null) {
            return null;
        }
        Assert.isTrue(uri.getScheme().equals(URI_FILE_SCHEME), "Cannot determine path of URI '%s' with non-file scheme", uri);
        try {
            return new File(uri).getCanonicalPath();
        } catch (IOException ioe) {
            throw new FatalIOException("Could not determine the canonical path for URI [" + uri + "]", ioe);
        }
    }

    /**
     * Verifies that the directory with the supplied <code>path</code> exists, and if it does not exist, an attempt
     * will be made to create it as well as any necessary parent directories. As a convenience, the supplied path will
     * be {@link #convertToCanonicalPath(String) converted} to a canonical path and returned.
     * 
     * @param path the directory path; must not be <code>null</code>
     * @return the canonical path to the directory
     */
    public static String createDirectoryIfNecessary(String path) {
        Assert.notNull(path, "'path' must not be null.");
        File dir = new File(path);
        if (!dir.exists()) {
            dir.mkdirs();
        }
        return convertToCanonicalPath(path);
    }

    /**
     * Delete the supplied {@link File} and, for directories, recursively delete any nested directories or files.
     * 
     * @param root the root <code>File</code> to delete.
     * @return <code>true</code> if the <code>File</code> was deleted, otherwise <code>false</code>.
     * @see #deleteRecursively(String)
     */
    @EntryExitTraceLevel(Level.DEBUG)
    public static boolean deleteRecursively(@NonNull File root) {
        boolean success = doRecursiveDelete(root);
        if (TRACER.isInfoEnabled()) {
            TRACER.traceInfo("Recursive delete of '%s' successful: %s", root.getAbsolutePath(), success);
        }
        return success;
    }

    /**
     * Delete the file referenced by the supplied path and, if the path refers to a directory, recursively delete any
     * nested directories or files.
     * 
     * @param path the path to the file or directory to delete.
     * @return <code>true</code> if the file or directory was deleted, otherwise <code>false</code>.
     * @see #deleteRecursively(File)
     */
    @EntryExitTraceLevel(Level.DEBUG)
    public static boolean deleteRecursively(@NonNull String path) {
        return deleteRecursively(new File(path));
    }

    @EntryExitTraceLevel(Level.DEBUG)
    private static boolean doRecursiveDelete(File root) {
        if (root.exists()) {
            if (root.isDirectory()) {
                File[] children = root.listFiles();
                if (children != null) {
                    for (File file : children) {
                        doRecursiveDelete(file);
                    }
                }
            }
            return root.delete();
        }
        return false;
    }
}
