/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.io.core;

import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Enumeration;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;

import org.springframework.util.FileCopyUtils;
import org.springframework.util.StringUtils;

import com.springsource.server.serviceability.ffdc.Assert;

/**
 * Utility code for working with JAR files.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * Threadsafe
 * 
 * @author Rob Harrop
 * @since 1.0
 */
public final class JarUtils {

    /**
     * Unpacks the JAR file at {@link PathReference jarFile} to the directory <code>dest</code>.<p/>
     * 
     * If the supplied <code>dest</code> {@link PathReference} does not exist, it is created as a directory and the
     * JAR file is unpacked <strong>directly</strong> into the newly created directory.<p/>
     * 
     * If the supplied <code>dest</code> <code>PathReference</code> already exists and is a directory then the JAR
     * file is unpacked as a subdirectory of the supplied directory this directory. The name of the generated
     * subdirectory is that of the JAR file without the file extension - so <code>foo.jar</code> is unpacked into a
     * directory <code>foo</code>.
     * 
     * @param jarFile the JAR file to unpack
     * @param dest the destination directory
     * @throws IOException if an error occurs during unpack.
     */
    public static void unpackTo(PathReference jarFile, PathReference dest) throws IOException {
        Assert.isTrue(jarFile.exists(), "Supplied JAR File '%s' must exist", jarFile);
        PathReference finalDest = determineFinalDestination(jarFile, dest);
        JarFile jar = new JarFile(jarFile.toFile());
        Enumeration<JarEntry> entries = jar.entries();
        while (entries.hasMoreElements()) {
            JarEntry entry = entries.nextElement();
            PathReference entryPath = finalDest.newChild(entry.getName());
            if (entry.isDirectory()) {
                entryPath.createDirectory();
            } else {
                PathReference filePath = entryPath.createFile();
                InputStream inputStream = jar.getInputStream(entry);
                FileCopyUtils.copy(inputStream, new FileOutputStream(filePath.toFile()));
            }
        }
        jar.close();
    }

    /**
     * Determine the final destination for the unpacked JAR file.
     * 
     * @param dest the supplied destination.
     * @return the final destination.
     */
    private static PathReference determineFinalDestination(PathReference jarFile, PathReference dest) {
        PathReference finalDest = dest;
        if (!dest.exists()) {
            dest.createDirectory();
        } else {
            if (dest.isDirectory()) {
                String destDir = StringUtils.stripFilenameExtension(StringUtils.getFilename(jarFile.getName()));
                finalDest = dest.newChild(destDir).createDirectory();
            } else {
                throw new FatalIOException("Destination path '" + dest + "' already exists and is not a directory");
            }
        }
        return finalDest;
    }
}
