/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.io.core.work;

import org.osgi.framework.BundleContext;
import org.springframework.beans.factory.BeanClassLoaderAware;
import org.springframework.beans.factory.FactoryBean;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.osgi.context.BundleContextAware;
import org.springframework.osgi.service.importer.support.OsgiServiceProxyFactoryBean;

import com.springsource.server.io.core.PathReference;
import com.springsource.server.kernel.core.Subsystem;
import com.springsource.server.kernel.core.SubsystemRegistry;

/**
 * {@link FactoryBean} that creates a {@link PathReference} for the work area for a given subsystem. Subsystems need
 * only to declare a bean of this type in their module context file to get access to the correct work area location.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
public final class WorkAreaFactoryBean implements FactoryBean, BundleContextAware, InitializingBean, BeanClassLoaderAware {

    private volatile BundleContext bundleContext;

    private volatile PathReference workAreaReference;

    private volatile ClassLoader beanClassLoader;

    /**
     * {@inheritDoc}
     */
    public Object getObject() throws Exception {
        return this.workAreaReference;
    }

    /**
     * {@inheritDoc}
     */
    public Class<?> getObjectType() {
        return PathReference.class;
    }

    /**
     * {@inheritDoc}
     */
    public boolean isSingleton() {
        return true;
    }

    /**
     * {@inheritDoc}
     */
    public void setBundleContext(BundleContext bundleContext) {
        this.bundleContext = bundleContext;
    }

    /**
     * {@inheritDoc}
     */
    public void afterPropertiesSet() throws Exception {
        SubsystemRegistry registry = createServiceProxy(SubsystemRegistry.class);
        WorkAreaManager manager = createServiceProxy(WorkAreaManager.class);

        Subsystem subsystem = registry.getSubsystemByBundle(this.bundleContext.getBundle());
        this.workAreaReference = manager.getWorkDirectory(subsystem);
    }

    /**
     * {@inheritDoc}
     */
    public void setBeanClassLoader(ClassLoader classLoader) {
        this.beanClassLoader = classLoader;
    }

    @SuppressWarnings("unchecked") private <T> T createServiceProxy(Class<T> serviceInterface) {
        OsgiServiceProxyFactoryBean fb = new OsgiServiceProxyFactoryBean();
        fb.setBundleContext(this.bundleContext);
        fb.setInterfaces(new Class[] { serviceInterface });
        fb.setBeanClassLoader(this.beanClassLoader);
        fb.setRetryTimes(20);
        fb.afterPropertiesSet();
        return (T) fb.getObject();
    }
}
