/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.kernel.bootstrap;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.springsource.json.parser.AntlrJSONParser;
import com.springsource.json.parser.JSONParser;
import com.springsource.json.parser.ListNode;
import com.springsource.json.parser.MapNode;
import com.springsource.json.parser.Node;
import com.springsource.json.parser.ScalarNode;
import com.springsource.json.parser.ScalarNodeType;
import com.springsource.server.bootstrap.BootstrapConfiguration;
import com.springsource.server.bootstrap.BootstrapConfigurationException;

/**
 * Standard implementation of the {@link BootstrapConfiguration} interface.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * StandardBootstrapConfiguration is immutable and therefore thread-safe.
 * 
 * @author Andy Wilkinson
 * @author Sam Brannen
 * @author Glyn Normington
 * @since 1.0
 */
public class StandardBootstrapConfiguration implements BootstrapConfiguration {
 
    private static final List<String> LOGS_DIRECTORY_PATH = createList("serviceability", "logs", "directory");

    private static final List<String> CONFIG_PATHS_PATH = createList("configPaths");

    private static final List<String> DUMP_DIRECTORY_PATH = createList("serviceability", "dump", "directory");

    private static final List<String> TRACE_DIRECTORY_PATH = createList("serviceability", "trace", "directory");

    private static final List<String> WORK_DIRECTORY_PATH = createList("workDirectory");

    private static final List<String> PICKUP_DIRECTORY_PATH = createList("pickupDirectory");

    private static final List<String> TRACE_LEVELS_PATH = createList("serviceability", "trace", "levels");
    
    private static final List<String> OSGI_PROFILE_LOCATION_PATH = createList("provisioning", "osgiProfile");
    
    private static final List<String> SUBSYSTEMS_DIRECTORY_PATH = createList("provisioning", "subsystemsPath");
    
    private static final List<String> SEARCH_PATHS_PATH = createList("provisioning", "searchPaths");
    
    private static final List<String> INSTALL_PATH_PATH = createList("provisioning", "installPath");
    
    private static final List<String> OSGI_CONSOLE_PORT_PATH = createList("osgiConsole", "port");
    
    private static final List<String> OSGI_CONSOLE_ENABLED_PATH = createList("osgiConsole", "enabled");
    
    private static final String JAVA_SPECIFICATION_VERSION_PROPERTY_NAME = "java.specification.version";
    
    private static final String JAVA_6_SPECIFICATION_VERSION = "1.6";

    private final String homeDirectory;

    private final String[] configPaths;

    private final String dumpDirectory;

    private final String logsDirectory;

    private final String traceDirectory;

    private final String workDirectory;

    private final String subsystemsDirectory;

    private final String pickUpDirectory;

    private final String installPath;
    
    private final String[] searchPaths;

    private final String osgiProfileLocation;

    private final Map<String, String> traceLevels;
    
    private final int osgiConsolePort;
    
    private final boolean osgiConsoleEnabled;

    public StandardBootstrapConfiguration() throws BootstrapConfigurationException {
        String rawHomeDirectory = System.getProperty(SERVER_HOME_PROPERTY);
        if (rawHomeDirectory == null) {
            throw new IllegalStateException("The " + SERVER_HOME_PROPERTY + " system property must be set prior to starting the Server.");
        }

        if (!rawHomeDirectory.endsWith(File.separator)) {
            rawHomeDirectory += File.separator;
        }
        
        this.homeDirectory = convertToCanonicalPath(rawHomeDirectory);
        
        String rawConfigDirectory = System.getProperty(SERVER_CONFIGDIR_PROPERTY);
        
        File configDirectory;

        if (rawConfigDirectory == null) {
            configDirectory = new File(rawHomeDirectory + "config");
        } else {
            configDirectory = new File(rawConfigDirectory);
            if (!configDirectory.isAbsolute()) {
                configDirectory = new File(this.homeDirectory, rawConfigDirectory);
            }
        }
        
        
        File configFile = new File(configDirectory, "server.config");

        Node root = null;

        if (configFile.exists()) {
            JSONParser parser = new AntlrJSONParser();
            try {
                root = parser.parse(configFile.toURL());
            } catch (Exception e) {
                throw new BootstrapConfigurationException("Unable to parse bootstrap configuration from '" + configFile + "'", e);
            }
        } else {
            throw new BootstrapConfigurationException("Configuration file not found at path '" + configFile + "'");
        }
        
        this.configPaths = readStringsFromListNode(root, CONFIG_PATHS_PATH, getDefaultConfigPaths(configDirectory));
        this.dumpDirectory = getAbsolutePathFromScalarStringNode(root, DUMP_DIRECTORY_PATH, getDefaultDumpDirectory(homeDirectory), homeDirectory);
        this.logsDirectory = getAbsolutePathFromScalarStringNode(root, LOGS_DIRECTORY_PATH, getDefaultLogsDirectory(homeDirectory), homeDirectory);
        this.traceDirectory = getAbsolutePathFromScalarStringNode(root, TRACE_DIRECTORY_PATH, getDefaultTraceDirectory(homeDirectory), homeDirectory);
        this.workDirectory = getAbsolutePathFromScalarStringNode(root, WORK_DIRECTORY_PATH, getDefaultWorkDirectory(homeDirectory), homeDirectory);
        this.pickUpDirectory = getAbsolutePathFromScalarStringNode(root, PICKUP_DIRECTORY_PATH, getDefaultPickupDirectory(homeDirectory),
            homeDirectory);
        this.traceLevels = readTraceLevels(root);

        // The following locations cannot be configured and are always located relative to the install root
        this.subsystemsDirectory = getAbsolutePathFromScalarStringNode(root, SUBSYSTEMS_DIRECTORY_PATH, getDefaultSubsystemsDirectory(homeDirectory), homeDirectory);
        this.osgiProfileLocation = getAbsolutePathFromScalarStringNode(root, OSGI_PROFILE_LOCATION_PATH, getDefaultOSGiProfileLocation(homeDirectory), homeDirectory);
        this.searchPaths = readStringsFromListNode(root, SEARCH_PATHS_PATH, getDefaultSearchPaths(homeDirectory));
        this.installPath = getAbsolutePathFromScalarStringNode(root, INSTALL_PATH_PATH, getDefaultInstallPath(homeDirectory), homeDirectory);
        
        this.osgiConsolePort = getIntFromNumericScalarNode(root, OSGI_CONSOLE_PORT_PATH, 2401);
        this.osgiConsoleEnabled = getBooleanFromBooleanScalarNode(root, OSGI_CONSOLE_ENABLED_PATH, true);
    }
    
    private static String getDefaultSubsystemsDirectory(String home) {
        return home + File.separatorChar + "repository" + File.separatorChar + "bundles" + File.separatorChar + "subsystems";
    }
    
    private static int getIntFromNumericScalarNode(Node root, List<String> pathToNode, int defaultValue) {
        if (root != null) {
            Node intNode = root.getNode(pathToNode);
            if (intNode != null) {
                if (intNode instanceof ScalarNode && ScalarNodeType.Number.equals(((ScalarNode)intNode).getType())) {
                    try {
                        return Integer.parseInt((String)((ScalarNode)intNode).getValue()); 
                    } catch (NumberFormatException nfe) {
                        throw new BootstrapConfigurationException("The configured port number could not be parsed as an int", nfe);
                    }
                } else {
                    throw new BootstrapConfigurationException("The node at " + pathToNode + " must be a numeric scalar node.");
                }
            } else {
                return defaultValue;
            }
        } else {
            return defaultValue;
        }
    }
    
    private static boolean getBooleanFromBooleanScalarNode(Node root, List<String> pathToNode, boolean defaultValue) {
        if (root != null) {
            Node booleanNode = root.getNode(pathToNode);
            if (booleanNode != null) {
                if (booleanNode instanceof ScalarNode && ScalarNodeType.Boolean.equals(((ScalarNode)booleanNode).getType())) {
                    return ((ScalarNode)booleanNode).getValue();
                } else {
                    throw new BootstrapConfigurationException("The node at " + pathToNode + " must be a boolean scalar node.");
                }
            } else {
                return defaultValue;
            }
        } else {
            return defaultValue;
        }
    }
    
    private static String[] readStringsFromListNode(Node root, List<String> pathToNode, String[] defaultValues) {
        if (root != null) {
            Node searchPathsNode = root.getNode(pathToNode);
            if (searchPathsNode != null) {
                if (searchPathsNode instanceof ListNode) {
                    List<String> searchPaths = new ArrayList<String>();
                    ListNode listNode = (ListNode)searchPathsNode;
                    List<Node> nodes = listNode.getNodes();                   
                    for (Node node: nodes) {
                        if (node instanceof ScalarNode && ScalarNodeType.String.equals(((ScalarNode)node).getType())) {
                            String path = ((ScalarNode)node).getValue();
                            searchPaths.add(path);
                        }
                    }       
                    return searchPaths.toArray(new String[searchPaths.size()]);
                } else {
                    throw new BootstrapConfigurationException("The node at " + pathToNode + " must be a list node.");
                }
            } else {
                return defaultValues;
            }
        } else {
            return defaultValues;
        }
    }

    /**
     * @param root
     * @return
     */
    private Map<String, String> readTraceLevels(Node root) {
        if (root != null) {
            Node specificLevelsNode = root.getNode(TRACE_LEVELS_PATH);
            if (specificLevelsNode != null) {
                if (specificLevelsNode instanceof MapNode) {
                    MapNode mapNode = (MapNode) specificLevelsNode;
                    Set<String> keys = mapNode.getKeys();
                    Map<String, String> specificTraceLevels = new HashMap<String, String>();
                    for (String key : keys) {
                        String level = getValueOfScalarStringNode(mapNode, createList(key), null);
    
                        if (level != null) {
                            specificTraceLevels.put(key, level);
                        }
                    }
                    return specificTraceLevels;
                } else {
                    throw new BootstrapConfigurationException("The node specific levels node for trace if not a map node.");
                }
            }
        } 
        return new HashMap<String, String>();     
    }

    private static String getValueOfScalarStringNode(Node root, List<String> path, String defaultValue) {
        if (root != null) {
            Node node = root.getNode(path);
            if (node != null) {
                if (node instanceof ScalarNode && ((ScalarNode) node).getType().equals(ScalarNodeType.String)) {
                    return ((ScalarNode) node).getValue();
                } else {
                    throw new BootstrapConfigurationException("The node at path '" + path + "' was not a scalar String node.");
                }
            }
        }
        return defaultValue;
    }

    private static String getAbsolutePathFromScalarStringNode(Node root, List<String> path, String defaultValue, String absoluteRoot) {
        String value = getValueOfScalarStringNode(root, path, defaultValue);
        return convertToAbsoluteCanonicalPath(absoluteRoot, value);
    }

    /**
     * Converts the supplied <code>path</code> to a {@link File#getCanonicalPath() canonical path}, wrapping any
     * {@link IOException IOExceptions} in a {@link BootstrapConfigurationException}.
     * </p>
     * 
     * @param path the path to convert; may be <code>null</code>
     * @return the canonical path or <code>null</code> if the supplied path was <code>null</code>
     */
    private static String convertToCanonicalPath(String path) throws BootstrapConfigurationException {
        if (path == null) {
            return null;
        }
        try {
            return new File(path).getCanonicalPath();
        } catch (IOException e) {
            throw new BootstrapConfigurationException("Could not determine the canonical path for [" + path + "]", e);
        }
    }

    private static String convertToAbsoluteCanonicalPath(String root, String path) throws BootstrapConfigurationException {
        if (path == null) {
            return null;
        }
        File file = new File(path);
        if (!file.isAbsolute()) {
            file = new File(root, path);
        }
        return convertToCanonicalPath(file.toString());

    }

    // -------------------------------------------------------------------------
    // --- BootstrapConfiguration implementation
    // -------------------------------------------------------------------------

    /**
     * {@inheritDoc}
     */
    public String getHomeDirectory() {
        return this.homeDirectory;
    }

    /**
     * {@inheritDoc}
     */
    public String[] getConfigPaths() {
        return this.configPaths;
    }

    /**
     * {@inheritDoc}
     */
    public String getDumpDirectory() {
        return this.dumpDirectory;
    }

    /**
     * {@inheritDoc}
     */
    public String getLogsDirectory() {
        return this.logsDirectory;
    }

    /**
     * {@inheritDoc}
     */
    public String getTraceDirectory() {
        return this.traceDirectory;
    }

    /**
     * {@inheritDoc}
     */
    public String getWorkDirectory() {
        return this.workDirectory;
    }

    /**
     * {@inheritDoc}
     */
    public String getSubsystemsDirectory() {
        return this.subsystemsDirectory;
    }

    /**
     * {@inheritDoc}
     */
    public String getPickUpDirectory() {
        return this.pickUpDirectory;
    }

    // -------------------------------------------------------------------------
    // --- StandardBootstrapConfiguration-specific implementation
    // -------------------------------------------------------------------------

    public String[] getSearchPaths() {
        return this.searchPaths;
    }
    
    public String getInstallPath() {
        return this.installPath;
    }

    /**
     * <p>
     * The <em>OSGi profile location</em> is the path to a server-specific file which defines the
     * <em>OSGi profile</em> to use when launching the Server.
     * </p>
     * <p>
     * The Server's OSGi profile is located in the &lt;server-home&gt;/lib directory.
     * </p>
     * 
     * @return The path to the Server's OSGi profile
     */
    public String getOSGiProfileLocation() {
        return this.osgiProfileLocation;
    }
    
    int getOSGiConsolePort() {
        return this.osgiConsolePort;
    }
    
    boolean isOsgiConsoleEnabled() {
        return this.osgiConsoleEnabled;
    }

    private static List<String> createList(String... strings) {
        return Arrays.asList(strings);
    }

    private static String[] getDefaultConfigPaths(File configDirectory) {
        return new String[] {configDirectory.getAbsolutePath()};
    }

    private static String getDefaultDumpDirectory(String home) {
        return home + File.separator + "serviceability" + File.separator + "dump";
    }

    private static String getDefaultLogsDirectory(String home) {
        return home + File.separator + "serviceability" + File.separator + "logs";
    }

    private static String getDefaultTraceDirectory(String home) {
        return home + File.separator + "serviceability" + File.separator + "dump";
    }

    private static String getDefaultWorkDirectory(String home) {
        return home + File.separator + "work";
    }

    private static String getDefaultPickupDirectory(String home) {
        return home + File.separator + "pickup";
    }
    
    private static String getDefaultOSGiProfileLocation(String home) {
        String specificationVersion = System.getProperty(JAVA_SPECIFICATION_VERSION_PROPERTY_NAME);
        if (JAVA_6_SPECIFICATION_VERSION.equals(specificationVersion)) {
            return home + File.separator + "lib/java6-server.profile";
        } else {
            return home + File.separator + "lib/java5-server.profile";
        }        
    }
    
    private static String[] getDefaultSearchPaths(String home) {
        return new String[] { home + "/repository/bundles/subsystems/{name}/{bundle}.jar", home + "/repository/bundles/ext/{bundle}",
            home + "/repository/libraries/ext/{library}", home + "/repository/bundles/usr/{bundle}", home + "/repository/libraries/usr/{library}" };
    }
    
    private static String getDefaultInstallPath(String home) {
        return home + "/repository/installed";
    }

    /**
     * {@inheritDoc}
     */
    public Map<String, String> getTraceLevels() {
        return this.traceLevels;
    }
}
