/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.kernel.core;

import java.util.List;

import org.osgi.framework.Bundle;

/**
 * The organizing element in the Kernel that groups and manages all components related to a specific functional area.
 * <p>
 * 
 * A <code>Subsystem</code> is a collection of OSGi {@link Bundle Bundles} that together provide a logical feature
 * set. The <code>Subsystem</code> manages the lifecycle of these <code>Bundles</code> internally and provides its
 * own state and lifecycle model to the <code>Kernel</code>.
 * <p>
 * 
 * Internal <code>Bundle</code> state is exposed to clients via {@link BundleDescriptor BundleDescriptors}, which
 * restrict access to mutative <code>Bundle</code> operations.
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * Implementations are <strong>required</strong> to be thread safe.
 * 
 * @author Rob Harrop
 * @author Glyn Normington
 * @since 1.0
 */
public interface Subsystem {

    /**
     * The subsystem state is closely related to the state of the bundles of the subsystem.
     * 
     * If the subsystem is in INSTALLING state, then all its bundles are in INSTALLED state, although the subsystem may
     * install more bundles before it transitions to the INSTALLED state.
     * 
     * If the subsystem is in INSTALLED state, then it has installed all its bundles and each of its bundles are in one
     * of the states INSTALLED or RESOLVED.
     * 
     * If the subsystem is in STARTING state, then each of its bundles is in one of the states INSTALLED, RESOLVED,
     * STARTING, or ACTIVE.
     * 
     * If the subsystem is in ACTIVE state, then all of its bundles are in ACTIVE state.
     * 
     * If the subsystem is in STOPPING state, then each of its bundles is in one of the states ACTIVE, STOPPING, or
     * RESOLVED.
     * 
     * If the subsystem is in UNINSTALLING state, then each of its bundles is in one of the states RESOLVED or
     * UNINSTALLED.
     * 
     * If the subsystem is in UNINSTALLED state, then all of its bundles are in UNINSTALLED state.
     * 
     * If the subsystem is in ERROR state, then its bundles may be in any states.
     * 
     */
    public enum State {
        INSTALLING, INSTALLED, STARTING, ACTIVE, ERROR, STOPPING, UNINSTALLING, UNINSTALLED
    }

    /**
     * Gets the name of the <code>Subsystem</code>.
     * 
     * @return the <code>Subsystem</code> name.
     */
    String getName();

    /**
     * Gets the {@link State} of the <code>Subsystem</code>.
     * 
     * @return the <code>Subsystem</code> <code>State</code>.
     */
    State getState();

    /**
     * Gets the {@link BundleDescriptor BundleDescriptors} for this <code>Subsystem</code>.
     * 
     * @return an unmodifiable {@link  List} of the <code>BundleDescriptors</code>.
     */
    List<BundleDescriptor> getBundleDescriptors();
    
    /**
     * Start the subsystem.
     */
    void start();

    /**
     * Stop the subsystem.
     */
    void stop();

    /**
     * Uninstall the subsystem.
     */
    void uninstall();
}
