/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.kernel.core;

import java.util.Collection;

import org.osgi.framework.Bundle;

/**
 * Maintains a registry of all installed {@link Subsystem Subsystems}.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * Implementations <strong>must</strong> be threadsafe.
 * 
 * @author Rob Harrop
 * @author Glyn Normington
 * @since 1.0
 */
public interface SubsystemRegistry extends SubsystemInstaller {

    /**
     * Gets the {@link Subsystem} representing the kernel itself.
     * 
     * @return the kernel <code>Subsystem</code>.
     */
    Subsystem getKernelSubsystem();

    /**
     * Gets all the {@link Subsystem Subsystems} installed into the <code>SubsystemRegistry</code>.
     * 
     * @return the {@link Collection} of installed <code>Subsystems</code>.
     */
    Collection<Subsystem> getSubsystems();

    /**
     * Retrieve the {@link Subsystem} identified by the supplied name.
     * 
     * @param name the name of the <code>Subsystem</code> to be returned
     * @return the named <code>Subsystem</code> or <code>null</code> if no such <code>Subsystem</code> exists.
     */
    Subsystem getSubsystemByName(String name);

    /**
     * Return the subsystem that the given class belongs to or null if the class does not belong to a subsystem.
     * 
     * @param clazz the class whose subsystem we want
     * @return the subsystem of the class or null if the class does not belong to a subsystem
     */
    Subsystem getSubsystemByClass(Class<?> clazz);

    /**
     * Return the <code>Subsystem</code> that the given {@link Bundle} belongs to.
     * 
     * @param bundle the <code>Bundle</code> whose subsystem we want.
     * @return the <code>Subsystem</code> of the <code>Bundle</code> or <code>null</code> if the
     *         <code>Bundle</code> does not belong to a <code>Subsystem</code>.
     */
    Subsystem getSubsystemByBundle(Bundle bundle);

    /**
     * Registers the supplied {@link SubsystemListener} with the <code>SubsystemRegistry</code>. Implementations <strong>must</strong>
     * guarantee that <code>SubsystemListeners</code> can be safely registered concurrently with any observable events.
     * 
     * @param listener the <code>SubsystemListener</code> to add.
     */
    void addSubsystemListener(SubsystemListener listener);
}
