/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.kernel.core.internal;

import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.ReentrantLock;

import org.springframework.context.ApplicationContext;

/**
 * {@link Future} that is used to access {@link ApplicationContext} instances that are created asynchronously by Spring
 * Dynamic Modules. Calls to {@link #get()} will correctly reflect any {@link Exception} raised during creation of the
 * <code>ApplicationContext</code>.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
final class ApplicationContextFuture implements Future<ApplicationContext> {

    private final ReentrantLock lock = new ReentrantLock();

    private final Condition available = this.lock.newCondition();

    private ApplicationContext context;

    private Throwable throwable;

    /**
     * Signals successful completion of the {@link ApplicationContext} construction.
     * 
     * @param context the created <code>ApplicationContext</code>.
     */
    public void handOff(ApplicationContext context) {
        this.lock.lock();
        try {
            this.context = context;
            this.throwable = null;
            this.available.signalAll();
        } finally {
            this.lock.unlock();
        }
    }

    /**
     * Signals failed {@link ApplicationContext} construction.
     * 
     * @param throwable the error that occurred.
     */
    public void handOff(Throwable throwable) {
        this.lock.lock();
        try {
            this.context = null;
            this.throwable = throwable;
            this.available.signalAll();
        } finally {
            this.lock.unlock();
        }
    }

    /**
     * {@inheritDoc}
     */
    public boolean cancel(boolean mayInterruptIfRunning) {
        return false;
    }

    /**
     * {@inheritDoc}
     */
    public ApplicationContext get() throws InterruptedException, ExecutionException {
        this.lock.lock();
        if (isDone()) {
            return processResult();
        }
        try {
            this.available.await();
            return processResult();
        } finally {
            this.lock.unlock();
        }
    }

    /**
     * {@inheritDoc}
     */
    public ApplicationContext get(long timeout, TimeUnit unit) throws InterruptedException, ExecutionException, TimeoutException {
        this.lock.lock();
        if (isDone()) {
            return processResult();
        }
        try {
            if (this.available.await(timeout, unit)) {
                return processResult();
            } else {
                throw new TimeoutException("Timeout while waiting.");
            }
        } finally {
            this.lock.unlock();
        }
    }

    /**
     * {@inheritDoc}
     */
    public boolean isCancelled() {
        return false;
    }

    /**
     * {@inheritDoc}
     */
    public boolean isDone() {
        this.lock.lock();
        try {
            return this.context != null || this.throwable != null;
        } finally {
            this.lock.unlock();
        }
    }

    private ApplicationContext processResult() throws ExecutionException {
        if (this.context != null) {
            return this.context;
        } else {
            throw new ExecutionException(this.throwable);
        }
    }
}
