/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.kernel.core.internal;

import java.util.ArrayList;
import java.util.Collections;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.ServiceReference;
import org.osgi.framework.ServiceRegistration;
import org.springframework.context.ApplicationContext;

import com.springsource.server.bootstrap.BootstrapConfiguration;
import com.springsource.server.kernel.bootstrap.SubsystemContents;
import com.springsource.server.kernel.bootstrap.SubsystemContents.SubsystemConfiguration;
import com.springsource.server.kernel.core.BundleDescriptor;
import com.springsource.server.kernel.core.BundleStarter;
import com.springsource.server.kernel.core.FatalKernelException;
import com.springsource.server.kernel.core.KernelException;
import com.springsource.server.kernel.core.Subsystem;
import com.springsource.server.kernel.core.SubsystemInstaller;
import com.springsource.server.kernel.core.SubsystemRegistry;
import com.springsource.server.kernel.core.diagnostics.KernelLogEvent;
import com.springsource.server.kernel.core.internal.install.StandardSubsystemResolver;
import com.springsource.server.osgi.framework.OsgiFramework;
import com.springsource.server.serviceability.logging.LogService;
import com.springsource.server.serviceability.logging.LogServiceFactory;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * {@link Subsystem} implementation for the kernel.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @author Glyn Normington
 * @since 1.0
 */
final class KernelSubsystem extends AbstractSubsystem implements SubsystemInstaller {

    private static final LogService LOGGER = LogServiceFactory.getLogService();

    private static final Tracer TRACER = TracerFactory.createTracer(KernelSubsystem.class);

    private static final String PRIMORDIAL_SUBSYSTEM_NAME = "com.springsource.server.primordial";

    private final Bundle kernelBundle;

    private final SubsystemInstaller subsystemInstaller;

    private final BundleStarter starter;

    private ServiceRegistration subsystemRegistryServiceRegistration;

    private Subsystem ffdcSubsystem;

    /**
     * Creates a new <code>KernelSubsystem</code>.
     * 
     * @param bundle the kernel {@link Bundle}.
     * @param osgiFramework the {@link OsgiFramework} the kernel is running in.
     * @param bootstrapConfiguration the {@link BootstrapConfiguration}.
     */
    public KernelSubsystem(Bundle bundle, OsgiFramework osgiFramework, BootstrapConfiguration bootstrapConfiguration, BundleStarter starter) {
        super(new StandardSubsystemRegistry(osgiFramework), StandardSubsystemRegistry.KERNEL_SUBSYSTEM_NAME);
        this.kernelBundle = bundle;
        this.starter = starter;
        setBundles(Collections.singletonList(bundle));
        this.subsystemInstaller = new StandardSubsystemInstaller(getSubsystemRegistry(), new StandardSubsystemResolver(bootstrapConfiguration));
        installBootSubsystems();
        LOGGER.log(KernelLogEvent.INSTALLED_BOOT_SUBSYSTEMS);
        installBaseSubsystems();
        LOGGER.log(KernelLogEvent.INSTALLED_BASE_SUBSYSTEMS);
        setInstalled();
    }

    @Override protected void doStart() {
        // do not call super.doStart(); as the kernel bundle is already starting
        this.subsystemRegistryServiceRegistration = this.kernelBundle.getBundleContext().registerService(SubsystemRegistry.class.getName(),
            this.getSubsystemRegistry(), null);

        // treat the ffdc subsystem specially so we can get it up before all the others
        this.ffdcSubsystem.start();
        for (BundleDescriptor bundle : this.ffdcSubsystem.getBundleDescriptors()) {
            Future<ApplicationContext> future = this.starter.awaitStart(bundle.getBundle());
            try {
                future.get(60, TimeUnit.SECONDS);
            } catch (InterruptedException e) {
                throw new FatalKernelException("Interrupted waiting for FFDC subsystem.", e);
            } catch (ExecutionException e) {
                throw new FatalKernelException("Error starting FFDC subsystem.", e.getCause());
            } catch (TimeoutException e) {
                throw new FatalKernelException("Timeout waiting for FFDC subsystem.");
            }
        }

        for (Subsystem s : getSubsystemRegistry().getSubsystems()) {
            if (s != this && s != this.ffdcSubsystem) {
                s.start();
            }
        }
    }

    /**
     * Fish subsystems out of the primordial soup and install them.
     */
    private void installBootSubsystems() {
        /*
         * Install a 'virtual' subsystem for classes such as the bootstrap and OSGi framework classes which are loaded
         * by the app. class loader This virtual subsystem has no bundles, but fortunately Bundle is not exposed on the
         * subsystem interface.
         */
        TRACER.traceInfo("Installing '%s' subsystem", PRIMORDIAL_SUBSYSTEM_NAME);
        ManagedSubsystem primordialSubsystem = new StaticSubsystem(this.getSubsystemRegistry(), PRIMORDIAL_SUBSYSTEM_NAME, new ArrayList<Bundle>(0));
        this.getSubsystemRegistry().setPrimordialSubsystem(primordialSubsystem);
        TRACER.traceInfo("Installed '%s' subsystem", PRIMORDIAL_SUBSYSTEM_NAME);

        // Install subsystems from bundles in the primordial soup
        TRACER.traceInfo("Installing boot subsystems");
        BundleContext bc = this.kernelBundle.getBundleContext();
        ServiceReference bootConfigRef = bc.getServiceReference(SubsystemContents.class.getName());
        SubsystemContents bootConfig = (SubsystemContents) bc.getService(bootConfigRef);

        for (SubsystemConfiguration sc : bootConfig.getSubsystemConfigurations()) {
            Subsystem subsystem = new StaticSubsystem(this.getSubsystemRegistry(), sc.getSubsystemName(), sc.getBundles());
            TRACER.traceInfo("Installed boot subsystem '%s'", subsystem.getName());
        }

        bc.ungetService(bootConfigRef);
        TRACER.traceInfo("Installed boot subsystems");
    }

    /**
     * Installs all the base subsystems.
     */
    private void installBaseSubsystems() {
        try {
            // install ffdc first so it can get it's failure handling in there
            this.ffdcSubsystem = this.subsystemInstaller.installSubsystem("com.springsource.server.ffdc");

            // Install the config subsystem early so that configuration is available for later subsystems.
            this.subsystemInstaller.installSubsystem("com.springsource.server.config");

            this.subsystemInstaller.installSubsystem("com.springsource.server.concurrent");
            this.subsystemInstaller.installSubsystem("com.springsource.server.io");
            this.subsystemInstaller.installSubsystem("com.springsource.server.control");
            this.subsystemInstaller.installSubsystem("com.springsource.server.deployer");
            this.subsystemInstaller.installSubsystem("com.springsource.server.system");
            this.subsystemInstaller.installSubsystem("com.springsource.server.management.proxy");

            // The profile manager is the last base subsystem to be installed. It
            // will install profile-specific subsystems.
            this.subsystemInstaller.installSubsystem("com.springsource.server.profile");
        } catch (KernelException e) {
            throw new FatalKernelException("Exception installing base subsystem", e);
        }
    }

    /**
     * Stop the Kernel and all installed {@link Subsystem Subsystems}.
     */
    @Override public void doStop() {
        for (Subsystem s : getSubsystemRegistry().getSubsystems()) {
            if (s != this) {
                s.stop();
            }
        }

        this.subsystemRegistryServiceRegistration.unregister();

        super.doStop();
    }

    @Override public void doUninstall() {
        for (Subsystem s : getSubsystemRegistry().getSubsystems()) {
            if (s != this) {
                s.uninstall();
            }
        }
        super.doUninstall();
    }

    /**
     * {@inheritDoc}
     */
    public Subsystem installSubsystem(String name) throws KernelException {
        return this.subsystemInstaller.installSubsystem(name);
    }

}
