/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.kernel.core.internal;

import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.Future;

import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.BundleEvent;
import org.osgi.framework.BundleException;
import org.osgi.framework.BundleListener;
import org.osgi.framework.InvalidSyntaxException;
import org.osgi.framework.ServiceReference;
import org.osgi.framework.ServiceRegistration;
import org.springframework.context.ApplicationContext;
import org.springframework.osgi.context.ConfigurableOsgiBundleApplicationContext;
import org.springframework.osgi.context.event.OsgiBundleApplicationContextEvent;
import org.springframework.osgi.context.event.OsgiBundleApplicationContextListener;
import org.springframework.osgi.context.event.OsgiBundleContextFailedEvent;
import org.springframework.osgi.context.event.OsgiBundleContextRefreshedEvent;

import com.springsource.server.kernel.core.BundleStarter;
import com.springsource.server.osgi.framework.OsgiFrameworkUtils;
import com.springsource.server.serviceability.ffdc.Assert;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * Standard implementation of {@link BundleStarter} that tracks {@link ApplicationContext} construction using an
 * {@link OsgiBundleApplicationContextListener}. <p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
final class StandardBundleStarter implements BundleStarter {

    private static final Tracer TRACER = TracerFactory.createTracer(StandardBundleStarter.class);

    private final Object monitor = new Object();

    private final Map<Bundle, ApplicationContext> contextMap = new HashMap<Bundle, ApplicationContext>();

    private final Map<Bundle, Throwable> failureMap = new HashMap<Bundle, Throwable>();

    private final ConcurrentMap<Bundle, ApplicationContextFuture> inProgress = new ConcurrentHashMap<Bundle, ApplicationContextFuture>();

    private final ServiceRegistration listenerRegistration;

    public StandardBundleStarter(BundleContext bundleContext) throws InvalidSyntaxException {
	this.listenerRegistration = bundleContext.registerService(OsgiBundleApplicationContextListener.class.getName(),
	        new BundleMonitorApplicationContextListener(), null);
	bundleContext.addBundleListener(new BundleStarterBundleListener());

	processExistingApplicationContexts(bundleContext, bundleContext.getServiceReferences(ApplicationContext.class.getName(), null));
    }

    public static Bundle getOsgiBundle(ApplicationContext applicationContext) {
	Bundle bundle = null;
	if (applicationContext instanceof ConfigurableOsgiBundleApplicationContext) {
	    ConfigurableOsgiBundleApplicationContext osgiBundleApplicationContext = (ConfigurableOsgiBundleApplicationContext) applicationContext;
	    bundle = osgiBundleApplicationContext.getBundle();
	}
	return bundle;
    }

    /**
     * {@inheritDoc}
     */
    public Future<ApplicationContext> start(Bundle bundle) throws BundleException {
	bundle.start();
	return awaitStart(bundle);
    }

    /**
     * {@inheritDoc}
     */
    public Future<ApplicationContext> awaitStart(Bundle bundle) {
	if (!OsgiFrameworkUtils.isSpringOsgiPoweredBundle(bundle)) {
	    return new StaticFuture<ApplicationContext>(null);
	}
	synchronized (this.monitor) {
	    ApplicationContext context = this.contextMap.get(bundle);
	    Throwable failure = this.failureMap.get(bundle);
	    Assert.isFalse(context != null && failure != null,
		    "Spring DM has notified an application context both successfully constructed and failed: " + failure);
	    if (context != null) {
		TRACER.traceInfo("Application context already available for bundle '%s' (StandardBundleStarter %s)", bundle, this);
		return new StaticFuture<ApplicationContext>(context);
	    } else if (failure != null) {
		TRACER.traceInfo("Failure already recorded for bundle '%s' (StandardBundleStarter %s)", bundle, this);
		return new FailingFuture<ApplicationContext>(failure);
	    } else {
		TRACER.traceInfo("Obtaining ApplicationContextFuture for bundle '%s' (StandardBundleStarter %s)", bundle, this);
		ApplicationContextFuture future = new ApplicationContextFuture();
		ApplicationContextFuture oldFuture = this.inProgress.putIfAbsent(bundle, future);
		if (oldFuture != null) {
		    future = oldFuture;
		    TRACER.traceInfo("Found existing ApplicationContextFuture '%s'", future);
		} else {
		    TRACER.traceInfo("Created ApplicationContextFuture '%s'", future);
		}
		return future;
	    }
	}
    }

    /**
     * {@inheritDoc}
     */
    public void destroy() throws Exception {
	if (this.listenerRegistration != null) {
	    this.listenerRegistration.unregister();
	}
    }

    private void processExistingApplicationContexts(BundleContext bundleContext, ServiceReference[] serviceReferences) {
	for (ServiceReference serviceReference : serviceReferences) {
	    ApplicationContext applicationContext = (ApplicationContext) bundleContext.getService(serviceReference);
	    processApplicationContext(applicationContext);
	}
    }

    private void processApplicationContext(ApplicationContext applicationContext) {
	Bundle bundle = getOsgiBundle(applicationContext);
	ApplicationContextFuture future = StandardBundleStarter.this.inProgress.remove(bundle);
	TRACER.traceInfo("Recording application context '%s' successfully created for bundle '%s' (StandardBundleStarter %s)", applicationContext,
	        bundle, StandardBundleStarter.this);
	StandardBundleStarter.this.contextMap.put(bundle, applicationContext);
	for (Bundle inProgressBundle : StandardBundleStarter.this.inProgress.keySet()) {
	    TRACER.traceInfo("Bundle still in progress: '%s'", inProgressBundle);
	}
	if (future != null) {
	    TRACER.traceInfo("Handing off creation of application context '%s' for bundle '%s'", applicationContext, bundle);
	    future.handOff(applicationContext);
	}
    }

    final class BundleMonitorApplicationContextListener implements OsgiBundleApplicationContextListener {

	/**
	 * {@inheritDoc}
	 */
	public void onOsgiApplicationEvent(OsgiBundleApplicationContextEvent event) {
	    synchronized (StandardBundleStarter.this.monitor) {
		ApplicationContext applicationContext = event.getApplicationContext();
		if (event instanceof OsgiBundleContextFailedEvent) {
		    Bundle bundle = getOsgiBundle(applicationContext);
		    ApplicationContextFuture future = StandardBundleStarter.this.inProgress.remove(bundle);
		    OsgiBundleContextFailedEvent failedEvent = (OsgiBundleContextFailedEvent) event;
		    TRACER.traceError("Recording application context construction failure '%s' for bundle '%s' (StandardBundleStarter %s)",
			    failedEvent, bundle, StandardBundleStarter.this);
		    StandardBundleStarter.this.failureMap.put(bundle, failedEvent.getFailureCause());
		    if (future != null) {
			TRACER.traceInfo("Handing off application context construction failure '%s' for bundle '%s'", failedEvent, bundle);
			future.handOff(failedEvent.getFailureCause());
		    }
		} else if (event instanceof OsgiBundleContextRefreshedEvent) {
		    processApplicationContext(applicationContext);
		}
	    }
	}

    }

    final class BundleStarterBundleListener implements BundleListener {

	/**
	 * {@inheritDoc}
	 */
	public void bundleChanged(BundleEvent event) {
	    if (event.getType() == BundleEvent.STOPPED) {
		synchronized (StandardBundleStarter.this.monitor) {
		    StandardBundleStarter.this.contextMap.remove(event.getBundle());
		    StandardBundleStarter.this.failureMap.remove(event.getBundle());
		}
	    }
	}

    }
}
