/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.kernel.core.internal;

import java.util.Set;
import java.util.concurrent.CopyOnWriteArraySet;

import com.springsource.server.kernel.core.Subsystem;
import com.springsource.server.kernel.core.SubsystemListener;

/**
 * Provides supporting functionality to manage a set of {@link SubsystemListener SubsystemListeners} and broadcast
 * Kernel events to them.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * Registering new <code>SubsystemListeners</code> and broadcasting events can both happen concurrently. Registering
 * new <code>SubsystemListeners</code> is a relatively expensive process and is expected to happen infrequently.
 * 
 * @author Rob Harrop
 * @author Glyn Normington
 * @since 1.0
 */
final class SubsystemListenerManager {

    private final Set<SubsystemListener> listeners = new CopyOnWriteArraySet<SubsystemListener>();

    /**
     * Routes a subsystem state change event to all registered {@link SubsystemListener KernelListeners}.
     * 
     * @param subsystem {@link Subsystem} that was installed
     * @param newState the new state of the subsystem
     */
    void subsystemStateChange(Subsystem subsystem, Subsystem.State newState) {
        for (SubsystemListener listener : this.listeners) {
            listener.subSystemStateChange(subsystem, newState);
        }
    }

    /**
     * Registers the supplied {@link SubsystemListener} for future kernel events.
     * 
     * @param listener the <code>SubsystemListener</code> to register.
     */
    public void registerSubsystemListener(SubsystemListener listener) {
        this.listeners.add(listener);
    }
}
