/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.kernel.dm;

import java.util.Collection;

import org.springframework.beans.factory.BeanFactoryUtils;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationEvent;
import org.springframework.context.ApplicationListener;
import org.springframework.context.event.ApplicationContextEvent;
import org.springframework.context.event.ContextClosedEvent;
import org.springframework.osgi.service.importer.support.OsgiServiceProxyFactoryBean;

import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * {@link ApplicationContextShutdownBean} manages the shutting down of application contexts in the dm Server. In
 * particular it is responsible for ensuring that Spring DM service proxies do not hold up application context shutdown.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * This class is thread safe.
 * 
 * @author Glyn Normington
 * @since 1.0
 */
final class ApplicationContextShutdownBean implements ApplicationListener {
    
    static final Tracer TRACER = TracerFactory.createTracer(ApplicationContextShutdownBean.class);

    /** 
     * {@inheritDoc}
     */
    public void onApplicationEvent(ApplicationEvent event) {
        if (event instanceof ContextClosedEvent) {
            TRACER.traceInfo("Processing ContextClosedEvent '%s'", event);
            ApplicationContext applicationContext = ((ApplicationContextEvent) event).getApplicationContext();
            disableServiceProxyRetry(applicationContext);
        }
    }
    
    @SuppressWarnings("unchecked")
    static void disableServiceProxyRetry(ApplicationContext applicationContext) {
        for (OsgiServiceProxyFactoryBean proxyBean : ((Collection<OsgiServiceProxyFactoryBean>) BeanFactoryUtils.beansOfTypeIncludingAncestors(
            applicationContext, OsgiServiceProxyFactoryBean.class, true, false).values())) {
            TRACER.traceInfo("Setting timeout to 0 for proxy '%s' of application context '%s'", proxyBean, applicationContext);
            proxyBean.setRetryTimes(0);
            proxyBean.setTimeout(0);
        }
    }
}
