/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.kernel.dm;

import static org.springframework.context.ConfigurableApplicationContext.LOAD_TIME_WEAVER_BEAN_NAME;

import org.osgi.framework.BundleContext;
import org.osgi.framework.BundleException;
import org.osgi.framework.InvalidSyntaxException;
import org.springframework.beans.BeansException;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.beans.factory.config.ConfigurableListableBeanFactory;
import org.springframework.jmx.export.MBeanExporter;
import org.springframework.jmx.export.annotation.AnnotationMBeanExporter;
import org.springframework.osgi.extender.OsgiBeanFactoryPostProcessor;

import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;
import com.springsource.server.serviceability.tracing.TracingService;

/**
 * {@link OsgiBeanFactoryPostProcessor} implementation that plugs in Server extensions implementation when needed.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @see ServerLoadTimeWeaver
 * @since 1.0
 */
public class ServerExtensionConfiguringOsgiPostProcessor implements OsgiBeanFactoryPostProcessor {

    private static final String APPLICATION_CONTEXT_SHUTDOWN_BEAN_NAME = "com.springsource.server.applicationContextShutdownBean";

    private static final Tracer TRACER = TracerFactory.createTracer(ServerExtensionConfiguringOsgiPostProcessor.class);

    private final TracingService tracingService;

    ServerExtensionConfiguringOsgiPostProcessor(TracingService tracingService) {
        this.tracingService = tracingService;
    }

    /**
     * {@inheritDoc}
     */
    public void postProcessBeanFactory(BundleContext bundleContext, ConfigurableListableBeanFactory beanFactory) throws BeansException,
        InvalidSyntaxException, BundleException {

        processLoadTimeWeaver(bundleContext, beanFactory);
        processMBeanExporter(bundleContext, beanFactory);
        addShutdownBean(beanFactory);
    }

    /**
     * Add an {@link ApplicationContextShutdownBean} to the given bean factory.
     * 
     * @param beanFactory the bean factory to which to add the bean
     */
    private void addShutdownBean(ConfigurableListableBeanFactory beanFactory) {
        beanFactory.registerSingleton(APPLICATION_CONTEXT_SHUTDOWN_BEAN_NAME, new ApplicationContextShutdownBean());
    }

    private void processLoadTimeWeaver(BundleContext bundleContext, ConfigurableListableBeanFactory beanFactory) {
        if (beanFactory.containsBean(LOAD_TIME_WEAVER_BEAN_NAME)) {
            BeanDefinition ltwBean = beanFactory.getBeanDefinition(LOAD_TIME_WEAVER_BEAN_NAME);
            ltwBean.setBeanClassName(ServerLoadTimeWeaver.class.getName());
            if (TRACER.isInfoEnabled()) {
                TRACER.traceInfo("Found load-time weaver bean for bundle '%s'. Switching to ServerLoadTimeWeaver.", bundleContext.getBundle());
            }
        } else {
            if (TRACER.isInfoEnabled()) {
                TRACER.traceInfo("Load-time weaving not enabled for bundle '%s',", bundleContext.getBundle());
            }
        }
    }

    private void processMBeanExporter(BundleContext bundleContext, ConfigurableListableBeanFactory beanFactory) {
        boolean foundExporter = false;
        String[] beanDefinitionNames = beanFactory.getBeanDefinitionNames();
        for (String beanDefinitionName : beanDefinitionNames) {
            BeanDefinition definition = beanFactory.getBeanDefinition(beanDefinitionName);
            if (MBeanExporter.class.getName().equals(definition.getBeanClassName())) {
                definition.setBeanClassName(ServerMBeanExporter.class.getName());
                definition.getConstructorArgumentValues().addGenericArgumentValue(tracingService);
                foundExporter = true;
            } else if (AnnotationMBeanExporter.class.getName().equals(definition.getBeanClassName())) {
                definition.setBeanClassName(ServerAnnotationMBeanExporter.class.getName());
                definition.getConstructorArgumentValues().addGenericArgumentValue(tracingService);
                foundExporter = true;
            }
        }
        if (TRACER.isInfoEnabled()) {
            if (foundExporter) {
                TRACER.traceInfo("Found MBean exporter bean for bundle '%s'. Switching to ServerMBeanExporter or ServerAnnotationMBeanExporter.",
                    bundleContext.getBundle());
            } else {
                TRACER.traceInfo("MBean exporting not enabled for bundle '%s',", bundleContext.getBundle());
            }
        }
    }

}
