/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.kernel.dm;

import java.lang.instrument.ClassFileTransformer;

import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.ServiceReference;
import org.springframework.beans.factory.BeanClassLoaderAware;
import org.springframework.instrument.classloading.LoadTimeWeaver;
import org.springframework.osgi.util.BundleDelegatingClassLoader;

import com.springsource.server.osgi.framework.InstrumentableClassLoader;
import com.springsource.server.osgi.framework.OsgiFramework;

/**
 * {@link LoadTimeWeaver} implementation that plugs into the
 * {@link InstrumentableClassLoader InstrumentableClassLoaders} created for all installed bundle.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
public final class ServerLoadTimeWeaver implements LoadTimeWeaver, BeanClassLoaderAware {

    private volatile InstrumentableClassLoader instrumentableClassLoader;

    /**
     * {@inheritDoc}
     */
    public void addTransformer(ClassFileTransformer transformer) {
        this.instrumentableClassLoader.addClassFileTransformer(transformer);
    }

    /**
     * {@inheritDoc}
     */
    public ClassLoader getInstrumentableClassLoader() {
        return (ClassLoader) this.instrumentableClassLoader;
    }

    /**
     * {@inheritDoc}
     */
    public ClassLoader getThrowawayClassLoader() {
        return this.instrumentableClassLoader.createThrowAway();
    }

    /**
     * {@inheritDoc}
     */
    public void setBeanClassLoader(ClassLoader classLoader) {
        InstrumentableClassLoader instrumentableClassLoader = null;
        if (classLoader instanceof InstrumentableClassLoader) {
            instrumentableClassLoader = (InstrumentableClassLoader) classLoader;
        } else if (classLoader instanceof BundleDelegatingClassLoader) {
            Bundle bundle = ((BundleDelegatingClassLoader) classLoader).getBundle();
            ClassLoader bundleClassLoader = getBundleClassLoader(bundle);
            if (bundleClassLoader instanceof InstrumentableClassLoader) {
                instrumentableClassLoader = (InstrumentableClassLoader) bundleClassLoader;
            }
        }
        if (instrumentableClassLoader == null) {
            throw new IllegalStateException("ClassLoader '" + classLoader + "' is not instrumentable.");
        }
        this.instrumentableClassLoader = instrumentableClassLoader;
    }

    /**
     * Gets the {@link ClassLoader} for the supplied {@link Bundle}.
     * 
     * @param bundle the <code>Bundle</code>.
     * @return the <code>Bundles</code> <code>ClassLoader</code>.
     */
    private ClassLoader getBundleClassLoader(Bundle bundle) {
        BundleContext bundleContext = bundle.getBundleContext();
        ServiceReference serviceReference = bundleContext.getServiceReference(OsgiFramework.class.getName());
        try {
            OsgiFramework framework = (OsgiFramework) bundleContext.getService(serviceReference);
            return framework.getBundleClassLoader(bundle);
        } finally {
            bundleContext.ungetService(serviceReference);
        }
    }

}
