/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.management.proxy.core;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.osgi.framework.BundleContext;
import org.osgi.framework.InvalidSyntaxException;
import org.osgi.framework.ServiceReference;

/**
 * <p>
 * Various functions needed to use the management information produced by the 
 * different subsystems of the server. Collected here to avoid code duplication.
 * </p>
 *
 * <strong>Concurrent Semantics</strong><br />
 *
 * Class is immutable - Thread Safe.
 *
 * @author Christopher Frost
 * @since 1.0
 */
public abstract class ManagementUtil {
    
    /**
     * To prevent instantiation of this util class
     */
    private ManagementUtil() {
    }
    
    /**
     * Goes through all the service references in the service registry getting 
     * the services that match the class of the provided T
     * 
     * @param <T>
     * @param clazz
     * @return the found services or an empty list if none can be found 
     */
    public static <T extends Info> List<T> listAll(BundleContext bundleContext, Class<T> clazz) {
        List<T> reply = null;
        try {
            ServiceReference[] serviceRefs = bundleContext.getAllServiceReferences(clazz.getName(), null);
            if (serviceRefs != null && serviceRefs.length > 0) {
                reply = new ArrayList<T>();
                for (ServiceReference serviceReference : serviceRefs) {
                    Object tempInfo = bundleContext.getService(serviceReference);
                    if (clazz.isInstance(tempInfo)) {
                        reply.add(clazz.cast(tempInfo));
                    }
                }
            }
        } catch (InvalidSyntaxException e) {
            /* no-op */
        }
        if (reply == null) {
            reply = Collections.emptyList();
        }
        return reply;
    }
    
}
