/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.framework;

import java.io.File;

import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.BundleException;
import org.osgi.framework.Version;

import com.springsource.server.osgi.manifest.VersionRange;
import com.springsource.server.osgi.provisioning.repository.MutableBundleRepository;

/**
 * Defines a contract for installing {@link Bundle Bundles} into an OSGi framework without knowing the exact
 * <code>Bundle</code> location.<p/>
 * 
 * <code>Bundles</code> can be identified using their {@link Bundle#getSymbolicName() symbolic name} and an optional
 * {@link VersionRange}. The <code>BundleInstaller</code> will search all known <code>Bundles</code> and install
 * the one with the highest matching {@link Version}.<p/>
 * 
 * If a <code>Bundle</code> with the given symbolic name and matching the supplied version range is already installed,
 * then the install phase is skipped and the existing <code>Bundle</code> is returned.<p/> Bundle location is
 * implementation-specific.<p/>
 * 
 * By default, installation of a given <code>Bundle</code> attempts to resolve all dependencies on other
 * <code>Bundles</code> eagerly by provisioning additional bundles on the fly. Optional dependencies are installed and
 * resolved where possible. If a required dependency cannot be satisfied, a {@link BundleException} is thrown.
 * Alternatively this processing my be deferred and then explicitly triggered via {@link #satisfyConstraints(Bundle)} or
 * {@link #satisfyConstraints(Bundle[])}.<p/>
 * 
 * All bundles that are implicitly installed, i.e. installed as a result of constraint satisfaction, will be started
 * automatically. Bundles that are installed explicitly are not started; their lifecycle management is the
 * responsibility of the caller.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Implementations <strong>must</strong> be thread safe.
 * 
 * @author Rob Harrop
 * @author Andy Wilkinson
 * @since 1.0
 */
public interface BundleInstaller {

    /**
     * Installs the {@link Bundle} identified by the given <code>bundleSymbolicName</code>. Equivalent to calling
     * {@link #installBundle(String, VersionRange)} passing in {@link VersionRange#NATURAL_NUMBER_RANGE}.<p/>
     * 
     * Satisfies constraints eagerly, starting any bundles that are installed as part of the satisfaction process.<p/>
     * 
     * @param bundleSymbolicName the name of the <code>Bundle</code> to install.
     * @return the installed <code>Bundle</code>.
     * @throws BundleException if an error occurs during installation.
     * @throws UnableToSatisfyDependenciesException if the bundle's dependencies cannot be satisfied
     * @see BundleContext#installBundle(String)
     */
    Bundle installBundle(String bundleSymbolicName) throws BundleException, UnableToSatisfyDependenciesException;

    /**
     * Installs the {@link Bundle} identified by the given <code>bundleSymbolicName</code>. Equivalent to calling
     * {@link #installBundle(String, VersionRange, boolean)} passing in {@link VersionRange#NATURAL_NUMBER_RANGE}.
     * 
     * @param bundleSymbolicName the name of the <code>Bundle</code> to install.
     * @param satisfyConstraints flag indicating whether or not constraints should be eagerly satisfied.
     * @return the installed <code>Bundle</code>.
     * @throws BundleException if an error occurs during installation.
     * @throws UnableToSatisfyDependenciesException if the bundle's dependencies cannot be satisfied
     * @see BundleContext#installBundle(String)
     */
    Bundle installBundle(String bundleSymbolicName, boolean satisfyConstraints) throws BundleException, UnableToSatisfyDependenciesException;

    /**
     * Installs the {@link Bundle} identified by the given <code>bundleSymbolicName</code> and with a {@link Version}
     * that {@link VersionRange#includes(Version) lies inside} the supplied {@link VersionRange}.<p/>
     * 
     * Satisfies constraints eagerly. starting any bundles that are installed as part of the satisfaction process.<p/>
     * 
     * In the event of multiple matching <code>Bundles</code>, the one with the highest <code>Version</code> is
     * installed.<p/P
     * 
     * @param bundleSymbolicName the name of the <code>Bundle</code> to install.
     * @param versionRange the range of <code>Versions</code> to match candidate <code>Bundles</code> against.
     * @return the installed <code>Bundle</code>.
     * @throws BundleException if an error occurs during installation.
     * @throws UnableToSatisfyDependenciesException if the bundle's dependencies cannot be satisfied
     * @see BundleContext#installBundle(String)
     */
    Bundle installBundle(String bundleSymbolicName, VersionRange versionRange) throws BundleException, UnableToSatisfyDependenciesException;

    /**
     * Installs the {@link Bundle} identified by the given <code>bundleSymbolicName</code> and with a {@link Version}
     * that {@link VersionRange#includes(Version) lies inside} the supplied {@link VersionRange}.<p/>
     * 
     * In the event of multiple matching <code>Bundles</code>, the one with the highest <code>Version</code> is
     * installed.
     * 
     * @param bundleSymbolicName the name of the <code>Bundle</code> to install.
     * @param versionRange the range of <code>Versions</code> to match candidate <code>Bundles</code> against.
     * @param satisfyConstraints flag indicating whether or not constraints should be eagerly satisfied.
     * @return the installed <code>Bundle</code>.
     * @throws BundleException if an error occurs during installation.
     * @throws UnableToSatisfyDependenciesException if the bundle's dependencies cannot be satisfied
     * @see BundleContext#installBundle(String)
     */
    Bundle installBundle(String bundleSymbolicName, VersionRange versionRange, boolean satisfyConstraints) throws BundleException,
        UnableToSatisfyDependenciesException;

    /**
     * Installs the {@link Bundle} at the given location and checks that its constraints can be satisfied. Equivalent to
     * calling <code>installBundleFromLocation(location, true)</code>. Any bundles that are installed as part of the
     * constraint satisfaction process will be started automatically.
     * 
     * @param location The location of the bundle to install
     * @return The installed bundle
     * @throws BundleException if an error occurs during installation.
     * @throws UnableToSatisfyDependenciesException if the bundle's dependencies cannot be satisfied
     */
    Bundle installBundleFromLocation(File location) throws UnableToSatisfyDependenciesException, BundleException;

    /**
     * Installs the Bundle at the given location and optionally satisfy the bundle's constraints.
     * 
     * @param location The location of the bundle to install.
     * @param satisfyConstraints Whether the bundle's constraints should be satisfied post install.
     * @throws BundleException if an error occurs during installation
     * @throws UnableToSatisfyDependenciesException if the bundle's dependencies cannot be satisfied
     */
    Bundle installBundleFromLocation(File location, boolean satisfyConstraints) throws UnableToSatisfyDependenciesException, BundleException;

    /**
     * Updates the supplied {@link Bundle} performing any necessary manifest re-writing, for example the expansion of
     * the bundle manifest's <code>Import-Library</code> header.
     * 
     * @param bundle The bundle to be updated.
     * @throws BundleException if an error ocurrs during update
     * @throws UnableToSatisfyDependenciesException if the updated bundle's dependencies cannot be satisfied
     */
    void updateBundle(Bundle bundle) throws BundleException, UnableToSatisfyDependenciesException;

    /**
     * Installs the {@link Library} identified by the given <code>libraryName</code> and with a {@link Version} that
     * {@link VersionRange#includes(Version) lies inside} the supplied {@link VersionRange}.<p/>
     * 
     * Satisfies constraints eagerly.<p/>
     * 
     * In the event of multiple matching libraries, the one with the highest <code>Version</code> is installed.<p/P
     * 
     * @param libraryName the name of the <code>Library</code> to install.
     * @param versionRange the range of <code>Versions</code> to match candidate <code>Library</code> against.
     * @return the installed <code>Library</code>.
     * @throws BundleException if an error occurs during installation.
     * @throws UnableToSatisfyDependenciesException if the library dependencies cannot be satisfied.
     */
    Library installLibrary(String libraryName, VersionRange versionRange) throws UnableToSatisfyDependenciesException, BundleException;

    /**
     * Attempts to satisfy the constraints of the given {@link Bundle Bundles}. An OsgiFrameworkException is thrown if
     * the constraints cannot be satisfied. Any bundles that are installed as part of the satisfaction process are
     * started.
     * 
     * @param bundles the bundles for which the constraints should be satisfied
     * @throws UnableToSatisfyDependenciesException if the bundle's constraints cannot be satisfied
     */
    void satisfyConstraints(Bundle[] bundles) throws UnableToSatisfyDependenciesException;

    /**
     * Attempts to satisfy the constraints of the given {@link Bundle}. An UnableToSatisfyBundleDependenciesException
     * is thrown if the constraints cannot be satisfied. Any bundles that are installed as part of the satisfaction
     * process are started.
     * 
     * @param bundle the bundle for which the constraints should be satisfied
     * @throws UnableToSatisfyDependenciesException if the bundle's constraints cannot be satisfied
     */
    void satisfyConstraints(Bundle bundle) throws UnableToSatisfyDependenciesException;

    /**
     * Gets the {@link MutableBundleRepository} used by this <code>BundleInstaller</code>.
     * 
     * @return the <code>MutableBundleRepository</code>.
     */
    MutableBundleRepository getBundleRepository();

    /**
     * Gets the import expander.
     * 
     * @return the import expander
     */
    ImportExpander getImportExpander();
}
