/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.framework;

import org.osgi.framework.Version;

/**
 * Signals an error resolving the dependencies of a bundle or library during installation.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Thread-safe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
@SuppressWarnings("serial")
public abstract class UnableToSatisfyDependenciesException extends Exception {

    private final String symbolicName;

    private final Version version;

    private final String failureDescription;

    /**
     * Creates a new <code>UnableToSatisfyDependenciesException</code>.
     * 
     * @param symbolicName the symbolic name of the entity that failed to install.
     * @param version the version of the entity that failed to install.
     * @param failureDescription the description of the dependency satisfaction problem.
     */
    protected UnableToSatisfyDependenciesException(String entity, String symbolicName, Version version, String failureDescription) {
        super(formatMessage(entity, symbolicName, version, failureDescription));
        this.symbolicName = symbolicName;
        this.version = version;
        this.failureDescription = failureDescription;
    }

    /**
     * Gets the symbolic name of the entity that failed to install.
     */
    public String getSymbolicName() {
        return this.symbolicName;
    }

    /**
     * Gets the version of the entity that failed to install.
     */
    public Version getVersion() {
        return this.version;
    }

    /**
     * Gets the full description of the dependencies that could not be satisfied.
     */
    public String getFailureDescription() {
        return this.failureDescription;
    }

    /**
     * Creates a formatted message to describe the dependency failure.
     */
    private static String formatMessage(String entity, String symbolicName, Version version, String failureDescription) {
        return "Unable to satisfy dependencies of " + entity + " '" + symbolicName + "' at version '" + version + "': " + failureDescription;
    }
}
