/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.framework.equinox;

import static com.springsource.server.osgi.framework.support.LocationUtils.convertToCanonicalFilePath;

import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.eclipse.osgi.service.resolver.BundleDescription;
import org.eclipse.osgi.service.resolver.BundleSpecification;
import org.eclipse.osgi.service.resolver.ExportPackageDescription;
import org.eclipse.osgi.service.resolver.HostSpecification;
import org.eclipse.osgi.service.resolver.ImportPackageSpecification;
import org.eclipse.osgi.service.resolver.PlatformAdmin;
import org.eclipse.osgi.service.resolver.State;
import org.eclipse.osgi.service.resolver.VersionConstraint;
import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.BundleException;
import org.osgi.framework.Constants;
import org.osgi.framework.Version;
import org.osgi.service.packageadmin.PackageAdmin;

import com.springsource.server.osgi.framework.BundleInstaller;
import com.springsource.server.osgi.framework.OsgiFrameworkException;
import com.springsource.server.osgi.framework.OsgiFrameworkUtils;
import com.springsource.server.osgi.framework.UnableToSatisfyBundleDependenciesException;
import com.springsource.server.osgi.framework.UnableToSatisfyDependenciesException;
import com.springsource.server.osgi.framework.support.AbstractBundleInstaller;
import com.springsource.server.osgi.framework.support.LocationUtils;
import com.springsource.server.osgi.manifest.VersionRange;
import com.springsource.server.osgi.provisioning.BundleDefinition;
import com.springsource.server.osgi.provisioning.repository.MutableBundleRepository;
import com.springsource.server.serviceability.ffdc.NonNull;

/**
 * {@link BundleInstaller} implementation for Equinox.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @author Andy Wilkinson
 * @since 1.0
 */
final class EquinoxBundleInstaller extends AbstractBundleInstaller {

    private final ResolutionFailureDetective detective;

    private final BundleContext bundleContext;

    private final PlatformAdmin serverAdmin;

    private final Object bundleContextAndResolverLock = LockManager.getMainLock();

    /**
     * Creates a new <code>EquinoxBundleInstaller</code>.
     * 
     * @param bundleRepository the {@link MutableBundleRepository} to search for <code>Bundles</code>.
     * @param packageAdmin the OSGi {@link PackageAdmin} service.
     * @param bundleContext the system {@link BundleContext}.
     * @param serverAdmin the Equinox {@link PlatformAdmin} service.
     */
    public EquinoxBundleInstaller(MutableBundleRepository bundleRepository, PackageAdmin packageAdmin, BundleContext bundleContext,
        PlatformAdmin serverAdmin) {
        super(bundleRepository, packageAdmin, bundleContext);
        this.bundleContext = bundleContext;
        this.serverAdmin = serverAdmin;
        this.detective = new ResolutionFailureDetective(this.serverAdmin);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected Bundle doInstallBundle(String resolvedBundleLocation, boolean satisfyConstraints) throws UnableToSatisfyDependenciesException,
        BundleException {

        InstallState installState = new InstallState();
        InstallResult installResult;
        boolean failed;

        synchronized (bundleContextAndResolverLock) {
            installResult = doInstallBundleInternal(resolvedBundleLocation, installState, satisfyConstraints);
            failed = !installResult.success;
            if (!failed) {
                if (satisfyConstraints) {
                    resolveBundle(installResult.bundle);
                    if (installResult.bundle.getState() == Bundle.INSTALLED) {
                        failed = true;
                    }
                }
            }
        }

        if (failed) {
            throw new UnableToSatisfyBundleDependenciesException(installResult.bundle.getSymbolicName(),
                extractVersionFromBundle(installResult.bundle), this.detective.generateFailureDescription(installResult.bundle));
        } else {
            for (Bundle bundle : installState.installedBundles) {
                if (!bundle.equals(installResult.bundle)) {
                    startBundle(bundle);
                }
            }
        }

        return installResult.bundle;
    }

    /**
     * Performs actual installation. Tracks locations that are being installed to avoid reentrancy problems when faced
     * with bundles that import and export the same package.
     */
    private InstallResult doInstallBundleInternal(String resolvedBundleLocation, InstallState state, boolean satisfyConstraints)
        throws UnableToSatisfyDependenciesException, BundleException {

        if (this.tracer.isInfoEnabled()) {
            this.tracer.traceInfo(String.format("doInstallBundleInternal(%s, %s, %s)", resolvedBundleLocation, state, satisfyConstraints));
        }
        InstallResult result = new InstallResult();

        Bundle bundle;
        synchronized (this.bundleContextAndResolverLock) {
            bundle = this.bundleContext.installBundle(resolvedBundleLocation);
        }
        result.bundle = bundle;

        resolveFragments(bundle, state);

        if (this.tracer.isInfoEnabled()) {
            this.tracer.traceInfo(String.format("Installed bundle %s", bundle));
        }

        if (!state.seenLocation(resolvedBundleLocation)) {
            state.recordLocation(resolvedBundleLocation);
            if (satisfyConstraints && !satisfyConstraints(bundle, state, new Bundle[] { bundle })) {
                result.success = false;
            } else {
                state.recordInstallation(bundle);
            }
        }

        return result;
    }

    /**
     * Attempts to eagerly install the fragments for a given {@link Bundle}.
     * 
     * @param bundle the <code>Bundle</code> to install fragments for.
     */
    private void resolveFragments(Bundle bundle, InstallState state) {
        Set<? extends BundleDefinition> fragments = this.bundleRepository.findByFragmentHost(bundle.getSymbolicName(),
            OsgiFrameworkUtils.getBundleVersion(bundle));
        for (BundleDefinition fragment : fragments) {
            try {
                doInstallBundleInternal(LocationUtils.convertToCanonicalFilePath(fragment.getLocation()), state, true);
            } catch (Exception ex) {
                throw new OsgiFrameworkException("Error installing fragment from '" + fragment.getLocation() + "' for bundle '"
                    + bundle.getSymbolicName() + "'", ex);
            }
        }
    }

    /**
     * Attempts to eagerly satisfy constraints for the supplied {@link Bundle}.
     * 
     * @param bundle the <code>Bundle</code> to satisfy constraints for.
     * @param installState the current state of this installation command.
     * @param groupedBundles the group of bundles including bundle
     * @throws BundleException if resolution fails, or if installation of a dependent <code>Bundle</code> fails.
     * @throws UnableToSatisfyDependenciesException if the bundle's dependencies cannot be satisfied
     * @return <code>true</code> if the constraints were successfully satisfied
     */
    private boolean satisfyConstraints(Bundle bundle, InstallState installState, @NonNull Bundle[] groupedBundles) throws BundleException,
        UnableToSatisfyDependenciesException {
        if (this.tracer.isInfoEnabled()) {
            this.tracer.traceInfo(String.format("satisfyConstraints(%s, %s, %s)", bundle, installState, groupedBundles));
        }

        VersionConstraint[] unsatisfiedConstraints;

        State state = this.serverAdmin.getState(true);
        synchronized (this.bundleContextAndResolverLock) {
            state.setResolver(this.serverAdmin.getResolver());
            state.resolve(true);
            unsatisfiedConstraints = findUnsatisfiedConstraints(bundle, state);
        }

        boolean success = true;

        for (VersionConstraint constraint : unsatisfiedConstraints) {
            String name = constraint.getName();
            VersionRange vr = EquinoxUtils.convertEquinoxVersionRange(constraint.getVersionRange());
            if (constraint instanceof ImportPackageSpecification) {

                Set<? extends BundleDefinition> definitions = this.bundleRepository.findByExportedPackage(name, vr);
                if (this.tracer.isInfoEnabled()) {
                    this.tracer.traceInfo(String.format("Found %d matches for constraint %s", definitions.size(), constraint));
                }

                Object resolution = ((ImportPackageSpecification) constraint).getDirective(Constants.RESOLUTION_DIRECTIVE);
                boolean optionalOrDynamic = ImportPackageSpecification.RESOLUTION_OPTIONAL.equals(resolution)
                    || ImportPackageSpecification.RESOLUTION_DYNAMIC.equals(resolution);

                if (success && definitions == null || definitions.isEmpty()) {
                    if (!optionalOrDynamic) {
                        boolean exists = existsInGroupedBundles(bundle, groupedBundles, state, constraint);
                        if (!exists) {
                            exists = existsInResolvedBundles(state, constraint);
                        }
                        success = success && exists;

                    }
                }
                for (BundleDefinition definition : definitions) {
                    InstallResult result = doInstallBundleInternal(convertToCanonicalFilePath(definition.getLocation()), installState, true);
                    if (!optionalOrDynamic) {
                        success = success && result.success;
                    }
                }
            } else if (constraint instanceof BundleSpecification) {
                BundleDefinition definition = this.bundleRepository.findBySymbolicName(name, vr);
                if (definition != null) {
                    if (this.tracer.isInfoEnabled()) {
                        this.tracer.traceInfo(String.format("Matched on Require-Bundle constraint: %s with: %s", constraint, definition));
                    }

                    InstallResult result = doInstallBundleInternal(convertToCanonicalFilePath(definition.getLocation()), installState, true);
                    success = success && result.success;
                } else {
                    if (this.tracer.isInfoEnabled()) {
                        this.tracer.traceInfo(String.format("Unable to satisfy Require-Bundle: %s", constraint));
                    }

                    success = false;
                }
            } else if (constraint instanceof HostSpecification) {
                BundleDefinition definition = this.bundleRepository.findBySymbolicName(name, vr);
                if (definition != null) {
                    if (this.tracer.isInfoEnabled()) {
                        this.tracer.traceInfo(String.format("Matched on Fragment-Host constraint: %s with: %s", constraint, definition));
                    }

                    InstallResult result = doInstallBundleInternal(convertToCanonicalFilePath(definition.getLocation()), installState, true);
                    success = success && result.success;
                } else {
                    if (this.tracer.isInfoEnabled()) {
                        this.tracer.traceInfo(String.format("Unable to satisfy Fragment-Host: %s", constraint));
                    }

                    success = false;
                }
            }
        }
        return success;
    }

    private boolean existsInResolvedBundles(State state, VersionConstraint constraint) {
        BundleDescription[] bundles = state.getBundles();
        for (BundleDescription bd : bundles) {
            if (bd.isResolved()) {
                ExportPackageDescription[] exportPackages = bd.getExportPackages();
                for (ExportPackageDescription epd : exportPackages) {
                    if (constraint.isSatisfiedBy(epd)) {
                        return true;
                    }
                }
            }
        }
        return false;
    }

    /**
     * Queries whether a {@link VersionConstraint} is satisfied by a bundle that is present in the same group (app) as
     * the constrained bundle.
     */
    private boolean existsInGroupedBundles(Bundle bundle, Bundle[] groupedBundles, State state, VersionConstraint constraint) {
        if (this.tracer.isInfoEnabled()) {
            this.tracer.traceInfo("Checking if constraint '" + constraint.getName() + " " + constraint.getVersionRange()
                + "' can be satisfied by grouped bundles.");
        }
        boolean existsInGroup = false;
        if (groupedBundles != null) {
            for (Bundle grouped : groupedBundles) {
                if (this.tracer.isDebugEnabled()) {
                    this.tracer.traceDebug("Checking bundle " + grouped);
                }
                BundleDescription bd = state.getBundle(grouped.getBundleId());
                ExportPackageDescription[] exportPackages = bd.getExportPackages();
                for (ExportPackageDescription exportPackageDescription : exportPackages) {
                    if (this.tracer.isDebugEnabled()) {
                        this.tracer.traceDebug("Exported package: " + exportPackageDescription.getName() + " "
                            + exportPackageDescription.getVersion());
                    }
                    if (constraint.isSatisfiedBy(exportPackageDescription)) {
                        existsInGroup = true;
                        break;
                    }
                }
            }
        }
        if (this.tracer.isInfoEnabled()) {
            this.tracer.traceInfo("Satisfied in group: " + existsInGroup);
        }
        return existsInGroup;
    }

    /**
     * {@inheritDoc}
     */
    public void satisfyConstraints(Bundle[] bundles) throws UnableToSatisfyDependenciesException {
        InstallState installState = new InstallState();
        for (Bundle bundle : bundles) {
            boolean failed = false;
            try {
                if (!satisfyConstraints(bundle, installState, bundles)) {
                    failed = true;
                }
            } catch (BundleException e) {
                processSatisfyConstraintsException(bundle, e);
            } catch (RuntimeException e) {
                processSatisfyConstraintsException(bundle, e);
            }
            if (failed) {
                String symbolicName = bundle.getSymbolicName();
                throw new UnableToSatisfyBundleDependenciesException(symbolicName, extractVersionFromBundle(bundle),
                    this.detective.generateFailureDescription(bundle));
            }
        }

        resolveBundle(bundles);

        for (Bundle bundle : bundles) {
            if (bundle.getState() == Bundle.INSTALLED) {
                String symbolicName = bundle.getSymbolicName();
                throw new UnableToSatisfyBundleDependenciesException(symbolicName, extractVersionFromBundle(bundle),
                    this.detective.generateFailureDescription(bundle));
            }
        }

        List<Bundle> suppliedBundles = Arrays.asList(bundles);

        for (Bundle bundle : installState.installedBundles) {
            if (!suppliedBundles.contains(bundle)) {
                startBundle(bundle);
            }
        }
    }

    private void startBundle(Bundle bundle) {
        if (bundle.getHeaders().get("Fragment-Host") == null) {
            String bundleActivationPolicy = (String) bundle.getHeaders().get("Bundle-ActivationPolicy");
            if (bundleActivationPolicy == null) {
                try {
                    bundle.start();
                } catch (BundleException be) {
                    throw new OsgiFrameworkException("Failed to start bundle '" + bundle.getSymbolicName() + "'", be);
                }
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    public void satisfyConstraints(Bundle bundle) throws UnableToSatisfyDependenciesException {
        satisfyConstraints(new Bundle[] { bundle });
    }

    /**
     * Throw an appropriately wrapped exception caught during constraint satisfaction.
     * 
     * @param bundle
     * @param e
     */
    private void processSatisfyConstraintsException(Bundle bundle, Exception e) {
        throw new OsgiFrameworkException("Error occurred during satisfaction of constraints for bundle '" + bundle.getSymbolicName() + "'.", e);
    }

    /**
     * Finds all unsatisfied constraints for the supplied {@link Bundle}.
     * 
     * @param bundle the <code>Bundle</code>.
     * @return the unsatisfied constraints.
     */
    private VersionConstraint[] findUnsatisfiedConstraints(Bundle bundle, State state) {
        BundleDescription bd = state.getBundle(bundle.getBundleId());
        return state.getStateHelper().getUnsatisfiedConstraints(bd);
    }

    /**
     * Resolves the supplied {@link Bundle}.
     */
    void resolveBundle(Bundle bundle) {
        synchronized (this.bundleContextAndResolverLock) {
            if (tracer.isInfoEnabled())
                tracer.traceInfo("Resolving bundle '%s'", bundle.getSymbolicName());
            this.packageAdmin.resolveBundles(new Bundle[] { bundle });
        }
    }

    /**
     * Resolves the supplied {@link Bundle Bundles}.
     */
    void resolveBundle(Bundle[] bundles) {
        synchronized (this.bundleContextAndResolverLock) {
            this.packageAdmin.resolveBundles(bundles);
        }
    }

    ResolutionFailureDetective getResolutionFailureDetective() {
        return this.detective;
    }

    /**
     * Stores the state of a given installation call chain.
     */
    private static class InstallState {

        private final Set<String> recordedLocations = new HashSet<String>(64);

        private final Set<Bundle> installedBundles = new HashSet<Bundle>(64);

        /**
         * Record a bundle location as being installed.
         * 
         * @param location the location to record.
         */
        public void recordLocation(String location) {
            this.recordedLocations.add(location);
        }

        /**
         * Queries whether or not a particular bundle location has been seen during installation.
         * 
         * @param location the bundle location.
         * @return <code>true</code> if the location has been seen, otherwise <code>false</code>.
         */
        public boolean seenLocation(String location) {
            return this.recordedLocations.contains(location);
        }

        /**
         * Record a bundle as having been installed.
         * 
         * @param bundle the bundle that has been installed.
         */
        public void recordInstallation(Bundle bundle) {
            installedBundles.add(bundle);
        }

        /**
         * Returns a Set of all the bundles that have been installed by the installation call chain
         * 
         * @return the installed bundles
         */
        public Set<Bundle> getInstalledBundles() {
            return this.installedBundles;
        }
    }

    class InstallResult {

        Bundle bundle = null;

        boolean success = true;
    }

    private Version extractVersionFromBundle(Bundle bundle) {
        String versionString = (String) bundle.getHeaders().get("Bundle-Version");
        if (versionString == null) {
            return Version.emptyVersion;
        } else {
            return new Version(versionString);
        }
    }
}
