/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.framework.equinox;

import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

import javax.management.MBeanServer;
import javax.management.MBeanServerFactory;
import javax.management.ObjectName;

import org.eclipse.osgi.service.resolver.BundleDescription;
import org.eclipse.osgi.service.resolver.ExportPackageDescription;
import org.eclipse.osgi.service.resolver.ImportPackageSpecification;
import org.eclipse.osgi.service.resolver.PlatformAdmin;
import org.eclipse.osgi.service.resolver.ResolverError;
import org.eclipse.osgi.service.resolver.State;
import org.eclipse.osgi.service.resolver.VersionConstraint;
import org.osgi.framework.Bundle;
import org.osgi.framework.Constants;
import org.springframework.util.StringUtils;

/**
 * Helper class that analyses resolution failures and generates a human-readable failure description.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
public final class ResolutionFailureDetective {

    private final PlatformAdmin serverAdmin;

    /**
     * Creates a new <code>ResolutionFailureDetective</code>.
     * 
     * @param serverAdmin the Equinox {@link PlatformAdmin} service.
     */
    public ResolutionFailureDetective(PlatformAdmin serverAdmin) {
        this.serverAdmin = serverAdmin;
    }

    /**
     * Generates a description of all the resolver errors for the supplied {@link Bundle}.
     * 
     * @param bundle the <code>Bundle</code> with resolve errors.
     * @return the failure description.
     */
    public String generateFailureDescription(Bundle bundle) {
        
        List<?> mBeanServers = MBeanServerFactory.findMBeanServer(null);
        if (mBeanServers != null && mBeanServers.size() > 0) {
            MBeanServer mBeanServer = (MBeanServer)mBeanServers.get(0);
            if (mBeanServer != null) {
                try {
                    ObjectName ffdcControlMBeanName = new ObjectName("com.springsource.server:type=SystemDumpControl");
                    mBeanServer.invoke(ffdcControlMBeanName, "generateDump", null, null);
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
        }
        
        if (bundle.getState() != Bundle.INSTALLED) {
            return null;
        }

        State state = this.serverAdmin.getState(false);
        BundleDescription bundleDescription = state.getBundle(bundle.getBundleId());

        StringBuilder sb = new StringBuilder();
        sb.append("Cannot resolve: ").append(bundleDescription.getSymbolicName()).append("\n");

        ResolverError[] resolverErrors = gatherResolverErrors(bundleDescription, state);
        if (resolverErrors.length > 0) {
            indent(sb, 1);
            sb.append("Resolver report:\n");
        }
        for (ResolverError resolverError : resolverErrors) {
            indent(sb, 2);
            formatResolverError(resolverError, sb, state);
            sb.append("\n");
        }

        if (resolverErrors.length == 0) {
            VersionConstraint[] unsatisfiedLeaves = this.serverAdmin.getStateHelper().getUnsatisfiedLeaves(
                new BundleDescription[] { bundleDescription });
            if (unsatisfiedLeaves.length > 0) {
                indent(sb, 1);
                sb.append("Unsatisfied leaf constraints:\n");
            }
            for (VersionConstraint versionConstraint : unsatisfiedLeaves) {
                if (!isOptional(versionConstraint)) {
                    indent(sb, 2);
                    formatConstraint(versionConstraint, sb);
                    sb.append("\n");
                }
            }
        }

        return sb.toString();

    }

    private ResolverError[] gatherResolverErrors(BundleDescription bundleDescription, State state) {
        Set<ResolverError> resolverErrors = new LinkedHashSet<ResolverError>();
        Collections.addAll(resolverErrors, state.getResolverErrors(bundleDescription));
        BundleDescription[] bundles = state.getBundles();
        for (BundleDescription bd : bundles) {
            if (!bd.isResolved()) {
                Collections.addAll(resolverErrors, state.getResolverErrors(bd));
            }
        }
        return resolverErrors.toArray(new ResolverError[resolverErrors.size()]);
    }

    /**
     * Finds the member of <code>candidates</code> that is the nearest match to <code>match</code>.
     * 
     * @param match the string to match against.
     * @param candidates the candidates to search.
     * @return the nearest match.
     */
    private String nearestMatch(String match) {
        Set<String> candidates = gatherExports();
        int nearestDistance = Integer.MAX_VALUE;
        String nearestMatch = null;

        for (String candidate : candidates) {
            int distance = calculateStringDistance(match, candidate);
            if (distance < nearestDistance) {
                nearestDistance = distance;
                nearestMatch = candidate;
            }
        }

        return nearestMatch;
    }

    /**
     * Calculate the distance between the given two Strings according to the Levenshtein algorithm.
     * 
     * @param s1 the first String
     * @param s2 the second String
     * @return the distance value
     */
    private int calculateStringDistance(String s1, String s2) {
        if (s1.length() == 0) {
            return s2.length();
        }
        if (s2.length() == 0) {
            return s1.length();
        }
        int d[][] = new int[s1.length() + 1][s2.length() + 1];

        for (int i = 0; i <= s1.length(); i++) {
            d[i][0] = i;
        }
        for (int j = 0; j <= s2.length(); j++) {
            d[0][j] = j;
        }

        for (int i = 1; i <= s1.length(); i++) {
            char s_i = s1.charAt(i - 1);
            for (int j = 1; j <= s2.length(); j++) {
                int cost;
                char t_j = s2.charAt(j - 1);
                if (Character.toLowerCase(s_i) == Character.toLowerCase(t_j)) {
                    cost = 0;
                } else {
                    cost = 1;
                }
                d[i][j] = Math.min(Math.min(d[i - 1][j] + 1, d[i][j - 1] + 1), d[i - 1][j - 1] + cost);
            }
        }

        return d[s1.length()][s2.length()];
    }

    /**
     * Gathers all the package exports.
     * 
     * @return the exported packages.
     */
    private Set<String> gatherExports() {
        State state = this.serverAdmin.getState(false);
        ExportPackageDescription[] exportedPackages = state.getExportedPackages();
        Set<String> exports = new HashSet<String>(exportedPackages.length);
        for (ExportPackageDescription epd : exportedPackages) {
            exports.add(epd.getName());
        }
        return exports;
    }

    private void formatResolverError(ResolverError resolverError, StringBuilder sb, State state) {
        VersionConstraint constraint = resolverError.getUnsatisfiedConstraint();
        sb.append("Bundle: ").append(resolverError.getBundle()).append(" - ");
        if (resolverError.getType() == ResolverError.IMPORT_PACKAGE_USES_CONFLICT) {
            sb.append("Uses Conflict: ").append(constraint);
            ExportPackageDescription[] exportedPackages = findMatchingExports((ImportPackageSpecification) constraint, state);
            for (ExportPackageDescription exportPackageDescription : exportedPackages) {
                sb.append("\n");
                indent(sb, 3);
                sb.append("Possible Supplier: ").append(exportPackageDescription.getExporter()).append(" - ").append(exportPackageDescription);
                String[] possibleUsesClashes = findPossibleUsesClashes(exportPackageDescription, constraint.getBundle());
                sb.append("\n");
                indent(sb, 4);
                sb.append("Possible Conflicts: ").append(StringUtils.arrayToCommaDelimitedString(possibleUsesClashes));
            }
        } else {
            sb.append(resolverError);
        }
    }

    private ExportPackageDescription[] findMatchingExports(ImportPackageSpecification ips, State state) {
        Set<ExportPackageDescription> exports = new HashSet<ExportPackageDescription>();
        ExportPackageDescription[] exportedPackages = state.getExportedPackages();
        for (ExportPackageDescription export : exportedPackages) {
            if (export.getName().equals(ips.getName())) {
                exports.add(export);
            }
        }
        return exports.toArray(new ExportPackageDescription[exports.size()]);
    }

    private String[] findPossibleUsesClashes(ExportPackageDescription ep, BundleDescription bd) {
        Set<String> possible = new HashSet<String>();
        Set<String> uses = extractUses(ep);
        ImportPackageSpecification[] importPackages = bd.getImportPackages();
        for (ImportPackageSpecification ips : importPackages) {
            if (uses.contains(ips.getName())) {
                possible.add(ips.getName());
            }
        }
        return uses.toArray(new String[uses.size()]);
    }

    private Set<String> extractUses(ExportPackageDescription ep) {
        String[] usesDirective = (String[]) ep.getDirective(Constants.USES_DIRECTIVE);
        if (usesDirective != null) {
            Set<String> uses = new HashSet<String>();
            Collections.addAll(uses, usesDirective);
            return uses;
        } else {
            return Collections.emptySet();
        }
    }

    private void formatConstraint(VersionConstraint versionConstraint, StringBuilder sb) {
        String constraintInformation = versionConstraint.toString();
        String bundleInQuestion = versionConstraint.getBundle().toString();
        sb.append("Bundle: ").append(bundleInQuestion).append(" - ").append(constraintInformation);
        if (versionConstraint instanceof ImportPackageSpecification) {
            sb.append("\n");
            indent(sb, 3);
            sb.append("Did you mean: '").append(nearestMatch(versionConstraint.getName())).append("'?");
        }
    }

    private boolean isOptional(VersionConstraint versionConstraint) {
        if (versionConstraint instanceof ImportPackageSpecification) {
            ImportPackageSpecification ips = (ImportPackageSpecification) versionConstraint;
            return !ImportPackageSpecification.RESOLUTION_STATIC.equals(ips.getDirective(Constants.RESOLUTION_DIRECTIVE));
        }
        return false;
    }

    /**
     * Indent the supplied {@link StringBuilder} to the supplied <code>level</code>.
     * 
     * @param out the <code>StringBuilder</code> to indent.
     * @param level the indentation level.
     */
    private void indent(StringBuilder out, int level) {
        for (int n = 0; n < level; n++) {
            out.append("  ");
        }
    }
}
