/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.framework.equinox;

import java.io.IOException;
import java.lang.instrument.ClassFileTransformer;
import java.lang.instrument.IllegalClassFormatException;
import java.net.URL;
import java.security.ProtectionDomain;
import java.util.Collection;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.CopyOnWriteArrayList;

import org.eclipse.osgi.baseadaptor.BaseData;
import org.eclipse.osgi.baseadaptor.bundlefile.BundleEntry;
import org.eclipse.osgi.baseadaptor.loader.ClasspathEntry;
import org.eclipse.osgi.baseadaptor.loader.ClasspathManager;
import org.eclipse.osgi.framework.adaptor.ClassLoaderDelegate;
import org.eclipse.osgi.internal.baseadaptor.DefaultClassLoader;
import org.eclipse.osgi.service.resolver.PlatformAdmin;
import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;

import com.springsource.server.osgi.framework.InstrumentableClassLoader;
import com.springsource.server.osgi.framework.OsgiFrameworkUtils;
import com.springsource.server.osgi.framework.ServerClassNotFoundException;
import com.springsource.server.osgi.framework.ServerNoClassDefFoundError;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * Extension to {@link DefaultClassLoader} that adds instrumentation support.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * As threadsafe as <code>DefaultClassLoader</code>.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
class ServerBundleClassLoader extends DefaultClassLoader implements InstrumentableClassLoader {

    private static final String[] EXCLUDED_PACKAGES = new String[] { "java.", "javax.", "sun.", "oracle." };

    private static final String HEADER_INSTRUMENT_PACKAGE = "Instrument-Package";

    private static final Tracer TRACER = TracerFactory.createTracer(ServerBundleClassLoader.class);

    private final List<ClassFileTransformer> classFileTransformers = new CopyOnWriteArrayList<ClassFileTransformer>();

    private final String[] instrumentedPackages;

    private final String[] classpath;

    private final String bundleScope;

    private volatile boolean instrumented;

    /**
     * Constructs a new <code>ServerBundleClassLoader</code>.
     * 
     * @param parent the parent <code>ClassLoader</code>.
     * @param delegate the delegate for this ClassLoader</code>
     * @param domain the domain for this ClassLoader</code>
     * @param bundledata the bundledata for this ClassLoader</code>
     * @param classpath the classpath for this ClassLoader</code>
     */
    public ServerBundleClassLoader(ClassLoader parent, ClassLoaderDelegate delegate, ProtectionDomain domain, BaseData bundledata,
        String[] classpath) {
        super(parent, delegate, domain, bundledata, classpath);
        this.classpath = classpath;
        this.bundleScope = OsgiFrameworkUtils.getScope(bundledata.getBundle());
        this.instrumentedPackages = findInstrumentedPackages(bundledata.getBundle());
    }

    /**
     * {@inheritDoc}
     */
    public void addClassFileTransformer(ClassFileTransformer transformer) {
        this.instrumented = true;
        synchronized (this.classFileTransformers) {
            if (this.classFileTransformers.contains(transformer)) {
                return;
            }
            this.classFileTransformers.add(transformer);
        }
        Bundle[] bundles = getDependencyBundles(false);
        for (Bundle bundle : bundles) {
            if (propagateInstrumentationTo(bundle)) {
                ClassLoader bundleClassLoader = getBundleClassLoader(bundle);
                if (bundleClassLoader instanceof ServerBundleClassLoader) {
                    ((ServerBundleClassLoader) bundleClassLoader).addClassFileTransformer(transformer);
                }
            }
        }
    }

    /**
     * @param bundle
     * @return
     */
    private ClassLoader getBundleClassLoader(Bundle bundle) {
        return EquinoxUtils.getBundleClassLoader(bundle, getPlatformAdmin());
    }

    /**
     * {@inheritDoc}
     */
    @Override protected Class<?> loadClass(String name, boolean resolve) throws ClassNotFoundException {
        try {
            return super.loadClass(name, resolve);
        } catch (ClassNotFoundException e) {
            throw new ServerClassNotFoundException(this, e);
        } catch (NoClassDefFoundError e) {
            throw new ServerNoClassDefFoundError(this, e);
        }
    }

    /**
     * {@inheritDoc}
     * 
     * Propagates requests for META-INF resources to bundles that this ClassLoader's bundle is wired to.
     */
    @Override public Enumeration<URL> getResources(String name) throws IOException {
        if (isMetaInfResource(name)) {
            Set<URL> found = new HashSet<URL>();
            addAll(found, doGetResources(name));
            Bundle[] bundles = getDependencyBundles(true);
            for (Bundle bundle : bundles) {
                if (!EquinoxUtils.isSystemBundle(bundle)) {
                    ClassLoader bundleClassLoader = getBundleClassLoader(bundle);
                    if (bundleClassLoader instanceof ServerBundleClassLoader) {
                        ServerBundleClassLoader pbcl = (ServerBundleClassLoader) bundleClassLoader;
                        addAll(found, pbcl.doGetResources(name));
                    }
                }
            }
            return new IteratorEnumerationAdaptor<URL>(found.iterator());
        } else {
            return doGetResources(name);
        }
    }

    /**
     * {@inheritDoc}
     * 
     * Propagates requests for META-INF resources to bundles that this ClassLoader's bundle is wired to.
     */
    @Override public URL getResource(String name) {
        if (isMetaInfResource(name)) {
            URL resource = super.getResource(name);
            if (resource == null) {
                Bundle[] bundles = getDependencyBundles(true);
                for (Bundle bundle : bundles) {
                    if (!EquinoxUtils.isSystemBundle(bundle)) {
                        ClassLoader bundleClassLoader = getBundleClassLoader(bundle);
                        if (bundleClassLoader instanceof ServerBundleClassLoader) {
                            ServerBundleClassLoader pbcl = (ServerBundleClassLoader) bundleClassLoader;
                            resource = pbcl.doGetResource(name);
                            if (resource != null) {
                                break;
                            }
                        }
                    }
                }
            }
            return resource;
        } else {
            return doGetResource(name);
        }
    }

    /**
     * Queries whether or not the supplied resource name is a META-INF resource.
     * 
     * @param name the resource name.
     * @return <code>true</code> if the resource is a META-INF resource.
     */
    private boolean isMetaInfResource(String name) {
        return name.startsWith("/META-INF") || name.startsWith("META-INF");
    }

    private Enumeration<URL> doGetResources(String name) throws IOException {
        return super.getResources(name);
    }

    private URL doGetResource(String name) {
        return super.getResource(name);
    }

    private void addAll(Collection<URL> target, Enumeration<URL> source) {
        while (source.hasMoreElements()) {
            target.add(source.nextElement());
        }
    }

    /**
     * Queries whether this <code>ServerBundleClassLoader</code> has been
     * {@link #addClassFileTransformer(ClassFileTransformer) instrumented}.
     * 
     * @return <code>true</code> if instrumented, otherwise <code>false</code>.
     */
    public boolean isInstrumented() {
        return this.instrumented;
    }

    /**
     * Finds the explicit list of packages to include in instrumentation (if specified).
     */
    private String[] findInstrumentedPackages(Bundle bundle) {
        String headerValue = (String) bundle.getHeaders().get(HEADER_INSTRUMENT_PACKAGE);
        if (headerValue == null || headerValue.length() == 0) {
            return new String[0];
        } else {
            String[] vals = headerValue.split(",");
            String[] packageNames = new String[vals.length];
            for (int x = 0; x < packageNames.length; x++) {
                packageNames[x] = vals[x].trim();
            }
            return packageNames;
        }
    }

    private boolean propagateInstrumentationTo(Bundle bundle) {
        return !EquinoxUtils.isSystemBundle(bundle) && this.bundleScope != null && this.bundleScope.equals(OsgiFrameworkUtils.getScope(bundle));
    }

    private boolean shouldInstrument(String className) {
        return includedForInstrumentation(className) && !excludedFromInstrumentation(className);
    }

    private boolean includedForInstrumentation(String className) {
        if (this.instrumentedPackages.length == 0) {
            return true;
        }
        for (String instrumentedPackage : this.instrumentedPackages) {
            if (className.startsWith(instrumentedPackage)) {
                return true;
            }
        }
        return false;
    }

    private boolean excludedFromInstrumentation(String className) {
        for (String excludedPackage : EXCLUDED_PACKAGES) {
            if (className.startsWith(excludedPackage)) {
                return true;
            }
        }
        return false;
    }

    /**
     * {@inheritDoc}
     */
    public ThrowAwayClassLoader createThrowAway() {
        if (TRACER.isInfoEnabled()) {
            TRACER.traceInfo("Creating throw away classloader for: " + this.manager.getBaseData().getSymbolicName());
        }
        final ClasspathManager manager = new ClasspathManager(this.manager.getBaseData(), this.classpath, this);
        manager.initialize();
        return new ThrowAwayClassLoader(manager);
    }

    /**
     * {@inheritDoc}
     */
    @Override public Class<?> defineClass(String name, byte[] classbytes, ClasspathEntry classpathEntry, BundleEntry entry) {

        byte[] transformedBytes = classbytes;
        if (shouldInstrument(name)) {
            for (ClassFileTransformer transformer : this.classFileTransformers) {
                try {
                    String transformName = name.replaceAll("\\.", "/");
                    if (TRACER.isInfoEnabled()) {
                        TRACER.traceInfo(String.format("Attempting transform for class %s using transformer %s", name, transformer));
                    }
                    byte[] transform = transformer.transform(this, transformName, null, this.domain, transformedBytes);
                    if (transform != null) {
                        if (TRACER.isInfoEnabled()) {
                            TRACER.traceInfo(String.format("Transformed class %s using transformer %s", name, transformer));
                        }
                        transformedBytes = transform;
                    }
                } catch (IllegalClassFormatException e) {
                    throw new ClassFormatError("Error reading class from bundle entry '" + entry.getName() + "'. " + e.getMessage());
                }
            }
        }
        try {
            return super.defineClass(name, transformedBytes, classpathEntry, entry);
        } catch (NoClassDefFoundError e) {
            throw new ServerNoClassDefFoundError(this, e);
        }
    }

    @Override public String toString() {
        return String.format("%s: [bundle=%s]", getClass().getSimpleName(), this.delegate);
    }

    private Bundle[] getDependencyBundles(boolean includeDependenciesFragments) {
        Bundle bundle = this.manager.getBaseData().getBundle();
        BundleContext systemBundleContext = getBundleContext();
        PlatformAdmin serverAdmin = getPlatformAdmin();
        Bundle[] deps = EquinoxUtils.getDirectDependencies(bundle, systemBundleContext, serverAdmin, includeDependenciesFragments);
        return deps;
    }

    /**
     * Gets the {@link BundleContext} for this ClassLoader's {@link Bundle}.
     * 
     * @return the <code>BundleContext</code>.
     */
    private BundleContext getBundleContext() {
        return this.manager.getBaseData().getAdaptor().getContext();
    }

    /**
     * Gets the {@link PlatformAdmin} service.
     * 
     * @return the <code>PlatformAdmin</code> service.
     */
    private PlatformAdmin getPlatformAdmin() {
        return this.manager.getBaseData().getAdaptor().getPlatformAdmin();
    }

    /**
     * Throwaway classloader for OSGi bundles.<p/>
     * 
     * <strong>Concurrent Semantics</strong><br />
     * 
     * As threadsafe as {@link ClassLoader}.
     * 
     * @author Rob Harrop
     * @since 1.0
     */
    final class ThrowAwayClassLoader extends ClassLoader {

        private final ConcurrentMap<String, Class<?>> loadedClasses = new ConcurrentHashMap<String, Class<?>>();

        private final ClasspathManager manager;

        /**
         * @param manager
         */
        private ThrowAwayClassLoader(ClasspathManager manager) {
            this.manager = manager;
        }

        /**
         * {@inheritDoc}
         */
        @Override public Class<?> loadClass(String name, boolean resolve) throws ClassNotFoundException {
            if (!shouldInstrument(name)) {
                return ServerBundleClassLoader.this.loadClass(name, resolve);
            }
            Class<?> cls = ServerBundleClassLoader.this.findLoadedClass(name);
            if (cls == null) {
                cls = this.loadedClasses.get(name);
                if (cls == null) {
                    cls = findClassInternal(name, true);
                    if (cls == null) {
                        cls = ServerBundleClassLoader.this.loadClass(name, resolve);
                    }
                }
            }
            if (cls == null) {
                throw new ClassNotFoundException(name);
            }
            if (resolve) {
                resolveClass(cls);
            }
            this.loadedClasses.putIfAbsent(name, cls);
            return cls;
        }

        /**
         * Attempts to find a <code>Class</code> from the enclosing bundle.
         * 
         * @param name the name of the <code>Class</code>
         * @param traverseDependencies should dependency bundles be checked for the class.
         */
        Class<?> findClassInternal(String name, boolean traverseDependencies) {
            String path = name.replaceAll("\\.", "/").concat(".class");

            BundleEntry entry = this.manager.findLocalEntry(path);
            if (entry == null) {
                if (traverseDependencies) {
                    return findClassFromImport(name);
                } else {
                    return null;
                }
            }
            byte[] bytes;
            try {
                bytes = entry.getBytes();
            } catch (IOException e) {
                bytes = null;
            }
            return bytes == null ? null : defineClass(name, bytes, 0, bytes.length);
        }

        /**
         * Attempts to locate a <code>Class</code> from one of the imported bundles.
         * 
         * @param name the <code>Class</code> name.
         * @return the located <code>Class</code>, or <code>null</code> if no <code>Class</code> can be found.
         */
        private Class<?> findClassFromImport(String name) {
            Bundle[] deps = getDependencyBundles(false);
            for (Bundle dep : deps) {
                if (propagateInstrumentationTo(dep)) {
                    ClassLoader depClassLoader = getBundleClassLoader(dep);
                    if (depClassLoader instanceof ServerBundleClassLoader) {
                        ServerBundleClassLoader pbcl = (ServerBundleClassLoader) depClassLoader;
                        ThrowAwayClassLoader throwAway = pbcl.createThrowAway();
                        Class<?> cls = throwAway.findClassInternal(name, false);
                        if (cls != null) {
                            return cls;
                        }
                    }
                }
            }
            return null;
        }

        /**
         * {@inheritDoc}
         */
        @Override public URL getResource(String name) {
            return ServerBundleClassLoader.this.getResource(name);
        }

        /**
         * {@inheritDoc}
         */
        @Override public Enumeration<URL> getResources(String name) throws IOException {
            return ServerBundleClassLoader.this.getResources(name);
        }
    }

    private static class IteratorEnumerationAdaptor<T> implements Enumeration<T> {

        private final Iterator<T> iterator;

        /**
         * @param iterator
         */
        public IteratorEnumerationAdaptor(Iterator<T> iterator) {
            this.iterator = iterator;
        }

        /**
         * {@inheritDoc}
         */
        public boolean hasMoreElements() {
            return this.iterator.hasNext();
        }

        /**
         * {@inheritDoc}
         */
        public T nextElement() {
            return this.iterator.next();
        }
    }
}
