/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.framework.support;

import static com.springsource.server.osgi.framework.support.LocationUtils.convertToCanonicalFilePath;

import java.io.File;
import java.util.HashSet;
import java.util.Set;

import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.BundleException;
import org.osgi.service.packageadmin.ExportedPackage;
import org.osgi.service.packageadmin.PackageAdmin;

import com.springsource.server.osgi.framework.BundleInstaller;
import com.springsource.server.osgi.framework.ImportExpander;
import com.springsource.server.osgi.framework.Library;
import com.springsource.server.osgi.framework.OsgiFrameworkException;
import com.springsource.server.osgi.framework.UnableToSatisfyDependenciesException;
import com.springsource.server.osgi.manifest.VersionRange;
import com.springsource.server.osgi.manifest.BundleManifest.ImportBundleHeader.BundleImport;
import com.springsource.server.osgi.provisioning.BundleDefinition;
import com.springsource.server.osgi.provisioning.LibraryDefinition;
import com.springsource.server.osgi.provisioning.repository.MutableBundleRepository;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * Base implementation of {@link BundleInstaller}.<p/>
 * 
 * Handles resolution of the first bundle location in the chain. Subclasses <strong>must</strong> handle installation
 * of dependencies as needed. <strong>Concurrent Semantics</strong><br/>
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @author Andy Wilkinson
 * @since 1.0
 */
public abstract class AbstractBundleInstaller implements BundleInstaller {

    private static final long SYSTEM_BUNDLE_ID = 0;

    protected final Tracer tracer = TracerFactory.createTracer(getClass());

    /**
     * The {@link MutableBundleRepository} to use to resolve bundle locations.
     */
    protected final MutableBundleRepository bundleRepository;

    /**
     * The OSGi {@link PackageAdmin} service.
     */
    protected final PackageAdmin packageAdmin;

    private final ImportExpansionHandler importExpansionHandler;

    /**
     * Creates a new <code>AbstractBundleInstaller</code>.
     * 
     * @param bundleRepository the {@link MutableBundleRepository} to use for loading bundles.
     * @param packageAdmin the OSGi {@link PackageAdmin} service.
     * @param bundleContext the system bundle context
     */
    public AbstractBundleInstaller(MutableBundleRepository bundleRepository, PackageAdmin packageAdmin, BundleContext bundleContext) {
        this.bundleRepository = bundleRepository;
        this.packageAdmin = packageAdmin;

        Set<String> packagesExportedBySystemBundle = new HashSet<String>(30);
        ExportedPackage[] exportedPackages = this.packageAdmin.getExportedPackages(bundleContext.getBundle(SYSTEM_BUNDLE_ID));
        for (ExportedPackage exportedPackage : exportedPackages) {
            packagesExportedBySystemBundle.add(exportedPackage.getName());
        }

        this.importExpansionHandler = new ImportExpansionHandler(bundleRepository, bundleContext, packagesExportedBySystemBundle);
    }

    /**
     * @throws BundleException {@inheritDoc}
     */
    public final Bundle installBundle(String bundleSymbolicName) throws UnableToSatisfyDependenciesException, BundleException {
        return installBundle(bundleSymbolicName, VersionRange.NATURAL_NUMBER_RANGE);
    }

    /**
     * @throws BundleException {@inheritDoc}
     */
    public final Bundle installBundle(String bundleSymbolicName, boolean satisfyConstraints) throws UnableToSatisfyDependenciesException,
        BundleException {
        return installBundle(bundleSymbolicName, VersionRange.NATURAL_NUMBER_RANGE, satisfyConstraints);
    }

    /**
     * @throws BundleException {@inheritDoc}
     */
    public final Bundle installBundle(String bundleSymbolicName, VersionRange versionRange) throws UnableToSatisfyDependenciesException,
        BundleException {
        return installBundle(bundleSymbolicName, versionRange, true);
    }

    /**
     * @throws BundleException {@inheritDoc}
     */
    public final Bundle installBundle(String bundleSymbolicName, VersionRange versionRange, boolean satisfyConstraints)
        throws UnableToSatisfyDependenciesException, BundleException {
        if (this.tracer.isInfoEnabled()) {
            this.tracer.traceInfo(String.format("installBundle(%s, %s)", bundleSymbolicName, versionRange));
        }
        Bundle[] bundles = this.packageAdmin.getBundles(bundleSymbolicName, versionRange.toParseString());
        if (bundles != null && bundles.length > 0) {
            Bundle bundle = bundles[0];
            if (this.tracer.isInfoEnabled()) {
                this.tracer.traceInfo(String.format("Found existing bundle %s", bundle));
            }
            return bundle;
        }
        BundleDefinition definition = findBundleDefinition(bundleSymbolicName, versionRange);
        return installBundleFromDefinition(definition, satisfyConstraints);
    }

    private BundleDefinition findBundleDefinition(String bundleSymbolicName, VersionRange versionRange) {
        BundleDefinition definition = this.bundleRepository.findBySymbolicName(bundleSymbolicName, versionRange);
        if (definition == null) {
            throw new OsgiFrameworkException("Unable to load '" + bundleSymbolicName + "' in range '" + versionRange + "'.");
        }
        return definition;
    }

    /**
     * {@inheritDoc}
     */
    public final Bundle installBundleFromLocation(File location) throws UnableToSatisfyDependenciesException, BundleException {
        Bundle bundle = installBundleFromLocation(location, true);
        return bundle;
    }

    /**
     * {@inheritDoc}
     */
    public final Bundle installBundleFromLocation(File location, boolean satisfyConstraints) throws UnableToSatisfyDependenciesException,
        BundleException {
        if (this.tracer.isInfoEnabled()) {
            this.tracer.traceInfo(String.format("installBundleFromLocation(%s, %s)", location, satisfyConstraints));
        }
        if (!location.exists()) {
            throw new OsgiFrameworkException("Unable to install bundle from location '" + location + "' - location does not exist.");
        }

        this.importExpansionHandler.handle(location);

        return doInstallBundle(convertToCanonicalFilePath(location), satisfyConstraints);
    }

    /**
     * {@inheritDoc}
     */
    public final void updateBundle(Bundle bundle) throws BundleException, UnableToSatisfyDependenciesException {
        String bundleLocation = bundle.getLocation();
        if (bundleLocation.startsWith("file:")) {
            bundleLocation = bundleLocation.substring(5);
        } else {
            throw new BundleException("Unexpected URL protocol encountered");
        }
        this.importExpansionHandler.handle(new File(bundleLocation));
        bundle.update();
    }

    /**
     * {@inheritDoc}
     */
    public final Library installLibrary(String libraryName, VersionRange versionRange) throws UnableToSatisfyDependenciesException, BundleException {
        if (this.tracer.isInfoEnabled()) {
            this.tracer.traceInfo(String.format("installLibrary(%s, %s)", libraryName, versionRange));
        }
        LibraryDefinition library = this.bundleRepository.findLibrary(libraryName, versionRange);
        if (library == null) {
            throw new OsgiFrameworkException("Unable to load library '" + libraryName + "' in range '" + versionRange + "'.");
        }
        if (this.tracer.isInfoEnabled()) {
            this.tracer.traceInfo(String.format("Found library definition %s", library));
        }

        BundleDefinition[] bundleDefinitions = lookupLibraryBundles(library);
        Bundle[] installedBundles = new Bundle[bundleDefinitions.length];
        for (int x = 0; x < bundleDefinitions.length; x++) {
            if (this.tracer.isInfoEnabled()) {
                this.tracer.traceInfo(String.format("Installing bundle %s for library %s", bundleDefinitions[x], library));
            }
            installedBundles[x] = installBundleFromDefinition(bundleDefinitions[x], true);
        }
        return new ImmutableLibrary(installedBundles);
    }

    /**
     * {@inheritDoc}
     */
    public final MutableBundleRepository getBundleRepository() {
        return this.bundleRepository;
    }

    /**
     * Subclasses <strong>must</strong> implement this method to perform actual bundle installation as per the
     * {@link BundleInstaller} contract.
     * 
     * @param resolvedBundleLocation the location of the <code>Bundle</code> to install.
     * @return the installed <code>Bundle</code>.
     * @throws BundleException if the <code>Bundle</code> cannot be installed.
     */
    protected abstract Bundle doInstallBundle(String resolvedBundleLocation, boolean validate) throws UnableToSatisfyDependenciesException,
        BundleException;

    /**
     * Installs the {@link Bundle} for the supplied {@link BundleDefinition}.
     * 
     * @param definition the <code>BundleDefinition</code>.
     * @param satisfyConstraints flag indicating whether or not constraints should be eagerly satisfied.
     * @return the installed <code>Bundle</code>.
     * @throws BundleException
     * @throws BundleException if an error occurs.
     */
    private Bundle installBundleFromDefinition(BundleDefinition definition, boolean satisfyConstraints) throws UnableToSatisfyDependenciesException,
        BundleException {
        return installBundleFromLocation(definition.getLocation(), satisfyConstraints);
    }

    /**
     * @param libraryBundles
     * @return
     */
    private BundleDefinition[] lookupLibraryBundles(LibraryDefinition library) {
        BundleImport[] bundleImports = library.getLibraryBundles();
        BundleDefinition[] definitions = new BundleDefinition[bundleImports.length];
        for (int x = 0; x < bundleImports.length; x++) {
            BundleImport libraryBundle = bundleImports[x];
            String symbolicName = libraryBundle.getBundleSymbolicName();
            VersionRange versionRange = new VersionRange(libraryBundle.getVersionRange());
            BundleDefinition definition = this.bundleRepository.findBySymbolicName(symbolicName, versionRange);
            if (definition == null) {
                throw new OsgiFrameworkException("Error installing library '" + library.getSymbolicName() + "'. Cannot find library bundle '"
                    + symbolicName + "' with version within range '" + versionRange + "'.");
            }
            definitions[x] = definition;
        }
        return definitions;
    }

    /**
     * {@inheritDoc}
     */
    public ImportExpander getImportExpander() {
        return this.importExpansionHandler;
    }
}
