/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.framework.support;

import java.io.File;
import java.util.Dictionary;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeUnit;

import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.InvalidSyntaxException;
import org.osgi.framework.ServiceEvent;
import org.osgi.framework.ServiceListener;
import org.osgi.framework.ServiceReference;
import org.osgi.framework.ServiceRegistration;
import org.osgi.service.packageadmin.PackageAdmin;
import org.springframework.util.StringUtils;

import com.springsource.server.osgi.framework.BundleInstaller;
import com.springsource.server.osgi.framework.OsgiConfiguration;
import com.springsource.server.osgi.framework.OsgiFramework;
import com.springsource.server.osgi.framework.OsgiFrameworkUtils;
import com.springsource.server.osgi.provisioning.repository.BundleRepository;
import com.springsource.server.osgi.provisioning.repository.BundleRepositoryFactory;
import com.springsource.server.osgi.provisioning.repository.MutableBundleRepository;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * Base implementation of {@link BundleInstaller} and {@link OsgiFramework}.<p/>
 * 
 * Core start and stop logic is handled by this base implementation. In particular, the requirement to self publish as
 * services in the service registration under <code>BundleInstaller</code> and <code>OsgiFramework</code> is
 * handled. Implementations are required to implement the two lifecycle callback methods {@link #doStart()} and
 * {@link #doStop()}. The contract of these methods <strong>must</strong> be implemented exactly as stated.<p/>
 * 
 * <code>Bundle</code> location is handled in conjunction with a {@link BundleRepository}. Subclasses are required to
 * implement {@link #createBundleInstaller()} to allow them the provide implementation-specific bundle installation.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * Implementation is thread safe.
 * 
 * @author Rob Harrop
 * @author Glyn Normington
 * @author Andy Wilkinson
 * @author Sam Brannen
 * @since 1.0
 */
public abstract class AbstractOsgiFramework implements OsgiFramework {

    private static final String DIRECTIVE_PUBLISH_CONTEXT = "publish-context";

    private static final String DIRECTIVE_EQUALS = ":=";

    public static final String DIRECTIVE_SEPARATOR = ";";

    public static final boolean DIRECTIVE_PUBLISH_CONTEXT_DEFAULT = true;

    protected final Tracer tracer = TracerFactory.createTracer(getClass());

    /**
     * The internally-managed <code<BundleRepository</code>.
     */
    private final MutableBundleRepository bundleRepository;

    /**
     * Framework monitor used for lifecycle events.
     */
    private final Object frameworkMonitor = new Object();

    private ServiceRegistration selfServiceRegistration;

    private ServiceRegistration bundleInstallerServiceRegistration;

    private ServiceReference packageAdminServiceReference;

    private volatile PackageAdmin packageAdmin;

    private volatile BundleInstaller bundleInstaller;

    protected AbstractOsgiFramework(OsgiConfiguration configuration) {
        String workDirectory = configuration.getWorkDirectory();
        String locationCachePath = workDirectory + File.separatorChar + "locationCache" + File.separatorChar;
        new File(locationCachePath).mkdirs();
        this.bundleRepository = BundleRepositoryFactory.createMutableBundleRepository(configuration.getInstallPath(), configuration.getSearchPaths(),
            locationCachePath, configuration.getAbsoluteRoot());
    }

    /**
     * Handles registration of the required OSGi services post-startup.<p/>
     * 
     * Delegates to the {@link #doStart} for actual starting of the framework implementation.
     */
    public final void start() {
        this.tracer.traceInfo("start()");
        synchronized (this.frameworkMonitor) {
            doStart();

            BundleContext bundleContext = getBundleContext();
            this.selfServiceRegistration = bundleContext.registerService(OsgiFramework.class.getName(), this, null);
            this.tracer.traceInfo("Registered OsgiFramework service.");

            this.packageAdminServiceReference = bundleContext.getServiceReference(PackageAdmin.class.getName());
            this.packageAdmin = (PackageAdmin) bundleContext.getService(this.packageAdminServiceReference);

            this.bundleInstaller = createBundleInstaller();
            this.bundleInstallerServiceRegistration = bundleContext.registerService(BundleInstaller.class.getName(), this.bundleInstaller, null);
            this.tracer.traceInfo("Registered BundleInstaller service.");
        }
    }

    /**
     * {@inheritDoc}<p/>
     * 
     * Handles de-registration of services and relinquishes necessary {@link ServiceReference ServiceReferences}.<p/>
     * 
     * Delegates to {@link #doStop()} to allow subclasses to shutdown the corresponding OSGi framework instance.
     */
    public final void stop() {
        this.tracer.traceInfo("stop()");
        synchronized (this.frameworkMonitor) {
            if (this.selfServiceRegistration != null) {
                this.selfServiceRegistration.unregister();
                this.selfServiceRegistration = null;
            }

            if (this.bundleInstallerServiceRegistration != null) {
                this.bundleInstallerServiceRegistration.unregister();
                this.bundleInstallerServiceRegistration = null;
            }

            if (this.packageAdminServiceReference != null) {
                getBundleContext().ungetService(this.packageAdminServiceReference);
                this.packageAdmin = null;
            }
            doStop();
        }
    }

    /**
     * {@inheritDoc}
     */
    public Bundle getClassBundle(Class<?> cls) {
        if (this.packageAdmin != null) {
            return this.packageAdmin.getBundle(cls);
        } else {
            return null;
        }
    }

    /**
     * {@inheritDoc}
     */
    public final BundleInstaller getBundleInstaller() {
        return this.bundleInstaller;
    }

    /**
     * Starts the OSGi framework instance.
     */
    protected abstract void doStart();

    /**
     * Stops the OSGi framework instance.
     */
    protected abstract void doStop();

    /**
     * Allows subclasses to create an implementation-specific {@link BundleInstaller}.
     * 
     * @return the newly created <code>BundleInstaller</code>.
     */
    protected abstract BundleInstaller createBundleInstaller();

    /**
     * Gets the {@link MutableBundleRepository} to use for bundle location resolution.
     * 
     * @return the <code>BundleRepository</code>.
     */
    protected final MutableBundleRepository getBundleRepository() {
        return this.bundleRepository;
    }

    /**
     * Gets the {@link PackageAdmin} service.
     * 
     * @return the <code>PackageAdmin</code> service.
     */
    protected final PackageAdmin getPackageAdmin() {
        return this.packageAdmin;
    }

    /**
     * {@inheritDoc}
     */
    public boolean bundlePublishesApplicationContext(Bundle bundle) {
        return OsgiFrameworkUtils.isSpringOsgiPoweredBundle(bundle) && getPublishContext(bundle.getHeaders());
    }

    /**
     * {@inheritDoc}
     */
    public boolean waitForBundleApplicationContext(Bundle bundle, int timeout) {
        if (this.tracer.isInfoEnabled()) {
            this.tracer.traceInfo(String.format("waitForBundleApplicationContext(%s, %s)", bundle, timeout));
        }
        if (bundlePublishesApplicationContext(bundle)) {
            return waitForBundleApplicationContext(bundle.getSymbolicName(), timeout);
        } else {
            return true;
        }
    }

    @SuppressWarnings("unchecked") private static boolean getPublishContext(Dictionary headers) {
        String value = getDirectiveValue(headers.toString(), DIRECTIVE_PUBLISH_CONTEXT);
        return value != null ? Boolean.valueOf(value).booleanValue() : DIRECTIVE_PUBLISH_CONTEXT_DEFAULT;
    }

    private static String getDirectiveValue(String header, String directive) {
        String[] directives = StringUtils.tokenizeToStringArray(header, DIRECTIVE_SEPARATOR);

        for (int i = 0; i < directives.length; i++) {
            String[] splittedDirective = StringUtils.delimitedListToStringArray(directives[i].trim(), DIRECTIVE_EQUALS);
            if (splittedDirective.length == 2 && splittedDirective[0].equals(directive)) {
                return splittedDirective[1];
            }
        }
        return null;
    }

    /**
     * {@inheritDoc}
     */
    public final boolean waitForBundleApplicationContext(String symbolicName, int timeout) {
        if (this.tracer.isInfoEnabled()) {
            this.tracer.traceInfo(String.format("waitForBundleApplicationContext(%s, %s)", symbolicName, timeout));
        }
        String filter = "(org.springframework.context.service.name=" + symbolicName + ")";

        final CountDownLatch latch = new CountDownLatch(1);
        ServiceListener listener = new ServiceListener() {

            public void serviceChanged(ServiceEvent event) {
                if (event.getType() == ServiceEvent.REGISTERED) {
                    latch.countDown();
                }
            }
        };

        BundleContext bundleContext = getBundleContext();
        try {
            bundleContext.addServiceListener(listener, filter);
        } catch (InvalidSyntaxException isex) {
            throw (RuntimeException) new IllegalArgumentException("invalid filter").initCause(isex);
        }

        ServiceReference[] alreadyRegistered = getServiceReferences(bundleContext, filter);

        for (int i = 0; i < alreadyRegistered.length; i++) {
            listener.serviceChanged(new ServiceEvent(ServiceEvent.REGISTERED, alreadyRegistered[i]));
        }

        boolean result;
        try {
            result = latch.await(timeout, TimeUnit.SECONDS);
        } catch (InterruptedException e) {
            result = false;
        }
        if (this.tracer.isInfoEnabled()) {
            this.tracer.traceInfo(String.format("waitForBundleApplicationContext(%s, %s) - returned %s", symbolicName, timeout, result));
        }
        return result;
    }

    private static ServiceReference[] getServiceReferences(BundleContext bundleContext, String filter) {
        try {
            ServiceReference[] refs = bundleContext.getServiceReferences(null, filter);
            return refs == null ? new ServiceReference[0] : refs;
        } catch (InvalidSyntaxException ise) {
            throw (RuntimeException) new IllegalArgumentException("invalid filter: " + ise.getFilter()).initCause(ise);
        }
    }
}
