/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.framework.support;

import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Set;
import java.util.jar.JarFile;

import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.BundleException;
import org.osgi.framework.Version;
import org.springframework.util.Assert;

import com.springsource.server.osgi.framework.ImportExpander;
import com.springsource.server.osgi.framework.ImportMergeException;
import com.springsource.server.osgi.framework.OsgiFrameworkException;
import com.springsource.server.osgi.framework.OsgiFrameworkLogEvent;
import com.springsource.server.osgi.framework.OsgiFrameworkUtils;
import com.springsource.server.osgi.framework.UnableToSatisfyBundleDependenciesException;
import com.springsource.server.osgi.framework.UnableToSatisfyDependenciesException;
import com.springsource.server.osgi.framework.support.merge.StandardTrackedPackageImportsFactory;
import com.springsource.server.osgi.framework.support.merge.TrackedPackageImports;
import com.springsource.server.osgi.framework.support.merge.TrackedPackageImportsFactory;
import com.springsource.server.osgi.manifest.BundleManifest;
import com.springsource.server.osgi.manifest.BundleManifestFactory;
import com.springsource.server.osgi.manifest.BundleManifestProcessor;
import com.springsource.server.osgi.manifest.Scoper;
import com.springsource.server.osgi.manifest.StandardBundleManifestFactory;
import com.springsource.server.osgi.manifest.VersionRange;
import com.springsource.server.osgi.manifest.BundleManifest.BundleSymbolicName;
import com.springsource.server.osgi.manifest.BundleManifest.ExportPackageHeader;
import com.springsource.server.osgi.manifest.BundleManifest.FragmentHostHeader;
import com.springsource.server.osgi.manifest.BundleManifest.ImportBundleHeader;
import com.springsource.server.osgi.manifest.BundleManifest.ImportLibraryHeader;
import com.springsource.server.osgi.manifest.BundleManifest.ImportPackageHeader;
import com.springsource.server.osgi.manifest.BundleManifest.ImportBundleHeader.BundleImport;
import com.springsource.server.osgi.manifest.BundleManifest.ImportLibraryHeader.LibraryImport;
import com.springsource.server.osgi.manifest.BundleManifest.ImportPackageHeader.PackageImport;
import com.springsource.server.osgi.manifest.parse.OSGiIntegrationLogService;
import com.springsource.server.osgi.provisioning.BundleDefinition;
import com.springsource.server.osgi.provisioning.LibraryDefinition;
import com.springsource.server.osgi.provisioning.repository.BundleRepository;
import com.springsource.server.serviceability.logging.LogService;
import com.springsource.server.serviceability.logging.LogServiceFactory;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * A helper class for handling the expansion of <code>Import-Library</code> and <code>Import-Bundle</code> headers
 * in a bundle manifest into <code>Import-Package</code> header entries.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * This class is <strong>thread-safe</strong>.
 * 
 * @author Andy Wilkinson
 * @author Christopher Frost
 * @author Glyn Normington
 * @since 1.0
 */
public final class ImportExpansionHandler implements ImportExpander {

    private static final LogService LOGGER = LogServiceFactory.getLogService();

    private static final String MANIFEST_LOCATION = "META-INF/MANIFEST.MF";

    private static final String MISSING_BUNDLE_SYMBOLIC_NAME = "<bundle symbolic name not present>";

    private static final Tracer TRACER = TracerFactory.createTracer(ImportExpansionHandler.class);

    private final BundleRepository bundleRepository;

    private final BundleManifestFactory bundleManifestFactory;

    private final BundleManifestProcessor manifestProcessor;

    private final TrackedPackageImportsFactory trackedPackageImportsFactory = new StandardTrackedPackageImportsFactory();

    private final BundleContext bundleContext;

    private final Set<String> packagesExportedBySystemBundle;

    public ImportExpansionHandler(BundleRepository bundleRepository, Set<String> packagesExportedBySystemBundle) {
        this(bundleRepository, null, packagesExportedBySystemBundle);
    }

    public ImportExpansionHandler(BundleRepository bundleRepository, BundleContext bundleContext, Set<String> packagesExportedBySystemBundle) {
        this.bundleRepository = bundleRepository;
        this.bundleManifestFactory = new StandardBundleManifestFactory(new OSGiIntegrationLogService());
        this.manifestProcessor = new BundleManifestProcessor(this.bundleManifestFactory);
        this.bundleContext = bundleContext;
        this.packagesExportedBySystemBundle = packagesExportedBySystemBundle;
    }

    /**
     * {@inheritDoc}
     * 
     * @throws ImportMergeException
     */
    public ImportPromotionVector expandImports(List<BundleManifest> bundleManifests) throws UnableToSatisfyDependenciesException,
        ImportMergeException {
        StandardImportPromotionVector importPromotionVector = new StandardImportPromotionVector(this.trackedPackageImportsFactory);
        TrackedPackageImports packageImportsToBePromoted = this.trackedPackageImportsFactory.createCollector();
        BundleRepository combinedRepository = new AdditionalManifestSearchingBundleRepository(bundleManifests, this.bundleRepository);
        for (BundleManifest bundleManifest : bundleManifests) {
            TrackedPackageImports bundlePackageImportsToBePromoted = this.trackedPackageImportsFactory.createCollector();
            bundleManifest.expandPackageLists();
            expandImportsIfNecessary(bundleManifest, combinedRepository, bundlePackageImportsToBePromoted, null);
            if (!bundlePackageImportsToBePromoted.isEmpty()) {
                packageImportsToBePromoted.merge(bundlePackageImportsToBePromoted);
                BundleSymbolicName bundleSymbolicNameHeader = bundleManifest.getBundleSymbolicName();
                Assert.notNull(bundleSymbolicNameHeader, "Bundle-SymbolicName must be present for import promotion tracking");
                importPromotionVector.put(bundleSymbolicNameHeader.getSymbolicName(), bundlePackageImportsToBePromoted);
            }
        }
        mergeImports(packageImportsToBePromoted, bundleManifests);
        return importPromotionVector;
    }

    /**
     * Merge the given {@list TrackPackageImports} into all the manifests in the given list.
     * 
     * @param importsToBemerged
     * @param bundleManifests
     * @throws ImportMergeException
     */
    private void mergeImports(TrackedPackageImports importsToBemerged, List<BundleManifest> bundleManifests) throws ImportMergeException {
        for (BundleManifest bundleManifest : bundleManifests) {
            mergeImports(importsToBemerged, bundleManifest);
        }
    }

    /**
     * @param importsToBemerged
     * @param bundleManifest
     */
    private void mergeImports(TrackedPackageImports importsToBemerged, BundleManifest bundleManifest) throws ImportMergeException {
        TrackedPackageImports bundleTrackedPackageImports = this.trackedPackageImportsFactory.create(bundleManifest);
        bundleTrackedPackageImports.merge(importsToBemerged);
        setMergedImports(bundleManifest, bundleTrackedPackageImports);
    }

    void handle(File bundleLocation) throws UnableToSatisfyDependenciesException, BundleException {
        if (bundleLocation.isDirectory()) {
            try {
                File manifestFile = new File(bundleLocation.getAbsolutePath() + File.separator + JarFile.MANIFEST_NAME);
                FileReader manifestReader = new FileReader(manifestFile);
                BundleManifest manifest = this.bundleManifestFactory.createBundleManifest(manifestReader);
                manifestReader.close();
                if (expandImportsIfNecessary(manifest, this.bundleRepository, this.trackedPackageImportsFactory.createCollector(), null)) {
                    manifestFile.delete();
                    FileWriter manifestWriter = new FileWriter(manifestFile);
                    manifest.write(manifestWriter);
                    manifestWriter.close();
                }
            } catch (IOException ioe) {
                throw new OsgiFrameworkException(ioe);
            }
        }
    }

    private boolean expandImportsIfNecessary(BundleManifest manifest, BundleRepository bundleRepository,
        TrackedPackageImports packageImportsToBePromoted, Scoper scoper) throws UnableToSatisfyDependenciesException {
        boolean expanded = false;
        ImportLibraryHeader importLibraryHeader = manifest.getImportLibraryHeader();
        ImportBundleHeader importBundleHeader = manifest.getImportBundleHeader();

        if (importLibraryHeader != null || importBundleHeader != null) {
            if (TRACER.isInfoEnabled()) {
                TRACER.traceInfo("Import-Library and/or Import-Bundle header found. Original manifest: %n%s", manifest);
            }

            List<BundleImport> directlyImportedBundles = importBundleHeader != null ? importBundleHeader.getImportedBundles()
                : new ArrayList<BundleImport>(0);

            List<LibraryImport> importedLibraries = importLibraryHeader != null ? importLibraryHeader.getImportedLibraries()
                : new ArrayList<LibraryImport>(0);

            expandImports(importedLibraries, directlyImportedBundles, manifest, bundleRepository, packageImportsToBePromoted, scoper);

            if (TRACER.isInfoEnabled()) {
                TRACER.traceInfo("Updated manifest: %n%s", manifest);
            }
            expanded = true;
        }
        return expanded;
    }

    UnableToSatisfyBundleDependenciesException createExceptionForMissingLibrary(String name, VersionRange versionRange, BundleManifest bundleManifest) {
        String description = String.format("A library with the name '%s' and a version within the range '%s' could not be found", name, versionRange);
        BundleSymbolicName bundleSymbolicName = bundleManifest.getBundleSymbolicName();
        return new UnableToSatisfyBundleDependenciesException(bundleSymbolicName != null ? bundleSymbolicName.getSymbolicName()
            : MISSING_BUNDLE_SYMBOLIC_NAME, new Version(bundleManifest.getBundleVersion()), description);
    }

    void expandImports(List<LibraryImport> libraryImports, List<BundleImport> directBundleImports, BundleManifest bundleManifest)
        throws UnableToSatisfyDependenciesException, BundleException {
        expandImports(libraryImports, directBundleImports, bundleManifest, this.bundleRepository,
            this.trackedPackageImportsFactory.createCollector(), null);
    }

    private void expandImports(List<LibraryImport> libraryImports, List<BundleImport> directBundleImports, BundleManifest bundleManifest,
        BundleRepository bundleRepository, TrackedPackageImports packageImportsToBePromoted, Scoper scoper)
        throws UnableToSatisfyDependenciesException {
        Assert.notNull(libraryImports);
        Assert.notNull(directBundleImports);

        mergeImports(getAdditionalPackageImports(libraryImports, directBundleImports, bundleManifest, bundleRepository, packageImportsToBePromoted,
            scoper), bundleManifest);

        libraryImports.clear();
        directBundleImports.clear();
    }

    /**
     * The bundle with the given bundle manifest imports the given list of libraries and the given list of bundles.
     * Dereference these imports and return a collection of the corresponding {@link TrackedPackageImports}. Any
     * promoted imports are added to the given <code>TrackedPackageImports</code> of imports to be promoted.
     * 
     * @param libraryImports
     * @param directBundleImports
     * @param bundleManifest
     * @param bundleRepository
     * @param packageImportsToBePromoted
     * @return
     * @throws UnableToSatisfyDependenciesException
     * @throws UnableToSatisfyBundleDependenciesException
     */
    private TrackedPackageImports getAdditionalPackageImports(List<LibraryImport> libraryImports, List<BundleImport> directBundleImports,
        BundleManifest bundleManifest, BundleRepository bundleRepository, TrackedPackageImports packageImportsToBePromoted, Scoper scoper)
        throws UnableToSatisfyDependenciesException, UnableToSatisfyBundleDependenciesException {
        TrackedPackageImports additionalPackageImports = this.trackedPackageImportsFactory.createCollector();
        TrackedPackageImports libraryPackageImports = getLibraryPackageImports(libraryImports, bundleRepository, packageImportsToBePromoted,
            bundleManifest, scoper);
        additionalPackageImports.merge(libraryPackageImports);
        for (BundleImport directBundleImport : directBundleImports) {
            additionalPackageImports.merge(getBundlePackageImports(directBundleImport, bundleRepository, packageImportsToBePromoted, bundleManifest,
                scoper));
        }
        return additionalPackageImports;
    }

    /**
     * Get a {@link TrackedPackageImports} instance representing the package imports that correspond to the given bundle
     * import. The imported bundle is looked up in the given {@link BundleRepository} and, if it is not found,
     * {@link UnableToSatisfiedBundleDependenciesException} is thrown. If the bundle import is to be promoted, then the
     * result is also merged into the given <code>TrackedPackageImports</code> instance representing the package
     * imports to be promoted.
     * 
     * @param bundleImport
     * @param bundleRepository
     * @param packageImportsToBePromoted
     * @param importingBundle
     * @return
     * @throws UnableToSatisfyBundleDependenciesException
     */
    private TrackedPackageImports getBundlePackageImports(BundleImport bundleImport, BundleRepository bundleRepository,
        TrackedPackageImports packageImportsToBePromoted, BundleManifest importingBundle, Scoper scoper)
        throws UnableToSatisfyBundleDependenciesException {
        String bundleSymbolicName = bundleImport.getBundleSymbolicName();
        String versionRange = bundleImport.getVersionRange();
        VersionRange importVersionRange = new VersionRange(versionRange);
        boolean mandatory = bundleImport.getResolution().equals(BundleImport.Resolution.MANDATORY);
        if (importingBundle.getBundleSymbolicName() != null && bundleSymbolicName.equals(importingBundle.getBundleSymbolicName().getSymbolicName())
            && importVersionRange.includes(new Version(importingBundle.getBundleVersion()))) {
            throw new UnableToSatisfyBundleDependenciesException(importingBundle.getBundleSymbolicName().getSymbolicName(), new Version(
                importingBundle.getBundleVersion()), "Import-Bundle must not import the importing bundle");
        }

        BundleDefinition bundleDefinition = findBundleDefinition(bundleSymbolicName, importVersionRange, bundleRepository, scoper);
        if (bundleDefinition != null) {
            return createTrackedPackageImportsFromImportedBundle(bundleDefinition, bundleImport.getApplicationImportScope(),
                packageImportsToBePromoted);
        } else if (mandatory) {
            throw new UnableToSatisfyBundleDependenciesException(
                importingBundle.getBundleSymbolicName() != null ? importingBundle.getBundleSymbolicName().getSymbolicName()
                    : MISSING_BUNDLE_SYMBOLIC_NAME, new Version(importingBundle.getBundleVersion()), "Import-Bundle with symbolic name '"
                    + bundleSymbolicName + "' in version range '" + versionRange + "' could not be satisfied");
        } else {
            return this.trackedPackageImportsFactory.createEmpty();
        }
    }

    /**
     * Return the {@link TrackedPackageImports} corresponding to importing the given {@link BundleDefinition} and, if
     * appropriate, merge them into the given imports to be promoted.
     * 
     * @param bundleDefinition
     * @param promoteExports
     * @param packageImportsToBePromoted
     * @return
     */
    private TrackedPackageImports createTrackedPackageImportsFromImportedBundle(BundleDefinition bundleDefinition, boolean promoteExports,
        TrackedPackageImports packageImportsToBePromoted) {
        BundleManifest bundleManifest = bundleDefinition.getManifest();
        TrackedPackageImports bundlePackageImports = createPackageImportsFromPackageExports(bundleDefinition, bundleManifest,
            bundleManifest.getExportPackageHeader());
        if (promoteExports) {
            packageImportsToBePromoted.merge(bundlePackageImports);
        }
        return bundlePackageImports;
    }

    /**
     * Create a {@link TrackedPackageImports} instance corresponding to the given package exports from the given bundle
     * manifest.
     * <p />
     * Importing fragment bundles is not encouraged but sometimes it's the only way out when reusing poorly packaged
     * bundles. Packages exported by a fragment are imported with the host's bundle symbolic name, an exact range
     * matching only the exported package version, and a bundle version range corresponding to that of the fragment host
     * header. Since the generated imports do not include a bundle version matching attribute with a value of the form
     * [v,v], there is a risk that the package import will wire to the wrong version of the host. This risk seems very
     * small and not worth the extra complexity of matching up hosts and fragments (some of which would need searching
     * for in the repository) during import expansion.
     * 
     * @param bundleDefinition
     * @param bundleManifest
     * @param exportPackageHeader
     * @return
     */
    private TrackedPackageImports createPackageImportsFromPackageExports(BundleDefinition bundleDefinition, BundleManifest bundleManifest,
        ExportPackageHeader exportPackageHeader) {
        if (exportPackageHeader == null) {
            return this.trackedPackageImportsFactory.createEmpty();
        }
        String bundleVersion = bundleManifest.getBundleVersion();
        BundleSymbolicName bundleSymbolicNameHeader = bundleManifest.getBundleSymbolicName();
        String bundleSymbolicNameString = bundleSymbolicNameHeader != null ? bundleSymbolicNameHeader.getSymbolicName() : null;
        FragmentHostHeader fragmentHostHeader = bundleManifest.getFragmentHostHeader();
        if (fragmentHostHeader != null) {
            bundleSymbolicNameString = fragmentHostHeader.getBundleSymbolicName();
            bundleVersion = fragmentHostHeader.getAttributes().get("bundle-version");
        }

        PackageImport[] packageImports = this.manifestProcessor.createPackageImportForEachPackageExport(exportPackageHeader.getPackageExports(),
            bundleSymbolicNameString, bundleVersion);

        diagnoseSystemBundleOverlap(packageImports, bundleSymbolicNameString, bundleVersion);

        return this.trackedPackageImportsFactory.create(Arrays.asList(packageImports), "Import-Bundle '"
            + bundleDefinition.getManifest().getBundleSymbolicName().getSymbolicName() + "' version '"
            + bundleDefinition.getManifest().getBundleVersion() + "'");
    }

    /**
     * Check whether the packages imported by importing a bundle are exported by the system bundle and, if they are,
     * issue a warning.
     * 
     * @param packageImports the imported packages
     * @param bundleSymbolicNameString the symbolic name of the imported bundle
     * @param bundleVersion the version of the imported bundle
     */
    private void diagnoseSystemBundleOverlap(PackageImport[] packageImports, String bundleSymbolicNameString, String bundleVersion) {
        boolean overlap = false;
        import_loop: for (PackageImport packageImport : packageImports) {
            for (String pkg : packageImport.getPackageNames()) {
                if (this.packagesExportedBySystemBundle.contains(pkg)) {
                    overlap = true;
                    break import_loop;
                }
            }
        }
        if (overlap) {
            StringBuffer imports = new StringBuffer();
            boolean first = true;
            for (PackageImport packageImport : packageImports) {
                for (String pkg : packageImport.getPackageNames()) {
                    if (!first) {
                        imports.append(",");
                    }
                    first = false;
                    imports.append(pkg);
                }
            }
            LOGGER.log(OsgiFrameworkLogEvent.SYSTEM_BUNDLE_OVERLAP, bundleSymbolicNameString, bundleVersion, imports.toString());
        }
    }

    /**
     * Get a {@link TrackedPackageImports} instance representing the package imports that correspond to the given
     * library imports. Each imported library is looked up in the given {@link BundleRepository} and, if it is not
     * found, {@link UnableToSatisfiedBundleDependenciesException} is thrown. If any package imports are to be promoted,
     * then the result is also merged into the given <code>TrackedPackageImports</code> instance representing the
     * package imports to be promoted.
     * 
     * @param libraryImports
     * @param bundleRepository
     * @param packageImportsToBePromoted
     * @param importingBundle
     * @return
     * @throws UnableToSatisfyBundleDependenciesException
     */
    private TrackedPackageImports getLibraryPackageImports(List<LibraryImport> libraryImports, BundleRepository bundleRepository,
        TrackedPackageImports packageImportsToBePromoted, BundleManifest importingBundle, Scoper scoper)
        throws UnableToSatisfyBundleDependenciesException {
        TrackedPackageImports allLibraryPackageImports = this.trackedPackageImportsFactory.createCollector();
        for (LibraryImport libraryImport : libraryImports) {
            VersionRange libraryVersionRange = new VersionRange(libraryImport.getVersionRange());
            String libraryName = libraryImport.getLibraryName();

            LibraryDefinition definition = findLibraryDefinition(libraryName, libraryVersionRange, bundleRepository);

            if (definition != null) {
                Version libraryVersion = definition.getVersion();
                TrackedPackageImports libraryPackageImports = this.trackedPackageImportsFactory.createContainer("Import-Library '"
                    + libraryImport.getLibraryName() + "' version '" + libraryVersion + "'");
                for (BundleImport bundleImport : definition.getLibraryBundles()) {

                    String bundleSymbolicName = bundleImport.getBundleSymbolicName();
                    String bundleVersionRange = bundleImport.getVersionRange();
                    BundleDefinition bundleDefinition = findBundleDefinition(bundleSymbolicName, new VersionRange(bundleVersionRange),
                        bundleRepository, scoper);
                    if (bundleDefinition != null) {
                        libraryPackageImports.merge(createTrackedPackageImportsFromImportedBundle(bundleDefinition,
                            bundleImport.getApplicationImportScope(), packageImportsToBePromoted));
                    } else if (bundleImport.getResolution().equals(BundleImport.Resolution.MANDATORY)) {
                        throw new UnableToSatisfyBundleDependenciesException(
                            importingBundle.getBundleSymbolicName() != null ? importingBundle.getBundleSymbolicName().getSymbolicName()
                                : MISSING_BUNDLE_SYMBOLIC_NAME, new Version(importingBundle.getBundleVersion()), "Imported library '" + libraryName
                                + "' version '" + libraryVersion + "' contains Import-Bundle for bundle '" + bundleSymbolicName
                                + "' in version range '" + bundleVersionRange + "' which could not be satisfied");
                    }

                }
                allLibraryPackageImports.merge(libraryPackageImports);
            } else if (libraryImport.getResolution().equals(LibraryImport.Resolution.MANDATORY)) {
                throw createExceptionForMissingLibrary(libraryName, libraryVersionRange, importingBundle);
            }
        }
        return allLibraryPackageImports;
    }

    /**
     * Set the package imports of the given {@link BundleManifest} to the merged imports of the given
     * {@link TrackedPackageImports}.
     * 
     * @param bundleManifest the bundle manifest to be modified
     * @param bundleTrackedPackageImports the <code>TrackedPackageImports</code> containing the merged imports
     */
    private void setMergedImports(BundleManifest bundleManifest, TrackedPackageImports bundleTrackedPackageImports) {
        ImportPackageHeader importPackageHeader = bundleManifest.getImportPackageHeader();
        if (importPackageHeader == null) {
            importPackageHeader = this.bundleManifestFactory.createImportPackageHeader(new ArrayList<PackageImport>());
            bundleManifest.setImportPackageHeader(importPackageHeader);
        }
        importPackageHeader.setPackageImports(bundleTrackedPackageImports.getMergedImports());
    }

    private LibraryDefinition findLibraryDefinition(String libraryName, VersionRange versionRange, BundleRepository bundleRepository) {
        return bundleRepository.findLibrary(libraryName, versionRange);
    }

    /**
     * Find the bundle with the given symbolic name and version range in the given bundle repository. If the given
     * scoper is non-null, scope the symbolic name before searching the repository. Return the {@link BundleDefinition}
     * of the bundle if it was found.
     * 
     * @param bundleSymbolicName
     * @param versionRange
     * @param bundleRepository
     * @param scoper
     * @return
     */
    private BundleDefinition findBundleDefinition(String bundleSymbolicName, VersionRange versionRange, BundleRepository bundleRepository,
        Scoper scoper) {
        BundleDefinition bundleDefinition = null;
        if (this.bundleContext != null) {
            Bundle[] installedBundles = this.bundleContext.getBundles();
            for (Bundle bundle : installedBundles) {
                if (bundleSymbolicName.equals(bundle.getSymbolicName()) && versionRange.includes(OsgiFrameworkUtils.getBundleVersion(bundle))) {
                    bundleDefinition = getBundleDefinition(bundle);
                    if (bundleDefinition != null) {
                        break;
                    }
                }
            }
        }

        if (bundleDefinition == null) {
            bundleDefinition = bundleRepository.findBySymbolicName(scoper == null ? bundleSymbolicName
                : scoper.getScopedSymbolicName(bundleSymbolicName), versionRange);
        }

        if (TRACER.isInfoEnabled()) {
            if (bundleDefinition != null) {
                TRACER.traceInfo("Found definition for bundle with symbolic name '%s' and version range '%s': %s", bundleSymbolicName, versionRange,
                    bundleDefinition.getManifest());
            } else {
                TRACER.traceInfo("Could not find definition for bundle with symbolic name '%s' and version range '%s'", bundleSymbolicName,
                    versionRange);
            }
        }

        return bundleDefinition;
    }

    /**
     * Get a {@link BundleDefinition} for the given {@link Bundle}. If a definition cannot be created, return
     * <code>null</code>.
     * 
     * @param bundle the bundle whose definition is required
     * @return the bundle definition or <code>null</code> if no definition can be created
     */
    private BundleDefinition getBundleDefinition(Bundle bundle) {
        URL manifestUrl = bundle.getResource(MANIFEST_LOCATION);
        if (manifestUrl != null) {
            try {
                InputStream manifestStream = manifestUrl.openStream();
                final BundleManifest bundleManifest = new StandardBundleManifestFactory(new OSGiIntegrationLogService()).createBundleManifest(new InputStreamReader(
                    manifestStream));
                manifestStream.close();

                return new BundleDefinition() {

                    public BundleManifest getManifest() {
                        return bundleManifest;
                    }

                    public File getLocation() {
                        throw new UnsupportedOperationException();
                    }

                };
            } catch (IOException e) {
            }
        }
        return null;
    }

    private class AdditionalManifestSearchingBundleRepository implements BundleRepository {

        private final Collection<BundleManifest> manifests;

        private final BundleRepository bundleRepository;

        AdditionalManifestSearchingBundleRepository(Collection<BundleManifest> manifests, BundleRepository bundleRepository) {
            this.manifests = manifests;
            this.bundleRepository = bundleRepository;
        }

        /**
         * {@inheritDoc}
         */
        public Set<? extends BundleDefinition> findByExportedPackage(String packageName, VersionRange versionRange) {
            return this.bundleRepository.findByExportedPackage(packageName, versionRange);
        }

        /**
         * {@inheritDoc}
         */
        public Set<? extends BundleDefinition> findByFragmentHost(String bundleSymbolicName, Version version) {
            return this.bundleRepository.findByFragmentHost(bundleSymbolicName, version);
        }

        /**
         * {@inheritDoc}
         */
        public BundleDefinition findBySymbolicName(String symbolicName, VersionRange versionRange) {
            BundleDefinition definition = this.bundleRepository.findBySymbolicName(symbolicName, versionRange);
            if (definition == null) {
                for (final BundleManifest manifest : this.manifests) {
                    BundleSymbolicName bsn = manifest.getBundleSymbolicName();
                    if (bsn != null && symbolicName.equals(bsn.getSymbolicName())) {
                        Version version = Version.parseVersion(manifest.getBundleVersion());
                        if (versionRange.includes(version)) {
                            definition = new BundleDefinition() {

                                public BundleManifest getManifest() {
                                    return manifest;
                                }

                                public File getLocation() {
                                    throw new UnsupportedOperationException();
                                }
                            };
                        }
                    }
                }
            }
            return definition;
        }

        /**
         * {@inheritDoc}
         */
        public LibraryDefinition findLibrary(String libraryName, VersionRange versionRange) {
            return this.bundleRepository.findLibrary(libraryName, versionRange);
        }

        /**
         * {@inheritDoc}
         */
        public Set<? extends BundleDefinition> getBundles() {
            return this.bundleRepository.getBundles();
        }

        /**
         * {@inheritDoc}
         */
        public Set<? extends LibraryDefinition> getLibraries() {
            return this.bundleRepository.getLibraries();
        }

        /**
         * {@inheritDoc}
         */
        public void refresh() {
            this.bundleRepository.refresh();
        }
    }

    /**
     * {@inheritDoc}
     */
    public boolean expandOnRefresh(List<BundleManifest> bundleManifests, String bundleSymbolicName, ImportPromotionVector importPromotionVector,
        Scoper scoper) throws UnableToSatisfyDependenciesException, ImportMergeException {
        Assert.isInstanceOf(StandardImportPromotionVector.class, importPromotionVector);
        StandardImportPromotionVector standardImportPromotionVector = (StandardImportPromotionVector) importPromotionVector;
        BundleRepository combinedRepository = new AdditionalManifestSearchingBundleRepository(bundleManifests, this.bundleRepository);
        for (BundleManifest bundleManifest : bundleManifests) {
            BundleSymbolicName symbolicName = bundleManifest.getBundleSymbolicName();
            if (symbolicName != null && bundleSymbolicName.equals(symbolicName.getSymbolicName())) {
                TrackedPackageImports bundlePackageImportsToBePromoted = this.trackedPackageImportsFactory.createCollector();
                bundleManifest.expandPackageLists();
                expandImportsIfNecessary(bundleManifest, combinedRepository, bundlePackageImportsToBePromoted, scoper);
                if (!bundlePackageImportsToBePromoted.isEquivalent(standardImportPromotionVector.get(bundleSymbolicName))) {
                    standardImportPromotionVector.put(bundleSymbolicName, bundlePackageImportsToBePromoted);
                    return false;
                }
                mergeImports(standardImportPromotionVector.getPromotedImports(), bundleManifest);
                return true;
            }
        }
        return false;
    }
}
