/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.framework.support;

import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import com.springsource.server.osgi.framework.ImportExpander.ImportPromotionVector;
import com.springsource.server.osgi.framework.support.merge.TrackedPackageImports;
import com.springsource.server.osgi.framework.support.merge.TrackedPackageImportsFactory;

/**
 * {@link StandardImportPromotionVector} maintains a vector of promoted TrackedPackageImports indexed by bundle symbolic
 * name.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * This class is thread safe.
 * 
 * @author Glyn Normington
 * @since 1.0
 */
public class StandardImportPromotionVector implements ImportPromotionVector {
    
    private final TrackedPackageImportsFactory trackedPackageImportsFactory;

    private final Map<String, TrackedPackageImports> imports = new ConcurrentHashMap<String, TrackedPackageImports>();

    private final Object monitor = new Object();
    
    /**
     * Construct a {@link StandardImportPromotionVector} with the given {@link TrackedPackageImportsFactory}.
     * 
     * @param trackedPackageImportsFactory a {@link TrackedPackageImportsFactory}
     */
    StandardImportPromotionVector(TrackedPackageImportsFactory trackedPackageImportsFactory) {
        this.trackedPackageImportsFactory = trackedPackageImportsFactory;
    }

    /**
     * Add the given {@link TrackedPackageImports} at the index of the given bundle symbolic name.
     * <p />
     * Any entry at the given index is replaced.
     * 
     * @param bundleSymbolicName the index into the vector
     * @param promotedImports the promoted imports to be added at the given index
     */
    void put(String bundleSymbolicName, TrackedPackageImports promotedImports) {
        synchronized (this.monitor) {
            this.imports.put(bundleSymbolicName, promotedImports);
        }
    }

    /**
     * Get the {@link TrackedPackageImports} at the index of the given bundle symbolic name.
     * 
     * @param bundleSymbolicName the index into the vector
     * @return the promoted imports at the given index or <code>null</code> if there are no such promoted imports
     */
    TrackedPackageImports get(String bundleSymbolicName) {
        return this.imports.get(bundleSymbolicName);
    }

    /**
     * Get a merged form of the promoted imports in this vector.
     * 
     * @return a merged {@link TrackedPackageImports}
     */
    TrackedPackageImports getPromotedImports() {
        synchronized (this.monitor) {
            TrackedPackageImports merged = this.trackedPackageImportsFactory.createCollector();
            for (TrackedPackageImports promotedImports : this.imports.values()) {
                merged.merge(promotedImports);
            }
            return merged;
        }
    }

}
