/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.framework.support.merge;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.springframework.util.Assert;

import com.springsource.server.osgi.framework.ImportMergeException;
import com.springsource.server.osgi.manifest.BundleManifestFactory;
import com.springsource.server.osgi.manifest.StandardBundleManifestFactory;
import com.springsource.server.osgi.manifest.VersionRange;
import com.springsource.server.osgi.manifest.BundleManifest.ImportPackageHeader.PackageImport;
import com.springsource.server.osgi.manifest.BundleManifest.ImportPackageHeader.PackageImport.Resolution;
import com.springsource.server.osgi.manifest.parse.OSGiIntegrationLogService;

/**
 * {@link AbstractTrackedPackageImports} provides the general implementations of {@link TrackedPackageImports}.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * This class is thread safe.
 * 
 * @author Glyn Normington
 * @since 1.0
 */
public abstract class AbstractTrackedPackageImports implements TrackedPackageImports {

    protected static final String SOURCE_SEPARATOR = ", ";

    private static final String VERSION_ATTRIBUTE_NAME = "version";

    private static final String VERSION_ALTERNATE_ATTRIBUTE_NAME = "specification-version";

    private static final String BUNDLE_VERSION_ATTRIBUTE_NAME = "bundle-version";

    private final Object monitor = new Object();

    /**
     * The current merged package imports are held as a map of package name to {@link PackageImport}. Each package
     * import contains one and only one imported package name which corresponds to the package name used to index the
     * package import in the map. The map is valid if and only if mergeException is <code>null</code>.
     */
    protected final Map<String, PackageImport> mergedPackageImports = new HashMap<String, PackageImport>();

    private ImportMergeException mergeException = null;

    protected final List<TrackedPackageImports> sources = new ArrayList<TrackedPackageImports>();

    private static final BundleManifestFactory bundleManifestFactory = new StandardBundleManifestFactory(new OSGiIntegrationLogService());

    private static final Object bundleManifestFactoryMonitor = new Object();

    /**
     * Construct an {@link AbstractTrackedPackageImports} from the given package imports.
     * 
     * @param initialPackageImports a map of package name to {@link PackageImport}
     */
    public AbstractTrackedPackageImports(Map<String, PackageImport> packageImports) {
        this.mergedPackageImports.putAll(packageImports);
    }

    /**
     * {@inheritDoc}
     */
    public void merge(TrackedPackageImports importsToMerge) throws ImportMergeException {
        synchronized (this.monitor) {
            checkMergeException();
            try {
                // Add the new imports before merging so they are included in any diagnostics.
                sources.add(importsToMerge);
                doMerge(importsToMerge);
            } catch (ImportMergeException e) {
                this.mergeException = e;
                throw e;
            }
        }
    }

    /**
     * Merge the given package imports into this collection of package imports. If there is a conflict, issue
     * diagnostics and throw {@link ImportMergeException}.
     * <p />
     * Pre-condition: the monitor is held and the current merged imports have no conflicts.
     * 
     * @param importsToMerge
     * @throws ImportMergeException
     */
    private void doMerge(TrackedPackageImports importsToMerge) throws ImportMergeException {
        List<PackageImport> mergedImportsToMerge = importsToMerge.getMergedImports();
        for (PackageImport packageImportToMerge : mergedImportsToMerge) {
            String pkg = packageImportToMerge.getPackageNames().get(0);
            PackageImport mergedPackageImport = this.mergedPackageImports.get(pkg);
            if (mergedPackageImport == null) {
                this.mergedPackageImports.put(pkg, packageImportToMerge);
            } else {
                mergePackageImport(mergedPackageImport, packageImportToMerge);
            }
        }

    }

    /**
     * Merge the given source package import into the given target package import. Throw {@link ImportMergeException} if
     * and only if there is a merge clash.
     * 
     * @param targetPackageImport the package import to be merged and updated
     * @param sourceImportToMerge the package import to be merged in
     * @throws ImportMergeException thrown if there is a merge clash
     */
    private void mergePackageImport(PackageImport targetPackageImport, PackageImport sourceImportToMerge) throws ImportMergeException {
        mergeAttributes(targetPackageImport, sourceImportToMerge);
        mergeDirectives(targetPackageImport, sourceImportToMerge);

    }

    /**
     * Merge the attributes of the source package import into those of the target package import. Throw
     * {@link ImportMergeException} if and only if there is a merge clash.
     * 
     * @param targetPackageImport the package import to be merged and updated
     * @param sourceImportToMerge the package import to be merged in
     * @throws ImportMergeException thrown if there is a merge clash
     */
    private void mergeAttributes(PackageImport targetPackageImport, PackageImport sourceImportToMerge) throws ImportMergeException {
        Map<String, String> targetAttributes = targetPackageImport.getAttributes();
        Map<String, String> sourceAttributes = sourceImportToMerge.getAttributes();

        mergeVersionRanges(targetPackageImport, sourceImportToMerge);

        mergeBundleVersionRanges(targetPackageImport, sourceImportToMerge);

        for (String sourceAttributeName : sourceAttributes.keySet()) {
            if (!isVersionAttribute(sourceAttributeName)) {
                String sourceAttributeValue = sourceAttributes.get(sourceAttributeName);
                String targetAttributeValue = targetAttributes.get(sourceAttributeName);
                if (targetAttributeValue != null) {
                    if (!targetAttributeValue.equals(sourceAttributeValue)) {
                        throw new ImportMergeException(getPackageName(targetPackageImport), getPackageSources(targetPackageImport),
                            "conflicting values '" + sourceAttributeValue + "', '" + targetAttributeValue + "' of attribute '" + sourceAttributeName
                                + "'");
                    }
                } else {
                    targetAttributes.put(sourceAttributeName, sourceAttributeValue);
                }
            }
        }

    }

    /**
     * Merge the version ranges of the given source and target attributes and update the target attributes if necessary
     * with the merged range. If the version ranges are disjoint, throw {@link ImportMergeException}.
     * 
     * @param targetPackageImport the package import to be merged and updated
     * @param sourceAttributes
     * @throws ImportMergeException
     */
    private void mergeVersionRanges(PackageImport targetPackageImport, PackageImport sourceImportToMerge) throws ImportMergeException {
        Map<String, String> sourceAttributes = sourceImportToMerge.getAttributes();
        VersionRange sourceVersionRange = getVersionRange(sourceAttributes);
        if (sourceVersionRange != null) {
            Map<String, String> targetAttributes = targetPackageImport.getAttributes();
            VersionRange targetVersionRange = getVersionRange(targetAttributes);
            VersionRange mergedVersionRange;
            if (targetVersionRange == null) {
                mergedVersionRange = sourceVersionRange;
            } else {
                mergedVersionRange = VersionRange.intersection(sourceVersionRange, targetVersionRange);
                if (mergedVersionRange == null) {
                    throw new ImportMergeException(getPackageName(targetPackageImport), getPackageSources(targetPackageImport),
                        "disjoint package version ranges");
                }
            }
            targetAttributes.put(VERSION_ATTRIBUTE_NAME, mergedVersionRange.toParseString());
        }
    }

    /**
     * Get the version range for the given attributes.
     * 
     * @param attributes the attributes which may specify a version range
     * @return the version range or <code>null</code> if no version range is specified
     */
    private VersionRange getVersionRange(Map<String, String> attributes) {
        String versionRangeString = attributes.get(VERSION_ATTRIBUTE_NAME);
        if (versionRangeString == null) {
            versionRangeString = attributes.get(VERSION_ALTERNATE_ATTRIBUTE_NAME);
        }
        return versionRangeString == null ? null : new VersionRange(versionRangeString);
    }

    /**
     * Merge the bundle version ranges of the given source and target attributes and update the target attributes if
     * necessary with the merged range. If the bundle version ranges are disjoint, throw {@link ImportMergeException}.
     * 
     * @param targetPackageImport the package import to be merged and updated
     * @param sourceAttributes
     * @throws ImportMergeException
     */
    private void mergeBundleVersionRanges(PackageImport targetPackageImport, PackageImport sourceImportToMerge) throws ImportMergeException {
        Map<String, String> sourceAttributes = sourceImportToMerge.getAttributes();
        VersionRange sourceVersionRange = getBundleVersionRange(sourceAttributes);
        if (sourceVersionRange != null) {
            Map<String, String> targetAttributes = targetPackageImport.getAttributes();
            VersionRange targetVersionRange = getBundleVersionRange(targetAttributes);
            VersionRange mergedVersionRange;
            if (targetVersionRange == null) {
                mergedVersionRange = sourceVersionRange;
            } else {
                mergedVersionRange = VersionRange.intersection(sourceVersionRange, targetVersionRange);
                if (mergedVersionRange == null) {
                    throw new ImportMergeException(getPackageName(targetPackageImport), getPackageSources(targetPackageImport),
                        "disjoint bundle version ranges");
                }
            }
            targetAttributes.put(BUNDLE_VERSION_ATTRIBUTE_NAME, mergedVersionRange.toParseString());
        }
    }

    /**
     * Get the bundle version range for the given attributes.
     * 
     * @param attributes the attributes which may specify a version range
     * @return the version range or <code>null</code> if no version range is specified
     */
    private VersionRange getBundleVersionRange(Map<String, String> attributes) {
        String versionRangeString = attributes.get(BUNDLE_VERSION_ATTRIBUTE_NAME);
        return versionRangeString == null ? null : new VersionRange(versionRangeString);
    }

    /**
     * Return <code>true</code> if and only if the given attribute name is that of a version or bundle version
     * attribute.
     * 
     * @param attributeName the attribute name
     * @return whether or not the given attribute name is that of a version or bundle version attribute
     */
    private boolean isVersionAttribute(String attributeName) {
        return VERSION_ATTRIBUTE_NAME.equals(attributeName) || VERSION_ALTERNATE_ATTRIBUTE_NAME.equals(attributeName)
            || BUNDLE_VERSION_ATTRIBUTE_NAME.equals(attributeName);
    }

    /**
     * Merge the directives of the source package import into those of the target package import.
     * 
     * @param targetPackageImport
     * @param sourceImportToMerge
     */
    private void mergeDirectives(PackageImport targetPackageImport, PackageImport sourceImportToMerge) {
        if (targetPackageImport.getResolution() == Resolution.OPTIONAL && sourceImportToMerge.getResolution() == Resolution.MANDATORY) {
            targetPackageImport.setResolution(Resolution.MANDATORY);
        }
    }

    /**
     * Get a string describing the sources of the package of the given package import. This should help in diagnosing
     * the root cause of a conflicting merge.
     * 
     * @param pkg the package whose sources are required
     * @return a string describing the given package's sources
     */
    private String getPackageSources(PackageImport packageImport) {
        return getSources(getPackageName(packageImport));
    }

    /**
     * {@inheritDoc}
     */
    public String getSources(String pkg) {
        synchronized (this.monitor) {
            StringBuffer sourcesDescription = new StringBuffer();
            boolean first = true;
            String source = getSource(pkg);
            if (source != null) {
                sourcesDescription.append(source);
                first = false;
            }
            for (TrackedPackageImports trackedPackageImports : this.sources) {
                String sources = trackedPackageImports.getSources(pkg);
                if (sources != null) {
                    if (!first) {
                        sourcesDescription.append(SOURCE_SEPARATOR);
                    }
                    sourcesDescription.append(sources);
                    first = false;
                }
            }
            return first ? null : sourcesDescription.toString();
        }
    }

    /**
     * Get the package name of a package import that contains only one package name.
     * 
     * @param packageImport the package import
     * @return the package name
     */
    private String getPackageName(PackageImport packageImport) {
        return packageImport.getPackageNames().get(0);
    }

    /**
     * {@inheritDoc}
     */
    public final List<PackageImport> getMergedImports() throws ImportMergeException {
        synchronized (this.monitor) {
            checkMergeException();
            List<PackageImport> mergedImports = new ArrayList<PackageImport>();
            for (PackageImport packageImport : this.mergedPackageImports.values()) {
                mergedImports.add(packageImport);
            }
            return mergedImports;
        }
    }

    /**
     * If a merge failure has occurred, re-throw the {@link ImportMergeException}.
     * 
     * @throws ImportMergeException
     */
    private void checkMergeException() throws ImportMergeException {
        if (this.mergeException != null) {
            throw this.mergeException;
        }
    }

    /**
     * Convert a given list of package imports with no duplicate package names to a map of package name to
     * {@link PackageImport}.
     * 
     * @param packageImports a list of <code>PackageImport</code>
     * @return a map of package name to <code>PackageImport</code>
     */
    protected static Map<String, PackageImport> convertPackageImportListToMap(List<PackageImport> packageImports) {
        Map<String, PackageImport> initialPackageImports = new HashMap<String, PackageImport>();
        for (PackageImport packageImport : packageImports) {
            if (packageImport.getPackageNames().size() > 1) {
                for (String packageName : packageImport.getPackageNames()) {
                    PackageImport expandedPackageImport;
                    synchronized (bundleManifestFactoryMonitor) {
                        expandedPackageImport = bundleManifestFactory.createPackageImport();
                    }
                    expandedPackageImport.getPackageNames().add(packageName);
                    expandedPackageImport.getAttributes().putAll(packageImport.getAttributes());
                    expandedPackageImport.setResolution(packageImport.getResolution());
                    Assert.isNull(initialPackageImports.put(packageName, expandedPackageImport),
                        "input packageImports must not contain duplicate items");
                }
            } else {
                Assert.isNull(initialPackageImports.put(packageImport.getPackageNames().get(0), packageImport),
                    "input packageImports must not contain duplicate items");
            }

        }
        return initialPackageImports;
    }

    /**
     * {@inheritDoc}
     */
    public boolean isEmpty() {
        synchronized (this.monitor) {
            return this.mergedPackageImports.isEmpty();
        }
    }

    /**
     * {@inheritDoc}
     */
    public boolean isEquivalent(TrackedPackageImports otherTrackedPackageImports) {
        if (otherTrackedPackageImports == null) {
            return isEmpty();
        }
        Assert.isInstanceOf(AbstractTrackedPackageImports.class, otherTrackedPackageImports);
        AbstractTrackedPackageImports otherAbstractTrackedPackageImports = (AbstractTrackedPackageImports) otherTrackedPackageImports;
        // Lock the object monitors in a predictable order to avoid deadlocks.
        if (System.identityHashCode(this) >= System.identityHashCode(otherTrackedPackageImports)) {
            synchronized (this.monitor) {
                synchronized (otherAbstractTrackedPackageImports.monitor) {
                    return this.mergedPackageImports.equals(otherAbstractTrackedPackageImports.mergedPackageImports);
                }
            }
        } else {
            synchronized (otherAbstractTrackedPackageImports.monitor) {
                synchronized (this.monitor) {
                    return this.mergedPackageImports.equals(otherAbstractTrackedPackageImports.mergedPackageImports);
                }
            }
        }
    }

}
