/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.framework.support.merge;

import java.util.HashMap;
import java.util.Map;

import org.springframework.util.Assert;

import com.springsource.server.osgi.framework.ImportMergeException;
import com.springsource.server.osgi.manifest.BundleManifest;
import com.springsource.server.osgi.manifest.BundleManifest.ImportPackageHeader;
import com.springsource.server.osgi.manifest.BundleManifest.ImportPackageHeader.PackageImport;

/**
 * {@link BundleTrackedPackageImports} provides merging and tracking for package imports originating in bundle
 * manifests.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * This class is thread safe.
 * 
 * @author Glyn Normington
 * @since 1.0
 */
public final class BundleTrackedPackageImports extends AbstractTrackedPackageImports {

    private final BundleManifest bundleManifest;

    /**
     * Construct a {@link TrackedPackageImports} sourced from a bundle manifest.
     * 
     * @param importPackageHeader
     */
    public BundleTrackedPackageImports(BundleManifest bundleManifest) {
        super(getInitialPackageImports(bundleManifest));
        this.bundleManifest = bundleManifest;
    }

    /**
     * Extract a map of package name to {@link PackageImport} from the given bundle manifest.
     * <p />
     * Each <code>PackageImport</code> in the map contains a single package name.
     * 
     * @param bundleManifest the bundle manifest containing the package imports
     * @return a map of package name to {@link PackageImport}
     */
    private static Map<String, PackageImport> getInitialPackageImports(BundleManifest bundleManifest) {
        Assert.notNull(bundleManifest, "bundleManifest must be non-null");
        ImportPackageHeader importPackageHeader = bundleManifest.getImportPackageHeader();
        return importPackageHeader == null ? new HashMap<String, PackageImport>()
            : convertPackageImportListToMap(importPackageHeader.getPackageImports());
    }

    /**
     * {@inheritDoc}
     */
    public String getSource(String pkg) {
        Assert.notNull(this.bundleManifest.getBundleSymbolicName(), "bundleManifest must have a symbolic name");
        return !getInitialPackageImports(this.bundleManifest).containsKey(pkg) ? null
            : ((this.bundleManifest.getFragmentHostHeader() == null ? "bundle " : "fragment ") + this.bundleManifest.getBundleSymbolicName().getSymbolicName());
    }
    
    /** 
     * {@inheritDoc}
     */
    @Override public final void merge(TrackedPackageImports importsToMerge) throws ImportMergeException {
        super.merge(importsToMerge);
    }

}
