/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.framework.support.merge;

import java.util.List;

import com.springsource.server.osgi.framework.ImportMergeException;
import com.springsource.server.osgi.manifest.BundleManifest;
import com.springsource.server.osgi.manifest.BundleManifest.ImportPackageHeader.PackageImport;

/**
 * {@link TrackedPackageImportsFactory} is an interface for creating {@link TrackedPackageImports} instances.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Implementations of this interface must be thread safe.
 * 
 * @author Glyn Normington
 * @since 1.0
 * @see TrackedPackageImports
 */
public interface TrackedPackageImportsFactory {

    /**
     * Create a {@link TrackedPackageImports} from the given {@link BundleManifest}.
     * <p />
     * This is typically used to take the imports of a bundle and merge in imports from other sources.
     * 
     * @param bundleManifest the bundle manifest
     * @return a <code>TrackedPackageImports</code> instance
     */
    TrackedPackageImports create(BundleManifest bundleManifest);

    /**
     * Create a {@link TrackedPackageImports} from the given list of {@link PackageImport} and given source.
     * <p />
     * This is typically used to represent a collection of imports which are to be added to one or more bundles.
     * 
     * @param packageImports a list of <code>PackageImport</code> which must not contain duplicate packages
     * @param source the source of the given package imports
     * @return a <code>TrackedPackageImports</code> instance
     */
    TrackedPackageImports create(List<PackageImport> packageImports, String source);

    /**
     * Create a {@link TrackedPackageImports} with no {@link PackageImport}s of its own.
     * <p />
     * This is typically used to gather together a collection of imports from multiple sources which are then to be
     * added to one or more bundles.
     * 
     * @return a <code>TrackedPackageImports</code> instance
     */
    TrackedPackageImports createCollector();

    /**
     * Create a <i>container</i> {@link TrackedPackageImports} which contains any {@link PackageImport}s which are
     * subsequently merged into it but which has no {@link PackageImport}s of its own. If the container is involved in
     * producing an {@link ImportMergeException}, the sources of the {@link PackageImport}s merged into the container
     * are wrapped in the given containing source.
     * <p />
     * This is typically used for named groups of imports such as the group of package imports corresponding to a
     * library (in which case the name of the group could be the library name).
     * 
     * @return a <code>TrackedPackageImports</code> instance
     */
    TrackedPackageImports createContainer(String containingSource);
    
    /**
     * Create an empty, immutable {@link TrackedPackageImports}.
     * 
     * @return a <code>TrackedPackageImports</code> instance
     */
    TrackedPackageImports createEmpty();

}
