/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.manifest;

import java.io.Writer;
import java.util.List;
import java.util.Map;
import java.util.jar.Manifest;

import org.osgi.framework.Version;

/**
 * This interface is an abstract syntax tree for an OSGi bundle manifest. Parsing is lazy, so an exception may be thrown
 * when a manifest header is accessed for the first time. The interface also supports certain kinds of modification and
 * unparsing to generate file contents for the modified manifest. <p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Implementations of this interface need not be thread safe.
 * 
 * @author Glyn Normington
 * @author Andy Wilkinson
 * @since 1.0
 */
public interface BundleManifest {

    public interface DynamicImportPackageHeader {

        /**
         * Gets the dynamic package imports in this dynamic import package header. Modifying the returned list modifies
         * the dynamic import package header.
         * 
         * @return a list of dynamic package imports
         */
        List<DynamicPackageImport> getDynamicPackageImports();

        /**
         * Replace the dynamic package imports of this dynamic import package header.
         * 
         * @param dynamicPackageImports a list of dynamic package imports
         */
        void setDynamicPackageImports(List<DynamicPackageImport> dynamicPackageImports);

        /**
         * Check that this header has no missing required components and throw an exception if there are omissions.
         */
        void checkWellFormed();

        /**
         * Abstract representation of a dynamic package import of one or more possible wildcarded packages with
         * attributes.
         */
        public interface DynamicPackageImport {

            /**
             * Gets the possibly wildcarded package names included in this declaration. Modifying the returned list
             * modifies the dynamic package import.
             * 
             * @return the wildcarded package names
             */
            List<String> getWildcardPackageNames();

            /**
             * Gets the attributes included in this declaration. Modifying the returned map modifies the dynamic package
             * import.
             * 
             * @return the attributes
             */
            Map<String, String> getAttributes();

            /**
             * Check that this dynamic package import has no missing required components and throw an exception if there
             * are omissions.
             */
            void checkWellFormed();
        }
    }

    public interface ExportPackageHeader {

        /**
         * Gets the package exports in this export package header. Modifying the returned list modifies the export
         * package header.
         * 
         * @return a list of package exports
         */
        List<PackageExport> getPackageExports();

        /**
         * Replace the package imports of this export package header.
         * 
         * @param packageExports a list of package exports
         */
        void setPackageExports(List<PackageExport> packageExports);

        /**
         * Check that this header has no missing required components and throw an exception if there are omissions.
         */
        void checkWellFormed();

        public interface PackageExport {

            /**
             * Gets the package names included in this declaration. Modifying the returned list modifies the package
             * import.
             * 
             * @return the package names
             */
            List<String> getPackageNames();

            /**
             * Gets the attributes included in this declaration. Modifying the returned map modifies the package import.
             * 
             * @return the attributes
             */
            Map<String, String> getAttributes();

            /**
             * Get the version of this declaration.
             * 
             * @return the version of the exported package(s)
             */
            Version getVersion();

            /**
             * Gets the package names included in this declaration. Modifying the returned list modifies the package
             * import.
             * 
             * @return the package names
             */
            List<String> getUses();

            /**
             * Gets the mandatory attributes listed by this declaration. Modifying the returned list modifies the
             * package import.
             * 
             * @return the attributes
             */
            List<String> getMandatory();

            /**
             * Get the resolution directive.
             * 
             * @return the resolution directive value
             */
            List<String> getInclude();

            /**
             * Gets the list of exclusions
             * 
             * @return the list of exclusions
             */
            List<String> getExclude();

            /**
             * Check that this header has no missing required components.
             */
            void checkWellFormed();
        }
    }

    public interface FragmentHostHeader {

        public enum Extension {
            FRAMEWORK, BOOTCLASSPATH
        };

        /**
         * Gets the bundle symbolic name included in this declaration.
         * 
         * @return the bundle symbolic names
         */
        String getBundleSymbolicName();

        /**
         * Set the bundle symbolic name included in this declaration.
         * 
         * @param hostName the bundle symbolic name of the host bundle
         */
        void setBundleSymbolicName(String hostName);

        /**
         * Gets the attributes included in this declaration. Modifying the returned map modifies the fragment host
         * header.
         * 
         * @return the attributes
         */
        Map<String, String> getAttributes();

        /**
         * Get the extension directive.
         * 
         * @return the extension directive value
         */
        Extension getExtension();

        /**
         * Set the extension directive.
         * 
         * @param extension the new extension directive value
         */
        void setExtension(Extension extension);

        /**
         * Check that this header has no missing required components and throw an exception if there are omissions.
         */
        void checkWellFormed();
    }

    public interface ImportPackageHeader {

        /**
         * Gets the package imports in this import package header. Modifying the returned list modifies the import
         * package header.
         * 
         * @return a list of package imports
         */
        List<PackageImport> getPackageImports();

        /**
         * Replace the package imports of this import package header.
         * 
         * @param packageImports a list of package imports
         */
        void setPackageImports(List<PackageImport> packageImports);

        /**
         * Check that this header has no missing required components and throw an exception if there are omissions.
         */
        void checkWellFormed();

        /**
         * Abstract representation of a package import of one or more packages with attributes and directives.
         */
        public interface PackageImport {

            public enum Resolution {
                MANDATORY, OPTIONAL
            };

            /**
             * Gets the package names included in this declaration. Modifying the returned list modifies the package
             * import.
             * 
             * @return the package names
             */
            List<String> getPackageNames();

            /**
             * Sets the package names included in this declaration.
             * 
             * @param packageNames the package names
             */
            void setPackageNames(List<String> packageNames);

            /**
             * Gets the attributes included in this declaration. Modifying the returned map modifies the package import.
             * 
             * @return the attributes
             */
            Map<String, String> getAttributes();

            /**
             * Get the resolution directive.
             * 
             * @return the resolution directive value
             */
            Resolution getResolution();

            /**
             * Set the resolution directive.
             * 
             * @param resolution the new resolution directive value
             */
            void setResolution(Resolution resolution);

            /**
             * Check that this header has no missing required components and throw an exception if there are omissions.
             */
            void checkWellFormed();
        }
    }

    public interface RequireBundleHeader {

        /**
         * Gets the list of required bundles
         * 
         * @return the list of required bundles
         */
        List<RequireBundle> getRequiredBundles();

        /**
         * 
         */
        void setRequiredBundles(List<RequireBundle> requireBundles);

        /**
         * Check that this header has no missing required components and throw an exception if there are omissions.
         */
        void checkWellFormed();

        /**
         * Abstract representation of a required bundle with attributes and directives.
         */
        public interface RequireBundle {

            public enum Resolution {
                MANDATORY, OPTIONAL
            };

            public enum Visibility {
                PRIVATE, REEXPORT
            };

            /**
             * Gets the bundle symbolic name included in this declaration.
             * 
             * @return the bundle symbolic name
             */
            String getBundleSymbolicName();

            /**
             * Set the bundle symbolic name for this declaration.
             * 
             * @param string the bundle symbolic name
             */
            void setBundleSymbolicName(String string);

            /**
             * Gets the attributes included in this declaration. Modifying the returned map modifies the attributes.
             * 
             * @return the attributes
             */
            Map<String, String> getAttributes();

            /**
             * Get the resolution directive.
             * 
             * @return the resolution directive value
             */
            Resolution getResolution();

            /**
             * Set the resolution directive.
             * 
             * @param resolution the new resolution directive value
             */
            void setResolution(Resolution resolution);

            /**
             * Get the visibility directive.
             * 
             * @return the visibility directive value
             */
            Visibility getVisibility();

            /**
             * Set the visibility directive.
             * 
             * @param visibility the new visibility directive value
             */
            void setVisibility(Visibility visibility);

            /**
             * Check that this header has no missing required components and throw an exception if there are omissions.
             */
            void checkWellFormed();

        }

    }

    /**
     * Non-standard header for expressing a dependency on a bundle. Bundles are referenced in the header by
     * <code>Bundle-SymbolicName</code> and a version range. Each referenced bundle is expanded into an
     * <code>Import-Package</code> header entry for each of the referenced bundle's exported packages.
     * 
     * <strong>Concurrent Semantics</strong><br /> Implementations <strong>must</strong> be thread-safe.
     * 
     * @author Andy Wilkinson
     * @since 1.0
     */
    public interface ImportBundleHeader {

        /**
         * Gets the list of bundles that are imported by this header. If no bundles are imported, an empty list is
         * returned.
         * 
         * @return the non-null but possibly empty list of imported bundles.
         */
        List<BundleImport> getImportedBundles();

        /**
         * Sets the list of bundles that are imported by this header. If no bundles are imported, an empty list must be
         * supplied.
         * 
         * @param bundles The list of bundles referenced in this header.
         */
        void setImportedBundles(List<BundleImport> bundles);

        /**
         * Check that this header has no missing required components and throw an exception if there are omissions.
         */
        void checkWellFormed();

        public interface BundleImport {

            public enum Resolution {
                MANDATORY, OPTIONAL
            }

            public enum Sharing {
                AUTOMATIC, SHARE, CLONE
            }

            /**
             * Gets the symbolic name of the bundle referenced by this import
             * 
             * @return the bundle symbolic name
             */
            String getBundleSymbolicName();

            /**
             * Gets the version range of this import
             * 
             * @return the (non-<code>null</code>) version range
             */
            String getVersionRange();

            /**
             * Gets the resolution of this import
             * 
             * @return the import's resolution
             */
            Resolution getResolution();

            /**
             * Check that there are no missing required components and throw an exception if there are omissions.
             */
            void checkWellFormed();

            /**
             * Get the sharing directive.
             * 
             * @return the sharing directive value
             */
            Sharing getSharing();

            /**
             * Gets the import-scope directive as a boolean.
             * 
             * @return <code>true</code> if and only if the import-scope:=application
             */
            boolean getApplicationImportScope();
        }
    }

    /**
     * Non-standard header for expressing a dependency on a library. A library is a collection of bundles that is
     * grouped together under a name and version.
     */
    public interface ImportLibraryHeader {

        /**
         * Gets the list of libraries that are imported by this header. If no libraries are imported, an empty list is
         * returned.
         * 
         * @return the non-null but possibly empty list of imported libraries.
         */
        List<LibraryImport> getImportedLibraries();

        /**
         * Sets the list of libraries that are imported by this header. If no libraries are imported, an empty list must
         * be supplied.
         * 
         * @param bundles The list of libraries referenced in this header.
         */
        void setImportedLibraries(List<LibraryImport> libraries);

        /**
         * Check that this header has no missing required components and throw an exception if there are omissions.
         */
        void checkWellFormed();

        /**
         * Represents a single import library declaration.
         */
        public interface LibraryImport {

            public enum Resolution {
                MANDATORY, OPTIONAL
            }

            public enum Sharing {
                AUTOMATIC, SHARE, CLONE
            }

            /**
             * Gets the library name included in this declaration.
             * 
             * @return the library
             */
            String getLibraryName();

            /**
             * Gets the library version range included in this declaration.
             * 
             * @return the (non-<code>null</code>) library version range.
             */
            String getVersionRange();

            /**
             * Get the resolution directive.
             * 
             * @return the resolution directive value
             */
            Resolution getResolution();

            /**
             * Get the sharing directive.
             * 
             * @return the sharing directive value
             */
            Sharing getSharing();

            /**
             * Gets the attributes included in this declaration. Modifying the returned map modifies the attributes.
             * 
             * @return the attributes
             */
            Map<String, String> getAttributes();

            /**
             * Check that this header has no missing required components and throw an exception if there are omissions.
             */
            void checkWellFormed();

        }
    }

    public interface BundleSymbolicName {

        public enum FragmentAttachment {
            ALWAYS, NEVER, RESOLVE_TIME;
        }

        /**
         * Gets the symbolic name of the declaration
         * 
         * @return the symbolic name of the declaration
         */
        String getSymbolicName();

        /**
         * Sets the symbolic name of the declaration
         * 
         * @param symbolicName the symbolic name of the declaration
         */
        void setSymbolicName(String symbolicName);

        /**
         * Gets the value of the singleton directive
         * 
         * @return the singleton directive
         */
        boolean isSingleton();

        /**
         * Gets the <code>fragment-attachment</code> directive
         * 
         * @return the <code>fragment-attachment</code> directive
         */
        FragmentAttachment getFragmentAttachment();

        /**
         * Gets the attributes included in the <code>Bundle-SymbolicName</code> declaration. Modifying the returned map
         * modifies the declaration's attributes
         * 
         * @return the <code>Bundle-SymbolicName</code> declaration's attributes
         */
        Map<String, String> getAttributes();
    }

    public interface BundleActivationPolicy {

        enum Policy {
            LAZY;
        }

        Policy getActivationPolicy();

        Map<String, String> getDirectives();
    }

    /**
     * Get the bundle class path as a list of JAR file path names. The returned value may be modified in order to change
     * the bundle class path of this bundle manifest.
     * 
     * @return the list of JAR file path names
     */
    List<String> getBundleClassPath();

    /**
     * Set the bundle class path as a list of JAR file path names. This replaces any current bundle class path.
     * 
     * @param bundleClassPath The list of entries in the bundle class path
     */
    void setBundleClassPath(List<String> bundleClassPath);

    /**
     * Get the bundle description.
     * 
     * @return the bundle description
     */
    String getBundleDescription();

    /**
     * Set the bundle description to the given value.
     * 
     * @param bundleDescription the bundle description to be set
     */
    void setBundleDescription(String bundleDescription);

    /**
     * Get the bundle documentation URL.
     * 
     * @return the bundle documentation URL string
     */
    String getBundleDocUrl();

    /**
     * Set the bundle documentation URL.
     * 
     * @param bundleDocUrl the bundle documentation URL string
     */
    void setBundleDocUrl(String bundleDocUrl);

    /**
     * Get the bundle's name
     * 
     * @return The bundle's name
     */
    String getBundleName();

    /**
     * Set the bundle's name
     * 
     * @param bundleName The bundle's name
     */
    void setBundleName(String bundleName);

    /**
     * Get the bundle's symbolic name
     * 
     * @return The bundle's symbolic name
     */
    BundleSymbolicName getBundleSymbolicName();

    /**
     * Set the bundle's symbolic name
     * 
     * @param bundleSymbolicName The bundle's symbolic name
     */
    void setBundleSymbolicName(BundleSymbolicName bundleSymbolicName);

    /**
     * Get the bundle's update location
     * 
     * @return the bundle's update location
     */
    String getBundleUpdateLocation();

    /**
     * Set the bundle's update location
     * 
     * @param bundleUpdateLocation the bundle's update location
     */
    void setBundleUpdateLocation(String bundleUpdateLocation);

    /**
     * Get the bundle's version
     * 
     * @return bundleVersion the bundle's version
     */
    String getBundleVersion();

    /**
     * Set the bundle's version
     * 
     * @param bundleVersion bundle's version
     */
    void setBundleVersion(String bundleVersion);

    /**
     * Get the bundle's DynamicImport-Package header
     * 
     * @return the bundle's DynamicImport-Package header
     */
    DynamicImportPackageHeader getDynamicImportPackageHeader();

    /**
     * Set the bundle's DynamicImport-Package header
     * 
     * @param dynamicImportPackageHeader bundle's DynamicImport-Package header
     */
    void setDynamicImportPackageHeader(DynamicImportPackageHeader dynamicImportPackageHeader);

    /**
     * Get the bundle's Export-Package header
     * 
     * @return the bundle's Export-Package header
     */
    ExportPackageHeader getExportPackageHeader();

    /**
     * Set the bundle's Export-Package header
     * 
     * @param exportPackageHeader the bundle's Export-Package header
     */
    void setExportPackageHeader(ExportPackageHeader exportPackageHeader);

    /**
     * Get the bundle's FragmentHost header
     * 
     * @return the bundle's FragmentHost header
     */
    FragmentHostHeader getFragmentHostHeader();

    /**
     * Set the bundle's FragmentHost header
     * 
     * @param fragmentHostHeader bundle's FragmentHost header
     */
    void setFragmentHostHeader(FragmentHostHeader fragmentHostHeader);

    /**
     * Get the bundle's Import-Package header
     * 
     * @return the bundle's Import-Package header
     */
    ImportPackageHeader getImportPackageHeader();

    /**
     * Set the bundle's Import-Package header
     * 
     * @param importPackageHeader the bundle's Import-Package header
     */
    void setImportPackageHeader(ImportPackageHeader importPackageHeader);

    /**
     * Get the bundle's Require-Bundle header
     * 
     * @return the bundle's Require-Bundle header
     */
    RequireBundleHeader getRequireBundleHeader();

    /**
     * Set the bundle's Require-Bundle header
     * 
     * @param requireBundleHeader bundle's Require-Bundle header
     */
    void setRequireBundleHeader(RequireBundleHeader requireBundleHeader);

    /**
     * Get the bundle's Import-Library header
     * 
     * @return the bundle's Import-Library header
     */
    ImportLibraryHeader getImportLibraryHeader();

    /**
     * Set the bundle's Import-Library header
     * 
     * @param importLibraryHeader the bundle's Import-Library header
     */
    void setImportLibraryHeader(ImportLibraryHeader importLibraryHeader);

    /**
     * Returns this manifest's <code>Import-Bundle</code> header or <code>null</code> if no such header is present in
     * the manifest.
     * 
     * @return the manifest's <code>Import-Bundle</code> header.
     */
    ImportBundleHeader getImportBundleHeader();

    /**
     * Set the bundle's Import-Bundle header
     * 
     * @param importBundleHeader the bundle's Import-Bundle header
     */
    void setImportBundleHeader(ImportBundleHeader importBundleHeader);

    /**
     * Get the bundle manifest version.
     * 
     * @return the bundle manifest version
     */
    int getBundleManifestVersion();

    /**
     * Set the bundle manifest version.
     * 
     * @param bundleManifestVersion the bundle manifest version to be set
     */
    void setBundleManifestVersion(int bundleManifestVersion);

    /**
     * Converts this <code>BundleManifest</code> to a raw {@link Manifest}.
     */
    Manifest toManifest();

    /**
     * Write the possibly updated bundle manifest to the given writer.
     * 
     * @param writer the writer to which the bundle manifest will be written
     */
    void write(Writer writer);

    /**
     * Compare the significant output values of two bundle manifests. Note that this differs from
     * {@link #equals(Object)} because it considers only the significant main section headers and completely ignores
     * named sections.
     */
    boolean isEquivalent(BundleManifest obj);

    /**
     * Set the manifest version to the given value.
     * 
     * @param string the new version string
     */
    void setManifestVersion(String string);

    /**
     * Get the manifest version number.
     * 
     * @return the manifest version number
     */
    String getManifestVersion();

    /**
     * Set the module scope to the given value.
     * 
     * @param scope the new module scope string
     */
    void setModuleScope(String scope);

    /**
     * Get the module scope.
     * 
     * @return the module scope
     */
    String getModuleScope();

    /**
     * Distribute parameters across semicolon separated packages in import-package, dynamicimport-package, and
     * export-package.
     */
    void expandPackageLists();

    /**
     * Indicates whether or not a fragment bundle should be auto-attached.
     * 
     * @return <code>true</code> if this fragment should be auto-attached, otherwise <code>false</code>.
     */
    boolean isFragmentAutoAttach();

    /**
     * Get the manifest's <code>Bundle-ActivationPolicy</code>. If no <code>Bundle-ActivationPolicy</code> is specified
     * the default of <code>null</code> is returned.
     * 
     * @return the manifest's <code>Bundle-ActivationPolicy</code>.
     */
    BundleActivationPolicy getBundleActivationPolicy();

    /**
     * Sets the <code>Bundle-ActivationPolicy</code> to the given value.
     * 
     * @param bundleActivationPolicy the new <code>Bundle-ActivationPolicy</code>.
     */
    void setBundleActivationPolicy(BundleActivationPolicy bundleActivationPolicy);
}
