/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.manifest;

import java.util.jar.Attributes;

import com.springsource.server.osgi.manifest.BundleManifest.ImportBundleHeader;


/**
 * A {@link Delegate} implementation for handling import bundle
 * <p />
 *
 * <strong>Concurrent Semantics</strong><br />
 * This class is <strong>not</strong thread-safe.
 * 
 * @author Andy Wilkinson
 * @since 1.0
 */
public class ImportBundleDelegate implements Delegate {
    
    private static final String IMPORT_BUNDLE_HEADER_NAME = "Import-Bundle";
    private final Attributes mainAttributes;
    private ImportBundleHeader importBundleHeader;
    
    ImportBundleDelegate(Attributes mainAttributes, BundleManifestFactory bundleManifestFactory) {
        this.mainAttributes = mainAttributes;
        String creationValue = this.mainAttributes.getValue(IMPORT_BUNDLE_HEADER_NAME);
        if (creationValue != null) {
            this.importBundleHeader = bundleManifestFactory.parseImportBundleHeader(creationValue);
        } else {
            this.importBundleHeader = null;
        }
    }
    
    ImportBundleHeader getImportBundleHeader() {
        return this.importBundleHeader;
    }
    
    void setImportBundleHeader(ImportBundleHeader importBundleHeader) {
        this.importBundleHeader = importBundleHeader;
    }
    
       
    /** 
     * {@inheritDoc}
     */
    public void checkHeaderValidity() {
        if (this.importBundleHeader != null) {
            this.importBundleHeader.checkWellFormed();
        }
    }

    /**
     * @param outputAttributes
     */
    void write(Attributes outputAttributes) {
        String value = null;
        if (this.importBundleHeader != null) {
            value = this.importBundleHeader.toString();
        }
        if (value != null) {
            outputAttributes.putValue(IMPORT_BUNDLE_HEADER_NAME, value);
        } else {
            outputAttributes.remove(new Attributes.Name(IMPORT_BUNDLE_HEADER_NAME));
        }
    }
    
    /**
     * {@inheritDoc}
     */
    @Override public int hashCode() {
        throw new UnsupportedOperationException();
    }

    @Override public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        ImportBundleDelegate other = (ImportBundleDelegate) obj;
        if (this.importBundleHeader == null) {
            if (other.importBundleHeader != null) {
                return false;
            }
        } else if (!this.importBundleHeader.equals(other.importBundleHeader)) {
            return false;
        }
        return true;
    }
}
