/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.manifest;

import org.springframework.util.ObjectUtils;
import org.springframework.util.StringUtils;

import com.springsource.server.osgi.manifest.BundleManifest.ImportBundleHeader.BundleImport;
import com.springsource.server.osgi.manifest.parse.FatalParseException;
import com.springsource.server.osgi.manifest.parse.HeaderDeclaration;

/**
 * Standard implementation of <code>BundleImport</code>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * This class is <strong>thread-safe<strong>.
 * 
 * @author Andy Wilkinson
 * @since 1.0
 */
public class StandardBundleImport extends AbstractImport implements BundleImport {

    private final String bundleSymbolicName;

    private final Resolution resolution;

    private final Sharing sharing;

    private final String versionRange;

    private final boolean applicationImportScope;
    
    public StandardBundleImport(String bundleSymbolicName) {
        this.bundleSymbolicName = bundleSymbolicName;
        this.versionRange = getDefaultedVersionRange(null);
        this.resolution = Resolution.MANDATORY;
        this.sharing = Sharing.AUTOMATIC;
        this.applicationImportScope = false;
    }

    public StandardBundleImport(String bundleSymbolicName, String versionRange, String resolutionString, String importScope) {
        this.bundleSymbolicName = bundleSymbolicName;
        this.versionRange = getDefaultedVersionRange(versionRange);
        this.resolution = this.resolveResolutionString(resolutionString);
        this.sharing = Sharing.AUTOMATIC;
        this.applicationImportScope = IMPORT_SCOPE_APPLICATION.equals(importScope);
    }

    StandardBundleImport(HeaderDeclaration declaration) {
        this.bundleSymbolicName = declaration.getNames().get(0);

        this.resolution = this.resolveResolutionString(declaration.getDirectives().get(RESOLUTION_DIRECTIVE_NAME));

        this.versionRange = getDefaultedVersionRange(declaration.getAttributes().get(ATTRIBUTE_VERSION));

        String sharingString = declaration.getDirectives().get(SHARING_DIRECTIVE_NAME);
        if (ObjectUtils.nullSafeEquals(SHARING_SHARE, sharingString)) {
            this.sharing = Sharing.SHARE;
        } else if (ObjectUtils.nullSafeEquals(SHARING_CLONE, sharingString)) {
            this.sharing = Sharing.CLONE;
        } else if (sharingString == null || ObjectUtils.nullSafeEquals(SHARING_AUTOMATIC, sharingString)) {
            this.sharing = Sharing.AUTOMATIC;
        } else {
            throw new FatalParseException("Import-Library directive " + SHARING_DIRECTIVE_NAME + ":=" + sharingString + " does not parse.");
        }

        String promoteString = declaration.getDirectives().get(IMPORT_SCOPE_DIRECTIVE_NAME);
        if (ObjectUtils.nullSafeEquals(IMPORT_SCOPE_APPLICATION, promoteString)) {
            this.applicationImportScope = true;
        } else if (ObjectUtils.nullSafeEquals(IMPORT_SCOPE_BUNDLE, promoteString)) {
            this.applicationImportScope = false;
        } else if (promoteString == null) {
            this.applicationImportScope = false;
        } else {
            throw new FatalParseException("Import-Bundle directive " + IMPORT_SCOPE_DIRECTIVE_NAME + ":=" + promoteString + " does not parse.");
        }

        this.attributes.putAll(declaration.getAttributes());
    }

    /**
     * {@inheritDoc}
     */
    public String getBundleSymbolicName() {
        return this.bundleSymbolicName;
    }

    /**
     * {@inheritDoc}
     */
    public Resolution getResolution() {
        return this.resolution;
    }

    /**
     * {@inheritDoc}
     */
    public String getVersionRange() {
        return this.versionRange;
    }

    /**
     * {@inheritDoc}
     */
    public Sharing getSharing() {
        return this.sharing;
    }

    /**
     * {@inheritDoc}
     */
    public boolean getApplicationImportScope() {
        return this.applicationImportScope;
    }

    public void checkWellFormed() {
        if (!StringUtils.hasText(this.bundleSymbolicName)) {
            throw new FatalParseException("Import-Bundle with an empty name is not valid.");
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override public String toString() {
        StringBuffer result = new StringBuffer(20);
        result.append(this.bundleSymbolicName);
        toStringOfAttributes(result);
        toStringOfDirectives(result);
        return result.toString();

    }

    private Resolution resolveResolutionString(String resolutionString) {
        Resolution resolution = null;
        if (ObjectUtils.nullSafeEquals(RESOLUTION_OPTIONAL, resolutionString)) {
            resolution = Resolution.OPTIONAL;
        } else if (resolutionString == null || ObjectUtils.nullSafeEquals(RESOLUTION_MANDATORY, resolutionString)) {
            resolution = Resolution.MANDATORY;
        } else {
            throw new FatalParseException("Import-Bundle directive " + RESOLUTION_DIRECTIVE_NAME + ":=" + resolutionString + " does not parse.");
        }
        return resolution;
    }

    private void toStringOfAttributes(StringBuffer result) {
        for (String attrName : this.attributes.keySet()) {
            result.append(ATTRIBUTE_SEPARATOR).append(attrName).append(EQUALS).append("\"").append(this.attributes.get(attrName)).append("\"");
        }
    }

    private void toStringOfDirectives(StringBuffer result) {
        if (this.resolution == Resolution.OPTIONAL) {
            result.append(ATTRIBUTE_SEPARATOR).append(RESOLUTION_DIRECTIVE_NAME).append(DIRECTIVE).append(RESOLUTION_OPTIONAL);
        }
        if (this.sharing == Sharing.CLONE) {
            result.append(ATTRIBUTE_SEPARATOR).append(SHARING_DIRECTIVE_NAME).append(DIRECTIVE).append(SHARING_CLONE);
        } else if (this.sharing == Sharing.SHARE) {
            result.append(ATTRIBUTE_SEPARATOR).append(SHARING_DIRECTIVE_NAME).append(DIRECTIVE).append(SHARING_SHARE);
        }
        if (this.applicationImportScope) {
            result.append(ATTRIBUTE_SEPARATOR).append(IMPORT_SCOPE_DIRECTIVE_NAME).append(DIRECTIVE).append(IMPORT_SCOPE_APPLICATION);
        }
    }
    
    @Override public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (!(obj instanceof StandardBundleImport)) {
            return false;
        }
        final StandardBundleImport other = (StandardBundleImport) obj;
        if (this.attributes == null) {
            if (other.attributes != null) {
                return false;
            }
        } else if (!this.attributes.equals(other.attributes)) {
            return false;
        }
        if (this.bundleSymbolicName == null) {
            if (other.bundleSymbolicName != null) {
                return false;
            }
        } else if (!this.bundleSymbolicName.equals(other.bundleSymbolicName)) {
            return false;
        }
        if (this.resolution == null) {
            if (other.resolution != null) {
                return false;
            }
        } else if (!this.resolution.equals(other.resolution)) {
            return false;
        }
        if (this.versionRange == null) {
            if (other.versionRange != null) {
                return false;
            }
        } else if (!this.versionRange.equals(other.versionRange)) {
            return false;
        }
        if (this.sharing == null) {
            if (other.sharing != null) {
                return false;
            }
        } else if (!this.sharing.equals(other.sharing)) {
            return false;
        }
        if (this.applicationImportScope != other.applicationImportScope) {
            return false;
        }
        return true;
    }
    
    /**
     * {@inheritDoc}
     */
    @Override public int hashCode() {
        throw new UnsupportedOperationException();
    }
}
