/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.manifest;

import java.util.Map;

import com.springsource.server.osgi.manifest.parse.FatalParseException;
import com.springsource.server.osgi.manifest.parse.HeaderDeclaration;
import com.springsource.server.osgi.manifest.parse.HeaderParser;
import com.springsource.server.osgi.manifest.parse.HeaderParserFactory;
import com.springsource.server.osgi.manifest.parse.ParserLogger;

/**
 * <strong>Concurrent Semantics</strong><br />
 * 
 * This class is not thread safe.
 * 
 * @author Glyn Normington
 * @since 1.0
 */
public class StandardFragmentHostHeader implements BundleManifest.FragmentHostHeader {
	
    private static final String QUOTE = "\"";

    private static final String EXTENSION_DIRECTIVE_NAME = "extension";

    private static final String EXTENSION_FRAMEWORK = "framework";

    private static final String EXTENSION_BOOT_CLASS_PATH = "bootclasspath";

    private static final String ATTRIBUTE_SEPARATOR = ";";

    private static final String EQUALS = "=";

    private static final String DIRECTIVE = ":=";

    public String symbolicName;

    private Map<String, String> attributes;

    private Extension extension;

    public StandardFragmentHostHeader(ParserLogger parserLogger, String fragmentHostHeaderString) {
        HeaderParser headerParser = HeaderParserFactory.newHeaderParser(parserLogger);
        HeaderDeclaration fragmentHostDeclaration = headerParser.parseFragmentHostHeader(fragmentHostHeaderString);
        this.symbolicName = fragmentHostDeclaration.getNames().get(0);
        String extensionString = fragmentHostDeclaration.getDirectives().get(EXTENSION_DIRECTIVE_NAME);
        this.attributes = fragmentHostDeclaration.getAttributes();
        if (extensionString != null) {
            if (extensionString.equals(EXTENSION_FRAMEWORK)) {
                this.extension = Extension.FRAMEWORK;
            } else if (extensionString.equals(EXTENSION_BOOT_CLASS_PATH)) {
                this.extension = Extension.BOOTCLASSPATH;
            }
        }
    }

    public StandardFragmentHostHeader(String symbolicBundleHost) {
        this.symbolicName = symbolicBundleHost;
    }

    public Map<String, String> getAttributes() {
        return this.attributes;
    }
    
    public void setAttributes(Map<String, String> attributes) {
		this.attributes = attributes;
	}

    public String getBundleSymbolicName() {
        return this.symbolicName;
    }

    public void setBundleSymbolicName(String hostName) {
        this.symbolicName = hostName;
    }

    public Extension getExtension() {
        return this.extension;
    }

    public void setExtension(Extension extension) {
        this.extension = extension;
    }

    public void checkWellFormed() {
        if (this.symbolicName == null) {
            throw new FatalParseException("A Bundle Fragment can not exist without a host");
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override public int hashCode() {
        throw new UnsupportedOperationException();
    }

    /**
     * {@inheritDoc}
     */
    @Override public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (!(obj instanceof StandardFragmentHostHeader)) {
            return false;
        }
        final StandardFragmentHostHeader other = (StandardFragmentHostHeader) obj;
        if (this.attributes == null) {
            if (other.attributes != null) {
                return false;
            }
        } else if (!this.attributes.equals(other.attributes)) {
            return false;
        }
        if (this.extension == null) {
            if (other.extension != null) {
                return false;
            }
        } else if (!this.extension.equals(other.extension)) {
            return false;
        }
        if (this.symbolicName == null) {
            if (other.symbolicName != null) {
                return false;
            }
        } else if (!this.symbolicName.equals(other.symbolicName)) {
            return false;
        }
        return true;
    }

    /**
     * {@inheritDoc}
     */
    @Override public String toString() {
        if (this.symbolicName != null && !this.symbolicName.equals("")) {
            StringBuffer result = new StringBuffer(20);
            result.append(this.symbolicName);
            toStringOfAttributes(result);
            toStringOfDirectives(result);
            return result.toString();
        } else {
            return null;
        }

    }

    private void toStringOfAttributes(StringBuffer result) {
        for (String attrName : this.attributes.keySet()) {
            result.append(ATTRIBUTE_SEPARATOR).append(attrName).append(EQUALS).append(QUOTE).append(this.attributes.get(attrName)).append(QUOTE);
        }
    }

    private void toStringOfDirectives(StringBuffer result) {
        if (this.extension == Extension.FRAMEWORK) {
            result.append(ATTRIBUTE_SEPARATOR).append(EXTENSION_DIRECTIVE_NAME).append(DIRECTIVE).append(EXTENSION_FRAMEWORK);
        }
        if (this.extension == Extension.BOOTCLASSPATH) {
            result.append(ATTRIBUTE_SEPARATOR).append(EXTENSION_DIRECTIVE_NAME).append(DIRECTIVE).append(EXTENSION_BOOT_CLASS_PATH);
        }
    }

}
