/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.manifest;

import java.util.ArrayList;
import java.util.List;

import org.springframework.util.Assert;

import com.springsource.server.osgi.manifest.BundleManifest.ImportBundleHeader;
import com.springsource.server.osgi.manifest.parse.HeaderDeclaration;
import com.springsource.server.osgi.manifest.parse.HeaderParser;
import com.springsource.server.osgi.manifest.parse.HeaderParserFactory;
import com.springsource.server.osgi.manifest.parse.ParseUtils;
import com.springsource.server.osgi.manifest.parse.ParserLogger;


/**
 * Standard implementation of <code>ImportBundleHeader</code>
 * <p />
 *
 * <strong>Concurrent Semantics</strong><br />
 *
 * This class is <strong>not<strong> thread-safe.
 *
 * @author Andy Wilkinson
 * @since 1.0
 */
final class StandardImportBundleHeader implements ImportBundleHeader {

     private List<BundleImport> importedBundles;
     
     private static final String SEPARATOR = ",";
     
     StandardImportBundleHeader(ParserLogger parserLogger, String creationValue) {
         importedBundles = new ArrayList<BundleImport>();
         HeaderParser headerParser = HeaderParserFactory.newHeaderParser(parserLogger);
         List<HeaderDeclaration> headerDeclarations = headerParser.parseImportBundleHeader(creationValue);
         for (HeaderDeclaration declaration : headerDeclarations) {
             importedBundles.add(new StandardBundleImport(declaration));
         }
     }
     
     public StandardImportBundleHeader(List<BundleImport> bundleImports) {
         this.importedBundles = bundleImports;
     }
     
    /** 
     * {@inheritDoc}
     */
    public List<BundleImport> getImportedBundles() {
        return importedBundles;
    }

    /** 
     * {@inheritDoc}
     */
    public void setImportedBundles(List<BundleImport> bundles) {
        Assert.notNull(bundles);
        this.importedBundles = bundles;      
    }
    
    /**
     * {@inheritDoc}
     */
    public void checkWellFormed() {
        for (BundleImport bundle : this.importedBundles) {
            bundle.checkWellFormed();
        }
    }
    
    /**
     * {@inheritDoc}
     */
    @Override public String toString() {
        if (this.importedBundles.size() != 0) {
            StringBuffer result = new StringBuffer(20);
            result.append(ParseUtils.collectionToDelimitedString(this.importedBundles, SEPARATOR));
            return result.toString();
        } else {
            return null;
        }
    }
    
    /**
     * {@inheritDoc}
     */
    @Override public int hashCode() {
        throw new UnsupportedOperationException();
    }
    
    /**
     * {@inheritDoc}
     */
    @Override public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (!(obj instanceof StandardImportBundleHeader)) {
            return false;
        }
        final StandardImportBundleHeader other = (StandardImportBundleHeader) obj;
        if (!this.importedBundles.equals(other.importedBundles)) {
            return false;
        }
        return true;
    }
}
