/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.manifest;

import java.util.Map;

import org.springframework.util.ObjectUtils;
import org.springframework.util.StringUtils;

import com.springsource.server.osgi.manifest.BundleManifest.ImportLibraryHeader.LibraryImport;
import com.springsource.server.osgi.manifest.parse.FatalParseException;
import com.springsource.server.osgi.manifest.parse.HeaderDeclaration;

/**
 * Standard implementation of <code>LibraryImport</code>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * This class is <strong>thread-safe<strong>.
 * 
 * @author Rob Harrop
 * @author Andy Wilkinson
 * @since 1.0
 */
final class StandardLibraryImport extends AbstractImport implements LibraryImport {

    private final String librarySymbolicName;

    private final Resolution resolution;

    private final Sharing sharing;

    private final String versionRange;

    public StandardLibraryImport(String librarySymbolicName) {
        this.librarySymbolicName = librarySymbolicName;
        this.resolution = Resolution.MANDATORY;
        this.sharing = Sharing.AUTOMATIC;
        this.versionRange = getDefaultedVersionRange(null);
    }

    public StandardLibraryImport(HeaderDeclaration declaration) {
        this.librarySymbolicName = declaration.getNames().get(0);
        this.attributes.putAll(declaration.getAttributes());

        String resolutionString = declaration.getDirectives().get(RESOLUTION_DIRECTIVE_NAME);
        if (ObjectUtils.nullSafeEquals(RESOLUTION_OPTIONAL, resolutionString)) {
            this.resolution = Resolution.OPTIONAL;
        } else if (resolutionString == null || ObjectUtils.nullSafeEquals(RESOLUTION_MANDATORY, resolutionString)) {
            this.resolution = Resolution.MANDATORY;
        } else {
            throw new FatalParseException("Import-Library directive " + RESOLUTION_DIRECTIVE_NAME + ":=" + resolutionString + " does not parse.");
        }

        String sharingString = declaration.getDirectives().get(SHARING_DIRECTIVE_NAME);
        if (ObjectUtils.nullSafeEquals(SHARING_SHARE, sharingString)) {
            this.sharing = Sharing.SHARE;
        } else if (ObjectUtils.nullSafeEquals(SHARING_CLONE, sharingString)) {
            this.sharing = Sharing.CLONE;
        } else if (sharingString == null || ObjectUtils.nullSafeEquals(SHARING_AUTOMATIC, sharingString)) {
            this.sharing = Sharing.AUTOMATIC;
        } else {
            throw new FatalParseException("Import-Library directive " + SHARING_DIRECTIVE_NAME + ":=" + sharingString + " does not parse.");
        }

        this.versionRange = getDefaultedVersionRange(declaration.getAttributes().get(ATTRIBUTE_VERSION));
    }

    /**
     * {@inheritDoc}
     */
    public Map<String, String> getAttributes() {
        return this.attributes;
    }

    /**
     * {@inheritDoc}
     */
    public String getLibraryName() {
        return this.librarySymbolicName;
    }

    /**
     * {@inheritDoc}
     */
    public String getVersionRange() {
        return this.versionRange;
    }

    /**
     * 
     * {@inheritDoc}
     */
    public Resolution getResolution() {
        return this.resolution;
    }

    /**
     * {@inheritDoc}
     */
    public void checkWellFormed() {
        if (!StringUtils.hasText(this.librarySymbolicName)) {
            throw new FatalParseException("Import-Library with an empty name is not valid.");
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override public String toString() {
        StringBuffer result = new StringBuffer(20);
        result.append(this.librarySymbolicName);
        toStringOfAttributes(result);
        toStringOfDirectives(result);
        return result.toString();

    }

    private void toStringOfAttributes(StringBuffer result) {
        for (String attrName : this.attributes.keySet()) {
            result.append(ATTRIBUTE_SEPARATOR).append(attrName).append(EQUALS).append("\"").append(this.attributes.get(attrName)).append("\"");
        }
    }

    private void toStringOfDirectives(StringBuffer result) {
        if (this.resolution == Resolution.OPTIONAL) {
            result.append(ATTRIBUTE_SEPARATOR).append(RESOLUTION_DIRECTIVE_NAME).append(DIRECTIVE).append(RESOLUTION_OPTIONAL);
        }
        if (this.sharing == Sharing.CLONE) {
            result.append(ATTRIBUTE_SEPARATOR).append(SHARING_DIRECTIVE_NAME).append(DIRECTIVE).append(SHARING_CLONE);
        } else if (this.sharing == Sharing.SHARE) {
            result.append(ATTRIBUTE_SEPARATOR).append(SHARING_DIRECTIVE_NAME).append(DIRECTIVE).append(SHARING_SHARE);
        }
    }

    /**
     * {@inheritDoc}
     */
    public Sharing getSharing() {
        return this.sharing;
    }
    
    @Override public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (!(obj instanceof StandardLibraryImport)) {
            return false;
        }
        final StandardLibraryImport other = (StandardLibraryImport) obj;
        if (this.attributes == null) {
            if (other.attributes != null) {
                return false;
            }
        } else if (!this.attributes.equals(other.attributes)) {
            return false;
        }
        if (this.librarySymbolicName == null) {
            if (other.librarySymbolicName != null) {
                return false;
            }
        } else if (!this.librarySymbolicName.equals(other.librarySymbolicName)) {
            return false;
        }
        if (this.resolution == null) {
            if (other.resolution != null) {
                return false;
            }
        } else if (!this.resolution.equals(other.resolution)) {
            return false;
        }
        if (this.versionRange == null) {
            if (other.versionRange != null) {
                return false;
            }
        } else if (!this.versionRange.equals(other.versionRange)) {
            return false;
        }
        if (this.sharing == null) {
            if (other.sharing != null) {
                return false;
            }
        } else if (!this.sharing.equals(other.sharing)) {
            return false;
        }
        return true;
    }
    
    /**
     * {@inheritDoc}
     */
    @Override public int hashCode() {
        throw new UnsupportedOperationException();
    }
}
