/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.manifest;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

import org.osgi.framework.Version;

import com.springsource.server.osgi.manifest.BundleManifest.ExportPackageHeader.PackageExport;
import com.springsource.server.osgi.manifest.parse.FatalParseException;
import com.springsource.server.osgi.manifest.parse.HeaderDeclaration;
import com.springsource.server.osgi.manifest.parse.ParseUtils;

/**
 * 
 * TODO Document StandardPackageExport
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * TODO Document concurrent semantics of StandardPackageExport
 * 
 * @author Christopher Frost
 * @since 1.0
 */
public class StandardPackageExport implements PackageExport {
    
    private static final String VERSION_ATTRIBUTE_NAME = "version";
    
    private static final String VERSION_ATTRIBUTE_ALTERNATE_NAME = "specification-version";

    private static final String INCLUDE_DIRECTIVE_NAME = "include";

    private static final String EXCLUDE_DIRECTIVE_NAME = "exclude";

    private static final String MANDATORIES_DIRECTIVE_NAME = "mandatory";

    private static final String USES_DIRECTIVE_NAME = "uses";

    private static final String PACKAGE_SEPARATOR = ";";

    private static final String ATTRIBUTE_SEPARATOR = ";";

    private static final String DIRECTIVE_SEPARATOR = ";";

    private static final String EQUALS = "=";

    private static final String DIRECTIVE = ":=";

    private static final String DIRECTIVE_ITEM_SEPARATOR = ",";

    private static final String QUOTE = "\"";

    private final List<String> packageNames = new ArrayList<String>(1);

    private final Map<String, String> attributes = new HashMap<String, String>();

    // Directives

    private List<String> excludes = new ArrayList<String>();

    private List<String> includes = new ArrayList<String>();

    private List<String> mandatories = new ArrayList<String>();

    private List<String> uses = new ArrayList<String>();

    public StandardPackageExport(HeaderDeclaration packageDeclaration) {
        this.packageNames.addAll(packageDeclaration.getNames());
        this.attributes.putAll(packageDeclaration.getAttributes());

        String excludeString = packageDeclaration.getDirectives().get(EXCLUDE_DIRECTIVE_NAME);
        if (excludeString != null) {
            this.excludes = generateList(excludeString);
        }

        String includeString = packageDeclaration.getDirectives().get(INCLUDE_DIRECTIVE_NAME);
        if (includeString != null) {
            this.includes = generateList(includeString);
        }

        String mandatoriesString = packageDeclaration.getDirectives().get(MANDATORIES_DIRECTIVE_NAME);
        if (mandatoriesString != null) {
            this.mandatories = generateList(mandatoriesString);
        }

        String usesString = packageDeclaration.getDirectives().get(USES_DIRECTIVE_NAME);
        if (usesString != null) {
            this.uses = generateList(usesString);
        }
    }

    public StandardPackageExport() {
    }

    public Map<String, String> getAttributes() {
        return this.attributes;
    }

    public List<String> getPackageNames() {
        return this.packageNames;
    }

    public List<String> getExclude() {
        return this.excludes;
    }

    public List<String> getInclude() {
        return this.includes;
    }

    public List<String> getMandatory() {
        return this.mandatories;
    }

    public List<String> getUses() {
        return this.uses;
    }

    private List<String> generateList(String commaSeperated) {
        List<String> list = new ArrayList<String>();

        StringTokenizer tokenizer = new StringTokenizer(commaSeperated, ",");
        while (tokenizer.hasMoreTokens()) {
            list.add(tokenizer.nextToken());
        }

        return list;
    }

    /**
     * {@inheritDoc}
     */
    @Override public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result + (this.attributes == null ? 0 : this.attributes.hashCode());
        result = prime * result + (this.excludes == null ? 0 : this.excludes.hashCode());
        result = prime * result + (this.includes == null ? 0 : this.includes.hashCode());
        result = prime * result + (this.mandatories == null ? 0 : this.mandatories.hashCode());
        result = prime * result + (this.packageNames == null ? 0 : this.packageNames.hashCode());
        result = prime * result + (this.uses == null ? 0 : this.uses.hashCode());
        return result;
    }

    /**
     * {@inheritDoc}
     */
    @Override public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (!(obj instanceof StandardPackageExport)) {
            return false;
        }
        final StandardPackageExport other = (StandardPackageExport) obj;
        if (this.attributes == null) {
            if (other.attributes != null) {
                return false;
            }
        } else if (!this.attributes.equals(other.attributes)) {
            return false;
        }
        if (this.excludes == null) {
            if (other.excludes != null) {
                return false;
            }
        } else if (!this.excludes.equals(other.excludes)) {
            return false;
        }
        if (this.includes == null) {
            if (other.includes != null) {
                return false;
            }
        } else if (!this.includes.equals(other.includes)) {
            return false;
        }
        if (this.mandatories == null) {
            if (other.mandatories != null) {
                return false;
            }
        } else if (!this.mandatories.equals(other.mandatories)) {
            return false;
        }
        if (this.packageNames == null) {
            if (other.packageNames != null) {
                return false;
            }
        } else if (!this.packageNames.equals(other.packageNames)) {
            return false;
        }
        if (this.uses == null) {
            if (other.uses != null) {
                return false;
            }
        } else if (!this.uses.equals(other.uses)) {
            return false;
        }
        return true;
    }

    /**
     * {@inheritDoc}
     */
    @Override public String toString() {
        StringBuffer result = new StringBuffer(20);
        result.append(ParseUtils.collectionToDelimitedString(this.packageNames, PACKAGE_SEPARATOR));
        toStringOfAttributes(result);
        toStringOfDirectives(result);
        return result.toString();
    }

    private void toStringOfAttributes(StringBuffer result) {
        for (String attrName : this.attributes.keySet()) {
            result.append(ATTRIBUTE_SEPARATOR).append(attrName).append(EQUALS).append(QUOTE).append(this.attributes.get(attrName)).append(QUOTE);
        }
    }

    private void toStringOfDirectives(StringBuffer result) {
        if (this.excludes.size() > 0) {
            result.append(DIRECTIVE_SEPARATOR).append(EXCLUDE_DIRECTIVE_NAME).append(DIRECTIVE).append(QUOTE);
            result.append(ParseUtils.collectionToDelimitedString(this.excludes, DIRECTIVE_ITEM_SEPARATOR)).append(QUOTE);
        }
        if (this.includes.size() > 0) {
            result.append(DIRECTIVE_SEPARATOR).append(INCLUDE_DIRECTIVE_NAME).append(DIRECTIVE).append(QUOTE);
            result.append(ParseUtils.collectionToDelimitedString(this.includes, DIRECTIVE_ITEM_SEPARATOR)).append(QUOTE);
        }
        if (this.mandatories.size() > 0) {
            result.append(DIRECTIVE_SEPARATOR).append(MANDATORIES_DIRECTIVE_NAME).append(DIRECTIVE).append(QUOTE);
            result.append(ParseUtils.collectionToDelimitedString(this.mandatories, DIRECTIVE_ITEM_SEPARATOR)).append(QUOTE);
        }
        if (this.uses.size() > 0) {
            result.append(DIRECTIVE_SEPARATOR).append(USES_DIRECTIVE_NAME).append(DIRECTIVE).append(QUOTE);
            result.append(ParseUtils.collectionToDelimitedString(this.uses, DIRECTIVE_ITEM_SEPARATOR)).append(QUOTE);
        }
    }

    /**
     * 
     * {@inheritDoc}
     */
    public void checkWellFormed() {
        if (this.packageNames.size() == 0) {
            throw new FatalParseException("Package export with no packages is not well formed");
        }
    }

    /** 
     * {@inheritDoc}
     */
    public Version getVersion() {
        String version = this.attributes.get(VERSION_ATTRIBUTE_NAME);
        if (version == null) {
            version = this.attributes.get(VERSION_ATTRIBUTE_ALTERNATE_NAME);
        }
        return version != null ? new Version(version) : Version.emptyVersion;
    }
}
