/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.manifest;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import com.springsource.server.osgi.manifest.parse.FatalParseException;
import com.springsource.server.osgi.manifest.parse.HeaderDeclaration;
import com.springsource.server.osgi.manifest.parse.ParseUtils;

/**
 * <strong>Concurrent Semantics</strong><br />
 * 
 * This class is not thread safe.
 * 
 * @author Glyn Normington
 * @since 1.0
 */
class StandardPackageImport implements BundleManifest.ImportPackageHeader.PackageImport {

    private static final String QUOTE = "\"";

    private static final String NO_PACKAGES_SPECIFIED = "**no packages specified**";

    private static final String RESOLUTION_OPTIONAL = "optional";

    private static final String RESOLUTION_DIRECTIVE_NAME = "resolution";

    private static final String PACKAGE_SEPARATOR = ";";

    private static final String ATTRIBUTE_SEPARATOR = ";";

    private static final String EQUALS = "=";

    private static final String DIRECTIVE = ":=";

    private List<String> packageNames = new ArrayList<String>(1);

    private final Map<String, String> attributes = new HashMap<String, String>();

    private Resolution resolution = Resolution.MANDATORY;

    StandardPackageImport(HeaderDeclaration packageDeclaration) {
        this.packageNames.addAll(packageDeclaration.getNames());
        this.attributes.putAll(packageDeclaration.getAttributes());

        String resolutionString = packageDeclaration.getDirectives().get(RESOLUTION_DIRECTIVE_NAME);
        if (resolutionString != null && resolutionString.equals(RESOLUTION_OPTIONAL)) {
            this.resolution = Resolution.OPTIONAL;
        }
    }

    StandardPackageImport() {
    }

    /**
     * {@inheritDoc}
     */
    public Map<String, String> getAttributes() {
        return this.attributes;
    }

    /**
     * {@inheritDoc}
     */
    public List<String> getPackageNames() {
        return this.packageNames;
    }

    /**
     * {@inheritDoc}
     */
    public void setPackageNames(List<String> packageNames) {
        this.packageNames = packageNames;

    }

    /**
     * {@inheritDoc}
     */
    public Resolution getResolution() {
        return this.resolution;
    }

    /**
     * {@inheritDoc}
     */
    public void setResolution(Resolution resolution) {
        this.resolution = resolution;
    }

    /**
     * {@inheritDoc}
     */
    @Override public int hashCode() {
        throw new UnsupportedOperationException();
    }

    /**
     * {@inheritDoc}
     */
    @Override public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (!(obj instanceof StandardPackageImport)) {
            return false;
        }
        final StandardPackageImport other = (StandardPackageImport) obj;
        if (this.attributes == null) {
            if (other.attributes != null) {
                return false;
            }
        } else if (!this.attributes.equals(other.attributes)) {
            return false;
        }
        if (this.packageNames == null) {
            if (other.packageNames != null) {
                return false;
            }
        } else if (!this.packageNames.equals(other.packageNames)) {
            return false;
        }
        if (this.resolution == null) {
            if (other.resolution != null) {
                return false;
            }
        } else if (!this.resolution.equals(other.resolution)) {
            return false;
        }
        return true;
    }

    /**
     * {@inheritDoc}
     */
    @Override public String toString() {
        StringBuffer result = new StringBuffer(20);
        if (this.packageNames.size() != 0) {
            result.append(ParseUtils.collectionToDelimitedString(this.packageNames, PACKAGE_SEPARATOR));
        } else {
            result.append(NO_PACKAGES_SPECIFIED);
        }
        toStringOfAttributes(result);
        toStringOfDirectives(result);
        return result.toString();

    }

    private void toStringOfAttributes(StringBuffer result) {
        for (String attrName : this.attributes.keySet()) {
            result.append(ATTRIBUTE_SEPARATOR).append(attrName).append(EQUALS).append(QUOTE).append(this.attributes.get(attrName)).append(QUOTE);
        }
    }

    private void toStringOfDirectives(StringBuffer result) {
        if (this.resolution == Resolution.OPTIONAL) {
            result.append(ATTRIBUTE_SEPARATOR).append(RESOLUTION_DIRECTIVE_NAME).append(DIRECTIVE).append(RESOLUTION_OPTIONAL);
        }
    }

    public void checkWellFormed() {
        if (this.packageNames.size() == 0) {
            throw new FatalParseException("Package import with no packages is not well formed");
        }

    }
}
