/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.manifest;

import java.util.HashMap;
import java.util.Map;

import com.springsource.server.osgi.manifest.parse.FatalParseException;
import com.springsource.server.osgi.manifest.parse.HeaderDeclaration;

class StandardRequireBundle implements BundleManifest.RequireBundleHeader.RequireBundle {

    private static final String RESOLUTION_OPTIONAL = "optional";

    private static final String RESOLUTION_DIRECTIVE_NAME = "resolution";

    private static final String VISIBILITY_REEXPORT = "reexport";

    private static final String VISIBILITY_DIRECTIVE_NAME = "visibility";
    
    private static final String BUNDLE_VERSION_ATTRIBUTE_NAME = "bundle-version";

    private static final String ATTRIBUTE_SEPARATOR = ";";

    private static final String EQUALS = "=";

    private static final String DIRECTIVE = ":=";

    private String bundleSymbolicName;

    private final Map<String, String> attributes = new HashMap<String, String>();

    private Resolution resolution = Resolution.MANDATORY;

    private Visibility visibility = Visibility.PRIVATE;

    /**
     * 
     * @param packageDeclaration
     */
    public StandardRequireBundle(HeaderDeclaration packageDeclaration) {
        this.bundleSymbolicName = packageDeclaration.getNames().get(0);
        this.attributes.putAll(packageDeclaration.getAttributes());

        String resolutionString = packageDeclaration.getDirectives().get(RESOLUTION_DIRECTIVE_NAME);
        if (resolutionString != null && resolutionString.equals(RESOLUTION_OPTIONAL)) {
            this.resolution = Resolution.OPTIONAL;
        }

        String visibilityString = packageDeclaration.getDirectives().get(VISIBILITY_DIRECTIVE_NAME);
        if (visibilityString != null && visibilityString.equals(VISIBILITY_REEXPORT)) {
            this.visibility = Visibility.REEXPORT;
        }

    }

    public StandardRequireBundle(String bundleSymbolicName) {
        this.bundleSymbolicName = bundleSymbolicName;
    }

    /**
     * 
     * {@inheritDoc}
     */
    public Map<String, String> getAttributes() {
        return this.attributes;
    }

    /**
     * 
     * {@inheritDoc}
     */
    public String getBundleSymbolicName() {
        return this.bundleSymbolicName;
    }

    /**
     * 
     * {@inheritDoc}
     */
    public Resolution getResolution() {
        return this.resolution;
    }

    /**
     * 
     * {@inheritDoc}
     */
    public void setResolution(Resolution resolution) {
        this.resolution = resolution;
    }

    /**
     * 
     * {@inheritDoc}
     */
    public Visibility getVisibility() {
        return this.visibility;
    }

    /**
     * 
     * {@inheritDoc}
     */
    public void setVisibility(Visibility visibility) {
        this.visibility = visibility;
    }

    /**
     * 
     * {@inheritDoc}
     */
    public void checkWellFormed() {
        if (this.bundleSymbolicName == null && this.bundleSymbolicName.length() == 0) {
            throw new FatalParseException("Required bundle with an empty name is not valid");
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override public int hashCode() {
        throw new UnsupportedOperationException();
    }

    /**
     * {@inheritDoc}
     */
    @Override public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (!(obj instanceof StandardRequireBundle)) {
            return false;
        }
        final StandardRequireBundle other = (StandardRequireBundle) obj;

        if (this.attributes == null) {
            if (other.attributes != null) {
                return false;
            }
        } else if (!this.attributes.equals(other.attributes)) {
            return false;
        }

        if (this.bundleSymbolicName == null) {
            if (other.bundleSymbolicName != null) {
                return false;
            }
        } else if (!this.bundleSymbolicName.equals(other.bundleSymbolicName)) {
            return false;
        }

        if (this.resolution == null) {
            if (other.resolution != null) {
                return false;
            }
        } else if (!this.resolution.equals(other.resolution)) {
            return false;
        }

        if (this.visibility == null) {
            if (other.visibility != null) {
                return false;
            }
        } else if (!this.visibility.equals(other.visibility)) {
            return false;
        }

        return true;
    }

    /**
     * {@inheritDoc}
     */
    @Override public String toString() {
        StringBuffer result = new StringBuffer(20);
        result.append(this.bundleSymbolicName);
        toStringOfAttributes(result);
        toStringOfDirectives(result);
        return result.toString();

    }

    private void toStringOfAttributes(StringBuffer result) {
		for (String attrName : this.attributes.keySet()) {
			boolean isBundleVersionAttribute = attrName.equals(BUNDLE_VERSION_ATTRIBUTE_NAME);
			result.append(ATTRIBUTE_SEPARATOR).append(attrName).append(EQUALS);
			if (isBundleVersionAttribute) {
				result.append("\"");
			}
			result.append(this.attributes.get(attrName));
			if (isBundleVersionAttribute) {
				result.append("\"");
			}
		}
	}

    private void toStringOfDirectives(StringBuffer result) {
        if (this.resolution == Resolution.OPTIONAL) {
            result.append(ATTRIBUTE_SEPARATOR).append(RESOLUTION_DIRECTIVE_NAME).append(DIRECTIVE).append(RESOLUTION_OPTIONAL);
        }
        if (this.visibility == Visibility.REEXPORT) {
            result.append(ATTRIBUTE_SEPARATOR).append(VISIBILITY_DIRECTIVE_NAME).append(DIRECTIVE).append(VISIBILITY_REEXPORT);
        }
    }

    /** 
     * {@inheritDoc}
     */
    public void setBundleSymbolicName(String symbolicName) {
        this.bundleSymbolicName = symbolicName;
    }

}
