/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.manifest.parse;

import org.antlr.runtime.RecognitionException;

/**
 * <p>
 * Provides a consistent handling of errors thrown from manifest header parser and 
 * lexer. All exit points should deferrer to this class for error handling.
 * </p>
 *
 * <strong>Concurrent Semantics</strong><br />
 *
 * This class is thread safe
 *
 * @author Christopher Frost
 * @since 1.0
 */
public class ManifestHeaderParserErrorHandler {

    private final ParserLogger logger;
    
    private static final int HEADER_SECTION_LENGTH = 20;

    /**
     * Sole constructor that takes a <code>ParserLogger</code> that will be used to report all messages to.
     * @param logger
     */
    public ManifestHeaderParserErrorHandler(ParserLogger logger) {
        this.logger = logger;
    }

    /**
     * 
     * @param re
     * @param headerName
     * @param insert
     */
    public void exitError(RecognitionException re, String headerName, String header, Object insert){
        this.logger.outputErrorMsg(re, "Unrecoverable " + this.formatMessage(re, headerName, header, insert));
    }
    
    /**
     * 
     * @param re
     * @param headerName
     * @param insert
     */
    public void recoverableError(RecognitionException re, String headerName, String header, Object insert){
        this.logger.outputWarnMsg(re, "Recovered from " + this.formatMessage(re, headerName, header, insert));
    }
    
    /**
     * Helper method to build a well formatted message string
     * 
     * @param re
     * @param headerName
     * @return
     */
    private String formatMessage(RecognitionException re, String headerName, String header, Object insert){
        StringBuilder msg = new StringBuilder();
        msg.append(re.getClass().getSimpleName()).append(" on ").append(headerName);
        String headerSection;
        int charPos = re.charPositionInLine;
        if(charPos >= 0){
            msg.append(" at postion [").append(charPos).append("], ");
            int beginIndex = charPos > HEADER_SECTION_LENGTH ? charPos - HEADER_SECTION_LENGTH : 0;
            int endIndex = header.length() > charPos + HEADER_SECTION_LENGTH ? charPos + HEADER_SECTION_LENGTH : header.length();
            headerSection = header.substring(beginIndex, endIndex);
        }else{
            msg.append(" at an unknown position, ");
            headerSection = header.length() > (HEADER_SECTION_LENGTH * 2) ? header.substring(header.length() - (HEADER_SECTION_LENGTH * 2), header.length()) : header ;
        }
        
        if(insert != null){
            msg.append("additional information [").append(insert.toString()).append("].");
        }else{
            msg.append("no additional information avaliable.");
        }
        
        msg.append(" Error may have occurred near '").append(headerSection).append("'.");
        
        return msg.toString();
    }
    
}
