/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.provisioning.repository;

import java.io.File;
import java.io.IOException;
import java.util.concurrent.atomic.AtomicLong;

import org.springframework.util.AntPathMatcher;

import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * A {@link FileSystemSearcher} implementation that searches based on an Ant-style search pattern. See
 * {@link AntPathMatcher} for more details of the syntax.
 * <p />
 * Files that match according to {@link AntPathMatcher#match(String, String)} are considered to be terminal matches.
 * Files that match according to {@link AntPathMatcher#matchStart(String, String)} are considered to be non-terminal
 * matches.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * This class is <strong>thread-safe</strong>.
 * 
 * @author Andy Wilkinson
 * @since 1.0
 */
class AntPathMatchingFileSystemSearcher implements FileSystemSearcher {

    private final File rootDir;

    private final String antPathPattern;

    private final AntPathMatcher antPathMatcher;
    
    private static final Tracer TRACE = TracerFactory.createTracer(AntPathMatchingFileSystemSearcher.class);

    /**
     * Creates a new instance that will locate files based on the supplied path pattern.
     * Should the pattern be relative, it will be made absolute using the supplied absolute
     * root.
     * @param antPathPattern The Ant path pattern to match against.
     * @param absoluteRoot The path to, if necessary, make the path pattern absolute.
     */
    AntPathMatchingFileSystemSearcher(String antPathPattern, String absoluteRoot) {
        String absoluteAntPathPattern = makeAbsoluteIfRequired(antPathPattern, absoluteRoot);
  
        String[] pathComponents;
        if (File.separator.equals("\\")) {
            pathComponents = absoluteAntPathPattern.split(File.separator + File.separator);
        } else {
            pathComponents = absoluteAntPathPattern.split(File.separator);
        }
        StringBuilder rootPathBuilder = new StringBuilder();
        StringBuilder antPatternBuilder = new StringBuilder();
        boolean rootPathBuilt = false;

        for (String component : pathComponents) {
            if (!rootPathBuilt && !component.contains("*") && !component.contains("?")) {
                rootPathBuilder.append(File.separator);
                rootPathBuilder.append(component);
            } else {
                rootPathBuilt = true;
                antPatternBuilder.append(File.separator);
                antPatternBuilder.append(component);
            }
        }
        try {
            this.rootDir = new File(rootPathBuilder.toString()).getCanonicalFile();
        } catch (IOException ioe) {
            throw new RuntimeException("Unable to determine canonical root directory from '" + rootPathBuilder.toString() + "'", ioe);
        }
        this.antPathMatcher = new AntPathMatcher();
        this.antPathMatcher.setPathSeparator(File.separator);
        this.antPathPattern = this.rootDir.getPath() + antPatternBuilder.toString();
        
        TRACE.traceInfo("Ant path pattern: %s", this.antPathPattern);
    }

    /**
     * {@inheritDoc}
     */
    public long getAgeOfDefinitionSpace() {
        final AtomicLong age = new AtomicLong(this.rootDir.lastModified());
        search(new SearchCallback() {

            public void found(File file, boolean terminal) {
                age.set(Math.max(age.get(), file.lastModified()));
            }

        });
        return age.get();
    }

    /**
     * {@inheritDoc}
     */
    public File getRootDir() {
        return this.rootDir;
    }

    /**
     * {@inheritDoc}
     */
    public void search(SearchCallback callback) {
        search(this.rootDir, callback);
    }

    private void search(File file, SearchCallback callback) {
        String absolutePath = file.getAbsolutePath();
        if (this.antPathMatcher.match(this.antPathPattern, absolutePath)) {
            callback.found(file, true);
        } else if (file.isDirectory() && this.antPathMatcher.matchStart(this.antPathPattern, absolutePath)) {
            callback.found(file, false);
            File[] files = file.listFiles();
            if (files != null) {
                for (File fileInDir : files) {
                    search(fileInDir, callback);
                }
            }
        }
    }
    
    static String makeAbsoluteIfRequired(String antPathPattern, String absoluteRoot) {
        String absolutePathPattern;
        if (!antPathPattern.startsWith("/") && !(antPathPattern.indexOf(":") > 0)) {
            absolutePathPattern = absoluteRoot + File.separator + antPathPattern;
        } else {
            absolutePathPattern = antPathPattern;
        }
        if (File.separator.equals("/")) {
            return absolutePathPattern.replace('\\', '/');
        } else {
            return absolutePathPattern.replace('/', '\\');
        }
    }
}
