/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.provisioning.repository;

/**
 * Factory for creating typical {@link BundleRepository} configuration.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @author Andy Wilkinson
 * @since 1.0
 */
public final class BundleRepositoryFactory {

    /**
     * Creates a {@link BundleRepository} for the supplied search paths.
     * 
     * @param searchPaths the search paths.
     * @param cacheLocationPath the path to the directory to store cache files in.
     * @return the new <code>BundleRepository</code>.
     * @see CompositeBundleRepository
     * @see FileSystemBundleRepository
     */
    public static BundleRepository createBundleRepository(String[] searchPaths, String cacheLocationPath, String absoluteRoot) {
        BundleRepository[] repos = new BundleRepository[searchPaths.length];

        for (int x = 0; x < searchPaths.length; x++) {
            repos[x] = new FileSystemBundleRepository(searchPaths[x], cacheLocationPath, absoluteRoot);
        }
        return new CompositeBundleRepository(repos);
    }

    /**
     * Creates a {@link MutableBundleRepository} for the supplied search paths, installing into <code>installPath</code>.
     * 
     * @param installPath the directory path where new bundles will be installed.
     * @param searchPaths the search paths.
     * @param cacheLocationPath the path to the directory to store cache files in.
     * @return the new <code>BundleRepository</code>.
     * @see CompositeBundleRepository
     * @see FileSystemBundleRepository
     */
    public static MutableBundleRepository createMutableBundleRepository(String installPath, String[] searchPaths, String cacheLocationPath, String absoluteRoot) {
        BundleRepository[] repos = new BundleRepository[searchPaths.length + 1];

        repos[0] = new MutableFileSystemBundleRepository(installPath, cacheLocationPath, absoluteRoot);
        for (int x = 0; x < searchPaths.length; x++) {
            repos[x + 1] = new FileSystemBundleRepository(searchPaths[x], cacheLocationPath, absoluteRoot);
        }
        return new CompositeBundleRepository(repos);
    }
}
