/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.provisioning.repository;

import java.io.File;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.Set;

import org.osgi.framework.Version;

import com.springsource.server.osgi.manifest.BundleManifest;
import com.springsource.server.osgi.manifest.VersionRange;
import com.springsource.server.osgi.manifest.BundleManifest.BundleSymbolicName;
import com.springsource.server.osgi.provisioning.BundleDefinition;
import com.springsource.server.osgi.provisioning.LibraryDefinition;

/**
 * Implementation of {@link BundleRepository} that supports ordered chaining across a list of
 * <code>BundleRepositories</code>.<p/>
 * 
 * <code>BundleRepositories</code> are searched in the order they are specified, with those earlier in the list being
 * preferred to those later in the list.
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @author Andy Wilkinson
 * @since 1.0
 */
final class CompositeBundleRepository implements MutableBundleRepository {

    /**
     * The <code>BundleRepositories</code> to search.
     */
    private final BundleRepository[] chainedRepositories;

    /**
     * Creates a new <code>CompositeBundleRepository</code>.
     * 
     * @param chainedRepositories the <code>BundleRepositories</code> to search.
     */
    public CompositeBundleRepository(BundleRepository[] chainedRepositories) {
        this.chainedRepositories = chainedRepositories;
    }

    /**
     * {@inheritDoc}
     */
    public BundleDefinition findBySymbolicName(String symbolicName, VersionRange versionRange) {
        for (BundleRepository repository : this.chainedRepositories) {
            BundleDefinition bundleDefinition = repository.findBySymbolicName(symbolicName, versionRange);
            if (bundleDefinition != null) {
                return bundleDefinition;
            }
        }
        return null;
    }

    /**
     * {@inheritDoc}
     */
    public Set<? extends BundleDefinition> findByExportedPackage(String packageName, VersionRange versionRange) {
        Set<BundleDefinition> bundleDefinitions = new LinkedHashSet<BundleDefinition>();
        for (BundleRepository repository : this.chainedRepositories) {
            Set<? extends BundleDefinition> definitionsFromRepository = repository.findByExportedPackage(packageName, versionRange);
            addAllNonDuplicateBundleDefinitions(bundleDefinitions, definitionsFromRepository);
        }
        return bundleDefinitions;
    }

    /**
     * {@inheritDoc}
     */
    public Set<? extends BundleDefinition> findByFragmentHost(String bundleSymbolicName, Version version) {
        Set<BundleDefinition> bundleDefinitions = new LinkedHashSet<BundleDefinition>();
        for (BundleRepository repository : this.chainedRepositories) {
            Set<? extends BundleDefinition> definitionsFromRepository = repository.findByFragmentHost(bundleSymbolicName, version);
            addAllNonDuplicateBundleDefinitions(bundleDefinitions, definitionsFromRepository);
        }
        return bundleDefinitions;
    }

    /**
     * {@inheritDoc}
     */
    public LibraryDefinition findLibrary(String libraryName, VersionRange versionRange) {
        for (BundleRepository repository : this.chainedRepositories) {
            LibraryDefinition library = repository.findLibrary(libraryName, versionRange);
            if (library != null) {
                return library;
            }
        }
        return null;
    }

    /**
     * {@inheritDoc}
     */
    public BundleDefinition addBundle(File location) {
        for (BundleRepository repository : this.chainedRepositories) {
            if (repository instanceof MutableBundleRepository) {
                return ((MutableBundleRepository) repository).addBundle(location);
            }
        }
        return null;
    }

    /**
     * {@inheritDoc}
     */
    public LibraryDefinition addLibrary(File location) {
        for (BundleRepository repository : this.chainedRepositories) {
            if (repository instanceof MutableBundleRepository) {
                return ((MutableBundleRepository) repository).addLibrary(location);
            }
        }
        return null;
    }

    /**
     * {@inheritDoc}
     */
    public void refresh() {
        for (BundleRepository repository : this.chainedRepositories) {
            repository.refresh();
        }
    }

    /**
     * Adds all of the non-duplicate <code>BundleDefinitions</code> in the supplied <code>Set</code> of new
     * definitions to the supplied <code>Set</code> of existing definitions. A definition is considered to be
     * duplicate if the <code>Set</code> of existing definitions already includes a definition with the same
     * <code>Bundle-SymbolicName</code> and <code>Bundle-Version</code>.
     * 
     * @param existingDefinitions The Set of existing definitions.
     * @param newDefinitions The Set of new definitions to add to the existing definitions.
     */
    private void addAllNonDuplicateBundleDefinitions(Set<BundleDefinition> existingDefinitions, Set<? extends BundleDefinition> newDefinitions) {
        if (newDefinitions != null && !newDefinitions.isEmpty()) {
            if (existingDefinitions.isEmpty()) {
                existingDefinitions.addAll(newDefinitions);
            } else {
                for (BundleDefinition newDefinition : newDefinitions) {
                    boolean duplicate = false;
                    BundleSymbolicName bundleSymbolicName = newDefinition.getManifest().getBundleSymbolicName();
                    if (bundleSymbolicName != null) {
                        Version newDefinitionVersion = getBundleVersion(newDefinition.getManifest());
                        for (BundleDefinition existingDefinition : existingDefinitions) {
                            BundleSymbolicName bsnOfExistingDefinition = existingDefinition.getManifest().getBundleSymbolicName();
                            if (bsnOfExistingDefinition != null && bsnOfExistingDefinition.equals(bundleSymbolicName)) {
                                Version existingDefinitionVersion = getBundleVersion(existingDefinition.getManifest());
                                if (existingDefinitionVersion.equals(newDefinitionVersion)) {
                                    duplicate = true;
                                    break;
                                }
                            }
                        }
                    }
                    if (!duplicate) {
                        existingDefinitions.add(newDefinition);
                    }
                }
            }
        }
    }
    
    private Version getBundleVersion(BundleManifest manifest) {
        String versionString =  manifest.getBundleVersion();
        if (versionString == null) {
            return Version.emptyVersion;
        } else {
            return new Version(versionString);
        }
    }

    /** 
     * {@inheritDoc}
     */
    public Set<? extends BundleDefinition> getBundles() {
        Set<BundleDefinition> allBundles = new HashSet<BundleDefinition>();
        for (BundleRepository repository : chainedRepositories) {
            allBundles.addAll(repository.getBundles());
        }
        return allBundles;
    }

    /** 
     * {@inheritDoc}
     */
    public Set<? extends LibraryDefinition> getLibraries() {
        Set<LibraryDefinition> allLibraries = new HashSet<LibraryDefinition>();
        for (BundleRepository repository : chainedRepositories) {
            allLibraries.addAll(repository.getLibraries());
        }
        return allLibraries;
    }
}
