/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.provisioning.repository;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Enumeration;
import java.util.Set;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;

import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;
import org.osgi.framework.Version;
import org.springframework.util.FileCopyUtils;
import org.springframework.util.StringUtils;

import com.springsource.server.osgi.manifest.VersionRange;
import com.springsource.server.osgi.provisioning.BundleDefinition;
import com.springsource.server.osgi.provisioning.BundleProvisioningException;
import com.springsource.server.osgi.provisioning.LibraryDefinition;
import com.springsource.server.osgi.provisioning.support.BundleDefinitionFactory;
import com.springsource.server.osgi.provisioning.support.LibraryDefinitionFactory;

/**
 * Implementation of {@link MutableBundleRepository} that delegates to a standard {@link FileSystemBundleRepository},
 * adding installation into the search path. <p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
final class MutableFileSystemBundleRepository implements MutableBundleRepository {

    private static final Tracer TRACER = TracerFactory.createTracer(MutableFileSystemBundleRepository.class);

    private final File installDirectory;

    private final FileSystemBundleRepository delegate;

    /**
     * Creates a new <code>MutableFileSystemBundleRepository</code>.
     * 
     * @param installPath the search path.
     * @param cacheLocationPath the cache location path.
     * @param absoluteRoot an absolute path to be used if the <code>installPath</code> is relative
     */
    MutableFileSystemBundleRepository(String installPath, String cacheLocationPath, String absoluteRoot) {
        File installDirectory = new File(installPath);
        if (!installDirectory.exists()) {
            if (TRACER.isDebugEnabled()) {
                TRACER.traceDebug("Creating install directory '" + installDirectory.getAbsolutePath() + "'.");
            }
            installDirectory.mkdirs();
        }
        if (!installDirectory.isDirectory()) {
            throw new IllegalArgumentException("Install path '" + installPath + " either does not exist or is not a directory.");
        }
        this.installDirectory = installDirectory;

        String searchPath = this.installDirectory.getAbsolutePath() + "/{name}";
        this.delegate = new FileSystemBundleRepository(searchPath, cacheLocationPath, absoluteRoot);
    }

    /**
     * {@inheritDoc}
     * @see FileSystemBundleRepository#findByExportedPackage(String, VersionRange)
     */
    public Set<? extends BundleDefinition> findByExportedPackage(String packageName, VersionRange versionRange) {
        return this.delegate.findByExportedPackage(packageName, versionRange);
    }

    /**
     * {@inheritDoc}
     * @see FileSystemBundleRepository#findBySymbolicName(String, VersionRange)
     */
    public BundleDefinition findBySymbolicName(String symbolicName, VersionRange versionRange) {
        return this.delegate.findBySymbolicName(symbolicName, versionRange);
    }

    /**
     * {@inheritDoc}
     * @see FileSystemBundleRepository#findByFragmentHost(String, Version)
     */
    public Set<? extends BundleDefinition> findByFragmentHost(String bundleSymbolicName, Version version) {
        return this.delegate.findByFragmentHost(bundleSymbolicName, version);
    }

    /**
     * {@inheritDoc}
     * @see FileSystemBundleRepository#findLibrary(String, VersionRange)
     */
    public LibraryDefinition findLibrary(String libraryName, VersionRange versionRange) {
        return this.delegate.findLibrary(libraryName, versionRange);
    }

    /**
     * {@inheritDoc}
     */
    public BundleDefinition addBundle(File location) {
        if (!location.exists()) {
            throw new IllegalArgumentException("Bundle location '" + location + "' does not exist.");
        }
        File installedLocation = copyLocation(location);
        BundleDefinition bundleDefinition = BundleDefinitionFactory.createBundleDefinition(installedLocation);
        this.delegate.registerBundleDefinition(bundleDefinition);
        this.delegate.updateCaches();
        return bundleDefinition;
    }

    /**
     * {@inheritDoc}
     */
    public LibraryDefinition addLibrary(File location) {
        if (!location.exists()) {
            throw new IllegalArgumentException("Library location '" + location + "' does not exist.");
        }
        File installedLocation = copyLocation(location);
        if (installedLocation.isFile() && installedLocation.getName().endsWith(".jar")) {
            File originalInstallLocation = installedLocation;
            installedLocation = expandJar(installedLocation);
            originalInstallLocation.delete();
        }
        LibraryDefinition libraryDefinition = LibraryDefinitionFactory.createLibraryDefinition(installedLocation);
        this.delegate.registerLibraryDefinition(libraryDefinition);
        this.delegate.updateCaches();
        return libraryDefinition;
    }

    private File expandJar(File installedLocation) {
        File targetDir = new File(this.installDirectory.getAbsolutePath() + File.separator
            + StringUtils.stripFilenameExtension(installedLocation.getName()));
        targetDir.mkdir();
        try {
            JarFile jar = new JarFile(installedLocation);
            Enumeration<JarEntry> entries = jar.entries();
            while (entries.hasMoreElements()) {
                JarEntry entry = entries.nextElement();
                expandEntry(targetDir, entry, jar);
            }
        } catch (IOException e) {
            throw new BundleProvisioningException("Unable to expand library JAR '" + installedLocation + "'.", e);
        }
        return targetDir;
    }

    /**
     * Expands the supplied {@link JarEntry} into the correct sub-directory under <code>targetDir</code>.
     * 
     * @param targetDir the target directory.
     * @param entry the entry to expand.
     * @param jar the encloing JAR file.
     * @throws IOException
     */
    private void expandEntry(File targetDir, JarEntry entry, JarFile jar) throws IOException {
        File file = new File(targetDir + File.separator + entry);
        if (entry.isDirectory()) {
            file.mkdirs();
        } else {
            file.createNewFile();
            InputStream in = jar.getInputStream(entry);
            OutputStream out = new FileOutputStream(file);
            try {
                FileCopyUtils.copy(in, out);
            } catch (IOException ex) {
                throw new BundleProvisioningException("Unable to expand JAR entry '" + entry.getName() + "' in JAR '" + jar.getName() + "'.", ex);
            }
        }
    }

    private File copyLocation(File location) {
        File target = new File(this.installDirectory.getPath() + File.separatorChar + location.getName());

        if (target.exists()) {
            throw new RuntimeException("Illegal attempt to install bundle to existing target: " + target);
        }

        if (location.isFile()) {
            copyFile(location, target);
        } else {
            copyDir(location, target);
        }
        return target;
    }

    private void copyDir(File location, File target) {
        target.mkdir();
        File[] files = location.listFiles();
        if (files != null) {
            for (File file : files) {
                File nestedTarget = new File(target.getAbsolutePath() + File.separatorChar + file.getName());
                if (file.isDirectory()) {
                    copyDir(file, nestedTarget);
                } else {
                    copyFile(file, nestedTarget);
                }
            }
        }
    }

    private void copyFile(File location, File target) {
        try {
            FileCopyUtils.copy(location, target);
        } catch (IOException ex) {
            throw new BundleProvisioningException("Unable to copy file at source '" + location + "'  into repository.", ex);
        }
    }

    public void refresh() {
        this.delegate.refresh();
    }

    /** 
     * {@inheritDoc}
     */
    public Set<? extends BundleDefinition> getBundles() {
        return delegate.getBundles();
    }

    /** 
     * {@inheritDoc}
     */
    public Set<? extends LibraryDefinition> getLibraries() {
        return delegate.getLibraries();
    }
}
