/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.provisioning.repository;

import java.io.File;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Describes a file system search pattern such as <code>${user.home}/ivy-cache/{org}/{name}/{bundle}.jar</code>.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
final class SearchPattern {

    private static final Pattern PROPERTY_PATTERN = Pattern.compile("(\\$\\{(([^\\}]+))\\})");

    private static final String PLACEHOLDER_END = "}";

    private static final String PLACEHOLDER_START = "{";

    private final String pattern;

    private final File rootDir;

    private final PatternElement[] patternElements;

    SearchPattern(String pattern, String absoluteRoot) {
        this.pattern = pattern;

        String expanded = expandProperties(pattern);
        String rootDirPath = extractRootDirPath(expanded);
        String matchPattern = expanded.substring(rootDirPath.length());

        if (absoluteRoot != null) {
            File rootDir = new File(rootDirPath);
            if (!rootDir.isAbsolute()) {
                this.rootDir = new File(absoluteRoot, rootDirPath);
            } else {
                this.rootDir = rootDir;
            }
        } else {
            this.rootDir = new File(rootDirPath);
        }

        validateRootDir();
        this.patternElements = parsePattern(matchPattern);
    }

    /**
     * Check that the root directory of the search pattern exists and is actually a directory.
     */
    private void validateRootDir() {
        if (!this.rootDir.exists()) {
            throw new IllegalArgumentException("Pattern '" + this.pattern + "' points to a non-existent root dir '" + this.rootDir.getAbsolutePath() + "'");
        }

        if (!this.rootDir.isDirectory()) {
            throw new IllegalArgumentException("Pattern '" + this.pattern + "' does not point to a valid root dir");
        }
    }

    /**
     * Parses the supplied pattern into an array of {@link PatternElement PatternElements}.
     */
    private PatternElement[] parsePattern(String matchPattern) {
        if (matchPattern == null || matchPattern.trim().length() == 0) {
            return new PatternElement[0];
        }
        if (matchPattern.startsWith("/")) {
            matchPattern = matchPattern.substring(1);
        }
        String[] patternPieces = matchPattern.split("/");
        PatternElement[] elements = new PatternElement[patternPieces.length];
        for (int x = 0; x < patternPieces.length; x++) {
            String piece = patternPieces[x];
            PatternElement pe;
            if (piece.contains(PLACEHOLDER_START) && piece.contains(PLACEHOLDER_END)) {
                pe = new DynamicPatternElement(piece);
            } else {
                pe = new StaticPatternElement(piece);
            }
            elements[x] = pe;
        }
        return elements;
    }

    /**
     * @param pattern
     * @return
     */
    private String extractRootDirPath(String pattern) {
        int firstPlaceholder = pattern.indexOf(PLACEHOLDER_START);
        if (firstPlaceholder < 0) {
            return pattern;
        }
        
        String upToFirstPlaceholder = pattern.substring(0, firstPlaceholder);
        
        int endOfRootDir = upToFirstPlaceholder.lastIndexOf("/");
        
        if (endOfRootDir < 0) {
            return upToFirstPlaceholder;
        }
        
        return pattern.substring(0, endOfRootDir);
    }

    File getRootDir() {
        return this.rootDir;
    }

    PatternElement getPatternElement(int index) {
        if (index < 0 || index >= this.patternElements.length) {
            return null;
        } else {
            return this.patternElements[index];
        }
    }

    int getPatternElementCount() {
        return this.patternElements.length;
    }

    /**
     * {@inheritDoc}
     */
    @Override public String toString() {
        return this.pattern;
    }

    private static String expandProperties(String value) {
        Pattern regex = PROPERTY_PATTERN;
        StringBuffer buffer = new StringBuffer(value.length());
        Matcher matcher = regex.matcher(value);
        int propertyGroup = matcher.groupCount();
        String key, property = "";
        while (matcher.find()) {
            key = matcher.group(propertyGroup);
            property = "";
            if(key.contains("::")){
                String[] keyDefault = key.split("::");
                property = System.getProperty(keyDefault[0]);
                if(property == null){
                    property = keyDefault[1];
                }else{
                    property.replace('\\', '/');
                }
            }else{
                property = System.getProperty(matcher.group(propertyGroup)).replace('\\', '/');
            }
            matcher.appendReplacement(buffer, property);
        }
        matcher.appendTail(buffer);
        return buffer.toString();
    }

    /**
     * A single element in a pattern.
     */
    public static interface PatternElement {

        /**
         * Tests whether this <code>PatternElement</code> matches the supplied {@link File}.
         * 
         * @param file the <code>File</code> to test.
         * @return <code>true</code> if it matches; otherwise <code>false</code>.
         */
        boolean matches(File file);
    }

    /**
     * A pattern element with a wildcard section.
     */
    private static class DynamicPatternElement implements PatternElement {

        private final String patternElement;

        private final String start;

        private final String end;

        public DynamicPatternElement(String patternElement) {
            int i = patternElement.indexOf('{');
            int j = patternElement.indexOf('}');
            this.start = patternElement.substring(0, i);
            this.end = patternElement.substring(j + 1);
            this.patternElement = patternElement;
        }

        /**
         * {@inheritDoc}
         */
        public boolean matches(File file) {
            String name = file.getName();
            return name.startsWith(this.start) && name.endsWith(this.end);
        }

        @Override public String toString() {
            return this.patternElement;
        }
    }

    /**
     * A pattern element with no wildcard section.
     */
    private static class StaticPatternElement implements PatternElement {

        private final String patternElement;

        public StaticPatternElement(String patternElement) {
            this.patternElement = patternElement;
        }

        public boolean matches(File file) {
            return this.patternElement.equals(file.getName());
        }

        @Override public String toString() {
            return this.patternElement;
        }
    }
}
