/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.provisioning.support;

import java.io.File;
import java.io.IOException;
import java.io.Reader;
import java.util.List;
import java.util.jar.Attributes;
import java.util.jar.Manifest;

import org.osgi.framework.Version;

import com.springsource.server.osgi.manifest.ManifestFactory;
import com.springsource.server.osgi.manifest.StandardBundleImport;
import com.springsource.server.osgi.manifest.BundleManifest.ImportBundleHeader.BundleImport;
import com.springsource.server.osgi.manifest.parse.HeaderDeclaration;
import com.springsource.server.osgi.manifest.parse.HeaderParser;
import com.springsource.server.osgi.manifest.parse.HeaderParserFactory;
import com.springsource.server.osgi.manifest.parse.OSGiIntegrationLogService;
import com.springsource.server.osgi.provisioning.BundleProvisioningException;
import com.springsource.server.osgi.provisioning.LibraryDefinition;

/**
 * Base implementation of {@link LibraryDefinition}. Supports both descriptor-defined bundle content and
 * artefact-defined bundle content.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
abstract class AbstractLibraryDefinition implements LibraryDefinition {

    private static final String DEFAULT_LIBRARY_VERSION = "0";

    protected static final String ATTR_LIBRARY_VERSION = "Library-Version";

    protected static final String ATTR_LIBRARY_SYMBOLICNAME = "Library-SymbolicName";

    protected static final String ATTR_IMPORT_BUNDLE = "Import-Bundle";

    protected static final String ATTR_LIBRARY_NAME = "Library-Name";

    protected static final String ATTR_LIBRARY_DESCRIPTION = "Library-Description";

    private static final HeaderParser PARSER = HeaderParserFactory.newHeaderParser(new OSGiIntegrationLogService());

    private final Object monitor = new Object();

    private final File libraryLocation;

    private final boolean artefactIncludesBundles;

    private volatile String symbolicName;

    private volatile String name;

    private volatile String description;

    private volatile Version version;

    private volatile BundleImport[] libraryBundles;

    private Sharing sharing;

    /**
     * Creates a new <code>AbstractLibraryDefinition</code>.
     * 
     * @param libraryFileLocation the location of the library file.
     */
    public AbstractLibraryDefinition(File libraryFileLocation, boolean artefactIncludesBundles) {
        try {
            this.libraryLocation = libraryFileLocation.getCanonicalFile();
        } catch (IOException ioe) {
            throw new IllegalStateException("'" + libraryFileLocation + "' could not be made canonical.", ioe);
        }
        this.artefactIncludesBundles = artefactIncludesBundles;
    }

    /**
     * {@inheritDoc}
     */
    public final BundleImport[] getLibraryBundles() {
        resolveState();
        return this.libraryBundles;
    }

    /**
     * {@inheritDoc}
     */
    public final String getName() {
        resolveState();
        return this.name;
    }

    /**
     * {@inheritDoc}
     */
    public final Version getVersion() {
        resolveState();
        return this.version;
    }

    /**
     * {@inheritDoc}
     */
    public final String getDescription() {
        resolveState();
        return this.description;
    }

    /**
     * {@inheritDoc}
     */
    public final String getSymbolicName() {
        resolveState();
        return this.symbolicName;
    }

    /**
     * {@inheritDoc}
     */
    public File getLocation() {
        return this.libraryLocation;
    }

    /**
     * {@inheritDoc}
     */
    public Sharing getSharing() {
        return this.sharing;
    }

    /**
     * Subclasses must override this to return a {@link Reader} for the library manifest when required.
     * 
     * @param libraryLocation the library file location.
     * @return the manifest <code>Reader</code>.
     */
    protected abstract Reader manifestReader(File libraryLocation);

    /**
     * Subclasses that support artefact-supplied bundles can override this to return those bundles when the state is
     * being loaded.
     * 
     * @param libraryLocation the library file location.
     * @return the artefact-supplied bundles.
     */
    protected BundleImport[] locateLibraryBundles(File libraryLocation) {
        throw new UnsupportedOperationException(getClass().getSimpleName() + "does not support artefact supplied bundles.");
    }

    private void resolveState() {
        synchronized (this.monitor) {
            if (this.symbolicName == null) {
                Reader reader = null;
                try {
                    reader = manifestReader(this.libraryLocation);
                    Manifest mf = new ManifestFactory(reader).getManifest();
                    Attributes attrs = mf.getMainAttributes();
                    String symbolicNameString = getRequiredHeader(ATTR_LIBRARY_SYMBOLICNAME, attrs);
                    String version = getLibraryVersion(attrs);

                    HeaderDeclaration symbolicNameDeclaration = HeaderParserFactory.newHeaderParser(new OSGiIntegrationLogService()).parseLibrarySymbolicName(
                        symbolicNameString);

                    this.symbolicName = symbolicNameString;

                    String sharing = symbolicNameDeclaration.getDirectives().get("sharing");
                    if (sharing != null) {
                        this.sharing = Sharing.valueOf(sharing.toUpperCase());
                    } else {
                        this.sharing = Sharing.SHAREABLE;
                    }

                    this.version = new Version(version);

                    if (this.artefactIncludesBundles) {
                        this.libraryBundles = locateLibraryBundles(this.libraryLocation);
                    } else {
                        this.libraryBundles = parseImportBundle(attrs);
                    }

                    this.name = attrs.getValue(ATTR_LIBRARY_NAME);
                    this.description = attrs.getValue(ATTR_LIBRARY_DESCRIPTION);

                } finally {
                    try {
                        if (reader != null) {
                            reader.close();
                        }
                    } catch (IOException ioe) {
                        throw new RuntimeException("Unable to close manifest reader", ioe);
                    }
                }
            }
        }
    }

    /**
     * Parse the library version and provide a default value if it is not specified.
     * 
     * @param attrs the manifest attributes to be parsed
     * @return the library version
     */
    private String getLibraryVersion(Attributes attrs) {
        String libraryVersionAttributevalue = attrs.getValue(ATTR_LIBRARY_VERSION);
        String version = libraryVersionAttributevalue == null || libraryVersionAttributevalue.trim().length() == 0 ? DEFAULT_LIBRARY_VERSION
            : libraryVersionAttributevalue;
        return version;
    }

    /**
     * Parses the <code>Import-Bundle</code> header in the supplied {@link Attributes}.
     * 
     * @param attrs the <code>Attributes</code>.
     * @param libraryFileLocation the location of the library file.
     * @return the library bundles.
     */
    private BundleImport[] parseImportBundle(Attributes attrs) {
        String headerValue = getRequiredHeader(ATTR_IMPORT_BUNDLE, attrs);
        List<HeaderDeclaration> headers = PARSER.parseImportBundleHeader(headerValue);
        BundleImport[] bundles = new BundleImport[headers.size()];
        for (int x = 0; x < headers.size(); x++) {
            HeaderDeclaration header = headers.get(x);
            String name = header.getNames().get(0);
            String version = header.getAttributes().get("version");
            String resolution = header.getDirectives().get("resolution");
            String importScope = header.getDirectives().get("import-scope");
            bundles[x] = new StandardBundleImport(name, version, resolution, importScope);
        }
        return bundles;
    }

    private String getRequiredHeader(String name, Attributes attrs) {
        String value = attrs.getValue(name);
        if (value == null || value.trim().length() == 0) {
            throw new BundleProvisioningException("Required attribute '" + name + "' is missing from library descriptor '"
                + this.libraryLocation.getAbsolutePath() + "'");
        }
        return value;
    }

    @Override public String toString() {
        return String.format("%s - %s, %s, %s", getClass().getSimpleName(), getSymbolicName(), getVersion().toString(),
            getLocation().getAbsolutePath());
    }

    @Override public int hashCode() {
        return this.libraryLocation.getName().hashCode();
    }

    @Override public boolean equals(Object other) {
        if (this == other) {
            return true;
        }

        if (other == null) {
            return false;
        }

        if (getClass() != other.getClass()) {
            return false;
        }

        AbstractLibraryDefinition that = (AbstractLibraryDefinition) other;

        return this.libraryLocation.equals(that.libraryLocation);
    }
}
