/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.provisioning.support;

import java.io.File;
import java.io.IOException;
import java.io.Reader;

import com.springsource.server.osgi.manifest.BundleManifest;
import com.springsource.server.osgi.manifest.StandardBundleManifestFactory;
import com.springsource.server.osgi.manifest.parse.FatalParseException;
import com.springsource.server.osgi.manifest.parse.OSGiIntegrationLogService;
import com.springsource.server.osgi.provisioning.BundleDefinition;
import com.springsource.server.osgi.provisioning.OsgiProvisioningLogEvent;
import com.springsource.server.serviceability.logging.LogService;
import com.springsource.server.serviceability.logging.LogServiceFactory;

/**
 * Base implementation of {@link BundleDefinition} that points to a resource on the local filesystem.
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Implementations must be threadsafe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
abstract class AbstractLocalBundleDefinition implements BundleDefinition {

    private static final LogService LOGGER  = LogServiceFactory.getLogService();
    
    private final File file;

    protected AbstractLocalBundleDefinition(File file) {
        try {
            this.file = file.getCanonicalFile();
        } catch (IOException ioe) {
            throw new IllegalStateException("'" + file + "' could not be made canonical.", ioe);
        }
    }

    /**
     * {@inheritDoc}
     */
    public final File getLocation() {
        return this.file;
    }

    /**
     * {@inheritDoc}
     */
    public final BundleManifest getManifest() {
        Reader reader = null;
        try {
            reader = getManifestReader();
            BundleManifest bundleManifest = null;
            
            if (reader != null) {
                try {
                    bundleManifest = new StandardBundleManifestFactory(new OSGiIntegrationLogService()).createBundleManifest(reader);
                } catch (FatalParseException fpe) {
                    LOGGER.log(OsgiProvisioningLogEvent.MALFORMED_MANIFEST, getLocation().getAbsolutePath(), fpe.getMessage());
                }
            }
            return bundleManifest;
        } finally {
            if (reader != null) {
                try {
                    reader.close();
                } catch (IOException ioe) {
                    throw new RuntimeException("Unable to close manifest reader", ioe);
                }
            }
        }
    }

    /**
     * Strategy method for retrieving a {@link Reader} for the manifest. Called lazily to improve startup performance.
     * 
     * @return A <code>Reader</code> for the manifest of the definition's bundle
     */
    abstract Reader getManifestReader();

    @Override public String toString() {
        return String.format("%s -> %s", getClass().getSimpleName(), this.file.getAbsolutePath());
    }

    @Override public int hashCode() {
        return this.file.getName().hashCode();
    }

    @Override public boolean equals(Object other) {
        if (this == other) {
            return true;
        }

        if (other == null) {
            return false;
        }

        if (getClass() != other.getClass()) {
            return false;
        }

        AbstractLocalBundleDefinition that = (AbstractLocalBundleDefinition) other;

        return this.file.equals(that.file);
    }
}
