/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.provisioning.support;

import java.io.File;

import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;
import org.springframework.util.StringUtils;

import com.springsource.server.osgi.provisioning.LibraryDefinition;

/**
 * Factory class for creating {@link LibraryDefinition} instances<p/>.
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
public final class LibraryDefinitionFactory {

    private static final Tracer TRACER = TracerFactory.createTracer(LibraryDefinitionFactory.class);

    private static final String LIBRARY_DESCRIPTOR_EXTENSION = "libd";

    /**
     * Creates a new {@link LibraryDefinition} from the supplied <code>libraryFile</code>. The library file can be
     * one of:
     * <ol>
     * <li>A library descriptor file <code>(*.libd)</code></li>
     * <li>A JAR-packaged library <code>(*.lib)</code></li>
     * <li>An exploded library folder</lib>
     * </ol>
     * 
     * @param libraryFile the library artefact file.
     */
    public static LibraryDefinition createLibraryDefinition(File libraryFile) {
        if (TRACER.isDebugEnabled()) {
            TRACER.traceDebug("Creating library definition for: " + libraryFile.getAbsolutePath());
        }
        if (libraryFile == null || !libraryFile.exists()) {
            throw new IllegalArgumentException("Library file '" + libraryFile + "' does not exist.");
        }
        if (libraryFile.isDirectory()) {
            if (TRACER.isDebugEnabled()) {
                TRACER.traceDebug("Created directory library definition for: " + libraryFile.getAbsolutePath());
            }
            return new DirectoryLibraryDefinition(libraryFile);
        } else {
            String extension = StringUtils.getFilenameExtension(libraryFile.getName());

            if (LIBRARY_DESCRIPTOR_EXTENSION.equals(extension)) {
                if (TRACER.isDebugEnabled()) {
                    TRACER.traceDebug("Created descriptor library definition for: " + libraryFile.getAbsolutePath());
                }
                return new DescriptorLibraryDefiniton(libraryFile);
            }
        }
        return null;
    }

    /**
     * Queries whether the supplied {@link File} is a library.
     * 
     * @param file the <code>File</code>.
     * @return <code>true</code> if the <code>File</code> is a library, otherwise <code>false</code>.
     */
    public static boolean isLibrary(File file) {
        if (file == null || !file.exists()) {
            return false;
        }
        if (file.isDirectory()) {
            return DirectoryLibraryDefinition.isDirectoryLibrary(file);
        } else {
            return LIBRARY_DESCRIPTOR_EXTENSION.equals(StringUtils.getFilenameExtension(file.getName()));
        }
    }

}
