/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.osgi.provisioning.support;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.StringReader;
import java.io.StringWriter;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;

import org.springframework.util.FileCopyUtils;

import com.springsource.server.osgi.provisioning.BundleProvisioningException;

/**
 * Utility class for extracting a {@link Reader} for manifest data in a JAR file and in exploded JAR directories. <p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
public final class ManifestUtils {

    private static final String MANIFEST_DIRECTORY_LOCATION = "META-INF" + File.separator + "MANIFEST.MF";

    private static final String MANIFEST_ENTRY = "META-INF/MANIFEST.MF";

    /**
     * Creates a {@link Reader} for the manifest in the supplied exploded JAR directory.
     * 
     * @param directory the exploded JAR directory.
     * @return the <code>Reader</code> or <code>null</code> if the manifest cannot be found.
     */
    public static final Reader manifestReaderFromExplodedDirectory(File directory) {
        if (directory == null || !directory.isDirectory()) {
            throw new BundleProvisioningException("Must supply a valid directory");
        }
        try {
            File manifestFile = new File(directory.getAbsolutePath() + File.separator + MANIFEST_DIRECTORY_LOCATION);
            if (manifestFile.exists()) {
                return new FileReader(manifestFile);
            } else {
                return null;
            }
        } catch (IOException e) {
            throw new BundleProvisioningException("Unable to read MANIFEST for exploded directory '" + directory.getAbsolutePath() + "'.", e);
        }
    }

    /**
     * Creates a {@link Reader} for the manifest in the supplied JAR file.
     * 
     * @param file the JAR file.
     * @return the <code>Reader</code> or <code>null</code> if the manifest cannot be found.
     */
    public static final Reader manifestReaderFromJar(File file) {
        try {
            JarFile jar = new JarFile(file);
            JarEntry entry = jar.getJarEntry(MANIFEST_ENTRY);
            if (entry != null) {
                StringWriter writer = new StringWriter();
                FileCopyUtils.copy(new InputStreamReader(jar.getInputStream(entry)), writer);
                jar.close();
                return new StringReader(writer.toString());
            } else {
                return null;
            }
        } catch (Exception e) {
            throw new BundleProvisioningException("Cannot read MANIFEST.MF from jar '" + file.getAbsolutePath() + "'.", e);
        }
    }
}
