/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.serviceability.ffdc;

import java.util.Collection;
import java.util.Map;

import org.springframework.util.StringUtils;

/**
 * A wrapper for the Spring framework's Assert class which will be advised by the Server's FFDC aspects.
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * This class is thread safe.
 * 
 * @author Glyn Normington
 * @since 1.0
 * @see org.springframework.util.Assert
 */
final public class Assert {

    
    /**
     * A <code>FatalAssertionException</code> is thrown when an assertion failure occurs and
     * will result in a dump being generated.
     *
     * <strong>Concurrent Semantics</strong><br />
     *
     *  This class is <strong>thread-safe</strong>.
     *
     * @author Andy Wilkinson
     * @since 1.0
     */
    final public static class FatalAssertionException extends FatalServerException {

        private static final long serialVersionUID = -4633344457818398425L;

        /**
         * Creates a new FatalAssertionException with the supplied message
         * @param message The exception message
         */
        public FatalAssertionException(String message) {
            super(message);
        }
    }

    /**
     * Assert a boolean expression, throwing a <code>FatalAssertionException</code> if the test result is
     * <code>false</code>.
     * 
     * <pre class="code">
     * Assert.isTrue(i &gt; 0, &quot;The value must be greater than zero&quot;);
     * </pre>
     * 
     * @param expression a boolean expression
     * @param message the exception message to use if the assertion fails
     * @param inserts any inserts to include if the message is a format string.
     * @throws FatalAssertionException if expression is <code>false</code>
     */
    public static void isTrue(boolean expression, String message, Object... inserts) {
        String formattedMessage = String.format(message, inserts);
        try {
            org.springframework.util.Assert.isTrue(expression, formattedMessage);
        } catch (RuntimeException e) {
            throw new FatalAssertionException(formattedMessage);
        }
    }
    
    /**
     * Assert a boolean expression, throwing a <code>FatalAssertionException</code> if the test result is
     * <code>true</code>.
     * 
     * <pre class="code">
     * Assert.isFalse(state.isBroken(), &quot;The state is broken&quot;);
     * </pre>
     * 
     * @param expression a boolean expression
     * @param message the exception message to use if the assertion fails
     * @throws FatalAssertionException if expression is <code>false</code>
     */
    public static void isFalse(boolean expression, String message) {
        try {
            org.springframework.util.Assert.isTrue(!expression, message);
        } catch (RuntimeException e) {
            throw new FatalAssertionException(message);
        }
    }

    /**
     * Assert that an object is <code>null</code>.
     * 
     * <pre class="code">
     * Assert.isNull(value, &quot;The value must be null&quot;);
     * </pre>
     * 
     * @param object the object to check
     * @param message the exception message to use if the assertion fails
     * @throws FatalAssertionException if the object is not <code>null</code>
     */
    public static void isNull(Object object, String message) {
        try {
            org.springframework.util.Assert.isNull(object, message);
        } catch (RuntimeException e) {
            throw new FatalAssertionException(message);
        }
    }

    /**
     * Assert that an object is not <code>null</code>.
     * 
     * <pre class="code">
     * Assert.notNull(clazz, &quot;The class must not be null&quot;);
     * </pre>
     * 
     * @param object the object to check
     * @param message the exception message to use if the assertion fails
     * @throws FatalAssertionException if the object is <code>null</code>
     */
    public static void notNull(Object object, String message) {
        try {
            org.springframework.util.Assert.notNull(object, message);
        } catch (RuntimeException e) {
            throw new FatalAssertionException(message);
        }
    }

    /**
     * Assert that the given String is not empty; that is, it must not be <code>null</code> and not the empty String.
     * 
     * <pre class="code">
     * Assert.hasLength(name, &quot;Name must not be empty&quot;);
     * </pre>
     * 
     * @param text the String to check
     * @param message the exception message to use if the assertion fails
     * @see StringUtils#hasLength
     */
    public static void hasLength(String text, String message) {
        try {
            org.springframework.util.Assert.hasLength(text, message);
        } catch (RuntimeException e) {
            throw new FatalAssertionException(message);
        }
    }

    /**
     * Assert that the given String has valid text content; that is, it must not be <code>null</code> and must contain
     * at least one non-whitespace character.
     * 
     * <pre class="code">
     * Assert.hasText(name, &quot;'name' must not be empty&quot;);
     * </pre>
     * 
     * @param text the String to check
     * @param message the exception message to use if the assertion fails
     * @see StringUtils#hasText
     */
    // public static void hasText(String text, String message) {
    // try {
    // org.springframework.util.Assert.hasText(text, message);
    // } catch (RuntimeException e) {
    // throw new FatalAssertionException(message);
    // }
    // }
    /**
     * Assert that the given text does not contain the given substring.
     * 
     * <pre class="code">
     * Assert.doesNotContain(name, &quot;rod&quot;, &quot;Name must not contain 'rod'&quot;);
     * </pre>
     * 
     * @param textToSearch the text to search
     * @param substring the substring to find within the text
     * @param message the exception message to use if the assertion fails
     */
    // public static void doesNotContain(String textToSearch, String substring, String message) {
    // try {
    // org.springframework.util.Assert.doesNotContain(textToSearch, substring, message);
    // } catch (RuntimeException e) {
    // throw new FatalAssertionException(message);
    // }
    // }
    /**
     * Assert that an array has elements; that is, it must not be <code>null</code> and must have at least one
     * element.
     * 
     * <pre class="code">
     * Assert.notEmpty(array, &quot;The array must have elements&quot;);
     * </pre>
     * 
     * @param array the array to check
     * @param message the exception message to use if the assertion fails
     * @throws IllegalArgumentException if the object array is <code>null</code> or has no elements
     */
    public static void notEmpty(Object[] array, String message) {
        try {
            org.springframework.util.Assert.notEmpty(array, message);
        } catch (RuntimeException e) {
            throw new FatalAssertionException(message);
        }
    }

    /**
     * Assert that a collection has elements; that is, it must not be <code>null</code> and must have at least one
     * element.
     * 
     * <pre class="code">
     * Assert.notEmpty(collection, &quot;Collection must have elements&quot;);
     * </pre>
     * 
     * @param collection the collection to check
     * @param message the exception message to use if the assertion fails
     * @throws FatalAssertionException if the collection is <code>null</code> or has no elements
     */
    public static <T> void notEmpty(Collection<T> collection, String message) {
        try {
            org.springframework.util.Assert.notEmpty(collection, message);
        } catch (RuntimeException e) {
            throw new FatalAssertionException(message);
        }
    }

    /**
     * Assert that a Map has entries; that is, it must not be <code>null</code> and must have at least one entry.
     * 
     * <pre class="code">
     * Assert.notEmpty(map, &quot;Map must have entries&quot;);
     * </pre>
     * 
     * @param map the map to check
     * @param message the exception message to use if the assertion fails
     * @throws FatalAssertionException if the map is <code>null</code> or has no entries
     */
    public static <K, V> void notEmpty(Map<K, V> map, String message) {
        try {
            org.springframework.util.Assert.notEmpty(map, message);
        } catch (RuntimeException e) {
            throw new FatalAssertionException(message);
        }
    }

    /**
     * Assert that the provided object is a non-null instance of the provided class.
     * 
     * <pre class="code">
     * Assert.instanceOf(Foo.class, foo);
     * </pre>
     * 
     * @param type the type to check against
     * @param obj the object to check
     * @param message a message which will be prepended to the message produced by the function itself, and which may be
     *        used to provide context. It should normally end in a ": " or ". " so that the function generate message
     *        looks ok when prepended to it.
     * @throws FatalAssertionException if the object is not an instance of clazz
     * @see Class#isInstance
     */
    public static <T> void isInstanceOf(Class<T> type, Object obj, String message) {
        try {
            org.springframework.util.Assert.isInstanceOf(type, obj, message);
        } catch (RuntimeException e) {
            throw new FatalAssertionException(message + "Object of class [" + (obj != null ? obj.getClass().getName() : "null")
                + "] must be an instance of " + type);
        }
    }

    /**
     * Assert that <code>superType.isAssignableFrom(subType)</code> is <code>true</code>.
     * 
     * <pre class="code">
     * Assert.isAssignable(Number.class, myClass);
     * </pre>
     * 
     * @param superType the super type to check against
     * @param subType the sub type to check
     * @param message a message which will be prepended to the message produced by the function itself, and which may be
     *        used to provide context. It should normally end in a ": " or ". " so that the function generate message
     *        looks ok when prepended to it.
     * @throws FatalAssertionException if the classes are not assignable
     */
    public static <T, U> void isAssignable(Class<T> superType, Class<U> subType, String message) {
        try {
            org.springframework.util.Assert.isAssignable(superType, subType, message);
        } catch (RuntimeException e) {
            throw new FatalAssertionException(message + subType + " is not assignable to " + superType);
        }
    }

    /*
     * Prevent instantiation - Java does not allow final abstract classes.
     */
    private Assert() {
    }
}
