/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.serviceability.ffdc;

import java.io.Writer;
/**
 * 
 * Provides a mechanism for constructing a structured dump document.<p/>.
 *
 * <strong>Concurrent Semantics</strong><br />
 *
 * Implementations needn't be threadsafe.
 *
 * @author Damilola Senbanjo
 * @since 1.0
 */
public interface DumpWriter {

    interface ListWriter<T> {

        /**
         * Writes a new item to a list, provided that the end method has not been called.
         * 
         * @param value an object
         */
        ListWriter<T> scalar(Object value);

        /**
         * Creates a new map inside a list.
         * 
         * @return a new {@link MapWriter}
         */
        MapWriter<ListWriter<T>> map();

        /**
         * Creates a list within a list.
         * 
         * @return a new {@link ListWriter}
         */
        ListWriter<ListWriter<T>> list();

        /**
         * closes a list.
         * 
         * @return
         */
        T end();
    }

    interface MapWriter<T> {

        /**
         * Creates a map.
         * 
         * @param key a string identifier
         * @param val an object
         * @return an {@link MapWriter}
         */
        MapWriter<T> scalar(String key, Object val);

        /**
         * Creates a list within a map.
         * 
         * @param key a string identifier
         * @return a {@link ListWriter}
         */
        ListWriter<MapWriter<T>> list(String key);

        /**
         * Closes the map
         * 
         * @return
         */
        T end();

        /**
         * Creates a map within a map
         * 
         * @param key a string
         * @return a new {@link MapWriter}
         */
        MapWriter<MapWriter<T>> map(String key);
    }

    interface SectionWriter {

        /**
         * Writes new map entries to a map
         * 
         * @param key a string
         * @param value an object
         */
        SectionWriter scalar(String key, Object value);

        /**
         * Creates a named list within a map.
         * 
         * @param key a string
         * @return a new {@link ListWriter}
         */
        ListWriter<SectionWriter> list(String key);

        /**
         * Creates a nested map.
         * 
         * @param key a string
         * @return a new {@link MapWriter}
         */
        MapWriter<SectionWriter> map(String key);

        /**
         * Ends the section.
         */
        void end();
    }

    /**
     * Writes to a writer
     * 
     * @param writer the Writer that the dump is written to
     */
    void write(Writer writer);

    /**
     * Starts the section to be written. Puts a name/value pair into the root object, thereby creating a map.
     * 
     * @param key the name of the new section being written
     * @return a new {@link SectionWriter}
     */
    SectionWriter startSection(String key);

}
