/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.serviceability.output;

import java.io.PrintWriter;
import java.io.StringWriter;

/**
 * Base implementation of {@link Entry}. Handles creation and formatting of timestamps and thread ids.
 * 
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * This class is <strong>thread-safe.</strong>
 * 
 * @author Andy Wilkinson
 * @since 1.0
 */
public abstract class AbstractEntry implements Entry {

    private static final int MAXIMUM_SOURCE_LENGTH = 64;

    private static final int MAXIMUM_THREADID_LENGTH = 24;

    private static final String TIMESTAMP_FORMAT_STRING = "[%1$tF %1$tT.%1$tL]";

    private static final String THREAD_ID_FORMAT_STRING = "%1$-" + MAXIMUM_THREADID_LENGTH + "." + MAXIMUM_THREADID_LENGTH + "s";

    private static final String SOURCE_FORMAT_STRING = "%1$" + MAXIMUM_SOURCE_LENGTH + "." + MAXIMUM_SOURCE_LENGTH + "s";

    private static final String LEVEL_IDENTIFIER_FORMAT_STRING = "%1$1.1s";

    protected static final String LINE_SEPARATOR = System.getProperty("line.separator");

    protected final long timestamp = System.currentTimeMillis();

    protected final String threadId = truncateString(Thread.currentThread().getName(), MAXIMUM_THREADID_LENGTH);

    private final String formattedTimestamp;

    private final String formattedThreadId;

    private final String formattedSource;

    private final String formattedLevelIdentifier;

    protected final Throwable throwable;

    private final String source;

    private final Level level;

    /**
     * Creates a new AbstractEntry. Equivalent to <code>AbstractEntry(level, null, null, level.toString()</code>;
     * 
     * @param level The output level of the entry
     * @see #AbstractEntry(Level, Throwable, String, String)
     */
    protected AbstractEntry(Level level) {
        this(level, null);
    }

    /**
     * Creates a new AbstractEntry. Equivalent to <code>AbstractEntry(level, throwable, null, level.toString()</code>;
     * 
     * @param level The output level of the entry
     * @param throwable The <code>Throwable</code> to be included in the entry
     * @see #AbstractEntry(Level, Throwable, String, String)
     */
    protected AbstractEntry(Level level, Throwable throwable) {
        this(level, throwable, null);
    }

    /**
     * Creates a new AbstractEntry. Equivalent to <code>AbstractEntry(level, throwable, source, level.toString()</code>;
     * @param level The output level of the entry
     * @param throwable The <code>Throwable</code> to be included in the entry
     * @param source A String identifying the source of the entry, typically a fully-qualified class or method name.
     * @see #AbstractEntry(Level, Throwable, String, String)
     */
    protected AbstractEntry(Level level, Throwable throwable, String source) {
        this(level, throwable, source, level.toString());
    }

    /**
     * Creates a new AbstractEntry.</p>
     * The <code>AbstractEntry</code> will be time stamped with the time at which it is created and will
     * have a thread id derived from the name of the thread upon which it is created.
     * @param level The output level of the entry.
     * @param throwable The <code>Throwable</code> to be included in the entry.
     * @param source A String identifying the source of the entry, typically a fully-qualified class or method name.
     * @param levelIdentifier A String identifying the level of the entry.
     * @see #AbstractEntry(Level, Throwable, String, String)
     * @see Thread#getName()
     */
    protected AbstractEntry(Level level, Throwable throwable, String source, String levelIdentifier) {
        this.source = source;
        this.throwable = throwable;
        this.level = level;

        this.formattedTimestamp = formatString(TIMESTAMP_FORMAT_STRING, this.timestamp);
        this.formattedThreadId = formatString(THREAD_ID_FORMAT_STRING, this.threadId);
        this.formattedSource = formatString(SOURCE_FORMAT_STRING, truncateString(this.source, MAXIMUM_SOURCE_LENGTH));
        this.formattedLevelIdentifier = formatString(LEVEL_IDENTIFIER_FORMAT_STRING, levelIdentifier);
    }

    /**
     * Returns the formatted timestamp of the entry
     * @return the entry's formatted timestamp
     */
    protected final String getFormattedTimestamp() {
        return formattedTimestamp;
    }

    /**
     * Returns the formatted thread id of the entry
     * @return the entry's formatted thread id
     */
    protected final String getFormattedThreadId() {
        return formattedThreadId;
    }

    /**
     * Returns the formatted source of the entry
     * @return the entry's formatted source
     */
    protected final String getFormattedSource() {
        return formattedSource;
    }

    /**
     * Returns the formatted level identifier of the entry
     * @return the entry's formatted level identifier
     */
    protected final String getFormattedLevelIdentifier() {
        return formattedLevelIdentifier;
    }

    /**
     * Returns the stack trace of the given <code>Throwable</code> as a formatted
     * <code>Strong</code>
     * @param throwable The <code>Throwable</code> for which a formatted stack trace is required
     * @return The formatted stack trace of the supplied <code>Throwable</code>.
     */
    protected static String getThrowableStackTraceAsString(Throwable throwable) {
        StringWriter stringWriter = new StringWriter();
        PrintWriter printWriter = new PrintWriter(stringWriter);
        throwable.printStackTrace(printWriter);

        return stringWriter.toString();
    }

    private static String truncateString(String string, int maximumLength) {
        if (string != null) {
            if (string.length() > maximumLength) {
                return string.substring(string.length() - maximumLength);
            } else {
                return string;
            }
        } else {
            return "";
        }
    }

    private static String formatString(String format, Object args) {
        if (args == null) {
            args = "";
        }
        return String.format(format, args);
    }

    /** 
     * {@inheritDoc}
     */
    public String getSource() {
        return source;
    }

    /** 
     * {@inheritDoc}
     */
    public Level getLevel() {
        return level;
    }
}
