/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.serviceability.output.internal;

import java.io.File;
import java.io.IOException;

import com.springsource.server.serviceability.output.Destination;

/**
 * A <code>Destination</code> implementation that writes its output to a file. <p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * This class is <strong>thread-safe</strong>
 * 
 * @author Christopher Frost
 * @author Glyn Normington
 * @author Andy Wilkinson
 * @since 1.0
 */
final class FileDestination extends WriterDestination implements Destination {

    private static final String DEFAULT_FILEPATH = "." + File.separator + "target" + File.separator;

    /**
     * Creates a new FileDestination that will create its files in the default location.
     * 
     * @param filename The basic filename to use
     * @param rolloverPolicy The rollover policy to use
     * @param maximumRolloverFiles The maximum number of rollover files to be retained by this destination
     * @throws IOException if a problem is encountered creating the output file.
     */
    FileDestination(String filename, RolloverPolicy rolloverPolicy, int maximumRolloverFiles, StandardDestinationRegistry destinationRegistry) throws IOException {
        super(new RollingOverFileWriter(DEFAULT_FILEPATH, filename, ".log", rolloverPolicy, maximumRolloverFiles), destinationRegistry);
    }

    /**
     * Creates a new FileDestination that will create its files in the supplied location
     * 
     * @param filename The basic filename to use
     * @param location The location in which the destination will create its files
     * @param rolloverPolicy The rollover policy to use
     * @param maximumRolloverFiles The maximum number of rollover files to be retained by this destination
     * @throws IOException if a problem is encountered creating the output file.
     */
    FileDestination(String filename, String location, RolloverPolicy rolloverPolicy, int maximumRolloverFiles, StandardDestinationRegistry destinationRegistry) throws IOException {
        super(new RollingOverFileWriter(handleRelativeLocation(location), filename, ".log", rolloverPolicy, maximumRolloverFiles), destinationRegistry);
    }

    private static String handleRelativeLocation(String location) {
        if (!new File(location).isAbsolute()) {
            return DEFAULT_FILEPATH + File.separator + location;
        } else {
            return location;
        }
    }

    String getOutputFilePath() {
        return ((RollingOverFileWriter) writer).getOutputFilePath();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String toString() {
        return this.getOutputFilePath();
    }
}
