/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.serviceability.output.internal;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import com.springsource.server.serviceability.output.Destination;
import com.springsource.server.serviceability.output.DestinationCreationListener;
import com.springsource.server.serviceability.output.DestinationRegistry;

/**
 * Standard implementation of {@link DestinationRegistry}
 * 
 * <strong>Concurrent Semantics</strong><br />
 * Thread-safe.
 * 
 * @author Andy Wilkinson
 * @since 1.0
 */
class StandardDestinationRegistry implements ClosureAwareDestinationRegistry {

    private final Map<String, Destination> destinationsMap = new HashMap<String, Destination>();

    private final Object destinationLock = new Object();

    private final RolloverPolicy defaultRolloverPolicy = new StandardRolloverPolicy(1024 * 1024 * 100);

    private final List<DestinationCreationListener> listeners = new ArrayList<DestinationCreationListener>();

    private final Object listenersLock = new Object();

    /**
     * Creates a new <code>StandardDestinationRegistry<code> and creates the
     * default destination for the console registering it under {@link DestinationRegistry#CONSOLE_DESTINATION_NAME}
     */
    StandardDestinationRegistry() {
        ConsoleDestination consoleDestination = new ConsoleDestination(this);
        registerDestination(DestinationRegistry.CONSOLE_DESTINATION_NAME, consoleDestination);
    }

    /**
     * {@inheritDoc}
     */
    public Destination getDestination(String name) {
        return destinationsMap.get(name);
    }

    /**
     * {@inheritDoc}
     */
    public Destination createFileDestination(String identifier, String filename) throws IOException {
        FileDestination destination = new FileDestination(filename, defaultRolloverPolicy, 4, this);
        registerDestination(identifier, destination);
        return destination;
    }

    /**
     * {@inheritDoc}
     */
    public Destination createFileDestination(String identifier, String filename, String location) throws IOException {
        FileDestination destination = new FileDestination(filename, location, defaultRolloverPolicy, 4, this);
        registerDestination(identifier, destination);
        return destination;
    }

    /**
     * Returns as an element in the returned array the location to which each of the file destinations known to this
     * registry is writing its output
     * 
     * @return an array of the output locations of all this registry's file destinations
     */
    public String[] getFileDestinationOutputLocations() {
        List<String> outputLocationsList = new ArrayList<String>();
        synchronized (destinationLock) {
            Collection<Destination> destinations = destinationsMap.values();
            for (Destination destination : destinations) {
                if (destination instanceof FileDestination) {
                    FileDestination fileDestination = (FileDestination) destination;
                    outputLocationsList.add(fileDestination.getOutputFilePath());
                }
            }
        }
        return outputLocationsList.toArray(new String[0]);
    }

    private void registerDestination(String identifier, Destination destination) {
        synchronized (destinationLock) {
            destinationsMap.put(identifier, destination);
            synchronized (listenersLock) {
                for (DestinationCreationListener listener : listeners) {
                    listener.destinationCreated(identifier, destination);
                }
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    public void addDestinationCreationListener(DestinationCreationListener listener) {
        synchronized (destinationLock) {
            for (String destinationName : destinationsMap.keySet()) {
                Destination destination = destinationsMap.get(destinationName);
                listener.destinationCreated(destinationName, destination);
            }
            synchronized (listenersLock) {
                this.listeners.add(listener);
            }
        }
    }
    
    void clearListeners() {
        synchronized (destinationLock) {
            synchronized (listenersLock) {
                this.listeners.clear();
            }
        }
    }
    
    public void destinationClosed(Destination destination) {
    	synchronized (destinationLock) {
    		Iterator<Entry<String, Destination>> entries = this.destinationsMap.entrySet().iterator();
    		
    		while (entries.hasNext()) {
    			Entry<String, Destination> entry = entries.next();    			
    			if (destination.equals(entry.getValue())) {
    				entries.remove();
    				synchronized(listenersLock) {
	    				for (DestinationCreationListener listener : listeners) {
	                        listener.destinationDestroyed(entry.getKey());
	                    }
    				}
    				break;
    			}
    		}
    	}
    }
}
