/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.serviceability.output.internal;

import java.io.IOException;
import java.io.Writer;
import java.util.concurrent.atomic.AtomicInteger;

import com.springsource.server.serviceability.output.Destination;
import com.springsource.server.serviceability.output.Entry;

/**
 * {@link Destination} implementation that writes to a {@link Writer}.
 * 
 * <strong>Concurrent Semantics</strong><br />
 * This class is <strong>thread-safe</strong>.
 * 
 * @author Glyn Normington
 * @author Rob Harrop
 * @author Andy Wilkinson
 * @since 1.0
 */
class WriterDestination implements Destination {

    private static final String LINE_SEPARATOR = System.getProperty("line.separator");

    protected final Writer writer;

    private final Object outputLock = new Object();

    private final Buffer entryBuffer;

    private final AtomicInteger refCount = new AtomicInteger();
    
    private final ClosureAwareDestinationRegistry destinationRegistry;

    WriterDestination(Writer writer, ClosureAwareDestinationRegistry destinationRegistry) {
        this.writer = writer;
        this.entryBuffer = new Buffer();
        this.destinationRegistry = destinationRegistry;
    }

    WriterDestination(Writer writer, int entryBufferSize, ClosureAwareDestinationRegistry destinationRegistry) {
        this.writer = writer;
        this.entryBuffer = new Buffer(entryBufferSize);
        this.destinationRegistry = destinationRegistry;
    }

    /**
     * {@inheritDoc}
     */
    public void write(Entry entry) {
        this.entryBuffer.buffer(entry);
    }

    private void handleIOException(IOException e) {
    }

    /**
     * {@inheritDoc}
     */
    public void flush() {
        this.entryBuffer.flush();
    }

    /**
     * {@inheritDoc}
     */
    public void close() {
        if (this.refCount.decrementAndGet() == 0) {
            synchronized (this.outputLock) {
                flush();
                try {
                    this.writer.close();
                } catch (IOException ioe) {
                    handleIOException(ioe);
                }
                this.destinationRegistry.destinationClosed(this);
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    public void open() {
        this.refCount.incrementAndGet();
    }

    private class Buffer {

        private static final int DEFAULT_BUFFER_SIZE = 10;

        private volatile BufferSegment active;

        private volatile BufferSegment inactive;

        private final Object monitor = new Object();

        public Buffer() {
            this(DEFAULT_BUFFER_SIZE);
        }

        public Buffer(int size) {
            this.active = new BufferSegment(size);
            this.inactive = new BufferSegment(size);
        }

        public void buffer(Entry value) {
            BufferSegment toFlush = null;
            synchronized (this.monitor) {
                if(this.active.buffer(value)) {
					toFlush = this.active;
					swapSegments();
				}
            }
            if (toFlush != null) {
                toFlush.flush();
            }
        }

        private void swapSegments() {
            BufferSegment tmp = this.inactive;
            this.inactive = this.active;
            this.active = tmp; // this allows writes to proceed
        }

        public void flush() {
			BufferSegment toFlush;
            synchronized (this.monitor) {
                this.active.suspend();
				toFlush = this.active;
                swapSegments();
            }
            toFlush.flush();
        }

        private class BufferSegment {

            private final Entry[] buffer;

            private final Object monitor = new Object();

            private int position;

            private boolean available = true;

            public BufferSegment(int size) {
                this.buffer = new Entry[size];
            }

            /**
             * Returns true if full
             * 
             * @param e
             * @return
             */
            public boolean buffer(Entry e) {
                synchronized (this.monitor) {
                    while (!this.available) {
                        try {
                            this.monitor.wait();
                        } catch (InterruptedException e1) {
                            continue;
                        }
                    }
                    this.buffer[this.position++] = e;
                    if (this.position == this.buffer.length) {
                        this.available = false;
                    }
                    return !this.available;
                }
            }

            public void suspend() {
                synchronized (this.monitor) {
                    this.available = false;
                }
            }

            public void flush() {
                synchronized (this.monitor) {
                    synchronized (WriterDestination.this.outputLock) {
                        for (int i = 0; i < this.position; i++) {
                            try {
                                WriterDestination.this.writer.append(this.buffer[i].getText()).append(LINE_SEPARATOR);
                            } catch (IOException e) {
                                handleIOException(e);
                            }
                        }

                        try {
                            WriterDestination.this.writer.flush();
                        } catch (IOException ioe) {
                            handleIOException(ioe);
                        }
                        this.position = 0;
                        this.available = true;
                        this.monitor.notify();
                    }
                }
            }
        }
    }
}
