/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.serviceability.tracing;


/**
 * Exposes the context in which method calls are executed.
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * Call context is thread local. Implementation is thread safe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
public final class CallContextHolder {

    private static final ThreadLocal<CallContext> CALL_CONTEXT = new ThreadLocal<CallContext>();

    private static final CallContext UNKNOWN_CONTEXT = new CallContext("unknown", -1);
    
    /**
     * Set the call context.
     * 
     * @param callingMethod the name of the calling method.
     * @param lineNumber the line number of call.
     */
    public static void setContext(String callingMethod, int lineNumber) {
        CALL_CONTEXT.set(new CallContext(callingMethod, lineNumber));
    }

    /**
     * Pop a call from the stack.
     */
    public static void clearContext() {
        CALL_CONTEXT.set(null);
    }

    /**
     * Get the current {@link CallContext}.
     * 
     * @return the current <code>CallContext</code>.
     */
    public static CallContext getContext() {
        return (CALL_CONTEXT.get() == null ? UNKNOWN_CONTEXT : CALL_CONTEXT.get());
    }


    /**
     * Describes the context of a single call. The context comprises the calling method and the line
     * number of the call. <p/>
     * 
     * <strong>Concurrent Semantics</strong><br/>
     * 
     * Implementation is immutable.
     */
    public static final class CallContext {

        private final String callingMethod;

        private final int lineNumber;

        /**
         * @param methodName
         * @param lineNumber
         */
        public CallContext(String methodName, int lineNumber) {
            this.callingMethod = methodName;
            this.lineNumber = lineNumber;
        }

        /**
         * Gets the name of the calling method.
         */
        public String getCallingMethod() {
            return callingMethod;
        }

        /**
         * Gets the line number that the call was made at.
         */
        public int getLineNumber() {
            return lineNumber;
        }
    }
}
