/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.serviceability.tracing;

import com.springsource.server.serviceability.output.Level;

/**
 * Tracer is used by Server code (or aspects) to trace the behaviour of the code.
 * 
 * <strong>Concurrent Semantics</strong><br />
 * Implementations of this interface <strong>must</strong> be thread-safe.
 * 
 * @see Level
 * @author Glyn Normington
 * @author Rob Harrop
 * @author Andy Wilkinson
 * @since 1.0
 */
public interface Tracer {

    /**
     * Trace the specified message at the specified trace level.
     * 
     * @param level the {@link Level} of the trace entry.
     * @param message the English message for the trace entry. Supports {@link java.util.Formatter} format strings.
     * @param inserts inserts for the message.
     */
    void trace(Level level, String message, Object... inserts);

    /**
     * Trace the specified message at the specified trace level. The stack trace of the supplied Throwable will be
     * included in the trace entry.
     * 
     * @param level the {@link Level} of the trace entry.
     * @param message the English message for the trace entry. Supports {@link java.util.Formatter} format strings.
     * @param throwable the throwable to be included in the trace entry
     * @param inserts inserts for the message.
     */
    void trace(Level level, String message, Throwable throwable, Object... inserts);

    /**
     * Trace a serious error such as incorrect behaviour or corrupted internal state. Equivalent to
     * <code>trace(ERROR, message, inserts).</code>
     * 
     * @param message the English message for the trace entry. Supports {@link java.util.Formatter} format strings.
     * @param inserts inserts for the message.
     * 
     * @see #trace(Level, String, Object...)
     * @see Level#ERROR
     */
    void traceError(String message, Object... inserts);

    /**
     * Trace a serious error such as incorrect behaviour or corrupted internal state. The stack trace of the supplied
     * Throwable will be included in the trace entry. Equivalent to
     * <code>trace(ERROR, message, throwable, inserts)</code>
     * 
     * @param message the English message for the trace entry. Supports {@link java.util.Formatter} format strings.
     * @param throwable the throwable to be included in the trace entry
     * @param inserts inserts for the message.
     * 
     * @see #trace(Level, String, Throwable, Object...)
     * @see Level#ERROR
     */
    void traceError(String message, Throwable throwable, Object... inserts);

    /**
     * Trace a warning. Equivalent to <code>trace(WARN, message, inserts)</code>
     * 
     * @param message the English message for the trace entry. Supports {@link java.util.Formatter} format strings.
     * @param inserts inserts for the message.
     * 
     * @see #trace(Level, String, Object...)
     * @see Level#WARN
     */
    void traceWarning(String message, Object... inserts);

    /**
     * Trace a warning. The stack trace of the supplied Throwable will be included in the trace entry. Equivalent to
     * <code>trace(WARN, message, inserts)</code>
     * 
     * @param message the English message for the trace entry. Supports {@link java.util.Formatter} format strings.
     * @param throwable the throwable to be included in the trace entry
     * @param inserts inserts for the message.
     * 
     * @see #trace(Level, String, Throwable, Object...)
     * @see Level#WARN
     */
    void traceWarning(String message, Throwable throwable, Object... inserts);

    /**
     * Trace an informational entry. Equivalent to <code>trace(INFO, message, inserts)</code>
     * 
     * @param message the English message for the trace entry. Supports {@link java.util.Formatter} format strings.
     * @param inserts inserts for the message.
     * 
     * @see #trace(Level, String, Object...)
     * @see Level#INFO
     */
    void traceInfo(String message, Object... inserts);

    /**
     * Trace an informational entry. The stack trace of the supplied Throwable will be included in the trace entry.
     * Equivalent to <code>trace(INFO, message, throwable, inserts)</code>
     * 
     * @param message the English message for the trace entry. Supports {@link java.util.Formatter} format strings.
     * @param throwable the throwable to be included in the trace entry
     * @param inserts inserts for the message.
     * 
     * @see #trace(Level, String, Object...)
     * @see Level#INFO
     */
    void traceInfo(String message, Throwable throwable, Object... inserts);

    /**
     * Trace debugging information. Equivalent to <code>trace(DEBUG, message, inserts)</code>
     * 
     * @param message the English message for the trace entry. Supports {@link java.util.Formatter} format strings.
     * @param inserts inserts for the message.
     * 
     * @see #trace(Level, String, Object...)
     * @see Level#DEBUG
     */
    void traceDebug(String message, Object... inserts);

    /**
     * Trace debugging information. The stack trace of the supplied Throwable will be included in the trace entry.
     * Equivalent to <code>trace(DEBUG, message, throwable, inserts)</code>
     * 
     * @param message the English message for the trace entry. Supports {@link java.util.Formatter} format strings.
     * @param throwable the throwable to be included in the trace entry
     * @param inserts inserts for the message.
     * 
     * @see #trace(Level, String, Throwable, Object...)
     * @see Level#DEBUG
     */
    void traceDebug(String message, Throwable throwable, Object... inserts);

    /**
     * Trace very detailed or costly to produce debugging information. Equivalent to
     * <code>trace(VERBOSE, message, inserts)</code>
     * 
     * @param message the English message for the trace entry. Supports {@link java.util.Formatter} format strings.
     * @param inserts inserts for the message.
     * 
     * @see #trace(Level, String, Throwable, Object...)
     * @see Level#VERBOSE
     */
    void traceVerbose(String message, Object... inserts);

    /**
     * Trace very detailed or costly to produce debugging information. The stack trace of the supplied Throwable will be
     * included in the trace entry. Equivalent to <code>trace(VERBOSE, message, throwable)</code>
     * 
     * @param message the English message for the trace entry. Supports {@link java.util.Formatter} format strings.
     * @param throwable the throwable to be included in the trace entry
     * @param inserts inserts for the message.
     * 
     * @see #trace(Level, String, Throwable, Object...)
     * @see Level#VERBOSE
     */
    void traceVerbose(String message, Throwable throwable, Object... inserts);

    /**
     * Trace method entry at the supplied level.
     * 
     * @param level The level of the trace entry
     * @param message The message for the trace entry. Supports {@link java.util.Formatter} format strings.
     * @param inserts The inserts to be applied to the message.
     */
    void traceEntry(Level level, String message, Object... inserts);

    /**
     * Trace method exit at the supplied level.
     * 
     * @param level The level of the trace entry
     * @param message The message for the trace entry. Supports {@link java.util.Formatter} format strings.
     * @param inserts The inserts to be applied to the message.
     */
    void traceExit(Level level, String message, Object... inserts);

    /**
     * Trace method exit due to the supplied exception at the supplied level.
     * 
     * @param level The level of the trace entry
     * @param message The message for the trace entry. Supports {@link java.util.Formatter} format strings.
     * @param throwable The cause of the method exit
     * @param inserts The inserts to be applied to the message.
     */
    void traceExit(Level level, String message, Throwable throwable, Object... inserts);

    /**
     * Returns true if error level trace is enabled, otherwise false.
     * 
     * @return true if error level trace is enabled
     * @see Level#ERROR
     */
    boolean isErrorEnabled();

    /**
     * Returns true if warning level trace is enabled, otherwise false.
     * 
     * @return true if warning level trace is enabled
     * @see Level#WARN
     */
    boolean isWarningEnabled();

    /**
     * Returns true if info level trace is enabled, otherwise false.
     * 
     * @return true if info level trace is enabled
     * @see Level#INFO
     */
    boolean isInfoEnabled();

    /**
     * Returns true if debug level trace is enabled, otherwise false.
     * 
     * @return true if debug level trace is enabled
     * @see Level#DEBUG
     */
    boolean isDebugEnabled();

    /**
     * Returns true if verbose level trace is enabled, otherwise false.
     * 
     * @return true if verbose level trace is enabled
     * @see Level#VERBOSE
     */
    boolean isVerboseEnabled();
}
