/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.serviceability.tracing;

import java.util.Map;

import com.springsource.server.serviceability.output.Level;


/**
 * <p>
 * The Server's tracing service. <p/>
 * <p>
 * An instance of this interface is available from the OSGi service registry.
 * </p>
 * <p>
 * <strong>Concurrent Semantics</strong><br />
 * Implementations <strong>must</strong> be thread safe.
 * </p>
 * 
 * @author Andy Wilkinson
 * @since 1.0
 */
public interface TracingService {
    /**
     * Informs the tracing service that the calling thread is within the scope of the application identified by
     * <code>applicationName</code>. This setting can be cleared i.e. because the calling thread is no longer in the
     * scope of an application, by calling this method with <code>null</code>.
     * 
     * @param applicationName The name of the application that is in scope on the calling thread.
     */
    void setCurrentApplicationName(String applicationName);

    /**
     * Returns the name of the application that is currently associated with the current thread
     * 
     * @return The name of the application associated with the thread.
     */
    String getCurrentApplicationName();
    
    /**
     * Informs the tracing service that the application with the supplied name has been undeployed. Calling this method
     * provides the tracing service with an opportunity to close and/or release any resources associated with tracing
     * for the application.
     * 
     * @param applicationName The name of the application which has been undeployed.
     */
    void applicationUndeployed(String applicationName);
    
    /**
     * Informs the tracing service that the application with the supplied name has been deployed and that the
     * given levels have been configured for the application.
     * 
     * @param configuredLevels The configured levels for the application's trace.
     */
    void applicationDeployed(String applicationName, Map<String, Level> configuredLevels);
}
