/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.serviceability.tracing.internal;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import com.springsource.server.serviceability.output.Destination;
import com.springsource.server.serviceability.output.DestinationRegistry;
import com.springsource.server.serviceability.output.Entry;
import com.springsource.server.serviceability.output.EntryFilter;
import com.springsource.server.serviceability.output.EntryPublisher;
import com.springsource.server.serviceability.output.EntryPublisherFactory;
import com.springsource.server.serviceability.output.Level;
import com.springsource.server.serviceability.output.internal.StandardEntryFilter;
import com.springsource.server.serviceability.tracing.TraceEntry;

/**
 * EntryPublisherAccessor provides access to the {@link EntryPublisher} that is currently in effect.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * This class is <strong>thread-safe</strong>.
 * 
 * @author Andy Wilkinson
 * @since 1.0
 */
public class EntryPublisherAccessor {

    private final InheritableThreadLocal<String> applicationNames;

    private final Object applicationTracingComponentsLock = new Object();

    private final DestinationRegistry destinationRegistry;

    private final EntryPublisherFactory entryPublisherFactory;

    private final String rootTraceFileLocation;

    private static final FilteringEntryPublisher CREATION_IN_PROGRESS = new FilteringEntryPublisher(null);

    private static final String TRACE_FILE_NAME = "trace";

    private static final Map<String, FilteringEntryPublisher> entryPublisherByApplicationName = new HashMap<String, FilteringEntryPublisher>();
    
    private final Destination globalDestination;
    
    private final FilteringEntryPublisher defaultEntryPublisher;
    
    private final EntryFilter defaultEntryFilter;
    
    private final EntryFilter defaultApplicationEntryFilter;
    
    private final Map<String, Level> globalTraceLevels;

    EntryPublisherAccessor(EntryFilter defaultEntryFilter, EntryPublisher defaultEntryPublisher, Destination globalDestination, InheritableThreadLocal<String> applicationNames,
        DestinationRegistry destinationRegistry, EntryPublisherFactory entryPublisherFactory, String rootTraceFileLocation,
        Map<String, Level> globalSpecificTraceLevels) {
        this.globalDestination = new FilteringDestination(defaultEntryFilter, globalDestination);
        this.defaultEntryPublisher = new FilteringEntryPublisher(defaultEntryPublisher, defaultEntryFilter);
        this.defaultEntryFilter = defaultEntryFilter;
        this.applicationNames = applicationNames;
        this.destinationRegistry = destinationRegistry;
        this.entryPublisherFactory = entryPublisherFactory;
        this.rootTraceFileLocation = rootTraceFileLocation;
        this.globalTraceLevels = globalSpecificTraceLevels;
        
        Map<String, Level> specificLevels = new HashMap<String, Level>();

        specificLevels.put("com.springsource.server.*", Level.WARN);

        if (globalSpecificTraceLevels != null) {
            for (String identifier : globalSpecificTraceLevels.keySet()) {
                specificLevels.put(identifier, globalSpecificTraceLevels.get(identifier));
            }
        }

        this.defaultApplicationEntryFilter = new StandardEntryFilter(specificLevels);
    }

    FilteringEntryPublisher getCurrentEntryPublisher() {
        if (this.applicationNames != null) {
            String applicationName = this.applicationNames.get();
            if (applicationName != null) {
                return getEntryPublisher(applicationName, this.defaultApplicationEntryFilter);
            }
        }
        return this.defaultEntryPublisher;
    }
    
    FilteringEntryPublisher getEntryPublisher(String applicationName, EntryFilter applicationEntryFilter) {
        synchronized (applicationTracingComponentsLock) {
            FilteringEntryPublisher applicationEntryPublisher = entryPublisherByApplicationName.get(applicationName);
            if (applicationEntryPublisher == null) {
                entryPublisherByApplicationName.put(applicationName, CREATION_IN_PROGRESS);
                applicationEntryPublisher = createEntryPublisherForApplication(applicationName, applicationEntryFilter);
                entryPublisherByApplicationName.put(applicationName, applicationEntryPublisher);
            }
            return applicationEntryPublisher;
        }
    }

    private FilteringEntryPublisher createEntryPublisherForApplication(String applicationName, EntryFilter applicationEntryFilter) {
        try {
            Destination applicationTraceDestination = createApplicationTraceDestination(applicationName, applicationEntryFilter);
            Destination[] destinations = new Destination[] { this.globalDestination, applicationTraceDestination };
            return new FilteringEntryPublisher(this.entryPublisherFactory.createEntryPublisher(destinations), this.defaultEntryFilter, applicationEntryFilter);
        } catch (IOException ioe) {
            return this.defaultEntryPublisher;
        }
    }

    private Destination createApplicationTraceDestination(String applicationName, EntryFilter applicationEntryFilter) throws IOException {
        Destination destination;
        if (this.rootTraceFileLocation != null) {
            destination = this.destinationRegistry.createFileDestination(TRACE_FILE_NAME + "-" + applicationName, TRACE_FILE_NAME,
                this.rootTraceFileLocation + File.separatorChar + applicationName);
        } else {
            destination = this.destinationRegistry.createFileDestination(TRACE_FILE_NAME + "-" + applicationName, TRACE_FILE_NAME, applicationName);
        }
        
        destination = new FilteringDestination(applicationEntryFilter, destination);
        return destination;
    }
    
    static class FilteringEntryPublisher implements EntryPublisher {
        
        EntryFilter[] entryFilters;
        
        EntryPublisher entryPublisher;
        
        public FilteringEntryPublisher(EntryPublisher entryPublisher, EntryFilter...entryFilters) {
            this.entryPublisher = entryPublisher;
            this.entryFilters = entryFilters;
        }

        /** 
         * {@inheritDoc}
         */
        public void close() {
            entryPublisher.close();           
        }

        /** 
         * {@inheritDoc}
         */
        public void flush() {
            entryPublisher.flush();
        }

        /** 
         * {@inheritDoc}
         */
        public void publish(Entry entry) {
            entryPublisher.publish(entry);            
        }
        
        boolean filter(Entry entry) {
            for (EntryFilter filter : entryFilters) {
                if (filter.filter(entry)) {
                    return true;
                }
            }
            return false;
        }
        
        boolean filter(String source, Level level) {
            for (EntryFilter filter : entryFilters) {
                if (filter == null || filter.filter(source, level)) {
                    return true;
                }
            }
            return false;
        }
    }

    static class FilteringDestination implements Destination {

        private Destination destination;

        private EntryFilter entryFilter;
        
        FilteringDestination(EntryFilter entryFilter, Destination destination) {
            this.entryFilter = entryFilter;
            this.destination = destination;
        }

        /**
         * {@inheritDoc}
         */
        public void close() {
            destination.close();
        }

        /**
         * {@inheritDoc}
         */
        public void flush() {
            destination.flush();
        }

        /**
         * {@inheritDoc}
         */
        public void write(Entry entry) {
            if (entryFilter.filter(entry)) {
                destination.write(entry);
            }
        }
        
        /**
         * {@inheritDoc}
         */
        public void open() {
            destination.open();
        }
    }

    void applicationUndeployed(String applicationName) {
        if (applicationName != null) {
            synchronized (applicationTracingComponentsLock) {
                EntryPublisher entryPublisher = entryPublisherByApplicationName.remove(applicationName);
                if (entryPublisher != null) {
                    entryPublisher.close();
                }
            }
        }
    }

    void applicationDeployed(String applicationName, Map<String, Level> configuredLevels) {
        Map<String, Level> applicationLevels = new HashMap<String, Level>();
        applicationLevels.putAll(this.globalTraceLevels);
        applicationLevels.putAll(configuredLevels);
        applicationLevels.put("com.springsource.server.*", Level.WARN);
        if (applicationLevels.get("*") == null) {
            applicationLevels.put("*", Level.INFO);
        }
        EntryFilter applicationEntryFilter = new StandardEntryFilter(applicationLevels);
        EntryPublisher publisher = getEntryPublisher(applicationName, applicationEntryFilter);
        TraceEntry entry = new TraceEntry(applicationName, "Trace configuration: " + applicationEntryFilter, null, Level.INFO.toString());
        publisher.publish(entry);
    }
}
