/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.serviceability.tracing.internal;

import com.springsource.server.serviceability.output.EntryFilter;
import com.springsource.server.serviceability.output.EntryPublisher;
import com.springsource.server.serviceability.tracing.Tracer;

/**
 * This class's sole reason for existence is to allow the {@link #setTracingComponentsAccessor(EntryPublisherAccessor)}
 * method to be hidden from users of the serviceability tracing component in an OSGi environment.
 * 
 * <strong>Concurrent Semantics</strong><br />
 * This class is <strong>not</strong> thread-safe.
 * 
 * @author Andy Wilkinson
 * @since 1.0
 */
public abstract class InternalTracerFactory {

    private static volatile EntryPublisherAccessor tracingComponentsAccessor;

    /**
     * Creates a new {@link Tracer} for the given {@link Class}. The {@link EntryFilter} and {@link EntryPublisher}
     * that are used are determined by the previously provided {@link EntryPublisherAccessor}.
     * 
     * @param clazz the name of the class for which a <code>Tracer</code> is required
     * @return the <code>Tracer</code>
     */
    public static Tracer createTracer(Class<?> clazz) {
        return InternalTracerFactory.createTracer(clazz.getName());
    }

    /**
     * Creates a new {@link Tracer} for the class with the given name. The {@link EntryFilter} and
     * {@link EntryPublisher} that are used are determined by the previously provided {@link EntryPublisherAccessor}.
     * 
     * @param className the name of the class for which a <code>Tracer</code> is required
     * @return the <code>Tracer</code>
     */
    public static Tracer createTracer(String className) {
        return new StandardTracer(className, tracingComponentsAccessor);
    }

    /**
     * Sets the {@link EntryPublisherAccessor} to be used by this factory when creating new {@link Tracer Tracers}.
     * 
     * @param accessor The <code>TracingComponentAccessor</code> to use.
     */
    static void setTracingComponentsAccessor(EntryPublisherAccessor accessor) {
        tracingComponentsAccessor = accessor;
    }
    
    static EntryPublisherAccessor getEntryPublisherAccessor() {
        return tracingComponentsAccessor;
    }
}
