/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.serviceability.tracing.internal;

import com.springsource.server.serviceability.output.EntryPublisher;
import com.springsource.server.serviceability.output.Level;
import com.springsource.server.serviceability.tracing.AbstractTracer;
import com.springsource.server.serviceability.tracing.CallContextHolder;
import com.springsource.server.serviceability.tracing.TraceEntry;
import com.springsource.server.serviceability.tracing.CallContextHolder.CallContext;
import com.springsource.server.serviceability.tracing.internal.EntryPublisherAccessor.FilteringEntryPublisher;

/**
 * The state of StandardTracer is immutable, and so the class is thread safe if the provided {@link EntryPublisher} is
 * thread safe. No explicit synchronization is required.
 * 
 * @author Glyn Normington
 * @author Rob Harrop
 * @author Andy Wilkinson
 * @since 1.0
 */
public final class StandardTracer extends AbstractTracer {

    private static final String EXIT_TRACE_INDICATOR = "<";

    private static final String ENTRY_TRACE_INDICATOR = ">";

    private final String className;
    
    private final EntryPublisherAccessor tracingComponentsAccessor;

    public StandardTracer(String className, EntryPublisherAccessor tracingComponentsAccessor) {
        this.className = className;
        this.tracingComponentsAccessor = tracingComponentsAccessor;
    }

    public void trace(Level level, String message, Throwable throwable, Object... inserts) {
        FilteringEntryPublisher publisher = this.tracingComponentsAccessor != null ? this.tracingComponentsAccessor.getCurrentEntryPublisher() : null;
        if (publisher != null && isTraceEnabled(level, className, publisher)) {
            CallContext context = CallContextHolder.getContext();
            String formattedMessage = (inserts.length == 0 ? message : String.format(message, inserts));
            TraceEntry entry = new TraceEntry(this.className, context.getCallingMethod(), context.getLineNumber(), formattedMessage, throwable, level);
            publisher.publish(entry);
        }
    }

    private boolean isTraceEnabled(Level level, String className) {
        if (this.tracingComponentsAccessor != null) {
            return this.tracingComponentsAccessor.getCurrentEntryPublisher().filter(className, level);
        }
        return isTraceEnabled(level, className, this.tracingComponentsAccessor != null ? this.tracingComponentsAccessor.getCurrentEntryPublisher() : null);
    }

    private static boolean isTraceEnabled(Level level, String className, FilteringEntryPublisher filteringEntryPublisher) {
        if (filteringEntryPublisher != null) {
            return filteringEntryPublisher.filter(className, level);
        } else {
            return true;
        }
    }

    public boolean isDebugEnabled() {
        return isTraceEnabled(Level.DEBUG, className);
    }

    public boolean isVerboseEnabled() {
        return isTraceEnabled(Level.VERBOSE, className);
    }

    public boolean isErrorEnabled() {
        return isTraceEnabled(Level.ERROR, className);
    }

    public boolean isInfoEnabled() {
        return isTraceEnabled(Level.INFO, className);
    }

    public boolean isWarningEnabled() {
        return isTraceEnabled(Level.WARN, className);
    }

    public void traceEntry(Level level, String message, Object... inserts) {
        traceEntryExit(level, message, ENTRY_TRACE_INDICATOR, null, inserts);
    }

    public void traceExit(Level level, String message, Object... inserts) {
        traceEntryExit(level, message, EXIT_TRACE_INDICATOR, null, inserts);
    }
    
    public void traceExit(Level level, String message, Throwable throwable, Object... inserts) {
        traceEntryExit(level, message, EXIT_TRACE_INDICATOR, throwable, inserts);
    }

    private void traceEntryExit(Level level, String message, String indicator, Throwable throwable, Object... inserts) {
        FilteringEntryPublisher publisher = this.tracingComponentsAccessor != null ? this.tracingComponentsAccessor.getCurrentEntryPublisher() : null;
        if (publisher != null && isTraceEnabled(level, className, publisher)) {
            String formattedMessage = (inserts.length == 0 ? message : String.format(message, inserts));
            TraceEntry entry = new TraceEntry(this.className, formattedMessage, throwable, level, indicator);
            publisher.publish(entry);
        }
    }
}
