/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.serviceability.tracing.internal;

import java.io.IOException;
import java.io.PrintStream;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;

import org.osgi.framework.BundleContext;
import org.osgi.framework.ServiceRegistration;

import com.springsource.server.serviceability.output.Destination;
import com.springsource.server.serviceability.output.DestinationRegistry;
import com.springsource.server.serviceability.output.EntryFilter;
import com.springsource.server.serviceability.output.EntryFilterFactory;
import com.springsource.server.serviceability.output.EntryPublisher;
import com.springsource.server.serviceability.output.EntryPublisherFactory;
import com.springsource.server.serviceability.output.Level;
import com.springsource.server.serviceability.tracing.TracingService;

/**
 * This class provides bootstrap functionality for the tracing component.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * This class is <strong>thread-safe</strong>
 * 
 * @author Andy Wilkinson
 * @since 1.0
 */
public class TracingBootstrap {

    private static final String TRACING_FILE_DESTINATION_NAME = "trace";

    private static final PrintStream ORIGINAL_SYSTEM_ERR = System.err;

    private static ServiceRegistration tracingServiceRegistration;

    private static EntryFilter tracingEntryFilter;

    private static TracingService tracingService;

    private static boolean started = false;

    private static final Object startedLock = new Object();

    private static boolean servicesPublished = false;

    private static final Object servicesPublishedLock = new Object();

    /**
     * Starts the tracing component. This method should only be invoked once. Duplicate invocations will be ignored.
     */
    public static void start(DestinationRegistry destinationRegistry, EntryFilterFactory entryFilterFactory,
        EntryPublisherFactory entryPublisherFactory, Map<String, String> rawTraceLevels, String tracingDirectory) {
        synchronized (startedLock) {
            if (!started) {
                
                Map<String, Level> traceLevels = convertToLevels(rawTraceLevels);

                Destination traceFile = createTraceDestination(destinationRegistry, tracingDirectory);
                tracingEntryFilter = createEntryFilter(entryFilterFactory, traceLevels);
                InheritableThreadLocal<String> applicationNameThreadLocal = new InheritableThreadLocal<String>();
                EntryPublisher entryPublisher = entryPublisherFactory.createEntryPublisher(new Destination[] { traceFile });

                EntryPublisherAccessor accessor = new EntryPublisherAccessor(tracingEntryFilter, entryPublisher, traceFile,
                    applicationNameThreadLocal, destinationRegistry, entryPublisherFactory, tracingDirectory,
                    traceLevels);

                InternalTracerFactory.setTracingComponentsAccessor(accessor);

                tracingService = new StandardTracingService(tracingEntryFilter, applicationNameThreadLocal);
                started = true;

                TracingPrintStreamWrapper wrapper = new TracingPrintStreamWrapper(System.out, accessor, "System.out", Level.INFO);
                System.setOut(wrapper);

                wrapper = new TracingPrintStreamWrapper(System.err, accessor, "System.err", Level.ERROR);
                System.setErr(wrapper);
            }
        }
    }

    /**
     * Publishes the tracing component's {@link TracingService} to the OSGi Service Registry available from the supplied
     * <code>BundleContext</code>.
     * <p>
     * This method should only be called after {@link #start} has been called. Subsequent invocations without a call to
     * {@link #retractServices()} will have no effect.
     * 
     * @param bundleContext The <code>BundleContext</code> to be used to publish the component's services
     * @see #start
     */
    public static void publishServices(BundleContext bundleContext) {
        synchronized (startedLock) {
            if (!started) {
                throw new IllegalStateException("Start must have been called prior to service publication.");
            }
        }
        synchronized (servicesPublishedLock) {
            if (!servicesPublished) {
                tracingServiceRegistration = bundleContext.registerService(TracingService.class.getName(), tracingService, null);
                servicesPublished = true;
            }
        }
    }

    /**
     * Retracts the service published by a call to {@link #publishServices(BundleContext)}.
     * <p>
     * This method should only be called once per invocation of {@link #publishServices(BundleContext)}
     */
    public static void retractServices() {
        synchronized (servicesPublishedLock) {
            if (servicesPublished) {
                tracingServiceRegistration.unregister();
                servicesPublished = false;
            }
        }
    }

    private static EntryFilter createEntryFilter(EntryFilterFactory entryFilterFactory, Map<String, Level> traceLevels) {

        EntryFilter tracingEntryFilter = entryFilterFactory.createEntryFilter(traceLevels);

        return tracingEntryFilter;
    }

    private static Destination createTraceDestination(DestinationRegistry destinationRegistry, String tracingDirectory) {
        Destination traceDestination;

        try {
            if (tracingDirectory != null) {
                traceDestination = destinationRegistry.createFileDestination(TRACING_FILE_DESTINATION_NAME, TRACING_FILE_DESTINATION_NAME,
                    tracingDirectory);
            } else {
                traceDestination = destinationRegistry.createFileDestination(TRACING_FILE_DESTINATION_NAME, TRACING_FILE_DESTINATION_NAME);
            }
        } catch (IOException ioe) {
            throw new RuntimeException("Unable to create file destination for tracing", ioe);
        }

        return traceDestination;
    }

    private static Map<String, Level> convertToLevels(Map<String, String> specificLevels) {
        Map<String, Level> convertedMap = new HashMap<String, Level>();

        if (specificLevels != null) {
            for (String identifier : specificLevels.keySet()) {
                try {
                    convertedMap.put(identifier, Level.valueOf(specificLevels.get(identifier).toUpperCase(Locale.ENGLISH)));
                } catch (IllegalArgumentException iae) {

                }
            }
        }

        return convertedMap;
    }

    /**
     * Returns the VM's original System.err <code>PrintStream</code> which can be used to output diagnostics in the
     * event of a catastrophic startup failure that prevents logging and/or tracing from being available.
     * 
     * @return The VM's original System.err <code>PrintStream</code>
     */
    public static PrintStream getOriginalSystemErr() {
        return ORIGINAL_SYSTEM_ERR;
    }
}
