/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.servlet.core;

import java.io.File;
import java.util.Set;

import org.osgi.framework.Bundle;

/**
 * <code>ServletContainer</code> defines a simple, generic API for deploying web applications in an OSGi environment.
 * 
 * <p>
 * <em>Context paths</em> must uniquely identify a single web application deployed in the servlet container. Methods
 * which accept a <code>contextPath</code> as an argument should transparently handle paths both with and without a
 * leading slash (i.e., &quot;/&quot;). Likewise, the <code>ROOT</code> context can be specified as &quot;/&quot;,
 * &quot;&quot; (i.e., an empty string), or <code>null</code>.
 * </p>
 * 
 * @author Sam Brannen
 * @author Rob Harrop
 * @since 1.0
 * @see DynamicContext
 */
public interface ServletContainer {

    /**
     * Gets the servlet container's {@link ServletContainerConfiguration configuration settings}, which may be used for
     * reporting, logging, etc.
     * 
     * @return the configuration; never <code>null</code>
     */
    ServletContainerConfiguration getConfiguration();

    /**
     * Gets the name of the servlet container.
     */
    String getName();

    /**
     * Returns the names of the packages upon which the servlet container depends that contain types which may be passed
     * to and/or used by an application running in the container.
     * 
     * @param the names of the packages upon which the servlet container depends that may be used by an application
     */
    String[] getPackageNamesOfDependenciesThatContainApplicationExposedTypes();

    /**
     * Starts the servlet container.
     * 
     * <p>
     * Calling this method has no effect if the servlet container is already in the {@link #isStarted() started state}.
     * </p>
     * 
     * @see #stop()
     * @see #isStarted()
     */
    void start() throws ServletContainerException;

    /**
     * Stops the servlet container.
     * 
     * <p>
     * Calling this method has no effect if the servlet container is not currently in the {@link #isStarted() started
     * state}.
     * </p>
     * 
     * @see #start()
     * @see #isStarted()
     */
    void stop() throws ServletContainerException;

    /**
     * Returns <code>true</code> if the servlet container has been {@link #start() started}.
     * 
     * @see #start()
     * @see #stop()
     */
    boolean isStarted();

    /**
     * Gets the context with the supplied <code>contextPath</code> which is currently associated with the servlet
     * container. Note that the context is not guaranteed to have been <em>deployed</em> or <em>started</em>.
     * 
     * @param contextPath the unique path of the context to retrieve. May be <code>null</code> or <em>empty</em> for the
     *        root context.
     * @return the context with the supplied path, or <code>null</code> if not found.
     * @see #getContexts()
     */
    Context getContext(String contextPath) throws ServletContainerException;

    /**
     * Gets the set of all contexts currently associated with the servlet container. Note that not all contexts in the
     * returned set are guaranteed to have been <em>deployed</em> or <em>started</em>.
     * 
     * @param <C> the type of context returned
     * @return a list of all contexts
     * @see #getContext(String)
     */
    <C extends Context> Set<C> getContexts() throws ServletContainerException;

    /**
     * Deploys the supplied WAR (web application archive) in the servlet container, using the default class loader. The
     * WAR may be supplied as either a compressed JAR (i.e., *.war file) or in an exploded WAR directory structure.
     * 
     * @param contextPath the context path for the new context. May be <code>null</code> or <em>empty</em> for the root
     *        context.
     * @param war the {@link File} referencing the WAR to deploy
     * 
     * @return a newly instantiated and configured context for the deployed WAR
     * @throws ServletContainerException if an error occurs while retrieving the WAR from the supplied {@link File} or
     *         if a context with the provided path already exists in the servlet container.
     * @throws IllegalStateException if the servlet container has not been started.
     * @see #deployWar(String, File, Bundle)
     */
    Context deployWar(String contextPath, File war) throws ServletContainerException, IllegalStateException;

    /**
     * Deploys the supplied WAR (web application archive) in the servlet container. The WAR may be supplied as either a
     * compressed JAR (i.e., *.war file) or in an exploded WAR directory structure. The supplied {@link Bundle} will be
     * used to load classes and retrieve resources within the web application.
     * 
     * @param contextPath the context path for the new context. May be <code>null</code> or <em>empty</em> for the root
     *        context.
     * @param war the {@link File} referencing the WAR to deploy
     * @param contextBundle the bundle to use for context class loading of classes from the OSGi application containing
     *        the web bundle
     * @param webBundle the bundle to use to load classes and retrieve resources within the web application; if
     *        <code>null</code>, the default class loader will be used.
     * 
     * @return a newly instantiated and configured context for the deployed WAR
     * @throws ServletContainerException if an error occurs while retrieving the WAR from the supplied {@link File} or
     *         if a context with the provided path already exists in the servlet container.
     * @throws IllegalStateException if the servlet container has not been started.
     * @see #deployWar(String, File)
     */
    Context deployWar(String contextPath, File war, Bundle contextBundle, Bundle webBundle) throws ServletContainerException, IllegalStateException;

    /**
     * Removes the web application context with the specified <code>contextPath</code>. Calling this method will have no
     * effect if no context is configured or deployed for the specified path.
     * <p>
     * Depending on the implementation, temporary files or directories (e.g., working directories, exploded WAR file
     * contents, etc.) may be removed from the repository (i.e., typically the file system) as well.
     * </p>
     * 
     * @param contextPath the context path of the context to remove; may be <code>null</code> or <em>empty</em> for the
     *        root context.
     * @return <code>true</code> if the context was successfully removed.
     * @throws IllegalStateException if the servlet container has not been started.
     */
    boolean removeContext(String contextPath) throws ServletContainerException, IllegalStateException;

    /**
     * Creates a new {@link DynamicContext context} in the servlet container, using the default <em>context root</em>
     * and the default class loader.
     * 
     * <p>
     * Note: this method does <strong>not</strong> add the created context to the servlet container. Thus, once the
     * context has been properly configured, it must be manually {@link #addContext(DynamicContext) added} to the
     * container.
     * </p>
     * 
     * @param contextPath the unique context path for the new context; may be <code>null</code> or <em>empty</em> for
     *        the root context.
     * @return a newly instantiated and configured context
     * @see #createContext(String, Bundle)
     * @see #createContext(String, File)
     * @see #addContext(DynamicContext)
     * @throws IllegalStateException if the servlet container has not been started.
     */
    DynamicContext createContext(String contextPath) throws ServletContainerException, IllegalStateException;

    /**
     * Creates a new {@link DynamicContext context} in the servlet container, using the default <em>context root</em>.
     * The supplied {@link Bundle} will be used to access classes and resources within the web application.
     * 
     * <p>
     * Note: this method does <strong>not</strong> add the created context to the servlet container. Thus, once the
     * context has been properly configured, it must be manually {@link #addContext(DynamicContext) added} to the
     * container.
     * </p>
     * 
     * @param contextPath the unique context path for the new context; may be <code>null</code> or <em>empty</em> for
     *        the root context.
     * @param contextBundle the bundle to use for context class loading of classes from the OSGi application containing
     *        the web bundle
     * @param webBundle the bundle to use to load classes and retrieve resources within the web application; if
     *        <code>null</code>, the default class loader will be used.
     * @return a newly instantiated and configured context
     * @see #createContext(String, File)
     * @see #addContext(DynamicContext)
     * @throws IllegalStateException if the servlet container has not been started.
     */
    DynamicContext createContext(String contextPath, Bundle contextBundle, Bundle webBundle) throws ServletContainerException, IllegalStateException;

    /**
     * Creates a new {@link DynamicContext context} in the servlet container, using the default class loader.
     * 
     * <p>
     * Note: this method does <strong>not</strong> add the created context to the servlet container. Thus, once the
     * context has been properly configured, it must be manually {@link #addContext(DynamicContext) added} to the
     * container.
     * </p>
     * 
     * @param contextPath the unique context path for the new context; may be <code>null</code> or <em>empty</em> for
     *        the root context.
     * @param contextRoot the {@link File} referencing the directory to be used as the <em>context root</em> for the
     *        context; if <code>null</code>, a default location will be used.
     * @return a newly instantiated and configured context
     * @see #createContext(String)
     * @see #createContext(String, Bundle)
     * @see #createContext(String, File, Bundle)
     * @see #addContext(DynamicContext)
     * @throws IllegalArgumentException if the supplied context root does not reference a directory
     * @throws IllegalStateException if the servlet container has not been started.
     */
    DynamicContext createContext(String contextPath, File contextRoot) throws ServletContainerException, IllegalStateException;

    /**
     * Creates a new {@link DynamicContext context} in the servlet container. The supplied {@link Bundle} will be used
     * to access classes and resources within the web application.
     * 
     * <p>
     * Note: this method does <strong>not</strong> add the created context to the servlet container. Thus, once the
     * context has been properly configured, it must be manually {@link #addContext(DynamicContext) added} to the
     * container.
     * </p>
     * 
     * @param contextPath the unique context path for the new context; may be <code>null</code> or <em>empty</em> for
     *        the root context.
     * @param contextRoot the {@link File} referencing the directory to be used as the <em>context root</em> for the
     *        context; if <code>null</code>, a default location will be used.
     * @param contextBundle the bundle to use for context class loading of classes from the OSGi application containing
     *        the web bundle
     * @param webBundle the bundle to use to load classes and retrieve resources within the web application; if
     *        <code>null</code>, the default class loader will be used.
     * @return a newly instantiated and configured context
     * @see #createContext(String)
     * @see #addContext(DynamicContext)
     * @throws IllegalArgumentException if the supplied context root does not reference a directory
     * @throws IllegalStateException if the servlet container has not been started.
     */
    DynamicContext createContext(String contextPath, File contextRoot, Bundle contextBundle, Bundle webBundle) throws ServletContainerException,
        IllegalStateException;

    /**
     * Adds the supplied web application context to the servlet container, effectively starting and deploying the
     * context.
     * 
     * @param context the context to add.
     * @return <code>true</code> if the context was successfully added; <code>false</code> otherwise
     * 
     * @see #createContext(String)
     * @see #createContext(String, Bundle)
     * @see #createContext(String, File)
     * @see #createContext(String, File, Bundle)
     * @throws ServletContainerException if a context with the provided path already exists.
     * @throws IllegalStateException if the servlet container has not been started.
     */
    boolean addContext(DynamicContext context) throws ServletContainerException, IllegalStateException;

    /**
     * Adds the supplied web application context to the servlet container, effectively starting and deploying the
     * context.
     * 
     * @param context the context to add.
     * @param webBundle the <em>web</em> bundle associated with this web application.
     * @return <code>true</code> if the context was successfully added; <code>false</code> otherwise
     * 
     * @see #createContext(String)
     * @see #createContext(String, Bundle)
     * @see #createContext(String, File)
     * @see #createContext(String, File, Bundle)
     * @throws ServletContainerException if a context with the provided path already exists.
     * @throws IllegalStateException if the servlet container has not been started.
     */
    boolean addContext(DynamicContext context, Bundle webBundle) throws ServletContainerException, IllegalStateException;

    // XXX Consider adding getDynamicContext().

}
