/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.servlet.core;

/**
 * <code>ServletContainerConfiguration</code> encapsulates all configuration parameters required to properly configure
 * and start a {@link ServletContainer}.
 * 
 * <p>
 * Concrete implementations should provide a reasonable {@link Object#toString()} which includes all configuration
 * parameters.
 * </p>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Implementations must be thread-safe and will typically be immutable.
 * 
 * @author Sam Brannen
 * @since 1.0
 */
public interface ServletContainerConfiguration {

    /**
     * Gets the name of the servlet container: used for logging, reporting, etc.
     * 
     * <p>
     * May <strong>not</strong> be <code>null</code> or <em>empty</em>.
     * </p>
     * 
     * @return the name of the servlet container
     */
    String getName();

    /**
     * Gets the path to the servlet container's <em>config</em> directory. This directory serves as the central location
     * for implementation-specific configuration files.
     * 
     * <p>
     * If the configured value does not represent an absolute path, it will be interpreted as a directory relative to
     * the server's home directory.
     * </p>
     * 
     * <p>
     * May <strong>not</strong> be <code>null</code> or <em>empty</em>.
     * </p>
     * 
     * @return the path to the servlet container's config directory
     */
    String getConfigDir();

    /**
     * Gets the path to the <em>work</em> directory where temporary files, compiled JSPs, etc. will be stored.
     * 
     * <p>
     * If the configured value does not represent an absolute path, it will be interpreted as a directory relative to
     * the server's home directory.
     * </p>
     * 
     * <p>
     * If the configured value is <code>null</code> or <em>empty</em>, the servlet container should use its own default
     * location, typically a &quot;work&quot; directory relative to the server's home directory.
     * </p>
     * 
     * @return the path to the work directory
     */
    String getWorkDir();

    /**
     * Gets the host name to use for this servlet container.
     * 
     * <p>
     * If the configured value is <code>null</code> or <em>empty</em>, the servlet container will typically default to
     * using &quot;localhost&quot;.
     * </p>
     * 
     * @return the host name
     */
    String getHostName();

    /**
     * Gets the path to the <em>access log</em> directory where HTTP requests to the servlet container will be logged.
     * 
     * <p>
     * If the configured value does not represent an absolute path, it will be interpreted as a directory relative to
     * the server's home directory.
     * </p>
     * 
     * <p>
     * If the configured value is <code>null</code> or <em>empty</em>, the servlet container should use its own default
     * location, typically a &quot;serviceability/logs/access&quot; directory relative to the server's home directory.
     * </p>
     * 
     * @return the path to the access log directory
     * @see #getAccessLogFormat()
     */
    String getAccessLogDir();

    /**
     * Gets the format to use for formatting <em>access log</em>.
     * 
     * @return the access log format
     * @see #getAccessLogDir()
     */
    String getAccessLogFormat();

    /**
     * Gets whether to use per application logging.
     */
    boolean perApplicationLogging();

}
