/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.servlet.core.util;

import com.springsource.server.serviceability.output.Level;
import com.springsource.server.serviceability.tracing.EntryExitTraceLevel;

/**
 * Utility methods for working in a servlet environment. <p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Thread-safe.
 * 
 * @author Sam Brannen
 * @since 1.0
 */
public final class ServletContainerUtils {

    private ServletContainerUtils() {
        /* no-op */
    }

    /**
     * Converts the supplied <code>contextPath</code> from a raw format to the format suitable for a typical Servlet
     * engine. Specifically, if the supplied value is <code>null</code>, it will be converted to an <em>empty</em>
     * string; if the supplied value is non-null, it will be trimmed and prepended with &quot;/&quot; if necessary.
     * 
     * @param contextPath the context path
     * @return a raw format version of the context path
     */
    @EntryExitTraceLevel(Level.DEBUG)
    public static String convertContextPath(String contextPath) {
        contextPath = contextPath == null ? "" : contextPath.trim();
        if (!contextPath.equals("") && !contextPath.startsWith("/")) {
            contextPath = "/" + contextPath;
        }
        return contextPath;
    }

    /**
     * Gets a <em>raw format</em> version of the supplied <code>contextPath</code>. Specifically, if the supplied
     * value is <code>null</code>, it will be converted to an <em>empty</em> string; if the supplied value is
     * non-null and begins with a &quot;/&quot;, the leading &quot;/&quot; will be removed.
     * 
     * @param contextPath the context path to convert
     * @return a potentially modified path
     */
    @EntryExitTraceLevel(Level.DEBUG)
    public static String getRawContextPath(String contextPath) {
        contextPath = contextPath == null ? "" : contextPath.trim();
        if (!contextPath.equals("") && contextPath.startsWith("/")) {
            contextPath = contextPath.substring(1);
        }
        return contextPath;
    }

    /**
     * Gets a <em>display suitable</em> version of the supplied <code>contextPath</code>. In contrast to the value
     * returned by {@link #convertContextPath(String)}, the context path returned by this method is suitable for
     * displaying to an end user (e.g., in a GUI, log file, etc.).
     * 
     * @param contextPath the raw context path
     * @return a version of the context path suitable for display purposes
     */
    @EntryExitTraceLevel(Level.DEBUG)
    public static String getDisplaySuitableContextPath(String contextPath) {
        contextPath = contextPath == null ? "/" : contextPath.trim();
        if (!contextPath.startsWith("/")) {
            contextPath = "/" + contextPath;
        }
        return contextPath;
    }

    /**
     * Returns <code>true</code> if the supplied port number is valid (i.e., in the range of 0 to 65535).
     */
    @EntryExitTraceLevel(Level.DEBUG)
    public static boolean isValidPort(int portNumber) {
        return portNumber >= 0 && portNumber <= 65535;
    }

}
