/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.servlet.tomcat;

import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;

import org.apache.catalina.Executor;
import org.apache.catalina.LifecycleException;
import org.apache.catalina.LifecycleListener;
import org.apache.catalina.util.LifecycleSupport;

import com.springsource.server.concurrent.core.ExecutorManager;
import com.springsource.server.concurrent.core.ServerExecutorService;
import com.springsource.server.kernel.core.Subsystem;
import com.springsource.server.kernel.core.SubsystemRegistry;
import com.springsource.server.serviceability.output.Level;
import com.springsource.server.serviceability.tracing.EntryExitTraceLevel;

/**
 * {@link Executor} that delegates to a {@link ServerExecutorService}.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Rob Harrop
 * @since 1.0
 */
final class DelegatingExecutor implements Executor {

    private final ExecutorManager executorManager;

    private final SubsystemRegistry subsystemRegistry;

    private final int coreSize;

    private final int maxSize;

    private final int keepAlivePeriod;

    private final String name;

    private final AtomicBoolean started = new AtomicBoolean();

    private volatile ServerExecutorService exec;

    private volatile WorkQueue queue;

    private final LifecycleSupport lifecycle = new LifecycleSupport(this);

    /**
     * Creates a new <code>DelegatingExecutor</code>.
     */
    public DelegatingExecutor(ExecutorManager executorService, SubsystemRegistry subsystemRegistry, int coreSize, int maxSize, int keepAlivePeriod,
        String name) {
        this.executorManager = executorService;
        this.subsystemRegistry = subsystemRegistry;
        this.coreSize = coreSize;
        this.maxSize = maxSize;
        this.keepAlivePeriod = keepAlivePeriod;
        this.name = name;
    }

    /**
     * {@inheritDoc}
     */
    public String getName() {
        return this.name;
    }

    /**
     * {@inheritDoc}
     */
    @EntryExitTraceLevel(Level.DEBUG)
    public void execute(Runnable command) {
        this.exec.execute(command);
    }

    /**
     * {@inheritDoc}
     */
    public void addLifecycleListener(LifecycleListener listener) {
        this.lifecycle.addLifecycleListener(listener);
    }

    /**
     * {@inheritDoc}
     */
    public LifecycleListener[] findLifecycleListeners() {
        return this.lifecycle.findLifecycleListeners();
    }

    /**
     * {@inheritDoc}
     */
    public void removeLifecycleListener(LifecycleListener listener) {
        this.lifecycle.removeLifecycleListener(listener);
    }

    /**
     * {@inheritDoc}
     */
    public void start() {
        if (this.started.compareAndSet(false, true)) {
            this.lifecycle.fireLifecycleEvent(BEFORE_START_EVENT, null);
            this.queue = new WorkQueue();
            this.lifecycle.fireLifecycleEvent(START_EVENT, null);
            this.exec = createExecutorService(this.queue);
            this.lifecycle.fireLifecycleEvent(AFTER_START_EVENT, null);
        }
    }

    /**
     * {@inheritDoc}
     */
    public void stop() throws LifecycleException {
        if (this.started.compareAndSet(true, false)) {
            this.lifecycle.fireLifecycleEvent(BEFORE_STOP_EVENT, null);
            this.lifecycle.fireLifecycleEvent(STOP_EVENT, null);
            if (this.exec != null) {
                this.exec.shutdown();
            }
            this.exec = null;
            this.queue = null;
            this.lifecycle.fireLifecycleEvent(AFTER_STOP_EVENT, null);
        }
    }

    /**
     * Creates the backing {@link ServerExecutorService}.
     */
    private ServerExecutorService createExecutorService(WorkQueue queue) {
        Subsystem subsystem = this.subsystemRegistry.getSubsystemByClass(TomcatServletContainer.class);
        ServerExecutorService executorService = this.executorManager.newExecutorService(this.name, subsystem, this.coreSize, this.maxSize,
            this.keepAlivePeriod, TimeUnit.MILLISECONDS, queue);
        queue.executorService = executorService;
        return executorService;
    }

    ServerExecutorService getBackingExecutor() {
        return this.exec;
    }

    BlockingQueue<Runnable> getQueue() {
        return this.queue;
    }

    /**
     * Unbounded work queue that forces thread creation beyond core size. <p/>
     * 
     * <strong>Concurrent Semantics</strong><br />
     * 
     * Threadsafe.
     * 
     * @author Rob Harrop
     * @since 1.0
     */
    private static class WorkQueue extends LinkedBlockingQueue<Runnable> {

        private static final long serialVersionUID = 34918240912834L;

        volatile ServerExecutorService executorService;

        /**
         * {@inheritDoc}
         */
        @EntryExitTraceLevel(Level.VERBOSE) @Override public boolean offer(Runnable o) {
            if (this.executorService != null) {
                if (this.executorService.getPoolSize() == this.executorService.getMaximumPoolSize()) {
                    // no free threads - enqueue
                    return super.offer(o);
                } else {
                    return false;
                }
            } else {
                return super.offer(o);
            }
        }

    }

}
