/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.servlet.tomcat;

import java.util.List;
import java.util.Map;

import com.springsource.server.servlet.core.ServletContainerConfiguration;

/**
 * <code>TomcatServletContainerConfiguration</code> encapsulates all configuration parameters required to properly
 * configure and start a {@link TomcatServletContainer}.
 * 
 * <p>
 * Concrete implementations should provide a reasonable {@link Object#toString()} which includes all configuration
 * parameters.
 * </p>
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Implementations must be thread-safe and will typically be immutable.
 * 
 * @author Sam Brannen
 * @author Ben Hale
 * @since 1.0
 */
public interface TomcatServletContainerConfiguration extends ServletContainerConfiguration {

    /**
     * Gets the path to the servlet container's <em>application</em> directory. This directory serves as the root
     * directory of all web applications deployed in the servlet container.
     * 
     * <p>
     * If the configured value does not represent an absolute path, it will be interpreted as a directory relative to
     * the server's home directory.
     * </p>
     * 
     * <p>
     * If the configured value is <code>null</code> or <em>empty</em>, the servlet container should use its own default
     * location, typically a &quot;webapps&quot; directory relative to the server's home directory.
     * </p>
     * 
     * @return the path to the application directory
     */
    String getApplicationDir();

    /**
     * Gets the list of property maps for configured connectors.
     * 
     * @return the connectors
     */
    List<Map<String, String>> getConnectors();

    /**
     * Gets the list of property maps for configured lifecycle listeners.
     * 
     * @return the listeners
     */
    List<Map<String, String>> getListeners();

    /**
     * Gets the JVM route to support load-balancing.
     * 
     * @return the JVM route
     */
    String getJvmRoute();

    /**
     * Gets the location of the &quot;<code>tomcat-users.xml</code>&quot; file.
     */
    String getTomcatUsersXml();

    /**
     * Gets the location of the default &quot;<code>web.xml</code>&quot; file.
     */
    String getDefaultWebXml();

    /**
     * Gets the location of the default &quot;<code>context.xml</code>&quot; file.
     */
    String getDefaultContextXml();

    /**
     * Gets the minimum size of the thread pool.
     */
    int getThreadPoolMinSize();

    /**
     * Gets the maximum size of the thread pool.
     */
    int getThreadPoolMaxSize();

    /**
     * Gets the period of time before an idle thread is removed.
     */
    int getThreadPoolKeepAlivePeriod();

    /**
     * Gets the property map for cluster configuration.
     */
    Map<String, ? extends Object> getCluster();
}
