/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.servlet.tomcat.internal;

import java.io.IOException;

import javax.servlet.ServletException;

import org.apache.catalina.connector.Request;
import org.apache.catalina.connector.Response;
import org.apache.catalina.valves.ValveBase;

import com.springsource.server.serviceability.output.Level;
import com.springsource.server.serviceability.tracing.EntryExitTraceLevel;
import com.springsource.server.serviceability.tracing.TracingService;
import com.springsource.server.servlet.core.DeployedWebApplicationNamesRegistry;

/**
 * A Tomcat <code>Valve</code> that sits in the request-response pipeline and sets the name of the application at
 * which the request is targeted during request processing and then clears the application name during response
 * processing.
 * 
 * <strong>Concurrent Semantics</strong><br />
 * This class is as thread-safe as <code>ValveBase</code>.
 * 
 * @author Andy Wilkinson
 * @author Sam Brannen
 * @since 1.0
 */
public class ApplicationNameTrackingValve extends ValveBase {

    private final TracingService tracingService;

    private final DeployedWebApplicationNamesRegistry deployedWebApplicationNamesRegistry;

    /**
     * Creates a new <code>ApplicationNameTrackingValve</code> that will inform the supplied {@link TracingService} of
     * the current application's name. The application name will be determined using the supplied
     * {@link DeployedWebApplicationNamesRegistry}.
     * 
     * @param tracingService The tracing service to be informed of the current application name.
     * @param deployedWebApplicationNamesRegistry The registry to use to determine the application's name.
     */
    public ApplicationNameTrackingValve(TracingService tracingService, DeployedWebApplicationNamesRegistry deployedWebApplicationNamesRegistry) {
        this.tracingService = tracingService;
        this.deployedWebApplicationNamesRegistry = deployedWebApplicationNamesRegistry;
    }

    /**
     * {@inheritDoc}
     */
    @EntryExitTraceLevel(Level.VERBOSE)
    @Override
    public void invoke(Request request, Response response) throws IOException, ServletException {
    	String contextPath = request.getContextPath().length() == 0 ? "/" : request.getContextPath();
        this.tracingService.setCurrentApplicationName(this.deployedWebApplicationNamesRegistry.getWebApplicationName(contextPath));
        getNext().invoke(request, response);
        this.tracingService.setCurrentApplicationName(null);
    }
}
