/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.servlet.tomcat.internal;

import org.springframework.util.ObjectUtils;

import com.springsource.server.serviceability.ffdc.DumpExtension;
import com.springsource.server.serviceability.ffdc.DumpWriter;
import com.springsource.server.serviceability.ffdc.DumpWriter.ListWriter;
import com.springsource.server.serviceability.ffdc.DumpWriter.SectionWriter;
import com.springsource.server.servlet.core.Context;
import com.springsource.server.servlet.core.ServletContainer;
import com.springsource.server.servlet.core.ServletContainerConfiguration;

/**
 * Custom implementation of {@link DumpExtension} which contributes {@link ServletContainer}-specific information for
 * the &quot;servlet&quot; subsystem generated dump files.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Thread-safe as long as {@link ServletContainer#getConfiguration()} and {@link ServletContainer#getContexts()} are
 * thread-safe.
 * 
 * @author Sam Brannen
 * @since 1.0
 */
public class ServletContainerDumpExtension implements DumpExtension {

    private final ServletContainer servletContainer;

    /**
     * Constructs a ServletContainerDumpExtension.
     */
    public ServletContainerDumpExtension(final ServletContainer servletContainer) {
        this.servletContainer = servletContainer;
    }

    /**
     * {@inheritDoc}
     */
    public void dump(SectionWriter sectionWriter) {
        dumpConfiguration(sectionWriter);
        dumpContexts(sectionWriter);
    }

    /**
     * Dumps the {@link ServletContainerConfiguration} to the supplied {@link DumpWriter}.
     * 
     * @param dumpWriter the DumpWriter to write to.
     */
    private void dumpConfiguration(SectionWriter sectionWriter) {
        ServletContainerConfiguration config = this.servletContainer.getConfiguration();
        sectionWriter//
        // start map
        .map("Configuration")//
        .scalar("Name", config.getName())//
        .scalar("Host Name", config.getHostName())//
        .scalar("Config Directory", config.getConfigDir())//
        .scalar("Work Directory", config.getWorkDir())//
        .scalar("Access Log Directory", config.getAccessLogDir())//
        .scalar("Access Log Format", config.getAccessLogFormat())//
        // end map
        .end();
    }

    /**
     * Dumps information regarding configured and/or deployed {@link Context contexts} to the supplied
     * {@link DumpWriter}.
     * 
     * @param dumpWriter the DumpWriter to write to.
     */
    private void dumpContexts(SectionWriter sectionWriter) {
        ListWriter<SectionWriter> list = sectionWriter.list("Web Application Contexts");
        int i = 1;
        for (Context context : this.servletContainer.getContexts()) {
            list.map().scalar(String.format("context #%03d", i++), ObjectUtils.nullSafeToString(context));
        }
        list.end();
    }

    /**
     * {@inheritDoc}
     */
    public String getName() {
        return this.getClass().getCanonicalName().replace(".tomcat.internal.ServletContainerDumpExtension", "");
    }

}
