/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.servlet.tomcat.loader;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.lang.instrument.ClassFileTransformer;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.sql.Driver;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.Vector;

import org.apache.catalina.Lifecycle;
import org.apache.catalina.LifecycleException;
import org.apache.catalina.LifecycleListener;
import org.apache.catalina.loader.Constants;
import org.apache.catalina.util.StringManager;
import org.apache.tomcat.util.IntrospectionUtils;
import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.ServiceReference;
import org.springframework.osgi.util.BundleDelegatingClassLoader;
import org.springframework.util.Assert;

import com.springsource.server.osgi.framework.InstrumentableClassLoader;
import com.springsource.server.osgi.framework.OsgiFramework;
import com.springsource.server.osgi.framework.support.FilePathSanitiser;
import com.springsource.server.serviceability.output.Level;
import com.springsource.server.serviceability.tracing.EntryExitTraceLevel;

/**
 * TODO Document OsgiWebappClassLoader
 * 
 * @author Sam Brannen
 * @author Rob Harrop
 * @author Andy Wilkinson
 * @author Remy Maucherat
 * @author Craig R. McClanahan
 */
public class OsgiWebappClassLoader extends URLClassLoader implements Lifecycle {

    /**
     * The string manager for this package.
     */
    protected static final StringManager sm = StringManager.getManager(Constants.Package);

    protected static org.apache.juli.logging.Log log = org.apache.juli.logging.LogFactory.getLog(OsgiWebappClassLoader.class);

    private static final String MODULE_INF_WEB_INF_CLASSES = "MODULE-INF/WEB-INF/classes/";

    private static final String WEB_INF_CLASSES = "WEB-INF/classes/";

    private static final String PACKED = "/packed/";

    private static final String FILE_SCHEME = "file:";

    private static final Set<String> TOMCAT_BUNDLE_NAMES = new HashSet<String>(Arrays.asList("com.springsource.org.apache.catalina"));

    private final Bundle bundle;

    private final Bundle webBundle;

    private final InstrumentableClassLoader instrumentable;

    /**
     * Has this class loader been started?
     */
    private volatile boolean started = false;

    /**
     * The delegating class loader for the web application's bundle(s).
     */
    private final BundleDelegatingClassLoader bundleDelegatingClassLoader;

    private final Bundle[] tomcatBundles;

    private final Bundle[] dependencies;

    private final Bundle[] webDependencies;

    // XXX Consider caching loaded resources (e.g., using org.apache.catalina.loader.ResourceEntry or similar).

    // ------------------------------------------------------------------------
    // --- Constructors
    // ------------------------------------------------------------------------

    /**
     * TODO Document OsgiWebappClassLoader(Bundle bundle, Bundle webBundle)
     * 
     * @param bundle
     * @param webBundle
     */
    public OsgiWebappClassLoader(Bundle bundle, Bundle webBundle) {
        super(getBundleURLs(bundle, webBundle));
        Assert.notNull(bundle, "'bundle' must not be null");
        Assert.notNull(webBundle, "'webBundle' must not be null");
        this.tomcatBundles = findTomcatBundles(bundle.getBundleContext());
        this.webDependencies = getDependencies(webBundle, false);
        this.dependencies = getDependencies(bundle, false);

        ServiceReference serviceReference = bundle.getBundleContext().getServiceReference(OsgiFramework.class.getName());
        OsgiFramework framework = (OsgiFramework) bundle.getBundleContext().getService(serviceReference);
        ClassLoader bundleClassLoader = framework.getBundleClassLoader(bundle);

        if (bundleClassLoader instanceof InstrumentableClassLoader) {
            this.instrumentable = (InstrumentableClassLoader) bundleClassLoader;
        } else {
            this.instrumentable = null;
        }

        this.bundleDelegatingClassLoader = BundleDelegatingClassLoader.createBundleClassLoaderFor(webBundle,
            BundleDelegatingClassLoader.createBundleClassLoaderFor(bundle));
        this.bundle = bundle;
        this.webBundle = webBundle;
    }

    private static URL[] getBundleURLs(Bundle bundle, Bundle webBundle) {

        List<URL> urls = new ArrayList<URL>();

        URL webBundleLocation = getBundleLocationAsURL(webBundle);
        if (webBundleLocation != null) {
            urls.add(webBundleLocation);
        }

        urls.addAll(getLocationsOfDependencies(webBundle));
        urls.addAll(getLocationsOfDependencies(bundle));

        String bundleClassPath = (String) webBundle.getHeaders().get("Bundle-Classpath");
        if (bundleClassPath != null) {
            File bundleFile = new File(webBundle.getLocation().replace(FILE_SCHEME, ""));
            for (String entry : bundleClassPath.split(",")) {
                if (entry.equals(".")) {
                    continue;
                }

                File entryFile = new File(bundleFile.getAbsolutePath() + File.separator + entry);
                if (!entryFile.exists()) {
                    continue;
                }
                try {
                    urls.add(entryFile.toURL());
                } catch (MalformedURLException e) {
                    throw new RuntimeException("Unable to get bundle classpath entry URL.", e);
                }
            }
        }

        if (log.isDebugEnabled())
            log.debug("Bundle URLs for bundle '" + bundle + "' with web bundle '" + webBundle + "' are: " + urls);

        return (URL[]) urls.toArray(new URL[urls.size()]);
    }

    private static URL getBundleLocationAsURL(Bundle bundle) {

        String location = FilePathSanitiser.sanitise(bundle.getLocation());

        try {
            if (new File(new URL(location).getFile()).isDirectory()) {

                String jarLocation = location.substring(0, location.lastIndexOf('/')) + PACKED + location.substring(location.lastIndexOf('/') + 1);

                if (new File(new URL(jarLocation).getFile()).exists()) {
                    return new URL(jarLocation);
                } else {
                    // Cope with the possibility that the web transform step has renamed a .jar to a .war
                    if (jarLocation.endsWith(".war")) {
                        String jarExtensionLocation = jarLocation.substring(0, jarLocation.length() - 4) + ".jar";
                        if (new File(new URL(jarExtensionLocation).getFile()).exists()) {
                            return new URL(jarExtensionLocation);
                        }
                    }

                    String alternativeJarLocation = location.substring(0, location.lastIndexOf('/')) + "/../.." + PACKED
                        + location.substring(location.lastIndexOf('/') + 1);

                    if (new File(new URL(alternativeJarLocation).getFile()).exists()) {
                        return new URL(alternativeJarLocation);
                    } else {
                        // XXX Consider packing exploded locations that contain TLDs so that Tomcat can find them
                        log.warn("The bundle at '" + location
                            + "' has been skipped as it is in exploded form and a packed equivalent could not be found at '" + jarLocation + "' or '"
                            + alternativeJarLocation + "'");
                        return null;
                    }
                }
            } else {
                return new URL(location);
            }
        } catch (MalformedURLException murle) {
            throw new RuntimeException("Unable to create location URL for bundle located at '" + location + "'");
        }
    }

    private static Bundle[] getDependencies(Bundle bundle, boolean includeFragments) {
        ServiceReference serviceReference = bundle.getBundleContext().getServiceReference(OsgiFramework.class.getName());
        OsgiFramework framework = (OsgiFramework) bundle.getBundleContext().getService(serviceReference);

        Bundle[] dependencies = framework.getDirectDependencies(bundle, includeFragments);

        bundle.getBundleContext().ungetService(serviceReference);

        return dependencies;
    }

    private static List<URL> getLocationsOfDependencies(Bundle bundle) {
        List<URL> locations = new ArrayList<URL>();

        Bundle[] dependencies = getDependencies(bundle, true);

        for (Bundle dependency : dependencies) {
            if (dependency.getBundleId() > 0) {
                URL location = getBundleLocationAsURL(dependency);
                if (location != null) {
                    locations.add(location);
                }
            }
        }

        return locations;
    }

    public void addTransformer(ClassFileTransformer transformer) {
        if (this.instrumentable != null) {
            this.instrumentable.addClassFileTransformer(transformer);
        }
    }

    public ClassLoader getThrowawayClassLoader() {
        return this.instrumentable.createThrowAway();
    }

    /**
     * @param bundleContext
     * @return
     */
    private Bundle[] findTomcatBundles(BundleContext bundleContext) {
        List<Bundle> tomcatBundles = new ArrayList<Bundle>();
        Bundle[] allBundles = bundleContext.getBundles();
        for (Bundle bundle : allBundles) {
            String symName = bundle.getSymbolicName();
            if (TOMCAT_BUNDLE_NAMES.contains(symName)) {
                tomcatBundles.add(bundle);
            }
        }
        return (Bundle[]) tomcatBundles.toArray(new Bundle[tomcatBundles.size()]);
    }

    // -------------------------------------------------------------------------
    // --- Lifecycle
    // -------------------------------------------------------------------------

    /**
     * {@inheritDoc}
     */
    public void addLifecycleListener(LifecycleListener listener) {
        /* no-op */
    }

    /**
     * {@inheritDoc}
     */
    public LifecycleListener[] findLifecycleListeners() {
        return new LifecycleListener[0];
    }

    /**
     * {@inheritDoc}
     */
    public void removeLifecycleListener(LifecycleListener listener) {
        /* no-op */
    }

    /**
     * {@inheritDoc}
     */
    public void start() throws LifecycleException {
        this.started = true;
    }

    /**
     * {@inheritDoc}
     */
    public void stop() throws LifecycleException {
        // Clearing references should be done before setting started to
        // false, due to possible side effects
        clearReferences();

        this.started = false;
    }

    // -------------------------------------------------------------------------
    // --- ClassLoader
    // -------------------------------------------------------------------------

    /**
     * Find the resource with the given name. A resource is some data (images, audio, text, etc.) that can be accessed
     * by class code in a way that is independent of the location of the code. The name of a resource is a "/"-separated
     * path name that identifies the resource. If the resource cannot be found, return <code>null</code>.
     * <p>
     * This method searches according to the following algorithm, returning as soon as it finds the appropriate URL. If
     * the resource cannot be found, returns <code>null</code>.
     * <ul>
     * <li>???</li>
     * </ul>
     * 
     * @param name Name of the resource to return a URL for
     */
    @EntryExitTraceLevel(Level.DEBUG)
    @Override
    public URL getResource(String name) {
        if (log.isDebugEnabled()) {
            log.debug("getResource(" + name + ")");
        }

        URL url = null;

        url = webBundle.getResource(MODULE_INF_WEB_INF_CLASSES + name);
        if (url != null) {
            return url;
        }

        url = webBundle.getResource(WEB_INF_CLASSES + name);
        if (url != null) {
            return url;
        }

        url = webBundle.getResource(name);
        if (url != null) {
            return url;
        }

        for (Bundle bundle : this.tomcatBundles) {
            url = bundle.getResource(name);
            if (url != null) {
                return url;
            }
        }

        url = bundle.getResource(name);
        if (url != null) {
            return url;
        }

        for (Bundle bundle : this.webDependencies) {
            url = bundle.getResource(name);
            if (url != null) {
                return url;
            }
        }

        for (Bundle bundle : this.dependencies) {
            url = bundle.getResource(name);
            if (url != null) {
                return url;
            }
        }

        // Resource was not found
        if (log.isDebugEnabled()) {
            log.debug("  --> Resource not found, returning null");
        }

        return null;
    }

    @EntryExitTraceLevel(Level.DEBUG)
    public InputStream getResourceAsStream(String name) {
        return super.getResourceAsStream(name);
    }

    /**
     * {@inheritDoc}
     */
    @EntryExitTraceLevel(Level.DEBUG)
    @Override
    public Enumeration<URL> getResources(String name) throws IOException {
        if (log.isDebugEnabled()) {
            log.debug("getResources(" + name + ")");
        }

        Vector<URL> found = new Vector<URL>();

        addResourcesFoundInBundle(name, webBundle, found);
        addResourcesFoundInBundle(name, bundle, found);
        for (Bundle bundle : this.webDependencies) {
            addResourcesFoundInBundle(name, bundle, found);
        }
        for (Bundle bundle : this.dependencies) {
            addResourcesFoundInBundle(name, bundle, found);
        }
        for (Bundle bundle : this.tomcatBundles) {
            addResourcesFoundInBundle(name, bundle, found);
        }

        // Resource was not found
        if (log.isDebugEnabled()) {
            log.debug("  --> Resource '" + name + " not found, returning null");
        }
        return found.elements();
    }

    @SuppressWarnings("unchecked")
    private void addResourcesFoundInBundle(String name, Bundle bundle, List<URL> found) throws IOException {
        if (bundle.getState() != Bundle.UNINSTALLED) {
            Enumeration<URL> resources = bundle.getResources(name);
            if (resources != null) {
                addAll(found, resources);
            }
        }
    }

    private <T> void addAll(Collection<T> collection, Enumeration<T> enumeration) {
        if (enumeration == null) {
            return;
        }
        while (enumeration.hasMoreElements()) {
            T t = (T) enumeration.nextElement();
            collection.add(t);
        }
    }

    /**
     * Load the class with the specified name, searching using the following algorithm until it finds and returns the
     * class. If the class cannot be found, throws <code>ClassNotFoundException</code>.
     * <ul>
     * <li>Check cache of previously loaded classes</li>
     * <li>If not found, attempt to load the class from the Application's synthetic context bundle (and indirectly from
     * the Web module's bundle)</li>
     * <li>If not found, attempt to load the class from Tomcat's bundles</li>
     * <li>If not found, throw a <code>ClassNotFoundException</code></li>
     * </ul>
     * If the class was found using the above steps, and the <code>resolve</code> flag is <code>true</code>, this
     * method will then call <code>resolveClass(Class)</code> on the resulting Class object.
     * 
     * @param name Name of the class to be loaded
     * @param resolve If <code>true</code> then resolve the class
     * 
     * @exception ClassNotFoundException if the class was not found
     */
    @EntryExitTraceLevel(Level.DEBUG)
    @Override
    public Class<?> loadClass(String name, boolean resolve) throws ClassNotFoundException {
        if (log.isDebugEnabled()) {
            log.debug("loadClass(" + name + ", " + resolve + ")");
        }

        Class<?> clazz = null;

        // Log access to stopped classloader
        if (!this.started) {
            try {
                throw new IllegalStateException();
            } catch (IllegalStateException e) {
                log.info(sm.getString("webappClassLoader.stopped", name), e);
            }
        }

        // Check our previously loaded class cache
        clazz = findLoadedClass(name);
        if (clazz != null) {
            if (log.isDebugEnabled()) {
                log.debug("  Returning class from cache");
            }
            if (resolve) {
                resolveClass(clazz);
            }
            return clazz;
        }

        // Search the application's bundle
        if (log.isDebugEnabled()) {
            log.debug("  Searching the application's bundle");
        }
        try {
            clazz = this.bundleDelegatingClassLoader.loadClass(name);
            if (clazz != null) {
                if (log.isDebugEnabled()) {
                    log.debug("  Loading class from the delegating classloader");
                }
                if (resolve) {
                    resolveClass(clazz);
                }
                return clazz;
            }
        } catch (ClassNotFoundException e) {
            /* no-op */
        }

        for (Bundle bundle : this.tomcatBundles) {
            try {
                clazz = bundle.loadClass(name);
                if (clazz != null) {
                    if (log.isDebugEnabled()) {
                        log.debug("  Loading class from bundle '" + bundle.getSymbolicName() + "'.");
                    }
                    if (resolve) {
                        resolveClass(clazz);
                    }
                    return clazz;
                }
            } catch (ClassNotFoundException ex) {
                /* no-op */
            }
        }

        throw new ClassNotFoundException(name);
    }

    // -------------------------------------------------------------------------
    // --- Protected Methods
    // -------------------------------------------------------------------------

    /**
     * Clear references.
     */
    protected void clearReferences() {

        // Unregister any JDBC drivers loaded by this classloader
        Enumeration<Driver> drivers = DriverManager.getDrivers();
        while (drivers.hasMoreElements()) {
            Driver driver = (Driver) drivers.nextElement();
            if (driver.getClass().getClassLoader() == this) {
                try {
                    DriverManager.deregisterDriver(driver);
                } catch (SQLException e) {
                    log.warn("SQL driver deregistration failed", e);
                }
            }
        }

        // XXX Determine if we need to null out static or final fields from loaded classes as in WebappClassLoader

        // Clear the IntrospectionUtils cache.
        IntrospectionUtils.clear();

        // Clear the classloader reference in common-logging
        org.apache.juli.logging.LogFactory.release(this);

        // Clear the classloader reference in the VM's bean introspector
        java.beans.Introspector.flushCaches();
    }
}
