/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.web.core;

import java.util.Set;
import java.util.jar.Attributes;

import com.springsource.server.deployer.core.DeploymentContext;
import com.springsource.server.deployer.core.Module;
import com.springsource.server.deployer.core.ModuleDeployer;
import com.springsource.server.kernel.core.BundleStarter;
import com.springsource.server.osgi.manifest.BundleManifestFactory;
import com.springsource.server.osgi.manifest.VersionRange;
import com.springsource.server.servlet.core.DeployedWebApplicationNamesRegistry;
import com.springsource.server.servlet.core.ServletContainer;

/**
 * This deployer implements {@link ModuleDeployer} with <em>web module</em> deployment capability.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * This class is thread-safe.
 * 
 * @author Glyn Normington
 * @author Sam Brannen
 * @since 1.0
 */
class WebDeployer implements ModuleDeployer {

    private static final String JAR_EXTENSION = ".jar";

    private final BundleManifestFactory bundleManifestFactory;

    private final ServletContainer servletContainer;

    private final DeployedWebApplicationNamesRegistry deployedWebApplicationNamesRegistry;

    private final BundleStarter bundleStarter;

    public WebDeployer(BundleManifestFactory bundleManifestFactory, ServletContainer servletContainer,
        DeployedWebApplicationNamesRegistry deployedWebApplicationNamesRegistry, BundleStarter bundleStarter) {
        this.bundleManifestFactory = bundleManifestFactory;
        this.servletContainer = servletContainer;
        this.deployedWebApplicationNamesRegistry = deployedWebApplicationNamesRegistry;
        this.bundleStarter = bundleStarter;
    }

    /**
     * Determines if the supplied {@link Module} is a <em>web module</em>.
     * 
     * @return <code>true</code> if the supplied module is <em>web module</em>, otherwise <code>false</code>
     */
    static boolean isWebModule(Module module) {
        Attributes attributes = module.getModuleData().getBundleManifest().toManifest().getMainAttributes();
        String moduleType = attributes.getValue(MODULE_TYPE_MANIFEST_HEADER);

        // Detecting web modules as JAR files which have the 'Module-Type' manifest header set to 'Web'.
        return module.getModuleData().getFilename().toLowerCase().endsWith(JAR_EXTENSION)
            && WebModulePersonality.PERSONALITY.equalsIgnoreCase(moduleType);
    }

    /**
     * {@inheritDoc}
     */
    public int getOrder() {
        return WEB_ORDER;
    }

    /**
     * {@inheritDoc}
     */
    public boolean canDeploy(DeploymentContext deploymentContext, Module module) {
        if (isWebModule(module)) {
            module.setPersonality(new WebModulePersonality(this.bundleManifestFactory, this.servletContainer, module,
                this.deployedWebApplicationNamesRegistry, deploymentContext, this.bundleStarter));
            return true;
        }
        return false;
    }

    /**
     * Always returns &quot;{@link WebModulePersonality#PERSONALITY}&quot;.
     */
    public final String getPersonality() {
        return WebModulePersonality.PERSONALITY;
    }

    /**
     * {@inheritDoc}
     */
    public Set<PersonalityBundle> getBundles(DeploymentContext deploymentContext) {
        // Set<PersonalityBundle> personalityBundles = new HashSet<PersonalityBundle>();
        // personalityBundles.add(new WebDmPersonalityBundle());
        // return personalityBundles;

        return null;
    }

    /**
     * {@link PersonalityBundle} descriptor for the &quot;<code>com.springsource.server.web.dm</code>&quot;
     * bundle.
     * <p />
     * 
     * <strong>Concurrent Semantics</strong><br />
     * 
     * Immutable and therefore thread-safe.
     * 
     * @author Sam Brannen
     * @since 1.0
     */
    @SuppressWarnings("unused")
    private static class WebDmPersonalityBundle implements PersonalityBundle {

        private final String bundleSymbolicName = "com.springsource.server.web.dm";

        private final VersionRange versionRange = new VersionRange("[1.0,2.0)");

        public final String getBundleSymbolicName() {
            return this.bundleSymbolicName;
        }

        public final com.springsource.server.osgi.manifest.VersionRange getBundleVersionRange() {
            return this.versionRange;
        }
    }

}
