/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.web.core.internal;

import static com.springsource.server.servlet.core.util.ServletContainerUtils.getDisplaySuitableContextPath;
import static com.springsource.server.web.core.internal.WebLogEvent.WEB_BUNDLE_FAILED_TO_PUBLISH_APPLICATION_CONTEXT;
import static com.springsource.server.web.core.internal.WebLogEvent.WEB_BUNDLE_FAILED_TO_PUBLISH_APPLICATION_CONTEXT_WITHIN_TIMEOUT;

import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import javax.servlet.ServletContext;

import org.osgi.framework.Bundle;
import org.springframework.context.ApplicationContext;
import org.springframework.util.Assert;

import com.springsource.server.deployer.core.DeploymentException;
import com.springsource.server.deployer.core.Module;
import com.springsource.server.kernel.core.BundleStarter;
import com.springsource.server.serviceability.ffdc.NonNull;
import com.springsource.server.serviceability.logging.LogService;
import com.springsource.server.serviceability.logging.LogServiceFactory;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;
import com.springsource.server.servlet.core.Context;

/**
 * <code>ApplicationContextRetrievingAndPublishingCallback</code> serves as a <em>callback</em> mechanism within the
 * deployment pipeline which attempts to retrieve a {@link Module module's} {@link ApplicationContext} from the OSGi
 * service registry and publish it via the web module's {@link ServletContext}. See
 * {@link #preContextDeployment(Module, Context)} for details.
 * 
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Immutable and therefore thread-safe.
 * 
 * @author Sam Brannen
 * @since 1.0
 */
public class ApplicationContextRetrievingAndPublishingCallback {

    /**
     * The suffix to be appended to the {@link ServletContext} attribute name for a published {@link ApplicationContext}:
     * &quot;<code>-ApplicationContext</code>&quot;.
     */
    public static final String APPLICATION_CONTEXT_SUFFIX = "-ApplicationContext";

    private static final LogService LOG_SERVICE = LogServiceFactory.getLogService();

    private static final Tracer TRACER = TracerFactory.createTracer(ApplicationContextRetrievingAndPublishingCallback.class);

    private final BundleStarter bundleStarter;

    /**
     * Creates a new <code>ApplicationContextRetrievingAndPublishingCallback</code> with the supplied
     * {@link BundleStarter}.
     * 
     * @param bundleStarter The <code>BundleStarter</code> to use to await the start of application contexts
     */
    public ApplicationContextRetrievingAndPublishingCallback(@NonNull BundleStarter bundleStarter) {
        this.bundleStarter = bundleStarter;
    }

    /**
     * Attempts to retrieve the supplied {@link Module module's} {@link ApplicationContext} from the OSGi service
     * registry and publish it via the {@link ServletContext} of the supplied {@link Context} under an attribute name
     * generated by concatenating the web bundle's <em>bundle symbolic name</em> with
     * {@link #APPLICATION_CONTEXT_SUFFIX}.
     * 
     * @param module the web module for which the application context should be retrieved and published; never
     *        <code>null</code>
     * @param context the web application context for which the application context should be retrieved and published;
     *        never <code>null</code>
     * @throws DeploymentException if the module is expected to publish its application context but fails to within the
     *         configured timeout period
     */
    public void preContextDeployment(@NonNull Module module, @NonNull Context context) throws DeploymentException {

        String contextPath = getDisplaySuitableContextPath(context.getPath());
        Bundle webBundle = module.getBundle();
        Assert.notNull(webBundle, "'webBundle' must not be null");

        Future<ApplicationContext> future = this.bundleStarter.awaitStart(webBundle);
        ApplicationContext applicationContext;
        try {
            applicationContext = future.get(module.getDeploymentTimeoutSeconds(), TimeUnit.SECONDS);
        } catch (InterruptedException e) {
            throw new DeploymentException("Interrupted while waiting for web bundle '" + webBundle.getSymbolicName() + "'", e);
        } catch (ExecutionException e) {
            LOG_SERVICE.log(WEB_BUNDLE_FAILED_TO_PUBLISH_APPLICATION_CONTEXT, contextPath, webBundle.getSymbolicName());
            throw new DeploymentException("Error starting web bundle '" + webBundle.getSymbolicName() + "'", e.getCause());
        } catch (TimeoutException e) {
            LOG_SERVICE.log(WEB_BUNDLE_FAILED_TO_PUBLISH_APPLICATION_CONTEXT_WITHIN_TIMEOUT, contextPath, webBundle.getSymbolicName(),
                module.getDeploymentTimeoutSeconds());
            throw new DeploymentException("Timeout while waiting for web bundle '" + webBundle.getSymbolicName() + "'", e);
        }

        String contextAttributeName = module.getUnscopedSymbolicName() + APPLICATION_CONTEXT_SUFFIX;
        TRACER.traceInfo("Publishing ApplicationContext for web bundle '%s' in the ServletContext under attribute '%s'.",
            webBundle.getSymbolicName(), contextAttributeName);
        context.getServletContext().setAttribute(contextAttributeName, applicationContext);
    }
}
