/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.web.core.internal.webxml;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * Object representation of <code>web.xml</code>'s <code>&lt;filter-mapping&gt;</code> element.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Immutable and therefore thread-safe.
 * 
 * @author Sam Brannen
 * @since 1.0
 * @see Filter
 */
class FilterMapping implements DeploymentDescriptorElement {

    protected static final String FILTER_MAPPING_ELEMENT = "filter-mapping";

    protected static final String FILTER_NAME_ELEMENT = "filter-name";

    protected static final String URL_PATTERN_ELEMENT = "url-pattern";

    protected static final String DISPATCHER_ELEMENT = "dispatcher";

    private final Filter filter;

    private final String urlPattern;

    private final List<String> dispatcherMappings = Collections.synchronizedList(new ArrayList<String>());

    /**
     * Creates a new <code>FilterMapping</code> for the supplied {@link Filter} and URL pattern.
     */
    public FilterMapping(Filter filter, String urlPattern) {
        this.filter = filter;
        this.urlPattern = urlPattern;
    }

    /**
     * {@inheritDoc}
     */
    public void appendToDocument(Document doc, Element root) {
        Element filterMappingElement = doc.createElement(FILTER_MAPPING_ELEMENT);
        WebXmlBuilder.appendTextElement(doc, filterMappingElement, FILTER_NAME_ELEMENT, getFilter().getName());
        WebXmlBuilder.appendTextElement(doc, filterMappingElement, URL_PATTERN_ELEMENT, getUrlPattern());
        for (String dispatcherMapping : this.dispatcherMappings) {
            WebXmlBuilder.appendTextElement(doc, filterMappingElement, DISPATCHER_ELEMENT, dispatcherMapping);
        }
        root.appendChild(filterMappingElement);
    }

    /**
     * Gets the {@link Filter} for this filter mapping.
     */
    public Filter getFilter() {
        return this.filter;
    }

    /**
     * Gets the URL pattern for this filter mapping.
     */
    public String getUrlPattern() {
        return this.urlPattern;
    }

    /**
     * Gets the list of dispatcher mappings for this filter mapping.
     */
    public List<String> getDispatcherMappings() {
        return this.dispatcherMappings;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result + (this.filter == null ? 0 : this.filter.hashCode());
        result = prime * result + (this.urlPattern == null ? 0 : this.urlPattern.hashCode());
        return result;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        FilterMapping other = (FilterMapping) obj;
        if (this.filter == null) {
            if (other.filter != null) {
                return false;
            }
        } else if (!this.filter.equals(other.filter)) {
            return false;
        }
        if (this.urlPattern == null) {
            if (other.urlPattern != null) {
                return false;
            }
        } else if (!this.urlPattern.equals(other.urlPattern)) {
            return false;
        }
        return true;
    }

}
