/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.web.core.internal.webxml;

import java.util.ArrayList;
import java.util.List;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * Object representation of <code>web.xml</code>'s <code>&lt;servlet&gt;</code> element.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Immutable and therefore thread-safe.
 * 
 * @author Sam Brannen
 * @since 1.0
 * @see DeploymentDescriptor
 * @see ServletMapping
 */
class Servlet implements DeploymentDescriptorElement {

    protected static final String SERVLET_ELEMENT = "servlet";

    protected static final String SERVLET_NAME_ELEMENT = "servlet-name";

    protected static final String SERVLET_CLASS_ELEMENT = "servlet-class";

    protected static final String LOAD_ON_STARTUP_ELEMENT = "load-on-startup";

    private final String name;

    private final String className;

    private final Integer loadOnStartup;

    private final List<InitParam> initParams = new ArrayList<InitParam>();

    private final List<ServletMapping> mappings = new ArrayList<ServletMapping>();

    /**
     * Creates a new <code>Servlet</code> element with the supplied name, class name, and <em>load on startup</em>
     * value.
     */
    public Servlet(String name, String className, Integer loadOnStartup) {
        this.name = name;
        this.className = className;
        this.loadOnStartup = loadOnStartup;
    }

    /**
     * {@inheritDoc}
     */
    public void appendToDocument(Document doc, Element root) {
        Element servletElement = doc.createElement(SERVLET_ELEMENT);
        WebXmlBuilder.appendTextElement(doc, servletElement, SERVLET_NAME_ELEMENT, getName());
        WebXmlBuilder.appendTextElement(doc, servletElement, SERVLET_CLASS_ELEMENT, getClassName());
        if (getLoadOnStartup() != null && getLoadOnStartup() >= 0) {
            WebXmlBuilder.appendTextElement(doc, servletElement, LOAD_ON_STARTUP_ELEMENT, getLoadOnStartup().toString());
        }
        for (InitParam initParam : getInitParams()) {
            initParam.appendToDocument(doc, servletElement);
        }
        root.appendChild(servletElement);

        for (ServletMapping mapping : getMappings()) {
            mapping.appendToDocument(doc, root);
        }
    }

    /**
     * Gets the name of this Servlet element.
     */
    public String getName() {
        return name;
    }

    /**
     * Gets the class name of this Servlet element.
     */
    public String getClassName() {
        return className;
    }

    /**
     * Gets the <em>load on startup</em> value of this Servlet element.
     */
    public Integer getLoadOnStartup() {
        return loadOnStartup;
    }

    /**
     * Gets the list of {@link InitParam init-params} for this Servlet element.
     */
    public List<InitParam> getInitParams() {
        return this.initParams;
    }

    /**
     * Adds an {@link InitParam} with the supplied name and value to this Servlet element.
     * 
     * @return the new <code>InitParam</code> as a convenience
     */
    public InitParam addInitParam(String name, String value) {
        InitParam initParam = new InitParam(name, value);
        getInitParams().add(initParam);
        return initParam;
    }

    /**
     * Adds a {@link ServletMapping} with the supplied URL pattern to this Servlet element.
     * 
     * @return the new <code>ServletMapping</code> as a convenience
     */
    public ServletMapping addMapping(String urlPattern) {
        ServletMapping servletMapping = new ServletMapping(this, urlPattern);
        getMappings().add(servletMapping);
        return servletMapping;
    }

    /**
     * Adds {@link ServletMapping servlet mappings} for each of the supplied URL patterns to this Servlet element.
     */
    public void addMappings(List<String> urlPatterns) {
        for (String urlPattern : urlPatterns) {
            getMappings().add(new ServletMapping(this, urlPattern));
        }
    }

    /**
     * Adds {@link ServletMapping servlet mappings} for each of the supplied URL patterns to this Servlet element.
     */
    public void addMappings(String... urlPatterns) {
        for (String urlPattern : urlPatterns) {
            getMappings().add(new ServletMapping(this, urlPattern));
        }
    }

    /**
     * Gets the list of {@link ServletMapping Servlet mappings} for this Servlet element.
     */
    public List<ServletMapping> getMappings() {
        return mappings;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result + ((this.name == null) ? 0 : this.name.hashCode());
        return result;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        Servlet other = (Servlet) obj;
        if (this.name == null) {
            if (other.name != null) {
                return false;
            }
        } else if (!this.name.equals(other.name)) {
            return false;
        }
        return true;
    }

}
