/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.web.core.internal.webxml;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * Object representation of <code>web.xml</code>'s <code>&lt;servlet-mapping&gt;</code> element.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Immutable and therefore thread-safe.
 * 
 * @author Sam Brannen
 * @since 1.0
 * @see Servlet
 */
class ServletMapping implements DeploymentDescriptorElement {

    protected static final String SERVLET_MAPPING_ELEMENT = "servlet-mapping";

    protected static final String SERVLET_NAME_ELEMENT = "servlet-name";

    protected static final String URL_PATTERN_ELEMENT = "url-pattern";

    private final Servlet servlet;

    private final String urlPattern;

    /**
     * Creates a new <code>ServletMapping</code> for the supplied {@link Servlet} and URL pattern.
     */
    public ServletMapping(Servlet servlet, String urlPattern) {
        this.servlet = servlet;
        this.urlPattern = urlPattern;
    }

    /**
     * {@inheritDoc}
     */
    public void appendToDocument(Document doc, Element root) {
        Element servletMappingElement = doc.createElement(SERVLET_MAPPING_ELEMENT);
        WebXmlBuilder.appendTextElement(doc, servletMappingElement, SERVLET_NAME_ELEMENT, getServlet().getName());
        WebXmlBuilder.appendTextElement(doc, servletMappingElement, URL_PATTERN_ELEMENT, getUrlPattern());
        root.appendChild(servletMappingElement);
    }

    /**
     * Gets the {@link Servlet} for this ServletMapping.
     */
    public Servlet getServlet() {
        return this.servlet;
    }

    /**
     * Gets the URL pattern for this ServletMapping.
     */
    public String getUrlPattern() {
        return this.urlPattern;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result + ((this.servlet == null) ? 0 : this.servlet.hashCode());
        result = prime * result + ((this.urlPattern == null) ? 0 : this.urlPattern.hashCode());
        return result;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        ServletMapping other = (ServletMapping) obj;
        if (this.servlet == null) {
            if (other.servlet != null) {
                return false;
            }
        } else if (!this.servlet.equals(other.servlet)) {
            return false;
        }
        if (this.urlPattern == null) {
            if (other.urlPattern != null) {
                return false;
            }
        } else if (!this.urlPattern.equals(other.urlPattern)) {
            return false;
        }
        return true;
    }

}
