/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.web.core.internal.webxml;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.springframework.util.StringUtils;
import org.w3c.dom.DOMImplementation;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;

import com.springsource.server.deployer.core.FatalDeploymentException;
import com.springsource.server.deployer.core.Module;
import com.springsource.server.deployer.core.DeploymentArtefact.Entry;
import com.springsource.server.osgi.manifest.BundleManifest;
import com.springsource.server.serviceability.ffdc.NonNull;
import com.springsource.server.serviceability.output.Level;
import com.springsource.server.serviceability.tracing.EntryExitTraceLevel;

/**
 * Builder for generating a Java EE web application deployment descriptor (i.e., &quot;<code>web.xml</code>&quot;)
 * for a given {@link Module web module}. Various configuration elements will either be inferred or generated from
 * explicit metadata configured in the module. If an existing &quot;<code>/WEB-INF/web.xml</code>&quot; is present,
 * its contents will be merged with the generated deployment descriptor.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Thread-safe
 * 
 * @author Sam Brannen
 * @since 1.0
 * @see StandardWebModuleDeploymentDescriptorBuilder
 * @see DeploymentDescriptor
 */
public class WebXmlBuilder {

    protected static final String MODULE_INF_WEB_INF_WEB_XML = "/MODULE-INF/WEB-INF/web.xml";

    protected static final String SERVLET_SPEC_NAMESPACE_URI = "http://java.sun.com/xml/ns/javaee";

    protected static final String SERVLET_SPEC_ROOT_ELEMENT = "web-app";

    protected static final String SERVLET_SPEC_SCHEMA_LOCATION = "http://java.sun.com/xml/ns/javaee http://java.sun.com/xml/ns/javaee/web-app_2_5.xsd";

    protected static final String SERVLET_SPEC_VERSION = "2.5";

    private final WebModuleDeploymentDescriptorBuilder webModuleDeploymentDescriptorBuilder;

    /**
     * Constructs a new <code>WebXmlBuilder</code> using the standard
     * {@link StandardWebModuleDeploymentDescriptorBuilder}.
     */
    public WebXmlBuilder() {
        this(new StandardWebModuleDeploymentDescriptorBuilder());
    }

    /**
     * Constructs a new <code>WebXmlBuilder</code> using the supplied
     * {@link StandardWebModuleDeploymentDescriptorBuilder}.
     */
    public WebXmlBuilder(WebModuleDeploymentDescriptorBuilder webModuleDeploymentDescriptorBuilder) {
        this.webModuleDeploymentDescriptorBuilder = webModuleDeploymentDescriptorBuilder;
    }

    /**
     * Creates a {@link DocumentBuilder}.
     * 
     * @return the <code>DocumentBuilder</code>.
     */
    @EntryExitTraceLevel(Level.DEBUG)
    private static DocumentBuilder createDocumentBuilder() {
        try {
            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            dbf.setNamespaceAware(true);
            return dbf.newDocumentBuilder();
        } catch (ParserConfigurationException e) {
            throw new FatalDeploymentException("Unable to create DocumentBuilder - JAXP parser configuration error.", e);
        }
    }

    /**
     * Creates and appends a new text {@link Element element} to the supplied {@link Element parent}.
     * 
     * @param doc the document in which to create the new text element
     * @param parent the parent element to which the new text element should be appended
     * @param name the name of the text element
     * @param value the textual value text element
     */
    @EntryExitTraceLevel(Level.VERBOSE)
    static void appendTextElement(Document doc, Element parent, String name, String value) {
        Element child = doc.createElement(name);
        child.setTextContent(value);
        parent.appendChild(child);
    }

    @EntryExitTraceLevel(Level.DEBUG)
    protected String retrieveModuleName(@NonNull Module module) {
        String name = null;
        try {
            BundleManifest bundleManifest = module.getModuleData().getBundleManifest();
            name = bundleManifest.getBundleName();
            if (!StringUtils.hasText(name)) {
                name = bundleManifest.getBundleSymbolicName().getSymbolicName();
            }
        } catch (Exception e) {
            /* ignore */
        }
        return StringUtils.hasText(name) ? name.trim() : null;
    }

    /**
     * Generates <code>web.xml</code> for the supplied {@link Module web module}.
     */
    public void generateWebXml(final Module module) {

        DeploymentDescriptor deploymentDescriptor = this.webModuleDeploymentDescriptorBuilder.buildDeploymentDescriptor(module);

        try {
            Entry webXmlEntry = module.getModuleData().getEntry(MODULE_INF_WEB_INF_WEB_XML);
            boolean mergeRequired = webXmlEntry.exists();
            DocumentBuilder documentBuilder = createDocumentBuilder();
            DOMImplementation impl = documentBuilder.getDOMImplementation();

            Document outputDoc = impl.createDocument(SERVLET_SPEC_NAMESPACE_URI, SERVLET_SPEC_ROOT_ELEMENT, null);
            Element root = outputDoc.getDocumentElement();
            setRootAttributes(root);
            deploymentDescriptor.writeToDocument(root, outputDoc);

            if (mergeRequired) {
                mergeWebXmlDocuments(documentBuilder.parse(new InputSource(webXmlEntry.getReader())), outputDoc);
            }

            DOMSource domSource = new DOMSource(outputDoc);
            StreamResult streamResult = new StreamResult(webXmlEntry.getWriter(false));
            Transformer transformer = TransformerFactory.newInstance().newTransformer();
            transformer.setOutputProperty(OutputKeys.ENCODING, "ISO-8859-1");
            transformer.setOutputProperty(OutputKeys.INDENT, "yes");
            transformer.transform(domSource, streamResult);
        } catch (Exception e) {
            throw new FatalDeploymentException("Failed to generate web.xml for web module '" + retrieveModuleName(module) + "'.", e);
        }
    }

    /**
     * Sets the attributes of the root element (i.e., &quot;web-app&quot;). The default implementation uses attributes
     * appropriate for version 2.5 of the Servlet Specification.
     * 
     * @param root
     */
    @EntryExitTraceLevel(Level.VERBOSE)
    protected void setRootAttributes(Element root) {
        root.setAttributeNS("http://www.w3.org/2000/xmlns/", "xmlns:xsi", "http://www.w3.org/1999/XMLSchema-instance");
        root.setAttribute("xsi:schemaLocation", SERVLET_SPEC_SCHEMA_LOCATION);
        root.setAttribute("version", SERVLET_SPEC_VERSION);
    }

    /**
     * Merges {@link Node nodes} from the supplied <code>inputDoc</code> into the supplied <code>outputDoc</code>.
     * 
     * @param inputDoc the input document
     * @param outputDoc the output document
     */
    protected void mergeWebXmlDocuments(Document inputDoc, Document outputDoc) {
        Element inputRoot = inputDoc.getDocumentElement();
        Element outputRoot = outputDoc.getDocumentElement();
        NodeList childNodes = inputRoot.getChildNodes();
        for (int i = 0; i < childNodes.getLength(); i++) {
            Node inputNode = childNodes.item(i);
            if (inputNode instanceof Element) {
                Node node = outputDoc.importNode(inputNode, true);
                outputRoot.appendChild(node);
            }
        }
    }

}
