/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.web.dm;

import static org.springframework.web.context.WebApplicationContext.ROOT_WEB_APPLICATION_CONTEXT_ATTRIBUTE;

import javax.servlet.ServletContext;

import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.springframework.osgi.context.DelegatedExecutionOsgiBundleApplicationContext;
import org.springframework.osgi.extender.OsgiApplicationContextCreator;
import org.springframework.osgi.extender.support.ApplicationContextConfiguration;
import org.springframework.osgi.util.OsgiStringUtils;
import org.springframework.util.Assert;
import org.springframework.util.StringUtils;

import com.springsource.server.deployer.core.Module;
import com.springsource.server.deployer.core.ModuleDeployer;
import com.springsource.server.deployer.core.ModuleRegistry;
import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;
import com.springsource.server.servlet.core.DynamicContext;

/**
 * OsgiWebApplicationContextCreator is a custom implementation of {@link OsgiApplicationContextCreator} which creates
 * {@link ServerOsgiBundleXmlWebApplicationContext}s for <em>web modules</em>. <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Immutable and therefore thread-safe.
 * 
 * @author Sam Brannen
 * @since 1.0
 */
public final class OsgiWebApplicationContextCreator implements OsgiApplicationContextCreator {

    private static final String WEB_PERSONALITY = "Web";

    private static final String DYNAMIC_CONTEXT_ATTRIBUTE = "com.springsource.server.web.core.WebModulePersonality.dynamicContext";

    private static final Tracer TRACER = TracerFactory.createTracer(OsgiWebApplicationContextCreator.class);

    private final ModuleRegistry moduleRegistry;

    /**
     * Constructs a new <code>OsgiWebApplicationContextCreator</code> with the supplied {@link ModuleRegistry}.
     */
    public OsgiWebApplicationContextCreator(ModuleRegistry moduleRegistry) {
        this.moduleRegistry = moduleRegistry;
    }

    /**
     * Creates a {@link ServerOsgiBundleXmlWebApplicationContext} for the supplied {@link BundleContext bundle context}
     * if it represents a <em>web module</em>.
     * 
     * @param bundleContext OSGi bundle context for which a web application context should be created; never
     *        <code>null</code>
     * @return <code>null</code> if no context should be created, non-<code>null</code> otherwise
     * @throws Exception if something goes wrong
     */
    public DelegatedExecutionOsgiBundleApplicationContext createApplicationContext(BundleContext bundleContext) throws Exception {

        Assert.notNull(bundleContext, "Can not create an application context for a null 'bundleContext'.");

        ServerOsgiBundleXmlWebApplicationContext applicationContext = null;
        Bundle bundle = bundleContext.getBundle();
        String symbolicName = OsgiStringUtils.nullSafeNameAndSymName(bundle);
        TRACER.traceVerbose("Considering creating a ServerOsgiBundleXmlWebApplicationContext for bundle [%s].", symbolicName);

        final String moduleType = (String) bundle.getHeaders().get(ModuleDeployer.MODULE_TYPE_MANIFEST_HEADER);

        if (StringUtils.hasText(moduleType) && WEB_PERSONALITY.equalsIgnoreCase(moduleType.trim())) {
            TRACER.traceInfo("Creating a ServerOsgiBundleXmlWebApplicationContext for web module bundle [%s].", symbolicName);

            Module module = this.moduleRegistry.getModule(bundle);
            DynamicContext context = (DynamicContext) module.getModuleData().getAttribute(DYNAMIC_CONTEXT_ATTRIBUTE);
            ServletContext servletContext = context.getServletContext();

            if (servletContext.getAttribute(ROOT_WEB_APPLICATION_CONTEXT_ATTRIBUTE) != null) {
                String msg = "Cannot create ServerOsgiBundleXmlWebApplicationContext for web module bundle [" + symbolicName
                    + "] because there is already an application context present in the ServletContext under attribute name ["
                    + ROOT_WEB_APPLICATION_CONTEXT_ATTRIBUTE + "]";
                TRACER.traceError(msg);
                throw new IllegalStateException(msg);
            }

            ApplicationContextConfiguration config = new ApplicationContextConfiguration(bundle);
            TRACER.traceInfo("Created ApplicationContextConfiguration [%s] for bundle [%s].", config, symbolicName);

            try {
                applicationContext = new ServerOsgiBundleXmlWebApplicationContext(config.getConfigurationLocations());
                applicationContext.setBundleContext(bundleContext);
                applicationContext.setServletContext(servletContext);
                applicationContext.setPublishContextAsService(config.isPublishContextAsService());
                servletContext.setAttribute(ROOT_WEB_APPLICATION_CONTEXT_ATTRIBUTE, applicationContext);
            } catch (RuntimeException ex) {
                servletContext.setAttribute(ROOT_WEB_APPLICATION_CONTEXT_ATTRIBUTE, ex);
                TRACER.traceError("ServerOsgiBundleXmlWebApplicationContext initialization for web module bundle [" + symbolicName + "] failed", ex);
                throw ex;
            } catch (Error err) {
                servletContext.setAttribute(ROOT_WEB_APPLICATION_CONTEXT_ATTRIBUTE, err);
                TRACER.traceError("ServerOsgiBundleXmlWebApplicationContext initialization for web module bundle [" + symbolicName + "] failed", err);
                throw err;
            }

        }

        return applicationContext;
    }

}
