/* 
 * This file is part of the SpringSource dm Server.
 * 
 * Copyright (C) 2008 SpringSource Inc.
 * 
 * The SpringSource dm Server is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * The SpringSource dm Server is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * the SpringSource dm Server. If not, see <http://www.gnu.org/licenses/>.
 */

package com.springsource.server.web.dm;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.URL;

import org.osgi.framework.Bundle;
import org.springframework.core.io.Resource;
import org.springframework.osgi.io.OsgiBundleResource;
import org.springframework.util.StringUtils;

import com.springsource.server.serviceability.tracing.Tracer;
import com.springsource.server.serviceability.tracing.TracerFactory;

/**
 * <code>ServerOsgiBundleResource</code> is a custom extension of {@link OsgiBundleResource}. See {@link #getFile()}
 * and {@link #getURL()}. <p />
 * 
 * @author Sam Brannen
 * @since 1.0
 */
public class ServerOsgiBundleResource extends OsgiBundleResource {

    private static final String MODULE_INF = "MODULE-INF/";

    private static final char PREFIX_SEPARATOR = ':';

    private static final Tracer TRACER = TracerFactory.createTracer(ServerOsgiBundleResource.class);

    private final OsgiBundleResource moduleInfOsgiBundleResource;

    private final Bundle bundle;

    private final String path;

    /**
     * Constructs a new <code>ServerOsgiBundleResource</code>.
     * 
     * @param bundle OSGi bundle used by this resource
     * @param path resource path inside the bundle
     */
    public ServerOsgiBundleResource(Bundle bundle, String path) {
        super(bundle, path);
        this.bundle = bundle;
        this.path = path;
        String prefix = parsePrefix(StringUtils.cleanPath(path));
        this.moduleInfOsgiBundleResource = new OsgiBundleResource(bundle, (StringUtils.hasText(prefix) ? prefix : "") + MODULE_INF
            + getPathWithinContext());
    }

    /**
     * Parses a prefix from the supplied path and returns it. A prefix is any string ending with a &quot;<code>:</code>
     * &quot;.
     * 
     * @return the prefix, or <code>null</code> if not present
     */
    private static String parsePrefix(String path) {
        int indexOfPrefix = path.indexOf(PREFIX_SEPARATOR);
        return indexOfPrefix == -1 ? null : path.substring(0, indexOfPrefix + 1);
    }

    /**
     * Delegates to the parent implementation for default behavior; however, if the parent implementation is
     * unsuccessful in locating the file and if the path associated with this resource is determined to be a bundle
     * resource located in the &quot;<code>MODULE-INF/</code>&quot; directory of the bundle and exists in the local file
     * system, an attempt will be made to locate the file in the unpacked bundle and return a reference to that file.
     * 
     * @see OsgiBundleResource#getFile()
     */
    @Override
    public File getFile() throws IOException {

        // 1) try default behavior.
        IOException originalException = null;
        try {
            return super.getFile();
        } catch (IOException ioe) {
            originalException = ioe;
        }

        // 2) search under MODULE-INF/...
        try {
            return this.moduleInfOsgiBundleResource.getFile();
        } catch (IOException ioe) {
            // Log, but essentially ignore
            TRACER.traceInfo(getDescription() + "[%s] cannot be resolved to an absolute file path.",
                this.moduleInfOsgiBundleResource.getDescription());
        }

        throw originalException != null ? originalException : new FileNotFoundException(getDescription()
            + " cannot be resolved to absolute file path");
    }

    /**
     * Delegates to the parent implementation for default behavior; however, if the parent implementation is
     * unsuccessful at retrieving a valid URL, an attempt will be made to retrieve a URL by prepending this resource's
     * path within the &quot;<code>MODULE-INF/</code>&quot; directory in the root of the bundle.
     * 
     * @see OsgiBundleResource#getURL()
     */
    @Override
    public URL getURL() throws IOException {

        // 1) try default behavior.
        IOException originalException = null;
        try {
            return super.getURL();
        } catch (IOException ioe) {
            originalException = ioe;
        }

        // 2) search under MODULE-INF/...
        try {
            return this.moduleInfOsgiBundleResource.getURL();
        } catch (IOException ioe) {
            // Log, but essentially ignore
            TRACER.traceInfo(getDescription() + "[%s] cannot be resolved to a URL because it does not exist.",
                this.moduleInfOsgiBundleResource.getDescription());
        }

        throw originalException != null ? originalException : new FileNotFoundException(getDescription()
            + " cannot be resolved to URL because it does not exist");
    }

    /**
     * Overrides {@link OsgiBundleResource#createRelative(String)} by returning a new {@link ServerOsgiBundleResource}
     * instead of an {@link OsgiBundleResource} but otherwise maintains the same behavior.
     */
    @Override
    public Resource createRelative(String relativePath) {
        String pathToUse = StringUtils.applyRelativePath(this.path, relativePath);
        return new ServerOsgiBundleResource(this.bundle, pathToUse);
    }

}
